/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ AiOpbg ]
 */

#if XM7_VER >= 2

#include <string.h>
#include "xm7.h"
#include "device.h"
#include "apalet.h"
#include "display.h"
#include "subctrl.h"

/*
 *	O[o [N
 */
WORD apalet_no;							/* Ipbgԍ */
BYTE apalet_b[4096];					/* Bx(0-15) */
BYTE apalet_r[4096];					/* Rx(0-15) */
BYTE apalet_g[4096];					/* Gx(0-15) */

/*
 *	ID
 */
#define	ID	MAKEID('A', 'P', 'A', 'L')
//#define APALET_LOG

/*
 *	AiOpbg
 *	
 */
BOOL FASTCALL apalet_init(void)
{
	memset(apalet_b, 0, sizeof(apalet_b));
	memset(apalet_r, 0, sizeof(apalet_r));
	memset(apalet_g, 0, sizeof(apalet_g));

	return TRUE;
}

/*
 *	AiOpbg
 *	N[Abv
 */
void FASTCALL apalet_cleanup(void)
{
}

/*
 *	AiOpbg
 *	Zbg
 */
void FASTCALL apalet_reset(void)
{
	LOG0(Normal, "Zbg");

	apalet_no = 0;
}

/*-[ }bvhI/O ]----------------------------------------------------*/

/*
 *	AiOpbg
 *	PoCgǂݏo
 */
BOOL FASTCALL apalet_readb(WORD addr, BYTE *dat)
{
	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}

	switch (addr) {
		/* pbgԍ(ǂݏołȂ) */
		case 0xfd30:
		case 0xfd31:
			*dat = 0xff;
			return TRUE;

		/* Bx */
		case 0xfd32:
			if (model_type <= MODEL_FM77AV) {
				/* FM77AV͓ǂݏołȂ */
				*dat = 0xff;
				return TRUE;
			}
			/* FM77AV40/20ȍ~RGBx̂ݓǂݏo\ */
			*dat = (BYTE)(0xf0 | (apalet_b[apalet_no] & 0x0f));
			return TRUE;

		/* Rx */
		case 0xfd33:
			if (model_type <= MODEL_FM77AV) {
				/* FM77AV͓ǂݏołȂ */
				*dat = 0xff;
				return TRUE;
			}
			/* FM77AV40/20ȍ~RGBx̂ݓǂݏo\ */
			*dat = (BYTE)(0xf0 | (apalet_r[apalet_no] & 0x0f));
			return TRUE;

		/* Gx */
		case 0xfd34:
			if (model_type <= MODEL_FM77AV) {
				/* FM77AV͓ǂݏołȂ */
				*dat = 0xff;
				return TRUE;
			}
			/* FM77AV40/20ȍ~RGBx̂ݓǂݏo\ */
			*dat = (BYTE)(0xf0 | (apalet_g[apalet_no] & 0x0f));
			return TRUE;
	}

	return FALSE;
}

/*
 *	AiOpbg
 *	PoCg
 */
BOOL FASTCALL apalet_writeb(WORD addr, BYTE dat)
{
	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}

	switch (addr) {
		/* pbgԍ */
		case 0xfd30:
			apalet_no &= (WORD)0xff;
			apalet_no |= (WORD)((dat & 0x0f) << 8);
			return TRUE;

		/* pbgԍ */
		case 0xfd31:
			apalet_no &= (WORD)(0xf00);
			apalet_no |= (WORD)(dat);
			return TRUE;

		/* Bx */
		case 0xfd32:
			if (apalet_b[apalet_no] != (dat & 0x0f)) {
				apalet_b[apalet_no] = (BYTE)(dat & 0x0f);
				/* ʒm */
				apalet_notify();
			}
			return TRUE;

		/* Rx */
		case 0xfd33:
			if (apalet_r[apalet_no] != (dat & 0x0f)) {
				apalet_r[apalet_no] = (BYTE)(dat & 0x0f);
				/* ʒm */
				apalet_notify();
			}
			return TRUE;

		/* Gx */
		case 0xfd34:
			if (apalet_g[apalet_no] != (dat & 0x0f)) {
				apalet_g[apalet_no] = (BYTE)(dat & 0x0f);
				/* ʒm */
				apalet_notify();
			}
			return TRUE;
	}

	return FALSE;
}

/*-[ t@CI/O ]----------------------------------------------------------*/

/*
 *	AiOpbg
 *	Z[u
 */
BOOL FASTCALL apalet_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_word_write(fileh, apalet_no)) {
		return FALSE;
	}
	if (!file_write(fileh, apalet_b, sizeof(apalet_b))) {
		return FALSE;
	}
	if (!file_write(fileh, apalet_r, sizeof(apalet_r))) {
		return FALSE;
	}
	if (!file_write(fileh, apalet_g, sizeof(apalet_g))) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	AiOpbg
 *	[h
 */
BOOL FASTCALL apalet_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_word_read(fileh, &apalet_no)) {
		return FALSE;
	}
	if (!file_read(fileh, apalet_b, sizeof(apalet_b))) {
		return FALSE;
	}
	if (!file_read(fileh, apalet_r, sizeof(apalet_r))) {
		return FALSE;
	}
	if (!file_read(fileh, apalet_g, sizeof(apalet_g))) {
		return FALSE;
	}

	return TRUE;
}

#endif	/* XM7_VER >= 2 */
