/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ DMAC(HD6844) AV40dlΉ ]
 */

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "xm7.h"
#include "device.h"
#include "mainetc.h"
#include "dmac.h"
#include "fdc.h"
#include "mmr.h"
#include "event.h"

/*
 *	O[o [N
 */
WORD dma_adr[4];						/* AhXWX^ */
WORD dma_bcr[4];						/* ]ꐔWX^ */
BYTE dma_chcr[4];						/* `l䃌WX^ */
BYTE dma_pcr;							/* D搧䃌WX^ */
BYTE dma_icr;							/* 荞ݐ䃌WX^ */
BYTE dma_dcr;							/* f[^`FC䃌WX^ */

#if XM7_VER >= 2
BYTE dma_reg;							/* ݑIĂ郌WX^ԍ */
#endif
BOOL dma_flag;							/* DMA]stO */
BOOL dma_burst_transfer;				/* DMAo[Xg]stO */

#if XM7_VER >= 2
const int dma_ch = 0;					/* DMA`l */
const BYTE dma_txrq = 0x01;				/* DMA TxRQl */
#elif XM7_VER == 1 && defined(SFDC)
const int dma_ch = 3;					/* DMA`l */
const BYTE dma_txrq = 0x08;				/* DMA TxRQl */
#endif

/*
 *	vg^Cv錾
 */
#if XM7_VER >= 2
static BOOL FASTCALL dmac_start_ch0(void);	/* DMA]JnCxg */
#else
static BOOL FASTCALL dmac_start_ch3(void);	/* DMA]JnCxg */
#endif

/*
 *	ID
 */
#define	ID	MAKEID('D', 'M', 'A', 'C')
//#define DMAC_LOG

/*
 *	DMAC
 *	
 */
BOOL FASTCALL dmac_init(void)
{
	return TRUE;
}

/*
 *	DMAC
 *	N[Abv
 */
void FASTCALL dmac_cleanup(void)
{
}

/*
 *	DMAC
 *	Zbg
 */
void FASTCALL dmac_reset(void)
{
	int i;

	LOG0(Normal, "Zbg");

	/* DMACWX^ */
	for (i=0; i<4; i++) {
		dma_adr[i] = 0xFFFF;
		dma_bcr[i] = 0xFFFF;
		dma_chcr[i] = 0x00;
	}
	dma_pcr = 0x00;
	dma_icr = 0x00;
	dma_dcr = 0x00;

	/* [N */
#if XM7_VER >= 2
	dma_reg = 0x00;
#endif
	dma_flag = FALSE;
	dma_burst_transfer = FALSE;
}


/*
 *	DMAC
 *	]Jn
 */
void FASTCALL dmac_start(void)
{
	/* f`FbN */
#if XM7_VER >= 2
	if (model_type <= MODEL_FM77AV20) {
		/* AV40/AV20EX/AV40EX̂ݗL */
		return;
	}
#else
	if (model_type != MODEL_FM8) {
		return;
	}
#endif

	/* DMA]쒆łΉȂ */
	if (dma_flag) {
		return;
	}

	/* ]꒷WX^(BCR)`FbN */
	if (dma_bcr[dma_ch] == 0) {
		return;
	}

	/* DMAgptO(TxRQ)`FbN */
	if (!(dma_pcr & dma_txrq)) {
		return;
	}

	/* DMA]Jn Cxgݒ */
#if XM7_VER >= 2
	schedule_setevent(EVENT_FDC_DMA, 50, dmac_start_ch0);
#else
	schedule_setevent(EVENT_FDC_DMA, 50, dmac_start_ch3);
#endif
}

/*
 *	DMAC Ch.0(FDC)
 *	]JnCxg
 */
#if XM7_VER >= 2
static BOOL FASTCALL dmac_start_ch0(void)
#else
static BOOL FASTCALL dmac_start_ch3(void)
#endif
{
	/* Cxg폜 */
	schedule_delevent(EVENT_FDC_DMA);

	/* DMAX^[g */
	dma_flag = TRUE;
	dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x40);

	return TRUE;
}

/*
 *	DMAC Ch.0(FDC)
 *	DMA]
 */
void FASTCALL dmac_exec(void)
{
	BYTE dat;
#if XM7_VER >= 2
	BYTE seg;
#endif

	/* TxRQ`FbN */
	if (!(dma_pcr & dma_txrq)) {
		return;
	}

	/* ̎_BCR0Ȃ]/荞ݔɏ𒆎~ */
	if (dma_bcr[dma_ch] == 0) {
		dma_flag = FALSE;
		dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x80);
		return;
	}

	/* FDCDRQ󂯁A] */
#if XM7_VER == 1
	fdc_mode = 1;
#endif
	fdc_readb(0xfd1f, &dat);
#if XM7_VER == 1
	fdc_mode = 0;
#endif
	if (!(dat & 0x80)) {
		if (dma_burst_transfer) {
			/* XPW[O̊֌WACCPUNbNi߂ */
			maincpu.total += (WORD)2;
		}
		return;
	}

	/* o[Xg[h̃`FbNȂ */
	if ((dma_chcr[dma_ch] & 0x02) && !dma_burst_transfer) {
		dma_burst_transfer = TRUE;
	}

#if XM7_VER >= 2
	/* DMA]sBMMR͏ɃZOg0ɂȂ */
	seg = mmr_seg;
	mmr_seg = 0;
#endif

	if (dma_chcr[dma_ch] & 0x01) {
		/* 1:FDC (Write) */
		dat = mainmem_readb(dma_adr[dma_ch]);
#if XM7_VER >= 2
		fdc_writeb(0xfd1b, dat);
#else
		ifdc_writeb(0xfd32, dat);
#endif
	}
	else {
		/* 0:FDC (Read) */
#if XM7_VER >= 2
		fdc_readb(0xfd1b, &dat);
#else
		ifdc_readb(0xfd32, &dat);
#endif
		mainmem_writeb(dma_adr[dma_ch], dat);
	}

#if XM7_VER >= 2
	/* MMRZOgA */
	mmr_seg = seg;
#endif

	/* AhXXV */
	if (dma_chcr[dma_ch] & 0x08) {
		/* AhXDOWN */
		dma_adr[dma_ch]--;
	}
	else {
		/* AhXUP */
		dma_adr[dma_ch]++;
	}

	/* CCPUNbN̈΂s */
	maincpu.total += (WORD)3;

	/* ]TCYXV */
	dma_bcr[dma_ch]--;

	/* ]I`FbN */
	if (dma_bcr[dma_ch] == 0) {
#if XM7_VER >= 3
		if ((model_type == MODEL_FM77AV40) &&
			(dma_dcr & 0x07) == 0x01) {
			/* f[^`FC */
			dma_adr[dma_ch] = dma_adr[3];
			dma_bcr[dma_ch] = dma_bcr[3];
			dma_bcr[3] = 0;
		}
		else {
			/* DMAI */
			dma_flag = FALSE;
			dma_burst_transfer = FALSE;
			dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x80);
			dma_dcr |= 0x10;

			/* 荞݂Cl[uȂA荞݂ */
			if (dma_icr & dma_txrq) {
				dma_icr |= 0x80;
				dma_irq_flag = TRUE;
				maincpu_irq();
			}
			return;
		}
		return;
#elif XM7_VER == 1 && defined(SFDC)
 #if 0
		if ((dma_dcr & 0x07) == 0x01) {
			/* f[^`FC */
			dma_adr[0] = dma_adr[3];
			dma_bcr[0] = dma_bcr[3];
			dma_bcr[3] = 0;
		}
		else if ((dma_dcr & 0x07) == 0x03) {
			/* f[^`FC */
			dma_adr[1] = dma_adr[3];
			dma_bcr[1] = dma_bcr[3];
			dma_bcr[3] = 0;
		}
		else if ((dma_dcr & 0x07) == 0x05) {
			/* f[^`FC */
			dma_adr[2] = dma_adr[3];
			dma_bcr[2] = dma_bcr[3];
			dma_bcr[3] = 0;
		}
		else {
			/* DMAI */
			dma_flag = FALSE;
			dma_burst_transfer = FALSE;
			dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x80);
			dma_dcr |= 0x10;

			/* 荞݂Cl[uȂA荞݂ */
			if (dma_icr & dma_txrq) {
				dma_icr |= 0x80;
				dma_irq_flag = TRUE;
				maincpu_irq();
			}
		}
 #else
		/* DMAI */
		dma_flag = FALSE;
		dma_burst_transfer = FALSE;
		dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x80);
		dma_dcr |= 0x10;

		/* 荞݂Cl[uȂA荞݂ */
		if (dma_icr & dma_txrq) {
			dma_icr |= 0x80;
			dma_irq_flag = TRUE;
			maincpu_irq();
		}
 #endif
		return;
#else
		/* DMAI */
		dma_flag = FALSE;
		dma_burst_transfer = FALSE;
		dma_chcr[dma_ch] = (BYTE)((dma_chcr[dma_ch] & 0x0f) | 0x80);
		dma_dcr |= 0x10;

		/* 荞݂Cl[uȂA荞݂ */
		if (dma_icr & dma_txrq) {
			dma_icr |= 0x80;
			dma_irq_flag = TRUE;
			maincpu_irq();
		}
		return;
#endif
	}
}

#if XM7_VER >= 2
/*
 *	DMAC
 *	WX^ǂݍ
 */
static BYTE FASTCALL dmac_readreg(WORD addr)
{
	BYTE	tmp;

	switch (addr) {
#if XM7_VER >= 3
		/* AhXWX^() */
		case 0x04:
		case 0x08:
		case 0x0c:
			if (model_type != MODEL_FM77AV40)
				break;
		case 0x00:
			return (BYTE)((dma_adr[addr>>2] >> 8) & 0xff);

		/* AhXWX^() */
		case 0x05:
		case 0x09:
		case 0x0d:
			if (model_type != MODEL_FM77AV40)
				break;
		case 0x01:
			return (BYTE)(dma_adr[addr>>2] & 0xff);

		/* ]ꐔWX^() */
		case 0x06:
		case 0x0a:
		case 0x0e:
			if (model_type != MODEL_FM77AV40)
				break;
		case 0x02:
			return (BYTE)((dma_bcr[addr>>2] >> 8) & 0xff);

		/* ]ꐔWX^() */
		case 0x07:
		case 0x0b:
		case 0x0f:
			if (model_type != MODEL_FM77AV40)
				break;
		case 0x03:
			return (BYTE)(dma_bcr[addr>>2] & 0xff);

		/* `l䃌WX^ */
		case 0x11:
		case 0x12:
		case 0x13:
			if (model_type != MODEL_FM77AV40)
				break;
		case 0x10:
			tmp = dma_chcr[addr-0x10];
			dma_chcr[addr-0x10] &= 0x7f;
			return tmp;
#else
		/* AhXWX^() */
		case 0x00:
			return (BYTE)((dma_adr[0] >> 8) & 0xff);

		/* AhXWX^() */
		case 0x01:
			return (BYTE)(dma_adr[0] & 0xff);

		/* ]ꐔWX^() */
		case 0x02:
			return (BYTE)((dma_bcr[0] >> 8) & 0xff);

		/* ]ꐔWX^() */
		case 0x03:
			return (BYTE)(dma_bcr[0] & 0xff);

		/* `l䃌WX^ */
		case 0x10:
			tmp = dma_chcr[0];
			dma_chcr[0] &= 0x7f;
			return tmp;
#endif

		/* D搧䃌WX^ */
		case 0x14:
			return dma_pcr;

		/* 荞ݐ䃌WX^ */
		case 0x15:
			tmp = (BYTE)(((dma_dcr >> 4) | 0x80) & dma_icr);
			dma_dcr &= 0x0f;

			/* IRQ */
			dma_icr &= (BYTE)~0x80;
			dma_irq_flag = FALSE;
			maincpu_irq();
			return tmp;

		/* f[^`FC䃌WX^ */
		case 0x16:
			return (BYTE)(dma_dcr & 0x0f);
	}

	/* ̑̃WX^̓fR[hĂȂ */
	return 0x00;
}

/*
 *	DMAC
 *	WX^
 */
static void FASTCALL dmac_writereg(WORD addr, BYTE dat)
{
	switch (addr) {
		/* AhXWX^() */
		case 0x00:
		case 0x04:
		case 0x08:
		case 0x0c:
			dma_adr[addr>>2] = (WORD)((dma_adr[addr>>2] & 0xff) | (dat << 8));
			return;

		/* AhXWX^() */
		case 0x01:
		case 0x05:
		case 0x09:
		case 0x0d:
			dma_adr[addr>>2] = (WORD)((dma_adr[addr>>2] & 0xff00) | dat);
			return;

		/* ]ꐔWX^() */
		case 0x02:
		case 0x06:
		case 0x0a:
		case 0x0e:
			dma_bcr[addr>>2] = (WORD)((dma_bcr[addr>>2] & 0xff) | (dat << 8));
			return;

		/* ]ꐔWX^() */
		case 0x03:
		case 0x07:
		case 0x0b:
		case 0x0f:
			dma_bcr[addr>>2] = (WORD)((dma_bcr[addr>>2] & 0xff00) | dat);
			return;

		/* `l䃌WX^ */
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
			dma_chcr[addr-0x10] = (BYTE)((dma_chcr[addr-0x10] & 0xc0) | (dat & 0x0f));
			return;

		/* D搧䃌WX^ */
		case 0x14:
			dma_pcr = (BYTE)(dat & 0x8f);
			return;

		/* 荞ݐ䃌WX^ */
		case 0x15:
			dma_icr = (BYTE)((dma_icr & 0x80) | (dat & 0x0f));
			return;

		/* f[^`FC䃌WX^ */
		case 0x16:
			dma_dcr = (BYTE)((dma_dcr & 0xf0) | (dat & 0x0f));
			return;
	}
}
#endif

/*
 *	DMAC
 *	PoCg擾
 */
BOOL FASTCALL dmac_readb(WORD addr,BYTE *dat)
{
	/* f`FbN */
#if XM7_VER >= 2
	if (model_type <= MODEL_FM77AV20) {
		/* AV40/AV20EX/AV40EX̂ݗL */
		return FALSE;
	}
#else
	if (model_type != MODEL_FM8) {
		return FALSE;
	}

	/* FDCڑȂ΃VXegjbgڑƌȂ */
#ifdef SFDC
	if (!fdc_enable[FDC_M] &&
		!fdc_enable[FDC_S]) {
#else
	if (!fdc_enable) {
#endif
		return FALSE;
	}
#endif

	switch (addr) {
#if XM7_VER >= 2
		/* DMACAhX */
		case 0xfd98:
			*dat = dma_reg;
			return TRUE;

		/* DMACf[^ */
		case 0xfd99:
			*dat = dmac_readreg(dma_reg);
			return TRUE;
#endif

#if XM7_VER == 1 && defined(SFDC)
		/* AhXWX^() */
		case 0xfd40:
		case 0xfd44:
		case 0xfd48:
		case 0xfd4c:
			*dat = (BYTE)((dma_adr[(addr&0xf)>>2] >> 8) & 0xff);
			return TRUE;

		/* AhXWX^() */
		case 0xfd41:
		case 0xfd45:
		case 0xfd49:
		case 0xfd4d:
			*dat = (BYTE)(dma_adr[(addr&0xf)>>2] & 0xff);
			return TRUE;

		/* ]ꐔWX^() */
		case 0xfd42:
		case 0xfd46:
		case 0xfd4a:
		case 0xfd4e:
			*dat = (BYTE)((dma_bcr[(addr&0xf)>>2] >> 8) & 0xff);
			return TRUE;

		/* ]ꐔWX^() */
		case 0xfd43:
		case 0xfd47:
		case 0xfd4b:
		case 0xfd4f:
			*dat = (BYTE)(dma_bcr[(addr&0xf)>>2] & 0xff);
			return TRUE;

		/* `l䃌WX^ */
		case 0xfd50:
		case 0xfd51:
		case 0xfd52:
		case 0xfd53:
			*dat = dma_chcr[addr&0xf];
			dma_chcr[addr&0xf] &= 0x7f;
			return TRUE;

		/* D搧䃌WX^ */
		case 0xfd54:
			*dat = dma_pcr;
			return TRUE;

		/* 荞ݐ䃌WX^ */
		case 0xfd55:
			*dat = (BYTE)(((dma_dcr >> 4) | 0x80) & dma_icr);
			dma_dcr &= 0x0f;

			/* IRQ */
			dma_icr &= (BYTE)~0x80;
			dma_irq_flag = FALSE;
			maincpu_irq();
			return TRUE;

		/* f[^`FC䃌WX^ */
		case 0xfd56:
			/* MSB:DMA DEADLOCK? */
			*dat = (BYTE)(dma_dcr & 0x0f);
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	DMAC
 *	PoCg
 */
BOOL FASTCALL dmac_writeb(WORD addr, BYTE dat)
{
	/* f`FbN */
#if XM7_VER >= 2
	if (model_type <= MODEL_FM77AV20) {
		/* AV40/AV20EX/AV40EX̂ݗL */
		return FALSE;
	}
#else
	if (model_type != MODEL_FM8) {
		return FALSE;
	}

	/* FDCڑȂ΃VXegjbgڑƌȂ */
#ifdef SFDC
	if (!fdc_enable[FDC_M] &&
		!fdc_enable[FDC_S]) {
#else
	if (!fdc_enable) {
#endif
		return FALSE;
	}
#endif

	switch (addr) {
#if XM7_VER >= 2
		/* DMACAhX */
		case 0xfd98:
			dma_reg = (BYTE)(dat & 0x1f);
			return TRUE;

		/* DMACf[^ */
		case 0xfd99:
			dmac_writereg(dma_reg, dat);
			return TRUE;
#endif

#if XM7_VER == 1 && defined(SFDC)
		/* AhXWX^() */
		case 0xfd40:
		case 0xfd44:
		case 0xfd48:
		case 0xfd4c:
			dma_adr[(addr&0xf)>>2] = (WORD)((dma_adr[(addr&0xf)>>2] & 0xff) | (dat << 8));
			return TRUE;

		/* AhXWX^() */
		case 0xfd41:
		case 0xfd45:
		case 0xfd49:
		case 0xfd4d:
			dma_adr[(addr&0xf)>>2] = (WORD)((dma_adr[(addr&0xf)>>2] & 0xff00) | dat);
			return TRUE;

		/* ]ꐔWX^() */
		case 0xfd42:
		case 0xfd46:
		case 0xfd4a:
		case 0xfd4e:
			dma_bcr[(addr&0xf)>>2] = (WORD)((dma_bcr[(addr&0xf)>>2] & 0xff) | (dat << 8));
			return TRUE;

		/* ]ꐔWX^() */
		case 0xfd43:
		case 0xfd47:
		case 0xfd4b:
		case 0xfd4f:
			dma_bcr[(addr&0xf)>>2] = (WORD)((dma_bcr[(addr&0xf)>>2] & 0xff00) | dat);
			return TRUE;

		/* `l䃌WX^ */
		case 0xfd50:
		case 0xfd51:
		case 0xfd52:
		case 0xfd53:
			dma_chcr[addr&0xf] = (BYTE)((dma_chcr[addr&0xf] & 0xc0) | (dat & 0x0f));
			return TRUE;

		/* D搧䃌WX^ */
		case 0xfd54:
			dma_pcr = (BYTE)(dat & 0x8f);
			return TRUE;

		/* 荞ݐ䃌WX^ */
		case 0xfd55:
			dma_icr = (BYTE)((dma_icr & 0x80) | (dat & 0x0f));
			return TRUE;

		/* f[^`FC䃌WX^ */
		case 0xfd56:
			dma_dcr = (BYTE)((dma_dcr & 0xf0) | (dat & 0x0f));
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	DMAC
 *	Z[u
 */
BOOL FASTCALL dmac_save(int fileh)
{
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC))
	int	i;
#endif

	LOG0(Normal, "Z[u");

	if (!file_word_write(fileh, dma_adr[0])) {
		return FALSE;
	}
	if (!file_word_write(fileh, dma_bcr[0])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, dma_chcr[0])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, dma_pcr)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, dma_icr)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_byte_write(fileh, dma_reg)) {
		return FALSE;
	}
#endif
	if (!file_bool_write(fileh, dma_flag)) {
		return FALSE;
	}

	/* Ver9g */
	if (!file_byte_write(fileh, dma_dcr)) {
		return FALSE;
	}

	/* Ver912g */
	if (!file_bool_write(fileh, dma_burst_transfer)) {
		return FALSE;
	}

	/* Ver9gEAV40dl (40EXdlƂ̌݊ȂȂ̂Œ) */
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC))
	for (i=1;i<4;i++) {
		if (!file_word_write(fileh, dma_adr[i])) {
			return FALSE;
		}
		if (!file_word_write(fileh, dma_bcr[i])) {
			return FALSE;
		}
		if (!file_byte_write(fileh, dma_chcr[i])) {
			return FALSE;
		}
	}
#endif

	return TRUE;
}

/*
 *	DMAC
 *	[h
 */
#if XM7_VER >= 3
BOOL FASTCALL dmac_load(int fileh, int ver, BOOL bflag)
#else	/* XM7_VER >= 3 */
BOOL FASTCALL dmac_load(int fileh, int ver)
#endif	/* XM7_VER >= 3 */
{
	int	i;

	LOG0(Normal, "[h");

	/* o[W`FbN */
#if XM7_VER >= 3
	if (ver < 800) {
#elif XM7_VER >= 2
	if (ver < 1200) {
#else
	if (ver < 405) {
#endif
		dmac_reset();
		return TRUE;
	}

	if (!file_word_read(fileh, &dma_adr[0])) {
		return FALSE;
	}
	if (!file_word_read(fileh, &dma_bcr[0])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &dma_chcr[0])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &dma_pcr)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &dma_icr)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_byte_read(fileh, &dma_reg)) {
		return FALSE;
	}
#endif
	if (!file_bool_read(fileh, &dma_flag)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver9g */
	if (ver >= 900) {
#elif XM7_VER >= 2
	if (ver >= 1200) {
#else
	if (ver >= 405) {
#endif
		if (!file_byte_read(fileh, &dma_dcr)) {
			return FALSE;
		}

		/* Ver912g */
#if XM7_VER >= 3
		if (ver >= 912) {
#elif XM7_VER >= 2
		if (ver >= 1200) {
#else
		if (ver >= 405) {
#endif
			if (!file_bool_read(fileh, &dma_burst_transfer)) {
				return FALSE;
			}
		}
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC))
#if XM7_VER >= 3
		if ((ver <= 999 && !bflag) || ((ver >= 1200) && (ver < 1400))) {
			for (i=1; i<4; i++) {
				dma_adr[i] = 0xFFFF;
				dma_bcr[i] = 0xFFFF;
				dma_chcr[i] = 0x00;
			}
		}
		else
#endif
		/* Ver9gEAV40dl (40EXdlƂ̌݊ȂȂ̂Œ) */
		for (i=1; i<4; i++) {
			if (!file_word_read(fileh, &dma_adr[i])) {
				return FALSE;
			}
			if (!file_word_read(fileh, &dma_bcr[i])) {
				return FALSE;
			}
			if (!file_byte_read(fileh, &dma_chcr[i])) {
				return FALSE;
			}
		}
#endif
	}
	else {
		dma_dcr = 0x08;
	}

	/* Ch1`3 */
#if XM7_VER == 2
	for (i=1; i<4; i++) {
		dma_adr[i] = 0xFFFF;
		dma_bcr[i] = 0xFFFF;
		dma_chcr[i] = 0x00;
	}
#endif

	return TRUE;
}
#endif
