/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ L[{[h ]
 */

#include <string.h>
#include "xm7.h"
#include "event.h"
#include "keyboard.h"
#include "mainetc.h"
#include "rtc.h"
#include "device.h"
#include "keyenc.h"

/*
 *	O[o [N
 */
BOOL caps_flag;							/* CAPS tO */
BOOL kana_flag;							/* Ji tO */
BOOL ins_flag;							/* INS  tO */

BOOL shift_flag;						/* SHIFTL[tO */
BOOL lshift_flag;						/* SHIFTL[tO */
BOOL rshift_flag;						/* ESHIFTL[tO */
BOOL ctrl_flag;							/* CTRLL[tO */
BOOL graph_flag;						/* GRAPHL[tO */
BOOL break_flag;						/* BreakL[tO */

BYTE key_scan;							/* L[R[h(R[hAMake/Breakp) */
WORD key_fm7;							/* L[R[h(FM-7݊R[h) */
BOOL key_repeat_flag;					/* L[s[gtO */

BYTE key_format;						/* R[htH[}bg */
DWORD key_repeat_time1;					/* L[s[gJn */
DWORD key_repeat_time2;					/* L[s[gԊu */

#if XM7_VER >= 2
BYTE simpose_mode;						/* X[p[C|[Y [h */
BOOL simpose_half;						/* X[p[C|[Y n[tg[ */

BOOL digitize_enable;					/* fBW^CYLEtO */
BOOL digitize_keywait;					/* fBW^CYL[҂ */

BYTE amnork_mode;						/* Am[NBZp */
BOOL keyenc_beep_flag;					/* L[GR[_PBEEPtO */
#else
BOOL key_mode_flag;						/* FM-8[hFM-7݊L[̓[htO */
WORD key_fm8;							/* L[R[h(FM-8) */
BYTE key_push_data[16];					/* L[͏ */
#endif

/*
 *	X^eBbN [N
 */
static BYTE key_repeat_code;			/* L[s[gR[h */
#if XM7_VER >= 2
static BYTE key_trans_send[16];			/* L[GR[_sMobt@ */
static BYTE key_trans_sendc;			/* L[GR[_sMJE^ */
static BYTE key_trans_recv[16];			/* L[GR[_Mobt@ */
static BYTE key_trans_recvc;			/* L[GR[_MJE^ */
static BYTE key_trans_pos;				/* L[GR[^̃f[^ʒu */
static BOOL key_trans_rxrdy;			/* L[GR[_̎Mf[^ */
static BYTE simpose_mode_bak;			/* fBW^CYỈʃ[hAp */
static BOOL key_trans_ack;				/* L[GR[_ ACKM */
static BOOL key_hidden_message_flag;	/* L[GR[_ BbZ[WsσtO */
static BOOL key_hidden_message_run;		/* L[GR[_ BbZ[WstO */
static BYTE key_hidden_message_count;	/* L[GR[_ BbZ[WJE^ */
#endif
static BOOL key_input_mask;				/* L[{[h͋֎~/tO */


/*-[ L[e[u(FM-7݊) ]-----------------------------------------------*/

/*
 *	CTRL[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD ctrl_key_table[] = {
	0x0c, 0x1e, 0x1e,
	0x0d, 0x1c, 0x1c,
	0x11, 0x11, 0x11,
	0x12, 0x17, 0x17,
	0x13, 0x05, 0x05,
	0x14, 0x12, 0x12,
	0x15, 0x14, 0x14,
	0x16, 0x19, 0x19,
	0x17, 0x15, 0x15,
	0x18, 0x09, 0x09,
	0x19, 0x0f, 0x09,
	0x1a, 0x10, 0x10,
	0x1b, 0x00, 0x00,
	0x1c, 0x1b, 0x1b,
	0x1e, 0x01, 0x01,
	0x1f, 0x13, 0x13,
	0x20, 0x04, 0x04,
	0x21, 0x06, 0x06,
	0x22, 0x07, 0x07,
	0x23, 0x08, 0x08,
	0x24, 0x0a, 0x0a,
	0x25, 0x0b, 0x0b,
	0x26, 0x0c, 0x0c,
	0x29, 0x1d, 0x1d,
	0x2a, 0x1a, 0x1a,
	0x2b, 0x18, 0x18,
	0x2c, 0x03, 0x03,
	0x2d, 0x16, 0x16,
	0x2e, 0x02, 0x02,
	0x2f, 0x0e, 0x0e,
	0x30, 0x0d, 0x0d,
	0x34, 0x1f, 0x1f
};

/*
 *	GRAPH[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD graph_key_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0xf9, 0xf9,
	0x03, 0xfa, 0xfa,
	0x04, 0xfb, 0xfb,
	0x05, 0xfc, 0xfc,
	0x06, 0xf2, 0xf2,
	0x07, 0xf3, 0xf3,
	0x08, 0xf4, 0xf4,
	0x09, 0xf5, 0xf5,
	0x0a, 0xf6, 0xf6,
	0x0b, 0xf7, 0xf7,
	0x0c, 0x8c, 0x8c,
	0x0d, 0x8b, 0x8b,
	0x0e, 0xf1, 0xf1,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0xfd, 0xfd,
	0x12, 0xf8, 0xf8,
	0x13, 0xe4, 0xe4,
	0x14, 0xe5, 0xe5,
	0x15, 0x9c, 0x9c,
	0x16, 0x9d, 0x9d,
	0x17, 0xf0, 0xf0,
	0x18, 0xe8, 0xe8,
	0x19, 0xe9, 0xe9,
	0x1a, 0x8d, 0x8d,
	0x1b, 0x8a, 0x8a,
	0x1c, 0xed, 0xed,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x95, 0x95,
	0x1f, 0x96, 0x96,
	0x20, 0xe6, 0xe6,
	0x21, 0xe7, 0xe7,
	0x22, 0x9e, 0x9e,
	0x23, 0x9f, 0x9f,
	0x24, 0xea, 0xea,
	0x25, 0xeb, 0xeb,
	0x26, 0x8e, 0x8e,
	0x27, 0x99, 0x99,
	0x28, 0x94, 0x94,
	0x29, 0xec, 0xec,

	0x2a, 0x80, 0x80,
	0x2b, 0x81, 0x81,
	0x2c, 0x82, 0x82,
	0x2d, 0x83, 0x83,
	0x2e, 0x84, 0x84,
	0x2f, 0x85, 0x85,
	0x30, 0x86, 0x86,
	0x31, 0x87, 0x87,
	0x32, 0x88, 0x88,
	0x33, 0x97, 0x97,
	0x34, 0xe0, 0xe0,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x98, 0x98,
	0x37, 0x91, 0x91,
	0x38, 0x99, 0x99,
	0x39, 0xee, 0xee,
	0x3a, 0xe1, 0xe1,
	0x3b, 0xe2, 0xe2,
	0x3c, 0xe3, 0xe3,
	0x3d, 0xef, 0xef,
	0x3e, 0x93, 0x93,
	0x3f, 0x8f, 0x8f,
	0x40, 0x92, 0x92,
	0x41, 0xffff, 0xffff,
	0x42, 0x9a, 0x9a,
	0x43, 0x90, 0x90,
	0x44, 0x9b, 0x9b,
	0x45, 0x0d, 0x0d,
	0x46, 0xffff, 0xffff,
	0x47, 0xffff, 0xffff,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*
 *	ȃ[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD kana_key_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0xc7, 0xffff,
	0x03, 0xcc, 0xffff,
	0x04, 0xb1, 0xa7,
	0x05, 0xb3, 0xa9,
	0x06, 0xb4, 0xaa,
	0x07, 0xb5, 0xab,
	0x08, 0xd4, 0xac,
	0x09, 0xd5, 0xad,
	0x0a, 0xd6, 0xae,
	0x0b, 0xdc, 0xa6,
	0x0c, 0xce, 0xffff,
	0x0d, 0xcd, 0xffff,
	0x0e, 0xb0, 0xffff,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0xc0, 0xffff,
	0x12, 0xc3, 0xffff,
	0x13, 0xb2, 0xa8,
	0x14, 0xbd, 0xffff,
	0x15, 0xb6, 0xffff,
	0x16, 0xdd, 0xffff,
	0x17, 0xc5, 0xffff,
	0x18, 0xc6, 0xffff,
	0x19, 0xd7, 0xffff,
	0x1a, 0xbe, 0xffff,
	0x1b, 0xde, 0xffff,
	0x1c, 0xdf, 0xa2,
	0x1d, 0x0d, 0x0d,

	0x1e, 0xc1, 0xffff,
	0x1f, 0xc4, 0xffff,
	0x20, 0xbc, 0xffff,
	0x21, 0xca, 0xffff,
	0x22, 0xb7, 0xffff,
	0x23, 0xb8, 0xffff,
	0x24, 0xcf, 0xffff,
	0x25, 0xc9, 0xffff,
	0x26, 0xd8, 0xffff,
	0x27, 0xda, 0xffff,
	0x28, 0xb9, 0xffff,
	0x29, 0xd1, 0xa3,

	0x2a, 0xc2, 0xaf,
	0x2b, 0xbb, 0xffff,
	0x2c, 0xbf, 0xffff,
	0x2d, 0xcb, 0xffff,
	0x2e, 0xba, 0xffff,
	0x2f, 0xd0, 0xffff,
	0x30, 0xd3, 0xffff,
	0x31, 0xc8, 0xa4,
	0x32, 0xd9, 0xa1,
	0x33, 0xd2, 0xa5,
	0x34, 0xdb, 0xffff,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*
 *	CAP[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD caps_key_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x51, 0x71,
	0x12, 0x57, 0x77,
	0x13, 0x45, 0x65,
	0x14, 0x52, 0x72,
	0x15, 0x54, 0x74,
	0x16, 0x59, 0x79,
	0x17, 0x55, 0x75,
	0x18, 0x49, 0x69,
	0x19, 0x4f, 0x6f,
	0x1a, 0x50, 0x70,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x41, 0x61,
	0x1f, 0x53, 0x73,
	0x20, 0x44, 0x64,
	0x21, 0x46, 0x66,
	0x22, 0x47, 0x67,
	0x23, 0x48, 0x68,
	0x24, 0x4a, 0x6a,
	0x25, 0x4b, 0x6b,
	0x26, 0x4c, 0x6c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x5a, 0x7a,
	0x2b, 0x58, 0x78,
	0x2c, 0x43, 0x63,
	0x2d, 0x56, 0x76,
	0x2e, 0x42, 0x62,
	0x2f, 0x4e, 0x6e,
	0x30, 0x4d, 0x6d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0x22, 0x5f,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*
 *	NORMAL[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD norm_key_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x71, 0x51,
	0x12, 0x77, 0x57,
	0x13, 0x65, 0x45,
	0x14, 0x72, 0x52,
	0x15, 0x74, 0x54,
	0x16, 0x79, 0x59,
	0x17, 0x75, 0x55,
	0x18, 0x69, 0x49,
	0x19, 0x6f, 0x4f,
	0x1a, 0x70, 0x50,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x61, 0x41,
	0x1f, 0x73, 0x53,
	0x20, 0x64, 0x44,
	0x21, 0x66, 0x46,
	0x22, 0x67, 0x47,
	0x23, 0x68, 0x48,
	0x24, 0x6a, 0x4a,
	0x25, 0x6b, 0x4b,
	0x26, 0x6c, 0x4c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x7a, 0x5a,
	0x2b, 0x78, 0x58,
	0x2c, 0x63, 0x43,
	0x2d, 0x76, 0x56,
	0x2e, 0x62, 0x42,
	0x2f, 0x6e, 0x4e,
	0x30, 0x6d, 0x4d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0x22, 0x5f,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*-[ L[e[u(FM-8݊) ]-----------------------------------------------*/

#if XM7_VER == 1
/*
 *	CTRL[hiFM-8[hj
 *	R[hASHIFTȂASHIFT̏
 */
#define	ctrl_fm8_table	(ctrl_key_table)

/*
 *	GRAPH[hiFM-8[hj
 *	R[hASHIFTȂASHIFT̏
 */
WORD graph_fm8_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0xf9, 0xf9,
	0x03, 0xfa, 0xfa,
	0x04, 0xfb, 0xfb,
	0x05, 0xfc, 0xfc,
	0x06, 0xf2, 0xf2,
	0x07, 0xf3, 0xf3,
	0x08, 0xf4, 0xf4,
	0x09, 0xf5, 0xf5,
	0x0a, 0xf6, 0xf6,
	0x0b, 0xf7, 0xf7,
	0x0c, 0x8c, 0x8c,
	0x0d, 0x8b, 0x8b,
	0x0e, 0xf1, 0xf1,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0xfd, 0xfd,
	0x12, 0xf8, 0xf8,
	0x13, 0xe4, 0xe4,
	0x14, 0xe5, 0xe5,
	0x15, 0x9c, 0x9c,
	0x16, 0x9d, 0x9d,
	0x17, 0xf0, 0xf0,
	0x18, 0xe8, 0xe8,
	0x19, 0xe9, 0xe9,
	0x1a, 0x8d, 0x8d,
	0x1b, 0x8a, 0x8a,
	0x1c, 0xed, 0xed,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x95, 0x95,
	0x1f, 0x96, 0x96,
	0x20, 0xe6, 0xe6,
	0x21, 0xe7, 0xe7,
	0x22, 0x9e, 0x9e,
	0x23, 0x9f, 0x9f,
	0x24, 0xea, 0xea,
	0x25, 0xeb, 0xeb,
	0x26, 0x8e, 0x8e,
	0x27, 0x99, 0x99,
	0x28, 0x94, 0x94,
	0x29, 0xec, 0xec,

	0x2a, 0x80, 0x80,
	0x2b, 0x81, 0x81,
	0x2c, 0x82, 0x82,
	0x2d, 0x83, 0x83,
	0x2e, 0x84, 0x84,
	0x2f, 0x85, 0x85,
	0x30, 0x86, 0x86,
	0x31, 0x87, 0x87,
	0x32, 0x88, 0x88,
	0x33, 0x97, 0x97,
	0x34, 0xe0, 0xe0,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0xffff, 0xffff,
	0x37, 0xffff, 0xffff,
	0x38, 0xef, 0xef,
	0x39, 0x9b, 0x9b,
	0x3a, 0x98, 0x98,
	0x3b, 0x91, 0x91,
	0x3c, 0x99, 0x99,
	0x3d, 0xee, 0xee,
	0x3e, 0xe1, 0xe1,
	0x3f, 0xe2, 0xe2,
	0x40, 0xe3, 0xe3,
	0x41, 0xee, 0xee,
	0x42, 0x93, 0x93,
	0x43, 0x8f, 0x8f,
	0x44, 0x92, 0x92,
	0x45, 0x0d, 0x0d,
	0x46, 0x9a, 0x9a,
	0x47, 0x90, 0x90,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*
 *	ȃ[hiFM-8[hj
 *	R[hASHIFTȂASHIFT̏
 */
#define	kana_fm8_table	(kana_key_table)

/*
 *	CAP[hiFM-8[hj
 *	R[hASHIFTȂASHIFT̏
 */
WORD caps_fm8_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x51, 0x71,
	0x12, 0x57, 0x77,
	0x13, 0x45, 0x65,
	0x14, 0x52, 0x72,
	0x15, 0x54, 0x74,
	0x16, 0x59, 0x79,
	0x17, 0x55, 0x75,
	0x18, 0x49, 0x69,
	0x19, 0x4f, 0x6f,
	0x1a, 0x50, 0x70,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x41, 0x61,
	0x1f, 0x53, 0x73,
	0x20, 0x44, 0x64,
	0x21, 0x46, 0x66,
	0x22, 0x47, 0x67,
	0x23, 0x48, 0x68,
	0x24, 0x4a, 0x6a,
	0x25, 0x4b, 0x6b,
	0x26, 0x4c, 0x6c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x5a, 0x7a,
	0x2b, 0x58, 0x78,
	0x2c, 0x43, 0x63,
	0x2d, 0x56, 0x76,
	0x2e, 0x42, 0x62,
	0x2f, 0x4e, 0x6e,
	0x30, 0x4d, 0x6d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0xffff, 0x5f,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};

/*
 *	NORMAL[hiFM-8[hj
 *	R[hASHIFTȂASHIFT̏
 */
WORD norm_fm8_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x71, 0x51,
	0x12, 0x77, 0x57,
	0x13, 0x65, 0x45,
	0x14, 0x72, 0x52,
	0x15, 0x74, 0x54,
	0x16, 0x79, 0x59,
	0x17, 0x75, 0x55,
	0x18, 0x69, 0x49,
	0x19, 0x6f, 0x4f,
	0x1a, 0x70, 0x50,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x61, 0x41,
	0x1f, 0x73, 0x53,
	0x20, 0x64, 0x44,
	0x21, 0x66, 0x46,
	0x22, 0x67, 0x47,
	0x23, 0x68, 0x48,
	0x24, 0x6a, 0x4a,
	0x25, 0x6b, 0x4b,
	0x26, 0x6c, 0x4c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x7a, 0x5a,
	0x2b, 0x78, 0x58,
	0x2c, 0x63, 0x43,
	0x2d, 0x76, 0x56,
	0x2e, 0x62, 0x42,
	0x2f, 0x6e, 0x4e,
	0x30, 0x6d, 0x4d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0xffff, 0x5f,

	0x48, 0x12, 0x12,
	0x4d, 0x1e, 0x19,
	0x4b, 0x7f, 0x7f,
	0x4f, 0x1d, 0x02,
	0x50, 0x1f, 0x1a,
	0x51, 0x1c, 0x06,

	0x49, 0x05, 0x05,
	0x4a, 0x0c, 0x0c,
	0x4c, 0x11, 0x11,
	0x4e, 0x0b, 0x0b,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x20, 0x20,
	0x58, 0x20, 0x20,
	0x35, 0x20, 0x20
};
#endif

/*-[ L[e[u(FM-16݊) ]--------------------------------------------*/

#if XM7_VER >= 2
/*
 *	CTRL[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD ctrl_16b_table[] = {
	0x0c, 0x1e, 0x1e,
	0x0d, 0x1c, 0x1c,
	0x11, 0x11, 0x11,
	0x12, 0x17, 0x17,
	0x13, 0x05, 0x05,
	0x14, 0x12, 0x12,
	0x15, 0x14, 0x14,
	0x16, 0x19, 0x19,
	0x17, 0x15, 0x15,
	0x18, 0x09, 0x09,
	0x19, 0x0f, 0x09,
	0x1a, 0x10, 0x10,
	0x1b, 0x00, 0x00,
	0x1c, 0x1b, 0x1b,
	0x1e, 0x01, 0x01,
	0x1f, 0x13, 0x13,
	0x20, 0x04, 0x04,
	0x21, 0x06, 0x06,
	0x22, 0x07, 0x07,
	0x23, 0x08, 0x08,
	0x24, 0x0a, 0x0a,
	0x25, 0x0b, 0x0b,
	0x26, 0x0c, 0x0c,
	0x29, 0x1d, 0x1d,
	0x2a, 0x1a, 0x1a,
	0x2b, 0x18, 0x18,
	0x2c, 0x03, 0x03,
	0x2d, 0x16, 0x16,
	0x2e, 0x02, 0x02,
	0x2f, 0x0e, 0x0e,
	0x30, 0x0d, 0x0d,
	0x34, 0x1f, 0x1f,

	0x57, 0x14e, 0x14e,
	0x58, 0x14d, 0x14d,

	0x49, 0x152, 0x152,
	0x4a, 0x151, 0x151,
	0x4c, 0x153, 0x153,
	0x4e, 0x155, 0x155,

	0x48, 0x150, 0x150,
	0x4d, 0x156, 0x156,
	0x4b, 0x154, 0x154,
	0x4f, 0x158, 0x158,
	0x50, 0x157, 0x157,
	0x51, 0x159, 0x159
};

/*
 *	GRAPH[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD graph_16b_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0xf9, 0xf9,
	0x03, 0xfa, 0xfa,
	0x04, 0xfb, 0xfb,
	0x05, 0xfc, 0xfc,
	0x06, 0xf2, 0xf2,
	0x07, 0xf3, 0xf3,
	0x08, 0xf4, 0xf4,
	0x09, 0xf5, 0xf5,
	0x0a, 0xf6, 0xf6,
	0x0b, 0xf7, 0xf7,
	0x0c, 0x8c, 0x8c,
	0x0d, 0x8b, 0x8b,
	0x0e, 0xf1, 0xf1,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0xfd, 0xfd,
	0x12, 0xf8, 0xf8,
	0x13, 0xe4, 0xe4,
	0x14, 0xe5, 0xe5,
	0x15, 0x9c, 0x9c,
	0x16, 0x9d, 0x9d,
	0x17, 0xf0, 0xf0,
	0x18, 0xe8, 0xe8,
	0x19, 0xe9, 0xe9,
	0x1a, 0x8d, 0x8d,
	0x1b, 0x8a, 0x8a,
	0x1c, 0xed, 0xed,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x95, 0x95,
	0x1f, 0x96, 0x96,
	0x20, 0xe6, 0xe6,
	0x21, 0xe7, 0xe7,
	0x22, 0x9e, 0x9e,
	0x23, 0x9f, 0x9f,
	0x24, 0xea, 0xea,
	0x25, 0xeb, 0xeb,
	0x26, 0x8e, 0x8e,
	0x27, 0x99, 0x99,
	0x28, 0x94, 0x94,
	0x29, 0xec, 0xec,

	0x2a, 0x80, 0x80,
	0x2b, 0x81, 0x81,
	0x2c, 0x82, 0x82,
	0x2d, 0x83, 0x83,
	0x2e, 0x84, 0x84,
	0x2f, 0x85, 0x85,
	0x30, 0x86, 0x86,
	0x31, 0x87, 0x87,
	0x32, 0x88, 0x88,
	0x33, 0x97, 0x97,
	0x34, 0xe0, 0xe0,

	0x48, 0x110, 0x130,
	0x4d, 0x116, 0x136,
	0x4b, 0x114, 0x134,
	0x4f, 0x118, 0x138,
	0x50, 0x117, 0x137,
	0x51, 0x119, 0x139,

	0x49, 0x112, 0x132,
	0x4a, 0x111, 0x131,
	0x4c, 0x113, 0x133,
	0x4e, 0x115, 0x135,

	0x36, 0x98, 0x98,
	0x37, 0x91, 0x91,
	0x38, 0x99, 0x99,
	0x39, 0xee, 0xee,
	0x3a, 0xe1, 0xe1,
	0x3b, 0xe2, 0xe2,
	0x3c, 0xe3, 0xe3,
	0x3d, 0xef, 0xef,
	0x3e, 0x93, 0x93,
	0x3f, 0x8f, 0x8f,
	0x40, 0x92, 0x92,
	0x41, 0xffff, 0xffff,
	0x42, 0x9a, 0x9a,
	0x43, 0x90, 0x90,
	0x44, 0x9b, 0x9b,
	0x45, 0x0d, 0x0d,
	0x46, 0xffff, 0xffff,
	0x47, 0xffff, 0xffff,

	0x57, 0x10e, 0x12e,
	0x58, 0x10d, 0x12d,
	0x35, 0x20, 0x20
};

/*
 *	ȃ[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD kana_16b_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0xc7, 0xffff,
	0x03, 0xcc, 0xffff,
	0x04, 0xb1, 0xa7,
	0x05, 0xb3, 0xa9,
	0x06, 0xb4, 0xaa,
	0x07, 0xb5, 0xab,
	0x08, 0xd4, 0xac,
	0x09, 0xd5, 0xad,
	0x0a, 0xd6, 0xae,
	0x0b, 0xdc, 0xa6,
	0x0c, 0xce, 0xffff,
	0x0d, 0xcd, 0xffff,
	0x0e, 0xb0, 0xffff,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0xc0, 0xffff,
	0x12, 0xc3, 0xffff,
	0x13, 0xb2, 0xa8,
	0x14, 0xbd, 0xffff,
	0x15, 0xb6, 0xffff,
	0x16, 0xdd, 0xffff,
	0x17, 0xc5, 0xffff,
	0x18, 0xc6, 0xffff,
	0x19, 0xd7, 0xffff,
	0x1a, 0xbe, 0xffff,
	0x1b, 0xde, 0xffff,
	0x1c, 0xdf, 0xa2,
	0x1d, 0x0d, 0x0d,

	0x1e, 0xc1, 0xffff,
	0x1f, 0xc4, 0xffff,
	0x20, 0xbc, 0xffff,
	0x21, 0xca, 0xffff,
	0x22, 0xb7, 0xffff,
	0x23, 0xb8, 0xffff,
	0x24, 0xcf, 0xffff,
	0x25, 0xc9, 0xffff,
	0x26, 0xd8, 0xffff,
	0x27, 0xda, 0xffff,
	0x28, 0xb9, 0xffff,
	0x29, 0xd1, 0xa3,

	0x2a, 0xc2, 0xaf,
	0x2b, 0xbb, 0xffff,
	0x2c, 0xbf, 0xffff,
	0x2d, 0xcb, 0xffff,
	0x2e, 0xba, 0xffff,
	0x2f, 0xd0, 0xffff,
	0x30, 0xd3, 0xffff,
	0x31, 0xc8, 0xa4,
	0x32, 0xd9, 0xa1,
	0x33, 0xd2, 0xa5,
	0x34, 0xdb, 0xffff,

	0x48, 0x110, 0x130,
	0x4d, 0x116, 0x136,
	0x4b, 0x114, 0x134,
	0x4f, 0x118, 0x138,
	0x50, 0x117, 0x137,
	0x51, 0x119, 0x139,

	0x49, 0x112, 0x132,
	0x4a, 0x111, 0x131,
	0x4c, 0x113, 0x133,
	0x4e, 0x115, 0x135,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x10e, 0x12e,
	0x58, 0x10d, 0x12d,
	0x35, 0x20, 0x20
};

/*
 *	CAP[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD caps_16b_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x51, 0x71,
	0x12, 0x57, 0x77,
	0x13, 0x45, 0x65,
	0x14, 0x52, 0x72,
	0x15, 0x54, 0x74,
	0x16, 0x59, 0x79,
	0x17, 0x55, 0x75,
	0x18, 0x49, 0x69,
	0x19, 0x4f, 0x6f,
	0x1a, 0x50, 0x70,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x41, 0x61,
	0x1f, 0x53, 0x73,
	0x20, 0x44, 0x64,
	0x21, 0x46, 0x66,
	0x22, 0x47, 0x67,
	0x23, 0x48, 0x68,
	0x24, 0x4a, 0x6a,
	0x25, 0x4b, 0x6b,
	0x26, 0x4c, 0x6c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x5a, 0x7a,
	0x2b, 0x58, 0x78,
	0x2c, 0x43, 0x63,
	0x2d, 0x56, 0x76,
	0x2e, 0x42, 0x62,
	0x2f, 0x4e, 0x6e,
	0x30, 0x4d, 0x6d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0x22, 0x5f,

	0x48, 0x110, 0x130,
	0x4d, 0x116, 0x136,
	0x4b, 0x114, 0x134,
	0x4f, 0x118, 0x138,
	0x50, 0x117, 0x137,
	0x51, 0x119, 0x139,

	0x49, 0x112, 0x132,
	0x4a, 0x111, 0x131,
	0x4c, 0x113, 0x133,
	0x4e, 0x115, 0x135,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x10e, 0x12e,
	0x58, 0x10d, 0x12d,
	0x35, 0x20, 0x20
};

/*
 *	NORMAL[h
 *	R[hASHIFTȂASHIFT̏
 */
WORD norm_16b_table[] = {
	0x5d, 0x0101, 0xffff,
	0x5e, 0x0102, 0xffff,
	0x5f, 0x0103, 0xffff,
	0x60, 0x0104, 0xffff,
	0x61, 0x0105, 0xffff,
	0x62, 0x0106, 0xffff,
	0x63, 0x0107, 0xffff,
	0x64, 0x0108, 0xffff,
	0x65, 0x0109, 0xffff,
	0x66, 0x010a, 0xffff,

	0x01, 0x1b, 0x1b,
	0x02, 0x31, 0x21,
	0x03, 0x32, 0x22,
	0x04, 0x33, 0x23,
	0x05, 0x34, 0x24,
	0x06, 0x35, 0x25,
	0x07, 0x36, 0x26,
	0x08, 0x37, 0x27,
	0x09, 0x38, 0x28,
	0x0a, 0x39, 0x29,
	0x0b, 0x30, 0xffff,
	0x0c, 0x2d, 0x3d,
	0x0d, 0x5e, 0x7e,
	0x0e, 0x5c, 0x7c,
	0x0f, 0x08, 0x08,

	0x10, 0x09, 0x09,
	0x11, 0x71, 0x51,
	0x12, 0x77, 0x57,
	0x13, 0x65, 0x45,
	0x14, 0x72, 0x52,
	0x15, 0x74, 0x54,
	0x16, 0x79, 0x59,
	0x17, 0x75, 0x55,
	0x18, 0x69, 0x49,
	0x19, 0x6f, 0x4f,
	0x1a, 0x70, 0x50,
	0x1b, 0x40, 0x60,
	0x1c, 0x5b, 0x7b,
	0x1d, 0x0d, 0x0d,

	0x1e, 0x61, 0x41,
	0x1f, 0x73, 0x53,
	0x20, 0x64, 0x44,
	0x21, 0x66, 0x46,
	0x22, 0x67, 0x47,
	0x23, 0x68, 0x48,
	0x24, 0x6a, 0x4a,
	0x25, 0x6b, 0x4b,
	0x26, 0x6c, 0x4c,
	0x27, 0x3b, 0x2b,
	0x28, 0x3a, 0x2a,
	0x29, 0x5d, 0x7d,

	0x2a, 0x7a, 0x5a,
	0x2b, 0x78, 0x58,
	0x2c, 0x63, 0x43,
	0x2d, 0x76, 0x56,
	0x2e, 0x62, 0x42,
	0x2f, 0x6e, 0x4e,
	0x30, 0x6d, 0x4d,
	0x31, 0x2c, 0x3c,
	0x32, 0x2e, 0x3e,
	0x33, 0x2f, 0x3f,
	0x34, 0x22, 0x5f,

	0x48, 0x110, 0x130,
	0x4d, 0x116, 0x136,
	0x4b, 0x114, 0x134,
	0x4f, 0x118, 0x138,
	0x50, 0x117, 0x137,
	0x51, 0x119, 0x139,

	0x49, 0x112, 0x132,
	0x4a, 0x111, 0x131,
	0x4c, 0x113, 0x133,
	0x4e, 0x115, 0x135,

	0x36, 0x2a, 0x2a,
	0x37, 0x2f, 0x2f,
	0x38, 0x2b, 0x2b,
	0x39, 0x2d, 0x2d,
	0x3a, 0x37, 0x37,
	0x3b, 0x38, 0x38,
	0x3c, 0x39, 0x39,
	0x3d, 0x3d, 0x3d,
	0x3e, 0x34, 0x34,
	0x3f, 0x35, 0x35,
	0x40, 0x36, 0x36,
	0x41, 0x2c, 0x2c,
	0x42, 0x31, 0x31,
	0x43, 0x32, 0x32,
	0x44, 0x33, 0x33,
	0x45, 0x0d, 0x0d,
	0x46, 0x30, 0x30,
	0x47, 0x2e, 0x2e,

	0x57, 0x10e, 0x12e,
	0x58, 0x10d, 0x12d,
	0x35, 0x20, 0x20
};
#endif

/*
 *	ID
 */
#define	ID	MAKEID('K', 'B', 'D', ' ')
//#define	KBD_LOG

/*-[ L[{[h ]-----------------------------------------------------------*/

/*
 *	L[{[h
 *	
 */
BOOL FASTCALL keyboard_init(void)
{
#if XM7_VER >= 2
	/* fBW^CYJ[h */
	digitize_enable = TRUE;
#else
	key_mode_flag = FALSE;
	memset(key_push_data, 0, sizeof(key_push_data));
#endif
	key_input_mask = FALSE;

	return TRUE;
}

/*
 *	L[{[h
 *	N[Abv
 */
void FASTCALL keyboard_cleanup(void)
{
}

/*
 *	L[{[h
 *	Zbg
 */
void FASTCALL keyboard_reset(void)
{
	LOG0(Normal, "Zbg");

	/* L[{[h͏Ԃ̏v */
	keyclear_request();

	/* L[ʂ̃tO */
	caps_flag = FALSE;
	kana_flag = FALSE;
	ins_flag = FALSE;

	shift_flag = FALSE;
	lshift_flag = FALSE;
	rshift_flag = FALSE;
	ctrl_flag = FALSE;
	graph_flag = FALSE;
	break_flag = FALSE;

	/* Zbgkey_fm7 != 0ɂ(fXtH[X΍) */
	key_scan = 0;
	key_fm7 = 0xffff;

	/* L[{[hp[^ */
	key_format = KEY_FORMAT_9BIT;
	key_repeat_time1 = 700000;
	key_repeat_time2 = 70000;

	key_repeat_flag = TRUE;
	key_repeat_code = 0;

#if XM7_VER >= 2
	/* L[{[hGR[_ */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
	key_trans_ack = TRUE;

	/* X[p[C|[Y */
	simpose_mode = 0;
	simpose_half = FALSE;

	/* fBW^CY */
	digitize_keywait = FALSE;
	simpose_mode_bak = 0;

	/* Am[N */
	amnork_mode = 0;

	/* BbZ[W */
	key_hidden_message_flag = FALSE;
	key_hidden_message_run = FALSE;
	key_hidden_message_count = 0;
#else
	key_fm8 = 0xffff;
#endif
}

/*
 *	L[͋/֎~
 */
void FASTCALL keyboard_inputmask(BOOL flag)
{
	/* L[͏Ԃۑ */
	key_input_mask = flag;

	/* L[͏ԂNA */
	keyclear_request();
}

/*-[ L[GR[_(FM77AV) ]-----------------------------------------------*/

#if XM7_VER >= 2
/*
 *	R[hnݒ
 */
static void FASTCALL key_set_format(void)
{
	/* p[^`FbN */
	if (key_trans_send[1] <= KEY_FORMAT_SCAN) {
		/* ݒ */
		key_format = key_trans_send[1];
		caps_flag = FALSE;
		kana_flag = FALSE;
		ins_flag = FALSE;
		key_repeat_flag = TRUE;
		key_repeat_code = 0;
		key_repeat_time1 = 700000;
		key_repeat_time2 = 70000;

		/* Vtg (F-BASIC V3.3/V3.4 {ꃂ[h) */
		shift_flag = FALSE;
		lshift_flag = FALSE;
		rshift_flag = FALSE;
		ctrl_flag = FALSE;
		graph_flag = FALSE;
	}

	/* I */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	R[hn擾
 */
static void FASTCALL key_get_format(void)
{
	/* X|Xݒ */
	key_trans_recv[0] = key_format;

	key_trans_sendc = 0;
	key_trans_recvc = 1;
	key_trans_pos = 0;
	key_trans_rxrdy = TRUE;
}

/*
 *	LEDݒ
 */
static void FASTCALL key_set_led(void)
{
	switch (key_trans_send[1]) {
		/* CAP ON */
		case 0:
			caps_flag = TRUE;
			break;

		/* CAP OFF */
		case 1:
			caps_flag = FALSE;
			break;

		/*  ON */
		case 2:
			kana_flag = TRUE;
			break;

		/*  OFF */
		case 3:
			kana_flag = FALSE;
			break;
	}

	/* I */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	LED擾
 */
static void FASTCALL key_get_led(void)
{
	/* X|Xݒ */
	key_trans_recv[0] = 0;
	if (caps_flag) {
		key_trans_recv[0] |= 0x01;
	}
	if (kana_flag) {
		key_trans_recv[0] |= 0x02;
	}

	key_trans_sendc = 0;
	key_trans_recvc = 1;
	key_trans_pos = 0;
	key_trans_rxrdy = TRUE;
}

/*
 *	L[s[gݒ
 */
static void FASTCALL key_set_repeat(void)
{
	switch (key_trans_send[1]) {
		/* s[gON */
		case 0:
			key_repeat_flag = TRUE;
			key_repeat_code = 0;
			break;

		/* s[gOFF */
		case 1:
			key_repeat_flag = FALSE;
			break;
	}

	/* I */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	L[s[gԐݒ
 */
static void FASTCALL key_set_time(void)
{
	if ((key_trans_send[1] == 0) || (key_trans_send[2] == 0)) {
		/* ǂ炩0ȂAftHgɐݒ */
		key_repeat_time1 = 700000;
		key_repeat_time2 = 70000;
	}
	else {
		/* p[^~10msɐݒ */
		key_repeat_time1 = (DWORD)key_trans_send[1];
		key_repeat_time1 *= 10000;
		key_repeat_time2 = (DWORD)key_trans_send[2];
		key_repeat_time2 *= 10000;
	}

	/* I */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	RTCݒ
 */
static void FASTCALL key_set_rtc(void)
{
	/* vZbg */
	rtc_set(&key_trans_send[2]);

	/* I */
	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	RTC擾
 */
static void FASTCALL key_get_rtc(void)
{
	/* v擾 */
	rtc_get(&key_trans_recv[0]);

	key_trans_sendc = 0;
	key_trans_recvc = 7;
	key_trans_pos = 0;
	key_trans_rxrdy = TRUE;
}

/*
 *	ʃfBW^CY
 */
static void FASTCALL key_capture(void)
{
	/* fBW^CYon */
	digitize_keywait = TRUE;
	simpose_mode_bak = simpose_mode;
	simpose_mode = 3;

	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	ʃ[hݒ
 */
static void FASTCALL key_set_scrmode(void)
{
	/* p[^󂯎 */
	switch (key_trans_send[1]) {
		/* p\R[h */
		case 0x00:
			simpose_mode = 0;
			break;
		/* X[p[C|[Y[h */
		case 0x01:
			simpose_mode = 1;
			break;
		/* er[h */
		case 0x02:
			simpose_mode = 2;
			break;
		/* fBW^CY[h */
		case 0x03:
			simpose_mode = 3;
			break;
		/* ȊÓA`p[^ƂăXLbv */
		default:
			break;
	}

	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	ʃ[h擾
 */
static void FASTCALL key_get_scrmode(void)
{
	/* X|Xݒ */
	if (simpose_half) {
		key_trans_recv[0] = 0x01;
	}
	else {
		key_trans_recv[0] = 0x00;
	}

	key_trans_sendc = 0;
	key_trans_recvc = 1;
	key_trans_pos = 0;
	key_trans_rxrdy = TRUE;
}

/*
 *	ʋPxݒ
 */
static void FASTCALL key_set_scrtone(void)
{
	/* p[^󂯎 */
	switch (key_trans_send[1]) {
		/* ʍPx[h */
		case 0x00:
			simpose_half = FALSE;
			break;
		/* n[tg[[h */
		case 0x01:
			simpose_half = TRUE;
			break;
		/* ȊÓA`p[^ƂăXLbv */
		default:
			break;
	}

	key_trans_sendc = 0;
	key_trans_recvc = 0;
	key_trans_pos = 0;
	key_trans_rxrdy = FALSE;
}

/*
 *	L[{[h
 *	ACKMM
 */
static BOOL FASTCALL key_send_ack(void)
{
	key_trans_ack = TRUE;
	schedule_delevent(EVENT_KEYENC_ACK);
	return TRUE;
}
#endif

/*-[ }bvhI/O ]----------------------------------------------------*/

/*
 *	L[{[h
 *	PoCgǂݍ
 */
BOOL FASTCALL keyboard_readb(WORD addr, BYTE *dat)
{
#if XM7_VER >= 2
	BYTE tmp;
#endif

	switch (addr) {
		/* L[R[h */
		case 0xd400:
			if (key_fm7 & 0x0100) {
				*dat = 0xff;
			}
			else {
				*dat = 0x7f;
			}
			return TRUE;

		/* L[R[h */
		case 0xd401:
			*dat = (BYTE)(key_fm7 & 0xff);
			/* 荞off */
			key_irq_flag = FALSE;
			maincpu_irq();
			subcpu_firq();
			return TRUE;

		/* INS LED ON */
		case 0xd40d:
			ins_flag = TRUE;
			return TRUE;

#if XM7_VER >= 2
		/* L[GR[_ f[^M */
		case 0xd431:
			if (model_type >= MODEL_FM77AV) {
				*dat = key_trans_recv[key_trans_pos];
				/* obt@f[^͈͓̔ŁAi߂ */
				if (key_trans_pos < key_trans_recvc) {
					key_trans_pos++;
					key_trans_rxrdy = FALSE;
				}
			}
			else {
				*dat = 0xff;
			}
			return TRUE;

		/* L[GR[_ Xe[^X */
		case 0xd432:
			if (model_type >= MODEL_FM77AV) {
				if (key_trans_rxrdy) {
					tmp = 0x7f;
				}
				else {
					tmp = 0xff;
				}
				/* fBW^CYL[҂̏ꍇACK SEND𑗐MȂ */
				if (!key_trans_ack || digitize_keywait) {
					tmp &= 0xfe;
				}
				*dat = tmp;

				/* f[^΁ARXRDY */
				if (key_trans_pos < key_trans_recvc) {
					key_trans_rxrdy = TRUE;
				}
			}
			else {
				*dat = 0xff;
			}
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	L[{[h
 *	PoCg
 */
BOOL FASTCALL keyboard_writeb(WORD addr, BYTE dat)
{
#if XM7_VER == 1
	UNUSED(dat);
#endif

	switch (addr) {
		/* L[R[h */
		case 0xd400:
			return TRUE;

		/* L[R[h */
		case 0xd401:
			return TRUE;

		/* INS LED OFF */
		case 0xd40d:
			ins_flag = FALSE;
			return TRUE;

#if XM7_VER >= 2
		/* L[GR[_ f[^M */
		case 0xd431:
			if (model_type >= MODEL_FM77AV) {
				/* obt@z̃`FbN */
				if (key_trans_sendc >= 16) {
					key_trans_sendc = 0;
					break;
				}
				/* f[^ZbgAI`FbN */
				key_trans_send[key_trans_sendc++] = dat;
				switch (key_trans_send[0]) {
					/* R[hn؂ւ */
					case 0:
						if (key_trans_sendc >= 2) {
							key_set_format();
							/* MakeԂł΍ēxMakeR[hĔs */
							if (!(key_scan & 0x80) && key_scan) {
								keyboard_make(key_scan);
							}
						}
						break;
					/* R[hnǂݎ */
					case 1:
						key_get_format();
						break;
					/* LEDݒ */
					case 2:
						if (key_trans_sendc >= 2) {
							key_set_led();

							if (key_format != KEY_FORMAT_SCAN) {
								/* MakeԂł΍ēxMakeR[hs */
								if (!(key_scan & 0x80) && key_scan) {
									keyboard_make(key_scan);
								}
							}
						}
						break;
					/* LED擾 */
					case 3:
						key_get_led();
						break;
					/* s[gݒ */
					case 4:
						if (key_trans_sendc >= 2) {
							key_set_repeat();
						}
						break;
					/* s[gԐݒ */
					case 5:
						if (key_trans_sendc >= 3) {
							key_set_time();
						}
						break;
					/* RTCM */
					case 0x80:
						if (key_trans_sendc >= 2) {
							if (key_trans_send[1] == 0) {
								key_get_rtc();
								break;
							}
							if (key_trans_send[1] == 1) {
								if (key_trans_sendc >= 9) {
									key_set_rtc();
								}
								break;
							}
							/* `p[^ */
							key_trans_sendc = 0;
						}
						break;
					/* ʃfBW^CY */
					case 0x81:
						if (key_trans_sendc >= 2) {
							key_capture();
						}
						break;
					/* ʃ[h */
					case 0x82:
						if (key_trans_sendc >= 2) {
							key_set_scrmode();
						}
						break;
					/* ʃ[h擾 */
					case 0x83:
						key_get_scrmode();
						break;
					/* erʋPx */
					case 0x84:
						if (key_trans_sendc >= 2) {
							key_set_scrtone();
						}
						break;
					/* ȊONOP */
					default:
						key_trans_sendc = 0;
				}

				/* 5soߌ(K)ACKM */
				key_trans_ack = FALSE;
				schedule_setevent(EVENT_KEYENC_ACK, 5, key_send_ack);
			}
			return TRUE;
#endif
	}

	return FALSE;
}

/*-[ L[R[hϊ ]-------------------------------------------------------*/

/*
 *	L[{[h
 *	VtgԃtOXV
 */
static BOOL FASTCALL keyboard_update_shift(BYTE dat)
{
	switch (dat) {
		/* SHIFT  */
		case 0x53:
			shift_flag = TRUE;
			lshift_flag = TRUE;
			return TRUE;

		/* SHIFT  */
		case 0x53 + 0x80:
			lshift_flag = FALSE;
			if (!rshift_flag) {
				shift_flag = FALSE;
			}
			return TRUE;

		/* ESHIFT  */
		case 0x54:
			shift_flag = TRUE;
			rshift_flag = TRUE;
			return TRUE;

		/* ESHIFT  */
		case 0x54 + 0x80:
			rshift_flag = FALSE;
			if (!lshift_flag) {
				shift_flag = FALSE;
			}
			return TRUE;

		/* CTRL  */
		case 0x52:
			ctrl_flag = TRUE;
			return TRUE;

		/* CTRL  */
		case 0x52 + 0x80:
			ctrl_flag = FALSE;
			return TRUE;

		/* GRAPH  */
		case 0x56:
			graph_flag = TRUE;
			return TRUE;

		/* GRAPH  */
		case 0x56 + 0x80:
			graph_flag = FALSE;
			return TRUE;

		/* CAP  */
		case 0x55:
			if (key_format != KEY_FORMAT_SCAN) {
				caps_flag = !(caps_flag);
			}
			return TRUE;

		/* Ji  */
		case 0x5a:
			if (key_format != KEY_FORMAT_SCAN) {
				kana_flag = !(kana_flag);
			}
			return TRUE;

		/* BREAK  */
		case 0x5c:
#if XM7_VER >= 2
		/* XL[hABREAKL[Ƒ̃L[ꂽꍇ̍׍H
		   (Am[N) */
			if (key_format == KEY_FORMAT_SCAN) {
				amnork_mode = 1;
			}
#endif
			break_flag = TRUE;
			maincpu_firq();
			return TRUE;

		/* BREAK  */
		case 0x5c + 0x80:
#if XM7_VER >= 2
			/* Am[N΍tO */
			if (key_format == KEY_FORMAT_SCAN) {
				amnork_mode = 0;
			}
#endif
			break_flag = FALSE;
			maincpu_firq();
			return TRUE;
	}

	return FALSE;
}

/*
 *	L[{[h
 *	Make/BreakR[hFM-7R[hϊ
 */
WORD FASTCALL keyboard_to_fm7(BYTE dat)
{
	int i;

	/* VtgԃtOXV */
	if (keyboard_update_shift(dat)) {
		return 0xffff;
	}

	/* MakeR[h */
	dat &= 0x7f;

	/* CTRL[h */
	if (ctrl_flag) {
		for (i=0; i<(sizeof(ctrl_key_table)/3); i++) {
			if (dat == ctrl_key_table[i * 3]) {
				if (!shift_flag) {
					return ctrl_key_table[i * 3 + 1];
				}
				else {
					return ctrl_key_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* GRAPH[h */
	if (graph_flag) {
		for (i=0; i<(sizeof(graph_key_table)/3); i++) {
			if (dat == graph_key_table[i * 3]) {
				if (!shift_flag) {
					return graph_key_table[i * 3 + 1];
				}
				else {
					return graph_key_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* ȃ[h */
	if (kana_flag) {
		for (i=0; i<(sizeof(kana_key_table)/3); i++) {
			if (dat == kana_key_table[i * 3]) {
				if (!shift_flag) {
					return kana_key_table[i * 3 + 1];
				}
				else {
					return kana_key_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* CAP[h */
	if (caps_flag) {
		for (i=0; i<(sizeof(caps_key_table)/3); i++) {
			if (dat == caps_key_table[i * 3]) {
				if (!shift_flag) {
					return caps_key_table[i * 3 + 1];
				}
				else {
					return caps_key_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* m[}[h */
	for (i=0; i<(sizeof(norm_key_table)/3); i++) {
		if (dat == norm_key_table[i * 3]) {
			if (!shift_flag) {
				return norm_key_table[i * 3 + 1];
			}
			else {
				return norm_key_table[i * 3 + 2];
			}
		}
	}

	return 0xffff;
}

#if XM7_VER >= 2
/*
 *	L[{[h
 *	Make/BreakR[hFM-16R[hϊ
 */
WORD FASTCALL keyboard_to_fm16b(BYTE dat)
{
	int i;

	/* VtgԃtOXV */
	if (keyboard_update_shift(dat)) {
		return 0xffff;
	}

	/* MakeR[h */
	dat &= 0x7f;

	/* CTRL[h */
	if (ctrl_flag) {
		for (i=0; i<(sizeof(ctrl_16b_table)/3); i++) {
			if (dat == ctrl_16b_table[i * 3]) {
				if (!shift_flag) {
					return ctrl_16b_table[i * 3 + 1];
				}
				else {
					return ctrl_16b_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* GRAPH[h */
	if (graph_flag) {
		for (i=0; i<(sizeof(graph_16b_table)/3); i++) {
			if (dat == graph_16b_table[i * 3]) {
				if (!shift_flag) {
					return graph_16b_table[i * 3 + 1];
				}
				else {
					return graph_16b_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* ȃ[h */
	if (kana_flag) {
		for (i=0; i<(sizeof(kana_16b_table)/3); i++) {
			if (dat == kana_16b_table[i * 3]) {
				if (!shift_flag) {
					return kana_16b_table[i * 3 + 1];
				}
				else {
					return kana_16b_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* CAP[h */
	if (caps_flag) {
		for (i=0; i<(sizeof(caps_16b_table)/3); i++) {
			if (dat == caps_16b_table[i * 3]) {
				if (!shift_flag) {
					return caps_16b_table[i * 3 + 1];
				}
				else {
					return caps_16b_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* m[}[h */
	for (i=0; i<(sizeof(norm_16b_table)/3); i++) {
		if (dat == norm_16b_table[i * 3]) {
			if (!shift_flag) {
				return norm_16b_table[i * 3 + 1];
			}
			else {
				return norm_16b_table[i * 3 + 2];
			}
		}
	}

	return 0xffff;
}

/*
 *	L[{[h
 *	Make/BreakR[hXLR[hϊ
 */
WORD FASTCALL keyboard_to_scan(BYTE dat)
{
	/* VtgԃtOXV */
	keyboard_update_shift(dat);

	/* XL[hABREAKL[Ƒ̃L[ꂽꍇ̍׍H */
	/* (Am[N) */
	if (key_format == KEY_FORMAT_SCAN) {
		if (((dat & 0x7f) != 0x5c) && (amnork_mode == 1)) {
			break_flag = FALSE;
			maincpu_firq();

			amnork_mode = 2;
			dat = (BYTE)((dat & 0x80) | 0x7f);
		}
	}

	/* R[h */
	return (WORD)dat;
}

/*
 *	L[GR[_
 *	BEEP~Cxg
 */
static BOOL FASTCALL keyencbeep_event(void)
{
	keyenc_beep_flag = FALSE;
	keyencbeep_notify();
	schedule_delevent(EVENT_KEYENC_BEEP);
	return TRUE;
}

/*
 *	L[GR[_
 *	BEEP
 */
static void FASTCALL keyenc_beep(void)
{
	keyenc_beep_flag = TRUE;
	keyencbeep_notify();
	schedule_setevent(EVENT_KEYENC_BEEP, 25000, keyencbeep_event);
}

/*
 *	L[GR[_
 *	BbZ[W
 */
static BOOL FASTCALL keyenc_hidden_message_event(void)
{
	key_fm7 = key_hidden_message[key_hidden_message_count];
	key_hidden_message_count ++;

	/* CPU֊荞 */
	key_irq_flag = TRUE;
	maincpu_irq();
	subcpu_firq();

	/* I */
	if (key_fm7 == 0x00) {
		schedule_delevent(EVENT_KEYENC_MESSAGE);
		key_hidden_message_flag = TRUE;
		key_hidden_message_run = FALSE;

		/* ISHIFTtO~낷 */
		lshift_flag = FALSE;
		rshift_flag = FALSE;
		shift_flag = FALSE;

		return TRUE;
	}

	/* BEEP */
	keyenc_beep();

	return TRUE;
}

#else
/*
 *	L[{[h
 *	Make/BreakR[hFM-8R[hϊ
 */
WORD FASTCALL keyboard_to_fm8(BYTE dat)
{
	int i;

	/* VtgԃtOXV */
	if (keyboard_update_shift(dat)) {
		return 0xffff;
	}

	/* MakeR[h */
	dat &= 0x7f;

	/* CTRL[h */
	if (ctrl_flag) {
		for (i=0; i<(sizeof(ctrl_fm8_table)/3); i++) {
			if (dat == ctrl_fm8_table[i * 3]) {
				if (!shift_flag) {
					return ctrl_fm8_table[i * 3 + 1];
				}
				else {
					return ctrl_fm8_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* GRAPH[h */
	if (graph_flag) {
		for (i=0; i<(sizeof(graph_fm8_table)/3); i++) {
			if (dat == graph_fm8_table[i * 3]) {
				if (!shift_flag) {
					return graph_fm8_table[i * 3 + 1];
				}
				else {
					return graph_fm8_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* ȃ[h */
	if (kana_flag) {
		for (i=0; i<(sizeof(kana_fm8_table)/3); i++) {
			if (dat == kana_fm8_table[i * 3]) {
				if (!shift_flag) {
					return kana_fm8_table[i * 3 + 1];
				}
				else {
					return kana_fm8_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* CAP[h */
	if (caps_flag) {
		for (i=0; i<(sizeof(caps_fm8_table)/3); i++) {
			if (dat == caps_fm8_table[i * 3]) {
				if (!shift_flag) {
					return caps_fm8_table[i * 3 + 1];
				}
				else {
					return caps_fm8_table[i * 3 + 2];
				}
			}
		}
		return 0xffff;
	}

	/* m[}[h */
	for (i=0; i<(sizeof(norm_fm8_table)/3); i++) {
		if (dat == norm_fm8_table[i * 3]) {
			if (!shift_flag) {
				return norm_fm8_table[i * 3 + 1];
			}
			else {
				return norm_fm8_table[i * 3 + 2];
			}
		}
	}

	return 0xffff;
}
#endif

/*-[ s[gAMakeABreak ]------------------------------------------------*/

/*
 *	L[{[h
 *	s[g^C}Cxg
 */
static BOOL FASTCALL keyboard_event(void)
{
	if (key_repeat_flag) {
		/* MakeR[hă[h(XL[h) */
		if (key_format == KEY_FORMAT_SCAN) {
			key_fm7 = key_repeat_code;
		}

		/* CPU֊荞 */
		key_irq_flag = TRUE;
		maincpu_irq();
		subcpu_firq();

		/* ̃Cxg */
		schedule_setevent(EVENT_KEYBOARD, key_repeat_time2, keyboard_event);
		return TRUE;
	}

	return FALSE;
}

#if XM7_VER == 1
/*
 *	L[{[h
 *	s[g^C}Cxg(FM-8)
 */
static BOOL FASTCALL keyboard_event_fm8(void)
{
	if (key_repeat_flag) {
		/* ŃL[R[hւ */
		if (key_fm8 != 0xffff) {
			key_fm7 = key_fm8;
		}

		/* CPU֊荞 */
		key_irq_flag = TRUE;
		maincpu_irq();
		subcpu_firq();

		/* ̃Cxg */
		schedule_setevent(EVENT_KEYBOARD, key_repeat_time1, keyboard_event);
		return TRUE;
	}

	return FALSE;
}
#endif

/*
 *	L[{[h
 *	L[s[g^C}ύX
 */
void FASTCALL keyboard_repeat(void)
{
	schedule_setevent(EVENT_KEYBOARD, key_repeat_time2, keyboard_event);
}

/*
 *	L[{[h
 *	Make
 */
void FASTCALL keyboard_make(BYTE dat)
{
	WORD code;
#if XM7_VER == 1
	BYTE num, key_dat;
	int n;
#endif

	/* L[͋֎~ȂA */
	if (key_input_mask) {
		return;
	}

#if XM7_VER >= 2
	/* fBW^CYL[҂̏ꍇ́AfBW^CY~߃L[ǂݎ̂Ă */
	if (digitize_keywait) {
		/* ʃ[hA */
		simpose_mode = simpose_mode_bak;

		/* fBW^CYs */
		if (digitize_enable) {
			digitize_notify();
		}

		digitize_keywait = FALSE;
		return;
	}
#else
	/* ϊ/ϊISPACEɊ蓖 */
	if ((dat == 0x57) || (dat == 0x58)) {
		dat = 0x35;
	}
#endif

	/* R[hL */
	key_scan = dat;

#if XM7_VER == 1
	/* L[ԕێ */
	key_push_data[dat >> 3] |= (0x80 >> (dat & 0x07));
#endif

	/* f[^ϊ */
#if XM7_VER >= 2
	switch (key_format) {
		/* FM-7݊[h */
		case KEY_FORMAT_9BIT:
			code = keyboard_to_fm7(dat);
			break;
		/* FM-16[h */
		case KEY_FORMAT_FM16B:
			code = keyboard_to_fm16b(dat);
			break;
		/* XL[h */
		case KEY_FORMAT_SCAN:
			code = keyboard_to_scan(dat);
			break;
		/* ȊO */
		default:
			ASSERT(FALSE);
			break;
	}
#else
	if (!key_mode_flag && (model_type == MODEL_FM8)) {
		code = keyboard_to_fm8(dat);
	}
	else {
		code = keyboard_to_fm7(dat);
	}
#endif

	/* VtgL[`FbN */
	if (shift_flag) {
		/* L[s[gon/off */
		if (ctrl_flag) {
			/* SHIFT+CTRL+0 L[s[goff */
			if ((dat == 0x0b) || (dat == 0x46)) {
				key_repeat_flag = FALSE;
				return;
			}
			/* SHIFT+CTRL+1 L[s[gon */
			if ((dat == 0x02) || (dat == 0x42)) {
				key_repeat_flag = TRUE;
				return;
			}
#if XM7_VER >= 2
			if (model_type >= MODEL_FM77AV) {
				if (lshift_flag && rshift_flag && caps_flag && kana_flag &&
					graph_flag && (dat == 0x15)) {
					if (!key_hidden_message_flag) {
						key_hidden_message_run = TRUE;
						schedule_setevent(EVENT_KEYENC_MESSAGE, 122880,
							keyenc_hidden_message_event);
					}
					else {
						keyenc_beep();
					}
				}
			}
#endif
		}

#if XM7_VER >= 2
		if (model_type >= MODEL_FM77AV) {
			/* X[p[C|[Y */
			switch (dat) {
				/* SHIFT+PF7 p\R[h */
				case 0x63:
					simpose_mode = 0;
					break;
				/* SHIFT+PF8 X[p[C|[Y(Px)[h */
				case 0x64:
					simpose_mode = 1;
					simpose_half = FALSE;
					break;
				/* SHIFT+PF9 X[p[C|[Y(Px)[h */
				case 0x65:
					simpose_mode = 1;
					simpose_half = TRUE;
					break;
				/* SHIFT+PF10 TV[h */
				case 0x66:
					simpose_mode = 2;
					break;
			}
		}
#endif
	}

#if XM7_VER >= 2
	/* BbZ[W\̓L[͂𖳎 */
	if (key_hidden_message_run) {
		return;
	}
#endif

#if XM7_VER == 1
	if (!key_mode_flag && (model_type == MODEL_FM8)) {
		n = 0;

		/* LL[o */
		for (num = 0x01; num <= 0x66; num ++) {
			/* VtgnL[̓JEgȂ */
			if (((num >= 0x52) && (num <= 0x56)) ||
				(num == 0x5a) || (num == 0x5c)) {
				continue;
			}
			if (key_push_data[num >> 3] & (0x80 >> (num & 0x07))) {
				n ++;
				key_dat = (BYTE)num;
			}
		}

		/* PL[͂Ȃ */
		if (n == 1) {
			code = keyboard_to_fm8(key_dat);

			/* s[gȂ */
			if ((event[EVENT_KEYBOARD].flag != EVENT_NOTUSE) &&
				(event[EVENT_KEYBOARD].reload == key_repeat_time2)) {
				schedule_delevent(EVENT_KEYBOARD);
				key_fm8 = code;
				key_repeat_code = key_dat;
				schedule_setevent(EVENT_KEYBOARD, key_repeat_time1, keyboard_event_fm8);
				return;
			}
		}
		else {
			/* PL[ȊO͖ */
			code = 0xffff;
		}
	}
#endif

	/* R[hLȂALĊ荞 */
	if (code != 0xffff) {
#ifdef KBD_LOG
		LOG1(Normal, "CN $%02X", code);
#endif

		key_fm7 = code;

		/* 荞 */
		key_irq_flag = TRUE;
		maincpu_irq();
		subcpu_firq();

		if ((code >= 0x101) && (code <= 0x10a)) {
			/* PFL[́As[gȂ */
			schedule_delevent(EVENT_KEYBOARD);
		}
		else {
			/* PFL[ȊOȂAs[g */
			key_repeat_code = dat;
			schedule_setevent(EVENT_KEYBOARD, key_repeat_time1, keyboard_event);
		}
	}
	else {
		/* Cxg폜 */
		schedule_delevent(EVENT_KEYBOARD);
	}
}

/*
 *	L[{[h
 *	Break
 */
void FASTCALL keyboard_break(BYTE dat)
{
#if XM7_VER == 1
	int rep_flag;
	WORD code;
	BYTE num, key_dat;
	int n;
#endif

	/* L[͋֎~ȂA */
	if (key_input_mask) {
		return;
	}

#if XM7_VER == 1
	/* ϊ/ϊISPACEɊ蓖 */
	if ((dat == 0x57) || (dat == 0x58)) {
		dat = 0x35;
	}
#endif

#ifdef KBD_LOG
	LOG1(Normal, "u[N $%02X", dat);
#endif

	/* [NɋL */
	key_scan = (BYTE)(dat | 0x80);

#if XM7_VER == 1
	/* L[ԕێ */
	key_push_data[dat >> 3] &= ~(0x80 >> (dat & 0x07));
#endif

	/* f[^ϊ */
#if XM7_VER >= 2
	switch (key_format) {
		/* FM-7݊[h */
		case KEY_FORMAT_9BIT:
			keyboard_to_fm7(key_scan);
			break;
		/* FM-16[h */
		case KEY_FORMAT_FM16B:
			keyboard_to_fm16b(key_scan);
			break;
		/* XL[h */
		case KEY_FORMAT_SCAN:
			keyboard_to_scan(key_scan);
			break;
		/* ȊO */
		default:
			ASSERT(FALSE);
			break;
	}
#else
	if (!key_mode_flag && (model_type == MODEL_FM8)) {
		keyboard_to_fm8(key_scan);
	}
	else {
		keyboard_to_fm7(key_scan);
	}
#endif

	/* XL[h̏ꍇ́ABreakf[^ */
	if (key_format == KEY_FORMAT_SCAN) {
#if XM7_VER >= 2
		/* BREAKL[Ƒ̃L[ꂽꍇ̍׍H (Am[N) */
		if (((key_scan & 0x7f) != 0x5c) && (amnork_mode == 2)) {
			amnork_mode = 0;
			key_fm7 = 0xff;
			keyenc_beep();
		}
		else {
			key_fm7 = key_scan;
		}
#else
		key_fm7 = key_scan;
#endif

		/* 荞 */
		key_irq_flag = TRUE;
		maincpu_irq();
		subcpu_firq();
	}

#if XM7_VER == 1
	/* s[gԂҔ */
	rep_flag = event[EVENT_KEYBOARD].flag;
#endif

	/* s[gȂ */
	if (dat == key_repeat_code) {
		schedule_delevent(EVENT_KEYBOARD);
	}

#if XM7_VER == 1
	if (!key_mode_flag && (model_type == MODEL_FM8)) {
		n = 0;

		/* LL[o */
		for (num = 0x01; num <= 0x66; num ++) {
			/* VtgnL[̓JEgȂ */
			if (((num >= 0x52) && (num <= 0x56)) ||
				(num == 0x5a) || (num == 0x5c)) {
				continue;
			}
			if (key_push_data[num >> 3] & (128 >> (num & 0x07))) {
				n ++;
				key_dat = (BYTE)num;
			}
		}

		/* PL[͂Ȃ */
		if (n == 1) {
			code = keyboard_to_fm8(key_dat);

			/* R[hLȂALĊ荞 */
			if (code != 0xffff) {
				/* s[gȂ */
				if ((rep_flag != EVENT_NOTUSE) &&
					(event[EVENT_KEYBOARD].reload == key_repeat_time2)) {
					schedule_delevent(EVENT_KEYBOARD);
					key_fm8 = code;
					key_repeat_code = key_dat;
					schedule_setevent(EVENT_KEYBOARD, key_repeat_time1, keyboard_event_fm8);
				}
				else {
					key_fm7 = code;

					/* 荞 */
					key_irq_flag = TRUE;
					maincpu_irq();
					subcpu_firq();

					if ((key_fm7 >= 0x101) && (key_fm7 <= 0x10a)) {
						/* PFL[́As[gȂ */
						schedule_delevent(EVENT_KEYBOARD);
					}
					else {
						/* PFL[ȊOȂAs[g */
						key_repeat_code = key_dat;
						schedule_setevent(EVENT_KEYBOARD, key_repeat_time1, keyboard_event);
					}
				}
			}
		}
	}
#endif
}

#ifdef KBDPASTE
/*
 *	L[{[h
 *	y[Xg
 */
BOOL FASTCALL keyboard_paste(BYTE dat)
{
	/* 荞݃`FbN */
	if (key_irq_flag) {
		/* łΎ񑗂 */
		return FALSE;
	}

	/* L */
	key_fm7 = dat;

	/* 荞 */
	key_irq_flag = TRUE;
	maincpu_irq();
	subcpu_firq();

	/* s[gȂ */
	schedule_delevent(EVENT_KEYBOARD);

	return TRUE;
}
#endif

/*
 *	L[{[h
 *	Z[u
 */
BOOL FASTCALL keyboard_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, caps_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, kana_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, ins_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, shift_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, ctrl_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, graph_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, break_flag)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, key_scan)) {
		return FALSE;
	}
	if (!file_word_write(fileh, key_fm7)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, key_repeat_flag)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, key_format)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, key_repeat_time1)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, key_repeat_time2)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, key_repeat_code)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_write(fileh, key_trans_send, sizeof(key_trans_send))) {
		return FALSE;
	}
	if (!file_byte_write(fileh, key_trans_sendc)) {
		return FALSE;
	}
	if (!file_write(fileh, key_trans_recv, sizeof(key_trans_recv))) {
		return FALSE;
	}
	if (!file_byte_write(fileh, key_trans_recvc)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, key_trans_pos)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, key_trans_rxrdy)) {
		return FALSE;
	}

	/* X[p[C|[YAfBW^CY(t@Co[W5Ŋg) */
	if (!file_byte_write(fileh, simpose_mode)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, simpose_half)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, digitize_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, digitize_keywait)) {
		return FALSE;
	}
#endif

	/* EVtgtO(Ver9.03/7.03Ŋg) */
	if (!file_bool_write(fileh, lshift_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, rshift_flag)) {
		return FALSE;
	}

#if XM7_VER >= 2
	/* L[{[hGR[_ACK(Ver9.06/7.06Ŋg) */
	if (!file_bool_write(fileh, key_trans_ack)) {
		return FALSE;
	}

	/* L[{[hGR[_BbZ[W(Ver9.19/7.19Ŋg) */
	if (!file_bool_write(fileh, key_hidden_message_flag)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, key_hidden_message_run)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, key_hidden_message_count)) {
		return FALSE;
	}
#else
	if (!file_word_write(fileh, key_fm8)) {
		return FALSE;
	}
	if (!file_write(fileh, key_push_data, sizeof(key_push_data))) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	L[{[h
 *	[h
 */
BOOL FASTCALL keyboard_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_bool_read(fileh, &caps_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &kana_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &ins_flag)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &shift_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &ctrl_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &graph_flag)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &break_flag)) {
		return FALSE;
	}

	if (!file_byte_read(fileh, &key_scan)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &key_fm7)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &key_repeat_flag)) {
		return FALSE;
	}

	if (!file_byte_read(fileh, &key_format)) {
		return FALSE;
	}
	if (!file_dword_read(fileh, &key_repeat_time1)) {
		return FALSE;
	}
	if (!file_dword_read(fileh, &key_repeat_time2)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &key_repeat_code)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_read(fileh, key_trans_send, sizeof(key_trans_send))) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &key_trans_sendc)) {
		return FALSE;
	}
	if (!file_read(fileh, key_trans_recv, sizeof(key_trans_recv))) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &key_trans_recvc)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &key_trans_pos)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &key_trans_rxrdy)) {
		return FALSE;
	}

	/* X[p[C|[YAfBW^CY(t@Co[W5Ŋg) */
	if (ver >= 500) {
		if (!file_byte_read(fileh, &simpose_mode)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &simpose_half)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &digitize_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &digitize_keywait)) {
			return FALSE;
		}
		if (model_type < MODEL_FM77AV) {
			simpose_mode = 0;
			simpose_half = FALSE;
			digitize_keywait = FALSE;
			simpose_mode_bak = 0;
		}
	}
#endif

	/* EVtgtO(Ver9.03/7.03Ŋg) */
#if XM7_VER >= 3
	if ((ver >= 903) || ((ver >= 703) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 703) {
#else
	if (ver >= 300) {
#endif
		if (!file_bool_read(fileh, &lshift_flag)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &rshift_flag)) {
			return FALSE;
		}
	}
	else {
		/* ۂ̏ԂƖ邪c */
		lshift_flag = FALSE;
		rshift_flag = FALSE;
	}

#if XM7_VER >= 2
	/* L[{[hGR[_ACK(Ver9.06/7.06Ŋg) */
#if XM7_VER >= 3
	if ((ver >= 906) || ((ver >= 706) && (ver <= 799))) {
#else
	if (ver >= 706) {
#endif
		if (!file_bool_read(fileh, &key_trans_ack)) {
			return FALSE;
		}
	}
	else {
		key_trans_ack = TRUE;
	}

	/* L[{[hGR[_BbZ[W(Ver9.19/7.19Ŋg) */
#if XM7_VER >= 3
	if ((ver >= 1401) || ((ver >= 1201) && (ver <= 1399)) ||
		((ver >= 919) && (ver <= 999)) || ((ver >= 719) && (ver <= 799))) {
#else
	if (((ver >= 1201) && (ver <= 1399)) || ((ver >= 719) && (ver <= 799))) {
#endif
		if (!file_bool_read(fileh, &key_hidden_message_flag)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &key_hidden_message_run)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &key_hidden_message_count)) {
			return FALSE;
		}
		if (model_type < MODEL_FM77AV) {
			key_hidden_message_flag = FALSE;
			key_hidden_message_run = FALSE;
			key_hidden_message_count = 0;
		}
	}
	else {
		key_hidden_message_flag = FALSE;
		key_hidden_message_run = FALSE;
		key_hidden_message_count = 0;
	}
#else
	if (((ver >= 310) && (ver <= 399)) ||
		((ver >= 1006) && (ver <= 1199))) {
		if (!file_word_read(fileh, &key_fm8)) {
			return FALSE;
		}
		if (!file_read(fileh, key_push_data, sizeof(key_push_data))) {
			return FALSE;
		}
	}
	else {
		/* ۂ̏ԂƖ邪c */
		key_fm8 = key_fm7;
		memset(key_push_data, 0, sizeof(key_push_data));
	}
#endif

#if XM7_VER >= 2
	/* Am[Np[NEL[GR[_BEEPtO͍~낷 */
	amnork_mode = 0;
	keyenc_beep_flag = FALSE;
	keyencbeep_notify();
#endif

	/* Cxg */
	schedule_delevent(EVENT_KEYBOARD);
#if XM7_VER >= 2
	schedule_delevent(EVENT_KEYENC_BEEP);
	schedule_handle(EVENT_KEYENC_ACK, key_send_ack);
	schedule_handle(EVENT_KEYENC_MESSAGE, keyenc_hidden_message_event);
#endif

	return TRUE;
}
