/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ CCPU ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "device.h"
#include "subctrl.h"
#include "ttlpalet.h"
#include "fdc.h"
#include "mainetc.h"
#include "multipag.h"
#include "kanji.h"
#include "tapelp.h"
#include "opn.h"
#include "mmr.h"
#include "apalet.h"
#include "rs232c.h"
#include "midi.h"
#include "jcard.h"
#include "dmac.h"
#include "jsubsys.h"
#include "bubble.h"
#include "ptm.h"
#include "mouse.h"

/*
 *	O[o [N
 */
BYTE *mainram_a;						/* RAM (\RAM)       $10000 */
BYTE *mainram_b;						/* RAM (RAM+)     $8000 */
BYTE *basic_rom;						/* ROM (F-BASIC)      $8000 */
BYTE *main_io;							/* CCPU I/O       $100 */
BOOL basicrom_en;						/* BASIC ROMCl[utO */
#if XM7_VER == 1
BYTE *basic_rom8;						/* ROM (F-BASIC V1.0) $8000 */
BYTE *boot_bas;							/* u[g(BASIC,FM-7)  $200 */
BYTE *boot_dos;							/* u[g(DOS,FM-7)    $200 */
BYTE *boot_bas8;						/* u[g(BASIC,FM-8)  $200 */
BYTE *boot_dos8;						/* u[g(DOS,FM-8)    $200 */
BYTE *boot_bbl8;						/* u[g(BUBBLE,FM-8) $200 */
BYTE *boot_bbl;							/* u[g(BUBBLE,FM-77)$200 */
BYTE *boot_sfd;							/* u[g(1MBDOS,FM-77)$200 */
BYTE *boot_dbg8;						/* u[g(DEBUG,FM-8)  $200 */
BYTE *boot_sfd8;						/* u[g(8"DOS,FM-8)  $200 */
#endif
BOOL bootram_rw;						/* u[gRAM ݃tO */
BOOL rom_ram_write;						/* FM-7[hROM̈揑ύX */

BYTE *extram_a;							/* gRAM           $10000 */
#if XM7_VER >= 3
BYTE *extram_c;							/* AV40gRAM       $C0000 */
#elif XM7_VER == 1
BYTE *extram_c;							/* gRAM           $C0000 */
#endif

#if XM7_VER >= 2
BYTE *init_rom;							/* CjVG[^ROM    $2000 */
BOOL initrom_en;						/* CjVG[^ROMCl[utO */
BOOL init_is_exsx;						/* CjVG[^ROM EX/SXtO */
#endif

/*
 *	X^eBbN [N
 */
#if XM7_VER >= 2
static BYTE *patch_branewboottfr;		/* Vu[g]pւBRA */
static BYTE *patch_jmpnewboot;			/* Vu[gւJMP */
#endif
static BYTE ioaccess_count;				/* I/ÖANZXJE^ */
static BOOL ioaccess_flag;				/* I/OANZXEFCgtO */
#if XM7_VER == 1
static BYTE *patch_bblboot8[7];
static BYTE *patch_bblboot[7];
#endif	/* XM7_VER == 1 */

/*
 *	vg^Cv錾
 */
#if XM7_VER == 1
static void FASTCALL mainmem_bubblemode_check(BYTE *patch[], BYTE *rom);
#endif	/* XM7_VER == 1 */

/*
 *	ID
 */
#define	ID	MAKEID('M', 'M', 'E', 'M')
//#define	MAINMEM_LOG

/*
 *	CCPU
 *	
 */
BOOL FASTCALL mainmem_init(void)
{
	int i;
	BYTE *p;
#if XM7_VER >= 3 || XM7_VER == 1
	BOOL basic_boot_flag;	/* basic_romboot_rom̑݃tO */
#endif
#if XM7_VER == 1
	int j;
#endif

	/* xASăNA */
	mainram_a = NULL;
	mainram_b = NULL;
	extram_a = NULL;
	main_io = NULL;
#if XM7_VER >= 3
	extram_c = NULL;
#elif XM7_VER == 1
	extram_c = NULL;
#endif
	basic_rom = NULL;
#if XM7_VER == 1
	boot_bas = NULL;
	boot_dos = NULL;
	boot_bas8 = NULL;
	boot_dos8 = NULL;
	boot_bbl8 = NULL;
	boot_bbl = NULL;
	boot_sfd = NULL;
	boot_dbg8 = NULL;
	boot_sfd8 = NULL;
#else
	patch_branewboottfr = NULL;
	patch_jmpnewboot = NULL;
	init_rom = NULL;
#endif
	rom_ram_write = FALSE;	/* EDPLAYWHGtɕQ̂ */

	/* RAM */
	mainram_a = (BYTE *)malloc(0x10000);
	if (mainram_a == NULL) {
		return FALSE;
	}
	memset(mainram_a, 0, 0x10000);
	mainram_b = (BYTE *)(&mainram_a[0x8000]);
#if XM7_VER >= 2
	extram_a = (BYTE *)malloc(0x10000);
	if (extram_a == NULL) {
		return FALSE;
	}
	memset(extram_a, 0, 0x10000);
#if XM7_VER >= 3
	extram_c = (BYTE *)malloc(0xc0000);
	if (extram_c == NULL) {
		return FALSE;
	}
	memset(extram_c, 0, 0xc0000);
#endif
#else
	/* V1 (400CZbgC192KB) */
	extram_a = (BYTE *)malloc(0x30000);
	if (extram_a == NULL) {
		return FALSE;
	}
	memset(extram_a, 0, 0x30000);
	extram_c = (BYTE *)malloc(0xc0000);
	if (extram_c == NULL) {
		return FALSE;
	}
	memset(extram_c, 0, 0xc0000);
#endif

	/* I/O */
	main_io = (BYTE *)malloc(0x0100);
	if (main_io == NULL) {
		return FALSE;
	}

	/* BASIC ROM */
	basic_rom = (BYTE *)malloc(0x8000);
	if (basic_rom == NULL) {
		return FALSE;
	}
	memset(basic_rom, 0, 0x8000);
#if XM7_VER == 1
	basic_rom8 = (BYTE *)malloc(0x8000);
	if (basic_rom8 == NULL) {
		return FALSE;
	}
	memset(basic_rom8, 0, 0x8000);
#endif

#if XM7_VER >= 2
	/* CjVG[^ROM */
	init_rom = (BYTE *)malloc(0x2000);
	if (init_rom == NULL) {
		return FALSE;
	}
	memset(init_rom, 0, 0x2000);
#endif

	/* u[gROM/RAM */
#if XM7_VER == 1
	boot_bas = (BYTE *)malloc(0x0800);
	if (boot_bas == NULL) {
		return FALSE;
	}
	memset(boot_bas, 0, 0x0800);
	boot_bbl = (BYTE *)(&boot_bas[0x0200]);
	boot_dos = (BYTE *)(&boot_bas[0x0400]);
	boot_sfd = (BYTE *)(&boot_bas[0x0600]);
	boot_bas8 = (BYTE *)malloc(0x0a00);
	if (boot_bas8 == NULL) {
		return FALSE;
	}
	memset(boot_bas8, 0, 0x0a00);
	boot_bbl8 = (BYTE *)(&boot_bas8[0x0200]);
	boot_dos8 = (BYTE *)(&boot_bas8[0x0400]);
	boot_dbg8 = (BYTE *)(&boot_bas8[0x0600]);
	boot_sfd8 = (BYTE *)(&boot_bas8[0x0800]);
#endif

	/* ROMt@Cǂݍ */
#if XM7_VER >= 3 || XM7_VER == 1
	basic_boot_flag = TRUE;
#endif
	if (!file_load(FBASIC_ROM, basic_rom, 0x8000)) {
		if (!file_load(FBASIC_ROM, basic_rom, 0x7c00)) {
#if XM7_VER == 1
			available_fm7roms = FALSE;
#else
			return FALSE;
#endif
		}
#if XM7_VER >= 3 || XM7_VER == 1
		if (!file_load(BOOTHDN_ROM, &basic_rom[0x7e00], 0x200)) {
			basic_boot_flag = FALSE;
		}
#endif
	}
#if XM7_VER == 1
	if (!file_load(FBASIC10_ROM, basic_rom8, 0x7c00)) {
		available_fm8roms = FALSE;
	}
#endif
#if XM7_VER >= 2
	if (!file_load(INITIATE_ROM, init_rom, 0x2000)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (!basic_boot_flag) {
		memcpy(&basic_rom[0x7e00], &init_rom[0x1800], 0x1e0);
		memset(&basic_rom[0x7fe0], 0, 30);
		basic_rom[0x7ffe] = 0xfe;
		basic_rom[0x7fff] = 0x00;
	}
#endif

	/* EX/SX̃CjVG[^ROM`FbN */
	if (init_rom[0xb10] == '1') {
		init_is_exsx = TRUE;
	}
	else {
		init_is_exsx = FALSE;
	}

	/* pb` */
	if (unpatch_rom_flag) {
		return TRUE;
	}

	/* N(AVV[Y) */
	for (i=0; i<2; i++) {
		p = &init_rom[0x1800 + i * 0x200];
		if (p[0x14f] == 0x26) {
			p[0x14f] = 0x21;
		}
		if (p[0x153] == 0x26) {
			p[0x153] = 0x21;
		}
	}

	p = &init_rom[0x1c00];

	/* N(AV20/40) */
	if ((p[0x74] == 0x10) && (p[0x75] == 0x23)) {
		if (p[0xf3] == 0x27) {
			p[0xf3] = 0x21;
		}
		if (p[0x196] == 0x26) {
			p[0x196] = 0x21;
		}
	}

	/* N(AV20EX/40EX/40SX) */
	if ((p[0x8f] == 0x10) && (p[0x90] == 0x23)) {
		if (p[0x166] == 0x27) {
			p[0x166] = 0x21;
		}
		if (p[0x1d6] == 0x26) {
			p[0x1d6] = 0x21;
		}
	}

	/* Vu[g֘ÃAhXEL */
	for (i=0; i<0xb00; i++) {
		/* Vu[g]ւBRA */
		if (!patch_branewboottfr) {
			if ((init_rom[i + 0] == 0x20) && (init_rom[i + 1] == 0xd7)) {
				patch_branewboottfr = &init_rom[i];
			}
		}

		/* Vu[gւJMP(CjVG[^̖ɑ݂) */
		if (!patch_jmpnewboot) {
			if ((init_rom[i + 0] == 0x7e) && (init_rom[i + 1] == 0x50) &&
				(init_rom[i + 2] == 0x00)) {
				patch_jmpnewboot = &init_rom[i];
			}
		}

		/* 烋[v𔲂 */
		if (patch_branewboottfr && patch_jmpnewboot) {
			break;
		}
	}
#else
	if (!file_load(BOOTBAS_ROM, boot_bas, 0x0200)) {
		available_fm7roms = FALSE;
	}
	if (!basic_boot_flag) {
		memcpy(&basic_rom[0x7e00], boot_bas, 0x01e0);
		memset(&basic_rom[0x7fe0], 0, 30);
		basic_rom[0x7ffe] = 0xfe;
		basic_rom[0x7fff] = 0x00;
	}
	if (!file_load(BOOTDOS_ROM, boot_dos, 0x0200)) {
		available_fm7roms = FALSE;
	}
	if (!file_load(BOOTBAS8_ROM, boot_bas8, 0x0200)) {
		available_fm8roms = FALSE;
	}
	if (!file_load(BOOTDOS8_ROM, boot_dos8, 0x0200)) {
		available_fm8roms = FALSE;
	}
	if (!file_load(BOOTBBL8_ROM, boot_bbl8, 0x0200)) {
		available_bblboot8 = FALSE;
	}
	else {
		mainmem_bubblemode_check(patch_bblboot8, boot_bbl8);
	}
	if (!file_load(BOOTSFD_ROM, boot_sfd, 0x0200)) {
		available_sfdboot = FALSE;
	}
	if (!file_load(BOOTBBL_ROM, boot_bbl, 0x0200)) {
		available_bblboot = FALSE;
	}
	else {
		mainmem_bubblemode_check(patch_bblboot, boot_bbl);
	}
	if (!file_load(BOOTSFD8_ROM, boot_sfd8, 0x0200)) {
		available_sfdboot8 = FALSE;
	}
	if (!file_load(BOOTDBG8_ROM, boot_dbg8, 0x0200)) {
		available_dbgboot8 = FALSE;
	}

	/* pb` */
	if (!unpatch_rom_flag) {
		/* u[gN */
		for (i=0; i<9; i++) {
			if (i < 4) {
				/* FM-7/NEW7/77 */
				p = &boot_bas[(i - 0) * 0x0200];
			}
			else {
				/* FM-8 */
				p = &boot_bas8[(i - 4) * 0x0200];
			}

			/* BOOT ROMEFCg[`ĎKp */
			for (j = 0; j <= 0x01dc; j++) {
				if ((p[j + 0] == 0x31) && (p[j + 1] == 0x3f) &&
					(p[j + 2] == 0x26) && (p[j + 3] == 0xfc)) {
					p[j + 2] = 0x21;
					break;
				}
			}
			p[0x1fe] = 0xfe;
			p[0x1ff] = 0x00;
		}
	}

	/* FM-7/FM-8ǂROMsSȏꍇAsƂ */
	if (!available_fm8roms && !available_fm7roms) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	CCPU
 *	N[Abv
 */
void FASTCALL mainmem_cleanup(void)
{
#if 0
	ASSERT(mainram_a);
	ASSERT(mainram_b);
	ASSERT(extram_a);
#if XM7_VER >= 3
	ASSERT(extram_c);
#elif XM7_VER == 1
	ASSERT(extram_c);
#endif
	ASSERT(main_io);
	ASSERT(basic_rom);
#if XM7_VER >= 2
	ASSERT(init_rom);
#else
	ASSERT(basic_rom8);
	ASSERT(boot_bas);
	ASSERT(boot_dos);
	ASSERT(boot_bas8);
	ASSERT(boot_dos8);
	ASSERT(boot_bbl8);
	ASSERT(boot_sfd);
	ASSERT(boot_bbl);
	ASSERT(boot_sfd8);
	ASSERT(boot_dbg8);
#endif
#endif

	/* rŎsꍇl */
	if (mainram_a) {
		free(mainram_a);
	}
	if (extram_a) {
		free(extram_a);
	}
#if XM7_VER >= 3
	if (extram_c) {
		free(extram_c);
	}
#elif XM7_VER == 1
	if (extram_c) {
		free(extram_c);
	}
#endif

	if (main_io) {
		free(main_io);
	}

	if (basic_rom) {
		free(basic_rom);
	}
#if XM7_VER >= 2
	if (init_rom) {
		free(init_rom);
	}
#else
	if (basic_rom8) {
		free(basic_rom8);
	}
	if (boot_bas) {
		free(boot_bas);
	}
	if (boot_bas8) {
		free(boot_bas8);
	}
#endif
}

/*
 *	CCPU
 *	Zbg
 */
void FASTCALL mainmem_reset(void)
{
	LOG0(Normal, "Zbg");

	/* I/OԁEI/OANZXJE^ */
	memset(main_io, 0xff, 0x0100);
	ioaccess_count = 0;
	ioaccess_flag = FALSE;

#if XM7_VER == 1
	switch (model_type) {
		case MODEL_FM8:
			if (!available_dbgboot8 &&
				boot_mode == BOOT_BASIC_DBG) {
				/* BASIC(DEBUG)[hBASIC[h */
				boot_mode = BOOT_BASIC;
			}
			else if (!available_bblboot8 &&
					 (boot_mode == BOOT_BUBBLE ||
					  boot_mode == BOOT_BUBBLE128)) {
				/* ou[hDOS[h */
				boot_mode = BOOT_DOS;
			}
			else if (!available_sfdboot8 &&
					 boot_mode == BOOT_DOS_SFD) {
				/* SFD[hDOS[h */
				boot_mode = BOOT_DOS;
			}
			break;
		case MODEL_FM7:
			if (boot_mode == BOOT_BASIC_DBG) {
				/* BASIC(DEBUG)[hBASIC[h */
				boot_mode = BOOT_BASIC;
			}
			else if (available_bblboot &&
					 (boot_mode == BOOT_BUBBLE ||
					  boot_mode == BOOT_BUBBLE128)) {
				/* ou[h̓ou[h */
				boot_mode = BOOT_BUBBLE;
			}
			else if (boot_mode != BOOT_BASIC) {
				/* BASIC[hȊODOS[h */
				boot_mode = BOOT_DOS;
			}
			break;
		case MODEL_FM77:
		default:
			if (!available_bblboot &&
				(boot_mode == BOOT_BUBBLE ||
				 boot_mode == BOOT_BUBBLE128)) {
				/* ou[hBASIC[h */
				boot_mode = BOOT_DOS;
			}
			else if (!available_sfdboot &&
					 boot_mode == BOOT_DOS_SFD) {
				/* SFD[hBASIC[h */
				boot_mode = BOOT_BASIC;
			}
			break;
	}
#endif	/* XM7_VER == 1 */

	/* BASIC[hł΁AF-BASIC ROMCl[u */
	if (boot_mode == BOOT_BASIC) {
		basicrom_en = TRUE;
	}
	else {
		basicrom_en = FALSE;
	}

#if XM7_VER >= 2
	if (model_type >= MODEL_FM77AV) {
		/* AV/40EX[h : CjVG[^ONAu[gRAM݉ */
		initrom_en = TRUE;
		bootram_rw = TRUE;
	}
	else {
		/* FM-7[h : CjVG[^OFFAu[gRAM݋֎~ */
		initrom_en = FALSE;
		bootram_rw = FALSE;
	}

	/* FM-7[hɂ̓u[gRAM̓e] */
	mainmem_transfer_boot();

	/* pb` */
	if (unpatch_rom_flag) {
		return;
	}

	/* CjVG[^ROM n[hEFAo[W */
	switch (model_type) {
		case MODEL_FM77AV:
			/* FM77AV */
			memset(&init_rom[0x0b0e], 0xff, 6);
			break;

		case MODEL_FM77AV20:
			/* FM77AV20 */
			init_rom[0x0b0e] = '2';
			init_rom[0x0b0f] = '0';
			init_rom[0x0b10] = '0';
			init_rom[0x0b11] = 'M';
			init_rom[0x0b12] = 'a';
			init_rom[0x0b13] = '.';
			break;

		case MODEL_FM77AV20EX:
			/* FM77AV20EX */
			init_rom[0x0b0e] = '2';
			init_rom[0x0b0f] = '0';
			init_rom[0x0b10] = '1';
			init_rom[0x0b11] = 'M';
			init_rom[0x0b12] = 'a';
			init_rom[0x0b13] = '.';
			break;

#if XM7_VER >= 3
		case MODEL_FM77AV40:
			/* FM77AV40 */
			init_rom[0x0b0e] = '4';
			init_rom[0x0b0f] = '0';
			init_rom[0x0b10] = '0';
			init_rom[0x0b11] = 'M';
			init_rom[0x0b12] = 'a';
			init_rom[0x0b13] = '.';
			break;

		case MODEL_FM77AV40EX:
			/* FM77AV40EX */
			init_rom[0x0b0e] = '4';
			init_rom[0x0b0f] = '0';
			init_rom[0x0b10] = '1';
			init_rom[0x0b11] = 'M';
			init_rom[0x0b12] = 'a';
			init_rom[0x0b13] = '.';
			break;
#endif

		default:
			return;
	}

	if (patch_branewboottfr && patch_jmpnewboot) {
		if (model_type == MODEL_FM77AV) {
			/* CjVG[^FM77AV̓ƂȂ悤Ƀpb` */
			patch_branewboottfr[0x0000] = 0x21;
			patch_jmpnewboot[0x0001] = 0xfe;
			patch_jmpnewboot[0x0002] = 0x00;
		}
		else {
			/* CjVG[^{̓ƂȂ悤ɖ߂ */
			patch_branewboottfr[0x0000] = 0x20;
			patch_jmpnewboot[0x0001] = 0x50;
			patch_jmpnewboot[0x0002] = 0x00;
		}
	}
#else
	/* u[g̈揑݋֎~ */
	bootram_rw = FALSE;
#endif
}

/*
 *	FM-7[hp u[gRAM]
 */
#if XM7_VER >= 2
void FASTCALL mainmem_transfer_boot(void)
{
	if (model_type < MODEL_FM77AV) {
		if (boot_mode == BOOT_BASIC) {
			memcpy(&mainram_b[0x7e00], &init_rom[0x1800], 0x1e0);
		}
		else {
			memcpy(&mainram_b[0x7e00], &init_rom[0x1a00], 0x1e0);
		}
		mainram_b[0x7ffe] = 0xfe;
		mainram_b[0x7fff] = 0x00;
	}
}
#endif

#if XM7_VER == 1
/*
 *	ou[hu[gROM`FbN
 */
static void FASTCALL mainmem_bubblemode_check(BYTE *patch[], BYTE *rom)
{
	int i, j, k;

	/* 32KB/128KBouJZbg֘ÃAhXEL */
	for (i = 0, j = 0; i < 0x1e0; i ++) {
		/* BSTAT */
		if (!patch[j]) {
			if ((rom[i + 0] == 0x96) &&
				((rom[i + 1] == 0x12) || (rom[i + 1] == 0xD2))) {
				patch[j] = &rom[i + 1];
				j++;
			}
		}
		/* BPGADH */
		if (!patch[3]) {
			if ((rom[i + 0] == 0x9F) &&
				((rom[i + 1] == 0x14) || (rom[i + 1] == 0xD4))) {
				patch[3] = &rom[i + 1];
			}
		}
		/* BPGCTH */
		if (!patch[4]) {
			if ((rom[i + 0] == 0x9F) &&
				((rom[i + 1] == 0x16) || (rom[i + 1] == 0xD6))) {
				patch[4] = &rom[i + 1];
			}
		}
		/* BCMD */
		if (!patch[5]) {
			if ((rom[i + 0] == 0x97) &&
				((rom[i + 1] == 0x11) || (rom[i + 1] == 0xD1))) {
				patch[5] = &rom[i + 1];
			}
		}
		/* BDATA */
		if (!patch[6]) {
			if ((rom[i + 0] == 0x96) &&
				((rom[i + 1] == 0x10) || (rom[i + 1] == 0xD0))) {
				patch[6] = &rom[i + 1];
			}
		}
		if ((patch[0] && patch[1] && patch[2] && patch[3] &&
			 patch[4] && patch[5] && patch[6]) || (j >= 3)) {
			break;
		}
	}
}

/*
 *	ou[hu[gROMpb`
 */
void FASTCALL mainmem_bubblemode_boot(void)
{
	/* pb` */
	if (unpatch_rom_flag) {
		return;
	}

	/* u[g[h */
	if (boot_mode == BOOT_BUBBLE) {
		/* 32KBou[h */
		if (patch_bblboot8[0] && patch_bblboot8[1] && patch_bblboot8[2] &&
			patch_bblboot8[3] && patch_bblboot8[4] && patch_bblboot8[5] &&
			patch_bblboot8[6]) {
			patch_bblboot8[0][0] = 0x12;
			patch_bblboot8[1][0] = 0x12;
			patch_bblboot8[2][0] = 0x12;
			patch_bblboot8[3][0] = 0x14;
			patch_bblboot8[4][0] = 0x16;
			patch_bblboot8[5][0] = 0x11;
			patch_bblboot8[6][0] = 0x10;
		}
		if (patch_bblboot[0] && patch_bblboot[1] && patch_bblboot[2] &&
			patch_bblboot[3] && patch_bblboot[4] && patch_bblboot[5] &&
			patch_bblboot[6]) {
			patch_bblboot[0][0] = 0x12;
			patch_bblboot[1][0] = 0x12;
			patch_bblboot[2][0] = 0x12;
			patch_bblboot[3][0] = 0x14;
			patch_bblboot[4][0] = 0x16;
			patch_bblboot[5][0] = 0x11;
			patch_bblboot[6][0] = 0x10;
		}
	}
	else if (boot_mode == BOOT_BUBBLE128) {
		/* 128KBou[h */
		if (patch_bblboot8[0] && patch_bblboot8[1] && patch_bblboot8[2] &&
			patch_bblboot8[3] && patch_bblboot8[4] && patch_bblboot8[5] &&
			patch_bblboot8[6]) {
			patch_bblboot8[0][0] = 0xD2;
			patch_bblboot8[1][0] = 0xD2;
			patch_bblboot8[2][0] = 0xD2;
			patch_bblboot8[3][0] = 0xD4;
			patch_bblboot8[4][0] = 0xD6;
			patch_bblboot8[5][0] = 0xD1;
			patch_bblboot8[6][0] = 0xD0;
		}
		if (patch_bblboot[0] && patch_bblboot[1] && patch_bblboot[2] &&
			patch_bblboot[3] && patch_bblboot[4] && patch_bblboot[5] &&
			patch_bblboot[6]) {
			patch_bblboot[0][0] = 0xD2;
			patch_bblboot[1][0] = 0xD2;
			patch_bblboot[2][0] = 0xD2;
			patch_bblboot[3][0] = 0xD4;
			patch_bblboot[4][0] = 0xD6;
			patch_bblboot[5][0] = 0xD1;
			patch_bblboot[6][0] = 0xD0;
		}
	}
}
#endif	/* XM7_VER == 1 */

/*
 *	I/OANZX̃fB
 */
static void FASTCALL mainmem_iowait(void)
{
	BYTE tmp;

#if XM7_VER == 1
	/* 1.2MHz[hłI/OEFCgȂ */
	if (lowspeed_mode) {
		return;
	}
#endif

	/* NbN΂ɕKvI/OANZX񐔂ݒ */
#if XM7_VER >= 2
	if ((mmr_flag || twr_flag) && !mmr_fastmode && ioaccess_flag) {
#else
	if ((mmr_flag || twr_flag) && ioaccess_flag) {
#endif
		/* MMRI(2)  3̃ANZX1TCN */
		tmp = 3;
	}
	else {
		/* MMRIt/[h/MMRI(1)  2̃ANZX1TCN */
		tmp = 2;
	}

	/* K񐔈ȏI/OANZX΁ANbN΂ */
	ioaccess_count ++;
	if (ioaccess_count >= tmp) {
		maincpu.total ++;
		ioaccess_count = 0;

		ioaccess_flag = !ioaccess_flag;
	}
}


/*
 *	CCPU
 *	PoCg擾
 */
BYTE FASTCALL mainmem_readb(WORD addr)
{
	BYTE dat;
#if XM7_VER >= 2
	WORD vaddr = addr;
#endif

	/* MMR, TWR`FbN */
#if XM7_VER == 1 && defined(Z80CARD)
	if ((mmr_flag || twr_flag) && !main_z80mode) {
#else
	if (mmr_flag || twr_flag) {
#endif
		/* MMRATWRʂ */
		if (mmr_extrb(&addr, &dat)) {
			return dat;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		return mainram_a[addr];
	}
	if (addr < 0x8000) {
		if (initrom_en) {
			return init_rom[addr - 0x6000];
		}
		else {
			return mainram_a[addr];
		}
	}
#else
	if (addr < 0x8000) {
		return mainram_a[addr];
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
#if XM7_VER == 1
			if (model_type == MODEL_FM8) {
				return basic_rom8[addr - 0x8000];
			}
#endif
			return basic_rom[addr - 0x8000];
		}
		else {
			return mainram_b[addr - 0x8000];
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		return mainram_b[addr - 0x8000];
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			return shared_ram[(WORD)(addr - 0xfc80)];
		}
		else {
#ifdef MAINMEM_LOG
#if XM7_VER >= 2
			if (vaddr != addr) {
				LOG2(Warning, "HALTLRAMւ̃oCgǂݍ $3%04X($%04X)", addr, vaddr);
			}
			else {
				LOG1(Warning, "HALTLRAMւ̃oCgǂݍ $%04X", addr);
			}
#else
			LOG1(Warning, "HALTLRAMւ̃oCgǂݍ $%04X", addr);
#endif
#endif
			return 0xff;
		}
	}

	/* u[gROM/RAM */
#if XM7_VER >= 2
	if ((addr >= 0xfffe) && (initrom_en)) {
		/* ZbgxN^ */
		mainmem_iowait();
		return init_rom[addr - 0xe000];
	}
#endif
	if (addr >= 0xfe00) {
		if (addr <= 0xffdf) {
			mainmem_iowait();
		}
#if XM7_VER >= 2
		return mainram_b[addr - 0x8000];
#else
		if (((addr >= 0xffe0) && (addr < 0xfffe)) ||
			((model_type == MODEL_FM77) && bootram_rw)) {
			/* FM-77 RAM */
			return mainram_b[addr - 0x8000];
		}
		else {
			if (model_type == MODEL_FM8) {
				/* FM-8 BOOT ROM */
				if (boot_mode == BOOT_BASIC) {
					return boot_bas8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS) {
					return boot_dos8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_BASIC_DBG) {
					return boot_dbg8[addr - 0xfe00];
				}
				else {
					return boot_bbl8[addr - 0xfe00];
				}
			}
			else {
				/* FM-7 BOOT ROM */
				if ((boot_mode == BOOT_BASIC) ||
					(boot_mode == BOOT_BASIC_DBG)) {
					return boot_bas[addr - 0xfe00];
				}
				else if ((boot_mode == BOOT_DOS)) {
					return boot_dos[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd[addr - 0xfe00];
				}
				else {
					return boot_bbl[addr - 0xfe00];
				}
			}
		}
#endif
	}

	/*
	 *	I/O
	 */
	mainmem_iowait();
	if (mainetc_readb(addr, &dat)) {
		return dat;
	}
	if (ttlpalet_readb(addr, &dat)) {
		return dat;
	}
	if (subctrl_readb(addr, &dat)) {
		return dat;
	}
	if (multipag_readb(addr, &dat)) {
		return dat;
	}
	if (fdc_readb(addr, &dat)) {
		return dat;
	}
#if XM7_VER == 1 && defined(SFDC)
	if (ifdc_readb(addr, &dat)) {
		return dat;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (kanji_readb(addr, &dat)) {
		return dat;
	}
	if (tapelp_readb(addr, &dat)) {
		return dat;
	}
	if (opn_readb(addr, &dat)) {
		return dat;
	}
	if (whg_readb(addr, &dat)) {
		return dat;
	}
	if (thg_readb(addr, &dat)) {
		return dat;
	}
	if (psg_readb(addr, &dat)) {
		return dat;
	}
	if (mmr_readb(addr, &dat)) {
		return dat;
	}
#if XM7_VER >= 2
	if (apalet_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef MIDI
	if (midi_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef RSC
	if (rs232c_readb(addr, &dat)) {
		return dat;
	}
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	if (dmac_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef JSUB
	if (jsub_readb(addr, &dat)) {
		return dat;
	}
#endif
#if XM7_VER == 1 && defined(BUBBLE)
	if (bmc_readb(addr, &dat)) {
		return dat;
	}
#endif
#ifdef MOUSE
	if (ptm_readb(addr, &dat)) {
		return dat;
	}
	if (mouse_readb(addr, &dat)) {
		return dat;
	}
#endif

#if XM7_VER >= 2
	if (vaddr != addr) {
		LOG2(Warning, "`oCgǂݍ $3%04X($%04X)", addr, vaddr);
	}
	else {
		LOG1(Warning, "`oCgǂݍ $%04X", addr);
	}
#else
	LOG1(Warning, "`oCgǂݍ $%04X", addr);
#endif

	return 0xff;
}

/*
 *	CCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL mainmem_readbnio(WORD addr)
{
	BYTE dat;

	/* MMR, TWR`FbN */
#if XM7_VER == 1 && defined(Z80CARD)
	if ((mmr_flag || twr_flag) && !main_z80mode) {
#else
	if (mmr_flag || twr_flag) {
#endif
		/* MMRATWRʂ */
		if (mmr_extbnio(&addr, &dat)) {
			return dat;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		return mainram_a[addr];
	}
	if (addr < 0x8000) {
		if (initrom_en) {
			return init_rom[addr - 0x6000];
		}
		else {
			return mainram_a[addr];
		}
	}
#else
	if (addr < 0x8000) {
		return mainram_a[addr];
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
#if XM7_VER == 1
			if (model_type == MODEL_FM8) {
				return basic_rom8[addr - 0x8000];
			}
#endif
			return basic_rom[addr - 0x8000];
		}
		else {
			return mainram_b[addr - 0x8000];
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		return mainram_b[addr - 0x8000];
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			return shared_ram[(WORD)(addr - 0xfc80)];
		}
		else {
			return 0xff;
		}
	}

	/* u[gROM/RAM */
#if XM7_VER >= 2
	if ((addr >= 0xfffe) && (initrom_en)) {
		/* ZbgxN^ */
		return init_rom[addr - 0xe000];
	}
#endif
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		return mainram_b[addr - 0x8000];
#else
		if (((addr >= 0xffe0) && (addr < 0xfffe)) ||
			((model_type == MODEL_FM77) && bootram_rw)) {
			return mainram_b[addr - 0x8000];
		}
		else {
			if (model_type == MODEL_FM8) {
				if (boot_mode == BOOT_BASIC) {
					return boot_bas8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS) {
					return boot_dos8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_BASIC_DBG) {
					return boot_dbg8[addr - 0xfe00];
				}
				else {
					return boot_bbl8[addr - 0xfe00];
				}
			}
			else {
				if ((boot_mode == BOOT_BASIC) ||
					(boot_mode == BOOT_BASIC_DBG)) {
					return boot_bas[addr - 0xfe00];
				}
				else if ((boot_mode == BOOT_DOS)) {
					return boot_dos[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd[addr - 0xfe00];
				}
				else {
					return boot_bbl[addr - 0xfe00];
				}
			}
		}
#endif
	}

	/* I/O */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	return main_io[addr - 0xfd00];
}

/*
 *	CCPU
 *	PoCg
 */
void FASTCALL mainmem_writeb(WORD addr, BYTE dat)
{
	WORD vaddr = addr;

	/* MMR, TWR`FbN */
#if XM7_VER == 1 && defined(Z80CARD)
	if ((mmr_flag || twr_flag) && !main_z80mode) {
#else
	if (mmr_flag || twr_flag) {
#endif
		/* MMRATWRʂ */
		if (mmr_extwb(&addr, dat)) {
			return;
		}
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		mainram_a[addr] = dat;
		return;
	}
	if (addr < 0x8000) {
		if (!initrom_en) {
			mainram_a[addr] = dat;
		}
		else {
			if (vaddr != addr) {
				LOG3(Warning, "CjVG[^ROMւ̃oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
			}
			else {
				LOG2(Warning, "CjVG[^ROMւ̃oCg $%04X <- $%02X", addr, dat);
			}
		}
		return;
	}
#else
	if (addr < 0x8000) {
		mainram_a[addr] = dat;
		return;
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
			/* ROMRCBBIOSĂяoP[X */
			if ((model_type == MODEL_FM7) && rom_ram_write) {
				/* FM-7łF-BASIC ROMIɂRAMɏ߂悤Ȃ̂ */
				mainram_b[addr - 0x8000] = dat;
			}
			if (vaddr != addr) {
				LOG3(Warning, "BASIC ROMւ̃oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
			}
			else {
				LOG2(Warning, "BASIC ROMւ̃oCg $%04X <- $%02X", addr, dat);
			}
			return;
		}
		else {
			mainram_b[addr - 0x8000] = dat;
			return;
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		mainram_b[addr - 0x8000] = dat;
		return;
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			shared_ram[(WORD)(addr - 0xfc80)] = dat;
			return;
		}
		else {
			/* BASE09΍ */
#ifdef MAINMEM_LOG
#if XM7_VER >= 2
			if (vaddr != addr) {
				LOG3(Warning, "HALTLRAMւ̃oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
			}
			else {
				LOG2(Warning, "HALTLRAMւ̃oCg $%04X <- $%02X", addr, dat);
			}
#else
			LOG2(Warning, "HALTLRAMւ̃oCg $%04X <- $%02X", addr, dat);
#endif
#endif
			return;
		}
	}

	/* u[gRAM */
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		if ((bootram_rw) && (model_type >= MODEL_FM77AV)) {
#else
		if ((model_type == MODEL_FM77) && bootram_rw) {
#endif
			mainram_b[addr - 0x8000] = dat;
			if ((addr <= 0xffdf) || (addr >= 0xfffe)) {
				mainmem_iowait();
			}
			return;
		}

		/* u[g[NRAMAxN^ */
		if ((addr >= 0xffe0) && (addr < 0xfffe)) {
			mainram_b[addr - 0x8000] = dat;
		}
		else {
#if XM7_VER >= 2
			if (model_type >= MODEL_FM77AV) {
				if (vaddr != addr) {
					LOG3(Warning, "Ǎp̃u[gRAMւ̃oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
				}
				else {
					LOG2(Warning, "Ǎp̃u[gRAMւ̃oCg $%04X <- $%02X", addr, dat);
				}
			}
			else {
				if (vaddr != addr) {
					LOG3(Warning, "u[gROMւ̃oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
				}
				else {
					LOG2(Warning, "u[gROMւ̃oCg $%04X <- $%02X", addr, dat);
				}
			}
#else
			LOG2(Warning, "u[gROMւ̃oCg $%04X <- $%02X", addr, dat);
#endif
		}
		return;
	}

	/*
	 *	I/O
	 */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	main_io[(WORD)(addr - 0xfd00)] = dat;
	mainmem_iowait();

	if (mainetc_writeb(addr, dat)) {
		return;
	}
	if (ttlpalet_writeb(addr, dat)) {
		return;
	}
	if (subctrl_writeb(addr, dat)) {
		return;
	}
	if (multipag_writeb(addr, dat)) {
		return;
	}
	if (fdc_writeb(addr, dat)) {
		return;
	}
#if XM7_VER == 1 && defined(SFDC)
	if (ifdc_writeb(addr, dat)) {
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (kanji_writeb(addr, dat)) {
		return;
	}
	if (tapelp_writeb(addr, dat)) {
		return;
	}
	if (opn_writeb(addr, dat)) {
		return;
	}
	if (whg_writeb(addr, dat)) {
		return;
	}
	if (thg_writeb(addr, dat)) {
		return;
	}
	if (psg_writeb(addr, dat)) {
		return;
	}
	if (mmr_writeb(addr, dat)) {
		return;
	}
#if XM7_VER >= 2
	if (apalet_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef MIDI
	if (midi_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef RSC
	if (rs232c_writeb(addr, dat)) {
		return;
	}
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	if (dmac_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef JSUB
	if (jsub_writeb(addr, dat)) {
		return;
	}
#endif
#if XM7_VER == 1 && defined(BUBBLE)
	if (bmc_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef MOUSE
	if (ptm_writeb(addr, dat)) {
		return;
	}
	if (mouse_writeb(addr, dat)) {
		return;
	}
#endif

#if XM7_VER >= 2
	if (vaddr != addr) {
		LOG3(Warning, "`oCg $3%04X($%04X) <- $%02X", addr, vaddr, dat);
	}
	else {
		LOG2(Warning, "`oCg $%04X <- $%02X", addr, dat);
	}
#else
	LOG2(Warning, "`oCg $%04X <- $%02X", addr, dat);
#endif

	return;
}

/*
 *	CCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL mainmem_readbsegnio(WORD addr, BYTE seg, int mode)
{
	BYTE dat;

	/* MMRATWRʂ */
	if (mmr_extrbsegnio(&addr, &dat, seg, mode)) {
		return dat;
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		return mainram_a[addr];
	}
	if (addr < 0x8000) {
		if (initrom_en) {
			return init_rom[addr - 0x6000];
		}
		else {
			return mainram_a[addr];
		}
	}
#else
	if (addr < 0x8000) {
		return mainram_a[addr];
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
#if XM7_VER == 1
			if (model_type == MODEL_FM8) {
				return basic_rom8[addr - 0x8000];
			}
#endif
			return basic_rom[addr - 0x8000];
		}
		else {
			return mainram_b[addr - 0x8000];
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		return mainram_b[addr - 0x8000];
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			return shared_ram[(WORD)(addr - 0xfc80)];
		}
		else {
			return 0xff;
		}
	}

	/* u[gROM/RAM */
#if XM7_VER >= 2
	if ((addr >= 0xfffe) && (initrom_en)) {
		/* ZbgxN^ */
		return init_rom[addr - 0xe000];
	}
#endif
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		return mainram_b[addr - 0x8000];
#else
		if (((addr >= 0xffe0) && (addr < 0xfffe)) ||
			((model_type == MODEL_FM77) && bootram_rw)) {
			return mainram_b[addr - 0x8000];
		}
		else {
			if (model_type == MODEL_FM8) {
				if (boot_mode == BOOT_BASIC) {
					return boot_bas8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS) {
					return boot_dos8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd8[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_BASIC_DBG) {
					return boot_dbg8[addr - 0xfe00];
				}
				else {
					return boot_bbl8[addr - 0xfe00];
				}
			}
			else {
				if ((boot_mode == BOOT_BASIC) ||
					(boot_mode == BOOT_BASIC_DBG)) {
					return boot_bas[addr - 0xfe00];
				}
				else if ((boot_mode == BOOT_DOS)) {
					return boot_dos[addr - 0xfe00];
				}
				else if (boot_mode == BOOT_DOS_SFD) {
					return boot_sfd[addr - 0xfe00];
				}
				else {
					return boot_bbl[addr - 0xfe00];
				}
			}
		}
#endif
	}

	/* I/O */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	return main_io[addr - 0xfd00];
}

/*
 *	CCPU
 *	PoCg
 */
void FASTCALL mainmem_writebseg(WORD addr, BYTE dat, BYTE seg, int mode)
{
	/* MMRATWRʂ */
	if (mmr_extwbseg(&addr, dat, seg, mode)) {
		return;
	}

	/* CRAM(\) */
#if XM7_VER >= 2
	if (addr < 0x6000) {
		mainram_a[addr] = dat;
		return;
	}
	if (addr < 0x8000) {
		if (!initrom_en) {
			mainram_a[addr] = dat;
		}
		return;
	}
#else
	if (addr < 0x8000) {
		mainram_a[addr] = dat;
		return;
	}
#endif

	/* BASIC ROM or CRAM() */
	if (addr < 0xfc00) {
		if (basicrom_en) {
			/* ROMRCBBIOSĂяoP[X */
			return;
		}
		else {
			mainram_b[addr - 0x8000] = dat;
			return;
		}
	}

	/* CROM̒ */
	if (addr < 0xfc80) {
		mainram_b[addr - 0x8000] = dat;
		return;
	}

	/* LRAM */
	if (addr < 0xfd00) {
		if (subhalt_flag) {
			shared_ram[(WORD)(addr - 0xfc80)] = dat;
			return;
		}
		else {
			/* BASE09΍ */
			return;
		}
	}

	/* u[gRAM */
	if (addr >= 0xfe00) {
#if XM7_VER >= 2
		if ((bootram_rw) && (model_type >= MODEL_FM77AV)) {
#else
		if ((model_type == MODEL_FM77) && bootram_rw) {
#endif
			mainram_b[addr - 0x8000] = dat;
			return;
		}

		/* u[g[NRAMAxN^ */
		if ((addr >= 0xffe0) && (addr < 0xfffe)) {
			mainram_b[addr - 0x8000] = dat;
		}
		return;
	}

	/*
	 *	I/O
	 */
	ASSERT((addr >= 0xfd00) && (addr < 0xfe00));
	main_io[(WORD)(addr - 0xfd00)] = dat;

	if (mainetc_writeb(addr, dat)) {
		return;
	}
	if (ttlpalet_writeb(addr, dat)) {
		return;
	}
	if (subctrl_writeb(addr, dat)) {
		return;
	}
	if (multipag_writeb(addr, dat)) {
		return;
	}
	if (fdc_writeb(addr, dat)) {
		return;
	}
#if XM7_VER == 1 && defined(SFDC)
	if (ifdc_writeb(addr, dat)) {
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (kanji_writeb(addr, dat)) {
		return;
	}
	if (tapelp_writeb(addr, dat)) {
		return;
	}
	if (opn_writeb(addr, dat)) {
		return;
	}
	if (whg_writeb(addr, dat)) {
		return;
	}
	if (thg_writeb(addr, dat)) {
		return;
	}
	if (mmr_writeb(addr, dat)) {
		return;
	}
#if XM7_VER >= 2
	if (apalet_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef MIDI
	if (midi_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef RSC
	if (rs232c_writeb(addr, dat)) {
		return;
	}
#endif
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(SFDC))
	if (dmac_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef JSUB
	if (jsub_writeb(addr, dat)) {
		return;
	}
#endif
#if (XM7_VER == 1 && defined(BUBBLE))
	if (bmc_writeb(addr, dat)) {
		return;
	}
#endif
#ifdef MOUSE
	if (ptm_writeb(addr, dat)) {
		return;
	}
	if (mouse_writeb(addr, dat)) {
		return;
	}
#endif

	return;
}

/*
 *	CCPU
 *	Z[u
 */
BOOL FASTCALL mainmem_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_write(fileh, mainram_a, 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, mainram_b, 0x7e00)) {
		return FALSE;
	}

	if (!file_write(fileh, main_io, 0x100)) {
		return FALSE;
	}

	if (!file_write(fileh, extram_a, 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x8000], 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_write(fileh, &extram_a[0x10000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x18000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x20000], 0x8000)) {
		return FALSE;
	}
	if (!file_write(fileh, &extram_a[0x28000], 0x8000)) {
		return FALSE;
	}
#endif
	if (!file_write(fileh, &mainram_b[0x7e00], 0x200)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, basicrom_en)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_bool_write(fileh, initrom_en)) {
		return FALSE;
	}
#endif
	if (!file_bool_write(fileh, bootram_rw)) {
		return FALSE;
	}

#if XM7_VER >= 2
	/* Ver8g */
	if (!file_write(fileh, &init_rom[0x0b0e], 3)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	CCPU
 *	[h
 */
BOOL FASTCALL mainmem_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_read(fileh, mainram_a, 0x8000)) {
		return FALSE;
	}
#if XM7_VER >= 3
	/* Ver1409g */
	if (((ver >= 1209) && (ver <= 1399)) || (ver >= 1409)) {
#elif XM7_VER >= 2
	/* Ver1209g */
	if (ver >= 1209) {
#else
	/* Ver302g */
	if (ver >= 302) {
#endif
		if (!file_read(fileh, mainram_b, 0x7e00)) {
			return FALSE;
		}
	}
	else {
		if (!file_read(fileh, mainram_b, 0x7c80)) {
			return FALSE;
		}
		memset(&mainram_b[0x7c80], 0, 0x180);
	}

	if (!file_read(fileh, main_io, 0x100)) {
		return FALSE;
	}

	if (!file_read(fileh, extram_a, 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x8000], 0x8000)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_read(fileh, &extram_a[0x10000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x18000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x20000], 0x8000)) {
		return FALSE;
	}
	if (!file_read(fileh, &extram_a[0x28000], 0x8000)) {
		return FALSE;
	}
#endif
	if (!file_read(fileh, &mainram_b[0x7e00], 0x200)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &basicrom_en)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_bool_read(fileh, &initrom_en)) {
		return FALSE;
	}
#endif
	if (!file_bool_read(fileh, &bootram_rw)) {
		return FALSE;
	}

#if XM7_VER >= 2
	/* Ver8g */
	if (ver >= 800) {
		if (!file_read(fileh, &init_rom[0x0b0e], 3)) {
			return FALSE;
		}
#if 0
		if (init_rom[0x0b0e] != 0xff) {
			init_rom[0x0b11] = 'M';
			init_rom[0x0b12] = 'a';
			init_rom[0x0b13] = '.';
		}
		else {
			memset(&init_rom[0x0b11], 0xff, 3);
		}
#endif
	}
#endif

#if XM7_VER == 1
	if (ver == 403) {
		if (!file_byte_read(fileh, &banksel_en)) {
			return FALSE;
		}
		if (banksel_en != BANKSEL_DISABLE) {
			banksel_en = BANKSEL_ENABLE;
		}
	}
#endif

	return TRUE;
}
