/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ MMR,TWR / I/O^RAMfBXNJ[h ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "device.h"
#include "mmr.h"
#include "subctrl.h"
#include "jcard.h"
#include "rs232c.h"

/*
 *	
 */
#ifdef RAMDISK
#define	MR2
#define	RD512
#define	R2D2
#endif

/*
 *	O[o [N
 */
BOOL mmr_flag;							/* MMRLtO */
BYTE mmr_seg;							/* MMRZOg */
BOOL mmr_modify;						/* MMRԕύXtO */
#if XM7_VER >= 2
BYTE mmr_reg[0x80];						/* MMRWX^ */
BOOL mmr_ext;							/* gMMRLtO */
BOOL mmr_fastmode;						/* MMRtO */
BOOL mmr_fast_refresh;					/* tbVtO */
BOOL twr_fastmode;						/* TWR[h */
#if XM7_VER >= 3
BYTE mmr_extram;						/* gRAM[h */
BOOL dsp_400linetiming;					/* 400C^C~Oo̓tO */
#endif
#else
BYTE mmr_reg[0x400];					/* MMRWX^ */
BYTE mmr_extram;						/* gRAM[h */
#endif
BOOL twr_flag;							/* TWRLtO */
BYTE twr_reg;							/* TWRWX^ */
BYTE wbr_reg;							/* WBRWX^ */
BYTE bsr_reg;							/* gRAMoNZNgWX^ */

/* I/O^RAMfBXNJ[h */
#if XM7_VER >= 3
#ifdef MR2
BYTE mr2_nowcnt;						/* MR2 ZN^JEg */
WORD mr2_secreg;						/* MR2 ZN^WX^ */
#endif
#ifdef RD512
WORD rd512_secreg;						/* RD512 ZN^WX^ */
#endif
#ifdef R2D2
BYTE r2d2_nowcnt;						/* R2D2 ZN^JEg */
WORD r2d2_secreg;						/* R2D2 ZN^WX^ */
#endif
#endif

/*
 *	ID
 */
#define	ID	MAKEID('M', 'M', 'R', ' ')
//#define MMR_LOG

/*
 *	MMR
 *	
 */
BOOL FASTCALL mmr_init(void)
{
	/* N̂MMRNA悤ɕύX */
	memset(mmr_reg, 0, sizeof(mmr_reg));

#if XM7_VER >= 3
	/* 768KB/256KB gRAMJ[hfBZ[u */
	mmr_extram = EXRAM_OFF;

	/* 400C^C~Oo͂fBZ[u */
	dsp_400linetiming = FALSE;
#elif XM7_VER == 1
	/* 192KB gRAMJ[hCl[u */
	mmr_extram = EXRAM_192KB;
#endif

	return TRUE;
}

/*
 *	MMR
 *	N[Abv
 */
void FASTCALL mmr_cleanup(void)
{
}

/*
 *	MMR
 *	Zbg
 */
void FASTCALL mmr_reset(void)
{
	LOG0(Normal, "Zbg");

	/* MMR/TWR */
	mmr_flag = FALSE;
	twr_flag = FALSE;
	mmr_seg = 0;
	twr_reg = 0;
	mmr_modify = FALSE;

	/* MMR/TWR */
#if XM7_VER >= 2
	mmr_ext = FALSE;
	mmr_fastmode = FALSE;
	mmr_fast_refresh = FALSE;
	twr_fastmode = FALSE;
	wbr_reg = 0;

#if XM7_VER >= 3
	/* I/O^RAMfBXN */
#ifdef MR2
	mr2_nowcnt = 0;
	mr2_secreg = 0;
#endif
#ifdef RD512
	rd512_secreg = 0;
#endif
#ifdef R2D2
	r2d2_nowcnt = 0;
	r2d2_secreg = 0;
#endif
#endif
#else
	wbr_reg = 2;
	bsr_reg = 2;
	if ((mmr_extram == EXRAM_960KB) && (init_flag)) {
		/* N1MB[ĥMMRݒ	*/
		/* F-BASIC V3.3L1xIPLɑ΂鉼΍			*/
		int i;
		for (i = 0; i < 16; i ++) {
			mmr_reg[i] = 0x30 | i;
		}
	}
#endif
}

/*-[ }l[W ]-----------------------------------------------------*/

/*
 *	TWRAhXϊ
 */
BOOL FASTCALL mmr_trans_twr(WORD addr, DWORD *taddr)
{
#if XM7_VER >= 2
	ASSERT(model_type >= MODEL_FM77AV);
#else
	ASSERT(model_type == MODEL_FM77);
#endif

	/* TWRL */
	if (!twr_flag) {
		return FALSE;
	}

	/* AhXv`FbN */
	if ((addr < 0x7c00) || (addr > 0x7fff)) {
		return FALSE;
	}

	/* TWRWX^ϊ */
	*taddr = (DWORD)twr_reg;
	*taddr *= 256;
	*taddr += addr;
	*taddr &= 0xffff;
	*taddr |= (wbr_reg << 16);

	return TRUE;
}

/*
 *	MMR
 *	AhXϊ
 */
DWORD FASTCALL mmr_trans_mmr(WORD addr)
{
	DWORD maddr;
	int offset;

#if XM7_VER >= 2
	ASSERT(model_type >= MODEL_FM77AV);
#else
	ASSERT(model_type == MODEL_FM77);
#endif

	/* MMRL */
	if (!mmr_flag) {
		return (DWORD)(0x30000 | addr);
	}

	/* MMRWX^擾 */
	offset = (int)addr;
	offset >>= 12;

#if XM7_VER >= 2
	/* gMMRoffȂAZOg0`3܂ */
	if (mmr_ext) {
		offset |= (mmr_seg * 0x10);
	}
	else {
		offset |= ((mmr_seg & 0x03) * 0x10);
	}
#else
	/* gMMRoffȂAZOg0`3܂ */
	if (mmr_extram == EXRAM_960KB) {
		offset |= (mmr_seg * 0x10);
	}
	else {
		offset |= ((mmr_seg & 0x03) * 0x10);
	}
#endif

	/* gMMRoffȂA6bit̂ݗL */
	maddr = (DWORD)mmr_reg[offset];
#if XM7_VER >= 2
	if (!mmr_ext || (model_type == MODEL_FM77AV20EX)) {
		maddr &= 0x3f;
	}
#else
	if (mmr_extram != EXRAM_960KB) {
		maddr &= 0x3f;
	}
#endif
	maddr <<= 12;

	/* 12rbgƍ */
	addr &= 0xfff;
	maddr |= addr;

	return maddr;
}

/*
 *	CCPUoX
 *	PoCgǂݏo
 */
BOOL FASTCALL mmr_extrb(WORD *addr, BYTE *dat)
{
	DWORD raddr, rsegment;

	/* $FC00`$FFFF͏풓 */
	if (*addr >= 0xfc00) {
		return FALSE;
	}

	/* TWR,MMRʂ */
#ifdef FMT
	raddr = mmr_trans_twr_and_mmr(*addr);
#else
	if (!mmr_trans_twr(*addr, &raddr)) {
		raddr = mmr_trans_mmr(*addr);
	}
#endif

	rsegment = (raddr & 0xf0000);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER >= 2
		/* MMR͍Ĕzu֎~ */
		if ((raddr >= 0x3fd80) && (raddr <= 0x3fd97)) {
			*dat = 0xff;
			LOG2(Warning, "ĔzuMMRւ̃oCgǂݍ $%05X($%04X)", raddr, *addr);
			return TRUE;
		}
#endif

#if XM7_VER == 1
		/* 0x3fc00-0x3ffff̍ĔzuɂBASIC ROM or RAMgp */
		if (raddr >= 0x3fc00) {
			if (basicrom_en) {
				*dat = basic_rom[raddr - 0x38000];
			}
			else {
				/* RAM */
				*dat = mainram_b[raddr - 0x38000];
			}
			return TRUE;
		}
#endif	/* XM7_VER == 1 */

		/* $30ZOg */
		*addr = (WORD)(raddr & 0xffff);
		return FALSE;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		*dat = extram_a[raddr & 0xffff];
		return TRUE;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			*dat = submem_readb((WORD)(raddr & 0xffff));
		}
		else {
#ifdef MMR_LOG
			LOG2(Warning, "HALTTuVXeւ̃oCgǂݍ $%05X($%04X)", raddr, *addr);
#endif
			*dat = 0xff;
		}
		return TRUE;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
		*dat = jcard_readb((WORD)(raddr & 0xffff));
		return TRUE;
	}

#if XM7_VER >= 3
	/* 768KB gRAM */
	if (rsegment >= 0x40000) {
		if ((mmr_extram == EXRAM_768KB) ||
			((mmr_extram == EXRAM_256KB) && (rsegment <= 0x70000))) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			LOG2(Warning, "gRAM̃oCgǂݍ $%05X($%04X)", raddr, *addr);
			*dat = 0xff;
		}
		return TRUE;
	}
#endif

	return FALSE;
#else
	/* FM-77 gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram == EXRAM_960KB) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			LOG2(Warning, "gRAM̃oCgǂݍ $%05X($%04X)", raddr, *addr);
			*dat = 0xff;
		}
		return TRUE;
	}

	/* FM-77 gRAM  */
	if (mmr_extram == EXRAM_OFF) {
		LOG2(Warning, "gRAM̃oCgǂݍ $%05X($%04X)", raddr, *addr);
		*dat = 0xff;
		return TRUE;
	}

	/* FM-77 gRAM 64KB */
	if (mmr_extram == EXRAM_64KB) {
		if (rsegment != (DWORD)(bsr_reg << 16)) {
			LOG2(Warning, "gRAM̃oCgǂݍ $%05X($%04X)", raddr, *addr);
			*dat = 0xff;
			return TRUE;
		}
		raddr &= (DWORD)0x0ffff;
		raddr |= (DWORD)0x20000;
	}

	*dat = extram_a[raddr];
	return TRUE;
#endif
}

/*
 *	CCPUoX
 *	PoCgǂݏo(I/OȂ)
 */
BOOL FASTCALL mmr_extbnio(WORD *addr, BYTE *dat)
{
	DWORD raddr, rsegment;

	/* $FC00`$FFFF͏풓 */
	if (*addr >= 0xfc00) {
		return FALSE;
	}

	/* TWR,MMRʂ */
#ifdef FMT
	raddr = mmr_trans_twr_and_mmr(*addr);
#else
	if (!mmr_trans_twr(*addr, &raddr)) {
		raddr = mmr_trans_mmr(*addr);
	}
#endif

	rsegment = (raddr & 0xf0000);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER >= 2
		/* MMR͍Ĕzu֎~ */
		if ((raddr >= 0x3fd80) && (raddr <= 0x3fd97)) {
			*dat = 0xff;
			return TRUE;
		}
#endif

#if XM7_VER == 1
		/* 0x3fc00-0x3ffff̍ĔzuɂBASIC ROM or RAMgp */
		if (raddr >= 0x3fc00) {
			if (basicrom_en) {
				*dat = basic_rom[raddr - 0x38000];
			}
			else {
				/* RAM */
				*dat = mainram_b[raddr - 0x38000];
			}
			return TRUE;
		}
#endif	/* XM7_VER == 1 */

		/* $30ZOg */
		*addr = (WORD)(raddr & 0xffff);
		return FALSE;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		*dat = extram_a[raddr & 0xffff];
		return TRUE;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			*dat = submem_readbnio((WORD)(raddr & 0xffff));
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
		*dat = jcard_readb((WORD)(raddr & 0xffff));
		return TRUE;
	}

#if XM7_VER >= 3
	/* 768KB gRAM */
	if (rsegment >= 0x40000) {
		if ((mmr_extram == EXRAM_768KB) ||
			((mmr_extram == EXRAM_256KB) && (rsegment <= 0x70000))) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}
#endif

	return FALSE;
#else
	/* FM-77 gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram == EXRAM_960KB) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}

	/* FM-77 gRAM  */
	if (mmr_extram == EXRAM_OFF) {
		*dat = 0xff;
		return TRUE;
	}

	/* FM-77 gRAM 64KB */
	if (mmr_extram == EXRAM_64KB) {
		if (rsegment != (DWORD)(bsr_reg << 16)) {
			*dat = 0xff;
			return TRUE;
		}
		raddr &= (DWORD)0x0ffff;
		raddr |= (DWORD)0x20000;
	}

	*dat = extram_a[raddr];
	return TRUE;
#endif
}

/*
 *	CCPUoX
 *	PoCg
 */
BOOL FASTCALL mmr_extwb(WORD *addr, BYTE dat)
{
	DWORD raddr, rsegment;

	/* $FC00`$FFFF͏풓 */
	if (*addr >= 0xfc00) {
		return FALSE;
	}

	/* TWR,MMRʂ */
#ifdef FMT
	raddr = mmr_trans_twr_and_mmr(*addr);
#else
	if (!mmr_trans_twr(*addr, &raddr)) {
		raddr = mmr_trans_mmr(*addr);
	}
#endif

	rsegment = (raddr & 0xf0000);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER >= 2
		/* MMR͍Ĕzu֎~ */
		if ((raddr >= 0x3fd80) && (raddr <= 0x3fd97)) {
			LOG3(Warning, "ĔzuMMRւ̃oCg $%05X($%04X) <- $%02X", raddr, *addr, dat);
			return TRUE;
		}
#endif

#if XM7_VER == 1
		/* 0x3fc00-0x3ffff̍ĔzuɂBASIC ROM or RAMgp */
		if (raddr >= 0x3fc00) {
			if (basicrom_en) {
				LOG3(Warning, "BASIC ROMւ̃oCg $%05X($%04X) <- $%02X", raddr, *addr, dat);
			}
			else {
				/* RAM */
				mainram_b[raddr - 0x38000] = dat;
			}
			return TRUE;
		}
#endif	/* XM7_VER == 1 */

		/* $30ZOg */
		*addr = (WORD)(raddr & 0xffff);
		return FALSE;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		extram_a[raddr & 0xffff] = dat;
		return TRUE;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			submem_writeb((WORD)(raddr & 0xffff), dat);
		}
#ifdef MMR_LOG
		else {
			LOG3(Warning, "HALTTuVXeւ̃oCg $%05X($%04X) <- $%02X", raddr, *addr, dat);
		}
#endif
		return TRUE;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
		jcard_writeb((WORD)(raddr & 0xffff), dat);
		return TRUE;
	}

#if XM7_VER >= 3
	/* AV40gRAM */
	if (rsegment >= 0x40000) {
		if ((mmr_extram == EXRAM_768KB) ||
			((mmr_extram == EXRAM_256KB) && (rsegment <= 0x70000))) {
			extram_c[raddr - 0x40000] = dat;
		}
		else {
			LOG3(Warning, "gRAM̃oCg $%05X($%04X) <- $%02X", raddr, *addr, dat);
		}
		return TRUE;
	}
#endif

	return FALSE;
#else
	/* FM-77 gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram == EXRAM_960KB) {
			extram_c[raddr - 0x40000] = dat;
		}
		return TRUE;
	}

	/* FM-77 gRAM  */
	if (mmr_extram == EXRAM_OFF) {
		return TRUE;
	}

	/* FM-77 gRAM 64KB */
	if (mmr_extram == EXRAM_64KB) {
		if (rsegment != (DWORD)(bsr_reg << 16)) {
			return TRUE;
		}
		raddr &= 0x0ffff;
		raddr |= 0x20000;
	}

	extram_a[raddr] = dat;
	return TRUE;
#endif

}

/*
 *	C(4rbgw)
 *	PoCgǂݏo(I/OȂ)
 */
BOOL FASTCALL mmr_extrbsegnio(WORD *addr, BYTE *dat, BYTE seg, int mode)
{
	DWORD raddr, rsegment;

#if XM7_VER >= 2
	ASSERT(model_type >= MODEL_FM77AV);
#else
	ASSERT(model_type == MODEL_FM77);
#endif

	if (mode == 1) {
		/* 4rbgw蕨AhX */
		raddr = ((seg << 16) | *addr);
	}
	else if (mode == 2) {
		int offset;

		/* $FC00`$FFFF͏풓 */
		if (*addr >= 0xfc00) {
			return FALSE;
		}

		/* MMRWX^擾 */
		offset = (int)*addr;
		offset >>= 12;

#if XM7_VER >= 3
		offset |= (seg * 0x10);
#elif XM7_VER == 1
		offset |= (seg * 0x10);
#else
		offset |= ((seg & 0x03) * 0x10);
#endif

		/* gMMRoffȂA6bit̂ݗL */
		raddr = (DWORD)mmr_reg[offset];
#if XM7_VER >= 2
		if (!mmr_ext || (model_type == MODEL_FM77AV20EX)) {
			raddr &= 0x3f;
		}
#else
		if (mmr_extram != EXRAM_960KB) {
			raddr &= 0x3f;
		}
#endif
		raddr <<= 12;

		/* 12rbgƍ */
		*addr &= 0xfff;
		raddr |= *addr;
	}

	rsegment = (raddr & 0xf0000);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER >= 2
		/* MMR͍Ĕzu֎~ */
		if ((raddr >= 0x3fd80) && (raddr <= 0x3fd97)) {
			*dat = 0xff;
			return TRUE;
		}
#endif

#if XM7_VER == 1
		/* 0x3fc00-0x3ffff̍ĔzuɂBASIC ROM or RAMgp */
		if (raddr >= 0x3fc00) {
			if (basicrom_en) {
				*dat = basic_rom[raddr - 0x38000];
			}
			else {
				/* RAM */
				*dat = mainram_b[raddr - 0x38000];
			}
			return TRUE;
		}
#endif	/* XM7_VER == 1 */

		/* $30ZOg */
		*addr = (WORD)(raddr & 0xffff);
		return FALSE;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		*dat = extram_a[raddr & 0xffff];
		return TRUE;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			*dat = submem_readbnio((WORD)(raddr & 0xffff));
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
		*dat = jcard_readb((WORD)(raddr & 0xffff));
		return TRUE;
	}

#if XM7_VER >= 3
	/* 768KB gRAM */
	if (rsegment >= 0x40000) {
		if ((mmr_extram == EXRAM_768KB) ||
			((mmr_extram == EXRAM_256KB) && (rsegment <= 0x70000))) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}
#endif

	return FALSE;
#else
	/* FM-77 gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram == EXRAM_960KB) {
			*dat = extram_c[raddr - 0x40000];
		}
		else {
			*dat = 0xff;
		}
		return TRUE;
	}

	/* FM-77 gRAM  */
	if (mmr_extram == EXRAM_OFF) {
		*dat = 0xff;
		return TRUE;
	}

	/* FM-77 gRAM 64KB */
	if (mmr_extram == EXRAM_64KB) {
		if (rsegment != (DWORD)(bsr_reg << 16)) {
			*dat = 0xff;
			return TRUE;
		}
		raddr &= (DWORD)0x0ffff;
		raddr |= (DWORD)0x20000;
	}

	*dat = extram_a[raddr];
	return TRUE;
#endif
}

/*
 *	C(4rbgw)
 *	PoCg
 */
BOOL FASTCALL mmr_extwbseg(WORD *addr, BYTE dat, BYTE seg, int mode)
{
	DWORD raddr, rsegment;

#if XM7_VER >= 2
	ASSERT(model_type >= MODEL_FM77AV);
#else
	ASSERT(model_type == MODEL_FM77);
#endif

	if (mode == 1) {
		/* 4rbgw蕨AhX */
		raddr = ((seg << 16) | *addr);
	}
	else if (mode == 2) {
		int offset;

		/* $FC00`$FFFF͏풓 */
		if (*addr >= 0xfc00) {
			return FALSE;
		}

		/* MMRWX^擾 */
		offset = (int)*addr;
		offset >>= 12;

#if XM7_VER >= 3
		offset |= (seg * 0x10);
#elif XM7_VER == 1
		offset |= (seg * 0x10);
#else
		offset |= ((seg & 0x03) * 0x10);
#endif

		/* gMMRoffȂA6bit̂ݗL */
		raddr = (DWORD)mmr_reg[offset];
#if XM7_VER >= 2
		if (!mmr_ext || (model_type == MODEL_FM77AV20EX)) {
			raddr &= 0x3f;
		}
#else
		if (mmr_extram != EXRAM_960KB) {
			raddr &= 0x3f;
		}
#endif
		raddr <<= 12;

		/* 12rbgƍ */
		*addr &= 0xfff;
		raddr |= *addr;
	}

	rsegment = (raddr & 0xf0000);

	/* W */
	if (rsegment == 0x30000) {
#if XM7_VER >= 2
		/* MMR͍Ĕzu֎~ */
		if ((raddr >= 0x3fd80) && (raddr <= 0x3fd97)) {
			return TRUE;
		}
#endif

#if XM7_VER == 1
		/* 0x3fc00-0x3ffff̍ĔzuɂBASIC ROM or RAMgp */
		if (raddr >= 0x3fc00) {
			if (basicrom_en) {
				return TRUE;
			}
			else {
				/* RAM */
				mainram_b[raddr - 0x38000] = dat;
				return TRUE;
			}
		}
#endif	/* XM7_VER == 1 */

		/* $30ZOg */
		*addr = (WORD)(raddr & 0xffff);
		return FALSE;
	}

#if XM7_VER >= 2
	/* FM77AV gRAM */
	if (rsegment == 0x00000) {
		extram_a[raddr & 0xffff] = dat;
		return TRUE;
	}

	/* TuVXe */
	if (rsegment == 0x10000) {
		if (subhalt_flag) {
			submem_writeb((WORD)(raddr & 0xffff), dat);
		}
		return TRUE;
	}

	/* {J[h */
	if (rsegment == 0x20000) {
		jcard_writeb((WORD)(raddr & 0xffff), dat);
		return TRUE;
	}

#if XM7_VER >= 3
	/* AV40gRAM */
	if (rsegment >= 0x40000) {
		if ((mmr_extram == EXRAM_768KB) ||
			((mmr_extram == EXRAM_256KB) && (rsegment <= 0x70000))) {
			extram_c[raddr - 0x40000] = dat;
		}
		return TRUE;
	}
#endif

	return FALSE;
#else
	/* FM-77 gRAM */
	if (rsegment >= 0x40000) {
		if (mmr_extram == EXRAM_960KB) {
			extram_c[raddr - 0x40000] = dat;
		}
		return TRUE;
	}

	/* FM-77 gRAM  */
	if (mmr_extram == EXRAM_OFF) {
		return TRUE;
	}

	/* FM-77 gRAM 64KB */
	if (mmr_extram == EXRAM_64KB) {
		if (rsegment != (DWORD)(bsr_reg << 16)) {
			return TRUE;
		}
		raddr &= 0x0ffff;
		raddr |= 0x20000;
	}

	extram_a[raddr] = dat;
	return TRUE;
#endif

}

/*-[ MR2 I/O^RAMfBXN ]-------------------------------------------------*/

#if (XM7_VER >= 3) && defined(MR2)
/*
 *	MR2
 *	AhXvZ
 */
static DWORD FASTCALL mr2_address(void)
{
	DWORD tmp;

	ASSERT (XM7_VER >= 3);

	if (mr2_secreg <= 0x0bff) {
		/* vZ@͓KȂ̂ňĂ\A(^^; 2002/07/29 */
		tmp = (0xbff - mr2_secreg) << 8;
		tmp |= mr2_nowcnt;
	}
	else {
		tmp = 0;
	}

	return tmp;
}

/*
 *	MR2
 *	f[^[h
 */
static BYTE FASTCALL mr2_read_data(void)
{
	BYTE dat;

	ASSERT (XM7_VER >= 3);

	/* ZN^WX^ُ̏ꍇ͓ǂݏoȂ */
	if (mr2_secreg <= 0x0bff) {
		dat = extram_c[mr2_address()];
		mr2_nowcnt ++;
	}
	else {
		dat = 0xff;
	}

	return dat;
}

/*
 *	MR2
 *	f[^Cg
 */
static void FASTCALL mr2_write_data(BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	/* ZN^WX^ُ̏ꍇ͏߂Ȃ */
	if (mr2_secreg <= 0x0bff) {
		extram_c[mr2_address()] = dat;
		mr2_nowcnt ++;
	}
}

/*
 *	MR2
 *	ZN^WX^XV
 */
static void FASTCALL mr2_update_sector(WORD addr, BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	/* ZN^WX^XV */
	if (addr == 0xfd9e) {
		mr2_secreg &= (WORD)0x00ff;
		mr2_secreg |= (WORD)((dat & 0x0f) << 8);
	}
	else {
		mr2_secreg &= (WORD)0x0f00;
		mr2_secreg |= (WORD)dat;
	}

	/* ͈̓`FbN */
	if (mr2_secreg >= 0x0c00) {
		ASSERT(FALSE);
		mr2_secreg = 0x0c00;
	}

	/* ZN^JE^Zbg */
	mr2_nowcnt = 0;
}
#endif

/*-[ RD512 I/O^RAMfBXN ]-----------------------------------------------*/

#if (XM7_VER >= 3) && defined(RD512)
/*
 *	RD512
 *	AhXvZ
 */
static DWORD FASTCALL rd512_address(WORD addr)
{
	ASSERT (XM7_VER >= 3);

	return (DWORD)(rd512_secreg + ((addr & 0x07) << 16) + 0x40000);
}

/*
 *	RD512
 *	f[^[h
 */
static BYTE FASTCALL rd512_read_data(WORD addr)
{
	BYTE dat;

	ASSERT (XM7_VER >= 3);

	if ((addr >= 0xfd48) && (addr <= 0xfd4f)) {
		dat = extram_c[rd512_address(addr)];
	}
	else {
		dat = 0xff;
	}

	return dat;
}

/*
 *	RD512
 *	f[^Cg
 */
static void FASTCALL rd512_write_data(WORD addr, BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	if ((addr >= 0xfd48) && (addr <= 0xfd4f)) {
		extram_c[rd512_address(addr)] = dat;
	}
}

/*
 *	RD512
 *	ZN^WX^XV
 */
static void FASTCALL rd512_update_sector(WORD addr, BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	/* ZN^WX^XV */
	if (addr == 0xfd41) {
		rd512_secreg &= (WORD)0x00ff;
		rd512_secreg |= (WORD)((dat & 0xff) << 8);
	}
	else {
		rd512_secreg &= (WORD)0xff00;
		rd512_secreg |= (WORD)dat;
	}
}
#endif

/*-[ R2D2 I/O^RAMfBXN ]------------------------------------------------*/

#if (XM7_VER >= 3) && defined(R2D2)
/*
 *	R2D2
 *	AhXvZ
 */
static DWORD FASTCALL r2d2_address(WORD addr)
{
	DWORD tmp;

	ASSERT (XM7_VER >= 3);

	tmp =  (DWORD)(r2d2_secreg << 8);
	tmp |= (DWORD)(r2d2_nowcnt & 0xfe);
	tmp |= (DWORD)(addr & 0x01);
	tmp += 0x40000;

	return tmp;
}

/*
 *	R2D2
 *	f[^[h
 */
static BYTE FASTCALL r2d2_read_data(WORD addr)
{
	BYTE dat;

	ASSERT (XM7_VER >= 3);

	if (r2d2_secreg <= 0x07ff) {
		dat = extram_c[r2d2_address(addr)];
	}
	r2d2_nowcnt ++;

	return dat;
}

/*
 *	R2D2
 *	f[^Cg
 */
static void FASTCALL r2d2_write_data(WORD addr, BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	if (r2d2_secreg <= 0x07ff) {
		extram_c[r2d2_address(addr)] = dat;
	}
	r2d2_nowcnt ++;
}

/*
 *	R2D2
 *	ZN^WX^XV
 */
static void FASTCALL r2d2_update_sector(WORD addr, BYTE dat)
{
	ASSERT (XM7_VER >= 3);

	/* ZN^WX^XV */
	if (addr == 0xfd62) {
		r2d2_secreg &= (WORD)0x00ff;
		r2d2_secreg |= (WORD)((dat & 0x07) << 8);
	}
	else {
		r2d2_secreg &= (WORD)0x0700;
		r2d2_secreg |= (WORD)dat;
	}

	/* JE^ */
	r2d2_nowcnt = 0;
}
#endif

/*-[ }bvhI/O ]----------------------------------------------------*/

/*
 *	MMR
 *	PoCgǂݏo
 */
BOOL FASTCALL mmr_readb(WORD addr, BYTE *dat)
{
	BYTE tmp;

	/* f`FbN */
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}
#else
	if (model_type != MODEL_FM77) {
		return FALSE;
	}
#endif

	switch (addr) {
#if XM7_VER >= 2
		/* u[gXe[^X */
		case 0xfd0b:
			if (boot_mode == BOOT_BASIC) {
				*dat = 0xfe;
			}
			else {
				*dat = 0xff;
			}

#ifdef RSC
			/* RS-232C CDM */
			if (rs_cd) {
				*dat &= (BYTE)~RSCB_CD;
			}
#endif
			return TRUE;

		/* CjVG[^ROM */
		case 0xfd10:
			*dat = 0xff;
			return TRUE;
#endif

#if XM7_VER == 1
		/* gRAMoNZNgWX^ */
		case 0xfd2f:
			if (mmr_extram == EXRAM_64KB) {
				*dat = (BYTE)(bsr_reg | 0xfc);
				return TRUE;
			}
			break;
#endif

		/* MMRZOg */
		case 0xfd90:
			*dat = 0xff;
			return TRUE;

#if XM7_VER == 1
		/* WBRoN */
		case 0xfd91:
			if (mmr_extram == EXRAM_960KB) {
				*dat = 0xff;
				return TRUE;
			}
			break;
#endif

		/* TWRItZbg */
		case 0xfd92:
			*dat = 0xff;
			return TRUE;

		/* [hZNg */
		case 0xfd93:
			tmp = 0xff;
			if (!mmr_flag) {
				tmp &= (BYTE)(~0x80);
			}
			if (!twr_flag) {
				tmp &= ~0x40;
			}
			if (!bootram_rw) {
				tmp &= ~1;
			}
			*dat = tmp;
			return TRUE;

#if XM7_VER >= 2
		/* gMMR/CPUXs[h */
		case 0xfd94:
			*dat = 0xff;
			return TRUE;

		/* [hZNgQ */
		case 0xfd95:
			tmp = 0xff;
#if XM7_VER >= 3
			/* AV40EX̂ */
			if (model_type == MODEL_FM77AV40EX) {
				/* bit7:gROMZNg */
				if (!extrom_sel) {
					tmp &= (BYTE)~0x80;
				}
			}
			/* AV20EX/AV40EX̂ */
			if ((model_type == MODEL_FM77AV20EX) ||
				(model_type == MODEL_FM77AV40EX)) {
#else
			/* AV20EX̂ */
			if (model_type == MODEL_FM77AV20EX) {
#endif
				/* bit4:MMRgp̑xቺ}~ */
				if (!mmr_fastmode) {
					tmp &= (BYTE)~0x08;
				}
			}
#if XM7_VER >= 3
			/* AV40EX̂ */
			if (model_type == MODEL_FM77AV40EX) {
				/* bit0:400C^C~Oo̓Xe[^X */
				/*      XM7ł1(200Co)Œ cB */
				if (dsp_400linetiming) {
					tmp &= (BYTE)~0x01;
				}
			}
#endif

			*dat = tmp;
			return TRUE;

#ifdef MR2
		/* MR2 f[^WX^ */
		case 0xfd9c:
			if (mmr_extram == EXRAM_768KB) {
				*dat = mr2_read_data();
			}
			else {
				*dat = 0xff;
			}
			return TRUE;

		/* MR2 ZN^WX^(Write Only) */
		case 0xfd9e:
		case 0xfd9f:
			*dat = 0xff;
			return TRUE;
#endif

#ifdef RD512
		/* RD512 ZN^WX^(Write Only) */
		case 0xfd40:
		case 0xfd41:
			*dat = 0xff;
			return TRUE;

		/* RD512 f[^WX^ */
		case 0xfd48:
		case 0xfd49:
		case 0xfd4a:
		case 0xfd4b:
		case 0xfd4c:
		case 0xfd4d:
		case 0xfd4e:
		case 0xfd4f:
			if ((mmr_extram != EXRAM_OFF)) {
				*dat = rd512_read_data(addr);
			}
			else {
				*dat = 0xff;
			}
			return TRUE;
#endif

#ifdef R2D2
		/* R2D2 f[^WX^ */
		case 0xfd60:
		case 0xfd61:
			if ((mmr_extram != EXRAM_OFF)) {
				*dat = r2d2_read_data(addr);
			}
			else {
				*dat = 0xff;
			}
			return TRUE;

		/* R2D2 ZN^WX^(Write Only) */
		case 0xfd62:
		case 0xfd63:
			*dat = 0xff;
			return TRUE;
#endif
#endif
	}

	/* MMRWX^ */
	if ((addr >= 0xfd80) && (addr <= 0xfd8f)) {
#if XM7_VER >= 3
		if (mmr_ext) {
			tmp = mmr_reg[mmr_seg * 0x10 + (addr - 0xfd80)];
		}
		else {
			tmp = mmr_reg[(mmr_seg & 3) * 0x10 + (addr - 0xfd80)];
			/* tmp &= 0x3f; */
		}
#elif XM7_VER == 1
		if (mmr_extram == EXRAM_960KB) {
			tmp = mmr_reg[mmr_seg * 0x10 + (addr - 0xfd80)];
		}
		else {
			tmp = mmr_reg[(mmr_seg & 3) * 0x10 + (addr - 0xfd80)];
			/* tmp &= 0x3f; */
		}
#else
		tmp = mmr_reg[(mmr_seg & 3) * 0x10 + (addr - 0xfd80)];
		/* tmp &= 0x3f; */
#endif
		*dat = tmp;
		return TRUE;
	}

	return FALSE;
}

/*
 *	MMR
 *	PoCg
 */
BOOL FASTCALL mmr_writeb(WORD addr, BYTE dat)
{
	/* f`FbN */
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}
#else
	if (model_type != MODEL_FM77) {
		return FALSE;
	}
#endif

	switch (addr) {
		/* CjVG[^ROM */
#if XM7_VER >= 2
		case 0xfd10:
			if (dat & 0x02) {
				initrom_en = FALSE;
			}
			else {
				initrom_en = TRUE;
			}
			return TRUE;
#endif

#if XM7_VER == 1
		/* gRAMoNZNgWX^ */
		case 0xfd2f:
			if (mmr_extram == EXRAM_64KB) {
				bsr_reg = (BYTE)(dat & 0x03);
				return TRUE;
			}
			break;
#endif

		/* MMRZOg */
		case 0xfd90:
#if XM7_VER == 1
			mmr_seg = (BYTE)(dat & 0x3f);
#else
			mmr_seg = (BYTE)(dat & 0x07);
#endif
			return TRUE;

#if XM7_VER == 1
		/* WBRoN */
		case 0xfd91:
			if (mmr_extram == EXRAM_960KB) {
				wbr_reg = (BYTE)(dat & 0x0f);
				return TRUE;
			}
			break;
#endif

		/* TWRItZbg */
		case 0xfd92:
			twr_reg = dat;
			return TRUE;

		/* [hZNg */
		case 0xfd93:
			if (dat & 0x80) {
				if (!mmr_flag) {
					mmr_flag = TRUE;
#if XM7_VER >= 2
					if (!mmr_fastmode) {
						mmr_modify = TRUE;
					}
#else
					mmr_modify = TRUE;
#endif
				}
			}
			else {
				if (mmr_flag) {
					mmr_flag = FALSE;
#if XM7_VER >= 2
					if (!mmr_fastmode) {
						mmr_modify = TRUE;
					}
#else
					mmr_modify = TRUE;
#endif
				}
			}
			if (dat & 0x40) {
				if (!twr_flag) {
					twr_flag = TRUE;
#if XM7_VER >= 2
					if (!mmr_fastmode) {
						mmr_modify = TRUE;
					}
#else
					mmr_modify = TRUE;
#endif
				}
			}
			else {
				if (twr_flag) {
					twr_flag = FALSE;
#if XM7_VER >= 2
					if (!mmr_fastmode) {
						mmr_modify = TRUE;
					}
#else
					mmr_modify = TRUE;
#endif
				}
			}
			if (dat & 0x01) {
				bootram_rw = TRUE;
			}
			else {
				bootram_rw = FALSE;
			}
			return TRUE;

#if XM7_VER >= 2
		/* gMMR/CPUXs[h */
		case 0xfd94:
#if XM7_VER >= 3
			/* AV40/AV20EX/AV40EXŗL */
			if ((model_type == MODEL_FM77AV20EX) ||
				(model_type == MODEL_FM77AV40) ||
				(model_type == MODEL_FM77AV40EX)) {
#else
			/* AV20EXŗL */
			if (model_type == MODEL_FM77AV20EX) {
#endif
				/* bit7:gMMR */
				if (dat & 0x80) {
					mmr_ext = TRUE;
				}
				else {
					mmr_ext = FALSE;
				}
			}
			else {
				mmr_ext = FALSE;
			}
#if XM7_VER >= 3
			/* AV20/AV40/AV20EX/AV40EXŗL */
			if ((model_type == MODEL_FM77AV20) ||
				(model_type == MODEL_FM77AV40) ||
				(model_type == MODEL_FM77AV20EX) ||
				(model_type == MODEL_FM77AV40EX)) {
#else
			/* AV20/AV20EXŗL */
			if ((model_type == MODEL_FM77AV20) ||
				(model_type == MODEL_FM77AV20EX)) {
#endif
				/* bit2:tbVXs[h */
				if (dat & 0x04) {
					if (!mmr_fast_refresh) {
						mmr_fast_refresh = TRUE;
						if (!mmr_fastmode) {
							mmr_modify = TRUE;
						}
					}
				}
				else {
					if (mmr_fast_refresh) {
						mmr_fast_refresh = FALSE;
						if (!mmr_fastmode) {
							mmr_modify = TRUE;
						}
					}
				}
				/* bit0:EBhEXs[h */
				if (dat & 0x01) {
					if (!twr_fastmode) {
						twr_fastmode = TRUE;
						if (!mmr_fastmode) {
							mmr_modify = TRUE;
						}
					}
				}
				else {
					if (twr_fastmode) {
						twr_fastmode = FALSE;
						if (!mmr_fastmode) {
							mmr_modify = TRUE;
						}
					}
				}
			}
			return TRUE;

		/* [hZNgQ */
		case 0xfd95:
#if XM7_VER >= 3
			/* AV40EXŗL */
			if (model_type == MODEL_FM77AV40EX) {
				/* bit7:gROMZNg */
				if (dat & 0x80) {
					extrom_sel = TRUE;
				}
				else {
					extrom_sel = FALSE;
				}
			}
			else {
				extrom_sel = FALSE;
			}
			/* AV20EX/AV40EXŗL */
			if ((model_type == MODEL_FM77AV20EX) ||
				(model_type == MODEL_FM77AV40EX)) {
#else
			/* AV20EXŗL */
			if (model_type == MODEL_FM77AV20EX) {
#endif
				/* bit4:MMRgp̑xቺ}~ */
				if (dat & 0x08) {
					if (!mmr_fastmode) {
						mmr_fastmode = TRUE;
						mmr_modify = TRUE;
					}
				}
				else {
					if (mmr_fastmode) {
						mmr_fastmode = FALSE;
						mmr_modify = TRUE;
					}
				}
			}
			return TRUE;

#ifdef MR2
		/* MR2 f[^WX^ */
		case 0xfd9c:
			if (mmr_extram == EXRAM_768KB) {
				mr2_write_data(dat);
			}
			return TRUE;

		/* MR2 ZN^WX^ */
		case 0xfd9e:
		case 0xfd9f:
			mr2_update_sector(addr, dat);
			return TRUE;
#endif

#ifdef RD512
		/* RD512 ZN^WX^ */
		case 0xfd40:
		case 0xfd41:
			rd512_update_sector(addr, dat);
			return TRUE;

		/* RD512 f[^WX^ */
		case 0xfd48:
		case 0xfd49:
		case 0xfd4a:
		case 0xfd4b:
		case 0xfd4c:
		case 0xfd4d:
		case 0xfd4e:
		case 0xfd4f:
			if ((mmr_extram != EXRAM_OFF)) {
				rd512_write_data(addr, dat);
			}
			return TRUE;
#endif

#ifdef R2D2
		/* R2D2 f[^WX^ */
		case 0xfd60:
		case 0xfd61:
			if ((mmr_extram != EXRAM_OFF)) {
				r2d2_write_data(addr, dat);
			}
			return TRUE;

		/* R2D2 ZN^WX^ */
		case 0xfd62:
		case 0xfd63:
			r2d2_update_sector(addr, dat);
			return TRUE;
#endif
#endif
	}

	/* MMRWX^ */
	if ((addr >= 0xfd80) && (addr <= 0xfd8f)) {
#if XM7_VER >= 2
		/* ł̃f[^8bitׂċL */
		if (mmr_ext) {
			mmr_reg[mmr_seg * 0x10 + (addr - 0xfd80)] = (BYTE)dat;
		}
		else {
			mmr_reg[(mmr_seg & 3) * 0x10 + (addr - 0xfd80)] = (BYTE)dat;
		}
#else
		/* ł̃f[^8bitׂċL */
		if (mmr_extram == EXRAM_960KB) {
			mmr_reg[mmr_seg * 0x10 + (addr - 0xfd80)] = (BYTE)dat;
		}
		else {
			mmr_reg[(mmr_seg & 3) * 0x10 + (addr - 0xfd80)] = (BYTE)dat;
		}
#endif
		return TRUE;
	}

	return FALSE;
}

/*-[ t@CI/O ]----------------------------------------------------------*/

/*
 *	MMR
 *	Z[u
 */
BOOL FASTCALL mmr_save(int fileh)
{
#if XM7_VER >= 3 || XM7_VER == 1
	DWORD addr;
#endif
#if XM7_VER >= 3
	BOOL tmp;
#endif

	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, mmr_flag)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, mmr_seg)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if (!file_write(fileh, mmr_reg, 0x80)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mmr_ext)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mmr_fastmode)) {
		return FALSE;
	}
	if (mmr_extram != EXRAM_OFF) {
		if (!file_bool_write(fileh, TRUE)) {
			return FALSE;
		}
		for (addr = 0x00000; addr < 0xc0000; addr += 0x8000) {
			if (!file_write(fileh, &extram_c[addr], 0x8000)) {
				return FALSE;
			}
		}
	}
	else {
		if (!file_bool_write(fileh, FALSE)) {
			return FALSE;
		}
	}
#ifdef MR2
	if (!file_byte_write(fileh, mr2_nowcnt)) {
		return FALSE;
	}
	if (!file_word_write(fileh, mr2_secreg)) {
		return FALSE;
	}
#else
	if (!file_byte_write(fileh, 0)) {
		return FALSE;
	}
	if (!file_word_write(fileh, 0)) {
		return FALSE;
	}
#endif
#elif XM7_VER >= 2
	/* Ver12g */
	if (!file_write(fileh, mmr_reg, 0x80)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mmr_ext)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mmr_fastmode)) {
		return FALSE;
	}
#else
	if (!file_byte_write(fileh, mmr_extram)) {
		return FALSE;
	}
	if (mmr_extram == EXRAM_960KB) {
		if (!file_write(fileh, mmr_reg, 0x100)) {
			return FALSE;
		}
		for (addr = 0x00000; addr < 0xc0000; addr += 0x8000) {
			if (!file_write(fileh, &extram_c[addr], 0x8000)) {
				return FALSE;
			}
		}
		if (!file_byte_write(fileh, wbr_reg)) {
			return FALSE;
		}
	}
	else {
		if (!file_write(fileh, mmr_reg, 0x40)) {
			return FALSE;
		}
	}
	if (mmr_extram == EXRAM_64KB) {
		if (!file_byte_write(fileh, bsr_reg)) {
			return FALSE;
		}
	}
#endif

	if (!file_bool_write(fileh, twr_flag)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, twr_reg)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_bool_write(fileh, twr_fastmode)) {
		return FALSE;
	}
#endif
#if XM7_VER >= 3
	if (mmr_extram == EXRAM_768KB) {
		tmp = TRUE;
	}
	else {
		tmp = FALSE;
	}
	if (!file_bool_write(fileh, tmp)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	MMR
 *	[h
 */
BOOL FASTCALL mmr_load(int fileh, int ver)
{
#if XM7_VER >= 3 || XM7_VER == 1
	DWORD addr;
#endif
#if XM7_VER >= 3
#ifndef MR2
	WORD tmp;
#endif
	BOOL bextram, b768kbmode;
#elif XM7_VER == 1
	BOOL tmp;
#endif

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_bool_read(fileh, &mmr_flag)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &mmr_seg)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* t@Co[W8Ŋg */
	if (((ver >= 800) && (ver <= 999)) || (ver >= 1400)) {
		if (!file_read(fileh, mmr_reg, 0x80)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_ext)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_fastmode)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &bextram)) {
			return FALSE;
		}
		if (bextram) {
			for (addr = 0x00000; addr < 0xc0000; addr+=0x8000) {
				if (!file_read(fileh, &extram_c[addr], 0x8000)) {
					return FALSE;
				}
			}
		}
		if (ver >= 902) {
#ifdef MR2
			if (!file_byte_read(fileh, &mr2_nowcnt)) {
				return FALSE;
			}
			if (!file_word_read(fileh, &mr2_secreg)) {
				return FALSE;
			}
#else
			if (!file_byte_read(fileh, (BYTE *)&tmp)) {
				return FALSE;
			}
			if (!file_word_read(fileh, &tmp)) {
				return FALSE;
			}
#endif
		}
#ifdef MR2
		else {
			mr2_nowcnt = 0;
			mr2_secreg = 0;
		}
#endif
		/* AV40/AV20EX/AV40EXȊO */
		if ((model_type != MODEL_FM77AV40) &&
			(model_type != MODEL_FM77AV20EX) &&
			(model_type != MODEL_FM77AV40EX)) {
			if (mmr_ext) {
				return FALSE;
			}
		}
	}
	/* t@Co[W12Ŋg */
	else if (ver >= 1200) {
		if (!file_read(fileh, mmr_reg, 0x80)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_ext)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_fastmode)) {
			return FALSE;
		}
		mmr_extram = EXRAM_OFF;
#ifdef MR2
		mr2_nowcnt = 0;
		mr2_secreg = 0;
#endif
		/* AV40/AV20EX/AV40EXȊO */
		if ((model_type != MODEL_FM77AV40) &&
			(model_type != MODEL_FM77AV20EX) &&
			(model_type != MODEL_FM77AV40EX)) {
			if (mmr_ext) {
				return FALSE;
			}
		}
	}
	else {
		/* Ver5݊ */
		if (!file_read(fileh, mmr_reg, 0x40)) {
			return FALSE;
		}
		mmr_ext = FALSE;
		mmr_fastmode = FALSE;
		mmr_extram = EXRAM_OFF;
#ifdef MR2
		mr2_nowcnt = 0;
		mr2_secreg = 0;
#endif
	}
#elif XM7_VER >= 2
	/* t@Co[W12Ŋg */
	if (ver >= 1200) {
		if (!file_read(fileh, mmr_reg, 0x80)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_ext)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &mmr_fastmode)) {
			return FALSE;
		}
		/* AV20EXȊO */
		if (model_type != MODEL_FM77AV20EX) {
			if (mmr_ext) {
				return FALSE;
			}
		}
	}
	else {
		/* Ver5݊ */
		if (!file_read(fileh, mmr_reg, 0x40)) {
			return FALSE;
		}
		mmr_ext = FALSE;
		mmr_fastmode = FALSE;
	}
#else
	/* Ver4.09g */
	if (ver >= 409) {
		if (!file_byte_read(fileh, &mmr_extram)) {
			return FALSE;
		}
		if (mmr_extram > EXRAM_960KB) {
			mmr_extram = EXRAM_192KB;
		}
	}
	else if (ver >= 406) {
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
		if (tmp) {
			mmr_extram = EXRAM_960KB;
		}
		else {
			mmr_extram = EXRAM_192KB;
		}
	}
	else {
		mmr_extram = EXRAM_192KB;
	}
	if (mmr_extram == EXRAM_960KB) {
		if (!file_read(fileh, mmr_reg, 0x100)) {
			return FALSE;
		}
		for (addr = 0x00000; addr < 0xc0000; addr+=0x8000) {
			if (!file_read(fileh, &extram_c[addr], 0x8000)) {
				return FALSE;
			}
		}
		if (!file_byte_read(fileh, &wbr_reg)) {
			return FALSE;
		}
	}
	else {
		if (!file_read(fileh, mmr_reg, 0x40)) {
			return FALSE;
		}
		wbr_reg = 2;
	}
	if ((ver >= 410) && (mmr_extram == EXRAM_64KB)) {
		if (!file_byte_read(fileh, &bsr_reg)) {
			return FALSE;
		}
	}
	else {
		bsr_reg = 2;
	}
	if ((ver >= 309) && (ver <= 399)) {
		if (!file_byte_read(fileh, &bsr_reg)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
		if (tmp) {
			mmr_extram = EXRAM_64KB;
		}
		else {
			mmr_extram = EXRAM_192KB;
		}
	}
#endif

	if (!file_bool_read(fileh, &twr_flag)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &twr_reg)) {
		return FALSE;
	}

	/* Ver9.16g */
#if XM7_VER >= 2
#if XM7_VER >= 3
	if (ver >= 916) {
#else
	if (ver >= 1200) {
#endif
		if (!file_bool_read(fileh, &twr_fastmode)) {
			return FALSE;
		}
	}
#endif

	/* Ver9.20g */
#if XM7_VER >= 3
	if (((ver >= 920) && (ver <= 999)) ||
		(ver >= 1409)) {
		if (!file_bool_read(fileh, &b768kbmode)) {
			return FALSE;
		}
		if (bextram) {
			if (b768kbmode) {
				mmr_extram = EXRAM_768KB;
			}
			else {
				mmr_extram = EXRAM_256KB;
			}
		}
		else {
			mmr_extram = EXRAM_OFF;
		}
	}
#endif

	return TRUE;
}
