/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ TuCPURg[ ]
 */

#include <string.h>
#include "xm7.h"
#include "keyboard.h"
#include "subctrl.h"
#include "device.h"
#include "display.h"
#include "aluline.h"
#include "ttlpalet.h"
#include "multipag.h"
#include "bubble.h"
#include "fdc.h"
#include "jsubsys.h"
#include "kanji.h"
#include "mainetc.h"
#include "midi.h"
#include "mmr.h"
#include "mouse.h"
#include "opn.h"
#include "rs232c.h"

/*
 *	O[o [N
 */
BOOL subhalt_flag;						/* TuHALTtO */
BOOL subbusy_flag;						/* TuBUSYtO */
BOOL subcancel_flag;					/* TuLZtO */
BOOL subattn_flag;						/* TuAeVtO */
BOOL subhalt_request;					/* TuHALTNGXgtO */
BOOL subcancel_request;					/* TuLZNGXgtO */
BYTE shared_ram[0x80];					/* LRAM */
BOOL subreset_flag;						/* TuċNtO */
BYTE busy_CLR_count;					/* BUSY($D40A) CLRߎsJE^ */
#if XM7_VER >= 2
BOOL mode320;							/* 320x200[h */
#if XM7_VER >= 3
BOOL mode400l;							/* 640x400[h */
BOOL mode256k;							/* 26F[h */
BOOL subram_protect;					/* Tuj^RAMveNg */
BOOL subreset_halt;						/* HALTċNtO */
BYTE subif_dat;							/* TuI/FWX^ ʃ[hۑ */
BOOL subkanji_flag;						/* ROM TuCPUڑtO */
#endif
#endif
#if XM7_VER == 1
BOOL parityerror;						/* peBG[tO(FM-8) */
#ifdef L4CARD
BOOL select_400line;					/* C400CItO */
BOOL subkanji_flag;						/* ROM TuCPUڑtO */
#endif
#endif
BOOL extdet_flag;						/* ANZXEXTDETLtO */

/*
 *	ID
 */
#define	ID	MAKEID('S', 'C', 'T', 'L')
//#define SUBCTRL_LOG

/*
 *	TuCPURg[
 *	
 */
BOOL FASTCALL subctrl_init(void)
{
#if XM7_VER == 1
	parityerror = FALSE;
#endif

	return TRUE;
}

/*
 *	TuCPURg[
 *	N[Abv
 */
void FASTCALL subctrl_cleanup(void)
{
}

/*
 *	TuCPURg[
 *	Zbg
 */
void FASTCALL subctrl_reset(void)
{
	LOG0(Normal, "Zbg");

	subhalt_request = FALSE;
	subhalt_flag = FALSE;
	subbusy_flag = TRUE;
	subcancel_request = FALSE;
	subcancel_flag = FALSE;
	subattn_flag = FALSE;
	subreset_flag = FALSE;
	busy_CLR_count = 0;
#if XM7_VER >= 2
	mode320 = FALSE;
#if XM7_VER >= 3
	mode400l = FALSE;
	mode256k = FALSE;
	screen_mode = SCR_200LINE;
	subram_protect = TRUE;
	subreset_halt = FALSE;
	subkanji_flag = FALSE;
	subif_dat = 0x08;
#endif
#endif

#if XM7_VER == 1 && defined(L4CARD)
	select_400line = FALSE;
	enable_400line = FALSE;
	subkanji_flag = FALSE;
#endif

	memset(shared_ram, 0xff, sizeof(shared_ram));
}

/*
 *	TuCPURg[
 *	HALT/CANCEL ANmbW
 */
void FASTCALL subctrl_halt_ack(void)
{
	subhalt_flag = subhalt_request;
	subcancel_flag = subcancel_request;

	/* HALTɂBUSYMONɂ */
	if (subhalt_request) {
		subbusy_flag = TRUE;
	}
}

/*
 *	TuCPURg[
 *	PoCgǂݏo
 */
BOOL FASTCALL subctrl_readb(WORD addr, BYTE *dat)
{
	BYTE ret;

	switch (addr) {
		/* TuCPU AeV荞݁ABreakL[荞 */
		case 0xfd04:
			/* BUSYtO */
			if (subbusy_flag) {
				ret = 0xff;
			}
			else {
				ret = 0x7f;
			}
			/* AeVtO */
			if (subattn_flag) {
				ret &= ~0x01;
				subattn_flag = FALSE;
			}
			/* BreakL[tO */
			if (break_flag || hotreset_flag) {
				ret &= ~0x02;

				/* zbgZbg */
				if (hotreset_flag) {
#if XM7_VER >= 2
					/* CjVG[^ROMȂzbgZbg */
					if (!initrom_en) {
						break_flag = FALSE;
						hotreset_flag = FALSE;
					}
#else
					break_flag = FALSE;
					hotreset_flag = FALSE;
#endif
				}
			}

#if XM7_VER == 1
			/* peBG[tO(FM-8) */
			if ((model_type == MODEL_FM8) && !parityerror) {
				ret &= ~0x04;
			}
#endif

#if XM7_VER == 1
			/* ȉAFM-77[ĥݗL */
			if (model_type == MODEL_FM77) {
#ifdef SFDC
				/* SFDCL`FbN, FDC 1MB[h */
				if (fdc_enable[FDC_S] && fdc_mode) {
					ret &= ~0x40;
				}
#endif	/* defined(SFDC) */

#ifdef L4CARD
				/* ROM؂芷 */
				if (subkanji_flag) {
					ret &= ~0x20;
				}

				/* 400CJ[h */
				if (enable_400linecard) {
					ret &= ~0x10;
				}

				/* 400CJ[h[h */
				if (enable_400linecard && select_400line) {
					ret &= ~0x08;
				}
#endif
			}
#endif	/* XM7_VER == 1 */

			*dat = ret;
			maincpu_firq();
			return TRUE;

		/* TuC^tF[X */
		case 0xfd05:
			ret = 0xff;
			if (!subbusy_flag) {
				ret &= ~0x80;
			}
			if (model_type == MODEL_FM7) {
				if (extdet_flag) {
					if (opn_use ||
						whg_use ||
						thg_use ||
#if XM7_VER == 1
						fmx_flag ||
#endif
#ifndef SFDC
						fdc_enable ||
#else
						fdc_enable[FDC_M] ||
#endif
#ifdef BUBBLE
						bmc_use[BMC_128] ||
#endif
#ifdef MIDI
						midi_use ||
#endif
#ifdef JSUB
						jsub_enable ||
#endif
#ifdef RSC
						rs_use ||
#endif
#ifdef MOUSE
						(mos_port == 3) ||
#endif
						kanji_enable) {
						ret &= ~0x01;
					}
				}
				else {
					if (opn_enable ||
						whg_enable ||
						thg_enable ||
#if XM7_VER == 1
						fmx_flag ||
#endif
#ifdef SFDC
						fdc_enable[FDC_M] ||
#else
						fdc_enable ||
#endif
#ifdef BUBBLE
						bmc_enable[BMC_128] ||
#endif
#ifdef MIDI
						midi_enable ||
#endif
#ifdef JSUB
						jsub_enable ||
#endif
#ifdef RSC
						rs_use ||
#endif
#ifdef MOUSE
						(mos_port == 3) ||
#endif
						kanji_enable) {
						ret &= ~0x01;
					}
				}
			}
#if XM7_VER == 1
			else if (model_type == MODEL_FM8) {
				if (extdet_flag) {
#ifdef SFDC
					if (fdc_enable[FDC_M] ||
						fdc_enable[FDC_S] ||
#else
					if (fdc_enable ||
#endif
#ifdef BUBBLE
						bmc_use[BMC_128] ||
#endif
#ifdef MIDI
						midi_use ||
#endif
						FALSE) {
						ret &= ~0x01;
					}
				}
				else {
#ifdef SFDC
					if (fdc_enable[FDC_M] ||
						fdc_enable[FDC_S] ||
#else
					if (fdc_enable ||
#endif
#ifdef BUBBLE
						bmc_enable[BMC_128] ||
#endif
#ifdef MIDI
						midi_enable ||
#endif
						FALSE) {
						ret &= ~0x01;
					}
				}
			}
#endif
			else {
				ret &= ~0x01;
			}
			*dat = ret;
			return TRUE;

		/* Tu[hXe[^X */
#if XM7_VER >= 2
		case 0xfd12:
			if (model_type < MODEL_FM77AV) {
				return FALSE;
			}
			ret = 0xff;
			/* 320/640 */
			if (!mode320) {
				ret &= ~0x40;
			}
			/* uNXe[^X */
			if (blank_flag) {
				ret &= ~0x02;
			}
			/* VSYNCXe[^X */
			if (!vsync_flag) {
				ret &= ~0x01;
			}
			*dat = ret;
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	TuCPURg[
 *	PoCg
 */
BOOL FASTCALL subctrl_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
#if XM7_VER >= 3
		/* TuC^tF[X AV40g */
		case 0xfd04:
			/* AV40/AV40EX̂ݗL */
			if (model_type < MODEL_FM77AV40) {
				return FALSE;
			}

			/* bit2:Tuj^veNg */
			if (dat & 0x04) {
				subram_protect = FALSE;
			}
			else {
				subram_protect = TRUE;
			}
			/* bit3:400C[hI */
			if (dat & 0x08) {
				mode400l = FALSE;
			}
			else {
				mode400l = TRUE;
			}
			/* bit4:26F[hI */
			/*      bit34̗ONꍇAmode400lD */
			if ((dat & 0x10) && !mode400l) {
				mode256k = TRUE;
			}
			else {
				mode256k = FALSE;
			}
			/* bit5:ROMڑ؂芷 */
			if (dat & 0x20) {
				subkanji_flag = FALSE;
			}
			else {
				subkanji_flag = TRUE;
			}

			/* ʃ[h؂ւꍇĕ` */
			if ((BYTE)(dat & 0x18) != subif_dat) {
				display_setpointer(TRUE);
			}
			subif_dat = (BYTE)(dat & 0x18);
			return TRUE;
#endif

#if XM7_VER == 1
		/* TuC^tF[X FM-77g */
		case 0xfd04:
			/* ȉAFM-77[ĥݗL */
			if (model_type != MODEL_FM77) {
				return FALSE;
			}

#ifdef SFDC
			/* bit2:荞ݐ؂芷 */
			if (dat & 0x04) {
				sfd_firq_mode = FALSE;
			}
			else {
				sfd_firq_mode = TRUE;
			}
#endif	/* SFDC */

#ifdef L4CARD
			/* bit3:400CJ[h[h */
			if (dat & 0x08) {
				select_400line = FALSE;
			}
			else {
				select_400line = TRUE;
			}

			/* bit5:ROMڑ؂芷 */
			if (dat & 0x20) {
				subkanji_flag = FALSE;
			}
			else {
				subkanji_flag = TRUE;
			}
#endif

#ifdef SFDC
			/* bit6:320KB/1MB FDC؂芷 */
			if (dat & 0x40) {
				/* 320KB FDC[h */
				fdc_mode = 0;
			}
			else {
				/* SFDCgptOI */
				fdc_use[FDC_S] = TRUE;
				/* SFDCL`FbN */
				if (fdc_enable[FDC_S]) {
					/* 1MB FDC[h */
					fdc_mode = 1;
				}
			}
#endif	/* SFDC */
			return TRUE;
#endif	/* XM7_VER == 1 */

		/* TuRg[ */
		case 0xfd05:
#if XM7_VER == 1 && defined(Z80CARD)
			if (dat & 0x01) {
				if (!main_z80mode) {
					mmr_modify = TRUE;
				}
				main_z80mode = TRUE;
			}
			else {
				if (main_z80mode) {
					mmr_modify = TRUE;
				}
				main_z80mode = FALSE;
			}
#endif
			if (dat & 0x80) {
				/* TuHALTNGXg */
				subhalt_request = TRUE;
			}
			else {
				/* TuRUNNGXg */
				subhalt_request = FALSE;

				/* oN؂ւ́AZbg */
#if XM7_VER >= 3
				if (subreset_halt) {
					subcpu_reset();
					subreset_halt = FALSE;
				}
#endif
			}
			if (dat & 0x40) {
				/* LZIRQ */
				subcancel_request = TRUE;
			}
			subcpu_irq();
			return TRUE;

		/* Tu[h؂ւ */
#if XM7_VER >= 2
		case 0xfd12:
			if (model_type < MODEL_FM77AV) {
				return FALSE;
			}

			/* ʃ[h؂ւꍇĕ` */
			if (dat & 0x40) {
				if (!mode320) {
					mode320 = TRUE;
					display_setpointer(TRUE);
				}
			}
			else {
				if (mode320) {
					mode320 = FALSE;
					display_setpointer(TRUE);
				}
			}
			return TRUE;

		/* TuoN؂ւ */
		case 0xfd13:
			if (model_type < MODEL_FM77AV) {
				return FALSE;
			}

			/* oN؂ւ */
#if XM7_VER >= 3
			/* AV40/AV40EX */
			if ((model_type >= MODEL_FM77AV40) &&
				(dat & 0x04)) {
				subrom_bank = 4;
			}
			else {
				subrom_bank = (BYTE)(dat & 0x03);
			}
#else
			subrom_bank = (BYTE)(dat & 0x03);
#endif

			/* Zbg */
#if XM7_VER >= 3
			if (!subhalt_flag) {
				subcpu_reset();
				subreset_halt = FALSE;
			}
			else {
				subreset_halt = TRUE;
			}
#else
			subcpu_reset();
#endif

			/* tOރZbg */
			subreset_flag = TRUE;
			subbusy_flag = TRUE;

			/* \nI/OZbg (FM77AV fvO) */
			aluline_reset();
			ttlpalet_reset();
			multipag_reset();

			/* CRTWX^Zbg */
			display_reset();
			display_notify();

			/* INS LED */
			ins_flag = FALSE;
			return TRUE;
#endif

#if XM7_VER >= 3
		/* CRTC(MB89321) */
		case 0xfd96:
		case 0xfd97:
			/* AV40̂ݗL */
			if (model_type == MODEL_FM77AV40) {
				/* ۂɂ͋@\Ȃ */
				return TRUE;
			}
			return FALSE;
#endif
	}

	return FALSE;
}

/*
 *	TuCPURg[
 *	Z[u
 */
BOOL FASTCALL subctrl_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, subhalt_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, subbusy_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, subcancel_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, subattn_flag)) {
		return FALSE;
	}

	if (!file_write(fileh, shared_ram, 0x80)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, subreset_flag)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_bool_write(fileh, mode320)) {
		return FALSE;
	}
#endif

	/* Ver7g */
	if (!file_bool_write(fileh, subhalt_request)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subcancel_request)) {
		return FALSE;
	}

	/* Ver711g */
	if (!file_byte_write(fileh, busy_CLR_count)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if (!file_bool_write(fileh, mode400l)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, mode256k)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subram_protect)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subreset_halt)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subkanji_flag)) {
		return FALSE;
	}
#endif

#if XM7_VER == 1 && defined(L4CARD)
	/* XM7 V1.1 / FM-77L4g */
	if (!file_bool_write(fileh, select_400line)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subkanji_flag)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	TuCPURg[
 *	[h
 */
BOOL FASTCALL subctrl_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_bool_read(fileh, &subhalt_flag)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &subbusy_flag)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &subcancel_flag)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &subattn_flag)) {
		return FALSE;
	}

	if (!file_read(fileh, shared_ram, 0x80)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &subreset_flag)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_bool_read(fileh, &mode320)) {
		return FALSE;
	}
#endif

	/* Ver7g */
#if XM7_VER >= 3
	if (((ver >= 700) && (ver <= 799)) || (ver >= 900)) {
#elif XM7_VER >= 2
	if (ver >= 700) {
#else
	if (ver >= 300) {
#endif
		if (!file_bool_read(fileh, &subhalt_request)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &subcancel_request)) {
			return FALSE;
		}
	}
	else {
		subhalt_request = subhalt_flag;
		subcancel_request = subcancel_flag;
	}

	/* Ver711g */
#if XM7_VER >= 3
	if (((ver >= 711) && (ver <= 799)) || (ver >= 911)) {
#elif XM7_VER >= 2
	if (ver >= 711) {
#else
	if (ver >= 300) {
#endif
		if (!file_byte_read(fileh, &busy_CLR_count)) {
			return FALSE;
		}
	}
	else {
		busy_CLR_count = 0;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if ((ver < 800) || ((ver >= 1000) && (ver < 1400))) {
		mode400l = FALSE;
		mode256k = FALSE;
		subif_dat = 0x08;
		subram_protect = FALSE;
		subreset_halt = FALSE;

		/* |C^\ */
		display_setpointer(TRUE);
		return TRUE;
	}

	if (!file_bool_read(fileh, &mode400l)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &mode256k)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &subram_protect)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &subreset_halt)) {
		return FALSE;
	}

	/* Ver9.06g */
	if (ver >= 906) {
		if (!file_bool_read(fileh, &subkanji_flag)) {
			return FALSE;
		}
	}
	else {
		if (mode400l) {
			subkanji_flag = TRUE;
		}
		else {
			subkanji_flag = FALSE;
		}
	}

	/* TuI/F Reg. ʃ[hA */
	subif_dat = 0x00;
	if (!mode400l) {
		subif_dat |= (BYTE)0x08;
	}
	if (mode256k) {
		subif_dat |= (BYTE)0x10;
	}
	/* AV40/AV40EXȊO */
	if ((model_type != MODEL_FM77AV40) &&
		(model_type != MODEL_FM77AV40EX)) {
		if (mode400l || mode256k || subkanji_flag) {
			return FALSE;
		}
	}
#endif

#if XM7_VER == 1 && defined(L4CARD)
	/* XM7 V1.1 / FM-77L4g */
	if (!file_bool_read(fileh, &select_400line)) {
		return FALSE;
	}
	/* 400CJ[h`FbN */
	if (select_400line && !detect_400linecard) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &subkanji_flag)) {
		return FALSE;
	}
#endif

	/* |C^\ */
	display_setpointer(TRUE);

	return TRUE;
}
