/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ Win32API VA|[g ]
 */

#ifdef _WIN32

#ifdef RSC

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <imm.h>
#include <shellapi.h>
#include <stdlib.h>
#include <assert.h>
#include <mmsystem.h>
#include <process.h>
#include "xm7.h"
#include "mainetc.h"
#include "device.h"
#include "rs232c.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_comm.h"


/*
 *	O[o [N
 */
int nCommPortNo;						/* gpVA|[gԍ */

/*
 *	X^eBbN [N
 */
static CRITICAL_SECTION CSection;		/* NeBJZNV */
static HANDLE hThreadRcv;				/* MXbhnh */
static UINT uThResult;					/* Xbh߂l */
static BOOL bCommCloseReq;				/* XbhIvtO */
static BOOL bCriticalSection;

static HANDLE hComm;					/* VA|[gnh */
static int nOpenPort;					/* gp̃|[gԍ */
static COMMPROP commprop;				/* */
static OVERLAPPED Overlap;				/* OVERLAPPED\ */
static DCB dcb;							/* DCB\ */
static COMMTIMEOUTS CommTimeout;		/* TIMEOUT\ */

extern BYTE rs_status;					/* Xe[^XWX^ */
static BYTE uRsRcvData[4096];			/* Mf[^obt@ */
static DWORD dwRsRcvCount;				/* Mf[^TCY */
static DWORD dwRsRcvPoint;				/* Mf[^ǂݍ݈ʒu */


/*
 *	vg^Cv錾
 */
static UINT WINAPI ThreadRcv(LPVOID);	/* Xbh֐ */

/*
 *	ʐMxe[u
 */
static const DWORD dwBaudrateSlow[8] = {
	CBR_300,	CBR_600,	CBR_1200,	CBR_2400,
	CBR_4800,	CBR_9600,	CBR_19200,	CBR_38400
};
static const DWORD dwBaudrateFast[8] = {
	CBR_1200,	CBR_2400,	CBR_4800,	CBR_9600,
	CBR_19200,	CBR_38400,	CBR_38400,	CBR_38400
};

/*
 *	VA|[g
 *	
 */
BOOL FASTCALL InitCommPort(void)
{
	char buf[256];
	char string[256];
	char comportname[256];

	/* ܂NA */
	if (hComm != INVALID_HANDLE_VALUE) {
		CloseHandle(hComm);
	}
	hComm = INVALID_HANDLE_VALUE;
	rs_mask = TRUE;
	bCriticalSection = FALSE;

	/* rs_useFALSEȂVA|[g͎gpȂ */
	if (!rs_use) {
		nOpenPort = 0;
		return FALSE;
	}

	/* nCommPortNo`FbN */
	if (nCommPortNo == 0) {
		nOpenPort = nCommPortNo;
		rs_mask = FALSE;
		return TRUE;
	}

	/* Ƃ肠AĂ݂ */
	if (nCommPortNo <= 9) {
		_snprintf(comportname, sizeof(comportname), "COM%d", nCommPortNo);
	}
	else {
		_snprintf(comportname, sizeof(comportname),
			"\\\\.\\COM%d", nCommPortNo);
	}
	hComm = CreateFile(comportname,
				GENERIC_READ | GENERIC_WRITE, 0, NULL,
				OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);

	/* ΉfoCX`FbN */
	if (hComm != INVALID_HANDLE_VALUE) {
		commprop.wPacketLength = sizeof(commprop);
		GetCommProperties(hComm, &commprop);
	}
	else {
		/* CreateFileɎsĂ̂ŃfoCXʂK */
		commprop.dwProvSubType = 0xffffffff;
	}
	if ((commprop.dwProvSubType != PST_RS232) &&
		(commprop.dwProvSubType != PST_MODEM) &&
		(commprop.dwProvSubType != PST_UNSPECIFIED)) {
		LoadString(hResInstance, IDS_COMM_OPENERROR, buf, sizeof(buf));
		_snprintf(string, sizeof(string), buf, comportname);
		MessageBox(hMainWnd, string, "XM7", MB_OK | MB_ICONERROR);

		if (hComm != INVALID_HANDLE_VALUE) {
			CloseHandle(hComm);
			hComm = INVALID_HANDLE_VALUE;
		}
		nOpenPort = 0;

		return FALSE;
	}

	/* ^CAEgԂݒ */
	CommTimeout.ReadIntervalTimeout = 0;
	CommTimeout.ReadTotalTimeoutMultiplier = 0;
	CommTimeout.ReadTotalTimeoutConstant = 0;
	CommTimeout.WriteTotalTimeoutMultiplier = 0;
	CommTimeout.WriteTotalTimeoutConstant = 0;
	SetCommTimeouts(hComm,&CommTimeout);

	/* ݒ */
	dcb.DCBlength = sizeof(dcb);
	GetCommState(hComm, &dcb);

	dcb.BaudRate = CBR_1200;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fErrorChar = TRUE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.fAbortOnError = FALSE;
	dcb.wReserved = 0;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	dcb.ErrorChar = '?';
	SetCommState(hComm, &dcb);

	/* MXbh쐬 */
	hThreadRcv = (HANDLE)_beginthreadex(NULL, 0, ThreadRcv, 0, 0, &uThResult);
	if (!hThreadRcv) {
		LoadString(hResInstance, IDS_COMM_THREADERROR, string, sizeof(string));
		MessageBox(hMainWnd, string, "XM7", MB_OK | MB_ICONERROR);

		CloseHandle(hComm);
		hComm = INVALID_HANDLE_VALUE;
		rs_mask = TRUE;
		nOpenPort = 0;

		return FALSE;
	}

	/* I[v|[gԍێ */
	nOpenPort = nCommPortNo;
	rs_mask = FALSE;

	return TRUE;
}

/*
 *	VA|[g
 *	N[Abv
 */
void FASTCALL CleanCommPort(void)
{
	DWORD dwExitCode;

	/* XbhIĂȂ΁AI点 */
	while (hThreadRcv && !uThResult) {
		bCommCloseReq = TRUE;
		while (TRUE) {
			GetExitCodeThread(hThreadRcv, &dwExitCode);
			if (dwExitCode == STILL_ACTIVE) {
				WaitForSingleObject(hThreadRcv, 10);
			}
			else {
				CloseHandle(hThreadRcv);
				hComm = INVALID_HANDLE_VALUE;
				rs_mask = TRUE;
				break;
			}
		}
	}
}

/*
 *	ZNgԃ`FbN
 */
BOOL FASTCALL SelectCheckCommPort(int port)
{
	/* gƂĂ|[gƎgp̃|[gvĂȂ`FbN */
	if (nOpenPort != port) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	ZbgfoCXZNg
 */
void FASTCALL rs232c_reset_notify(void)
{
	if (nCommPortNo == 0) {
		/* ߂ď */
		InitCommPort();
		return;
	}

	/* gƂĂ|[gƎgp̃|[gvĂȂ`FbN */
	if (nOpenPort != nCommPortNo) {
		/* |[gJĂꍇ͂N[Abv */
		if (nOpenPort) {
			CleanCommPort();
		}

		/* ߂ď */
		InitCommPort();
	}
}


/*
 *	bN
 */
void FASTCALL LockRS(void)
{
	if (bCriticalSection) {
		EnterCriticalSection(&CSection);
	}
}

/*
 *	AbN
 */
void FASTCALL UnlockRS(void)
{
	if (bCriticalSection) {
		LeaveCriticalSection(&CSection);
	}
}

/*
 *	f[^M
 */
void FASTCALL rs232c_senddata(BYTE dat)
{
	DWORD dwErrors;
	DWORD dwTransfer;
	COMSTAT comstat;

	/* RS-232C̏ꍇ͉Ȃ */
	if (rs_mask || hComm == INVALID_HANDLE_VALUE) {
		return;
	}

	/* TxE=1,CTS=L̏ꍇAM\ */
	if (!rs_cts && (rs_command & RSC_TXEN) && (rs_status & RSS_TXRDY)) {
		/* |[OXbhbN */
		/* LockRS(); */	/* Ȃc */

		/* TxEMPTY/TxRDY𗎂Ƃ */
		rs232c_txrdy(FALSE);

		/* 1oCgM */
		ClearCommError(hComm, &dwErrors, &comstat);
		if (!WriteFile(hComm, &dat, 1, NULL, &Overlap)) {
			if (GetLastError() == ERROR_IO_PENDING) {
				GetOverlappedResult(hComm, &Overlap, &dwTransfer, TRUE);
			}
		}

		/* TxRDY ON/TxEMPTYv */
		rs232c_txempty_request();

		/* |[OXbhAbN */
		/* UnlockRS(); */	/* Ȃc */
	}
}

/*
 *	f[^M
 */
BYTE FASTCALL rs232c_receivedata(void)
{
	BYTE dat;

	/* RS-232C̏ꍇ͉Ȃ */
	if (rs_mask || hComm == INVALID_HANDLE_VALUE) {
		return 0xff;
	}

	/* |[OXbhbN */
	/* LockRS(); */	/* ȂCc */

	/* obt@f[^肾 */
	dat = uRsRcvData[dwRsRcvPoint];

	/* RxE=1̏ꍇAM\ */
	if (rs_command & RSC_RXE) {
		dwRsRcvPoint ++;

		/* f[^܂ꍇARxRDYv */
		/* (MXbhdwRsRcvCountωĂCɂȂ) */
		if (dwRsRcvPoint < dwRsRcvCount) {
			rs232c_rxrdy_request();
		}
		else {
			uRsRcvData[0] = uRsRcvData[dwRsRcvPoint];
			dwRsRcvCount = 0;
			dwRsRcvPoint = 0;
		}
	}

	/* RxRDY𗎂Ƃ */
	rs232c_rxrdy(FALSE);

	/* |[OXbhAbN */
	/* UnlockRS(); */	/* ȂCc */

	return dat;
}

/*
 *	Xe[^XWX^ǂݏo
 */
BYTE FASTCALL rs232c_readstatus(void)
{
	/* ł͒PɃXe[^XWX^̓eԂ */
	return rs_status;
}

/*
 *	[hR}hWX^
 */
void FASTCALL rs232c_writemodecmd(BYTE dat)
{
	/* RS-232C̏ꍇ͉Ȃ */
	if (rs_mask || hComm == INVALID_HANDLE_VALUE) {
		return;
	}

	/* |[OXbhbN */
	LockRS();

	/* ݂̐ݒ擾 */
	dcb.DCBlength = sizeof(dcb);
	GetCommState(hComm, &dcb);

	/* Xgbvrbg */
	switch (dat & RSM_STOPBITM) {
		case RSM_STOPBIT1	:	dcb.StopBits = ONESTOPBIT;
								break;
		case RSM_STOPBIT15	:	dcb.StopBits = ONE5STOPBITS;
								break;
		case RSM_STOPBIT2	:	dcb.StopBits = TWOSTOPBITS;
								break;
	}

	/* peB */
	if (dat & RSM_PARITYEN) {
		dcb.fParity = TRUE;
		if (dat & RSM_PARITYEVEN) {
			dcb.Parity = EVENPARITY;
		}
		else {
			dcb.Parity = ODDPARITY;
		}
	}
	else {
		dcb.fParity = FALSE;
		dcb.Parity = NOPARITY;
	}

	/* LN^ */
	switch (dat & RSM_CHARLENM) {
		case RSM_CHARLEN5	:	dcb.ByteSize = 5;
								break;
		case RSM_CHARLEN6	:	dcb.ByteSize = 6;
								break;
		case RSM_CHARLEN7	:	dcb.ByteSize = 7;
								break;
		case RSM_CHARLEN8	:	dcb.ByteSize = 8;
								break;
	}

	/* NbN */
	if ((dat & RSM_BAUDDIVM) == RSM_BAUDDIV64) {
		dcb.BaudRate = dwBaudrateSlow[(rs_baudrate & RSCB_BAUDM) >> 2];
	}
	else {
		dcb.BaudRate = dwBaudrateFast[(rs_baudrate & RSCB_BAUDM) >> 2];
	}

	/* ݒ */
	SetCommState(hComm, &dcb);

	/* |[OXbhAbN */
	UnlockRS();
}

/*
 *	R}hWX^
 */
void FASTCALL rs232c_writecommand(BYTE dat)
{
	/* RS-232C̏ꍇ͉Ȃ */
	if (rs_mask || hComm == INVALID_HANDLE_VALUE) {
		return;
	}

	/* |[OXbhbN */
	LockRS();

	/* G[NA */
	if (dat & RSC_ER) {
		rs_status &= (BYTE)~(RSS_FRAMEERR | RSS_OVERRUN | RSS_PARITYERR);
	}

	/* RTS */
	if (dat & RSC_RTS) {
		EscapeCommFunction(hComm, SETRTS);
	}
	else {
		EscapeCommFunction(hComm, CLRRTS);
	}

	/* u[N */
	if (dat & RSC_SBRK) {
		EscapeCommFunction(hComm, SETBREAK);
	}
	else {
		EscapeCommFunction(hComm, CLRBREAK);
	}

	/* DTR */
	if ((dat & RSC_DTR) && !rs_dtrmask) {
		EscapeCommFunction(hComm, SETDTR);
	}
	else {
		EscapeCommFunction(hComm, CLRDTR);
	}

	/* |[OXbhAbN */
	UnlockRS();
}

/*
 *	NbNE{[[gݒ背WX^
 */
void FASTCALL rs232c_setbaudrate(BYTE dat)
{
	/* RS-232C̏ꍇ͉Ȃ */
	if (rs_mask || hComm == INVALID_HANDLE_VALUE) {
		return;
	}

	/* |[OXbhbN */
	LockRS();

	/* ݂̐ݒ擾 */
	dcb.DCBlength = sizeof(dcb);
	GetCommState(hComm, &dcb);

	/* {[[gݒ */
	if ((rs_modecmd & RSM_BAUDDIVM) == RSM_BAUDDIV64) {
		dcb.BaudRate = dwBaudrateSlow[(dat & RSCB_BAUDM) >> 2];
	}
	else {
		dcb.BaudRate = dwBaudrateFast[(dat & RSCB_BAUDM) >> 2];
	}

	/* ݒ */
	SetCommState(hComm, &dcb);

	/* |[OXbhAbN */
	UnlockRS();
}

/*
 *	Xbh֐
 */
static UINT WINAPI ThreadRcv(LPVOID param)
{
	DWORD dwEvent;
	DWORD dwErrors;
	DWORD dwStat;
	DWORD dwTransfer;
	DWORD dwRcvCount;
	COMSTAT comstat;

	UNUSED(param);

	/* NeBJZNV쐬 */
	InitializeCriticalSection(&CSection);

	/* [NGA */
	bCommCloseReq = FALSE;
	rs_status |= (RSS_TXRDY | RSS_TXEMPTY);
	dwRsRcvCount = 0;
	dwRsRcvPoint = 0;

	/* 񓯊I/OpI[o[bv\̂ */
	memset(&Overlap, 0, sizeof(Overlap));
	Overlap.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

	/* Cxg}XNݒ */
	SetCommMask(hComm,
		EV_BREAK | EV_CTS | EV_DSR | EV_ERR | EV_RLSD | EV_RXCHAR | EV_TXEMPTY);

	/* NeBJZNV䋖 */
	bCriticalSection = TRUE;

	while (!bCloseReq) {
		WaitCommEvent(hComm, &dwEvent, &Overlap);
		if (GetLastError() == ERROR_IO_PENDING) {
			GetOverlappedResult(hComm, &Overlap, &dwTransfer, TRUE);
			ClearCommError(hComm, &dwErrors, &comstat);

			/* u[No */
			if (dwEvent & EV_BREAK) {
				if (rs_command & RSC_EH) {
					rs232c_syndet(TRUE);
				}
			}

			/* CTSMω */
			if (dwEvent & EV_CTS) {
				GetCommModemStatus(hComm, &dwStat);
				if (dwStat & MS_CTS_ON) {
					rs_cts = TRUE;
				}
				else {
					rs_cts = FALSE;
				}
			}

			/* DSRMω */
			if (dwEvent & EV_DSR) {
				GetCommModemStatus(hComm, &dwStat);
				if (dwStat & MS_DSR_ON) {
					rs_status |= RSS_DSR;
				}
				else {
					rs_status &= (BYTE)~RSS_DSR;
				}
			}

			/* G[ */
			if (dwEvent & EV_ERR) {
				if (dwErrors & CE_FRAME) {
					rs_status |= RSS_FRAMEERR;
				}
				if (dwErrors & CE_RXOVER) {
					rs_status |= RSS_OVERRUN;
				}
				if (dwErrors & CE_RXPARITY) {
					rs_status |= RSS_PARITYERR;
				}
			}

			/* CDMω */
			if (dwEvent & EV_RLSD) {
				GetCommModemStatus(hComm, &dwStat);
				if (dwStat & MS_RLSD_ON) {
					rs_cd = TRUE;
				}
				else {
					rs_cd = FALSE;
				}
			}

			/* f[^M */
			if (dwEvent & EV_RXCHAR) {
				if (!ReadFile(hComm, &uRsRcvData[dwRsRcvCount],
					comstat.cbInQue, &dwRcvCount, &Overlap)) {
					if (GetLastError() == ERROR_IO_PENDING) {
						GetOverlappedResult(hComm, &Overlap, &dwTransfer,
							TRUE);
					}
				}

				/* Mf[^ꍇAIRQ𔭐 */
				dwRsRcvCount += dwRcvCount;
				if ((dwRsRcvCount > 0) && (rs_command & RSC_RXE)) {
					rs232c_rxrdy(TRUE);
				}
			}

			/* f[^M */
			if (dwEvent & EV_TXEMPTY) {
				rs_status |= RSS_TXEMPTY;
			}
		}

		Sleep(1);
	}

	/* XbhIvtO~낷 */
	bCommCloseReq = FALSE;

	/* I[o[bvCxg폜 */
	DeleteObject(Overlap.hEvent);

	/* NeBJZNV폜 */
	DeleteCriticalSection(&CSection);
	bCriticalSection = FALSE;

	/* XbhI */
	_endthreadex(TRUE);

	return TRUE;
}

#endif	/* RSC */
#endif	/* _WIN32 */
