/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *							ROMEO support by usalin
 *
 *	[ Win32API TEh ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#define DIRECTSOUND_VERSION		0x300	/* DirectX3w */
#include <dsound.h>
#include <assert.h>
#include <math.h>
#include <objbase.h>
#include <tchar.h>
#include "xm7.h"
#include "device.h"
#include "mainetc.h"
#include "opn.h"
#include "tapelp.h"
#include "cisc.h"
#include "opna.h"
#include "psg.h"
#include "keyboard.h"
#include "w32.h"
#include "w32_sch.h"
#include "w32_snd.h"
#include "juliet.h"
#include "w32_midi.h"
#ifdef FMDLL
#include "fmdll.h"
#endif
#include "scci.h"
#include "SCCIDefines.h"

/*
 *	O[o [N
 */
UINT nSampleRate;						/* TvO[g */
UINT nSoundBuffer;						/* TEhobt@TCY */
UINT uStereoOut;						/* o̓[h */
BOOL bInterpolation;					/* TEhԃ[h */
BOOL bForceStereo;						/* XeIo */
UINT nBeepFreq;							/* BEEPg */
BOOL bTapeMon;							/* e[vj^ */
int hWavCapture;						/* WAVLv`nh */
BOOL bWavCapture;						/* WAVLv`Jn */
BOOL bUseRomeo;							/* ݂gptO */
BOOL bUseSCCI;							/* SCCIgptO */
int nSCCI_ChipInfo[4];					/* SCCI TEh`bv */
DWORD dwSoundTotal;						/* TEhg[^ */
WORD nChannelMask[4];					/* `l}XN */
UINT uChSeparation;						/* `lZp[V */
int nFMVolume;							/* FM{[ */
int nPSGVolume;							/* PSG{[ */
int nBeepVolume;						/* BEEP{[ */
int nCMTVolume;							/* CMTj^{[ */
int nWaveVolume;						/* eʉ{[ */
BOOL bGlobalFocusFlag;					/* O[otH[JXtO */
#ifdef FMDLL
BOOL bFmDll;							/* ODLLFtO */
BOOL bUseFmDll;							/* ODLLgptO */
char szFmDll[_MAX_FNAME];				/* ODLLt@C */
#endif
int pan_set[5][4];						/* p` */
const int pan_set_def[5][4] = {			/* p`ftHg */
	/*	STD		WHG		THG		PSG		*/
	{	PAN_C,	PAN_C,	PAN_C,	PAN_C,	},	/* m */
	{	PAN_L,	PAN_R,	PAN_C,	PAN_C,	},	/* `#1 */
	{	PAN_R,	PAN_L,	PAN_C,	PAN_C,	},	/* `#2 */
	{	PAN_C,	PAN_R,	PAN_L,	PAN_C,	},	/* `#3 */
	{	PAN_C,	PAN_L,	PAN_R,	PAN_C,	},	/* `#4 */
};


/*
 *	X^eBbN [N
 */
static LPDIRECTSOUND lpds;				/* DirectSound */
static LPDIRECTSOUNDBUFFER lpdsp;		/* DirectSoundBuffer(vC}) */
static LPDIRECTSOUNDBUFFER lpdsb;		/* DirectSoundBuffer(ZJ_) */
static DWORD *lpsbuf;					/* TEh쐬obt@ */
static BOOL bNowBank;					/* ݍĐ̃oN */
static UINT uBufSize;					/* TEhobt@TCY */
static UINT uRate;						/* [g */
static UINT uTick;						/* obt@TCY̒ */
static BOOL bMode;						/* TEhԃ[h */
static UINT uStereo;					/* o̓[h */
static UINT uSample;					/* TvJE^ */
static UINT uBeep;						/* BEEPg`JE^ */
#if XM7_VER >= 2
static UINT uKeyEncBeep;				/* L[GR[_BEEPg`JE^ */
#endif
#ifdef OPNAMODE
static FM::OPNA *pOPN[4];				/* OPNAfoCX */
#else
static FM::OPN *pOPN[4];				/* OPNfoCX */
#endif
#ifdef FMDLL
static CFMDLL	*pFmDll;
static LPVOID	*pOPNA[4];				/* OPNAfoCX */
static int32	*pMix;
#endif
static int nScale[3];					/* OPNvXP[ */
static BYTE uCh3Mode[3];				/* OPN Ch.3[h */
static HMODULE hSCCI;					/* SCCI DLLnh */
static SoundInterfaceManager *pManager;	/* SCCI TEhI/F}l[W */
static SoundChip *pSoundChip[4];		/* SCCI TEh`bv */
static BOOL bInitFlag;					/* tO */
static WORD *pWavCapture;				/* Lv`obt@(64KB) */
static UINT nWavCapture;				/* Lv`obt@f[^ */
static DWORD dwWavCapture;				/* Lv`t@CTCY */
static WORD uChannels;					/* o̓`l */
static BOOL bBeepFlag;					/* BEEPo */
#if XM7_VER >= 2
static BOOL bKeyEncBeepFlag;			/* L[GR[_BEEPo */
#endif
static int nBeepLevel;					/* BEEPo̓x */
static int nCMTLevel;					/* CMTj^o̓x */
static int nWaveLevel;					/* eʉo̓x */
static BOOL bTapeFlag;					/* ݂̃e[vo͏ */
static BOOL bTapeFlag2;					/* Õe[vo͏ */
static BYTE uTapeDelta;					/* e[vg`ԃJE^ */

static BOOL bRomeoMute;					/* ݂~[gtO */
static BOOL bSCCIMute;					/* SCCI~[gtO */

static BOOL bGlobalFlag;				/* O[otH[JXtO */
#ifdef FMDLL
static BOOL bUseDll;					/* ODLLgptO */
static uint uMask;						/* ODLLgp̃}XN */
#endif

#if defined(LPRINT) && defined(JASTSOUND)
static BYTE *pDACbuf;					/* WXgTEh f[^obt@ */
static DWORD dwDACptr;					/* WXgTEh obt@|C^ */
#endif

#ifndef FMGEN_DIRECT_OUTPUT
/*
 *	X^eBbN [N (FM݃L[)
 */
typedef struct _FMqueue {
	WORD	reg;						/* WX^ԍ(glWORD^) */
	BYTE	dat;						/* ݃f[^ */
	BYTE	padding;					/* pfBO */
} FMqueue;

static FMqueue OPNqueue[FMQUEUE_SIZE];	/* OPNWX^݃L[ */
static int uOPNqueuePtr;				/* OPNWX^݃L[|C^ */
static FMqueue WHGqueue[FMQUEUE_SIZE];	/* WHGWX^݃L[ */
static int uWHGqueuePtr;				/* WHGWX^݃L[|C^ */
static FMqueue THGqueue[FMQUEUE_SIZE];	/* THGWX^݃L[ */
static int uTHGqueuePtr;				/* THGWX^݃L[|C^ */
static FMqueue PSGqueue[FMQUEUE_SIZE];	/* PSGWX^݃L[ */
static int uPSGqueuePtr;				/* PSGWX^݃L[|C^ */
#endif

/*
 *	X^eBbN [N (WAVĐ)
 */
static struct _WAVDATA {
	short *p;							/* g`f[^|C^ */
	DWORD size;							/* f[^TCY(Tv) */
	DWORD freq;							/* TvOg */
} Wav[3];

static struct _WAVPLAY {
	BOOL	bPlay;						/* WAVĐtO */
	DWORD	dwWaveNo;					/* WAVŁ[Ȃ΁[ */
	DWORD	dwCount1;					/* WAVŁ[() */
	DWORD	dwCount2;					/* WAVŁ[() */
	DWORD	dwCount3;					/* WAVĐ */
} WavP[SNDBUF];

static LPSTR WavName[] = {				/* WAVt@C */
	"RELAY_ON.WAV",
	"RELAYOFF.WAV",
	"FDDSEEK.WAV",
#if 0
	"HEADUP.WAV",
	"HEADDOWN.WAV",
#endif
};

/* XeIo͎̍EoXe[u */
static struct PAN_SET {
	int	l;
	int	r;
} vol_set[] = {
	{	30,	2	},
	{	23,	9	},
	{	16,16	},
	{	 9,23	},
	{	 2,30	},
};

/*
 *	vg^Cv錾
 */
#ifdef __cplusplus
extern "C" {
#endif
void (*CopySoundBuffer)(DWORD *src, WORD *dst, int count);
#ifndef _M_AMD64
extern void CopySndBufMMX(DWORD *src, WORD *dst, int count);
#endif
extern void CopySndBuf(DWORD *src, WORD *dst, int count);
#ifdef __cplusplus
}
#endif

/*
 *	~[gp`l}XNrbg
 */
static const WORD nChannelMaskBit[6] = {
	0x000001, 0x000002, 0x000004, 0x000040, 0x000080, 0x000100
};

/*
 *	`
 */
#ifdef OPNAMODE
#define ENABLE_ROMEO_OPN	(bUseRomeo)
#define ENABLE_ROMEO_WHG	(0)
#else
#define ENABLE_ROMEO_OPN	(bUseRomeo)
#define ENABLE_ROMEO_WHG	(bUseRomeo)
#endif
#define ENABLE_SCCI_OPN	(pSoundChip[OPN_STD] && bUseSCCI)
#define ENABLE_SCCI_WHG	(pSoundChip[OPN_WHG] && bUseSCCI)
#define ENABLE_SCCI_THG	(pSoundChip[OPN_THG] && bUseSCCI)
#define ENABLE_SCCI_PSG	(pSoundChip[OPN_PSG] && bUseSCCI)

/*
 *	
 */
void FASTCALL InitSnd(void)
{
	int i;

	/* [NGA */
	nSampleRate = 44100;
	nSoundBuffer = 100;
	bInterpolation = FALSE;
	uChSeparation = 9;
	nFMVolume = 0;
	nPSGVolume = -2;
	nBeepVolume = -24;
	nCMTVolume = -24;
	nWaveVolume = 0;
	nBeepFreq = 1200;
	uStereoOut = 0;
	bForceStereo = TRUE;
	bTapeMon = TRUE;

	memcpy(pan_set, pan_set_def, sizeof(pan_set));

	hWavCapture = -1;
	bWavCapture = FALSE;
	pWavCapture = NULL;
	nWavCapture = 0;
	dwWavCapture = 0;

	lpds = NULL;
	lpdsp = NULL;
	lpdsb = NULL;
	lpsbuf = NULL;
	bNowBank = FALSE;
	uBufSize = 0;
	uRate = 0;
	uSample = 0;
	uBeep = 0;
#if XM7_VER >= 2
	uKeyEncBeep = 0;
#endif
	bMode = FALSE;
	uStereo = 0;
	uChannels = 1;

	bBeepFlag = FALSE;
#if XM7_VER >= 2
	bKeyEncBeepFlag = FALSE;
#endif
	bTapeFlag = FALSE;
	bTapeFlag2 = FALSE;
	uTapeDelta = 0;

	for (i=0; i<4; i++) {
		pOPN[i] = NULL;
#ifdef FMDLL
		pOPNA[i] = NULL;
#endif
		pSoundChip[i] = NULL;
	}
	memset(nScale, 0, sizeof(nScale));
	memset(uCh3Mode, 0xff, sizeof(uCh3Mode));
	bInitFlag = FALSE;
	bRomeoMute = FALSE;
	bSCCIMute = FALSE;
	bUseSCCI = FALSE;
	dwSoundTotal = 0;
	memset(nChannelMask, 0x00, sizeof(nChannelMask));
#ifndef FMGEN_DIRECT_OUTPUT
	memset(OPNqueue, 0x00, sizeof(OPNqueue));
	uOPNqueuePtr = 0;
	memset(WHGqueue, 0x00, sizeof(WHGqueue));
	uWHGqueuePtr = 0;
	memset(THGqueue, 0x00, sizeof(THGqueue));
	uTHGqueuePtr = 0;
	memset(PSGqueue, 0x00, sizeof(PSGqueue));
	uPSGqueuePtr = 0;
#endif

#ifdef MIDI
	/* MIDI[N */
	nMidiDelay = 100;
	bMidiDelayMode = TRUE;
#endif

	/* g`NbsO֐ݒ */
#ifdef _M_AMD64
	CopySoundBuffer = CopySndBuf;
#else
	if (bMMXflag) {
		CopySoundBuffer = CopySndBufMMX;
	}
	else {
		CopySoundBuffer = CopySndBuf;
	}
#endif

	bGlobalFlag = TRUE;

#ifdef FMDLL
	pFmDll = NULL;
	pMix = NULL;
	bUseDll = TRUE;
	uMask = 0;
#endif
}

/*
 *	N[Abv
 */
void FASTCALL CleanSnd(void)
{
	int i;

	/* TEh~ */
	StopSnd();

	/* OPN */
	for (i=0; i<4; i++) {
		if (pOPN[i]) {
			delete pOPN[i];
			pOPN[i] = NULL;
		}
#ifdef FMDLL
		if (pOPNA[i]) {
			pFmDll->Release(pOPNA[i]);
			pOPNA[i] = NULL;
		}
#endif
	}

#ifdef FMDLL
	delete pFmDll;
	pFmDll = NULL;
#endif

	/* TEh쐬obt@ */
	if (lpsbuf) {
		free(lpsbuf);
		lpsbuf = NULL;
	}

#ifdef FMDLL
	if (pMix) {
		free(pMix);
		pMix = NULL;
	}
#endif

	/* DirectSoundBuffer */
	if (lpdsb) {
		lpdsb->Release();
		lpdsb = NULL;
	}
	if (lpdsp) {
		lpdsp->Release();
		lpdsp = NULL;
	}

	/* DirectSound */
	if (lpds) {
		lpds->Release();
		lpds = NULL;
	}

	/* uRateNA */
	uRate = 0;

	/* Lv`֘A */
	if (hWavCapture >= 0) {
		CloseCaptureSnd();
	}
	if (pWavCapture) {
		free(pWavCapture);
		pWavCapture = NULL;
	}
	hWavCapture = -1;
	bWavCapture = FALSE;

#if defined(LPRINT) && defined(JASTSOUND)
	if (pDACbuf) {
		free(pDACbuf);
		pDACbuf = NULL;
	}
#endif
}

/*
 *	SCCI
 */
void FASTCALL InitSCCI(void)
{
	char path[_MAX_PATH];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];

	GetModuleFileName(NULL, path, sizeof(path));
	_splitpath(path, drive, dir, NULL, NULL);
	strncpy(path, drive, sizeof(path));
	strncat(path, dir, sizeof(path) - strlen(path) - 1);
	strncat(path, "scci.dll", sizeof(path) - strlen(path) - 1);
	hSCCI = LoadLibrary(path);
	if (hSCCI) {
		SCCIFUNC getSoundInterfaceManager =
			(SCCIFUNC)(GetProcAddress(hSCCI,"getSoundInterfaceManager"));

		pManager = getSoundInterfaceManager();
		if(pManager == NULL){
			/* SCCIBDLL */
			bSCCI_available = FALSE;

			FreeLibrary(hSCCI);
			hSCCI = NULL;;
		}
		else {
			/* SCCIL */
			bSCCI_available = TRUE;

			/* CX^X */
			pManager->initializeInstance();
		}
	}
	else {
		pManager = NULL;
		bSCCI_available = FALSE;
	}
}

/*
 *	SCCIN[Abv
 */
void FASTCALL CleanSCCI(void)
{
	int i;

	if (hSCCI) {
		/* TEh`bv */
		for (i=0; i<4; i++) {
			if (pSoundChip[i]) {
				delete pSoundChip[i];
				pSoundChip[i] = NULL;
			}
		}

		/* TEhI/F}l[W */
		if (pManager){
			pManager->releaseInstance();
			pManager = NULL;
		}

		/* DLL */
		FreeLibrary(hSCCI);
		hSCCI = NULL;
	}
}

/*
 *	WAVEt@Cǂݍ (16rbgmf[^p)
 */
static BOOL FASTCALL LoadWav(LPSTR fname, struct _WAVDATA *wav)
{
	WAVEFORMATEX wfex;
	TCHAR path[_MAX_PATH];
	TCHAR drvname[_MAX_DRIVE];
	TCHAR dirname[_MAX_DIR];
	TCHAR filename[_MAX_FNAME];
	TCHAR extname[_MAX_EXT];
	BYTE buf[16];
	DWORD filSize;
	DWORD hdrSize;
	DWORD datSize;
	int fileh;

	ASSERT(fname);
	ASSERT(wav);

	/* t@CƊgqo */
	_tsplitpath(fname, NULL, NULL, filename, extname);

	/* XM7̃pX𓾂 */
	memset(path, 0, sizeof(path));
	GetModuleFileName(NULL, path, sizeof(path));
	_tsplitpath(path, drvname, dirname, NULL, NULL);

	/* pX */
	_tmakepath(path, drvname, dirname, filename, extname);

	/* t@CI[v */
	fileh = file_open(path, OPEN_R);
	if (fileh < 0) {
		return FALSE;
	}

	/* RIFFwb_`FbN */
	file_read(fileh, buf, 4);
	file_read(fileh, (BYTE *)&filSize, 4);
	buf[4] = '\0';
	if (strcmp((char *)buf, "RIFF")) {
		file_close(fileh);
		return FALSE;
	}
	filSize += 8;

	/* WAVEwb_`FbN */
	file_read(fileh, buf, 8);
	file_read(fileh, (BYTE *)&hdrSize, 4);
	buf[8] = '\0';
	if (strcmp((char *)buf, "WAVEfmt ")) {
		file_close(fileh);
		return FALSE;
	}
	hdrSize += (12 + 8);

	/* WAVEFORMATEX`FbN */
	file_read(fileh, (BYTE *)&wfex, sizeof(wfex));
	if ((wfex.wFormatTag != WAVE_FORMAT_PCM) ||
		(wfex.nChannels != 1) || (wfex.wBitsPerSample != 16)) {
		/* 16rbgmEjAPCMȊO͕s */
		file_close(fileh);
		return FALSE;
	}

	/* data`N */
	while (hdrSize < filSize) {
		/* `Nwb_ǂݍ */
		file_seek(fileh, hdrSize);
		file_read(fileh, buf, 4);
		file_read(fileh, (BYTE *)&datSize, 4);
		buf[4] = '\0';

		/* ̃`Nwb_ItZbgvZ */
		hdrSize += (datSize + 8);

		if (strcmp((char *)buf, "data") == 0) {
			/* data`Nǂݍ */
			wav->size = datSize / 2;
			wav->freq = wfex.nSamplesPerSec;
			wav->p = (short *)malloc(datSize);
			if (wav->p == NULL) {
				file_close(fileh);
				return FALSE;
			}
			if (!file_read(fileh, (BYTE *)wav->p, wav->size)) {
				file_close(fileh);
				free(wav->p);
				wav->p = NULL;
				return FALSE;
			}
			file_close(fileh);
			return TRUE;
		}
	}

	/* data`Nł */
	file_close(fileh);
	return FALSE;
}

/*
 *	FDDTEh 
 */
void FASTCALL InitFDDSnd(void)
{
	int i;

	/* [N */
	for (i=0; i<SNDBUF; i++) {
		memset(&WavP[i], 0, sizeof(_WAVPLAY));
	}

	/* WAVt@Cǂݍ */
	for (i=0; i<countof(Wav); i++) {
		if (!LoadWav(WavName[i], &Wav[i])) {
			Wav[i].size = 0;
			Wav[i].freq = 0;
		}
	}
}

/*
 *	FDDTEh N[Abv
 */
void FASTCALL CleanFDDSnd(void)
{
	int i;

	for (i=0; i<countof(Wav); i++) {
		if (Wav[i].p) {
			free(Wav[i].p);
			Wav[i].p = NULL;
		}
	}
}

/*
 *	WX^ݒ
 */
#ifdef OPNAMODE
static void FASTCALL SetReg(FM::OPNA *pOPN, BYTE *reg)
#else
static void FASTCALL SetReg(FM::OPN *pOPN, BYTE *reg)
#endif
{
	int i;

	/* PSG */
	for (i=0; i<16; i++) {
#ifdef OPNAMODE
		pOPN->SetReg((WORD)i, reg[i]);
#else
		pOPN->SetReg((BYTE)i, reg[i]);
#endif
	}

	/* FML[It */
#ifdef OPNAMODE
	for (i=0; i<6; i++) {
		pOPN->SetReg(0x28, (BYTE)(i + ((i >= 3) ? (1) : (0))));
	}
#else
	for (i=0; i<3; i++) {
		pOPN->SetReg(0x28, (BYTE)i);
	}
#endif

	/* FMWX^ */
#ifdef OPNAMODE
	for (i=0x30; i<0xb8; i++) {
		pOPN->SetReg((WORD)i, reg[i]);
	}
	for (i=0x130; i<0x1b8; i++) {
		pOPN->SetReg((WORD)i, reg[i]);
	}
#else
	for (i=0x30; i<0xb4; i++) {
		pOPN->SetReg((BYTE)i, reg[i]);
	}
#endif

	/* FM샂[h */
	pOPN->SetReg(0x27, reg[0x27] & 0xc0);
}

#ifdef FMDLL
static void FASTCALL SetReg2(LPVOID pOPNA, BYTE *reg)
{
	int i;

	/* PSG */
	for (i=0; i<16; i++) {
#ifdef OPNAMODE
		pFmDll->SetReg(pOPNA, (WORD)i, reg[i]);
#else
		pFmDll->SetReg(pOPNA, (BYTE)i, reg[i]);
#endif
	}

	/* FML[It */
#ifdef OPNAMODE
	for (i=0; i<6; i++) {
		pFmDll->SetReg(pOPNA, 0x28, (BYTE)(i + ((i >= 3) ? (1) : (0))));
	}
#else
	for (i=0; i<3; i++) {
		pFmDll->SetReg(pOPNA, 0x28, (BYTE)i);
	}
#endif

	/* FMWX^ */
#ifdef OPNAMODE
	for (i=0x30; i<0xb8; i++) {
		pFmDll->SetReg(pOPNA, (WORD)i, reg[i]);
	}
	for (i=0x130; i<0x1b8; i++) {
		pFmDll->SetReg(pOPNA, (WORD)i, reg[i]);
	}
#else
	for (i=0x30; i<0xb4; i++) {
		pFmDll->SetReg(pOPNA, (BYTE)i, reg[i]);
	}
#endif

	/* FM샂[h */
	pFmDll->SetReg(pOPNA, 0x27, reg[0x27] & 0xc0);
}
#endif

/*
 *	SCCIWX^ݒ
 */
static void FASTCALL SetRegSCCI(SoundChip *pSoundChip, BYTE *reg)
{
	int i;

	/* SCCIgpȂꍇ͉ɋA */
	if (!bSCCI_available || !bUseSCCI) {
		return;
	}

	/* PSG */
	for (i=0; i<16; i++) {
		pSoundChip->setRegister((BYTE)i, reg[i]);
	}

	/* FML[It */
	for (i=0; i<3; i++) {
		pSoundChip->setRegister(0x28, (BYTE)i);
	}

	/* FMWX^ */
	for (i=0x30; i<0xb4; i++) {
		pSoundChip->setRegister((BYTE)i, reg[i]);
	}

	/* FM샂[h */
	pSoundChip->setRegister(0x27, reg[0x27] & 0xc0);
}

/*
 *	SCCI[h
 */
void FASTCALL LoadSCCI(void)
{
	int i, j;
	static const struct _SCCICHIP {
		int		nScType;
		int		ncycle;
	} ScciChip[] = {
		{SC_TYPE_NONE,		0},
		{SC_TYPE_YM2203,	100},	/* OPN		*/
		{SC_TYPE_YM2608,	200},	/* OPNA		*/
		{SC_TYPE_YM2610B,	200},	/* OPNB-B	*/
		{SC_TYPE_AY8910,	100},
	};

	/* TEh`bv擾 */
	if (hSCCI && pManager) {
		for (i = OPN_STD; i <= OPN_PSG; i ++) {
			nSCCI_ChipInfo[i] = SCCICHIP_UNKNOWN;
			for (j = 1; j <= countof(ScciChip); j ++) {
				if ((i != OPN_PSG) && (ScciChip[j].nScType == SC_TYPE_AY8910)) {
					continue;
				}
				pSoundChip[i] = pManager->getSoundChip(
									ScciChip[j].nScType,
									OPN_CLOCK * ScciChip[j].ncycle);
				if (pSoundChip[i]) {
					nSCCI_ChipInfo[i] = j;
					break;
				}
			}
		}
	}
}

/*
 *	ZNg
 */
BOOL FASTCALL SelectSnd(HWND hWnd)
{
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfex;
	int i;
#ifdef FMDLL
	DWORD dwflg;
#endif
	int j;

	/* assert */
	ASSERT(hWnd);

	/* NtOĂ */
	bInitFlag = TRUE;

	/* p[^ݒ */
	uRate = nSampleRate;
	uTick = nSoundBuffer;
	bMode = bInterpolation;
	uStereo = uStereoOut;
	bGlobalFlag = bGlobalFocusFlag;
#ifdef FMDLL
	bUseDll = bUseFmDll;
#endif
	uTapeDelta = 0;
	if ((uStereo > 0) || bForceStereo) {
		uChannels = 2;
	}
	else {
		uChannels = 1;
	}

	/* rate==0ȂAȂ */
	if (uRate == 0) {
		return TRUE;
	}

	/* DiectSoundIuWFNg쐬 */
	if (FAILED(DirectSoundCreate(NULL, &lpds, NULL))) {
		/* ftHgfoCXȂAgp */
		return TRUE;
	}

	/* xݒ(D拦) */
	if (FAILED(lpds->SetCooperativeLevel(hWnd, DSSCL_PRIORITY))) {
		return FALSE;
	}

	/* vC}obt@쐬 */
	memset(&dsbd, 0, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	if (FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsp, NULL))) {
		return FALSE;
	}

	/* vC}obt@̃tH[}bgw */
	memset(&wfex, 0, sizeof(wfex));
	wfex.wFormatTag = WAVE_FORMAT_PCM;
	wfex.nChannels = uChannels;
	wfex.nSamplesPerSec = uRate;
	wfex.nBlockAlign = (WORD)(2 * uChannels);
	wfex.nAvgBytesPerSec = wfex.nSamplesPerSec * wfex.nBlockAlign;
	wfex.wBitsPerSample = 16;
	if (FAILED(lpdsp->SetFormat(&wfex))) {
		return FALSE;
	}

	/* ZJ_obt@쐬 */
	memset(&pcmwf, 0, sizeof(pcmwf));
	pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmwf.wf.nChannels = uChannels;
	pcmwf.wf.nSamplesPerSec = uRate;
	pcmwf.wf.nBlockAlign = (WORD)(2 * uChannels);
	pcmwf.wf.nAvgBytesPerSec = pcmwf.wf.nSamplesPerSec * pcmwf.wf.nBlockAlign;
	pcmwf.wBitsPerSample = 16;
	memset(&dsbd, 0, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	if (bGlobalFlag) {
		dsbd.dwFlags = DSBCAPS_GLOBALFOCUS | DSBCAPS_GETCURRENTPOSITION2;
	}
	else {
		dsbd.dwFlags = DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;
	}
	dsbd.dwBufferBytes = (pcmwf.wf.nAvgBytesPerSec * uTick) / 1000;
	dsbd.dwBufferBytes += (DWORD)7;
	dsbd.dwBufferBytes &= (DWORD)0xfffffff8;	/* 8oCgE */
	uBufSize = dsbd.dwBufferBytes;
	dsbd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;
	if (FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsb, NULL))) {
		return FALSE;
	}

	/* TEhobt@쐬(ZJ_obt@̔̎ԂŁADWORD) */
	lpsbuf = (DWORD *)malloc(uBufSize);
	if (lpsbuf == NULL) {
		return FALSE;
	}
	memset(lpsbuf, 0, uBufSize);

#ifdef FMDLL
	pMix = (int32*)malloc(uBufSize * 2 * uChannels);
	if (pMix == NULL) {
		return FALSE;
	}
#endif

	/* TvJE^ATEhԂNA */
	uSample = 0;
	dwSoundTotal = 0;

#ifdef FMDLL
	pFmDll = new CFMDLL(szFmDll);
	bFmDll = pFmDll->IsDll();
	dwflg = 0;
#endif

	/* OPNfoCX쐬 */
	for (i=0; i<4; i++) {
#ifdef OPNAMODE
		pOPN[i] = new FM::OPNA;
#else
		pOPN[i] = new FM::OPN;
#endif
#if XM7_VER == 1
		if ((i == 1) && fmx_flag) {
			pOPN[i]->Init(PSG_CLOCK_MSX, uRate, bMode, NULL);
		}
		else {
			pOPN[i]->Init(OPN_CLOCK * 100, uRate, bMode, NULL);
		}
#else
		pOPN[i]->Init(OPN_CLOCK * 100, uRate, bMode, NULL);
#endif
		pOPN[i]->Reset();
		pOPN[i]->SetReg(0x27, 0);

#ifdef FMDLL
		if (!bUseDll) {
			continue;
		}
		if ((i>0) && !(dwflg & SUPPORT_MULTIPLE)) {
			/* chips\ȂCreate͍ŏ1񂾂 */
			continue;
		}
		pFmDll->Create((LPVOID*)&pOPNA[i], OPN_CLOCK*100*2, uRate);
		if (pOPNA[i]) {
			dwflg = pFmDll->GetCaps(pOPNA[i]);
#if 0	/* chip\`FbN */
			if (!(dwflg & SUPPORT_MULTIPLE)) {
				pFmDll->Release(pOPNA[i]);
				pOPNA[i] = NULL;
				continue;
			}
#endif
			uMask = 0;
			if ((dwflg & SUPPORT_FM_A) == SUPPORT_FM_A) {
				uMask |= 0007;
			}
			if ((dwflg & SUPPORT_PSG) == SUPPORT_PSG) {
				uMask |= 0700;
			}
#ifdef OPNAMODE
			if ((dwflg & SUPPORT_FM_B) == SUPPORT_FM_B) {
				uMask |= 0070;
			}
			if ((dwflg & SUPPORT_ADPCM_B) == SUPPORT_ADPCM_B) {
				uMask |= 01000;
			}
			if ((dwflg & SUPPORT_RHYTHM) == SUPPORT_RHYTHM) {
				uMask |= 0xfc00;
			}
#endif
			pOPN[i]->SetChannelMask(uMask);	// fmgen̓}XN

			pFmDll->Reset(pOPNA[i]);
			pFmDll->SetReg(pOPNA[i], 0x27, 0);
		}
#endif
	}

	/* PSG̃vXP[𒲐 */
	pOPN[OPN_PSG]->SetReg(0x2d, 0);
	pOPN[OPN_PSG]->SetReg(0x2e, 0);
	if (pSoundChip[OPN_PSG] && (nSCCI_ChipInfo[OPN_PSG] != SCCICHIP_AY8910)) {
		pSoundChip[OPN_PSG]->setRegister(0x2d, 0);
		pSoundChip[OPN_PSG]->setRegister(0x2e, 0);
	}
#ifdef FMDLL
	pFmDll->SetReg(pOPNA[OPN_PSG], 0x2d, 0);
	pFmDll->SetReg(pOPNA[OPN_PSG], 0x2e, 0);
#endif

	/* PSG̃Gx[v[hPSG݊ɐݒ */
	pOPN[OPN_PSG]->psg.SetSSGEnvMode(FALSE);

	/* ăZNgɔAWX^ݒ */
	for (i=0; i<3; i++) {
		nScale[i] = 0;
	}
	opn_notify(0x27, 0);
	whg_notify(0x27, 0);
	thg_notify(0x27, 0);
	for (i=0; i<3; i++) {
		SetReg(pOPN[i], opn_reg[i]);
#ifdef FMDLL
		SetReg2(pOPNA[i], opn_reg[i]);
#endif
		if (pSoundChip[i]) {
			SetRegSCCI(pSoundChip[i], opn_reg[i]);

			/* OPNA/OPNBBgp͒特o悤ɐݒ */
			if ((nSCCI_ChipInfo[i] == SCCICHIP_YM2608) ||
				(nSCCI_ChipInfo[i] == SCCICHIP_YM2610B)) {
				for (j=0xb4; j<0xb7; j++) {
					pSoundChip[i]->setRegister(j, 0xC0);
				}
			}
		}
	}
	/* PSG */
	for (i=0; i<16; i++) {
#ifdef OPNAMODE
		pOPN[OPN_PSG]->SetReg((WORD)i, psg_reg[i]);
#else
		pOPN[OPN_PSG]->SetReg((BYTE)i, psg_reg[i]);
#endif
#ifdef FMDLL
#ifdef OPNAMODE
		pFmDll->SetReg(pOPNA[OPN_PSG], (WORD)i, psg_reg[i]);
#else
		pFmDll->SetReg(pOPNA[OPN_PSG], (BYTE)i, psg_reg[i]);
#endif
#endif
		if (pSoundChip[OPN_PSG]) {
#ifdef OPNAMODE
			pSoundChip[OPN_PSG]->setRegister((WORD)i, psg_reg[i]);
#else
			pSoundChip[OPN_PSG]->setRegister((BYTE)i, psg_reg[i]);
#endif
		}
	}

	/* Lv`֘A */
	if (!pWavCapture) {
		pWavCapture = (WORD *)malloc(sizeof(WORD) * 0x8000);
	}
	ASSERT(hWavCapture == -1);
	ASSERT(!bWavCapture);

#if defined(LPRINT) && defined(JASTSOUND)
	/* WXgTEh  */
	pDACbuf = (BYTE *)malloc((uRate * uTick) / 100);
	pDACbuf[0] = 0x80;
	dwDACptr = 0;
#endif

	/* TEhX^[g */
	bNowBank = FALSE;
	PlaySnd();

	/* {[ݒ */
	SetSoundVolume();

	return TRUE;
}

/*
 *	Kp
 */
void FASTCALL ApplySnd(void)
{
#ifdef FMDLL
	int i;
#endif

	/* ŃA^[ */
	if (!bInitFlag) {
		return;
	}

	/* ROMEOgp̓~[g */
	if (bRomeo) {
		ROMEO_Mute(!bUseRomeo);
	}

	/* SCCIgp̓~[g */
	if (bSCCI_available) {
		SCCI_Mute(!bUseSCCI);
	}

	/* p[^v`FbN */
	if ((uRate == nSampleRate) && (uTick == nSoundBuffer) &&
		(bGlobalFlag == bGlobalFocusFlag) &&
#ifdef FMDLL
		(bUseDll == bUseFmDll) &&
#endif
		(bMode == bInterpolation) && (uStereo == uStereoOut)) {
		return;
	}

	/* ɏłĂȂA */
	if (uRate != 0) {
		CleanSnd();
	}

	/* ăZNg */
	SelectSnd(hMainWnd);

	/* {[ݒ */
	SetSoundVolume();

#ifdef FMDLL
	/* ~[gĐݒ */
#ifdef OPNAMODE
	for (i=0; i<3 * 16 + 3; i++) {
		SetMute2(i / 16, i % 16, GetMute2(i / 16, i % 16));
	}
#else
	for (i=0; i<21; i++) {
		SetMute(i, GetMute(i));
	}
#endif
#endif
}

/*
 *	{[ݒ
 */
void FASTCALL SetSoundVolume(void)
{
	int i;

	/* FM/PSG{[ݒ */
	for (i=0; i<4; i++) {
		ASSERT(pOPN[i]);

		if (pOPN[i]) {
			pOPN[i]->SetVolumeFM(nFMVolume * 2);
			pOPN[i]->psg.SetVolume(nPSGVolume * 2);
#ifdef FMDLL
			if (pOPNA[i]) {
				pOPN[i]->SetChannelMask(nChannelMask[i] | uMask);
				pFmDll->SetVolumeFM(pOPNA[i], nFMVolume * 2);
				pFmDll->SetVolumePSG(pOPNA[i], nPSGVolume * 2);
			}
#endif
		}
	}

	/* BEEP/CMT/eʉ{[ݒ */
	nBeepLevel = (int)(32767.0 * pow(10.0, nBeepVolume / 20.0));
	nCMTLevel = (int)(32767.0 * pow(10.0, nCMTVolume / 20.0));
	nWaveLevel = (int)(32767.0 * pow(10.0, nWaveVolume / 20.0));

	/* `lZp[Vݒ */
	vol_set[0].l = vol_set[4].r = 32;
	vol_set[1].l = vol_set[3].r = 16 + uChSeparation;
	vol_set[3].l = vol_set[1].r = 16 - uChSeparation;
	vol_set[4].l = vol_set[0].r = 0;
}

/*
 *	{[ݒ2(ݒ_CAOp)
 */
void FASTCALL SetSoundVolume2(UINT uSp, int nFM, int nPSG,
							  int nBeep, int nCMT, int nWav)
{
	int i;

	/* FM/PSG{[ݒ */
	for (i=0; i<4; i++) {
		ASSERT(pOPN[i]);

		if (pOPN[i]) {
			pOPN[i]->SetVolumeFM(nFM * 2);
			pOPN[i]->psg.SetVolume(nPSG * 2);
#ifdef FMDLL
			if (pOPNA[i]) {
				pOPN[i]->SetChannelMask(nChannelMask[i] | uMask);
				pFmDll->SetVolumeFM(pOPNA[i], nFM * 2);
				pFmDll->SetVolumePSG(pOPNA[i], nPSG * 2);
			}
#endif
		}
	}

	/* BEEP/CMT/eʉ{[ݒ */
	nBeepLevel = (int)(32767.0 * pow(10.0, nBeep / 20.0));
	nCMTLevel = (int)(32767.0 * pow(10.0, nCMT / 20.0));
	nWaveLevel = (int)(32767.0 * pow(10.0, nWav / 20.0));

	/* `lZp[Vݒ */
	vol_set[0].l = vol_set[4].r = 32;
	vol_set[1].l = vol_set[3].r = 16 + uSp;
	vol_set[3].l = vol_set[1].r = 16 - uSp;
	vol_set[4].l = vol_set[0].r = 0;
}

/*
 *	tJn
 */
void FASTCALL PlaySnd()
{
	HRESULT hr;
	WORD *ptr1, *ptr2;
	DWORD size1, size2;

	if (lpdsb) {
		/* obt@ׂăNA */
		if (lpsbuf) {
			memset(lpsbuf, 0, uBufSize);
		}

		/* bN */
		hr = lpdsb->Lock(0, uBufSize, (void **)&ptr1, &size1,
						(void**)&ptr2, &size2, 0);
		/* obt@Ă΁AXgA */
		if (hr == DSERR_BUFFERLOST) {
			lpdsb->Restore();
		}
		/* bNꍇ̂݁AZbg */
		if (SUCCEEDED(hr)) {
			if (ptr1) {
				memset(ptr1, 0, size1);
			}
			if (ptr2) {
				memset(ptr2, 0, size2);
			}

			/* AbN */
			lpdsb->Unlock(ptr1, size1, ptr2, size2);
		}

		/* tJn */
		lpdsb->Play(0, 0, DSBPLAY_LOOPING);
	}

	/* TvJE^ATEhԂNA */
	uSample = 0;
	dwSoundTotal = 0;
}

/*
 *	t~
 */
void FASTCALL StopSnd()
{
	if (lpdsb) {
		lpdsb->Stop();
	}
}

/*
 *	BEEP
 */
static void FASTCALL BeepSnd(int32 *buf, int samples)
{
	int sf;
	int i;

	/* BEEPo̓`FbN */
	if (!bBeepFlag) {
		return;
	}

	/* Tv */
	for (i=0; i<samples; i++) {
		/* `g쐬 */
		sf = (int)(uBeep * nBeepFreq * 2);
		sf /= (int)uRate;

		/* EɉăTv */
		if (uChannels == 1) {
			if (sf & 1) {
				*buf++ += nBeepLevel;
			}
			else {
				*buf++ -= nBeepLevel;
			}
		}
		else {
			if (sf & 1) {
				*buf++ += nBeepLevel;
				*buf++ += nBeepLevel;
			}
			else {
				*buf++ -= nBeepLevel;
				*buf++ -= nBeepLevel;
			}
		}

		/* JE^Abv */
		uBeep++;
		if (uBeep >= uRate) {
			uBeep = 0;
		}
	}
}

#if XM7_VER >= 2
/*
 *	BEEP
 */
static void FASTCALL KeyEncBeepSnd(int32 *buf, int samples)
{
	int sf;
	int i;

	/* BEEPo̓`FbN */
	if (!bKeyEncBeepFlag) {
		return;
	}

	/* Tv */
	for (i=0; i<samples; i++) {
		/* `g쐬 */
		sf = (int)(uKeyEncBeep * 2350 * 2);
		sf /= (int)uRate;

		/* EɉăTv */
		if (uChannels == 1) {
			if (sf & 1) {
				*buf++ += nBeepLevel;
			}
			else {
				*buf++ -= nBeepLevel;
			}
		}
		else {
			if (sf & 1) {
				*buf++ += nBeepLevel;
				*buf++ += nBeepLevel;
			}
			else {
				*buf++ -= nBeepLevel;
				*buf++ -= nBeepLevel;
			}
		}

		/* JE^Abv */
		uKeyEncBeep++;
		if (uKeyEncBeep >= uRate) {
			uKeyEncBeep = 0;
		}
	}
}
#endif

/*
 *	e[v
 */
static void FASTCALL TapeSnd(int32 *buf, int samples)
{
	DWORD dat;
	int i;
	int tmp;

	/* e[vo̓`FbN */
	if (!tape_motor || !bTapeMon) {
		return;
	}

	/* g`߂ */
	if ((uRate % 11025) == 0) {
		tmp = (uRate * 5) / 44100;
	}
	else {
		tmp = (uRate * 5) / 48000;
#ifdef FMDLL
//		tmp = (uRate * 5) / 55466;
#endif
	}

	/* o͏ԂωꍇAg`ԂJn */
	if (bTapeFlag != bTapeFlag2) {
		if (!uTapeDelta) {
			uTapeDelta = 1;
		}
		else {
			uTapeDelta = (BYTE)(tmp - uTapeDelta + 1);
		}
	}

	/* Tv */
	for (i=0; i<samples; i++) {
		if (uTapeDelta) {
			/* g`Ԃ */
			dat = (0x1000 / tmp) * uTapeDelta;
			if (bTapeFlag) {
				dat = dat - nCMTLevel;
			}
			else {
				dat = nCMTLevel - dat;
			}

			uTapeDelta ++;
			if (uTapeDelta > tmp) {
				uTapeDelta = 0;
			}
		}
		else {
			/* g`ԂȂ */
			if (bTapeFlag) {
				dat = nCMTLevel;
			}
			else {
				dat = -nCMTLevel;
			}
		}

		*buf++ += dat;
		if (uChannels == 2) {
			*buf++ += dat;
		}
	}

	/* ݂̃e[vo͏Ԃۑ */
	bTapeFlag2 = bTapeFlag;
}

/*
 *	WAVf[^ (FDD/CMT)
 */
static void FASTCALL WaveSnd(int32 *buf, int samples)
{
	int i, j, dat;

	/* Tv */
	for (i=0; i<samples; i++) {
		for (j=0; j<SNDBUF; j++) {
			if (WavP[j].bPlay) {
				dat = Wav[WavP[j].dwWaveNo].p[WavP[j].dwCount1];
				dat = (short)(((int)dat * nWaveLevel) >> 16);
				*buf += (int)dat;
				if (uChannels == 2) {
					*(buf+1) += (int)dat;
				}

				/* JEgAbv */
				WavP[j].dwCount2 += (Wav[WavP[j].dwWaveNo].freq << 16) / uRate;
				if (WavP[j].dwCount2 > 0x10000) {
					WavP[j].dwCount1 += (WavP[j].dwCount2 >> 16);
					WavP[j].dwCount2 &= 0xFFFF;
					if (WavP[j].dwCount1 >= Wav[WavP[j].dwWaveNo].size) {
						WavP[j].bPlay = FALSE;
					}
				}
				WavP[j].dwCount3 ++;
			}
		}
		buf ++;
		if (uChannels == 2) {
			buf ++;
		}
	}
}

#ifdef FMDLL
static void Mix(int id, int32 *dst, int samples)
{
	if (!samples) return;
	if (samples * sizeof(int32) * 2 > uBufSize * 2 * uChannels) return;

	ZeroMemory(pMix, samples*sizeof(int32)*2);
	pFmDll->Mix(pOPNA[id], pMix, samples);
	int	i;
	int32	*src = pMix;
	for(i=0; i<samples; i++) {
		*dst += *src;
		dst++;
		src += 2;
	}

//	pOPN[id]->psg.Mix(dst, samples);
}

static void Mix2(int id, int32 *dst, int samples, int l, int r)
{
	if (!samples) return;
	if (samples * sizeof(int32) * 2 > uBufSize * 2 * uChannels) return;

	ZeroMemory(pMix, samples*sizeof(int32)*2);
	pFmDll->Mix(pOPNA[id], pMix, samples);
	int	i;
	int32	*src = pMix;
	for(i=0; i<samples; i++) {
		*dst += (*src * l) >> 4;
		dst++;
		src++;
		*dst += (*src * r) >> 4;
		dst++;
		src++;
	}

//	pOPN[id]->psg.Mix2( dst, samples, l, r);
}
#endif

/*
 *	WXgTEh DACf[^
 */
#if defined(LPRINT) && defined(JASTSOUND)
static void FASTCALL DACSnd(int32 *buf, int samples)
{
	int i;
	DWORD ptr;
	int ratio;
	short dat;
	short dat1, dat2;

	if (!samples || (lp_mode != LP_JASTSOUND)) {
		return;
	}

	/* Tv */
	ratio = (dwDACptr << 16) / samples;
	ptr = 0;
	for (i=0; i<samples; i++) {
		dat1 = (short)pDACbuf[(ptr >> 16)];
		dat2 = (short)pDACbuf[(ptr >> 16) + 1];
		dat = (short)(((dat1 + (((dat2 - dat1) * (ptr & 0xffff)) >> 16)) - 0x80) << 8);
		ptr += ratio;

		*buf++ += (int)((dat * nWaveLevel) >> 16);
		if (uChannels == 2) {
			*buf++ += (int)((dat * nWaveLevel) >> 16);
		}
	}

	pDACbuf[0] = pDACbuf[dwDACptr];
	dwDACptr = 0;
}
#endif

/*
 *	g`
 */
static void FASTCALL MixingSound(DWORD *q, int samples, BOOL bZero)
{
	memset(q, 0, sizeof(DWORD) * samples * uChannels);
	if (!bZero) {
		if (uChannels == 1) {
			/* m */
			if (opn_use) {
				if (pOPN[OPN_STD] && !ENABLE_ROMEO_OPN && !ENABLE_SCCI_OPN) {
					pOPN[OPN_STD]->Mix((int32*)q, samples);
#ifdef FMDLL
					Mix(OPN_STD, (int32*)q, samples);
#endif
				}
			}
#if XM7_VER == 1
			if (fmx_flag) {
				pOPN[OPN_WHG]->psg.Mix((int32*)q, samples);
#ifdef FMDLL
				Mix(OPN_WHG, (int32*)q, samples);
#endif
			}
			else if (whg_use) {
#else
			if (whg_use) {
#endif
				if (!ENABLE_SCCI_WHG) {
#ifdef OPNAMODE
					pOPN[OPN_WHG]->Mix((int32*)q, samples);
#else
					if (!ENABLE_ROMEO_WHG) {
						pOPN[OPN_WHG]->Mix((int32*)q, samples);
					}
					else {
						pOPN[OPN_WHG]->psg.Mix((int32*)q, samples);
					}
#endif
#ifdef FMDLL
					Mix(OPN_WHG, (int32*)q, samples);
#endif
				}
			}
			if (thg_use) {
				if (!ENABLE_SCCI_THG) {
					pOPN[OPN_THG]->Mix((int32*)q, samples);
#ifdef FMDLL
					Mix(OPN_THG, (int32*)q, samples);
#endif
				}
			}
#if XM7_VER >= 2
			if (model_type < MODEL_FM77AV) {
				if (!ENABLE_SCCI_PSG) {
					pOPN[OPN_PSG]->psg.Mix((int32*)q, samples);
#ifdef FMDLL
					Mix(OPN_PSG, (int32*)q, samples);
#endif
				}
			}
#else
			if (!ENABLE_SCCI_PSG) {
				pOPN[OPN_PSG]->psg.Mix((int32*)q, samples);
#ifdef FMDLL
				Mix(OPN_PSG, (int32*)q, samples);
#endif
			}
#endif
		}
		else {
			/* XeI */
			if (!whg_use && !thg_use &&
				((!opn_use) ||
#if XM7_VER >= 2
				 (model_type >= MODEL_FM77AV) ||
#endif
				 (pan_set[uStereo][OPN_PSG] == PAN_C))) {
				/* WHG/THGgpĂȂ(m) */
				if (opn_use) {
					if (!ENABLE_ROMEO_OPN && !ENABLE_SCCI_OPN) {
						pOPN[OPN_STD]->Mix2((int32*)q, samples, 16, 16);
#ifdef FMDLL
						Mix2(OPN_STD, (int32*)q, samples, 16, 16);
#endif
					}
				}
#if XM7_VER == 1
				if (fmx_flag) {
					pOPN[OPN_WHG]->psg.Mix2((int32*)q, samples, 16, 16);
#ifdef FMDLL
					Mix2(OPN_WHG, (int32*)q, samples, 16, 16);
#endif
				}
#endif
#if XM7_VER >= 2
				if (model_type < MODEL_FM77AV) {
					if (!ENABLE_SCCI_PSG) {
						pOPN[OPN_PSG]->psg.Mix2((int32*)q, samples, 16, 16);
#ifdef FMDLL
						Mix2(OPN_PSG, (int32*)q, samples, 16, 16);
#endif
					}
				}
#else
				if (!ENABLE_SCCI_PSG) {
					pOPN[OPN_PSG]->psg.Mix2((int32*)q, samples, 16, 16);
#ifdef FMDLL
					Mix2(OPN_PSG, (int32*)q, samples, 16, 16);
#endif
				}
#endif
			}
			else {
				/* WHG܂THGgp */
				if (opn_use) {
					if (!ENABLE_ROMEO_OPN && !ENABLE_SCCI_OPN) {
						pOPN[OPN_STD]->Mix2((int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_STD]].l,
							vol_set[pan_set[uStereo][OPN_STD]].r);
#ifdef FMDLL
						Mix2(OPN_STD, (int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_STD]].l,
							vol_set[pan_set[uStereo][OPN_STD]].r);
#endif
					}
				}
				if (whg_use) {
					if (!ENABLE_SCCI_WHG) {
#ifdef OPNAMODE
						pOPN[OPN_WHG]->Mix2((int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_WHG]].l,
							vol_set[pan_set[uStereo][OPN_WHG]].r);
#else
						if (!ENABLE_ROMEO_WHG) {
							pOPN[OPN_WHG]->Mix2((int32*)q, samples,
								vol_set[pan_set[uStereo][OPN_WHG]].l,
								vol_set[pan_set[uStereo][OPN_WHG]].r);
						}
						else {
							pOPN[OPN_WHG]->psg.Mix2((int32*)q, samples,
								vol_set[pan_set[uStereo][OPN_WHG]].l,
								vol_set[pan_set[uStereo][OPN_WHG]].r);
						}
#endif
#ifdef FMDLL
						Mix2(OPN_WHG, (int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_WHG]].l,
							vol_set[pan_set[uStereo][OPN_WHG]].r);
#endif
					}
				}
				if (thg_use) {
					if (!ENABLE_SCCI_THG) {
						pOPN[OPN_THG]->Mix2((int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_THG]].l,
							vol_set[pan_set[uStereo][OPN_THG]].r);
#ifdef FMDLL
						Mix2(OPN_THG, (int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_THG]].l,
							vol_set[pan_set[uStereo][OPN_THG]].r);
#endif
					}
				}
#if XM7_VER >= 2
				if (model_type < MODEL_FM77AV) {
					if (!ENABLE_SCCI_OPN) {
						pOPN[OPN_PSG]->psg.Mix2((int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_WHG]].l,
							vol_set[pan_set[uStereo][OPN_WHG]].r);
#ifdef FMDLL
						Mix2(OPN_PSG, (int32*)q, samples,
							vol_set[pan_set[uStereo][OPN_PSG]].l,
							vol_set[pan_set[uStereo][OPN_PSG]].r);
#endif
					}
				}
#else
				if (!ENABLE_SCCI_OPN) {
					pOPN[OPN_PSG]->psg.Mix2((int32*)q, samples,
						vol_set[pan_set[uStereo][OPN_WHG]].l,
						vol_set[pan_set[uStereo][OPN_WHG]].r);
#ifdef FMDLL
					Mix2(OPN_PSG, (int32*)q, samples,
						vol_set[pan_set[uStereo][OPN_PSG]].l,
						vol_set[pan_set[uStereo][OPN_PSG]].r);
#endif
				}
#endif
			}
		}

		/* e[v */
		TapeSnd((int32*)q, samples);

		/* r[v */
		BeepSnd((int32*)q, samples);
#if XM7_VER >= 2
		KeyEncBeepSnd((int32*)q, samples);
#endif

		/* WAVTEh */
		WaveSnd((int32*)q, samples);

#if defined(LPRINT) && defined(JASTSOUND)
		/* WXgTEh */
		if (lp_mode == LP_JASTSOUND) {
			DACSnd((int32*)q, samples);
		}
#endif
	}
}

#ifndef FMGEN_DIRECT_OUTPUT
/*
 *	fL[WX^
 */
static void FASTCALL FMReg_Dequeue_sub(int no, FMqueue *queue)
{
#ifdef OPNAMODE
	pOPN[no]->SetReg((uint16)queue->reg, (uint8)queue->dat);
#ifdef FMDLL
	if (queue->reg != 0xff) {
		pFmDll->SetReg(pOPNA[no], (uint16)queue->reg, (uint8)queue->dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[no]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[no], (uint16)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[no], (uint16)0x28, (uint8)0xf2);
	}
#endif
#else
	pOPN[no]->SetReg((uint8)queue->reg, (uint8)queue->dat);
#ifdef FMDLL
	if (queue->reg != 0xff) {
		pFmDll->SetReg(pOPNA[no], (uint8)queue->reg, (uint8)queue->dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[no]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[no], (uint8)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[no], (uint8)0x28, (uint8)0xf2);
	}
#endif
#endif
}

/*
 *	fL[
 */
static void FASTCALL FMReg_Dequeue(int queue)
{
	int ptr;

	if (queue & 1) {
		for (ptr = 0; ptr < uOPNqueuePtr; ptr++) {
			FMReg_Dequeue_sub(OPN_STD, &OPNqueue[ptr]);
		}
		uOPNqueuePtr = 0;
	}

	if (queue & 2) {
		for (ptr = 0; ptr < uWHGqueuePtr; ptr++) {
			FMReg_Dequeue_sub(OPN_WHG, &WHGqueue[ptr]);
		}
		uWHGqueuePtr = 0;
	}

	if (queue & 4) {
		for (ptr = 0; ptr < uTHGqueuePtr; ptr++) {
			FMReg_Dequeue_sub(OPN_THG, &THGqueue[ptr]);
		}
		uTHGqueuePtr = 0;
	}

	if (queue & 8) {
		for (ptr = 0; ptr < uPSGqueuePtr; ptr++) {
			FMReg_Dequeue_sub(OPN_PSG, &PSGqueue[ptr]);
		}
		uPSGqueuePtr = 0;
	}
}
#endif

/*
 *	TEh쐬obt@֒ǉ
 */
static void FASTCALL AddSnd(BOOL bFill, BOOL bZero)
{
	int i;
	int samples;

	/* OPNfoCX쐬ĂȂ΁AȂ */
	if (!pOPN[OPN_STD] || !pOPN[OPN_WHG] || !pOPN[OPN_THG] || !pOPN[OPN_PSG]) {
		return;
	}

	/* bFill̏ꍇ̃Tv */
	/* (m2byte/sampleEXeI4byte/sample) */
	samples = (uBufSize / uChannels) >> 2;
	samples -= uSample;

	/* !bFillȂAԂv */
	if (!bFill) {
		/* Ԍo߂狁߂_Tv */
		/* vZʂI[o[t[ɑ΍ 2002/11/25 */
		i = (uRate / 25);
		i *= dwSoundTotal;
		i /= 40000;

		/* uSampleƔrAvĂΉȂ */
		if (i <= (int)uSample) {
			return;
		}

		/* uSampleƂ̍񐶐Tv */
		i -= (int)(uSample);

		/* samples΍i */
		if (i <= samples) {
			samples = i;
		}
	}

	/* L[fmgenւ̃f[^荞 */
#ifndef FMGEN_DIRECT_OUTPUT
	FMReg_Dequeue(0x0f);
#endif

	/* ~LVO */
	MixingSound(&lpsbuf[uSample * uChannels], samples, bZero);

	/* XV */
	if (bFill) {
		dwSoundTotal = 0;
		uSample = 0;
	}
	else {
		uSample += samples;
	}
}

/*
 *	WAVLv`
 */
static void FASTCALL WavCapture(void)
{
	UINT nSize;
	DWORD *p;
	WORD *q;
	int j;

	/* WAVLv`łȂ΁A^[ */
	if (hWavCapture < 0) {
		return;
	}
	ASSERT(pWavCapture);

	/* |C^ATCY(nSizeWORDϊBYTEl) */
	p = lpsbuf;
	nSize = uBufSize / 2;

	/* bWavCaptureFALSEȂ */
	if (!bWavCapture) {
		/* o`FbN */
		while (nSize > 0) {
			if (uChannels == 1) {
				if (*p != 0) {
					break;
				}
				else {
					nSize -= 2;
					p++;
				}
			}
			else {
				if ((p[0] != 0) || (p[1] != 0)) {
					break;
				}
				else {
					nSize -= 4;
					p += 2;
				}
			}
		}
		/*  */
		if (nSize == 0) {
			return;
		}
	}

	/* nWavCapturel */
	if ((nWavCapture + nSize) >= 0x8000) {
		/* 32KBς܂ŃRs[ */
		j = (0x8000 - nWavCapture) >> 1;
		q = &pWavCapture[nWavCapture >> 1];
		CopySoundBuffer(p, q, j);
		p += j;

		/* cTCYXV */
		nSize -= (0x8000 - nWavCapture);

		/*  */
		file_write(hWavCapture, (BYTE*)pWavCapture, 0x8000);
		dwWavCapture += 0x8000;
		nWavCapture = 0;
	}

	/* ]Rs[ */
	j = nSize >> 1;
	q = &pWavCapture[nWavCapture >> 1];
	CopySoundBuffer(p, q, j);
	nWavCapture += nSize;

	/* Ș^ */
	bWavCapture = TRUE;
}

/*
 *	
 */
BOOL FASTCALL ProcessSnd(BOOL bZero)
{
	DWORD dwPlayC, dwWriteC;
	BOOL bWrite;
	DWORD dwOffset;
	WORD *ptr1, *ptr2;
	DWORD size1, size2;
	DWORD *p;
	int i, j;
	HRESULT hr;

	/* ݂R}hs by  */
	if (bUseRomeo) {
		juliet_YMF288EXEC((DWORD)uTick);
	}

	/* MIDIR}hs */
#ifdef MIDI
	if (bMidiDelayMode) {
		MidiSendData((DWORD)uTick);
	}
	else {
		MidiSendData((DWORD)nMidiDelay);
	}
#endif

	/* ĂȂ΁AȂ */
	if (!lpdsb) {
		return TRUE;
	}

	/* ݈ʒu𓾂 */
	if (FAILED(lpdsb->GetCurrentPosition(&dwPlayC, &dwWriteC))) {
		return FALSE;
	}

	/* ݈ʒuƃoNAKv𔻒f */
	bWrite = FALSE;
	if (bNowBank) {
		if (dwPlayC >= (uBufSize / 2)) {
			dwOffset = 0;
			bWrite = TRUE;
		}
	}
	else {
		if (dwPlayC < (uBufSize / 2)) {
			dwOffset = uBufSize / 2;
			bWrite = TRUE;
		}
	}

	/* ޕKvȂ΁A^[ */
	if (!bWrite) {
		/* e[v */
		if (tape_motor && bTapeMon) {
			bWrite = TRUE;
		}
		/* BEEP */
		if (beep_flag && speaker_flag) {
			bWrite = TRUE;
		}

		/* ǂ炩ONȂAobt@[U */
		if (bWrite) {
			AddSnd(FALSE, bZero);
		}

		return FALSE;
	}

	/* ݁B܂TEh쐬obt@S߂ */
	AddSnd(TRUE, bZero);

	/* ŃbN */
	hr = lpdsb->Lock(dwOffset, uBufSize / 2, (void **)&ptr1, &size1,
						(void**)&ptr2, &size2, 0);

	/* obt@Ă΁AXgA */
	if (hr == DSERR_BUFFERLOST) {
		lpdsb->Restore();
		hr = lpdsb->Lock(dwOffset, uBufSize / 2, (void **)&ptr1, &size1,
							(void**)&ptr2, &size2, 0);
	}
	/* bNȂ΁AĂӖȂ */
	if (FAILED(hr)) {
		return FALSE;
	}

	/* TEh쐬obt@ZJ_obt@ */
	p = lpsbuf;
	i = (int)(size1 / 2);
	j = (int)(size2 / 2);
	CopySoundBuffer(p, ptr1, i);
	CopySoundBuffer(p + i, ptr2, j);

	/* AbN */
	lpdsb->Unlock(ptr1, size1, ptr2, size2);

	/* oN] */
	bNowBank = (!bNowBank);

	/* WAVLv` */
	WavCapture();

	return TRUE;
}

/*
 *	TEho͊ҋ@
 */
void FASTCALL WaitSnd(void)
{
	DWORD dwPlayC, dwWriteC;
	DWORD dwLength;

	while (TRUE) {
		/* ĂȂ΁AȂ */
		if (!lpdsb) {
			return;
		}

		/* ݈ʒu𓾂 */
		if (FAILED(lpdsb->GetCurrentPosition(&dwPlayC, &dwWriteC))) {
			break;
		}

		/* ݈ʒuƃoNAKv𔻒f */
		if (bNowBank) {
			if (dwPlayC >= (uBufSize / 2)) {
				break;
			}
			dwLength = (uBufSize / 2);
		}
		else {
			if (dwPlayC < (uBufSize / 2)) {
				break;
			}
			dwLength = uBufSize;
		}

		if ((dwLength - dwPlayC) * 1000 > uRate) {
			Sleep(1);
		}
	}
}

/*
 *	x擾
 */
int FASTCALL GetLevelSnd(int ch)
{
#ifdef OPNAMODE
	FM::OPNA *p;
#else
	FM::OPN *p;
#endif
	int i;
	double s;
	double t;
	int *buf;
#ifdef FMDLL
	LPVOID *pa;
	BOOL b;
#endif

	ASSERT((ch >= 0) && (ch < 21));
	p = NULL;
#ifdef FMDLL
	pa = NULL;
	b = FALSE;
#endif

	/* OPN */
	if (ch < 6) {
		p = pOPN[OPN_STD];
#ifdef FMDLL
		pa = pOPNA[OPN_STD];
#endif
		/* ݂gp͎擾łȂ̂0 */
		if (bUseRomeo) {
			return 0;
		}

		/* SCCIgp0 */
		if (bUseSCCI && pSoundChip[OPN_STD]) {
			return 0;
		}

		/* ۂɎgĂȂ0 */
#if XM7_VER >= 2
		if ((!opn_enable || !opn_use) && (model_type < MODEL_FM77AV)) {
#else
		if (!opn_enable || !opn_use) {
#endif
			return 0;
		}
	}

	/* WHG */
	if ((ch >= 6) && (ch < 12)) {
		p = pOPN[OPN_WHG];
#ifdef FMDLL
		pa = pOPNA[OPN_WHG];
#endif
		ch -= 6;

		/* ۂɎgĂȂ0 */
		if (!whg_enable || !whg_use) {
			return 0;
		}

		/* ݂gpFM͎擾łȂ̂0 */
		if (bUseRomeo && (ch < 3)) {
			return 0;
		}

		/* SCCIgp0 */
		if (bUseSCCI && pSoundChip[OPN_WHG]) {
			return 0;
		}
	}

	/* THG */
	if ((ch >= 12) && (ch < 18)) {
		p = pOPN[OPN_THG];
#ifdef FMDLL
		pa = pOPNA[OPN_THG];
#endif
		ch -= 12;

		/* ۂɎgĂȂ0 */
		if (!thg_enable || !thg_use) {
			return 0;
		}

		/* SCCIgp0 */
		if (bUseSCCI && pSoundChip[OPN_THG]) {
			return 0;
		}
	}

	/* PSG */
	if ((ch >= 18) && (ch < 21)) {
		p = pOPN[OPN_PSG];
#ifdef FMDLL
		pa = pOPNA[OPN_PSG];
#endif
		ch -= 15;

		/* ۂɎgĂȂ0 */
#if XM7_VER >= 2
		if (model_type >= MODEL_FM77AV) {
#else
		if ((!psg_enable || !psg_use) && (model_type == MODEL_FM8)) {
#endif
			return 0;
		}

		/* SCCIgp0 */
		if (bUseSCCI && pSoundChip[OPN_PSG]) {
			return 0;
		}
	}

	/* ݃`FbN */
	if (!p) {
		return 0;
	}

	/* FM,PSG̋ */
	if (ch < 3) {
		/* FM:512Tv2avZ */
#ifdef FMDLL
		if (pa) {
			buf = pFmDll->GetBuffer(pa, ch);
			b = TRUE;
		}
		if (!pa || !buf) {
			buf = p->rbuf[ch];
			b = FALSE;
		}
#else
		buf = p->rbuf[ch];
#endif

		s = 0;
		for (i=0; i<512; i++) {
			t = (double)*buf++;
			t *= t;
			s += t;
		}
		s /= 512;

		/* [`FbN */
		if (s == 0) {
			return 0;
		}

		/* log10 */
		s = log10(s);

		/* FM␳ */
		s *= 40.0;
#ifdef FMDLL
		if (b) {
			s *= 318834.0 / (160746.0 + int(16384.0 * pow(10.0, nFMVolume / 40.0)));
		}
#endif
	}
	else {
		/* PSG:512Tv2avZ */
#ifdef FMDLL
		if (pa) {
			buf = pFmDll->GetBuffer(pa, ch + 3);
		}
		if (!pa || !buf) {
			buf = p->psg.rbuf[ch - 3];
		}
#else
		buf = p->psg.rbuf[ch - 3];
#endif

		s = 0;
		for (i=0; i<512; i++) {
			t = (double)*buf++;
			t *= t;
			s += t;
		}
		s /= 512;

		/* [`FbN */
		if (s == 0) {
			return 0;
		}

		/* log10 */
		s = log10(s);

		/* PSG␳ */
		s *= 60.0;
	}

	return (int)s;
}

#ifdef OPNAMODE
/*
 *	x擾
 */
int FASTCALL GetLevelSnd2(int no, int ch)
{
	FM::OPNA *p;
	int i;
	double s;
	double t;
	int *buf;
#ifdef FMDLL
	LPVOID *pa;
#endif

	ASSERT((no >= 0) && (no < 4));
	ASSERT((ch >= 0) && (ch < 16));

	/* OPN */
	if (no == 0) {
		p = pOPN[OPN_STD];
#ifdef FMDLL
		pa = pOPNA[OPN_STD];
#endif
		/* ݂gp͎擾łȂ̂0 */
		if (bUseRomeo) {
			return 0;
		}

		/* ۂɎgĂȂ0 */
#if XM7_VER >= 2
		if ((!opn_enable || !opn_use) && (model_type < MODEL_FM77AV)) {
#else
		if (!opn_enable || !opn_use) {
#endif
			return 0;
		}
	}

	/* WHG */
	if (no == 1) {
		p = pOPN[OPN_WHG];
#ifdef FMDLL
		pa = pOPNA[OPN_WHG];
#endif

		/* ۂɎgĂȂ0 */
		if (!whg_enable || !whg_use) {
			return 0;
		}

		/* ݂gpFM͎擾łȂ̂0 */
		if (bUseRomeo && (ch < 3)) {
			return 0;
		}
	}

	/* THG */
	if (no == 2) {
		p = pOPN[OPN_THG];
#ifdef FMDLL
		pa = pOPNA[OPN_THG];
#endif

		/* ۂɎgĂȂ0 */
		if (!thg_enable || !thg_use) {
			return 0;
		}
	}

	/* PSG */
	if (no == 3) {
		p = pOPN[OPN_PSG];
#ifdef FMDLL
		pa = pOPNA[OPN_PSG];
#endif
		ch += 6;

		/* ۂɎgĂȂ0 */
#if XM7_VER >= 2
		if (model_type >= MODEL_FM77AV) {
#else
		if ((!psg_enable || !psg_use) && (model_type == MODEL_FM8)) {
#endif
			return 0;
		}
	}

	/* ݃`FbN */
	if (!p) {
		return 0;
	}

#ifdef FMDLL
	if (pa) {
		buf = pFmDll->GetBuffer(pa, ch);
		if (!buf) {
			pa = NULL;
		}
	}
	if (!pa) {
		if (ch < 6) {
			/* FM */
			buf = p->rbuf[ch];
		}
		else if (ch < 9) {
			/* PSG */
			buf = p->psg.rbuf[ch - 6];
		}
		else if (ch < 15) {
			/* RHYDZM */
			buf = p->rhbuf[ch - 9];
		}
		else {
			/* ADPCM */
			buf = p->apbuf;
		}
	}
#else
	if (ch < 6) {
		/* FM */
		buf = p->rbuf[ch];
	}
	else if (ch < 9) {
		/* PSG */
		buf = p->psg.rbuf[ch - 6];
	}
	else if (ch < 15) {
		/* RHYDZM */
		buf = p->rhbuf[ch - 9];
	}
	else {
			/* ADPCM */
		buf = p->apbuf;
	}
#endif

	/* 512Tv2avZ */
	s = 0;
	if (buf) {
		for (i=0; i<512; i++) {
			t = (double)*buf++;
			t *= t;
			s += t;
		}
		s /= 512;
	}

	/* [`FbN */
	if (s == (double)0.0) {
		return 0;
	}

	/* log10 */
	s = log10(s);

	if (ch < 6) {
		/* FM␳ */
		s *= 40.0;
#ifdef FMDLL
		if (pa) {
			s *= 318834.0 / (160746.0 + int(16384.0 * pow(10.0, nFMVolume / 40.0)));
		}
#endif
	}
	else if (ch < 9) {
		/* PSG␳ */
		s *= 60.0;
	}
	else if (ch < 15) {
		/* RHYDZM␳ */
		s *= 58.0;
	}
	else {
		/* ADPCM␳ */
		if (s > (double)8) {
			return 0;
		}
		s *= 58.0;
	}

	return (int)s;
}
#endif

/*
 *	WAVLv`Jn
 */
void FASTCALL OpenCaptureSnd(char *fname)
{
	WAVEFORMATEX wfex;
	DWORD dwSize;
	int fileh;

	ASSERT(fname);
	ASSERT(hWavCapture < 0);
	ASSERT(!bWavCapture);

	/* łȂ΁A^[ */
	if (!pOPN[OPN_STD] || !pOPN[OPN_WHG] || !pOPN[OPN_THG]) {
		return;
	}

	/* obt@΁A^[ */
	if (!pWavCapture) {
		return;
	}

	/* uBufSize / 20x8000ȉłȂƃG[ */
	if ((uBufSize / 2) > 0x8000) {
		return;
	}

	/* t@CI[v(݃[h) */
	fileh = file_open(fname, OPEN_W);
	if (fileh < 0) {
		return;
	}

	/* RIFFwb_ */
	if (!file_write(fileh, (BYTE*)"RIFFŎEWAVEfmt ", 16)) {
		file_close(fileh);
		return;
	}

	/* WAVEFORMATEX */
	dwSize = sizeof(wfex);
	if (!file_write(fileh, (BYTE*)&dwSize, sizeof(dwSize))) {
		file_close(fileh);
		return;
	}
	memset(&wfex, 0, sizeof(wfex));
	wfex.cbSize = sizeof(wfex);
	wfex.wFormatTag = WAVE_FORMAT_PCM;
	wfex.nChannels = uChannels;
	wfex.nSamplesPerSec = uRate;
	wfex.nBlockAlign = (WORD)(2 * uChannels);
	wfex.nAvgBytesPerSec = wfex.nSamplesPerSec * wfex.nBlockAlign;
	wfex.wBitsPerSample = 16;
	if (!file_write(fileh, (BYTE *)&wfex, sizeof(wfex))) {
		file_close(fileh);
		return;
	}

	/* dataTuwb_ */
	if (!file_write(fileh, (BYTE *)"dataŎE", 8)) {
		file_close(fileh);
		return;
	}

	/* ok */
	nWavCapture = 0;
	dwWavCapture = 0;
	bWavCapture = FALSE;
	hWavCapture = fileh;
}

/*
 *	WAVLv`I
 */
void FASTCALL CloseCaptureSnd(void)
{
	DWORD dwLength;

	ASSERT(hWavCapture >= 0);

	/* obt@Ɏc */
	file_write(hWavCapture, (BYTE*)pWavCapture, nWavCapture);
	dwWavCapture += nWavCapture;
	nWavCapture = 0;

	/* t@COX */
	file_seek(hWavCapture, 4);
	dwLength = dwWavCapture + sizeof(WAVEFORMATEX) + 20;
	file_write(hWavCapture, (BYTE *)&dwLength, sizeof(dwLength));

	/* dataOX */
	file_seek(hWavCapture, sizeof(WAVEFORMATEX) + 24);
	file_write(hWavCapture, (BYTE *)&dwWavCapture, sizeof(dwWavCapture));

	/* t@CN[Y */
	file_close(hWavCapture);

	/* [NGANA */
	hWavCapture = -1;
	bWavCapture = FALSE;
}

/*
 *	ROMEO ~[g
 */
extern "C" {
void FASTCALL ROMEO_Mute(BOOL flag)
{
	int i;

	/* Ԃωꍇ̂ݏs */
	if (bRomeo && (bRomeoMute != flag)) {
		if (flag) {
			/* ~[g */
			juliet_YMF288Mute(FALSE);

			/* PSG */
			for (i=0; i<16; i++) {
				juliet_YMF288A_B((BYTE)i, opn_reg[OPN_STD][i]);
			}

			/* FML[It */
			for (i=0; i<3; i++) {
				juliet_YMF288A_B(0x28, (BYTE)i);
				juliet_YMF288A_B(0x28, (BYTE)(i | 0x04));
			}

			/* FMWX^ */
			for (i=0x30; i<0xb4; i++) {
				juliet_YMF288A_B((BYTE)i, opn_reg[OPN_STD][i]);
				juliet_YMF288B_B((BYTE)i, opn_reg[OPN_WHG][i]);
			}

			/* FM샂[h */
			juliet_YMF288A_B((BYTE)0x27, (BYTE)(opn_reg[OPN_STD][0x27] & 0xc0));
		}
		else {
			/* ~[gݒ */
			juliet_YMF288Mute(TRUE);
		}

		/* tOۑ */
		bRomeoMute = flag;
	}
}
}

/*
 *	ROMEO vXP[ύXpWX^Đݒ
 */
static void FASTCALL ROMEO_ChangePrescaler(void)
{
	int i;

	/* PSG */
	for (i=0; i<13; i++) {
		if ((i < 7) || (i > 10)) {
			juliet_YMF288A_B((BYTE)i, opn_reg[OPN_STD][i]);
		}
	}

	/* FM */
	for (i=0xa0; i<0xb0; i++) {
		juliet_YMF288A_B((BYTE)i, opn_reg[OPN_STD][i]);
		juliet_YMF288B_B((BYTE)i, opn_reg[OPN_WHG][i]);
	}
}

/*
 *	SCCI ~[g
 */
extern "C" {
void FASTCALL SCCI_Mute(BOOL flag)
{
	int i, j;

	LockVM();

	/* Ԃωꍇ̂ݏs */
	if (bSCCI_available && (bSCCIMute != flag)) {
		if (!flag) {
			/* FM */
			for (i=0; i<3; i++) {
				if (pSoundChip[i]) {
					SetRegSCCI(pSoundChip[i], opn_reg[i]);
					switch (opn_scale[i]) {
						case 2:
							pSoundChip[i]->setRegister(0x2f,0);
							break;
						case 3:
							pSoundChip[i]->setRegister(0x2e,0);
							break;
						case 6:
							pSoundChip[i]->setRegister(0x2d,0);
							break;
					}
				}
			}
			/* PSG */
			if (pSoundChip[OPN_PSG]) {
				for (i=0; i<16; i++) {
					pSoundChip[OPN_PSG]->setRegister((BYTE)i, psg_reg[i]);
				}
			}
		}
		else {
			/* ~[gݒ */
			for (i=0; i<3; i++) {
				if (pSoundChip[i]) {
					for (j=0x40; j<0x4f; j++) {
						if ((j & 0x03) != 0x03) {
							pSoundChip[i]->setRegister(j, 0x7f);
						}
					}
					for (j=0; j<3; j++) {
						pSoundChip[i]->setRegister(0x28, j);
					}
					for (j=8; j<=10; j++) {
						pSoundChip[i]->setRegister(j, 0);
					}
				}
			}
			for (j=8; j<=10; j++) {
				pSoundChip[OPN_PSG]->setRegister(j, 0);
			}
		}

		/* tOۑ */
		bSCCIMute = flag;
	}

	UnlockVM();
}
}

/*
 *	~[gԎ擾
 */
extern "C" {
BOOL FASTCALL GetMute(int ch)
{
	/* 4ԖڂPSĜ */
	if (ch >= 6 * 3) {
		ch += 3;
	}
	return (BOOL)(nChannelMask[ch / 6] & nChannelMaskBit[ch % 6]);
}
}

/*
 *	~[gԐݒ
 */
extern "C" {
void FASTCALL SetMute(int ch, BOOL mute)
{
	/* 4ԖڂPSĜ */
	if (ch >= 6 * 3) {
		ch += 3;
	}

	/* tO̐ݒ */
	if (mute) {
		nChannelMask[ch / 6] |= (WORD)nChannelMaskBit[ch % 6];
	}
	else {
		nChannelMask[ch / 6] &= (WORD)~nChannelMaskBit[ch % 6];
	}

	/* fmgenɑ΂ă~[gs */
	if (pOPN[ch / 6]) {
		pOPN[ch / 6]->SetChannelMask(nChannelMask[ch / 6]);
#ifdef FMDLL
		if (pOPNA[ch / 6]) {
			//	Ȃ炸FM̓}XN
			pOPN[ch / 6]->SetChannelMask(nChannelMask[ch / 6] | uMask);
			pFmDll->SetChannelMask(pOPNA[ch / 6], nChannelMask[ch / 6]);
		}
#endif
	}
}
}

#ifdef OPNAMODE
/*
 *	~[gԎ擾
 */
extern "C" {
BOOL FASTCALL GetMute2(int no, int ch)
{
	/* 4ԖڂPSĜ */
	if (no >= 3) {
		ch += 6;
	}
	return (BOOL)(nChannelMask[no] & (1 << ch));
}
}

/*
 *	~[gԐݒ
 */
extern "C" {
void FASTCALL SetMute2(int no, int ch, BOOL mute)
{
	/* 4ԖڂPSĜ */
	if (no >= 3) {
		ch += 6;
	}

	/* tO̐ݒ */
	if (mute) {
		nChannelMask[no] |= (WORD)(1 << ch);
	}
	else {
		nChannelMask[no] &= (WORD)~(1 << ch);
	}

	/* fmgenɑ΂ă~[gs */
	if (pOPN[no]) {
		pOPN[no]->SetChannelMask(nChannelMask[no]);
#ifdef FMDLL
		if (pOPNA[no]) {
			//	Ȃ炸FM̓}XN
			pOPN[no]->SetChannelMask(nChannelMask[no] | uMask);
			pFmDll->SetChannelMask(pOPNA[no], nChannelMask[no]);
		}
#endif
	}
}
}
#endif

/*
 *	OPNo
 */
extern "C" {
#ifdef OPNAMODE
void FASTCALL opn_notify(WORD reg, BYTE dat)
#else
void FASTCALL opn_notify(BYTE reg, BYTE dat)
#endif
{
	/* OPNȂ΁AȂ */
	if (!pOPN[OPN_STD]) {
		return;
	}

	/* vXP[𒲐 */
	if (opn_scale[OPN_STD] != nScale[OPN_STD]) {
		nScale[OPN_STD] = opn_scale[OPN_STD];
		if (bUseRomeo) {
			ROMEO_ChangePrescaler();
		}

		switch (opn_scale[OPN_STD]) {
			case 2:
				pOPN[OPN_STD]->SetReg(0x2f, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_STD], 0x2f, 0);
#endif
				if (pSoundChip[OPN_STD]) {
					pSoundChip[OPN_STD]->setRegister(0x2f,0);
				}
				break;
			case 3:
				pOPN[OPN_STD]->SetReg(0x2e, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_STD], 0x2d, 0);
				pFmDll->SetReg(pOPNA[OPN_STD], 0x2e, 0);
#endif
				if (pSoundChip[OPN_STD]) {
					pSoundChip[OPN_STD]->setRegister(0x2e,0);
				}
				break;
			case 6:
				pOPN[OPN_STD]->SetReg(0x2d, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_STD], 0x2d, 0);
#endif
				if (pSoundChip[OPN_STD]) {
					pSoundChip[OPN_STD]->setRegister(0x2d,0);
				}
				break;
		}
	}

	/* Ch3샂[h`FbN */
	if (reg == 0x27) {
		if (uCh3Mode[OPN_STD] == dat) {
			return;
		}
		uCh3Mode[OPN_STD] = dat;
	}

	/* 0xffWX^̓`FbN */
	if (reg == 0xff) {
		if ((opn_reg[OPN_STD][0x27] & 0xc0) != 0x80) {
			return;
		}
		else if (bUseSCCI && pSoundChip[OPN_STD]) {
			pSoundChip[OPN_STD]->setRegister(0x28, 0x02);
			pSoundChip[OPN_STD]->setRegister(0x28, 0xf2);
		}
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* óEL[o^ */
#ifdef FMGEN_DIRECT_OUTPUT
#ifdef OPNAMODE
	pOPN[OPN_STD]->SetReg((uint16)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_STD], (uint16)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_STD]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_STD], (uint16)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_STD], (uint16)0x28, (uint8)0xf2);
	}
#endif
#else
	pOPN[OPN_STD]->SetReg((uint8)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_STD], (uint8)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_STD]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_STD], (uint8)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_STD], (uint8)0x28, (uint8)0xf2);
	}
#endif
#endif
#else
	if (uOPNqueuePtr >= FMQUEUE_SIZE) {
		FMReg_Dequeue(0x01);
	}
	OPNqueue[uOPNqueuePtr].reg = reg;
	OPNqueue[uOPNqueuePtr].dat = dat;
	uOPNqueuePtr ++;
#endif

	/* ݂o */
	if (bUseRomeo) {
#ifdef OPNAMODE
		if (reg <= 0xff) {
			juliet_YMF288A_B((BYTE)reg, dat);
		}
		else {
			juliet_YMF288B_B((BYTE)(reg & 0xff), dat);
		}
#else
		juliet_YMF288A_B(reg, dat);
#endif
	}

	/* SCCIo */
	if (bUseSCCI && pSoundChip[OPN_STD] && !bSCCIMute) {
		if ((reg <= 0x0f) || ((reg >= 0x20) && (reg <= 0xb3))) {
			pSoundChip[OPN_STD]->setRegister((uint8)reg, (uint8)dat);
		}
	}
}
}

/*
 *	WHGo
 */
extern "C" {
#ifdef OPNAMODE
void FASTCALL whg_notify(WORD reg, BYTE dat)
#else
void FASTCALL whg_notify(BYTE reg, BYTE dat)
#endif
{
	/* WHGȂ΁AȂ */
	if (!pOPN[OPN_WHG]) {
		return;
	}

	/* vXP[𒲐 */
	if (opn_scale[OPN_WHG] != nScale[OPN_WHG]) {
		nScale[OPN_WHG] = opn_scale[OPN_WHG];
		if (bUseRomeo) {
			ROMEO_ChangePrescaler();
		}

		switch (opn_scale[OPN_WHG]) {
			case 2:
				pOPN[OPN_WHG]->SetReg(0x2f, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_WHG], 0x2f, 0);
#endif
				if (pSoundChip[OPN_WHG]) {
					pSoundChip[OPN_WHG]->setRegister(0x2f,0);
				}
				break;
			case 3:
				pOPN[OPN_WHG]->SetReg(0x2e, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_WHG], 0x2d, 0);
				pFmDll->SetReg(pOPNA[OPN_WHG], 0x2e, 0);
#endif
				if (pSoundChip[OPN_WHG]) {
					pSoundChip[OPN_WHG]->setRegister(0x2e,0);
				}
				break;
			case 6:
				pOPN[OPN_WHG]->SetReg(0x2d, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_WHG], 0x2d, 0);
#endif
				if (pSoundChip[OPN_WHG]) {
					pSoundChip[OPN_WHG]->setRegister(0x2d,0);
				}
				break;
		}
	}

	/* Ch3샂[h`FbN */
	if (reg == 0x27) {
		if (uCh3Mode[OPN_WHG] == dat) {
			return;
		}
		uCh3Mode[OPN_WHG] = dat;
	}

	/* 0xffWX^̓`FbN */
	if (reg == 0xff) {
		if ((opn_reg[OPN_WHG][0x27] & 0xc0) != 0x80) {
			return;
		}
		else if (bUseSCCI && pSoundChip[OPN_WHG] && !bSCCIMute) {
			pSoundChip[OPN_WHG]->setRegister(0x28, 0x02);
			pSoundChip[OPN_WHG]->setRegister(0x28, 0xf2);
		}
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* óEL[o^ */
#ifdef FMGEN_DIRECT_OUTPUT
#ifdef OPNAMODE
	pOPN[OPN_WHG]->SetReg((uint16)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint16)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_WHG]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint16)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint16)0x28, (uint8)0xf2);
	}
#endif
#else
	pOPN[OPN_WHG]->SetReg((uint8)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint8)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_WHG]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint8)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_WHG], (uint8)0x28, (uint8)0xf2);
	}
#endif
#endif
#else
	if (uWHGqueuePtr >= FMQUEUE_SIZE) {
		FMReg_Dequeue(0x02);
	}
	WHGqueue[uWHGqueuePtr].reg = reg;
	WHGqueue[uWHGqueuePtr].dat = dat;
	uWHGqueuePtr ++;
#endif

	/* ݂o */
#ifndef OPNAMODE
	if (bUseRomeo && ((reg >= 0x20) && (reg <= 0xfe))) {
		juliet_YMF288B_B(reg, dat);
	}
#endif

	/* SCCIo */
	if (bUseSCCI && pSoundChip[OPN_WHG] && !bSCCIMute) {
		if ((reg <= 0x0f) || ((reg >= 0x20) && (reg <= 0xb3))) {
			pSoundChip[OPN_WHG]->setRegister((uint8)reg, (uint8)dat);
		}
	}
}
}

/*
 *	THGo
 */
extern "C" {
#ifdef OPNAMODE
void FASTCALL thg_notify(WORD reg, BYTE dat)
#else
void FASTCALL thg_notify(BYTE reg, BYTE dat)
#endif
{
	/* THGȂ΁AȂ */
	if (!pOPN[OPN_THG]) {
		return;
	}

	/* vXP[𒲐 */
	if (opn_scale[OPN_THG] != nScale[OPN_THG]) {
		nScale[OPN_THG] = opn_scale[OPN_THG];

		switch (opn_scale[OPN_THG]) {
			case 2:
				pOPN[OPN_THG]->SetReg(0x2f, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_THG], 0x2f, 0);
#endif
				if (pSoundChip[OPN_THG]) {
					pSoundChip[OPN_THG]->setRegister(0x2f,0);
				}
				break;
			case 3:
				pOPN[OPN_THG]->SetReg(0x2e, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_THG], 0x2d, 0);
				pFmDll->SetReg(pOPNA[OPN_THG], 0x2e, 0);
#endif
				if (pSoundChip[OPN_THG]) {
					pSoundChip[OPN_THG]->setRegister(0x2e,0);
				}
				break;
			case 6:
				pOPN[OPN_THG]->SetReg(0x2d, 0);
#ifdef FMDLL
				pFmDll->SetReg(pOPNA[OPN_THG], 0x2d, 0);
#endif
				if (pSoundChip[OPN_THG]) {
					pSoundChip[OPN_THG]->setRegister(0x2d,0);
				}
				break;
		}
	}

	/* Ch3샂[h`FbN */
	if (reg == 0x27) {
		if (uCh3Mode[OPN_THG] == dat) {
			return;
		}
		uCh3Mode[OPN_THG] = dat;
	}

	/* 0xffWX^̓`FbN */
	if (reg == 0xff) {
		if ((opn_reg[OPN_THG][0x27] & 0xc0) != 0x80) {
			return;
		}
		else if (bUseSCCI && pSoundChip[OPN_THG] && !bSCCIMute) {
			pSoundChip[OPN_THG]->setRegister(0x28, 0x02);
			pSoundChip[OPN_THG]->setRegister(0x28, 0xf2);
		}
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* óEL[o^ */
#ifdef FMGEN_DIRECT_OUTPUT
#ifdef OPNAMODE
	pOPN[OPN_THG]->SetReg((uint16)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_THG], (uint16)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_THG]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_THG], (uint16)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_THG], (uint16)0x28, (uint8)0xf2);
	}
#endif
#else
	pOPN[OPN_THG]->SetReg((uint8)reg, (uint8)dat);
#ifdef FMDLL
	if (reg != 0xff) {
		pFmDll->SetReg(pOPNA[OPN_THG], (uint8)reg, (uint8)dat);
	}
	else if (!(pFmDll->GetCaps(pOPNA[OPN_THG]) & SUPPORT_CSM)) {
		pFmDll->SetReg(pOPNA[OPN_THG], (uint8)0x28, (uint8)0x02);
		pFmDll->SetReg(pOPNA[OPN_THG], (uint8)0x28, (uint8)0xf2);
	}
#endif
#endif
#else
	if (uTHGqueuePtr >= FMQUEUE_SIZE) {
		FMReg_Dequeue(0x04);
	}
	THGqueue[uTHGqueuePtr].reg = reg;
	THGqueue[uTHGqueuePtr].dat = dat;
	uTHGqueuePtr ++;
#endif

	/* SCCIo */
	if (bUseSCCI && pSoundChip[OPN_THG] && !bSCCIMute) {
		if ((reg <= 0x0f) || ((reg >= 0x20) && (reg <= 0xb3))) {
			pSoundChip[OPN_THG]->setRegister((uint8)reg, (uint8)dat);
		}
	}
}
}

/*
 *	PSGo
 */
extern "C" {
void FASTCALL psg_notify(BYTE reg, BYTE dat)
{
	/* PSGȂ΁AȂ */
	if (!pOPN[OPN_PSG]) {
		return;
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* óEL[o^ */
#ifdef FMGEN_DIRECT_OUTPUT
	pOPN[OPN_PSG]->SetReg((uint8)reg, (uint8)dat);
#ifdef FMDLL
	pFmDll->SetReg(pOPNA[OPN_PSG], (uint8)reg, (uint8)dat);
#endif
#else
	if (uPSGqueuePtr >= FMQUEUE_SIZE) {
		FMReg_Dequeue(0x08);
	}
	PSGqueue[uPSGqueuePtr].reg = reg;
	PSGqueue[uPSGqueuePtr].dat = dat;
	uPSGqueuePtr ++;
#endif

	/* SCCIo */
	if (bUseSCCI && pSoundChip[OPN_PSG] && !bSCCIMute) {
		if (reg <= 0x0f) {
			pSoundChip[OPN_PSG]->setRegister((uint8)reg, (uint8)dat);
		}
	}
}
}

/*
 *	BEEPo
 */
extern "C" {
void FASTCALL beep_notify(void)
{
	/* o͏ԂωĂȂ΃^[ */
	if (!((beep_flag & speaker_flag) ^ bBeepFlag)) {
		return;
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* tOێ */
	if (beep_flag && speaker_flag) {
		bBeepFlag = TRUE;
	}
	else {
		bBeepFlag = FALSE;
	}
}
}

#if XM7_VER >= 2
/*
 *	L[GR[_BEEPo
 */
extern "C" {
void FASTCALL keyencbeep_notify(void)
{
	/* o͏ԂωĂȂ΃^[ */
	if (!(keyenc_beep_flag ^ bKeyEncBeepFlag)) {
		return;
	}

	/* TEh */
	AddSnd(FALSE, FALSE);

	/* tOێ */
	if (keyenc_beep_flag) {
		bKeyEncBeepFlag = TRUE;
	}
	else {
		bKeyEncBeepFlag = FALSE;
	}
}
}
#endif

/*
 *	e[vo
 */
extern "C" {
void FASTCALL tape_notify(BOOL flag)
{
	/* o͏Ԃω`FbN */
	if (bTapeFlag == flag) {
		return;
	}

	if (bTapeMon) {
		/* TEh */
		AddSnd(FALSE, FALSE);
	}

	/* tOێ */
	bTapeFlag = flag;
}
}

/*
 *	WAVo
 */
extern "C" {
void FASTCALL wav_notify(BYTE no)
{
	int i, j;
	DWORD k;

	/* TEh */
	AddSnd(FALSE, FALSE);

	if (no == SOUND_STOP) {
		/* ~ */
		for (i=0; i<SNDBUF; i++) {
			WavP[i].bPlay = FALSE;
		}
	}
	else {
		if (Wav[no].freq) {
			j = 0;
			k = 0;
			for (i=0; i<SNDBUF; i++) {
				/* Đ~̃`l */
				if (!WavP[i].bPlay) {
					j = i;
					break;
				}
				else {
					/* ԍŏɍĐJnꂽ`l */
					if (k < WavP[i].dwCount3) {
						k = WavP[i].dwCount3;
						j = i;
					}
				}
			}

			/* f[^Zbg */
			WavP[j].dwWaveNo	= no;
			WavP[j].dwCount1	= 0;
			WavP[j].dwCount2	= 0;
			WavP[j].dwCount3	= 0;
			WavP[j].bPlay		= TRUE;
		}
	}
}
}

#ifdef FMDLL
/*
 *	FMDLLgptO擾
 */
extern "C" {
BOOL FASTCALL GetUseFmDllFlag(int d)
{
	return (BOOL)(pOPNA[d]);
}
}
#endif

/*
 *	WXgTEh DACf[^
 */
#if defined(LPRINT) && defined(JASTSOUND)
extern "C" {
void FASTCALL dac_notify(BYTE dat)
{
	if (dwDACptr < ((uRate * uTick) / 100)) {
		pDACbuf[++dwDACptr] = dat;
	}
}
}
#endif

#endif	/* _WIN32 */
