/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *	Copyright (C) 2021-2023 GIMONS
 *
 *	[ Win32API o[W_CAO ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <assert.h>
#include "xm7.h"
#include "keyboard.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_draw.h"
#include "w32_kbd.h"

#define	URL_NUM	0

/*
 *	O[o [N
 */
#if XM7_VER >= 3
BOOL bGravestone;						/* !? */
#endif

/*
 *	X^eBbN [N
 */
#if URL_NUM >= 1
static RECT AboutURLRect[URL_NUM];		/* URLeLXg` */
static BOOL bAboutURLHit[URL_NUM];		/* URLtH[JXtO */
static char pszAboutURL[URL_NUM][128];	/* URLeLXg */
#endif	/* URL_NUM >= 1 */
static RECT IconRect;					/* ACR` */
static HICON hVerIcon;					/* ACRnh */


/*
 *	_CAO
 */
static void FASTCALL AboutDlgInit(HWND hDlg)
{
	/* o[W */
	#if (XM7_VER == 2) && defined(FMTV151)
		#ifdef REVISION
			/* rW */
			#define	VER		VERSION""SBVRSN" "REVISION"-V2߂"
			#define	VER_IN	VERSION""SBVRSN" "REVISION"-V2"
		#else
			/* rWȂ */
			#define	VER		VERSION""SBVRSN"-V2߂"
			#define	VER_IN	VERSION""SBVRSN"-V2"
		#endif
	#else
		#ifdef REVISION
			/* rW */
			#define	VER		VERSION""SBVRSN" "REVISION
			#define	VER_IN	VER
		#else
			/* rWȂ */
			#define	VER		VERSION""SBVRSN
			#define	VER_IN	VER
		#endif
	#endif

	#ifdef BETAVER
		#ifdef BETANO
			/* 2ȍ~ */
			#define	VERSTR		VER""BETANO" ("DATE")"
			#define	VERSTR_IN	VER_IN"-beta"BETANO" ("DATE")"
		#else
			/* 1(ԍ\LȂ) */
			#define	VERSTR		VER" ("DATE")"
			#define	VERSTR_IN	VER_IN"-beta ("DATE")"
		#endif
	#else
		/*  */
		#define	VERSTR			VER" ("DATE")"
		#define	VERSTR_IN		VER_IN" ("DATE")"
	#endif

	#ifdef LEVEL
		/* x */
		#define	BASE_VER		"based on XM7 "LEVEL
	#endif

	HWND hWnd;
	RECT prect;
	RECT drect;
	POINT point;
	int i;

	ASSERT(hDlg);

	/* o[Wݒ */
	hWnd = GetDlgItem(hDlg, IDC_VERSION);
#if (XM7_VER == 2 && defined(FMTV151)) || defined(BETAVER)
	if (hAppInstance == hResInstance) {
		SetWindowText(hWnd, VERSTR);
	}
	else {
		SetWindowText(hWnd, VERSTR_IN);
	}
#else
	SetWindowText(hWnd, VERSTR);
#endif

	/* o[Wݒ */
	hWnd = GetDlgItem(hDlg, IDC_BASE_VERSION);
	SetWindowText(hWnd, BASE_VER);

#if URL_NUM >= 1
	/* 񃊃\[X[h */
	for (i=0; i<URL_NUM; i++) {
		pszAboutURL[i][0] = '\0';
		LoadString(hResInstance, IDS_ABOUTURL + i, pszAboutURL[i],
			sizeof(pszAboutURL[i]));
	}
#endif	/* URL_NUM >= 1 */

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

#if URL_NUM >= 1
	/* IDC_URL̃NCAgW擾 */
	for (i=0; i<URL_NUM; i++) {
		hWnd = GetDlgItem(hDlg, IDC_URL + i);
		ASSERT(hWnd);
		GetWindowRect(hWnd, &prect);

		point.x = prect.left;
		point.y = prect.top;
		ScreenToClient(hDlg, &point);
		AboutURLRect[i].left = point.x;
		AboutURLRect[i].top = point.y;

		point.x = prect.right;
		point.y = prect.bottom;
		ScreenToClient(hDlg, &point);
		AboutURLRect[i].right = point.x;
		AboutURLRect[i].bottom = point.y;

		/* IDC_URL폜 */
		DestroyWindow(hWnd);

		/* ̑ */
		bAboutURLHit[i] = FALSE;
	}
#endif	/* URL_NUM >= 1 */

	/* IDC_ABOUTICOÑNCAgW擾 */
	hWnd = GetDlgItem(hDlg, IDC_ABOUTICON);
	ASSERT(hWnd);
	GetWindowRect(hWnd, &prect);

	point.x = prect.left;
	point.y = prect.top;
	ScreenToClient(hDlg, &point);
	IconRect.left = point.x;
	IconRect.top = point.y;

	point.x = prect.right;
	point.y = prect.bottom;
	ScreenToClient(hDlg, &point);
	IconRect.right = point.x;
	IconRect.bottom = point.y;

	/* IDC_ABOUTICON폜 */
	DestroyWindow(hWnd);

	/* ACR[h */
	switch (model_type) {
		case MODEL_FM7:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM7ICON));
			break;
#if XM7_VER >= 2
		case MODEL_FM77AV:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_77AVICON));
			break;
		case MODEL_FM77AV20:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_AV20ICON));
			break;
		case MODEL_FM77AV20EX:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_20EXICON));
			break;
#if XM7_VER >= 3
		case MODEL_FM77AV40:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_AV40ICON));
			break;
		case MODEL_FM77AV40EX:
			if (bGravestone) {
				hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_40SXICON));
			}
			else {
				hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_40EXICON));
			}
			break;
#endif
#else
		case MODEL_FM77:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM77ICON2));
			break;
		case MODEL_FM8:
			hVerIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM8ICON));
			break;
#endif
		default:
			hVerIcon = NULL;
			ASSERT(FALSE);
			break;
	}

	ASSERT(hVerIcon);

	#undef VER
}

/*
 *	_CAO`
 */
static void AboutDlgPaint(HWND hDlg)
{
	HDC hDC;
	PAINTSTRUCT ps;
	HFONT hFont;
	HFONT hDefFont;
	TEXTMETRIC tm;
	LOGFONT lf;
#if URL_NUM >= 1
	int i;
#endif	/* URL_NUM >= 1 */

	ASSERT(hDlg);

	/* DC擾 */
	hDC = BeginPaint(hDlg, &ps);

	/* GUItHg̃gbN𓾂 */
	hFont = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
	hDefFont = (HFONT)SelectObject(hDC, hFont);
	GetTextMetrics(hDC, &tm);
	memset(&lf, 0, sizeof(lf));
	GetTextFace(hDC, LF_FACESIZE, lf.lfFaceName);
	SelectObject(hDC, hDefFont);

	/* A_[CttHg쐬 */
	lf.lfHeight = tm.tmHeight;
	lf.lfWidth = 0;
	lf.lfEscapement = 0;
	lf.lfOrientation = 0;
	lf.lfWeight = FW_DONTCARE;
	lf.lfItalic = tm.tmItalic;
	lf.lfUnderline = TRUE;
	lf.lfStrikeOut = tm.tmStruckOut;
	lf.lfCharSet = tm.tmCharSet;
	lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	lf.lfQuality = DEFAULT_QUALITY;
	lf.lfPitchAndFamily = tm.tmPitchAndFamily;
	hFont = CreateFontIndirect(&lf);
	hDefFont = (HFONT)SelectObject(hDC, hFont);

#if URL_NUM >= 1
	/* ` */
	for (i=0; i<URL_NUM; i++) {
		if (bAboutURLHit[i]) {
			SetTextColor(hDC, RGB(255, 0, 0));
		}
		else {
			SetTextColor(hDC, RGB(0, 0, 255));
		}
		SetBkColor(hDC, GetSysColor(COLOR_3DFACE));
		SetBkMode(hDC, OPAQUE);
		DrawText(hDC, pszAboutURL[i], strlen(pszAboutURL[i]),
			&AboutURLRect[i], DT_LEFT | DT_SINGLELINE);
	}
#endif	/* URL_NUM >= 1 */

	/* tHg߂ */
	SelectObject(hDC, hDefFont);
	DeleteObject(hFont);

	/* ACR` */
	DrawIcon(hDC, IconRect.left, IconRect.top, hVerIcon);

	/* DC */
	EndPaint(hDlg, &ps);
}

/*
 *	_CAOvV[W
 */
static BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
#if URL_NUM >= 1
	POINT point;
	BOOL bFlag;
	HCURSOR hCursor;
	int i;
#else
	UNUSED(lParam);
#endif	/* URL_NUM >= 1 */

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			AboutDlgInit(hDlg);
			return TRUE;

		/* ĕ` */
		case WM_PAINT:
			AboutDlgPaint(hDlg);
			return 0;

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* I */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					DestroyIcon(hVerIcon);
					EndDialog(hDlg, 0);
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;

#if URL_NUM >= 1
		/* ̈`FbN */
		case WM_NCHITTEST:
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			ScreenToClient(hDlg, &point);
			for (i=0; i<URL_NUM; i++) {
				bFlag = PtInRect(&AboutURLRect[i], point);
				/* tOقȂAXVčĕ` */
				if (bFlag != bAboutURLHit[i]) {
					bAboutURLHit[i] = bFlag;
					InvalidateRect(hDlg, &AboutURLRect[i], FALSE);
				}
			}
			break;

		/* J[\ݒ */
		case WM_SETCURSOR:
			for (i=0; i<URL_NUM; i++) {
				if (bAboutURLHit[i]) {
					/* OS̃o[WɂĂIDC_HAND͎s */
					hCursor = LoadCursor(NULL, MAKEINTRESOURCE(32649));
					if (!hCursor) {
						hCursor = LoadCursor(NULL, IDC_IBEAM);
					}
					SetCursor(hCursor);
					return TRUE;
				}
			}
			break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			/* tXN[̓uEUNȂ */
			if (bFullScreen) {
				break;
			}

			for (i=0; i<URL_NUM; i++) {
				if (bAboutURLHit[i]) {
					ShellExecute(hDlg, NULL, pszAboutURL[i], NULL, NULL,
						SW_SHOWNORMAL);
					return TRUE;
				}
			}
			break;
#endif	/* URL_NUM >= 1 */
	}

	/* ȊOFALSE */
	return FALSE;
}

/*
 *	o[W
 */
void FASTCALL OnAbout(HWND hWnd)
{
	ASSERT(hWnd);

	/* [__CAOs */
	keyboard_inputmask(TRUE);
	DialogBox(hResInstance, MAKEINTRESOURCE(IDD_ABOUTDLG), hWnd, AboutDlgProc);
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();
}

#endif	/* _WIN32 */
