/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *	Copyright (C) 2021-2023 GIMONS
 *
 *	[ tbs[fBXN Rg[(MB8877A) ]
 */

#include <string.h>
#include <stdlib.h>
#include "xm7.h"
#include "device.h"
#include "fdc.h"
#include "mainetc.h"
#include "dmac.h"
#include "event.h"

#if XM7_VER == 1 && defined(SFDC)
/*
 *	O[o [N
 */
BYTE fdc_command[2];				/* FDCR}h */
BYTE fdc_status[2];					/* FDCXe[^X */
BYTE fdc_trkreg[2];					/* gbNWX^ */
BYTE fdc_secreg[2];					/* ZN^WX^ */
BYTE fdc_datareg[2];				/* f[^WX^ */
BYTE fdc_dden[2];					/* P/{x(FM/MFM) */
BYTE fdc_sidereg[2];				/* TChWX^ */
BYTE fdc_drvreg[2];					/* _hCu */
BYTE fdc_dsenable[2];				/* DSCl[u */
BYTE fdc_motor[2];					/* [^ */
BYTE fdc_drqirq[2];					/* DRQIRQ */

BYTE fdc_cmdtype[2];				/* R}h^Cv */
WORD fdc_totalcnt[2];				/* g[^JE^ */
BYTE fdc_onetime[2];				/* 1oCg] */
WORD fdc_nowcnt[2];					/* JgJE^ */
BYTE fdc_ready[2][FDC_DRIVES];		/* fB */
BOOL fdc_teject[2][FDC_DRIVES];		/* ꎞCWFNg */
BOOL fdc_writep[2][FDC_DRIVES];		/* CgveNg */
BYTE fdc_track[2][FDC_DRIVES];		/* gbN */

char fdc_fname[2][FDC_DRIVES][256+1];/* t@C */
char fdc_name[2][FDC_DRIVES][FDC_MEDIAS][17];
BOOL fdc_fwritep[2][FDC_DRIVES];	/* CgveNg(t@Cx) */
BYTE fdc_header[2][FDC_DRIVES][0x2b0];/* D77t@Cwb_ */
BYTE fdc_medias[2][FDC_DRIVES];		/* fBA */
BYTE fdc_media[2][FDC_DRIVES];		/* fBAZNg */
BYTE fdc_access[2][FDC_DRIVES];		/* ANZXLED */
BYTE fdc_doorlock[FDC_DRIVES];		/* hAbN */
BYTE fdc_sided[FDC_DRIVES];			/* ЖʁE */
BOOL fdc_enable[2];					/* FDCLEtO */
BOOL fdc_use[2];					/* FDCgptO */

BOOL fdc_waitmode;					/* FDCANZXEFCg */
BOOL fdc_sound;						/* FDDV[NtO */

BYTE fdc_mode;						/* FDC[htO */
DWORD fdc_irq_time;					/* FDCG[荞݂܂ł̎ */
DWORD fdc_eod_time;					/* FDCf[^]荞݂܂ł̎ */

#if 1	/* FM-8 8" IFDCn */
BYTE ifdc_ssr;						/* IFDC VXeXe[^XWX^(SSR) */
BYTE ifdc_fsr;						/* IFDC FDCXe[^XWX^(FSR) */
BYTE ifdc_rrr;						/* IFDC WX^U[uWX^(RRR) */
BYTE ifdc_rdr;						/* IFDC WX^f[^WX^(RDR) */

BYTE ifdc_cmdreg;					/* IFDC R}hWX^(CMDREG) */
BYTE ifdc_trkreg;					/* IFDC gbNWX^(TRKREG) */
BYTE ifdc_ssreg;					/* IFDC ZN^X^[gWX^(SSREG) */
BYTE ifdc_sereg;					/* IFDC ZN^GhWX^(SEREG) */
BYTE ifdc_sidreg;					/* IFDC TChWX^(SIDREG) */
BYTE ifdc_dnsreg;					/* IFDC xWX^(DNSREG) */
BYTE ifdc_drvreg;					/* IFDC hCuWX^(DRVREG) */
BYTE ifdc_dlcreg;					/* IFDC hAbNWX^(DLCREG) */
BYTE ifdc_sformt;					/* IFDC ZN^tH[}bg(SFORMT) */
BYTE ifdc_secdat;					/* IFDC ZN^OX(SECDAT) */
BYTE ifdc_modreg;					/* IFDC [hZbgWX^(MODREG) */
#endif


/*
 *	X^eBbN [N
 */
static BYTE fdc_buffer0[FDC_BUFFER_SIZE_M];		/* f[^obt@ */
static BYTE fdc_buffer1[FDC_BUFFER_SIZE_S];		/* f[^obt@ */
static BYTE * const fdc_buffer[2] = {&fdc_buffer0[0], &fdc_buffer1[0]};
static BYTE *fdc_dataptr[2];					/* f[^|C^ */
static DWORD fdc_seekofs[2][FDC_DRIVES];		/* V[NItZbg */
static DWORD fdc_secofs[2][FDC_DRIVES];			/* ZN^ItZbg */
static DWORD fdc_foffset[2][FDC_DRIVES][FDC_MEDIAS];
static DWORD fdc_trklen[2][FDC_DRIVES];			/* gbNf[^ */
static BOOL fdc_seekvct[2];						/* V[N(Trk0:TRUE) */
static BYTE fdc_indexcnt[2];					/* INDEXz[ JE^ */
static BOOL fdc_wait;							/* EFCg[hstO */
static int fdc_seek_track[2];					/* waitmodepV[NJE^ */
static int fdi_init[2];							/* C[W̏tO*/
static int fdi_secs[2];							/* C[W̃ZN^ */
static int fdi_idpos[2][96];					/* C[W̃ZN^ʒu */
static int fdi_dtpos[2][96];					/* C[W̃ZN^ʒu */


/*
 *	Xebv[ge[u
 */
static const int fdc_steprate[2][4] = {	{ 6000, 12000, 20000, 30000 },
										{ 3000, 6000, 10000, 15000 }};


/*
 *	f[^obt@TCYe[u
 */
static const DWORD fdc_buffer_size[2] = {	FDC_BUFFER_SIZE_M,
											FDC_BUFFER_SIZE_S	};


/*
 *	hCu]ԃe[u
 */
static const int fdd_rotation_time[2] = {	FDD_ROTATION_TIME,
											SFDD_ROTATION_TIME	};


/*
 *	[^[Cxge[u
 */
static const int fdd_event_motor[2] = {	EVENT_FDD_MOTOR,
										EVENT_SFDD_MOTOR	};


/*
 *	vg^Cv錾
 */
static void FASTCALL fdc_readbuf(int type, int drive);	/* PgbNǂݍ */
static BOOL FASTCALL mfdc_drq_event(void);		/* FDC DataRequestCxg (WAIT) */
static BOOL FASTCALL sfdc_drq_event(void);		/* FDC DataRequestCxg (WAIT) */
static BOOL FASTCALL mfdd_seek_event(void);		/* FDC V[NCxg (WAIT) */
static BOOL FASTCALL sfdd_seek_event(void);		/* FDC V[NCxg (WAIT) */
static BOOL FASTCALL mfdc_multi_event(void);	/* }`ZN^Cxg */
static BOOL FASTCALL sfdc_multi_event(void);	/* }`ZN^Cxg */
static BOOL FASTCALL mfdc_lost_event(void);		/* LOST DATACxg */
static BOOL FASTCALL sfdc_lost_event(void);		/* LOST DATACxg */
static void FASTCALL sfdc_drq_sub(void);		/* f[^NGXgTu[` */
static BOOL FASTCALL sfdc_nmi_event(void);		/* f[^]ICxg */
static BOOL FASTCALL fdd_motor_event(void);		/* [^[Cxg */

/*
 *	ID
 */
#define	ID	MAKEID('F', 'D', 'C', ' ')
//#define FDC_LOG

/*
 *	FDC
 *	
 */
BOOL FASTCALL fdc_init(void)
{
	/* tbs[t@C֌WZbg */
	memset(fdc_ready, FDC_TYPE_NOTREADY, sizeof(fdc_ready));
	memset(fdc_teject, FALSE, sizeof(fdc_teject));
	memset(fdc_fwritep, FALSE, sizeof(fdc_fwritep));
	memset(fdc_medias, 0, sizeof(fdc_medias));

	/* t@CItZbgSăNA */
	memset(fdc_foffset, 0, sizeof(fdc_foffset));

	/* ftHgFDCL */
	fdc_enable[FDC_M] = TRUE;
	fdc_enable[FDC_S] = TRUE;
	fdc_use[FDC_M] = FALSE;
	fdc_use[FDC_S] = FALSE;

	/* EFCg}[htO */
	fdc_waitmode = TRUE;
	fdc_sound = FALSE;
	fdc_wait = FALSE;

	/* [^[Cxgo^ */
	schedule_handle(EVENT_FDD_MOTOR, fdd_motor_event);
	schedule_handle(EVENT_SFDD_MOTOR, fdd_motor_event);

	/* C[W񖢏 */
	fdi_init[FDC_M] = FALSE;
	fdi_init[FDC_S] = FALSE;

	return TRUE;
}

/*
 *	FDC
 *	N[Abv
 */
void FASTCALL fdc_cleanup(void)
{
	/* tbs[t@C֌WZbg */
	memset(fdc_ready, FDC_TYPE_NOTREADY, sizeof(fdc_ready));
}

/*
 *	FDC
 *	Zbg
 */
void FASTCALL fdc_reset(void)
{
	LOG0(Normal, "Zbg");

	/* FDCWX^Zbg */
	memset(fdc_command, 0x03, sizeof(fdc_command));
	memset(fdc_status, 0, sizeof(fdc_status));
	memset(fdc_trkreg, 0, sizeof(fdc_trkreg));
	memset(fdc_secreg, 0x01, sizeof(fdc_secreg));
	memset(fdc_datareg, 0, sizeof(fdc_datareg));
	memset(fdc_sidereg, 0, sizeof(fdc_sidereg));
	memset(fdc_dsenable, 0, sizeof(fdc_dsenable));
	memset(fdc_drvreg, 0, sizeof(fdc_drvreg));
	memset(fdc_motor, 0, sizeof(fdc_motor));
	memset(fdc_doorlock, 0, sizeof(fdc_doorlock));
	memset(fdc_sided, 0x80, sizeof(fdc_sided));

	memset(fdc_cmdtype, 0, sizeof(fdc_cmdtype));
	memset(fdc_seekvct, 0, sizeof(fdc_seekvct));
	memset(fdc_track, 0, sizeof(fdc_track));
	memset(fdc_dataptr, 0, sizeof(fdc_dataptr));
	memset(fdc_access, 0, sizeof(fdc_access));

	/* FDCANZX֌WZbg */
	fdc_use[FDC_M] = FALSE;
	fdc_use[FDC_S] = FALSE;
	fdc_dden[FDC_M] = 0x40;				/* { */
	fdc_dden[FDC_S] = 0x00;				/* P */
	fdc_mode = 0;

#if 1	/* FM-8 8" IFDCn */
	ifdc_ssr = 0;
	ifdc_fsr = 0;
	ifdc_rrr = 0xff;
	ifdc_rdr = 0;

	ifdc_cmdreg = 0;
	ifdc_trkreg = 0;
	ifdc_ssreg = 0;
	ifdc_sereg = 0;
	ifdc_sidreg = 0;
	ifdc_dnsreg = 0;
	ifdc_drvreg = 0;
	ifdc_dlcreg = 0;
	ifdc_sformt = 0;
	ifdc_secdat = 0;
	ifdc_modreg = 0;
#endif

	/* f[^obt@֓ǂݍ */
	fdc_readbuf(FDC_M, fdc_drvreg[FDC_M]);
	fdc_readbuf(FDC_S, fdc_drvreg[FDC_S]);
}

/*-[ CRCvZ ]--------------------------------------------------------------*/

/*
 *	CRCvZe[u
 */
static WORD crc_table[256] = {
	0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
	0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
	0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
	0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
	0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
	0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
	0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
	0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
	0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
	0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
	0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
	0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
	0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
	0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
	0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
	0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
	0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
	0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
	0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
	0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
	0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
	0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
	0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
	0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
	0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
	0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
	0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
	0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
	0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
	0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
	0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
	0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

/*
 *	16bit CRCvZAZbg
 */
static void FASTCALL calc_set_crc(BYTE *addr, int size)
{
	WORD crc;

	/*  */
	crc = 0xffff;	/* CRC̏l0xffff */

	/* vZ */
	while (size > 0) {
		crc = (WORD)((crc << 8) ^ crc_table[(BYTE)(crc >> 8) ^ (BYTE)*addr++]);
		size--;
	}

	/* QoCgɃZbg(rbOGfBA) */
	*addr++ = (BYTE)(crc >> 8);
	*addr = (BYTE)(crc & 0xff);
}

/*
 *	vZ
 *	http://ja.wikipedia.org/wiki/Xorshift
 */
static DWORD FASTCALL rand_xor(void)
{
	static DWORD y = 2463534242;

	y = y ^ (y << 13);
	y = y ^ (y >> 17);
	y = y ^ (y << 15);

	return y;
}

/*-[ t@CǗ ]---------------------------------------------------------*/

/*
 *	Read Trackf[^쐬
 */
static void FASTCALL fdc_make_track(int type)
{
	int i;
	int j;
	int gap0;
	int gap1;
	int gap2;
	int sync;
	int gap;
	int gap3;
	int track;
	WORD count;
	WORD secs;
	WORD size;
	BOOL flag;
	BYTE *p;
	BYTE *q;
	WORD diff;
	BOOL ddm;
	int max_track, track_totalcnt;
	DWORD track_size;

	if (type) {
		max_track = 77;
		if (fdc_dden[type]) {
			track_totalcnt = 0x2A00;
		}
		else {
			track_totalcnt = 0x1500;
		}
	}
	else {
		if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
			max_track = 80;
		}
		else {
			max_track = 40;
		}
		track_totalcnt = 0x1800;
	}
	track_size = fdc_buffer_size[type];

	/* AtH[}bg`FbN */
	flag = FALSE;
	if (fdc_ready[type][fdc_drvreg[type]] != FDC_TYPE_D77) {
		/* 2D/2DD/2HD/VFDt@C̃AtH[}bg`FbN͋ */
		if (fdc_track[type][fdc_drvreg[type]] >= max_track) {
			/* AtH[}bg */
			flag = TRUE;
		}
	}
	else {
		if ((fdc_buffer[type][4] == 0) && (fdc_buffer[type][5] == 0)) {
			/* AtH[}bg */
			flag = TRUE;
		}
	}

	/* AtH[}bgȂA_f[^쐬 */
	if (flag) {
		p = fdc_buffer[type];
		for (i=0; i<track_totalcnt; i++) {
			*p++ = (BYTE)rand_xor();
		}

		/* f[^|C^AJE^ݒ */
		fdc_dataptr[type] = fdc_buffer[type];
		fdc_totalcnt[type] = track_totalcnt;
		fdc_onetime[type] = 32;
		fdc_nowcnt[type] = 0;
		return;
	}

	/* ZN^ZoAf[^ړ */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		secs = 16;
		q = &fdc_buffer[type][0x1000];
		memcpy(q, fdc_buffer[type], 0x1000);
	}
	else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
		secs = 26;
		q = &fdc_buffer[type][0x1A00];
		memcpy(q, fdc_buffer[type], 0x1A00);
	}
	else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		track = fdc_track[type][fdc_drvreg[type]];
		secs = fdc_header[type][fdc_drvreg[type]][track * 6 + 5];
		count = fdc_header[type][fdc_drvreg[type]][track * 6 + 4];
		if (count == 0) {
			count = 128;
		}
		else {
			count *= (WORD)256;
		}
		count *= (WORD)secs;

		/* f[^Rs[ */
		q = &fdc_buffer[type][track_size - count];
		for (j=(count-1); j>=0; j--) {
			q[j] = fdc_buffer[type][j];
		}
	}
	else {
		secs = (WORD)(fdc_buffer[type][0x0005] * 256);
		secs += (WORD)(fdc_buffer[type][0x0004]);
		count = 0;
		/* SZN^܂āATCYv */
		for (j=0; j<secs; j++) {
			p = &fdc_buffer[type][count];
			count += (WORD)((p[0x000f] * 256 + p[0x000e]));
			count += (WORD)0x10;
		}
		/* f[^Rs[ */
		q = &fdc_buffer[type][track_size - count];
		for (j=(count-1); j>=0; j--) {
			q[j] = fdc_buffer[type][j];
		}
	}

	/* GAP/SYNC */
	if (fdc_dden[type]) {
		gap0 = 80;
		gap1 = 50;
		gap2 = 22;
		sync = 12;
		gap  = 0x4e;
	}
	else {
		gap0 = 40;
		gap1 = 26;
		gap2 = 11;
		sync = 6;
		gap  = 0xff;
	}
	if (type) {
		if (fdc_dden[type]) {
			if (secs <= 8) {
				gap3 = 116;
			}
			else {
				if (secs <= 15) {
					gap3 = 84;
				}
				else {
					if (secs <= 26) {
						gap3 = 54;
					}
					else {
						gap3 = 16;
					}
				}
			}
		}
		else {
			if (secs <= 8) {
				gap3 = 58;
			}
			else {
				if (secs <= 15) {
					gap3 = 42;
				}
				else {
					if (secs <= 26) {
						gap3 = 27;
					}
					else {
						gap3 = 8;
					}
				}
			}
		}
	}
	else {
		if (secs <= 5) {
			gap3 = 0x74;
		}
		else {
			if (secs <= 10) {
				gap3 = 0x54;
			}
			else {
				if (secs <= 16) {
					gap3 = 0x33;
				}
				else {
					gap3 = 0x10;
				}
			}
		}
	}

	/* obt@ */
	p = fdc_buffer[type];
	count = 0;

	/* GAP0 */
	for (i=0; i<gap0; i++) {
		*p++ = gap;
	}
	count += (WORD)gap0;

	/* SYNC */
	for (i=0; i<sync; i++) {
		*p++ = 0;
	}
	count += (WORD)sync;

	/* INDEX MARK */
	if (fdc_dden[type]) {
		*p++ = 0xc2;
		*p++ = 0xc2;
		*p++ = 0xc2;
		count += (WORD)3;
	}
	*p++ = 0xfc;
	count += (WORD)1;

	/* GAP1 */
	for (i=0; i<gap1; i++) {
		*p++ = gap;
	}
	count += (WORD)gap1;

	/* ZN^[v */
	for (j=0; j<secs; j++) {
		/* SYNC */
		for (i=0; i<sync; i++) {
			*p++ = 0;
		}
		count += (WORD)sync;

		/* ID ADDRESS MARK */
		if (fdc_dden[type]) {
			p[0] = 0xa1;
			p[1] = 0xa1;
			p[2] = 0xa1;
			p[3] = 0xfe;
			diff = 4;
		}
		else {
			p[0] = 0xfe;
			diff = 1;
		}

		/* ID */
		if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
			(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
			if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
				/* fBA^Cv(2DD) != hCu[h(2D) */
				p[diff + 0] = (BYTE)(fdc_track[type][fdc_drvreg[type]] << 1);
			}
			else {
				/* fBA^Cv == hCu[h ␳Ȃ */
				p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
			}
			p[diff + 1] = fdc_sidereg[type];
			p[diff + 2] = (BYTE)(j + 1);
			p[diff + 3] = 1;
			size = 0x100;
			ddm = FALSE;
		}
		else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
			p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
			p[diff + 1] = fdc_sidereg[type];
			p[diff + 2] = (BYTE)(j + 1);
			if ((p[diff + 0] == 0) && (p[diff + 1] == 0)) {
				p[diff + 3] = 0;
				size = 0x080;
			}
			else {
				p[diff + 3] = 1;
				size = 0x100;
			}
			ddm = FALSE;
		}
		else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
			p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
			p[diff + 1] = fdc_sidereg[type];
			p[diff + 2] = (BYTE)(j + 1);
			p[diff + 3] = fdc_header[type][fdc_drvreg[type]][p[diff + 0] * 6 + 4];
			if (p[diff + 3] == 0) {
				size = 128;
			}
			else {
				size = (WORD)(p[diff + 3] * 256);
			}
			ddm = FALSE;
		}
		else {
			memcpy(p + diff, q, 4);
			size = (WORD)(q[0x000f] * 256 + q[0x000e]);
			if (q[0x0007] != 0x00) {
				ddm = TRUE;
			}
			else {
				ddm = FALSE;
			}
			q += 0x0010;
		}
		calc_set_crc(p, diff + 4);
		p += (diff + 4 + 2);
		count += (WORD)(diff + 4 + 2);

		/* GAP2 */
		for (i=0; i<gap2; i++) {
			*p++ = gap;
		}
		count += (WORD)gap2;

		/* SYNC */
		for (i=0; i<sync; i++) {
			*p++ = 0;
		}
		count += (WORD)sync;

		/* DATA ADDRESS MARK */
		if (fdc_dden[type]) {
			p[0] = 0xa1;
			p[1] = 0xa1;
			p[2] = 0xa1;
			diff = 3;
		}
		else {
			diff = 0;
		}
		if (ddm) {
			/* DELETED DATA MARK */
			p[diff] = 0xf8;
		}
		else {
			/* DATA MARK */
			p[diff] = 0xfb;
		}
		diff += 1;

		/* f[^ */
		memcpy(p + diff, q, size);
		q += size;
		calc_set_crc(p, diff + size);
		p += (diff + size + 2);
		count += (WORD)(diff + size + 2);

		/* GAP3 */
		for (i=0; i<gap3; i++) {
			*p++ = gap;
		}
		count += (WORD)gap3;
	}

	/* GAP4 */
	j = (track_totalcnt - count);
	if (j < track_totalcnt) {
		for (i=0; i<j; i++) {
			*p++ = gap;
		}
		count += (WORD)j;
	}

	/* f[^|C^AJE^ݒ */
	fdc_dataptr[type] = fdc_buffer[type];
	fdc_totalcnt[type] = count;
	fdc_onetime[type] = 32;
	fdc_nowcnt[type] = 0;
}

/*
 *	IDtB[hobt@ɍ
 *	JE^Af[^|C^ݒ
 */
static void FASTCALL fdc_makeaddr(int type, int index)
{
	int i;
	BYTE *p;
	WORD diff;
	WORD offset;
	WORD size;
	int tmp;

	/* ]̂߂̃e|obt@́Aʏobt@ */
	/* ŌׂĐ݂ */
	tmp = fdc_buffer_size[type] - 0x0010;
	p = &fdc_buffer[type][tmp];

	/* ID ADDRESS MARK */
	if (fdc_dden[type]) {
		p[0] = 0xa1;
		p[1] = 0xa1;
		p[2] = 0xa1;
		p[3] = 0xfe;
		diff = 4;
	}
	else {
		p[0] = 0xfe;
		diff = 1;
	}

	/* ID */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		/* 2D/2DD̏ꍇAC,H,R,N͊m肷 */
		if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			p[diff + 0] = (BYTE)(fdc_track[type][fdc_drvreg[type]] << 1);
		}
		else {
			/* fBA^Cv == hCu[h ␳Ȃ */
			p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
		}
		p[diff + 1] = fdc_sidereg[type];
		p[diff + 2] = (BYTE)(index + 1);
		p[diff + 3] = 1;
	}
	else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
		/* 2HD̏ꍇAC,H,R,N͊m肷 */
		p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
		p[diff + 1] = fdc_sidereg[type];
		p[diff + 2] = (BYTE)(index + 1);
		if ((p[diff + 0] == 0) && (p[diff + 1] == 0)) {
			p[diff + 3] = 0;
		}
		else {
			p[diff + 3] = 1;
		}
	}
	else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		/* VFD̏ꍇAC,H,R͊mAN̓wb_擾 */
		p[diff + 0] = fdc_track[type][fdc_drvreg[type]];
		p[diff + 1] = fdc_sidereg[type];
		p[diff + 2] = (BYTE)(index + 1);
		p[diff + 3] = fdc_header[type][fdc_drvreg[type]][p[diff + 0] * 6 + 4];
	}
	else {
		/* obt@ړĨZN^܂Ői߂ */
		offset = 0;
		i = 0;
		while (i < index) {
			/* ZN^TCY擾 */
			size = fdc_buffer[type][offset + 0x000f];
			size *= (WORD)256;
			size |= fdc_buffer[type][offset + 0x000e];

			/* ̃ZN^֐i߂ */
			offset += size;
			offset += (WORD)0x10;

			i++;
		}

		/* C,H,R,NRs[ */
		memcpy(&p[diff], &fdc_buffer[type][offset], 4);
	}

	/* CRC */
	calc_set_crc(p, diff + 4);

	/* f[^|C^AJE^ݒ */
	fdc_dataptr[type] = &fdc_buffer[type][tmp + diff];
	fdc_totalcnt[type] = 6;
	fdc_onetime[type] = 32;
	fdc_nowcnt[type] = 0;

	/* ZN^WX^ɃgbNԍݒ */
	fdc_secreg[type] = p[diff];
}

/*
 *	CfbNXJE^̃ZN^ֈڂ
 */
static int FASTCALL fdc_next_index(int type)
{
	int max_track;
	int track;
	int secs;

	ASSERT(fdc_ready[type][fdc_drvreg[type]] != FDC_TYPE_NOTREADY);

	/* fBXN^Cv`FbN */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		/* 2D/2DDȂ16ZN^Œ */
		fdc_indexcnt[type] = (BYTE)((fdc_indexcnt[type] + 1) & 0x0f);
		track = fdc_track[type][fdc_drvreg[type]];
		if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
		max_track = 40;
		if (track >= max_track) {
			return -1;
		}
		return fdc_indexcnt[type];
	}

	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
		/* 2HDȂ26ZN^Œ */
		fdc_indexcnt[type] = (BYTE)((fdc_indexcnt[type] + 1) % 26);
		max_track = 77;
		if (fdc_track[type][fdc_drvreg[type]] >= max_track) {
			return -1;
		}
		return fdc_indexcnt[type];
	}

	/* D77/VFD */
	ASSERT( (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_D77) ||
			(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD));

	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		track = fdc_track[type][fdc_drvreg[type]];
		secs = fdc_header[type][fdc_drvreg[type]][track * 6 + 5];
	}
	else if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_D77) {
		secs = fdc_buffer[type][0x0005];
		secs *= 256;
		secs |= fdc_buffer[type][0x0004];
	}
	if (secs == 0) {
		/* AtH[}bg */
		if (type) {
			fdc_indexcnt[type] = (BYTE)((fdc_indexcnt[type] + 1) % 26);
		}
		else {
			fdc_indexcnt[type] = (BYTE)((fdc_indexcnt[type] + 1) & 0x0f);
		}
		return -1;
	}
	else {
		fdc_indexcnt[type]++;
		if (fdc_indexcnt[type] >= secs) {
			fdc_indexcnt[type] = 0;
		}
		return fdc_indexcnt[type];
	}
}

/*
 *	ID}[NT
 */
static BOOL FASTCALL fdc_idmark(int type, WORD *p)
{
	WORD offset;
	BYTE dat;

	offset = *p;

	if (fdc_dden[type]) {
		/* {x */
		while (offset < fdc_totalcnt[type]) {
			/* id address mark : A1 A1 A1 FEF5 F5 F5 FE */
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat == 0xfe) {
				*p = offset;
				return TRUE;
			}
		}
	}
	else {
		/* Px */
		while (offset < fdc_totalcnt[type]) {
			/* sync : 00 00 00 00 00 00 */
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			/* id address mark : FE */
			dat = fdc_buffer[type][offset++];
			if (dat == 0xfe) {
				*p = offset;
				return TRUE;
			}
		}
	}

	*p = offset;
	return FALSE;
}

/*
 *	f[^}[NT
 */
static BOOL FASTCALL fdc_datamark(int type, WORD *p, BOOL *deleted_mark)
{
	WORD offset;
	BYTE dat;

	offset = *p;

	if (fdc_dden[type]) {
		/* {x */
		while (offset < fdc_totalcnt[type]) {
			/* data mark : A1 A1 A1 FBF5 F5 F5 FB */
			/* deleted data mark : A1 A1 A1 F8F5 F5 F5 F8 */
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if ((dat != 0xa1) && (dat != 0xf5)) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if ((dat == 0xfb) || (dat == 0xf8)) {
				*p = offset;
				if (dat == 0xfb) {
					/* 0xFB : DATA MARK */
					*deleted_mark = FALSE;
				}
				else {
					/* 0xF8 : DELETED DATA MARK */
					*deleted_mark = TRUE;
				}
				return TRUE;
			}
		}
	}
	else {
		/* Px */
		while (offset < fdc_totalcnt[type]) {
			/* sync : 00 00 00 00 00 00 */
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			dat = fdc_buffer[type][offset++];
			if (dat != 0x00) {
				continue;
			}
			/* data mark : FB */
			/* deleted data mark : F8 */
			dat = fdc_buffer[type][offset++];
			if ((dat == 0xfb) || (dat == 0xf8)) {
				*p = offset;
				if (dat == 0xfb) {
					/* 0xFB : DATA MARK */
					*deleted_mark = FALSE;
				}
				else {
					/* 0xF8 : DELETED DATA MARK */
					*deleted_mark = TRUE;
				}
				return TRUE;
			}
		}
	}

	*p = offset;
	return FALSE;
}

/*
 *	gbNݏI
 */
static BOOL FASTCALL fdc_writetrk(int type)
{
	int total;
	int sectors;
	WORD offset;
	WORD seclen;
	WORD writep;
	int i;
	int handle;
	BOOL ddm;

	/*  */
	total = 0;
	sectors = 0;

	/* ZN^ƁAf[^̃g[^TCY𐔂 */
	offset = 0;
	while (offset < fdc_totalcnt[type]) {
		/* ID}[NT */
		if (!fdc_idmark(type, &offset)) {
			break;
		}
		/* C,H,R,N̎$F7 */
		offset += (WORD)4;
		if (offset >= fdc_totalcnt[type]) {
			return FALSE;
		}
		if (fdc_buffer[type][offset] != 0xf7) {
			return FALSE;
		}
		offset++;
		/* f[^}[NT */
		if (!fdc_datamark(type, &offset, &ddm)) {
			return FALSE;
		}
		/* OXvZA$F7T */
		seclen = 0;
		while(offset < fdc_totalcnt[type]) {
			if (fdc_buffer[type][offset] == 0xf7) {
				break;
			}
			offset++;
			seclen++;
		}
		if (offset >= fdc_totalcnt[type]) {
			return FALSE;
		}
		/* ZN^ok */
		total += seclen;
		sectors++;
		offset++;
	}

	/* 2D/2DDfBȀꍇAtotal=0x1000, sectors=0x10K{ */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		/* total, sectorš */
		if (total != 0x1000) {
			return FALSE;
		}
		if (sectors != 16) {
			return FALSE;
		}

		/* {C,H,NȂǃ`FbNׂ */
		return TRUE;
	}

	/* 2HDfBȀꍇAtotal=0x1A00, sectors=0x1AK{ */
	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
		/* total, sectorš */
		if ((fdc_trkreg[type] == 0) && (fdc_sidereg[type] == 0)) {
			if (total != 0x0D00) {
				return FALSE;
			}
		}
		else {
			if (total != 0x1A00) {
				return FALSE;
			}
		}
		if (sectors != 26) {
			return FALSE;
		}

		/* {C,H,NȂǃ`FbNׂ */
		return TRUE;
	}

	/* VFDfBȀꍇA܂ */
	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		/* {C,H,NȂǃ`FbNׂ */
		return TRUE;
	}

	/* ZN^Ƃ0x10ԂA]vɂ */
	if ((DWORD)(total + (sectors * 0x10)) > fdc_trklen[type][fdc_drvreg[type]]) {
		return FALSE;
	}

	/* ޕKvȂꍇ͐I */
	if ((sectors == 0) && (total == 0)) {
		return TRUE;
	}

	/* ܂邱Ƃ킩̂ŁAf[^쐬 */
	writep = 0;
	offset = 0;
	for (i=0; i<sectors; i++) {
		/* ID}[N */
		fdc_idmark(type, &offset);

		/* 0x10wb_̍쐬 */
		fdc_buffer[type][writep++] = fdc_buffer[type][offset++];
		fdc_buffer[type][writep++] = fdc_buffer[type][offset++];
		fdc_buffer[type][writep++] = fdc_buffer[type][offset++];
		fdc_buffer[type][writep++] = fdc_buffer[type][offset++];
		fdc_buffer[type][writep++] = (BYTE)(sectors & 0xff);
		fdc_buffer[type][writep++] = (BYTE)(sectors >> 8);
		if (type) {
			fdc_buffer[type][writep++] = (BYTE)(fdc_dden[type]^0x40);
		}
		else {
			fdc_buffer[type][writep++] = 0x00;
		}
		offset++;

		/* f[^}[N̏ */
		fdc_datamark(type, &offset, &ddm);
		if (ddm) {
			fdc_buffer[type][writep++] = 0x10;
		}
		else {
			fdc_buffer[type][writep++] = 0x00;
		}

		/* U[uGANA */
		/* ZN^OX͌ŏ */
		memset(&fdc_buffer[type][writep], 0, 8);
		writep += (WORD)8;

		/* OX𐔂Rs[ */
		seclen = 0;
		while (fdc_buffer[type][offset] != 0xf7) {
			fdc_buffer[type][writep++] = fdc_buffer[type][offset++];
			seclen++;
		}
		offset++;

		/* ZN^OXݒ */
		fdc_buffer[type][writep - seclen - 2] = (BYTE)(seclen & 0xff);
		fdc_buffer[type][writep - seclen - 1] = (BYTE)(seclen >> 8);
	}

	/* t@CɃf[^ */
	handle = file_open(fdc_fname[type][fdc_drvreg[type]], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, fdc_seekofs[type][fdc_drvreg[type]])) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, fdc_buffer[type], (sectors * 0x10) + total)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* I */
	return TRUE;
}

/*
 *	ZN^ݏI
 */
static BOOL FASTCALL fdc_writesec(int type)
{
	DWORD offset;
	int handle;

	/* assert */
	ASSERT(fdc_drvreg[type] < FDC_DRIVES);
	ASSERT(fdc_ready[type][fdc_drvreg[type]] != FDC_TYPE_NOTREADY);
	ASSERT(fdc_dataptr[type]);
	ASSERT(fdc_totalcnt[type] > 0);

	/* ItZbgZo */
	offset = fdc_seekofs[type][fdc_drvreg[type]];
	offset += fdc_secofs[type][fdc_drvreg[type]];

	/*  */
	handle = file_open(fdc_fname[type][fdc_drvreg[type]], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, fdc_dataptr[type], fdc_totalcnt[type])) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	gbNAZN^ATChƈvZN^
 *	JE^Af[^|C^ݒ
 */
static BYTE FASTCALL fdc_readsec(
	int type, BYTE track, BYTE sector, BYTE side, BOOL sidecmp, int *secidx, int cpos)
{
	int secs;
	int	len;
	int i;
	WORD offset;
	WORD size;
	BYTE stat;
	DWORD vfdoffset;
	int fdctrack;

	/* assert */
	ASSERT(fdc_drvreg[type] < FDC_DRIVES);
	ASSERT(fdc_ready[type][fdc_drvreg[type]] != FDC_TYPE_NOTREADY);
	UNUSED(cpos);

	/* wb_̂gbNԍ(rp)ݒ */
	fdctrack = fdc_track[type][fdc_drvreg[type]];

	/* 2D/2DDt@C̏ꍇ */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			fdctrack = (BYTE)(fdctrack << 1);
		}
		if (track != fdctrack) {
			return FDC_ST_RECNFND;
		}
		if (side != fdc_sidereg[type]) {
			return FDC_ST_RECNFND;
		}
		if ((sector < 1) || (sector > 16)) {
			return FDC_ST_RECNFND;
		}

		/* f[^|C^ݒ */
		fdc_dataptr[type] = &fdc_buffer[type][(sector - 1) * 0x0100];
		fdc_secofs[type][fdc_drvreg[type]] = (sector - 1) * 0x0100;

		/* JE^ݒ */
		fdc_totalcnt[type] = 0x0100;
		fdc_onetime[type] = 32;
		fdc_nowcnt[type] = 0;

		/* ZN^̃CfbNXԂ */
		*secidx = sector - 1;

		return 0;
	}

	/* 2HDt@C̏ꍇ */
	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD) {
		if (track != fdctrack) {
			return FDC_ST_RECNFND;
		}
		if (side != fdc_sidereg[type]) {
			return FDC_ST_RECNFND;
		}
		if ((sector < 1) || (sector > 26)) {
			return FDC_ST_RECNFND;
		}

		if ((track == 0) && (side == 0)) {
			size = 0x0080;
		}
		else {
			size = 0x0100;
		}

		/* f[^|C^ݒ */
		fdc_dataptr[type] = &fdc_buffer[type][(sector - 1) * size];
		fdc_secofs[type][fdc_drvreg[type]] = (sector - 1) * size;

		/* JE^ݒ */
		fdc_totalcnt[type] = size;
		fdc_onetime[type] = 32;
		fdc_nowcnt[type] = 0;

		/* ZN^̃CfbNXԂ */
		*secidx = sector - 1;

		return 0;
	}

	/* VFDt@C̏ꍇ */
	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		if (track != fdctrack) {
			return FDC_ST_RECNFND;
		}
		if (side != fdc_sidereg[type]) {
			return FDC_ST_RECNFND;
		}

		/* gbNւ̃ItZbg`FbN */
		vfdoffset  = fdc_header[type][fdc_drvreg[type]][track * 6 + 3];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[type][fdc_drvreg[type]][track * 6 + 2];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[type][fdc_drvreg[type]][track * 6 + 1];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[type][fdc_drvreg[type]][track * 6 + 0];
		if (vfdoffset == 0) {
			return FDC_ST_RECNFND;
		}

		/* gbÑZN^`FbN */
		len = fdc_header[type][fdc_drvreg[type]][track * 6 + 4];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}
		size = fdc_header[type][fdc_drvreg[type]][track * 6 + 5];
		if ((sector < 1) || (sector > size)) {
			return FDC_ST_RECNFND;
		}

		/* f[^|C^ݒ */
		fdc_dataptr[type] = &fdc_buffer[type][(sector - 1) * len];
		fdc_secofs[type][fdc_drvreg[type]] = (sector - 1) * len;

		/* JE^ݒ */
		fdc_totalcnt[type] = (WORD)len;
		fdc_onetime[type] = 32;
		fdc_nowcnt[type] = 0;

		/* ZN^̃CfbNXԂ */
		*secidx = sector - 1;

		return 0;
	}

	/* D77t@C̏ꍇ */
	secs = fdc_buffer[type][0x0005];
	secs *= (WORD)256;
	secs |= fdc_buffer[type][0x0004];
	if (secs == 0) {
		return FDC_ST_RECNFND;
	}

	offset = 0;
	/* ZN^[v */
	for (i=0; i<secs; i++) {
		/* ̃ZN^̃TCYɎ擾 */
		size = fdc_buffer[type][offset + 0x000f];
		size *= (WORD)256;
		size |= fdc_buffer[type][offset + 0x000e];

		/* C,H,RvZN^邩 */
		if (fdc_buffer[type][offset + 0] != track) {
			offset += size;
			offset += (WORD)0x10;
			continue;
		}
		/* TChbit0ȊO͍lȂ悤ɕύX */
		if (sidecmp) {
			if ((BYTE)(fdc_buffer[type][offset + 1] & 1) != side) {
				offset += size;
				offset += (WORD)0x10;
				continue;
			}
		}
		if (fdc_buffer[type][offset + 2] != sector) {
			offset += size;
			offset += (WORD)0x10;
			continue;
		}

		/* P`FbN */
		if (type) {
			if (fdc_dden[type]) {
				if (fdc_buffer[type][offset + 0x0006] != 0) {
					offset += size;
					offset += (WORD)0x10;
					continue;
				}
			}
			else {
				if (fdc_buffer[type][offset + 0x0006] != 0x40) {
					offset += size;
					offset += (WORD)0x10;
					continue;
				}
			}
		}
		else {
			if (fdc_buffer[type][offset + 0x0006] != 0) {
				offset += size;
				offset += (WORD)0x10;
				continue;
			}
		}

		/* f[^|C^AJE^ݒ */
		fdc_dataptr[type] = &fdc_buffer[type][offset + 0x0010];
		fdc_secofs[type][fdc_drvreg[type]] = offset + 0x0010;
		fdc_totalcnt[type] = size;
		fdc_onetime[type] = 32;
		fdc_nowcnt[type] = 0;

		/* Xe[^Xݒ */
		stat = 0;
		if (fdc_buffer[type][offset + 0x0007] != 0) {
			stat |= FDC_ST_RECTYPE;
		}
		if (fdc_buffer[type][offset + 0x0008] == 0xb0) {
			/* Oɐݒ肵Xe[^XɂȂC (ݼ޼) */
			stat |= FDC_ST_CRCERR;
		}

		/* ZN^̃CfbNXԂ */
		*secidx = i;

		return stat;
	}

	/* SZN^AȂ */
	return FDC_ST_RECNFND;
}

/*
 *	gbNǂݍ
 */
static void FASTCALL fdc_readbuf(int type, int drive)
{
	DWORD offset;
	DWORD min_offset;
	DWORD len;
	DWORD secs;
	int trkside;
	int handle;
	int max_track;
	int trk;
	DWORD track_size;

	/* fBA`FbN */
	if ((type != FDC_M) && (type != FDC_S)) {
		LOG1(Warning, "type:%d\n", type);
		return;
	}

	/* C[W͖ */
	fdi_init[type] = FALSE;

	/* hCu`FbN */
	if ((drive < 0) || (drive >= FDC_DRIVES)) {
		LOG1(Warning, "drive:%d\n", drive);
		return;
	}

	/* fB`FbN */
	if (fdc_ready[type][drive] == FDC_TYPE_NOTREADY) {
		return;
	}

	/* CfbNXz[JE^NA */
	fdc_indexcnt[type] = 0;

	/* gbN~Q{TCh */
	trkside = fdc_track[type][drive] * 2 + fdc_sidereg[type];

	/*
	 * 2D/2DDt@C
	 */
	if ((fdc_ready[type][drive] == FDC_TYPE_2D) ||
		(fdc_ready[type][drive] == FDC_TYPE_2DD)) {
		if (fdc_ready[type][drive] == FDC_TYPE_2D) {
			max_track = 80;
		}
		else {
			max_track = 160;
			trkside = fdc_track[type][drive] * 4 + fdc_sidereg[type];
		}

		/* ͈̓`FbN */
		if (trkside >= max_track) {
			memset(fdc_buffer[type], 0, 0x1000);
			return;
		}

		/* ItZbgZo */
		offset = (DWORD)trkside;
		offset *= 0x1000;
		fdc_seekofs[type][drive] = offset;
		fdc_trklen[type][drive] = 0x1000;

		/* ǂݍ */
		memset(fdc_buffer[type], 0, 0x1000);
		handle = file_open(fdc_fname[type][drive], OPEN_R);
		if (handle == -1) {
			return;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return;
		}
		file_read(handle, fdc_buffer[type], 0x1000);
		file_close(handle);
		return;
	}

	/*
	 * 2HDt@C
	 */
	if (fdc_ready[type][drive] == FDC_TYPE_2HD) {
		/* ͈̓`FbN */
		if (trkside >= 154) {
			memset(fdc_buffer[type], 0, 0x1A00);
			return;
		}

		/* ItZbgZo */
		offset = (DWORD)trkside;
		offset *= 0x1A00;
		if (trkside == 0) {
			len = 0x0D00;
		}
		else {
			len = 0x1A00;
			offset -= 0x0D00;
		}
		fdc_seekofs[type][drive] = offset;
		fdc_trklen[type][drive] = len;

		/* ǂݍ */
		memset(fdc_buffer[type], 0, 0x1A00);
		handle = file_open(fdc_fname[type][drive], OPEN_R);
		if (handle == -1) {
			return;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return;
		}
		file_read(handle, fdc_buffer[type], len);
		file_close(handle);
		return;
	}

	/*
	 * VFDt@C
	 */
	if (fdc_ready[type][drive] == FDC_TYPE_VFD) {

		/* ͈̓`FbN */
		if (trkside >= 80) {
			memset(fdc_buffer[type], 0, 0x1000);
			return;
		}

		/* ItZbgZo */
		offset  = fdc_header[type][drive][trkside * 6 + 3];
		offset *= 256;
		offset |= fdc_header[type][drive][trkside * 6 + 2];
		offset *= 256;
		offset |= fdc_header[type][drive][trkside * 6 + 1];
		offset *= 256;
		offset |= fdc_header[type][drive][trkside * 6 + 0];
		fdc_seekofs[type][drive] = offset;

		len		= fdc_header[type][drive][trkside * 6 + 4];
		secs	= fdc_header[type][drive][trkside * 6 + 5];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}
		fdc_trklen[type][drive] = len * secs;

		/* ǂݍ */
		memset(fdc_buffer[type], 0, fdc_trklen[type][drive]);
		handle = file_open(fdc_fname[type][drive], OPEN_R);
		if (handle == -1) {
			return;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return;
		}
		file_read(handle, fdc_buffer[type], fdc_trklen[type][drive]);
		file_close(handle);
		return;
	}

	/*
	 * D77t@C
	 */
	if (fdc_header[type][drive][0x001b] == 0x00) {
		max_track = 84;
	}
	else {
		max_track = 164;
	}
	if (fdc_header[type][drive][0x001b] != 0x20) {
		track_size = fdc_buffer_size[FDC_M];
	}
	else {
		track_size = fdc_buffer_size[FDC_S];
	}

	/* ANZXĂŏIgbN̎Zo */
	min_offset = (0x0020 + max_track * 4);
	for (trk = 0; trk < max_track; trk++) {
		if ((DWORD)(0x0020 + trk * 4) >= min_offset) {
			break;
		}

		offset  = fdc_header[type][drive][0x0020 + trk * 4 + 3];
		offset *= 256;
		offset |= fdc_header[type][drive][0x0020 + trk * 4 + 2];
		offset *= 256;
		offset |= fdc_header[type][drive][0x0020 + trk * 4 + 1];
		offset *= 256;
		offset |= fdc_header[type][drive][0x0020 + trk * 4 + 0];
		if ((offset < min_offset) && (offset >= 0x20)) {
			min_offset = offset;
		}
	}
	max_track = (min_offset - 0x20) / 4;

	/* 2DDt@C2Dǂ݂ꍇAŕ␳ (AV40pOS-9) */
	if (fdc_header[type][drive][0x001b] == 0x10) {
		trkside = fdc_track[type][drive] * 4 + fdc_sidereg[type];
	}

	/* I[o[`FbN */
	if (trkside >= max_track) {
		/* gbNI[o[AZN^0 */
		fdc_buffer[type][4] = 0;
		fdc_buffer[type][5] = 0;
		return;
	}

	/* wb_t@Cɏ]AItZbgEOXZo */
	offset = *(DWORD *)(&fdc_header[type][drive][0x0020 + trkside * 4]);
	if (offset == 0) {
		/* ݂ȂgbN */
		fdc_buffer[type][4] = 0;
		fdc_buffer[type][5] = 0;
		return;
	}

	len = *(DWORD *)(&fdc_header[type][drive][0x0020 + (trkside + 1) * 4]);
	if ((len == 0) || (trkside + 1 >= max_track)) {
		/* ŏIgbN */
		len = *(DWORD *)(&fdc_header[type][drive][0x0014]);
	}
	len -= offset;
	if (len > track_size) {
		len = track_size;
	}
	fdc_seekofs[type][drive] = offset;
	fdc_trklen[type][drive] = len;

	/* V[NAǂݍ */
	memset(fdc_buffer[type], 0, track_size);
	handle = file_open(fdc_fname[type][drive], OPEN_R);
	if (handle == -1) {
		return;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return;
	}
	file_read(handle, fdc_buffer[type], len);
	file_close(handle);
}

/*
 *	D77t@C wb_ǂݍ
 */
static BOOL FASTCALL fdc_readhead(int type, int drive, int index)
{
	int i;
	DWORD offset;
	DWORD temp;
	int handle;
	DWORD pretemp;
	BOOL flag;

	/* assert */
	ASSERT((type == FDC_M) || (type == FDC_S));
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((index >= 0) && (index < FDC_MEDIAS));
	ASSERT(fdc_ready[type][drive] == FDC_TYPE_D77);

	/* ItZbg */
	offset = fdc_foffset[type][drive][index];

	/* V[NAǂݍ */
	handle = file_open(fdc_fname[type][drive], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, fdc_header[type][drive], 0x2b0)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* ^Cv`FbNACgveNgݒ */
	if ((type == FDC_M) &&
		(fdc_header[type][drive][0x001b] != 0x00) &&
		(fdc_header[type][drive][0x001b] != 0x10)) {
		/* 2D/2DDłȂ */
		return FALSE;
	}
	if ((type == FDC_S) &&
		(fdc_header[type][drive][0x001b] != 0x20)) {
		/* 2HDłȂ */
		return FALSE;
	}
	if (fdc_fwritep[type][drive]) {
		fdc_writep[type][drive] = TRUE;
	}
	else {
		if (fdc_header[type][drive][0x001a] & 0x10) {
			fdc_writep[type][drive] = TRUE;
		}
		else {
			fdc_writep[type][drive] = FALSE;
		}
	}

	/* ЖʃfBAɉݒ */
	pretemp = 0;
	flag = FALSE;

	/* gbNItZbgݒ */
	for (i=0; i<164; i++) {
		temp = 0;
		temp |= fdc_header[type][drive][0x0020 + i * 4 + 3];
		temp *= 256;
		temp |= fdc_header[type][drive][0x0020 + i * 4 + 2];
		temp *= 256;
		temp |= fdc_header[type][drive][0x0020 + i * 4 + 1];
		temp *= 256;
		temp |= fdc_header[type][drive][0x0020 + i * 4 + 0];

		if (temp != 0) {
			/* f[^ */
			temp += offset;
			*(DWORD *)(&fdc_header[type][drive][0x0020 + i * 4]) = temp;

			/* gbNɃf[^ΗʃfBAƐݒ */
			if ((i % 2) &&
				(fdc_header[type][drive][0x001b] == 0x20)) {
				fdc_sided[drive] = (BYTE)0x80;
			}
		}

		/* 8"ΉȂ̂2HDfBÂݑΏۂƂ */
		if (fdc_header[type][drive][0x001b] == 0x20) {
			if ((i > 0) && ((i % 2) == 0)) {
				/* gbN݃`FbN */
				if ((pretemp != 0) &&		/* ItZbg */
					(pretemp != temp)) {	/* ̂ */
					flag = TRUE;			/* ʃfBAƔ */
				}
			}
		}

		pretemp = temp;
	}

	/* Ж/ʃfBAݒ */
	if (fdc_header[type][drive][0x001b] == 0x20) {
		if (flag) {
			fdc_sided[drive] = (BYTE)0x80;
		}
		else {
			fdc_sided[drive] = (BYTE)0;
		}
	}

	/* fBXNTCY{ItZbg */
	temp = 0;
	temp |= fdc_header[type][drive][0x001c + 3];
	temp *= 256;
	temp |= fdc_header[type][drive][0x001c + 2];
	temp *= 256;
	temp |= fdc_header[type][drive][0x001c + 1];
	temp *= 256;
	temp |= fdc_header[type][drive][0x001c + 0];
	temp += offset;
	*(DWORD *)(&fdc_header[type][drive][0x0014]) = temp;

	return TRUE;
}

/*
 *	VFDt@C wb_ǂݍ
 */
static BOOL FASTCALL fdc_readhead_vfd(int drive)
{
	int handle;

	/* assert */
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT(fdc_ready[FDC_M][drive] == FDC_TYPE_VFD);

	/* V[NAǂݍ */
	handle = file_open(fdc_fname[FDC_M][drive], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, 0)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, fdc_header[FDC_M][drive], 0x1e0)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	݂̃fBÃCgveNg؂ւ
 */
BOOL FASTCALL fdc_setwritep(int type, int drive, BOOL writep)
{
	BYTE header[0x2b0];
	DWORD offset;
	int handle;

	/* assert */
	ASSERT((type == FDC_M) || (type == FDC_S));
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((writep == TRUE) || (writep == FALSE));

	/* fBłȂ΂ȂȂ */
	if (fdc_ready[type][drive] == FDC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* t@CݕsȂ_ */
	if (fdc_fwritep[type][drive]) {
		return FALSE;
	}

	/* ǂݍ݁AݒA */
	if (fdc_ready[type][drive] == FDC_TYPE_D77) {
		offset = fdc_foffset[type][drive][fdc_media[type][drive]];
		handle = file_open(fdc_fname[type][drive], OPEN_RW);
		if (handle == -1) {
			return FALSE;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_read(handle, header, 0x2b0)) {
			file_close(handle);
			return FALSE;
		}
		if (writep) {
			header[0x1a] |= 0x10;
		}
		else {
			header[0x1a] &= ~0x10;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_write(handle, header, 0x2b0)) {
			file_close(handle);
			return FALSE;
		}
	}

	/*  */
	file_close(handle);
	fdc_writep[type][drive] = writep;
	return TRUE;
}

/*
 *	fBAԍݒ
 */
BOOL FASTCALL fdc_setmedia(int type, int drive, int index)
{
	/* assert */
	ASSERT((type == FDC_M) || (type == FDC_S));
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((index >= 0) && (index < FDC_MEDIAS));

	/* fBԂ */
	if (fdc_ready[type][drive] == FDC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* 2D/2DDt@C̏ꍇAindex = 0 */
	if (((fdc_ready[type][drive] == FDC_TYPE_2D) ||
		 (fdc_ready[type][drive] == FDC_TYPE_2DD)) && ((index != 0) || (type != FDC_M))) {
		return FALSE;
	}
	/* 2HDt@C̏ꍇAindex = 0 */
	if ((fdc_ready[type][drive] == FDC_TYPE_2HD) && ((index != 0) || (type != FDC_S))) {
		return FALSE;
	}

	/* VFDt@C̏ꍇAindex = 0`FbNăwb_ǂݍ */
	if (fdc_ready[type][drive] == FDC_TYPE_VFD) {
		if (type != FDC_M) {
			return FALSE;
		}
		if (index != 0) {
			return FALSE;
		}
		if (!fdc_readhead_vfd(drive)) {
			return FALSE;
		}
	}

	/* index > 0 ȂAfdc_foffset𒲂ׂ>0Kv */
	if (index > 0) {
		if (fdc_foffset[type][drive][index] == 0) {
			return FALSE;
		}
	}

	/* D77t@C̏ꍇAwb_ǂݍ */
	if (fdc_ready[type][drive] == FDC_TYPE_D77) {
		/* CgveNg͓Őݒ */
		if (!fdc_readhead(type, drive, index)) {
			return FALSE;
		}
	}
	else {
		/* 2D/2HD/VFDt@CȂAt@Cɏ] */
		fdc_writep[type][drive] = fdc_fwritep[type][drive];
	}

	/* fBAꂽꍇAꎞCWFNg */
	if (fdc_media[type][drive] != index) {
		fdc_teject[type][drive] = FALSE;
	}

	/* f[^obt@ǂݍ݁A[NZ[u */
	fdc_media[type][drive] = (BYTE)index;
	if (drive == fdc_drvreg[type]) {
		/* JghCuƈvꍇ̂݃obt@ǂݍ݂s */
		fdc_readbuf(type, drive);
	}

	return TRUE;
}

/*
 *	D77t@ĆAfBAі̎擾
 */
static int FASTCALL fdc_chkd77(int type, int drive)
{
	int i;
	int handle;
	int count;
	int max_track;
	int trk;
	DWORD offset;
	DWORD len;
	DWORD trkoffset;
	DWORD min_offset;
	BYTE buf[0x2b0];

	/*  */
	for (i=0; i<FDC_MEDIAS; i++) {
		fdc_foffset[type][drive][i] = 0;
		fdc_name[type][drive][i][0] = '\0';
	}
	count = 0;
	offset = 0;

	/* t@CI[v */
	handle = file_open(fdc_fname[type][drive], OPEN_R);
	if (handle == -1) {
		return count;
	}

	/* fBA[v */
	while (count < FDC_MEDIAS) {
		/* V[N */
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return count;
		}

		/* ǂݍ */
		if (!file_read(handle, buf, 0x02b0)) {
			file_close(handle);
			return count;
		}

		/* ^Cv`FbNB2D,2DD,2HDɂꂼΉ */
		if (((type == FDC_M) && (buf[0x1b] != 0x00) && (buf[0x1b] != 0x10)) ||
			((type == FDC_S) && (buf[0x1b] != 0x20))) {
			file_close(handle);
			return count;
		}
		if (buf[0x1b] == 0x00) {
			max_track = 84;
		}
		else {
			max_track = 164;
		}

		/* ANZXĂŏIgbN̎Zo */
		min_offset = (0x0020 + max_track * 4);
		for (trk = 0; trk < max_track; trk++) {
			if ((DWORD)(0x0020 + trk * 4) >= min_offset) {
				break;
			}

			trkoffset = buf[0x20 + trk * 4 + 3];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 2];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 1];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 0];
			if ((trkoffset < min_offset) && (trkoffset >= 0x20)) {
				min_offset = trkoffset;
			}
		}
		max_track = (min_offset - 0x20) / 4;

		/* ok,t@CAItZbgi[ */
		buf[0x10] = '\0';
		memcpy(fdc_name[type][drive][count], buf, 17);
		fdc_foffset[type][drive][count] = offset;

		/* next */
		len = 0;
		len |= buf[0x1f];
		len *= 256;
		len |= buf[0x1e];
		len *= 256;
		len |= buf[0x1d];
		len *= 256;
		len |= buf[0x1c];
		if (len == 0) {
			file_close(handle);
			return count;
		}
		for (trk = 0; trk < max_track; trk++) {
			trkoffset = buf[0x20 + trk * 4 + 3];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 2];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 1];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 0];
			if (trkoffset >= len) {
				file_close(handle);
				return count;
			}
		}
		offset += len;
		count++;
	}

	/* ő僁fBAɒB */
	file_close(handle);
	return count;
}

/*
 *	VFDt@C`FbN
 */
static int FASTCALL fdc_chkvfd(int drive)
{
	int	i;
	int handle;
	BYTE buf[0x20];

	/*  */
	for (i=0; i<FDC_MEDIAS; i++) {
		fdc_foffset[FDC_M][drive][i] = 0;
		fdc_name[FDC_M][drive][i][0] = '\0';
	}

	/* t@CI[v */
	handle = file_open(fdc_fname[FDC_M][drive], OPEN_R);
	if (handle == -1) {
		return 0;
	}

	/* wb_̃`FbN */
	if (!file_seek(handle, 0)) {
		file_close(handle);
		return 0;
	}
	if (!file_read(handle, buf, 0x0020)) {
		file_close(handle);
		return 0;
	}
	file_close(handle);

	if (	(buf[0x00] != 0xe0) || (buf[0x01] != 0x01) ||
			(buf[0x02] != 0x00) || (buf[0x03] != 0x00)) {
		/* gbN0̃ItZbg0x01e0ȊOȂVFDł͂Ȃ(蔲c) */
		return 0;
	}

	return 1;
}

/*
 *	fBXNt@Cݒ
 */
int FASTCALL fdc_setdisk(int type, int drive, char *fname)
{
	BOOL writep;
	int handle;
	DWORD fsize;
	int count;
	BOOL flag;
	int i;

	ASSERT((type == FDC_M) || (type == FDC_S));
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));

	/* mbgfBɂꍇ */
	if (fname == NULL) {
		fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
		fdc_fname[type][drive][0] = '\0';
		if (type) {
			fdc_doorlock[drive] = (BYTE)0;
			fdc_sided[drive] = (BYTE)0x80;
		}
		return 1;
	}

	/* t@CI[vAt@CTCY𒲂ׂ */
	if (strlen(fname) >= sizeof(fdc_fname[drive])) {
		fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
		fdc_fname[type][drive][0] = '\0';
		if (type) {
			fdc_doorlock[drive] = (BYTE)0;
			fdc_sided[drive] = (BYTE)0x80;
		}
		return 1;
	}
	strcpy(fdc_fname[type][drive], fname);
	writep = FALSE;
	handle = file_open(fdc_fname[type][drive], OPEN_RW);
	if (handle == -1) {
		handle = file_open(fdc_fname[type][drive], OPEN_R);
		if (handle == -1) {
			fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
			return 0;
		}
		writep = TRUE;
	}
	fsize = file_getsize(handle);
	file_close(handle);

	/*
	 * 2Dt@C
	 */
	if ((type == FDC_M) && (fsize == 327680)) {
		/* ^CvAݑݒ */
		fdc_ready[type][drive] = FDC_TYPE_2D;
		fdc_fwritep[type][drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(type, drive, 0)) {
			fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[type][drive] = FALSE;
		fdc_medias[type][drive] = 1;
		return 1;
	}

	/*
	 * 2DDt@C
	 */
	if ((type == FDC_M) && (fsize == 655360)) {
		/* ^CvAݑݒ */
		fdc_ready[type][drive] = FDC_TYPE_2DD;
		fdc_fwritep[type][drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(type, drive, 0)) {
			fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[type][drive] = FALSE;
		fdc_medias[type][drive] = 1;
		return 1;
	}

	/*
	 * 2HDt@C
	 */
	if ((type == FDC_S) && (fsize == 1021696)) {
		/* ^CvAݑݒ */
		fdc_ready[type][drive] = FDC_TYPE_2HD;
		fdc_fwritep[type][drive] = writep;
		fdc_doorlock[drive] = (BYTE)0;
		fdc_sided[drive] = (BYTE)0x80;

		/* fBAݒ */
		if (!fdc_setmedia(type, drive, 0)) {
			fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[type][drive] = FALSE;
		fdc_medias[type][drive] = 1;
		return 1;
	}

	/*
	 * VFDt@C
	 */
	/* t@C */
	if ((type == FDC_M) && fdc_chkvfd(drive) != 0) {
		fdc_ready[type][drive] = FDC_TYPE_VFD;
		fdc_fwritep[type][drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(type, drive, 0)) {
			fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[type][drive] = FALSE;
		fdc_medias[type][drive] = 1;
		return 1;
	}

	/*
	 * D77t@C
	 */
	fdc_ready[type][drive] = FDC_TYPE_D77;
	fdc_fwritep[type][drive] = writep;
	if (type) {
		fdc_doorlock[drive] = (BYTE)0;
	}

	/* t@C */
	count = fdc_chkd77(type, drive);
	if (count == 0){
		fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
		return 0;
	}

	/* fBAݒ */
	flag = FALSE;
	for (i = 0; i < 16; i++) {
		if (fdc_setmedia(type, drive, i)) {
			flag = TRUE;
			break;
		}
	}
	if (!flag) {
		fdc_ready[type][drive] = FDC_TYPE_NOTREADY;
		return 0;
	}

	/* BꎞCWFNg */
	fdc_teject[type][drive] = FALSE;
	fdc_medias[type][drive] = (BYTE)count;
	return count;
}

/*-[ FDD Cxg ]-----------------------------------------------------*/

/*
 *	FDD [^[Cxg (WAIT)
 */
static BOOL FASTCALL fdd_motor_event(void)
{
	return TRUE;
}

/*-[ FDCR}h EFCg[h ]---------------------------------------*/

/*
 *	FDC DataRequestCxg (WAIT)
 */
static BOOL FASTCALL fdc_drq_event(int type)
{
	/* Xgf[^`FbN */
	if (fdc_drqirq[type] & 0x80) {
		if (type) {
			return sfdc_lost_event();
		}
		else {
			return mfdc_lost_event();
		}
	}

	/* f[^Ă̂DRQݒ肵Čp */
	if (fdc_dataptr) {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] |= 0x80;
		if (type && sfd_firq_mode) {
			/* f[^NGXg */
			sfdc_drq_sub();
		}
		return TRUE;
	}

	/* f[^ĂȂ̂RECORD NOT FOUNDŏI */
	fdc_status[type] |= FDC_ST_RECNFND;
	fdc_status[type] &= ~FDC_ST_BUSY;

	/* 荞ݔ */
	if (type && sfd_firq_mode) {
		/* G[荞 */
		schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

		/* Xe[^X */
		fdc_status[type] |= FDC_ST_BUSY;
		fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
	}
	else {
		if (type) {
			mainetc_sfdc();
		}
		else {
			mainetc_fdc();
		}
		fdc_drqirq[type] = 0x40;
		fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
	}

	return FALSE;
}

/*
 *	MFD
 *	FDC DataRequestCxg (WAIT)
 */
static BOOL FASTCALL mfdc_drq_event(void)
{
	return fdc_drq_event(FDC_M);
}

/*
 *	SFD
 *	FDC DataRequestCxg (WAIT)
 */
static BOOL FASTCALL sfdc_drq_event(void)
{
	return fdc_drq_event(FDC_S);
}

/*
 *	FDC V[NCxg (WAIT)
 */
static BOOL FASTCALL fdd_seek_event(int type)
{
	/* V[NI */
	if (fdc_seek_track[type] < 0) {
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= (FDC_ST_BUSY | FDC_ST_HEADENG);
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;

			/* BUSY/HEAD ENGAGEDtO𗎂Ƃ */
			fdc_status[type] &= ~(FDC_ST_BUSY | FDC_ST_HEADENG);
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
		}
		if (type) {
			schedule_delevent(EVENT_SFDD_SEEK);
		}
		else {
			schedule_delevent(EVENT_FDD_SEEK);
		}
	}
	else if (fdc_seek_track[type]-- == 0) {
		/* V[NI̓EFCg} */
		if (type) {
			schedule_setevent(EVENT_SFDD_SEEK, 20000, sfdd_seek_event);
		}
		else {
			schedule_setevent(EVENT_FDD_SEEK, 20000, mfdd_seek_event);
		}
	}
	else {
		/* TEh */
		if (fdc_sound) {
			wav_notify(SOUND_FDDSEEK);
		}
	}

	return TRUE;
}

/*
 *	MFD
 *	FDC V[NCxg (WAIT)
 */
static BOOL FASTCALL mfdd_seek_event(void)
{
	return fdd_seek_event(FDC_M);
}

/*
 *	SFD
 *	FDC V[NCxg (WAIT)
 */
static BOOL FASTCALL sfdd_seek_event(void)
{
	return fdd_seek_event(FDC_S);
}

/*
 *	FDC V[NCxgݒ (SOUND ON)
 */
static void FASTCALL fdc_setseekevent(int type, BYTE track)
{
	/* V[NgbNۑ */
	fdc_seek_track[type] = track;

	/* Cxgݒ */
	if (track > 0) {
		if (type) {
			schedule_setevent(EVENT_SFDD_SEEK, fdc_steprate[type][fdc_command[type] & 3], sfdd_seek_event);
		}
		else {
			schedule_setevent(EVENT_FDD_SEEK, fdc_steprate[type][fdc_command[type] & 3], mfdd_seek_event);
		}
	}
	else {
		/* wbhړȂĂ΂炭̊ԂBUSYԂɂ */
		if (type) {
			schedule_setevent(EVENT_SFDD_SEEK, 300, sfdd_seek_event);
		}
		else {
			schedule_setevent(EVENT_FDD_SEEK, 300, mfdd_seek_event);
		}
	}
}

/*-[ FDCR}h ]----------------------------------------------------------*/

/*
 *	[^[]ʒu擾
 */
static int FASTCALL fdd_get_currentpos(int type)
{
	event_t *ev;
	DWORD pos;
	DWORD exec;

	/* ݈ʒu */
	ev = schedule_getevent(fdd_event_motor[type]);
	pos = ev->reload - ev->current;

	/* sȂ玞Ԃ𑫂 */
	if (run_flag) {
		exec = (DWORD)maincpu.total;
		exec *= 10000;
		exec /= schedule_get_cycle();
		pos += exec;
		if (pos >= ev->reload) {
			pos -= ev->reload;
		}
	}

	return (int)pos;
}

/*
 *	ZN^ʒuvZ
 */
static void FASTCALL fdc_calc_pos(int type)
{
	int i;
	int max_track;
	int track;
	int gap0;
	int gap1;
	int gap2;
	int sync;
	int am;
	int gap3;
	int len;
	BYTE *p;

	ASSERT(fdc_ready[type][fdc_drvreg[type]] != FDC_TYPE_NOTREADY);

	/* ς݂ȂXLbv */
	if (fdi_init[type]) {
		return;
	}

	/*  */
	fdi_init[type] = TRUE;
	fdi_secs[type] = 0;

	/* GAP/SYNC */
	if (fdc_dden[type]) {
		gap0 = 80;
		gap1 = 50;
		gap2 = 22;
		sync = 12;
		am = 4;
	}
	else {
		gap0 = 40;
		gap1 = 26;
		gap2 = 11;
		sync = 6;
		am = 1;
	}

	/* fBXN^Cv`FbN */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) ||
		(fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD)) {
		track = fdc_track[type][fdc_drvreg[type]];
#if XM7_VER >= 3
		if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2D) && fdc_2ddmode) {
			/* fBA^Cv(2D) && hCu[h(2DD) */
			track >>= 1;
		}
		else if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) && !fdc_2ddmode) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#else
		if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#endif

#if XM7_VER >= 3
		if (fdc_2ddmode) {
			max_track = 80;
		}
		else {
			max_track = 40;
		}
#else
		max_track = 40;
#endif
		if (track >= max_track) {
			return;
		}

		/* ZN^=16CTCY=256oCgAGAP3=51oCg */
		fdi_secs[type] = 16;
		len = 256;
		gap3 = 51;

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[type][0] = (gap0 + sync + am + gap1) * 32;
		fdi_dtpos[type][0] = fdi_idpos[type][0] + (sync + am + 6 + gap2) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs[type]; i++) {
			fdi_idpos[type][i] = fdi_idpos[type][i - 1] + (sync + am + 6 + gap2 + sync + am + len + 2 + gap3) * 32;
			fdi_dtpos[type][i] = fdi_idpos[type][i] + (sync + am + 6 + gap2) * 32;
		}
		return;
	}

	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_2HD)) {
		track = fdc_track[type][fdc_drvreg[type]];

		max_track = 77;
		if (track >= max_track) {
			return;
		}

		if (fdc_dden[type]) {
			/* ZN^=26CTCY=256oCgAGAP3=54oCg */
			fdi_secs[type] = 26;
			len = 256;
			gap3 = 54;
		}
		else {
			/* ZN^=26CTCY=128oCgAGAP3=27oCg */
			fdi_secs[type] = 26;
			len = 128;
			gap3 = 27;
		}

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[type][0] = (gap0 + sync + am + gap1) * 32;
		fdi_dtpos[type][0] = fdi_idpos[type][0] + (sync + am + 6 + gap2) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs[type]; i++) {
			fdi_idpos[type][i] = fdi_idpos[type][i - 1] + (sync + am + 6 + gap2 + sync + am + len + 2 + gap3) * 32;
			fdi_dtpos[type][i] = fdi_idpos[type][i] + (sync + am + 6 + gap2) * 32;
		}
		return;
	}

	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_VFD) {
		track = fdc_track[type][fdc_drvreg[type]];

		/* ZN^ */
		fdi_secs[type] = fdc_header[type][fdc_drvreg[type]][track * 6 + 5];
		if (fdi_secs[type] > 96) {
			fdi_secs[type] = 96;
		}

		if (fdi_secs[type] == 0) {
			return;
		}

		/* TCY */
		len = fdc_header[type][fdc_drvreg[type]][track * 6 + 4];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}

		/* GAP3 */
		if (fdi_secs[type] <= 5) {
			gap3 = 0x74;
		}
		else {
			if (fdi_secs[type] <= 10) {
				gap3 = 0x54;
			}
			else {
				if (fdi_secs[type] <= 16) {
					gap3 = 0x33;
				}
				else {
					gap3 = 0x10;
				}
			}
		}

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[type][0] = (80 + 12 + 4 + 50) * 32;
		fdi_dtpos[type][0] = fdi_idpos[type][0] + (12 + 4 + 6 + 22) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs[type]; i++) {
			fdi_idpos[type][i] = fdi_idpos[type][i - 1] + (12 + 10 + 22 + 12 + 4 + len + 2 + gap3) * 32;
			fdi_dtpos[type][i] = fdi_idpos[type][i] + (12 + 4 + 6 + 22) * 32;
		}
		return;
	}

	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_D77) {
		/* ZN^ */
		p = fdc_buffer[type];
		fdi_secs[type] = p[0x0005];
		fdi_secs[type] *= 256;
		fdi_secs[type] |= p[0x0004];
		if (fdi_secs[type] > 96) {
			fdi_secs[type] = 96;
		}

		if (fdi_secs[type] == 0) {
			return;
		}

		/* GAP3 */
		if (type) {
			if (fdc_dden[type]) {
				if (fdi_secs[type] <= 8) {
					gap3 = 116;
				}
				else {
					if (fdi_secs[type] <= 15) {
						gap3 = 84;
					}
					else {
						if (fdi_secs[type] <= 26) {
							gap3 = 54;
						}
						else {
							gap3 = 16;
						}
					}
				}
			}
			else {
				if (fdi_secs[type] <= 8) {
					gap3 = 58;
				}
				else {
					if (fdi_secs[type] <= 15) {
						gap3 = 42;
					}
					else {
						if (fdi_secs[type] <= 26) {
							gap3 = 27;
						}
						else {
							gap3 = 8;
						}
					}
				}
			}
		}
		else {
			if (fdi_secs[type] <= 5) {
				gap3 = 0x74;
			}
			else {
				if (fdi_secs[type] <= 10) {
					gap3 = 0x54;
				}
				else {
					if (fdi_secs[type] <= 16) {
						gap3 = 0x33;
					}
					else {
						gap3 = 0x10;
					}
				}
			}
		}

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[type][0] = (gap0 + sync + am + gap1) * 32;
		fdi_dtpos[type][0] = fdi_idpos[type][0] + (sync + am + 6 + gap2) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs[type]; i++) {
			/* TCY */
			len = p[0x000f];
			len *= 256;
			len |= p[0x000e];
			fdi_idpos[type][i] = fdi_idpos[type][i - 1] + (sync + am + 6 + gap2 + sync + am + len + 2 + gap3) * 32;
			fdi_dtpos[type][i] = fdi_idpos[type][i] + (sync + am + 6 + gap2) * 32;
			p += len + 0x10;
		}
		return;
	}
}

/*
 *	FDC Xe[^X쐬
 */
static void FASTCALL fdc_make_stat(int type)
{
	int pos;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_status[type] |= FDC_ST_NOTREADY;
		return;
	}

	/* LȃhCu */
	if (fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_NOTREADY) {
		fdc_status[type] |= FDC_ST_NOTREADY;
	}
	else {
		fdc_status[type] &= (BYTE)(~FDC_ST_NOTREADY);
	}

	/* ꎞCWFNg */
	if (fdc_teject[type][fdc_drvreg[type]]) {
		fdc_status[type] |= FDC_ST_NOTREADY;
	}

	/* CgveNg(ReadnR}h0) */
	if ((fdc_cmdtype[type] == 2) || (fdc_cmdtype[type] == 4) ||
		(fdc_cmdtype[type] == 6)) {
		fdc_status[type] &= ~FDC_ST_WRITEP;
	}
	else {
		if (fdc_writep[type][fdc_drvreg[type]] &&
			!(fdc_status[type] & FDC_ST_NOTREADY)) {
			fdc_status[type] |= FDC_ST_WRITEP;
		}
		else {
			fdc_status[type] &= ~FDC_ST_WRITEP;
		}
	}

	/* TYPE I ̂ */
	if (fdc_cmdtype[type] != 1) {
		return;
	}

	/* TRACK00 */
	if (fdc_track[type][fdc_drvreg[type]] == 0) {
		fdc_status[type] |= FDC_ST_TRACK00;
	}
	else {
		fdc_status[type] &= ~FDC_ST_TRACK00;
	}

	/* index */
	if (!(fdc_status[type] & FDC_ST_NOTREADY)) {
		if (fdc_wait) {
			/* UNA */
			fdc_status[type] &= ~FDC_ST_INDEX;

			/* ݂̉]ʒu擾 */
			pos = fdd_get_currentpos(type);

			/* CfbNXM̎(5000us) */
			if (pos < 5000) {
				fdc_status[type] |= FDC_ST_INDEX;
			}
		} else {
			if (fdc_indexcnt[type] == 0) {
				fdc_status[type] |= FDC_ST_INDEX;
			}
			else {
				fdc_status[type] &= ~FDC_ST_INDEX;
			}
			fdc_next_index(type);
		}
	}
}

/*
 *	TYPE I
 *	RESTORE
 */
static void FASTCALL fdc_restore(int type)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype[type] = 1;
	fdc_status[type] = 0;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_make_stat(type);
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[type][fdc_drvreg[type]];

	/* gbNrAXgAAǂݍ */
	if (fdc_track[type][fdc_drvreg[type]] != 0) {
		fdc_track[type][fdc_drvreg[type]] = 0;
		fdc_readbuf(type, fdc_drvreg[type]);
	}
	fdc_seekvct[type] = TRUE;

	/* Abvf[g(f[^WX^0ɂ) */
	fdc_trkreg[type] = 0;
	fdc_datareg[type] = 0;

	/* FDC荞 */
	fdc_drqirq[type] = 0x00;
	if (fdc_wait) {
		fdc_setseekevent(type, prevtrk);
	}
	else if (type) {
		if (sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (!sfd_irq_mask) {
			mainetc_sfdc();
			fdc_drqirq[type] |= (BYTE)0x40;
		}
	}
	else if (!mfd_irq_mask) {
		mainetc_fdc();
		fdc_drqirq[type] |= (BYTE)0x40;
	}

	/* Xe[^X */
	fdc_status[type] = FDC_ST_BUSY;
	if (fdc_command[type] & 0x08) {
		fdc_status[type] |= FDC_ST_HEADENG;
	}
	fdc_make_stat(type);

	/* ANZX(SEEK) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	SEEK
 */
static void FASTCALL fdc_seek(int type)
{
	BYTE target;
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype[type] = 1;
	fdc_status[type] = 0;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_make_stat(type);
		return;
	}

	/* Ƀxt@C */
	if (fdc_command[type] & 0x04) {
		if (fdc_trkreg[type] != fdc_track[type][fdc_drvreg[type]]) {
			fdc_status[type] |= FDC_ST_SEEKERR;
			/* C(ՂĂ΍) */
			fdc_trkreg[type] = fdc_track[type][fdc_drvreg[type]];
		}
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[type][fdc_drvreg[type]];

	/* ΃V[N */
	target = (BYTE)(fdc_track[type][fdc_drvreg[type]] + fdc_datareg[type] - fdc_trkreg[type]);
	if (fdc_datareg[type] > fdc_trkreg[type]) {
		fdc_seekvct[type] = FALSE;
		if (type) {
			if (target > 81) {
				target = 81;
			}
		}
		else {
			if (target > 41) {
				target = 41;
			}
		}
	}
	else {
		fdc_seekvct[type] = TRUE;
		/* gbNԍ̃I[o[t[h~ (OS-9΍) */
		if (fdc_track[type][fdc_drvreg[type]] < (fdc_trkreg[type] - fdc_datareg[type])) {
			target = 0;
		}
	}

	/* gbNrAV[NAǂݍ */
	if (fdc_track[type][fdc_drvreg[type]] != target) {
		fdc_track[type][fdc_drvreg[type]] = target;
		fdc_readbuf(type, fdc_drvreg[type]);
	}

	/* Abvf[g */
	fdc_trkreg[type] = fdc_datareg[type];	/* f[^WX^̒lf͗l (OS-9΍) */

	/* FDC荞 */
	fdc_drqirq[type] = 0x00;
	if (fdc_wait) {
		fdc_setseekevent(type, (BYTE)abs(target - prevtrk));
	}
	else if (type) {
		if (sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (!sfd_irq_mask) {
			mainetc_sfdc();
			fdc_drqirq[type] |= (BYTE)0x40;
		}
	}
	else if (!mfd_irq_mask) {
		mainetc_fdc();
		fdc_drqirq[type] |= (BYTE)0x40;
	}

	/* Xe[^X */
	fdc_status[type] |= FDC_ST_BUSY;
	if (fdc_command[type] & 0x08) {
		fdc_status[type] |= FDC_ST_HEADENG;
	}
	fdc_make_stat(type);

	/* ANZX(SEEK) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP IN
 */
static void FASTCALL fdc_step_in(int type)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype[type] = 1;
	fdc_status[type] = 0;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_make_stat(type);
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[type][fdc_drvreg[type]];

	/* Ƀxt@C */
	if (fdc_command[type] & 0x04) {
		if (fdc_trkreg[type] != fdc_track[type][fdc_drvreg[type]]) {
			fdc_status[type] |= FDC_ST_SEEKERR;
		}
	}

	/* XebvAǂݍ */
	if (type) {
		if (fdc_track[type][fdc_drvreg[type]] < 81) {
			fdc_track[type][fdc_drvreg[type]]++;
			fdc_readbuf(type, fdc_drvreg[type]);
		}
	}
	else {
		if (fdc_track[type][fdc_drvreg[type]] < 41) {
			fdc_track[type][fdc_drvreg[type]]++;
			fdc_readbuf(type, fdc_drvreg[type]);
		}
	}
	fdc_seekvct[type] = FALSE;

	/* Abvf[g */
	if (fdc_command[type] & 0x10) {
		if (fdc_trkreg[type] < 255) {
			fdc_trkreg[type]++;	/* gbNWX^̂̍XV݂̂Ǝv */
		}
	}

	/* FDC荞 */
	fdc_drqirq[type] = 0x00;
	if (fdc_wait && (prevtrk != fdc_track[type][fdc_drvreg[type]])) {
		fdc_setseekevent(type, 1);
	}
	else if (type) {
		if (sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (!sfd_irq_mask) {
			mainetc_sfdc();
			fdc_drqirq[type] |= (BYTE)0x40;
		}
	}
	else if (!mfd_irq_mask) {
		mainetc_fdc();
		fdc_drqirq[type] |= (BYTE)0x40;
	}

	/* Xe[^X */
	fdc_status[type] |= FDC_ST_BUSY;
	if (fdc_command[type] & 0x08) {
		fdc_status[type] |= FDC_ST_HEADENG;
	}
	fdc_make_stat(type);

	/* ANZX(SEEK) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP OUT
 */
static void FASTCALL fdc_step_out(int type)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype[type] = 1;
	fdc_status[type] = 0;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_make_stat(type);
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[type][fdc_drvreg[type]];

	/* Ƀxt@C */
	if (fdc_command[type] & 0x04) {
		if (fdc_trkreg[type] != fdc_track[type][fdc_drvreg[type]]) {
			fdc_status[type] |= FDC_ST_SEEKERR;
		}
	}

	/* XebvAǂݍ */
	if (fdc_track[type][fdc_drvreg[type]] != 0) {
		fdc_track[type][fdc_drvreg[type]]--;
		fdc_readbuf(type, fdc_drvreg[type]);
	}
	fdc_seekvct[type] = TRUE;

	/* Abvf[g */
	if (fdc_command[type] & 0x10) {
		if (fdc_trkreg[type] > 0) {
			fdc_trkreg[type]--;	/* gbNWX^̂̍XV݂̂Ǝv */
		}
	}

	/* FDC荞 */
	fdc_drqirq[type] = 0x00;
	if (fdc_wait && (prevtrk != fdc_track[type][fdc_drvreg[type]])) {
		fdc_setseekevent(type, 1);
	}
	else if (type) {
		if (sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (!sfd_irq_mask) {
			mainetc_sfdc();
			fdc_drqirq[type] |= (BYTE)0x40;
		}
	}
	else if (!mfd_irq_mask) {
		mainetc_fdc();
		fdc_drqirq[type] |= (BYTE)0x40;
	}

	/* Xe[^X */
	fdc_status[type] |= FDC_ST_BUSY;
	if (fdc_command[type] & 0x08) {
		fdc_status[type] |= FDC_ST_HEADENG;
	}
	fdc_make_stat(type);

	/* ANZX(SEEK) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP
 */
static void FASTCALL fdc_step(int type)
{
	if (fdc_seekvct[type]) {
		fdc_step_out(type);
	}
	else {
		fdc_step_in(type);
	}
}

/*
 *	TYPE II, III
 *	READ/WRITE Tu
 */
static BOOL FASTCALL fdc_rw_sub(int type)
{
	fdc_status[type] = 0;

	/* hCu`FbN */
	if (fdc_drvreg[type] >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
		fdc_make_stat(type);
		return FALSE;
	}

	/* NOT READY`FbN */
	if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_NOTREADY) || fdc_teject[type][fdc_drvreg[type]]) {
		fdc_make_stat(type);
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
		}
		return FALSE;
	}

	return TRUE;
}

/*
 *	TYPE II
 *	READ DATA
 */
static void FASTCALL fdc_read_data(int type)
{
	BYTE stat;
	int secidx;
	int time;

	/* TYPE II, Read */
	fdc_cmdtype[type] = 2;

	/* {`FbN */
	if (!fdc_rw_sub(type)) {
		return;
	}

	if (fdc_wait) {
		/* |WVݒ */
		fdc_calc_pos(type);
	}

	/* ZN^ */
	if (fdc_command[type] & 0x02) {
		stat = fdc_readsec(
				type, fdc_trkreg[type], fdc_secreg[type],
				(BYTE)((fdc_command[type] & 0x08) >> 3),
				TRUE, &secidx, fdd_get_currentpos(type));
	}
	else {
		stat = fdc_readsec(
				type, fdc_trkreg[type], fdc_secreg[type], fdc_sidereg[type],
				FALSE, &secidx, fdd_get_currentpos(type));
	}

	/* ɃXe[^Xݒ肷 */
	fdc_status[type] = stat;

	/* RECORD NOT FOUND ? */
	if (fdc_status[type] & FDC_ST_RECNFND) {
		/* ]f[^ */
		fdc_dataptr[type] = NULL;
		fdc_datareg[type] = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status[type] |= FDC_ST_BUSY;

			/* tONAăCxgōĐݒ肷 */
			fdc_status[type] &= ~FDC_ST_RECNFND;
			if (type) {
				schedule_setevent(EVENT_SFDC_L, fdd_rotation_time[type] * 6,
									sfdc_drq_event);
			}
			else {
				schedule_setevent(EVENT_FDC_L, fdd_rotation_time[type] * 6,
									mfdc_drq_event);
			}

			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
			return;
		}
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
		}
		return;
	}

	/* ŏ̃f[^ݒ */
	fdc_status[type] |= FDC_ST_BUSY;
	fdc_datareg[type] = fdc_dataptr[type][0];
	if (fdc_wait) {
		/* YZN^܂ł̎ */
		time = 0;
		if (secidx < fdi_secs[type]) {
			time = fdi_dtpos[type][secidx] - fdd_get_currentpos(type);
		}
		while (time <= 0) {
			time += fdd_rotation_time[type];
		}

		/* ZN^܂ł̗\莞Ԃݒ */
		if (type) {
			schedule_setevent(EVENT_SFDC_L, (DWORD)time, sfdc_drq_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, (DWORD)time, mfdc_drq_event);
		}
		fdc_drqirq[type] = 0x00;
	}
	else {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] = 0x80;
	}
	if (type && sfd_firq_mode) {
		/* f[^NGXg */
		sfdc_drq_sub();
	}

	/* ANZX(READ) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
	dmac_start();
	if (!fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, 30 * 1000, sfdc_lost_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, 30 * 1000, mfdc_lost_event);
		}
	}
}

/*
 *	TYPE II
 *	WRITE DATA
 */
static void FASTCALL fdc_write_data(int type)
{
	BYTE stat;
	int secidx;
	int time;

	/* TYPE II, Write */
	fdc_cmdtype[type] = 3;

	/* {`FbN */
	if (!fdc_rw_sub(type)) {
		return;
	}

	/* WRITE PROTECT`FbN */
	if (fdc_writep[type][fdc_drvreg[type]] != 0) {
		fdc_make_stat(type);
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
		}
		return;
	}

	if (fdc_wait) {
		/* |WVݒ */
		fdc_calc_pos(type);
	}

	/* ZN^ */
	if (fdc_command[type] & 0x02) {
		stat = fdc_readsec(
			type, fdc_trkreg[type], fdc_secreg[type],
			(BYTE)((fdc_command[type] & 0x08) >> 3),
			TRUE, &secidx, fdd_get_currentpos(type));
	}
	else {
		stat = fdc_readsec(
			type, fdc_trkreg[type], fdc_secreg[type], fdc_sidereg[type],
			FALSE, &secidx, fdd_get_currentpos(type));
	}

	/* ɃXe[^Xݒ肷 */
	fdc_status[type] = stat;

	/* RECORD NOT FOUND ? */
	if (fdc_status[type] & FDC_ST_RECNFND) {
		/* ]f[^ */
		fdc_dataptr[type] = NULL;
		fdc_datareg[type] = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status[type] |= FDC_ST_BUSY;

			/* tONAăCxgŐݒ肷 */
			fdc_status[type] &= ~FDC_ST_RECNFND;
			if (type) {
				schedule_setevent(EVENT_SFDC_L, fdd_rotation_time[type] * 6, sfdc_drq_event);
			}
			else {
				schedule_setevent(EVENT_FDC_L, fdd_rotation_time[type] * 6, mfdc_drq_event);
			}
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_WRITE;
			return;
		}
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
		}
		return;
	}

	/* DRQ */
	fdc_status[type] |= FDC_ST_BUSY;
	if (fdc_wait) {
		/* YZN^܂ł̎ */
		time = 0;
		if (secidx < fdi_secs[type]) {
			time = fdi_dtpos[type][secidx] - fdd_get_currentpos(type);
		}
		while (time <= 0) {
			time += fdd_rotation_time[type];
		}

		/* ZN^܂ł̗\莞Ԃݒ */
		if (type) {
			schedule_setevent(EVENT_SFDC_L, (DWORD)time, sfdc_drq_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, (DWORD)time, mfdc_drq_event);
		}
		fdc_drqirq[type] = 0x00;
	}
	else {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] = 0x80;
	}
	if (type && sfd_firq_mode) {
		/* f[^NGXg */
		sfdc_drq_sub();
	}

	/* ANZX(WRITE) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_WRITE;

	/* DMA]JnAXgf[^Cxgݒ */
	dmac_start();
	if (!fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, 30 * 1000, sfdc_lost_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, 30 * 1000, mfdc_lost_event);
		}
	}
}

/*
 *	TYPE III
 *	READ ADDRESS
 */
static void FASTCALL fdc_read_addr(int type)
{
	int i;
	int cur;
	int idx;
	int time;

	/* TYPE III, Read Address */
	fdc_cmdtype[type] = 4;

	/* {`FbN */
	if (!fdc_rw_sub(type)) {
		return;
	}

	if (fdc_wait) {
		/* C[W̃ZN^ʒuvZ */
		fdc_calc_pos(type);

		/* ݂̉]ʒu */
		cur = fdd_get_currentpos(type);

		/* ߂̃ZN^T */
		idx = -1;
		time = 0;
		if (fdi_secs[type] > 0) {
			for (i = 0; i < fdi_secs[type]; i++) {
				if (fdi_idpos[type][i] < fdd_rotation_time[type] && fdi_idpos[type][i] > cur) {
					break;
				}
			}

			/* ZN^m */
			idx = i;
			if (idx == fdi_secs[type]) {
				idx = 0;
			}

			/* ݈ʒu̎ */
			time = fdi_idpos[type][idx] - cur;
			while (time <= 0) {
				time += fdd_rotation_time[type];
			}
		}
	} else {
		/* gbN擪̑΃ZN^ԍ擾 */
		idx = fdc_next_index(type);
	}

	/* AtH[}bg`FbN */
	if (idx == -1) {
		/* ]f[^ */
		fdc_dataptr[type] = NULL;
		fdc_datareg[type] = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status[type] |= FDC_ST_BUSY;

			/* tONAăCxgŐݒ肷 */
			fdc_status[type] &= ~FDC_ST_RECNFND;
			if (type) {
				schedule_setevent(EVENT_SFDC_L, fdd_rotation_time[type] * 6, sfdc_drq_event);
			}
			else {
				schedule_setevent(EVENT_FDC_L, fdd_rotation_time[type] * 6, mfdc_drq_event);
			}
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
			return;
		}
		/* RECORD NOT FOUND */
		fdc_status[type] |= FDC_ST_RECNFND;

		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
		}
		return;
	}

	/* f[^쐬 */
	fdc_makeaddr(type, idx);

	/* ŏ̃f[^ݒ */
	fdc_status[type] |= FDC_ST_BUSY;
	fdc_datareg[type] = fdc_dataptr[type][0];
	if (fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, (DWORD)time, sfdc_drq_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, (DWORD)time, mfdc_drq_event);
		}
		fdc_drqirq[type] = 0x00;
	}
	else {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] = 0x80;
	}
	if (type && sfd_firq_mode) {
		/* f[^NGXg */
		sfdc_drq_sub();
	}

	/* ANZX(READ) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
	dmac_start();
	if (!fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, 10 * 1000, sfdc_lost_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, 10 * 1000, mfdc_lost_event);
		}
	}
}

/*
 *	TYPE III
 *	READ TRACK
 */
static void FASTCALL fdc_read_track(int type)
{
	int time;

	/* TYPE III, Read Track */
	fdc_cmdtype[type] = 6;

	/* {`FbN */
	if (!fdc_rw_sub(type)) {
		return;
	}

	/* f[^쐬 */
	fdc_make_track(type);

	/* ŏ̃f[^ݒ */
	fdc_status[type] |= FDC_ST_BUSY;
	fdc_datareg[type] = fdc_dataptr[type][0];
	if (fdc_wait) {
		time = fdd_rotation_time[type] - fdd_get_currentpos(type);
		while (time <= 0) {
			time += fdd_rotation_time[type];
		}
		if (type) {
			schedule_setevent(EVENT_SFDC_L, (DWORD)time, sfdc_drq_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, (DWORD)time, mfdc_drq_event);
		}
		fdc_drqirq[type] = 0x00;
	}
	else {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] = 0x80;
	}
	if (type && sfd_firq_mode) {
		/* f[^NGXg */
		sfdc_drq_sub();
	}

	/* ANZX(READ) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
	dmac_start();
	if (!fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, 150 * 1000, sfdc_lost_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, 150 * 1000, mfdc_lost_event);
		}
	}
}

/*
 *	TYPE III
 *	WRITE TRACK
 */
static void FASTCALL fdc_write_track(int type)
{
	int time;

	/* TYPE III, Write */
	fdc_cmdtype[type] = 5;

	/* {`FbN */
	if (!fdc_rw_sub(type)) {
		return;
	}

	/* WRITE PROTECT`FbN */
	if (fdc_writep[type][fdc_drvreg[type]] != 0) {
		fdc_make_stat(type);
		/* FDC荞 */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else {
			if (type) {
				mainetc_sfdc();
			}
			else {
				mainetc_fdc();
			}
			fdc_drqirq[type] = 0x40;
		}
		return;
	}

	/* DRQ */
	fdc_status[type] |= FDC_ST_BUSY;
	if (fdc_wait) {
		time = fdd_rotation_time[type] - fdd_get_currentpos(type);
		while (time <= 0) {
			time += fdd_rotation_time[type];
		}
		if (type) {
			schedule_setevent(EVENT_SFDC_L, (DWORD)time, sfdc_drq_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, (DWORD)time, mfdc_drq_event);
		}
		fdc_drqirq[type] = 0x00;
	}
	else {
		fdc_status[type] |= FDC_ST_DRQ;
		fdc_drqirq[type] = 0x80;
	}
	if (type && sfd_firq_mode) {
		/* f[^NGXg */
		sfdc_drq_sub();
	}

	fdc_dataptr[type] = fdc_buffer[type];
	if (type) {
		if (fdc_dden[type]) {
			fdc_totalcnt[type] = 0x2A00;
		}
		else {
			fdc_totalcnt[type] = 0x1500;
		}
	}
	else {
		fdc_totalcnt[type] = 0x1800;
	}
	fdc_onetime[type] = 32;
	fdc_nowcnt[type] = 0;

	/* ANZX(WRITE) */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_WRITE;

	/* DMA]JnAXgf[^Cxgݒ */
	dmac_start();
	if (!fdc_wait) {
		if (type) {
			schedule_setevent(EVENT_SFDC_L, 150 * 1000, sfdc_lost_event);
		}
		else {
			schedule_setevent(EVENT_FDC_L, 150 * 1000, mfdc_lost_event);
		}
	}
}

/*
 *	TYPE IV
 *	FORCE INTERRUPT
 */
static void FASTCALL fdc_force_intr(int type)
{
	/* WRITE TRACKI */
	if (fdc_cmdtype[type] == 5) {
		if (!fdc_writetrk(type)) {
			fdc_status[type] |= FDC_ST_WRITEFAULT;
		}
	}

	/* ANZX~ */
	if (!fdc_wait) {
		if (type) {
			schedule_delevent(EVENT_SFDC_L);
		}
		else {
			schedule_delevent(EVENT_FDC_L);
		}
	}

	/* obt@C */
	fdc_readbuf(type, fdc_drvreg[type]);
	fdc_dataptr[type] = NULL;

	/* Xe[^X */
	switch (fdc_access[type][fdc_drvreg[type]]) {
		case FDC_ACCESS_READ:
		case FDC_ACCESS_WRITE:
			/* s̃R}hƓXe[^X */
			fdc_status[type] &= ~FDC_ST_BUSY;
			break;

		case FDC_ACCESS_SEEK:
		case FDC_ACCESS_READY:
			fdc_status[type] = 0;
			fdc_indexcnt[type] = 0;
			fdc_make_stat(type);
			break;
	}

	/* ŃANZXREADY */
	fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;

	/* ꂩĂ΁AIRQ(ƂĂ͕s\) */
	if (fdc_command[type] & 0x0f) {
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (type) {
			mainetc_sfdc();
		}
		else {
			mainetc_fdc();
		}
		fdc_drqirq[type] = 0x40;
	}
	else {
		fdc_drqirq[type] = 0x00;
	}
}

/*
 *	}`ZN^
 *	Cxg
 */
static BOOL FASTCALL fdc_multi_event(int type)
{
	if (fdc_drqirq[type] & 0x10) {
		fdc_drqirq[type] = 0x08;
		/* ZN^WX^UpAZN^ */
		fdc_secreg[type]++;
		if (type) {
			schedule_setevent(EVENT_SFDC_M, 30, sfdc_multi_event);
		}
		else {
			schedule_setevent(EVENT_FDC_M, 30, mfdc_multi_event);
		}
		return TRUE;
	}

	if (fdc_drqirq[type] & 0x08) {
		fdc_drqirq[type] = 0x00;
		/* ZN^ */
		if (fdc_cmdtype[type] == 2) {
			fdc_read_data(type);
			fdc_drqirq[type] |= 0x20;
		}
		if (fdc_cmdtype[type] == 3) {
			fdc_write_data(type);
			fdc_drqirq[type] |= 0x20;
		}
	}

	/* Cxg폜ďI */
	if (type) {
		schedule_delevent(EVENT_SFDC_M);
	}
	else {
		schedule_delevent(EVENT_FDC_M);
	}
	return TRUE;
}

/*
 *	MFD
 *	}`ZN^
 *	Cxg
 */
static BOOL FASTCALL mfdc_multi_event(void)
{
	return fdc_multi_event(FDC_M);
}

/*
 *	SFD
 *	}`ZN^
 *	Cxg
 */
static BOOL FASTCALL sfdc_multi_event(void)
{
	return fdc_multi_event(FDC_S);
}

/*
 *	Xgf[^
 *	Cxg
 */
static BOOL FASTCALL fdc_lost_event(int type)
{
	if (fdc_dataptr[type] && (fdc_drqirq[type] & 0x80)) {
		/* R}hł؂AXgf[^ */
		fdc_dataptr[type] = NULL;
		fdc_status[type] |= FDC_ST_LOSTDATA;
		fdc_status[type] &= ~FDC_ST_DRQ;
		fdc_status[type] &= ~FDC_ST_BUSY;
		fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;

		/* 荞ݔ */
		if (type && sfd_firq_mode) {
			/* G[荞 */
			schedule_setevent(EVENT_FDC_NMI, fdc_irq_time, sfdc_nmi_event);

			/* Xe[^X */
			fdc_status[type] |= FDC_ST_BUSY;
			fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READ;
		}
		else if (type) {
			mainetc_sfdc();
		}
		else {
			mainetc_fdc();
		}
		fdc_drqirq[type] = 0x40;
	}

	/* Cxg폜ďI */
	if (type) {
		schedule_delevent(EVENT_SFDC_L);
	}
	else {
		schedule_delevent(EVENT_FDC_L);
	}
	return TRUE;
}

/*
 *	MFD
 *	Xgf[^
 *	Cxg
 */
static BOOL FASTCALL mfdc_lost_event(void)
{
	return fdc_lost_event(FDC_M);
}

/*
 *	SFD
 *	Xgf[^
 *	Cxg
 */
static BOOL FASTCALL sfdc_lost_event(void)
{
	return fdc_lost_event(FDC_S);
}

/*
 *	SFD
 *	f[^NGXg^LZ
 *	Tu[`
 */
static void FASTCALL sfdc_drq_sub(void)
{
	if (fdc_drqirq[FDC_S] & 0x80) {
		/* f[^]̊荞ݔ */
		sfd_drq_flag = TRUE;
		maincpu_firq();

		/* 1MB FDCłDRQ/IRQ`FbNȂ̂łŃtO𗧂Ă */
		fdc_drqirq[FDC_S] |= (BYTE)0x20;
	}
	else {
		/* f[^]̊荞݉ */
		sfd_drq_flag = FALSE;
		maincpu_firq();
	}
}

/*
 *	SFD
 *	f[^]IEG[I
 *	Cxg
 */
static BOOL FASTCALL sfdc_nmi_event(void)
{
	/* f[^]NMI */
	maincpu_nmi();

	/* IRQ On */
	fdc_drqirq[FDC_S] = 0x40;

	/* BUSYtO𗎂Ƃ */
	fdc_status[FDC_S] &= ~FDC_ST_BUSY;
	fdc_access[FDC_S][fdc_drvreg[FDC_S]] = FDC_ACCESS_READY;

	/* Cxg폜ďI */
	schedule_delevent(EVENT_FDC_NMI);
	return TRUE;
}


/*
 *	R}h
 */
static void FASTCALL fdc_process_cmd(int type)
{
	BYTE high;

	high = (BYTE)(fdc_command[type] >> 4);

	/* EFCgtOݒ */
	fdc_wait = fdc_waitmode;

	/* f[^]sĂ΁A~߂ */
	fdc_dataptr[type] = NULL;

	/*  */
	switch (high) {
		/* restore */
		case 0x00:
			fdc_restore(type);
			break;
		/* seek */
		case 0x01:
			fdc_seek(type);
			break;
		/* step */
		case 0x02:
		case 0x03:
			fdc_step(type);
			break;
		/* step in */
		case 0x04:
		case 0x05:
			fdc_step_in(type);
			break;
		/* step out */
		case 0x06:
		case 0x07:
			fdc_step_out(type);
			break;
		/* read data */
		case 0x08:
		case 0x09:
			fdc_read_data(type);
			break;
		/* write data */
		case 0x0a:
		case 0x0b:
			fdc_write_data(type);
			break;
		/* read address */
		case 0x0c:
			fdc_read_addr(type);
			break;
		/* force interrupt */
		case 0x0d:
			fdc_force_intr(type);
			break;
		/* read track */
		case 0x0e:
			fdc_read_track(type);
			break;
		/* write track */
		case 0x0f:
			fdc_write_track(type);
			break;
		/* ȊO */
		default:
			ASSERT(FALSE);
			break;
	}
}

#if 1	/* FM-8 8" IFDCn */
/*
 *	IFDCWX^f[^R}h
 */
static void FASTCALL ifdc_data_cmd(void)
{
	/* EFCgtOݒ */
	fdc_wait = fdc_waitmode;

	/* f[^]sĂ΁A~߂ */
	fdc_dataptr[FDC_S] = NULL;

	/* p[^ݒ */
	if (fdc_drvreg[FDC_S] != ifdc_drvreg) {
		fdc_drvreg[FDC_S] = (BYTE)ifdc_drvreg;
		fdc_readbuf(FDC_S, fdc_drvreg[FDC_S]);
	}

	/*  */
	switch (ifdc_cmdreg) {
		/* IFDC INITIALIZE */
		case 0x00:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/IFDC INITIALIZE", ifdc_cmdreg);
#endif
			break;
		/* MAKE SOFT SECTOR FORMAT */
		case 0x01:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/MAKE SOFT SECTOR FORMAT", ifdc_cmdreg);
#endif
			break;
		/* SEEK TRACK */
		case 0x02:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/SEEK TRACK", ifdc_cmdreg);
#endif
			fdc_trkreg[FDC_S] = fdc_track[FDC_S][fdc_drvreg[FDC_S]];
			fdc_datareg[FDC_S] = ifdc_trkreg;
			if (ifdc_trkreg) {
				fdc_command[FDC_S] = 0x18;
			}
			else {
				fdc_command[FDC_S] = 0x08;
			}
			fdc_process_cmd(FDC_S);
			break;
		/* READ SECTOR (FLOPPY --> IFDC BUFFER) */
		case 0x03:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/READ SECTOR  (FLOPPY --> IFDC BUFFER)", ifdc_cmdreg);
#endif
			break;
		/* WRITE SECTOR(FLOPPY <-- IFDC BUFFER) */
		case 0x04:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/WRITE SECTOR (FLOPPY <-- IFDC BUFFER)", ifdc_cmdreg);
#endif
			break;
		/* DATA GET(PROGRAM TRANSFER) FM8 <-- IFDC */
		case 0x05:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/DATA GET (PROGRAM TRANSFER) FM8 <-- IFDC", ifdc_cmdreg);
#endif
			break;
		/* DATA PUT(PROGRAM TRANSFER) FM8 --> IFDC */
		case 0x06:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/DATA PUT (PROGRAM TRANSFER) FM8 --> IFDC", ifdc_cmdreg);
#endif
			break;
		/* DATA GET(DMA TRANSFER)     FM8 <-- IFDC */
		case 0x07:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/DATA GET (DMA TRANSFER)     FM8 <-- IFDC", ifdc_cmdreg);
#endif
			break;
		/* DATA PUT(DMA TRANSFER)     FM8 --> IFDC */
		case 0x08:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/DATA PUT (DMA TRANSFER)     FM8 --> IFDC", ifdc_cmdreg);
#endif
			break;
		/* DIRECT READ SECTOR FM8 <-- FLOPPY */
		case 0x09:	/* (PROGRAM) */
		case 0x0b:	/* (DMA) */
#ifdef FDC_LOG
			if (ifdc_cmdreg == 0x09) {
				LOG1(Normal, "rdr:%02x/DIRECT READ SECTOR (PROGRAM) FM8 <-- FLOPPY", ifdc_cmdreg);
			}
			else {
				LOG1(Normal, "rdr:%02x/DIRECT READ SECTOR (DMA)     FM8 <-- FLOPPY", ifdc_cmdreg);
			}
#endif
			fdc_trkreg[FDC_S] = ifdc_trkreg;
			fdc_secreg[FDC_S] = ifdc_ssreg;
			fdc_dden[FDC_S] = ifdc_dnsreg;
			if (ifdc_sidreg != fdc_sidereg[FDC_S]) {
				fdc_sidereg[FDC_S] = ifdc_sidreg;
				fdc_readbuf(FDC_S, fdc_drvreg[FDC_S]);
			}
			if (ifdc_ssreg == ifdc_sereg) {
				fdc_command[FDC_S] = 0x80;
			}
			else {
				fdc_command[FDC_S] = 0x90;
			}
			fdc_process_cmd(FDC_S);
			break;
		/* DIRECT WRITE SECTOR FM8 --> FLOPPY */
		case 0x0a:	/* (PROGRAM) */
		case 0x0c:	/* (DMA) */
#ifdef FDC_LOG
			if (ifdc_cmdreg == 0x0a) {
				LOG1(Normal, "rdr:%02x/DIRECT WRITE SECTOR(PROGRAM) FM8 --> FLOPPY", ifdc_cmdreg);
			}
			else {
				LOG1(Normal, "rdr:%02x/DIRECT WRITE SECTOR(DMA)     FM8 --> FLOPPY", ifdc_cmdreg);
			}
#endif
			fdc_trkreg[FDC_S] = ifdc_trkreg;
			fdc_secreg[FDC_S] = ifdc_ssreg;
			fdc_dden[FDC_S] = ifdc_dnsreg;
			if (ifdc_sidreg != fdc_sidereg[FDC_S]) {
				fdc_sidereg[FDC_S] = ifdc_sidreg;
				fdc_readbuf(FDC_S, fdc_drvreg[FDC_S]);
			}
			if (ifdc_ssreg == ifdc_sereg) {
				fdc_command[FDC_S] = 0xa0;
			}
			else {
				fdc_command[FDC_S] = 0xb0;
			}
			fdc_process_cmd(FDC_S);
			break;
		/* FLOPPY DISK SEEK TRACK OPERATION */
		case 0x0d:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/FLOPPY DISK SEEK TRACK OPERATION", ifdc_cmdreg);
#endif
			break;
		/* OTHERS */
		default:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x/OTHERS", ifdc_cmdreg);
#endif
			ifdc_ssr |= IFDC_ST_CMD_ERR;
			break;
	}
}

/*
 *	IFDCWX^U[uR}h
 */
static void FASTCALL ifdc_reserve_cmd(void)
{
	/* EFCgtOݒ */
	fdc_wait = fdc_waitmode;

	/* xBUSYɂ */
	fdc_status[FDC_S] |= FDC_ST_BUSY;
	ifdc_ssr = 0;

	/*  */
	switch (ifdc_rrr) {
		/* CMDREG */
		case 0x00:
			ifdc_cmdreg = ifdc_rdr;
			ifdc_data_cmd();
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* TRKREG */
		case 0x01:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->TRKREG", ifdc_rdr);
#endif
			ifdc_trkreg = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* SSREG */
		case 0x02:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->SSREG", ifdc_rdr);
#endif
			ifdc_ssreg = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* SEREG */
		case 0x03:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->SEREG", ifdc_rdr);
#endif
			ifdc_sereg = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* SIDREG */
		case 0x04:
#ifdef FDC_LOG
			LOG1(Normal, "rrr:%02x->SIDREG", ifdc_rdr);
#endif
			ifdc_sidreg = (ifdc_rdr & 0x01);
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* DNSREG */
		case 0x05:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->DNSREG", ifdc_rdr);
#endif
			ifdc_dnsreg = (BYTE)((ifdc_rdr & 0x01) << 6);
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* DRVREG */
		case 0x06:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->DRVREG", ifdc_rdr);
#endif
			ifdc_drvreg = (BYTE)(ifdc_rdr & 0x03);
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* DLCREG */
		case 0x07:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->DLCREG", ifdc_rdr);
#endif
			ifdc_dlcreg = ifdc_rdr;
			/* hAbN */
			if (fdc_doorlock[ifdc_dlcreg & 0x03] != (BYTE)((ifdc_dlcreg >> 3) & 0x10)) {
				fdc_doorlock[ifdc_dlcreg & 0x03] = (BYTE)((ifdc_dlcreg >> 3) & 0x10);
			}
			/* FDCNOT READY̎̓bNȂ */
			if ((fdc_ready[FDC_S][ifdc_dlcreg & 0x03] == FDC_TYPE_NOTREADY) ||
				fdc_teject[FDC_S][ifdc_dlcreg & 0x03]) {
				fdc_doorlock[ifdc_dlcreg & 0x03] = (BYTE)0;
			}
			/* B7`4:ЖʁEʃtOA󗼖ʃC[WT|[gĂȂ̂ŌŒl */
			ifdc_fsr = (BYTE)(0xF0 | (fdc_status[FDC_S] & 0x0F));
			break;
		/* ERESET */
		case 0x08:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->ERESET", ifdc_rdr);
#endif
			ifdc_fsr = (BYTE)fdc_status[FDC_S] = 0;
			break;
		/* SFORMT */
		case 0x09:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->SFORMT", ifdc_rdr);
#endif
			ifdc_sformt = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* SECDAT */
		case 0x0a:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->SECDAT", ifdc_rdr);
#endif
			ifdc_secdat = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* MODREG */
		case 0x0b:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->MODREG", ifdc_rdr);
#endif
			ifdc_modreg = ifdc_rdr;
			ifdc_fsr = (BYTE)fdc_status[FDC_S];
			break;
		/* OTHERS */
		default:
#ifdef FDC_LOG
			LOG1(Normal, "rdr:%02x->OTHERS", ifdc_rdr);
#endif
			ifdc_ssr |= IFDC_ST_DAT_ERR;
			break;
	}
}
#endif

/*
 *	FDC
 *	PoCgǂݏo
 */
BOOL FASTCALL fdc_readb(WORD addr, BYTE *dat)
{
	int type = FDC_M;
	event_t *ev;

	/* 1MB[hݒ */
	if (fdc_mode) {
		type = FDC_S;
	}
	else {
		/* FM-7/8[hFDC͂ȂɂȂ */
		if (!fdc_enable[type]) {
			if ((model_type == MODEL_FM7) ||
				(model_type == MODEL_FM8)) {
				return FALSE;
			}
		}
	}

	switch (addr) {
		/* Xe[^XWX^ */
		case 0xfd18:
			fdc_make_stat(type);
			*dat = fdc_status[type];
			/* Xe[^X`FbN() */
			fdc_drqirq[type] |= (BYTE)0x20;
			/* BUSY */
			if ((fdc_status[type] & FDC_ST_BUSY) && (fdc_dataptr[type] == NULL)) {
				if (!fdc_wait && (fdc_cmdtype[type] == 1)) {
					/* IRQ On */
					fdc_drqirq[type] |= (BYTE)0x40;

					/* HEAD ENGAGEDtO𗎂Ƃ */
					fdc_status[type] &= ~FDC_ST_HEADENG;
				}
				if (!fdc_wait || (fdc_cmdtype[type] != 1) || (fdc_drqirq[type] & 0x40)) {
					/* BUSYtO𗎂Ƃ */
					fdc_status[type] &= ~FDC_ST_BUSY;
					fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
				}
				/* SFD荞݃[h̏ꍇ(Ô) */
				if (!type || !sfd_firq_mode) {
					/* xBUSY */
					return TRUE;
				}
			}
			/* FDC荞݂AŎ~߂ */
			if (type && sfd_firq_mode) {
				sfd_drq_flag = FALSE;
				fdc_drqirq[type] &= ~0x40;
				maincpu_firq();
			}
			else {
				if (type) {
					sfd_irq_flag = FALSE;
				}
				else {
					mfd_irq_flag = FALSE;
				}
				fdc_drqirq[type] &= ~0x40;
				maincpu_irq();
			}
			return TRUE;

		/* gbNWX^ */
		case 0xfd19:
			*dat = fdc_trkreg[type];
			return TRUE;

		/* ZN^WX^ */
		case 0xfd1a:
			*dat = fdc_secreg[type];
			return TRUE;

		/* f[^WX^ */
		case 0xfd1b:
			*dat = fdc_datareg[type];
			/* JE^ */
			if (fdc_dataptr[type] && (fdc_drqirq[type] & 0x20)) {
				fdc_nowcnt[type]++;
				/* DRQ,IRQ */
				if (fdc_nowcnt[type] == fdc_totalcnt[type]) {
					if (fdc_wait) {
						if (type) {
							schedule_delevent(EVENT_SFDC_L);
						}
						else {
							schedule_delevent(EVENT_FDC_L);
						}
					}
					fdc_status[type] &= ~FDC_ST_BUSY;
					fdc_status[type] &= ~FDC_ST_DRQ;
					fdc_drqirq[type] &= (BYTE)~0x80;
					if (type && sfd_firq_mode) {
						/* f[^NGXg */
						sfdc_drq_sub();
					}

					if ((fdc_cmdtype[type] == 2) && (fdc_command[type] & 0x10)) {
						/* }`ZN^ */
						fdc_status[type] |= FDC_ST_BUSY;
						fdc_dataptr[type] = NULL;
						fdc_drqirq[type] = 0x10;
						if (type) {
							schedule_setevent(EVENT_SFDC_M, 30, sfdc_multi_event);
						}
						else {
							schedule_setevent(EVENT_FDC_M, 30, mfdc_multi_event);
						}
						return TRUE;
					}
					/* VOZN^ */
					fdc_dataptr[type] = NULL;
					if (type && sfd_firq_mode) {
						/* f[^]荞 */
						schedule_setevent(EVENT_FDC_NMI, fdc_eod_time, sfdc_nmi_event);
					}
					else {
						if (type) {
							mainetc_sfdc();
						}
						else {
							mainetc_fdc();
						}
						fdc_drqirq[type] = 0x40;
						fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
					}
					/* Read Track͕KLOST DATAAobt@C */
					if (fdc_cmdtype[type] == 6) {
						fdc_status[type] |= FDC_ST_LOSTDATA;
						fdc_readbuf(type, fdc_drvreg[type]);
					}
				}
				else {
					fdc_datareg[type] = fdc_dataptr[type][fdc_nowcnt[type]];
					if (fdc_wait) {
						/* 1oCg]ԂقȂƂĐݒ */
						if (type) {
							ev = schedule_getevent(EVENT_SFDC_L);
							if (!ev->flag || ev->reload != (DWORD)fdc_onetime[type]) {
								schedule_setevent(
									EVENT_SFDC_L, (DWORD)fdc_onetime[type], sfdc_drq_event);
							}
						}
						else {
							ev = schedule_getevent(EVENT_FDC_L);
							if (!ev->flag || ev->reload != (DWORD)fdc_onetime[type]) {
								schedule_setevent(
									EVENT_FDC_L, (DWORD)fdc_onetime[type], mfdc_drq_event);
							}
						}

						fdc_status[type] &= (BYTE)~FDC_ST_DRQ;
						fdc_drqirq[type] &= (BYTE)~0x80;
					}
					else {
						fdc_drqirq[type] |= (BYTE)0x80;
					}
					if (type && sfd_firq_mode) {
						/* f[^NGXg */
						sfdc_drq_sub();
					}
				}
			}
			return TRUE;

		/* wbhWX^ */
		case 0xfd1c:
			if (type) {
				/* PxE{x */
				if (fdc_dden[type]) {
					*dat = (BYTE)(fdc_sidereg[type] | 0xfe);
				}
				else {
					*dat = (BYTE)(fdc_sidereg[type] | 0xbe);
				}
				/* hAbN */
				if (!fdc_doorlock[fdc_drvreg[type]]) {
					*dat &= ~0x10;
				}
			}
			else {
				*dat = (BYTE)(fdc_sidereg[type] | 0xfe);
			}
			return TRUE;

		/* hCuWX^ */
		case 0xfd1d:
			if (fdc_motor[type]) {
				*dat = (BYTE)(0xbc | fdc_drvreg[type]);
			}
			else {
				*dat = (BYTE)(0x3c | fdc_drvreg[type]);
			}
			return TRUE;

		/* [hWX^ */
		case 0xfd1e:
			if (type) {
				/* MSB:ЖʁEʃtO */
				*dat = (BYTE)(fdc_sided[fdc_drvreg[type]] | 0x7f);
			}
			else {
				*dat = 0xFF;
			}
			return TRUE;

		/* DRQ,IRQ */
		case 0xfd1f:
			*dat = (BYTE)(fdc_drqirq[type] | 0x3f);
			/* DRQ܂IRQ`FbN() */
			fdc_drqirq[type] |= (BYTE)0x20;
			/* f[^]ŖΏI(z̐_a) */
			if (fdc_dataptr[type] == NULL) {
				if (!fdc_wait && (fdc_cmdtype[type] == 1) && (fdc_status[type] & FDC_ST_BUSY)) {
					/* IRQ On */
					fdc_drqirq[type] |= (BYTE)0x40;

					/* HEAD ENGAGEDtO𗎂Ƃ */
					fdc_status[type] &= ~FDC_ST_HEADENG;
				}
				if (!fdc_wait || (fdc_cmdtype[type] != 1) || (fdc_drqirq[type] & 0x40)) {
					/* BUSYtO𗎂Ƃ */
					fdc_status[type] &= ~FDC_ST_BUSY;
					fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
				}
			}
			return TRUE;
	}

	return FALSE;
}

/*
 *	FDC
 *	PoCg
 */
BOOL FASTCALL fdc_writeb(WORD addr, BYTE dat)
{
	int type = FDC_M;
	event_t *ev;

	/* 1MB[hݒ */
	if (fdc_mode) {
		type = FDC_S;
	}
	else {
		/* FM-7/8[hFDC͂ȂɂȂ */
		if (!fdc_enable[type]) {
			if ((model_type == MODEL_FM7) ||
				(model_type == MODEL_FM8)) {
				return FALSE;
			}
		}

		/* AhX`FbN */
		if ((addr & 0xfff8) == 0xfd18) {
			/* 1MB[h0xfd04쎞ɗLɂȂ */
			fdc_use[type] = TRUE;
		}
	}

	switch (addr) {
		/* R}hWX^ */
		case 0xfd18:
			fdc_dsenable[type] = 0x00;	/* R}hsDS ENABLEԂƂ */
			fdc_command[type] = dat;
			fdc_process_cmd(type);
			/* motor onɂȂ܂܂̃R}hs΍(F-BASIC V3.5) */
			if ((type == FDC_S) && (fdc_motor[type] == 0)) {
				fdc_motor[type] = (BYTE)0x80;
				ev = schedule_getevent(fdd_event_motor[type]);
				if (ev->flag == EVENT_NOTUSE) {
					schedule_setevent(fdd_event_motor[type], fdd_rotation_time[type], fdd_motor_event);
				}
			}
			return TRUE;

		/* gbNWX^ */
		case 0xfd19:
			fdc_trkreg[type] = dat;
			/* R}hs̃gbNWX^΍ */
			/* (F-BASIC V3.3L3x/V3.4L2x VOLCOPY/SUBSET(VCOPYEB)) */
			if (((fdc_status[type] & FDC_ST_BUSY) && (fdc_nowcnt[type] == 0)) &&
				((fdc_cmdtype[type] == 2) || (fdc_cmdtype[type] == 3))) {
				fdc_process_cmd(type);
			}
			return TRUE;

		/* ZN^WX^ */
		case 0xfd1a:
			fdc_secreg[type] = dat;
			/* R}hs̃ZN^WX^΍ */
			/* (F-BASIC V3.3L3x/V3.4L2x VOLCOPY/SUBSET(VCOPYEB)) */
			if (((fdc_status[type] & FDC_ST_BUSY) && (fdc_nowcnt[type] == 0)) &&
				((fdc_cmdtype[type] == 2) || (fdc_cmdtype[type] == 3))) {
				fdc_process_cmd(type);
			}
			return TRUE;

		/* f[^WX^ */
		case 0xfd1b:
			fdc_datareg[type] = dat;
			/* JE^ */
			if (fdc_dataptr[type] && (fdc_drqirq[type] & 0x20)) {
				fdc_dataptr[type][fdc_nowcnt[type]] = fdc_datareg[type];
				fdc_nowcnt[type]++;
				/* DRQ,IRQ */
				if (fdc_nowcnt[type] == fdc_totalcnt[type]) {
					if (fdc_wait) {
						if (type) {
							schedule_delevent(EVENT_SFDC_L);
						}
						else {
							schedule_delevent(EVENT_FDC_L);
						}
					}
					fdc_status[type] &= ~FDC_ST_DRQ;
					fdc_drqirq[type] &= (BYTE)(~0x80);
					fdc_status[type] &= ~FDC_ST_BUSY;
					if (type && sfd_firq_mode) {
						/* f[^NGXg */
						sfdc_drq_sub();
					}

					if (fdc_cmdtype[type] == 3) {
						/* ZN^ݏI */
						if (!fdc_writesec(type)) {
							fdc_status[type] |= FDC_ST_WRITEFAULT;
						}
					}
					if (fdc_cmdtype[type] == 5) {
						/* gbNݏI */
						if (!fdc_writetrk(type)) {
							fdc_status[type] |= FDC_ST_WRITEFAULT;
						}
						/* tH[}bĝߎgpobt@C */
						fdc_readbuf(type, fdc_drvreg[type]);
					}
					if ((fdc_cmdtype[type] == 3) && (fdc_command[type] & 0x10)) {
						/* }`ZN^ */
						fdc_status[type] |= FDC_ST_BUSY;
						fdc_dataptr[type] = NULL;
						fdc_drqirq[type] = 0x10;
						if (type) {
							schedule_setevent(EVENT_SFDC_M, 30, sfdc_multi_event);
						}
						else {
							schedule_setevent(EVENT_FDC_M, 30, mfdc_multi_event);
						}
						return TRUE;
					}
					fdc_dataptr[type] = NULL;
					if (type && sfd_firq_mode) {
						/* f[^]荞 */
						schedule_setevent(EVENT_FDC_NMI, fdc_eod_time, sfdc_nmi_event);
					}
					else {
						if (type) {
							mainetc_sfdc();
						}
						else {
							mainetc_fdc();
						}
						fdc_drqirq[type] = 0x40;
						fdc_access[type][fdc_drvreg[type]] = FDC_ACCESS_READY;
					}
				}
				else {
					if (fdc_wait) {
						/* 1oCg]ԂقȂƂĐݒ */
						if (type) {
							ev = schedule_getevent(EVENT_SFDC_L);
							if (!ev->flag || ev->reload != (DWORD)fdc_onetime[type]) {
								schedule_setevent(
									EVENT_SFDC_L, (DWORD)fdc_onetime[type], sfdc_drq_event);
							}
						}
						else {
							ev = schedule_getevent(EVENT_FDC_L);
							if (!ev->flag || ev->reload != (DWORD)fdc_onetime[type]) {
								schedule_setevent(
									EVENT_FDC_L, (DWORD)fdc_onetime[type], mfdc_drq_event);
							}
						}

						fdc_status[type] &= (BYTE)~FDC_ST_DRQ;
						fdc_drqirq[type] &= (BYTE)~0x80;
					}
					else {
						fdc_drqirq[type] |= (BYTE)0x80;
					}
					if (type && sfd_firq_mode) {
						/* f[^NGXg */
						sfdc_drq_sub();
					}
				}
			}
			return TRUE;

		/* wbhWX^ */
		case 0xfd1c:
			if (type) {
				/* PxE{x */
				if ((dat & 0x40) != fdc_dden[type]) {
					fdc_dden[type] = (BYTE)(dat & 0x40);
				}
				/* hAbN */
				if ((dat & 0x10) != fdc_doorlock[fdc_drvreg[type]]) {
					fdc_doorlock[fdc_drvreg[type]] = (BYTE)(dat & 0x10);
				}
				/* FDCNOT READY̎̓bNȂ */
				if ((fdc_ready[type][fdc_drvreg[type]] == FDC_TYPE_NOTREADY) ||
					fdc_teject[type][fdc_drvreg[type]]) {
					fdc_doorlock[fdc_drvreg[type]] = (BYTE)0;
				}
			}
			if ((dat & 0x01) != fdc_sidereg[type]) {
				fdc_sidereg[type] = (BYTE)(dat & 0x01);
				fdc_readbuf(type, fdc_drvreg[type]);
			}
			return TRUE;

		/* hCuWX^ */
		case 0xfd1d:
			/* hCuύXȂAfdc_readbuf */
			if (fdc_drvreg[type] != (dat & 0x03)) {
				fdc_drvreg[type] = (BYTE)(dat & 0x03);
				fdc_readbuf(type, fdc_drvreg[type]);
			}
			fdc_motor[type] = (BYTE)(dat & 0x80);
			/* hCuȂA[^~߂ */
			if (fdc_drvreg[type] >= FDC_DRIVES) {
				LOG1(Warning, "drvreg:%d\n", fdc_drvreg[type]);
				fdc_motor[type] = 0;
			}

			/* [^[Cxg */
			ev = schedule_getevent(fdd_event_motor[type]);
			if (fdc_motor[type]) {
				if (ev->flag == EVENT_NOTUSE) {
					schedule_setevent(fdd_event_motor[type], fdd_rotation_time[type], fdd_motor_event);
				}
			} else {
				if (ev->flag == EVENT_ENABLED) {
					schedule_delevent(fdd_event_motor[type]);
				}
			}

			if (type == FDC_S) {
				/* DS ENABLE͋L̂(ۂɎgpɂ͋@ˑ̑ΉKv) */
				fdc_dsenable[type] = (dat & 0x40);
			}
			return TRUE;

		/* [hWX^ */
		case 0xfd1e:
			return TRUE;
	}

	return FALSE;
}

#if 1	/* FM-8 8" IFDCn */
/*
 *	IFDC
 *	PoCgǂݏo
 */
BOOL FASTCALL ifdc_readb(WORD addr, BYTE *dat)
{
	if (addr < 0xfd30 || addr > 0xfd32) {
		return FALSE;
	}

	/* IFDCFM-8[h */
	if (model_type != MODEL_FM8 || !fdc_enable[FDC_S]) {
		return FALSE;
	}

	switch (addr) {
		/* VXeXe[^XWX^(SSR) */
		case 0xfd30:
			fdc_mode = 1;
			fdc_readb(0xfd18, &ifdc_fsr);
			fdc_mode = 0;
			/* BUSY */
			if (ifdc_fsr & FDC_ST_BUSY) {
				/* BUSY */
				ifdc_ssr &= ~IFDC_ST_READY;
			}
			else {
				/* READY */
				ifdc_ssr |= IFDC_ST_READY;
			}
			/* ERROR */
			if (ifdc_fsr & (FDC_ST_CRCERR | FDC_ST_NOTREADY)) {
				/* ERROR */
				ifdc_ssr |= IFDC_ST_FDD_ERR;
			}
			else {
				/* NO ERROR */
				ifdc_ssr &= ~IFDC_ST_FDD_ERR;
			}
			if ((ifdc_fsr & FDC_ST_WRITEP) &&
				((fdc_cmdtype[FDC_S] == 3) || (fdc_cmdtype[FDC_S] == 5))) {
				/* ERROR */
				ifdc_ssr |= IFDC_ST_FDD_ERR;
			}
			*dat = ifdc_ssr;
			return TRUE;

		/* FDCXe[^XWX^(FSR) */
		case 0xfd31:
			*dat = ifdc_fsr;
			return TRUE;

		/* f[^WX^ */
		case 0xfd32:
			fdc_mode = 1;
			fdc_readb(0xfd1b, dat);
			fdc_mode = 0;
			return TRUE;
	}

	return FALSE;
}

/*
 *	IFDC
 *	PoCg
 */
BOOL FASTCALL ifdc_writeb(WORD addr, BYTE dat)
{
	if (addr < 0xfd30 || addr > 0xfd32) {
		return FALSE;
	}

	/* IFDCFM-8[h */
	if (model_type != MODEL_FM8 || !fdc_enable[FDC_S]) {
		return FALSE;
	}

	fdc_use[FDC_S] = TRUE;

	switch (addr) {
		/* WX^U[uWX^(RRR) */
		case 0xfd30:
			ifdc_rrr = dat;
			fdc_mode = 1;
			ifdc_reserve_cmd();
			fdc_mode = 0;
			return TRUE;

		/* WX^f[^WX^(RDR) */
		case 0xfd31:
			ifdc_rdr = dat;
			return TRUE;

		/* f[^WX^ */
		case 0xfd32:
			fdc_mode = 1;
			fdc_writeb(0xfd1b, dat);
			fdc_mode = 0;
			return TRUE;
	}

	return FALSE;
}
#endif

#if 1	/* o[W݊p */
/*
 *	FDC
 *	[hio[W݊pj
 */
static BOOL FASTCALL fdc_load_old(int fileh, int ver)
{
	int i;
	int j;
	BYTE ready[2][FDC_DRIVES];
	char fname[2][FDC_DRIVES][256 + 1];
	BYTE media[2][FDC_DRIVES];
	WORD offset;
	BOOL tmp;
	int pathlen;

	/* t@C̍ő啶 */
	if (((ver >= 305) && (ver <= 399)) || (ver >= 404)) {
		pathlen = 256;
	}
	else {
		pathlen = 128;
	}

	/* t@C֌WɎĂ */
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &ready[j][i])) {
				return FALSE;
			}
		}
	}
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_read(fileh, (BYTE*)fname[j][i], pathlen + 1)) {
				return FALSE;
			}
		}
	}
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &media[j][i])) {
				return FALSE;
			}
		}
	}

	/* ă}Eg݂ */
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			fdc_setdisk(j, i, NULL);
			if (ready[j][i] != FDC_TYPE_NOTREADY) {
				fdc_setdisk(j, i, fname[j][i]);
				if (fdc_ready[j][i] != FDC_TYPE_NOTREADY) {
					if (fdc_medias[j][i] >= (media[j][i] + 1)) {
						fdc_setmedia(j, i, media[j][i]);
					}
				}
			}
		}
	}

	/* Ver4g că`FbNӖȂł(` */
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_bool_read(fileh, &fdc_teject[j][i])) {
				return FALSE;
			}
		}
	}

	/* t@CXe[^X */
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_track[j][i])) {
				return FALSE;
			}
		}
	}
	if (!file_read(fileh, fdc_buffer[FDC_M], fdc_buffer_size[FDC_M])) {
		return FALSE;
	}
	if (!file_read(fileh, fdc_buffer[FDC_S], fdc_buffer_size[FDC_S])) {
		return FALSE;
	}

	/* fdc_dataptr͊Ɉˑf[^|C^ */
	if (!file_word_read(fileh, &offset)) {
		return FALSE;
	}
	if ((DWORD)offset >= fdc_buffer_size[FDC_M]) {
		fdc_dataptr[FDC_M] = NULL;
	}
	else {
		fdc_dataptr[FDC_M] = &fdc_buffer[FDC_M][offset];
	}
	if (!file_word_read(fileh, &offset)) {
		return FALSE;
	}
	if ((DWORD)offset >= fdc_buffer_size[FDC_S]) {
		fdc_dataptr[FDC_S] = NULL;
	}
	else {
		fdc_dataptr[FDC_S] = &fdc_buffer[FDC_S][offset];
	}

	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_dword_read(fileh, &fdc_seekofs[j][i])) {
				return FALSE;
			}
		}
	}
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_dword_read(fileh, &fdc_secofs[j][i])) {
				return FALSE;
			}
		}
	}

	/* I/O */
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_command[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_status[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_trkreg[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_secreg[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_datareg[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_sidereg[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_drvreg[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_motor[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_drqirq[j])) {
			return FALSE;
		}
		if (ver < 600) {
			fdc_drqirq[j] |= (BYTE)0x20;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_cmdtype[j])) {
			return FALSE;
		}
	}

	/* ̑ */
	for (j=0; j<2; j++) {
		if (!file_word_read(fileh, &fdc_totalcnt[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_word_read(fileh, &fdc_nowcnt[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_bool_read(fileh, &fdc_seekvct[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		if (!file_byte_read(fileh, &fdc_indexcnt[j])) {
			return FALSE;
		}
	}
	for (j=0; j<2; j++) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_access[j][i])) {
				return FALSE;
			}
		}
	}
	/* o[WƂ̌݊p(u[gtO) */
	for (i=0; i<2; i++) {
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
	}
	if (ver >= 403) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_doorlock[i])) {
				return FALSE;
			}
		}
	}
	else {
		memset(fdc_doorlock, 0, sizeof(fdc_doorlock));
	}
	if (ver >= 1000) {
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_sided[i])) {
				return FALSE;
			}
		}
	}
	else {
		memset(fdc_sided, 0, sizeof(fdc_sided));
	}

	/* Cxg */
	schedule_handle(EVENT_FDD_MOTOR, fdd_motor_event);
	schedule_handle(EVENT_FDC_M, mfdc_multi_event);
	if (ver >= 300) {
		/* Ver9.05/7.05g */
		if (!file_bool_read(fileh, &fdc_wait)) {
			return FALSE;
		}
		schedule_handle(EVENT_FDD_SEEK, mfdd_seek_event);
	}
	else {
		/* EFCg[hFALSE */
		fdc_wait = FALSE;

		/* V[NCxg(zbgZbgCxg)폜 */
		schedule_delevent(EVENT_FDD_SEEK);
	}
	if (fdc_wait) {
		schedule_handle(EVENT_FDC_L, mfdc_drq_event);
	}
	else {
		schedule_handle(EVENT_FDC_L, mfdc_lost_event);
	}

	/* XM7 V1.2 - Ver4g 1MB FDC֘A */
	if (ver >= 401) {
		if (!file_byte_read(fileh, &fdc_mode)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
		sfd_firq_mode = !tmp;
		if (!file_byte_read(fileh, &fdc_dden[FDC_S])) {
			return FALSE;
		}
	}
	else {
		fdc_mode = 0;
		sfd_firq_mode = FALSE;
		fdc_dden[FDC_S] = 0;
	}
	if (ver >= 403) {
		if (!file_bool_read(fileh, &fdc_enable[FDC_S])) {
			return FALSE;
		}
	}
	else {
		fdc_enable[FDC_S] = TRUE;
	}
	if (ver >= 405) {
		if (!file_bool_read(fileh, &fdc_use[FDC_S])) {
			return FALSE;
		}
	}
	else {
		fdc_use[FDC_S] = fdc_enable[FDC_S];
	}
	fdc_dsenable[FDC_S] = 0x00;


	return TRUE;
}
#endif

/*
 *	FDC
 *	Xe[gZ[uʏ
 */
static BOOL FASTCALL fdc_save_common(int type, int fileh)
{
	int i;
	DWORD size;

	size = fdc_buffer_size[type];

	/* t@C֌WɎĂ */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_ready[type][i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_write(fileh, (BYTE*)fdc_fname[type][i], 256 + 1)) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_media[type][i])) {
			return FALSE;
		}
	}

	/* Ver4g */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_bool_write(fileh, fdc_teject[type][i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_track[type][i])) {
			return FALSE;
		}
	}
	if (!file_write(fileh, fdc_buffer[type], size)) {
		return FALSE;
	}

	/* fdc_dataptr͊Ɉˑf[^|C^ */
	if (!fdc_dataptr[type]) {
		if (!file_word_write(fileh, size)) {
			return FALSE;
		}
	}
	else {
		if (!file_word_write(fileh, (WORD)(fdc_dataptr[type] - &fdc_buffer[type][0]))) {
			return FALSE;
		}
	}

	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_write(fileh, fdc_seekofs[type][i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_write(fileh, fdc_secofs[type][i])) {
			return FALSE;
		}
	}

	/* I/O */
	if (!file_byte_write(fileh, fdc_command[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_status[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_trkreg[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_secreg[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_datareg[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_sidereg[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_drvreg[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_motor[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_drqirq[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_cmdtype[type])) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_write(fileh, fdc_totalcnt[type])) {
		return FALSE;
	}
	if (!file_word_write(fileh, fdc_nowcnt[type])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, fdc_seekvct[type])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_indexcnt[type])) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_access[type][i])) {
			return FALSE;
		}
	}
	if (!file_bool_write(fileh, fdc_enable[type])) {
		return FALSE;
	}
	if (!file_bool_write(fileh, fdc_use[type])) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	FDC
 *	Xe[g[hʏ
 */
static BOOL FASTCALL fdc_load_common(int type, int fileh, int ver)
{
	int i;
	BYTE ready[2][FDC_DRIVES];
	char fname[2][FDC_DRIVES][256 + 1];
	BYTE media[2][FDC_DRIVES];
	WORD offset;
	int pathlen;
	DWORD size;

	/* t@C̍ő啶 */
	if (((ver >= 305) && (ver <= 399)) || (ver >= 404)) {
		pathlen = 256;
	}
	else {
		pathlen = 128;
	}

	size = fdc_buffer_size[type];

	/* t@C֌WɎĂ */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &ready[type][i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_read(fileh, (BYTE*)fname[type][i], pathlen + 1)) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &media[type][i])) {
			return FALSE;
		}
	}

	/* ă}Eg݂ */
	for (i=0; i<FDC_DRIVES; i++) {
		fdc_setdisk(type, i, NULL);
		if (ready[type][i] != FDC_TYPE_NOTREADY) {
			fdc_setdisk(type, i, fname[type][i]);
			if (fdc_ready[type][i] != FDC_TYPE_NOTREADY) {
				if (fdc_medias[type][i] >= (media[type][i] + 1)) {
					fdc_setmedia(type, i, media[type][i]);
				}
			}
		}
	}

	/* Ver4g că`FbNӖȂł(` */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_bool_read(fileh, &fdc_teject[type][i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &fdc_track[type][i])) {
			return FALSE;
		}
	}
	if (!file_read(fileh, fdc_buffer[type], size)) {
		return FALSE;
	}

	/* fdc_dataptr͊Ɉˑf[^|C^ */
	if (!file_word_read(fileh, &offset)) {
		return FALSE;
	}
	if ((DWORD)offset >= size) {
		fdc_dataptr[type] = NULL;
	}
	else {
		fdc_dataptr[type] = &fdc_buffer[type][offset];
	}

	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_read(fileh, &fdc_seekofs[type][i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_read(fileh, &fdc_secofs[type][i])) {
			return FALSE;
		}
	}

	/* I/O */
	if (!file_byte_read(fileh, &fdc_command[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_status[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_trkreg[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_secreg[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_datareg[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_sidereg[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_drvreg[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_motor[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_drqirq[type])) {
		return FALSE;
	}
	if (ver < 600) {
		fdc_drqirq[type] |= (BYTE)0x20;	/* Kv? */
	}
	if (!file_byte_read(fileh, &fdc_cmdtype[type])) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_read(fileh, &fdc_totalcnt[type])) {
		return FALSE;
	}
	if (!file_word_read(fileh, &fdc_nowcnt[type])) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &fdc_seekvct[type])) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_indexcnt[type])) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &fdc_access[type][i])) {
			return FALSE;
		}
	}
	if (ver >= 1003) {
		if (!file_bool_read(fileh, &fdc_enable[type])) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &fdc_use[type])) {
			return FALSE;
		}
	}
	else {
		if (((ver >= 310) && (ver <= 399)) && type == 0) {
			if (!file_bool_read(fileh, &fdc_enable[type])) {
				return FALSE;
			}
		}
		else {
			fdc_enable[type] = TRUE;
		}
		fdc_use[type] = TRUE;
	}

	/* g */
	fdi_init[type] = FALSE;
	fdc_onetime[type] = 32;

	return TRUE;
}

/*
 *	FDC
 *	Z[u
 */
BOOL FASTCALL fdc_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!fdc_save_common(FDC_M, fileh)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, TRUE)) {
		return FALSE;
	}

	/* Ver9.05/7.05g */
	if (!file_bool_write(fileh, fdc_wait)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	FDC
 *	[h
 */
BOOL FASTCALL fdc_load(int fileh, int ver)
{
	BOOL tmp;

	LOG0(Normal, "[h");

#if 1	/* o[W݊p */
	/* o[W`FbN */
	if ((ver >= 401) && (ver <= 499)) {
		return fdc_load_old(fileh, ver);
	}
#endif

	/* Ver10dlύX */
	if (!fdc_load_common(FDC_M, fileh, ver)) {
		return FALSE;
	}
	/* o[WƂ̌݊p(u[gtO) */
	if (!file_bool_read(fileh, &tmp)) {
		return FALSE;
	}

	/* Cxg */
	schedule_handle(EVENT_FDD_MOTOR, fdd_motor_event);
	schedule_handle(EVENT_FDC_M, mfdc_multi_event);
	if (ver >= 300) {
		/* Ver9.05/7.05g */
		if (!file_bool_read(fileh, &fdc_wait)) {
			return FALSE;
		}
		schedule_handle(EVENT_FDD_SEEK, mfdd_seek_event);
	}
	else {
		/* EFCg[hFALSE */
		fdc_wait = FALSE;

		/* V[NCxg(zbgZbgCxg)폜 */
		schedule_delevent(EVENT_FDD_SEEK);
	}
	/* FDANZXEFCgLȏꍇ̓CxgnhύX */
	if (fdc_wait) {
		schedule_handle(EVENT_FDC_L, mfdc_drq_event);
	}
	else {
		schedule_handle(EVENT_FDC_L, mfdc_lost_event);
	}

	return TRUE;
}

/*
 *	SFDC
 *	Z[u
 */
BOOL FASTCALL sfdc_save(int fileh)
{
	int i;

	LOG0(Normal, "Z[u");

	/* FM-77 SFDC */
	if (!fdc_save_common(FDC_S, fileh)) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_doorlock[i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_sided[i])) {
			return FALSE;
		}
	}
	if (!file_byte_write(fileh, fdc_mode)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, sfd_firq_mode)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_dden[FDC_S])) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_dsenable[FDC_S])) {
		return FALSE;
	}

#if 1	/* FM-8 8" IFDCn */
	if (!file_byte_write(fileh, ifdc_ssr)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_fsr)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_rrr)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_rdr)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_cmdreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_trkreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_ssreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_sereg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_sidreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_dnsreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_drvreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_dlcreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_sformt)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_secdat)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, ifdc_modreg)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	SFDC
 *	[h
 */
BOOL FASTCALL sfdc_load(int fileh, int ver)
{
	int i;

	LOG0(Normal, "[h");

	/* FM-77 SFDC Ver10dlύX */
	if (ver >= 1000) {
		if (!fdc_load_common(FDC_S, fileh, ver)) {
			return FALSE;
		}
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_doorlock[i])) {
				return FALSE;
			}
		}
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_sided[i])) {
				return FALSE;
			}
		}

		/* XM7 V1.2 - Ver4g 1MB FDC֘A */
		if (!file_byte_read(fileh, &fdc_mode)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &sfd_firq_mode)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &fdc_dden[FDC_S])) {
			return FALSE;
		}
		if (ver < 1003) {
			if (!file_bool_read(fileh, &fdc_enable[FDC_S])) {
				return FALSE;
			}
			if (!file_bool_read(fileh, &fdc_use[FDC_S])) {
				return FALSE;
			}
		}
		if (ver >= 1009) {
			if (!file_byte_read(fileh, &fdc_dsenable[FDC_S])) {
				return FALSE;
			}
		}
	}
#if 1	/* o[W݊p */
	else if (ver >= 401) {
		/* fdc_load(fdc_load_old)Ń[h */
	}
#endif
	else {
		for (i=0; i<FDC_DRIVES; i++) {
			fdc_setdisk(FDC_S, i, NULL);
			fdc_teject[FDC_S][i] = FALSE;
			fdc_track[FDC_S][i] = 0;
			fdc_seekofs[FDC_S][i] = 0;
			fdc_secofs[FDC_S][i] = 0;
			fdc_access[FDC_S][i] = 0;
			fdc_doorlock[i] = 0;
			fdc_sided[i] = 0x80;
		}

		fdc_command[FDC_S] = 0xff;
		fdc_status[FDC_S] = 0;
		fdc_trkreg[FDC_S] = 0;
		fdc_secreg[FDC_S] = 1;
		fdc_datareg[FDC_S] = 0;
		fdc_sidereg[FDC_S] = 0;
		fdc_drvreg[FDC_S] = 0;
		fdc_motor[FDC_S] = 0;

		fdc_drqirq[FDC_S] = 0xff;
		fdc_cmdtype[FDC_S] = 0;
		fdc_totalcnt[FDC_S] = 0;
		fdc_nowcnt[FDC_S] = 0;
		fdc_seekvct[FDC_S] = 0;
		fdc_indexcnt[FDC_S] = 0;

		memset(fdc_buffer[FDC_S], 0, fdc_buffer_size[FDC_S]);
		fdc_dataptr[FDC_S] = NULL;

		fdc_mode = 0;
		sfd_firq_mode = FALSE;
		fdc_dden[FDC_S] = 0;
		fdc_enable[FDC_S] = FALSE;
		fdc_use[FDC_S] = FALSE;
		fdc_dsenable[FDC_S] = FALSE;
	}

	/* Cxg */
	schedule_handle(EVENT_SFDD_MOTOR, fdd_motor_event);
	schedule_handle(EVENT_FDC_NMI, sfdc_nmi_event);
	schedule_handle(EVENT_SFDC_M, sfdc_multi_event);
	if (fdc_wait) {
		schedule_handle(EVENT_SFDC_L, sfdc_drq_event);
	}
	else {
		schedule_handle(EVENT_SFDC_L, sfdc_lost_event);
	}
	if (ver >= 401) {
		schedule_handle(EVENT_SFDD_SEEK, sfdd_seek_event);
	}
	else {
		schedule_delevent(EVENT_SFDD_SEEK);
	}

#if 1	/* FM-8 8" IFDCn */
	if (ver >= 405) {
		if (!file_byte_read(fileh, &ifdc_ssr)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_fsr)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_rrr)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_rdr)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_cmdreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_trkreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_ssreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_sereg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_sidreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_dnsreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_drvreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_dlcreg)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_sformt)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_secdat)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &ifdc_modreg)) {
			return FALSE;
		}
	}
	else {
		ifdc_ssr = 0;
		ifdc_fsr = 0;
		ifdc_rrr = 0xff;
		ifdc_rdr = 0;

		ifdc_cmdreg = 0;
		ifdc_trkreg = 0;
		ifdc_ssreg = 0;
		ifdc_sereg = 0;
		ifdc_sidreg = 0;
		ifdc_dnsreg = 0;
		ifdc_drvreg = 0;
		ifdc_dlcreg = 0;
		ifdc_sformt = 0;
		ifdc_secdat = 0;
		ifdc_modreg = 0;
	}
#endif

	return TRUE;
}
#else	/* XM7_VER == 1 && defined(SFDC) */
/*
 *	O[o [N
 */
BYTE fdc_command;					/* FDCR}h */
BYTE fdc_status;					/* FDCXe[^X */
BYTE fdc_trkreg;					/* gbNWX^ */
BYTE fdc_secreg;					/* ZN^WX^ */
BYTE fdc_datareg;					/* f[^WX^ */
BYTE fdc_sidereg;					/* TChWX^ */
BYTE fdc_drvreg;					/* _hCu */
#if XM7_VER >= 2
BYTE fdc_dsenable;					/* DSCl[u */
BYTE fdc_drvregP;					/* hCu */
#endif
BYTE fdc_motor;						/* [^ */
BYTE fdc_drqirq;					/* DRQIRQ */

BYTE fdc_cmdtype;					/* R}h^Cv */
WORD fdc_totalcnt;					/* g[^JE^ */
BYTE fdc_onetime;					/* 1oCg] */
WORD fdc_nowcnt;					/* JgJE^ */
BYTE fdc_ready[FDC_DRIVES];			/* fB */
BOOL fdc_teject[FDC_DRIVES];		/* ꎞCWFNg */
BOOL fdc_writep[FDC_DRIVES];		/* CgveNg */
BYTE fdc_track[FDC_DRIVES];			/* gbN */

#if XM7_VER >= 2
BYTE fdc_logidrv;					/* _hCuԍ */
BYTE fdc_physdrv[FDC_DRIVES];		/* _/hCȗΉ */
BYTE fdc_2ddmode;					/* 2D[hI */
#endif

char fdc_fname[FDC_DRIVES][256+1];	/* t@C */
char fdc_name[FDC_DRIVES][FDC_MEDIAS][17];
BOOL fdc_fwritep[FDC_DRIVES];		/* CgveNg(t@Cx) */
BYTE fdc_header[FDC_DRIVES][0x2b0];	/* D77t@Cwb_ */
BYTE fdc_medias[FDC_DRIVES];		/* fBA */
BYTE fdc_media[FDC_DRIVES];			/* fBAZNg */
BYTE fdc_access[FDC_DRIVES];		/* ANZXLED */
BOOL fdc_enable;					/* FDCLEtO */
BOOL fdc_use;						/* FDCgptO */

BOOL fdc_waitmode;					/* FDCANZXEFCg */
BOOL fdc_sound;						/* FDDV[NtO */


/*
 *	X^eBbN [N
 */
static BYTE fdc_buffer[0x2000];					/* f[^obt@ */
static BYTE *fdc_dataptr;						/* f[^|C^ */
static DWORD fdc_seekofs[FDC_DRIVES];			/* V[NItZbg */
static DWORD fdc_secofs[FDC_DRIVES];			/* ZN^ItZbg */
static DWORD fdc_foffset[FDC_DRIVES][FDC_MEDIAS];
static DWORD fdc_trklen[FDC_DRIVES];			/* gbNf[^ */
static BOOL fdc_seekvct;						/* V[N(Trk0:TRUE) */
static BYTE fdc_indexcnt;						/* INDEXz[ JE^ */
static BOOL fdc_wait;							/* EFCg[hstO */
static int fdc_seek_track;						/* waitmodepV[NJE^ */
static int fdi_init;							/* C[W̏tO*/
static int fdi_secs;							/* C[W̃ZN^ */
static int fdi_idpos[96];						/* C[W̃ZN^ʒu */
static int fdi_dtpos[96];						/* C[W̃ZN^ʒu */


/*
 *	Xebv[ge[u
 */
static const int fdc_steprate[4] = { 6000, 12000, 20000, 30000 };


/*
 *	vg^Cv錾
 */
static void FASTCALL fdc_readbuf(int drive);	/* PgbNǂݍ */
static BOOL FASTCALL fdc_lost_event(void);		/* LOST DATACxg */
static BOOL FASTCALL fdd_motor_event(void);		/* [^[Cxg */

/*
 *	ID
 */
#define	ID	MAKEID('F', 'D', 'C', ' ')
//#define FDC_LOG

/*
 *	FDC
 *	
 */
BOOL FASTCALL fdc_init(void)
{
	int i;

	/* tbs[t@C֌WZbg */
	for (i=0; i<FDC_DRIVES; i++) {
		fdc_ready[i] = FDC_TYPE_NOTREADY;
		fdc_teject[i] = FALSE;
		fdc_fwritep[i] = FALSE;
		fdc_medias[i] = 0;
	}

	/* t@CItZbgSăNA */
	memset(fdc_foffset, 0, sizeof(fdc_foffset));

	/* ftHgFDCL */
	fdc_enable = TRUE;
	fdc_use = FALSE;

	/* EFCg}[htO */
	fdc_waitmode = TRUE;
	fdc_sound = FALSE;
	fdc_wait = FALSE;

	/* [^[Cxgo^ */
	schedule_handle(EVENT_FDD_MOTOR, fdd_motor_event);

	/* C[W񖢏 */
	fdi_init = FALSE;

	return TRUE;
}

/*
 *	FDC
 *	N[Abv
 */
void FASTCALL fdc_cleanup(void)
{
	int i;

	/* tbs[t@C֌WZbg */
	for (i=0; i<FDC_DRIVES; i++) {
		fdc_ready[i] = FDC_TYPE_NOTREADY;
	}
}

/*
 *	FDC
 *	Zbg
 */
void FASTCALL fdc_reset(void)
{
#if XM7_VER >= 2
	int i;
#endif

	LOG0(Normal, "Zbg");

	/* FDCWX^Zbg */
	fdc_command = 0x03;
	fdc_status = 0;
	fdc_trkreg = 0;
	fdc_secreg = 0x01;
	fdc_datareg = 0;
	fdc_sidereg = 0;
	fdc_drvreg = 0;
#if XM7_VER >= 2
	fdc_dsenable = 0x00;
	fdc_drvregP = 0;
#endif
	fdc_motor = 0;

#if XM7_VER >= 2
	fdc_logidrv = 0;
	fdc_2ddmode = FALSE;

	/* _hCuhCuɐݒ */
	for (i=0; i<FDC_DRIVES; i++) {
		fdc_physdrv[i] = (BYTE)i;
	}
#endif

	fdc_cmdtype = 0;
	fdc_seekvct = 0;
	memset(fdc_track, 0, sizeof(fdc_track));
	fdc_dataptr = NULL;
	memset(fdc_access, 0, sizeof(fdc_access));

	/* FDCANZX֌WZbg */
	fdc_use = FALSE;

	/* f[^obt@֓ǂݍ */
	fdc_readbuf(fdc_drvreg);
}

/*-[ CRCvZ ]--------------------------------------------------------------*/

/*
 *	CRCvZe[u
 */
static WORD crc_table[256] = {
	0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
	0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
	0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
	0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
	0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
	0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
	0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
	0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
	0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
	0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
	0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
	0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
	0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
	0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
	0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
	0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
	0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
	0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
	0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
	0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
	0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
	0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
	0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
	0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
	0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
	0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
	0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
	0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
	0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
	0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
	0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
	0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

/*
 *	16bit CRCvZAZbg
 */
static void FASTCALL calc_set_crc(BYTE *addr, int size)
{
	WORD crc;

	/*  */
	crc = 0xffff;	/* CRC̏l0xffff */

	/* vZ */
	while (size > 0) {
		crc = (WORD)((crc << 8) ^ crc_table[(BYTE)(crc >> 8) ^ (BYTE)*addr++]);
		size--;
	}

	/* QoCgɃZbg(rbOGfBA) */
	*addr++ = (BYTE)(crc >> 8);
	*addr = (BYTE)(crc & 0xff);
}

/*
 *	vZ
 *	http://ja.wikipedia.org/wiki/Xorshift
 */
static DWORD FASTCALL rand_xor(void)
{
	static DWORD y = 2463534242;

	y = y ^ (y << 13);
	y = y ^ (y >> 17);
	y = y ^ (y << 15);

	return y;
}

/*-[ t@CǗ ]---------------------------------------------------------*/

/*
 *	Read Trackf[^쐬
 */
static void FASTCALL fdc_make_track(void)
{
	int i;
	int j;
	int gap3;
	int track;
	WORD count;
	WORD secs;
	WORD size;
	BOOL flag;
	BYTE *p;
	BYTE *q;
	BOOL ddm;
	int max_track;

	/* AtH[}bg`FbN */
	flag = FALSE;
	if (fdc_ready[fdc_drvreg] != FDC_TYPE_D77) {
		/* 2D/2DD/VFDt@C̃AtH[}bg`FbN͋ */
		if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
			max_track = 80;
		}
		else {
			max_track = 40;
		}

		if (fdc_track[fdc_drvreg] >= max_track) {
			/* AtH[}bg */
			flag = TRUE;
		}
#if XM7_VER >= 2
		if (((fdc_ready[fdc_drvreg] != FDC_TYPE_2DD) && fdc_2ddmode) &&
			((fdc_track[fdc_drvreg] % 2) == 1)) {
			/* 2DC[W2DDANZXAgbN̓AtH[}bg */
			flag = TRUE;
		}
#endif
	}
	else {
		if ((fdc_buffer[4] == 0) && (fdc_buffer[5] == 0)) {
			/* AtH[}bg */
			flag = TRUE;
		}
#if XM7_VER >= 2
		if (((fdc_header[fdc_drvreg][0x001b] == 0x00) && fdc_2ddmode) &&
			((fdc_track[fdc_drvreg] % 2) == 1)) {
			/* 2DC[W2DDANZXAgbN̓AtH[}bg */
			flag = TRUE;
		}
#endif
	}

	/* AtH[}bgȂA_f[^쐬 */
	if (flag) {
		p = fdc_buffer;
		for (i=0; i<0x1800; i++) {
			*p++ = (BYTE)rand_xor();
		}

		/* f[^|C^AJE^ݒ */
		fdc_dataptr = fdc_buffer;
		fdc_totalcnt = 0x1800;
		fdc_onetime = 32;
		fdc_nowcnt = 0;
		return;
	}

	/* ZN^ZoAf[^ړ */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
		secs = 16;
		q = &fdc_buffer[0x1000];
		memcpy(q, fdc_buffer, 0x1000);
	}
	else if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
#if XM7_VER >= 2
		if (fdc_2ddmode) {
			/* fBA^Cv(2D) != hCu[h(2DD) */
			track = (fdc_track[fdc_drvreg] >> 1);
		}
		else {
			/* fBA^Cv == hCu[h ␳Ȃ */
			track = fdc_track[fdc_drvreg];
		}
#else
		track = fdc_track[fdc_drvreg];
#endif
		secs = fdc_header[fdc_drvreg][track * 6 + 5];
		count = fdc_header[fdc_drvreg][track * 6 + 4];
		if (count == 0) {
			count = 128;
		}
		else {
			count *= (WORD)256;
		}
		count *= (WORD)secs;

		/* f[^Rs[ */
		q = &fdc_buffer[0x2000 - count];
		for (j = (count - 1); j >= 0; j--) {
			q[j] = fdc_buffer[j];
		}
	}
	else {
		secs = (WORD)(fdc_buffer[0x0005] * 256);
		secs += (WORD)(fdc_buffer[0x0004]);
		count = 0;
		/* SZN^܂āATCYv */
		for (j=0; j<secs; j++) {
			p = &fdc_buffer[count];
			count += (WORD)((p[0x000f] * 256 + p[0x000e]));
			count += (WORD)0x10;
		}
		/* f[^Rs[ */
		q = &fdc_buffer[0x2000 - count];
		for (j = (count - 1); j >= 0; j--) {
			q[j] = fdc_buffer[j];
		}
	}

	/* GAP3 */
	if (secs <= 5) {
		gap3 = 0x74;
	}
	else {
		if (secs <= 10) {
			gap3 = 0x54;
		}
		else {
			if (secs <= 16) {
				gap3 = 0x33;
			}
			else {
				gap3 = 0x10;
			}
		}
	}

	/* obt@ */
	p = fdc_buffer;
	count = 0;

	/* GAP0 */
	for (i=0; i<80; i++) {
		*p++ = 0x4e;
	}
	count += (WORD)80;

	/* SYNC */
	for (i=0; i<12; i++) {
		*p++ = 0;
	}
	count += (WORD)12;

	/* INDEX MARK */
	*p++ = 0xc2;
	*p++ = 0xc2;
	*p++ = 0xc2;
	*p++ = 0xfc;
	count += (WORD)4;

	/* GAP1 */
	for (i=0; i<50; i++) {
		*p++ = 0x4e;
	}
	count += (WORD)50;

	/* ZN^[v */
	for (j=0; j<secs; j++) {
		/* SYNC */
		for (i=0; i<12; i++) {
			*p++ = 0;
		}
		count += (WORD)12;

		/* ID ADDRESS MARK */
		p[0] = 0xa1;
		p[1] = 0xa1;
		p[2] = 0xa1;
		p[3] = 0xfe;

		/* ID */
		if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
			(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
#if XM7_VER >= 3
			if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) && fdc_2ddmode) {
				/* fBA^Cv(2D) != hCu[h(2DD) */
				p[4] = (BYTE)(fdc_track[fdc_drvreg] >> 1);
			}
			else if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) && !fdc_2ddmode) {
				/* fBA^Cv(2DD) != hCu[h(2D) */
				p[4] = (BYTE)(fdc_track[fdc_drvreg] << 1);
			}
			else {
				/* fBA^Cv == hCu[h ␳Ȃ */
				p[4] = fdc_track[fdc_drvreg];
			}
#else
			if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
				/* fBA^Cv(2DD) != hCu[h(2D) */
				p[4] = (BYTE)(fdc_track[fdc_drvreg] << 1);
			}
			else {
				/* fBA^Cv == hCu[h ␳Ȃ */
				p[4] = fdc_track[fdc_drvreg];
			}
#endif
			p[5] = fdc_sidereg;
			p[6] = (BYTE)(j + 1);
			p[7] = 1;
			size = 0x100;
			ddm = FALSE;
		}
		else if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
#if XM7_VER >= 2
			if (fdc_2ddmode) {
				/* fBA^Cv(2D) != hCu[h(2DD) */
				p[4] = (BYTE)(fdc_track[fdc_drvreg] >> 1);
			}
			else {
				/* fBA^Cv == hCu[h ␳Ȃ */
				p[4] = fdc_track[fdc_drvreg];
			}
#else
			p[4] = fdc_track[fdc_drvreg];
#endif
			p[5] = fdc_sidereg;
			p[6] = (BYTE)(j + 1);
			p[7] = fdc_header[fdc_drvreg][p[4] * 6 + 4];
			if (p[7] == 0) {
				size = 128;
			}
			else {
				size = (WORD)(p[7] * 256);
			}
			ddm = FALSE;
		}
		else {
			memcpy(&p[4], q, 4);
			size = (WORD)(q[0x000f] * 256 + q[0x000e]);
			if (q[0x0007] != 0x00) {
				ddm = TRUE;
			}
			else {
				ddm = FALSE;
			}
			q += 0x0010;
		}
		calc_set_crc(p, 8);
		p += (4 + 4 + 2);
		count += (WORD)(4 + 4 + 2);

		/* GAP2 */
		for (i=0; i<22; i++) {
			*p++ = 0x4e;
		}
		count += (WORD)22;

		/* SYNC */
		for (i=0; i<12; i++) {
			*p++ = 0;
		}
		count += (WORD)12;

		/* DATA ADDRESS MARK */
		p[0] = 0xa1;
		p[1] = 0xa1;
		p[2] = 0xa1;
		if (ddm) {
			/* DELETED DATA MARK */
			p[3] = 0xf8;
		}
		else {
			/* DATA MARK */
			p[3] = 0xfb;
		}

		/* f[^ */
		memcpy(p + 4, q, size);
		q += size;
		calc_set_crc(p, 4 + size);
		p += (4 + size + 2);
		count += (WORD)(4 + size + 2);

		/* GAP3 */
		for (i=0; i<gap3; i++) {
			*p++ = 0x4e;
		}
		count += (WORD)gap3;
	}

	/* GAP4 */
	j = (0x1800 - count);
	if (j < 0x1800) {
		for (i=0; i<j; i++) {
			*p++ = 0x4e;
		}
		count += (WORD)j;
	}

	/* f[^|C^AJE^ݒ */
	fdc_dataptr = fdc_buffer;
	fdc_totalcnt = count;
	fdc_onetime = 32;
	fdc_nowcnt = 0;
}

/*
 *	IDtB[hobt@ɍ
 *	JE^Af[^|C^ݒ
 */
static void FASTCALL fdc_makeaddr(int index)
{
	int i;
	BYTE *p;
	WORD offset;
	WORD size;

	/* ]̂߂̃e|obt@́Aʏobt@ */
	/* ŌׂĐ݂ */
	p = &fdc_buffer[0x1ff0];

	/* ID ADDRESS MARK */
	p[0] = 0xa1;
	p[1] = 0xa1;
	p[2] = 0xa1;
	p[3] = 0xfe;

	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
#if XM7_VER >= 3
		/* 2D/2DD̏ꍇAC,H,R,N͊m肷 */
		if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) && fdc_2ddmode) {
			/* fBA^Cv(2D) != hCu[h(2DD) */
			p[4] = (BYTE)(fdc_track[fdc_drvreg] >> 1);
		}
		else if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) && !fdc_2ddmode) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			p[4] = (BYTE)(fdc_track[fdc_drvreg] << 1);
		}
		else {
			/* fBA^Cv == hCu[h ␳Ȃ */
			p[4] = fdc_track[fdc_drvreg];
		}
#else
		if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			p[4] = (BYTE)(fdc_track[fdc_drvreg] << 1);
		}
		else {
			/* fBA^Cv == hCu[h ␳Ȃ */
			p[4] = fdc_track[fdc_drvreg];
		}
#endif
		p[5] = fdc_sidereg;
		p[6] = (BYTE)(index + 1);
		p[7] = 1;
	}
	else if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
		/* VFD̏ꍇAC,H,R͊mAN̓wb_擾 */
#if XM7_VER >= 2
		if (fdc_2ddmode) {
			/* fBA^Cv(2D) != hCu[h(2DD) */
			p[4] = (BYTE)(fdc_track[fdc_drvreg] >> 1);
		}
		else {
			/* fBA^Cv == hCu[h ␳Ȃ */
			p[4] = fdc_track[fdc_drvreg];
		}
#else
		p[4] = fdc_track[fdc_drvreg];
#endif
		p[5] = fdc_sidereg;
		p[6] = (BYTE)(index + 1);
		p[7] = fdc_header[fdc_drvreg][p[4] * 6 + 4];
	}
	else {
		/* obt@ړĨZN^܂Ői߂ */
		offset = 0;
		i = 0;
		while (i < index) {
			/* ZN^TCY擾 */
			size = fdc_buffer[offset + 0x000f];
			size *= (WORD)256;
			size |= fdc_buffer[offset + 0x000e];

			/* ̃ZN^֐i߂ */
			offset += size;
			offset += (WORD)0x10;

			i++;
		}

		/* C,H,R,NRs[ */
		memcpy(&p[4], &fdc_buffer[offset], 4);
	}

	/* CRC */
	calc_set_crc(p, 8);

	/* f[^|C^AJE^ݒ */
	fdc_dataptr = &fdc_buffer[0x1ff0 + 4];
	fdc_totalcnt = 6;
	fdc_onetime = 32;
	fdc_nowcnt = 0;

	/* ZN^WX^ɃgbNԍݒ */
	fdc_secreg = p[4];
}

/*
 *	CfbNXJE^̃ZN^ֈڂ
 */
static int FASTCALL fdc_next_index(void)
{
	int max_track;
	int track;
	int secs;

	ASSERT(fdc_ready[fdc_drvreg] != FDC_TYPE_NOTREADY);

	/* fBXN^Cv`FbN */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
		/* 2D/2DDȂ16ZN^Œ */
		fdc_indexcnt = (BYTE)((fdc_indexcnt + 1) & 0x0f);
		track = fdc_track[fdc_drvreg];
#if XM7_VER >= 2
		if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) && fdc_2ddmode) {
			/* fBA^Cv(2D) && hCu[h(2DD) */
			track >>= 1;
		}
		else if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) && !fdc_2ddmode) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#else
		if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#endif

#if XM7_VER >= 3
		if (fdc_2ddmode) {
			max_track = 80;
		}
		else {
			max_track = 40;
		}
#else
		max_track = 40;
#endif
		if (track >= max_track) {
			return -1;
		}
		return fdc_indexcnt;
	}

	/* D77/VFD */
	ASSERT( (fdc_ready[fdc_drvreg] == FDC_TYPE_D77) ||
			(fdc_ready[fdc_drvreg] == FDC_TYPE_VFD));

	if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
		track = fdc_track[fdc_drvreg];
#if XM7_VER >= 2
		if (fdc_2ddmode) {
			track >>= 1;
		}
#endif
		secs = fdc_header[fdc_drvreg][track * 6 + 5];
	}
	else if (fdc_ready[fdc_drvreg] == FDC_TYPE_D77) {
		secs = fdc_buffer[0x0005];
		secs *= 256;
		secs |= fdc_buffer[0x0004];
	}
	if (secs == 0) {
		/* AtH[}bg */
		fdc_indexcnt = (BYTE)((fdc_indexcnt + 1) & 0x0f);
		return -1;
	}
	else {
		fdc_indexcnt++;
		if (fdc_indexcnt >= secs) {
			fdc_indexcnt = 0;
		}
		return fdc_indexcnt;
	}
}

/*
 *	ID}[NT
 */
static BOOL FASTCALL fdc_idmark(WORD *p)
{
	WORD offset;
	BYTE dat;

	/* A1 A1 A1 FEF5 F5 F5 FE */
	offset = *p;

	while (offset < fdc_totalcnt) {
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if (dat == 0xfe) {
			*p = offset;
			return TRUE;
		}
	}

	*p = offset;
	return FALSE;
}

/*
 *	f[^}[NT
 */
static BOOL FASTCALL fdc_datamark(WORD *p, BOOL *deleted_mark)
{
	WORD offset;
	BYTE dat;

	/* data mark : A1 A1 A1 FBF5 F5 F5 FB */
	/* deleted data mark : A1 A1 A1 F8F5 F5 F5 F8 */
	offset = *p;

	while (offset < fdc_totalcnt) {
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if ((dat != 0xa1) && (dat != 0xf5)) {
			continue;
		}
		dat = fdc_buffer[offset++];
		if ((dat == 0xfb) || (dat == 0xf8)) {
			*p = offset;
			if (dat == 0xfb) {
				/* 0xFB : DATA MARK */
				*deleted_mark = FALSE;
			}
			else {
				/* 0xF8 : DELETED DATA MARK */
				*deleted_mark = TRUE;
			}
			return TRUE;
		}
	}

	*p = offset;
	return FALSE;
}

/*
 *	gbNݏI
 */
static BOOL FASTCALL fdc_writetrk(void)
{
	int total;
	int sectors;
	WORD offset;
	WORD seclen;
	WORD writep;
	int i;
	int handle;
	BOOL ddm;

	/*  */
	total = 0;
	sectors = 0;

	/* ZN^ƁAf[^̃g[^TCY𐔂 */
	offset = 0;
	while (offset < fdc_totalcnt) {
		/* ID}[NT */
		if (!fdc_idmark(&offset)) {
			break;
		}
		/* C,H,R,N̎$F7 */
		offset += (WORD)4;
		if (offset >= fdc_totalcnt) {
			return FALSE;
		}
		if (fdc_buffer[offset] != 0xf7) {
			return FALSE;
		}
		offset++;
		/* f[^}[NT */
		if (!fdc_datamark(&offset, &ddm)) {
			return FALSE;
		}
		/* OXvZA$F7T */
		seclen = 0;
		while(offset < fdc_totalcnt) {
			if (fdc_buffer[offset] == 0xf7) {
				break;
			}
			offset++;
			seclen++;
		}
		if (offset >= fdc_totalcnt) {
			return FALSE;
		}
		/* ZN^ok */
		total += seclen;
		sectors++;
		offset++;
	}

	/* 2D/2DDfBȀꍇAtotal=0x1000, sectors=0x10K{ */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
		/* total, sectorš */
		if (total != 0x1000) {
			return FALSE;
		}
		if (sectors != 16) {
			return FALSE;
		}

		/* {C,H,NȂǃ`FbNׂ */
		return TRUE;
	}

	/* VFDfBȀꍇA܂ */
	if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
		/* {C,H,NȂǃ`FbNׂ */
		return TRUE;
	}

	/* ZN^Ƃ0x10ԂA]vɂ */
	if ((DWORD)(total + (sectors * 0x10)) > fdc_trklen[fdc_drvreg]) {
		return FALSE;
	}

	/* ޕKvȂꍇ͐I */
	if ((sectors == 0) && (total == 0)) {
		return TRUE;
	}

	/* ܂邱Ƃ킩̂ŁAf[^쐬 */
	writep = 0;
	offset = 0;
	for (i=0; i<sectors; i++) {
		/* ID}[N */
		fdc_idmark(&offset);

		/* 0x10wb_̍쐬 */
		fdc_buffer[writep++] = fdc_buffer[offset++];
		fdc_buffer[writep++] = fdc_buffer[offset++];
		fdc_buffer[writep++] = fdc_buffer[offset++];
		fdc_buffer[writep++] = fdc_buffer[offset++];
		fdc_buffer[writep++] = (BYTE)(sectors & 0xff);
		fdc_buffer[writep++] = (BYTE)(sectors >> 8);
		fdc_buffer[writep++] = 0x00;
		offset++;

		/* f[^}[N̏ */
		fdc_datamark(&offset, &ddm);
		if (ddm) {
			fdc_buffer[writep++] = 0x10;
		}
		else {
			fdc_buffer[writep++] = 0x00;
		}

		/* U[uGANA */
		/* ZN^OX͌ŏ */
		memset(&fdc_buffer[writep], 0, 8);
		writep += (WORD)8;

		/* OX𐔂Rs[ */
		seclen = 0;
		while (fdc_buffer[offset] != 0xf7) {
			fdc_buffer[writep++] = fdc_buffer[offset++];
			seclen++;
		}
		offset++;

		/* ZN^OXݒ */
		fdc_buffer[writep - seclen - 2] = (BYTE)(seclen & 0xff);
		fdc_buffer[writep - seclen - 1] = (BYTE)(seclen >> 8);
	}

	/* t@CɃf[^ */
	handle = file_open(fdc_fname[fdc_drvreg], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, fdc_seekofs[fdc_drvreg])) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, fdc_buffer, (sectors * 0x10) + total)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* I */
	return TRUE;
}

/*
 *	ZN^ݏI
 */
static BOOL FASTCALL fdc_writesec(void)
{
	DWORD offset;
	int handle;

	/* assert */
	ASSERT(fdc_drvreg < FDC_DRIVES);
	ASSERT(fdc_ready[fdc_drvreg] != FDC_TYPE_NOTREADY);
	ASSERT(fdc_dataptr);
	ASSERT(fdc_totalcnt > 0);

	/* ItZbgZo */
	offset = fdc_seekofs[fdc_drvreg];
	offset += fdc_secofs[fdc_drvreg];

	/*  */
	handle = file_open(fdc_fname[fdc_drvreg], OPEN_RW);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_write(handle, fdc_dataptr, fdc_totalcnt)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	gbNAZN^ATChƈvZN^
 *	JE^Af[^|C^ݒ
 */
static BYTE FASTCALL fdc_readsec(
	BYTE track, BYTE sector, BYTE side, BOOL sidecmp, int *secidx, int cpos)
{
	int secs;
	int	len;
	int i;
	WORD offset;
	WORD size;
	BYTE stat;
	DWORD vfdoffset;
	int fdctrack;

	/* assert */
	ASSERT(fdc_drvreg < FDC_DRIVES);
	ASSERT(fdc_ready[fdc_drvreg] != FDC_TYPE_NOTREADY);
	UNUSED(cpos);

	/* wb_̂gbNԍ(rp)ݒ */
	fdctrack = fdc_track[fdc_drvreg];

	/* 2D/2DDt@C̏ꍇ */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
#if XM7_VER >= 3
		if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) && fdc_2ddmode) {
			/* fBA^Cv(2D) != hCu[h(2DD) */
			if ((fdctrack % 2) == 1) {
				/* 2DC[W2DDANZXAgbNANZX̓G[ */
				return FDC_ST_RECNFND;
			}
			fdctrack = (BYTE)(fdctrack >> 1);
		}
		else if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) && !fdc_2ddmode) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			fdctrack = (BYTE)(fdctrack << 1);
		}
#else
		if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) != hCu[h(2D) */
			fdctrack = (BYTE)(fdctrack << 1);
		}
#endif
		if (track != fdctrack) {
			return FDC_ST_RECNFND;
		}
		if (side != fdc_sidereg) {
			return FDC_ST_RECNFND;
		}
		if ((sector < 1) || (sector > 16)) {
			return FDC_ST_RECNFND;
		}

		/* f[^|C^ݒ */
		fdc_dataptr = &fdc_buffer[(sector - 1) * 0x0100];
		fdc_secofs[fdc_drvreg] = (sector - 1) * 0x0100;

		/* JE^ݒ */
		fdc_totalcnt = 0x0100;
		fdc_onetime = 32;
		fdc_nowcnt = 0;

		/* ZN^̃CfbNXԂ */
		*secidx = sector - 1;

		return 0;
	}

	/* VFDt@C̏ꍇ */
	if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
#if XM7_VER >= 2
		if (fdc_2ddmode) {
			/* fBA^Cv(2D) != hCu[h(2DD) */
			if ((fdctrack % 2) == 1) {
				/* 2DC[W2DDANZXAgbNANZX̓G[ */
				return FDC_ST_RECNFND;
			}
			fdctrack = (BYTE)(fdctrack >> 1);
		}
#endif
		if (track != fdctrack) {
			return FDC_ST_RECNFND;
		}
		if (side != fdc_sidereg) {
			return FDC_ST_RECNFND;
		}

		/* gbNւ̃ItZbg`FbN */
		vfdoffset  = fdc_header[fdc_drvreg][track * 6 + 3];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[fdc_drvreg][track * 6 + 2];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[fdc_drvreg][track * 6 + 1];
		vfdoffset *= (DWORD)256;
		vfdoffset |= fdc_header[fdc_drvreg][track * 6 + 0];
		if (vfdoffset == 0) {
			return FDC_ST_RECNFND;
		}

		/* gbÑZN^`FbN */
		len = fdc_header[fdc_drvreg][track * 6 + 4];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}
		size = fdc_header[fdc_drvreg][track * 6 + 5];
		if ((sector < 1) || (sector > size)) {
			return FDC_ST_RECNFND;
		}

		/* f[^|C^ݒ */
		fdc_dataptr = &fdc_buffer[(sector - 1) * len];
		fdc_secofs[fdc_drvreg] = (sector - 1) * len;

		/* JE^ݒ */
		fdc_totalcnt = (WORD)len;
		fdc_onetime = 32;
		fdc_nowcnt = 0;

		/* ZN^̃CfbNXԂ */
		*secidx = sector - 1;

		return 0;
	}

	/* D77t@C̏ꍇ */
	secs = fdc_buffer[0x0005];
	secs *= (WORD)256;
	secs |= fdc_buffer[0x0004];
	if (secs == 0) {
		return FDC_ST_RECNFND;
	}

#if XM7_VER >= 2
	/* 2DC[W2DDANZXAgbNANZX̓G[ */
	if ((fdc_header[fdc_drvreg][0x001b] == 0x00) && fdc_2ddmode) {
		if ((fdctrack % 2) == 1) {
			return FDC_ST_RECNFND;
		}
	}
#endif

	offset = 0;
	/* ZN^[v */
	for (i=0; i<secs; i++) {
		/* ̃ZN^̃TCYɎ擾 */
		size = fdc_buffer[offset + 0x000f];
		size *= (WORD)256;
		size |= fdc_buffer[offset + 0x000e];

		/* C,H,RvZN^邩 */
		if (fdc_buffer[offset + 0] != track) {
			offset += size;
			offset += (WORD)0x10;
			continue;
		}
		/* TChbit0ȊO͍lȂ悤ɕύX */
		if (sidecmp) {
			if ((BYTE)(fdc_buffer[offset + 1] & 1) != side) {
				offset += size;
				offset += (WORD)0x10;
				continue;
			}
		}
		if (fdc_buffer[offset + 2] != sector) {
			offset += size;
			offset += (WORD)0x10;
			continue;
		}

		/* P`FbN */
		if (fdc_buffer[offset + 0x0006] != 0) {
			continue;
		}

		/* f[^|C^AJE^ݒ */
		fdc_dataptr = &fdc_buffer[offset + 0x0010];
		fdc_secofs[fdc_drvreg] = offset + 0x0010;
		fdc_totalcnt = size;
		fdc_onetime = 32;
		fdc_nowcnt = 0;

		/* Xe[^Xݒ */
		stat = 0;
		if (fdc_buffer[offset + 0x0007] != 0) {
			stat |= FDC_ST_RECTYPE;
		}
		if (fdc_buffer[offset + 0x0008] == 0xb0) {
			/* Oɐݒ肵Xe[^XɂȂC (ݼ޼) */
			stat |= FDC_ST_CRCERR;
		}

		/* ZN^̃CfbNXԂ */
		*secidx = i;

		return stat;
	}

	/* SZN^AȂ */
	return FDC_ST_RECNFND;
}

/*
 *	gbNǂݍ
 */
static void FASTCALL fdc_readbuf(int drive)
{
	DWORD offset;
	DWORD min_offset;
	DWORD len;
	DWORD secs;
	int trkside;
	int handle;
	int max_track;
	int trk;

	/* C[W͖ */
	fdi_init = FALSE;

	/* hCu`FbN */
	if ((drive < 0) || (drive >= FDC_DRIVES)) {
		LOG1(Warning, "drive:%d\n", drive);
		return;
	}

	/* fB`FbN */
	if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
		return;
	}

	/* CfbNXz[JE^NA */
	fdc_indexcnt = 0;

	/* gbN~Q{TCh */
	trkside = fdc_track[drive] * 2 + fdc_sidereg;

	/*
	 * 2D/2DDt@C
	 */
	if ((fdc_ready[drive] == FDC_TYPE_2D) ||
		(fdc_ready[drive] == FDC_TYPE_2DD)) {
		if (fdc_ready[drive] == FDC_TYPE_2D) {
			max_track = 80;

#if XM7_VER >= 2
			/* 2Dt@C2DDǂݏȍꍇ̕␳ */
			if (fdc_2ddmode) {
				trkside = (fdc_track[drive] >> 1) * 2 + fdc_sidereg;
			}
#endif
		}
		else {
			max_track = 160;

#if XM7_VER >= 2
			/* 2DDt@C2Dǂݏȍꍇ̕␳ */
			if (!fdc_2ddmode) {
				trkside = fdc_track[drive] * 4 + fdc_sidereg;
			}
#else
			trkside = fdc_track[drive] * 4 + fdc_sidereg;
#endif
		}

		/* I[o[`FbN */
		if (trkside >= max_track) {
			memset(fdc_buffer, 0, 0x1000);
			return;
		}

		/* ItZbgZo */
		offset = (DWORD)trkside;
		offset *= 0x1000;
		fdc_seekofs[drive] = offset;
		fdc_trklen[drive] = 0x1000;

		/* ǂݍ */
		memset(fdc_buffer, 0, 0x1000);
		handle = file_open(fdc_fname[drive], OPEN_R);
		if (handle == -1) {
			return;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return;
		}
		file_read(handle, fdc_buffer, 0x1000);
		file_close(handle);
		return;
	}

	/*
	 * VFDt@C
	 */
	if (fdc_ready[drive] == FDC_TYPE_VFD) {
#if XM7_VER >= 2
		/* 2DD[h̏ꍇA␳ */
		if (fdc_2ddmode) {
			trkside = (fdc_track[drive] >> 1) * 2 + fdc_sidereg;
		}
#endif

		/* ͈̓`FbN */
		if (trkside >= 80) {
			memset(fdc_buffer, 0, 0x1000);
			return;
		}

		/* ItZbgZo */
		offset  = fdc_header[drive][trkside * 6 + 3];
		offset *= 256;
		offset |= fdc_header[drive][trkside * 6 + 2];
		offset *= 256;
		offset |= fdc_header[drive][trkside * 6 + 1];
		offset *= 256;
		offset |= fdc_header[drive][trkside * 6 + 0];
		fdc_seekofs[drive] = offset;

		len		= fdc_header[drive][trkside * 6 + 4];
		secs	= fdc_header[drive][trkside * 6 + 5];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}
		fdc_trklen[drive] = len * secs;

		/* ǂݍ */
		memset(fdc_buffer, 0, fdc_trklen[drive]);
		handle = file_open(fdc_fname[drive], OPEN_R);
		if (handle == -1) {
			return;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return;
		}
		file_read(handle, fdc_buffer, fdc_trklen[drive]);
		file_close(handle);
		return;
	}

	/*
	 * D77t@C
	 */
	if (fdc_header[drive][0x001b] == 0x00) {
		max_track = 84;
	}
	else {
		max_track = 164;
	}

	/* ANZXĂŏIgbN̎Zo */
	min_offset = (0x0020 + max_track * 4);
	for (trk = 0; trk < max_track; trk++) {
		if ((DWORD)(0x0020 + trk * 4) >= min_offset) {
			break;
		}

		offset  = fdc_header[drive][0x0020 + trk * 4 + 3];
		offset *= 256;
		offset |= fdc_header[drive][0x0020 + trk * 4 + 2];
		offset *= 256;
		offset |= fdc_header[drive][0x0020 + trk * 4 + 1];
		offset *= 256;
		offset |= fdc_header[drive][0x0020 + trk * 4 + 0];
		if ((offset < min_offset) && (offset >= 0x20)) {
			min_offset = offset;
		}
	}
	max_track = (min_offset - 0x20) / 4;

#if XM7_VER >= 2
	if (fdc_2ddmode) {
		/* 2Dt@C2DDǂ݂ꍇAŕ␳ */
		if (fdc_header[drive][0x001b] == 0x00) {
			if ((fdc_track[drive] % 2) == 1) {
				/* gbNANZXAZN^0 */
				fdc_buffer[4] = 0;
				fdc_buffer[5] = 0;
				return;
			}
			trkside = (fdc_track[drive] >> 1) * 2 + fdc_sidereg;
		}
	}
	else {
		/* 2DDt@C2Dǂ݂ꍇAŕ␳ (AV40pOS-9) */
		if (fdc_header[drive][0x001b] == 0x10) {
			trkside = fdc_track[drive] * 4 + fdc_sidereg;
		}
	}
#else
	/* 2DDt@C2Dǂ݂ꍇAŕ␳ (AV40pOS-9) */
	if (fdc_header[drive][0x001b] == 0x10) {
		trkside = fdc_track[drive] * 4 + fdc_sidereg;
	}
#endif

	/* I[o[`FbN */
	if (trkside >= max_track) {
		/* gbNI[o[AZN^0 */
		fdc_buffer[4] = 0;
		fdc_buffer[5] = 0;
		return;
	}

	/* wb_ɏ]AItZbgEOXZo */
	offset = *(DWORD *)(&fdc_header[drive][0x0020 + trkside * 4]);
	if (offset == 0) {
		/* ݂ȂgbN */
		fdc_buffer[4] = 0;
		fdc_buffer[5] = 0;
		return;
	}

	len = *(DWORD *)(&fdc_header[drive][0x0020 + (trkside + 1) * 4]);
	if ((len == 0) || ((trkside + 1) >= max_track)) {
		/* ŏIgbN */
		len = *(DWORD *)(&fdc_header[drive][0x0014]);
	}
	len -= offset;
	if (len > 0x2000) {
		len = 0x2000;
	}
	fdc_seekofs[drive] = offset;
	fdc_trklen[drive] = len;

	/* V[NAǂݍ */
	memset(fdc_buffer, 0, 0x2000);
	handle = file_open(fdc_fname[drive], OPEN_R);
	if (handle == -1) {
		return;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return;
	}
	file_read(handle, fdc_buffer, len);
	file_close(handle);
}

/*
 *	D77t@C wb_ǂݍ
 */
static BOOL FASTCALL fdc_readhead(int drive, int index)
{
	int i;
	DWORD offset;
	DWORD temp;
	int handle;

	/* assert */
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((index >= 0) && (index < FDC_MEDIAS));
	ASSERT(fdc_ready[drive] == FDC_TYPE_D77);

	/* ItZbg */
	offset = fdc_foffset[drive][index];

	/* V[NAǂݍ */
	handle = file_open(fdc_fname[drive], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, offset)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, fdc_header[drive], 0x2b0)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	/* ^Cv`FbNACgveNgݒ */
	if ((fdc_header[drive][0x001b] != 0x00) &&
		(fdc_header[drive][0x001b] != 0x10)) {
		/* 2D/2DDłȂ */
		return FALSE;
	}
	if (fdc_fwritep[drive]) {
		fdc_writep[drive] = TRUE;
	}
	else {
		if (fdc_header[drive][0x001a] & 0x10) {
			fdc_writep[drive] = TRUE;
		}
		else {
			fdc_writep[drive] = FALSE;
		}
	}

	/* gbNItZbgݒ */
	for (i=0; i<164; i++) {
		temp = 0;
		temp |= fdc_header[drive][0x0020 + i * 4 + 3];
		temp *= 256;
		temp |= fdc_header[drive][0x0020 + i * 4 + 2];
		temp *= 256;
		temp |= fdc_header[drive][0x0020 + i * 4 + 1];
		temp *= 256;
		temp |= fdc_header[drive][0x0020 + i * 4 + 0];

		if (temp != 0) {
			/* f[^ */
			temp += offset;
			*(DWORD *)(&fdc_header[drive][0x0020 + i * 4]) = temp;
		}
	}

	/* fBXNTCY{ItZbg */
	temp = 0;
	temp |= fdc_header[drive][0x001c + 3];
	temp *= 256;
	temp |= fdc_header[drive][0x001c + 2];
	temp *= 256;
	temp |= fdc_header[drive][0x001c + 1];
	temp *= 256;
	temp |= fdc_header[drive][0x001c + 0];
	temp += offset;
	*(DWORD *)(&fdc_header[drive][0x0014]) = temp;

	return TRUE;
}

/*
 *	VFDt@C wb_ǂݍ
 */
static BOOL FASTCALL fdc_readhead_vfd(int drive)
{
	int handle;

	/* assert */
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT(fdc_ready[drive] == FDC_TYPE_VFD);

	/* V[NAǂݍ */
	handle = file_open(fdc_fname[drive], OPEN_R);
	if (handle == -1) {
		return FALSE;
	}
	if (!file_seek(handle, 0)) {
		file_close(handle);
		return FALSE;
	}
	if (!file_read(handle, fdc_header[drive], 0x1e0)) {
		file_close(handle);
		return FALSE;
	}
	file_close(handle);

	return TRUE;
}

/*
 *	݂̃fBÃCgveNg؂ւ
 */
BOOL FASTCALL fdc_setwritep(int drive, BOOL writep)
{
	BYTE header[0x2b0];
	DWORD offset;
	int handle;

	/* assert */
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((writep == TRUE) || (writep == FALSE));

	/* fBłȂ΂ȂȂ */
	if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* t@CݕsȂ_ */
	if (fdc_fwritep[drive]) {
		return FALSE;
	}

	/* ǂݍ݁AݒA */
	if (fdc_ready[drive] == FDC_TYPE_D77) {
		offset = fdc_foffset[drive][fdc_media[drive]];
		handle = file_open(fdc_fname[drive], OPEN_RW);
		if (handle == -1) {
			return FALSE;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_read(handle, header, 0x2b0)) {
			file_close(handle);
			return FALSE;
		}
		if (writep) {
			header[0x1a] |= 0x10;
		}
		else {
			header[0x1a] &= ~0x10;
		}
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return FALSE;
		}
		if (!file_write(handle, header, 0x2b0)) {
			file_close(handle);
			return FALSE;
		}
	}

	/*  */
	file_close(handle);
	fdc_writep[drive] = writep;
	return TRUE;
}

/*
 *	fBAԍݒ
 */
BOOL FASTCALL fdc_setmedia(int drive, int index)
{
	/* assert */
	ASSERT((drive >= 0) && (drive < FDC_DRIVES));
	ASSERT((index >= 0) && (index < FDC_MEDIAS));

	/* fBԂ */
	if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
		return FALSE;
	}

	/* 2D/2DDt@C̏ꍇAindex = 0 */
	if (((fdc_ready[drive] == FDC_TYPE_2D) ||
		 (fdc_ready[drive] == FDC_TYPE_2DD)) && (index != 0)) {
		return FALSE;
	}

	/* VFDt@C̏ꍇAindex = 0`FbNăwb_ǂݍ */
	if (fdc_ready[drive] == FDC_TYPE_VFD) {
		if (index != 0) {
			return FALSE;
		}
		if (!fdc_readhead_vfd(drive)) {
			return FALSE;
		}
	}

	/* index > 0 ȂAfdc_foffset𒲂ׂ>0Kv */
	if (index > 0) {
		if (fdc_foffset[drive][index] == 0) {
			return FALSE;
		}
	}

	/* D77t@C̏ꍇAwb_ǂݍ */
	if (fdc_ready[drive] == FDC_TYPE_D77) {
		/* CgveNg͓Őݒ */
		if (!fdc_readhead(drive, index)) {
			return FALSE;
		}
	}
	else {
		/* 2D/2DD/VFDt@CȂAt@Cɏ] */
		fdc_writep[drive] = fdc_fwritep[drive];
	}

	/* fBAꂽꍇAꎞCWFNg */
	if (fdc_media[drive] != index) {
		fdc_teject[drive] = FALSE;
	}

	/* f[^obt@ǂݍ݁A[NZ[u */
	fdc_media[drive] = (BYTE)index;
	if (drive == fdc_drvreg) {
		/* JghCuƈvꍇ̂݃obt@ǂݍ݂s */
		fdc_readbuf(drive);
	}

	return TRUE;
}

/*
 *	D77t@ĆAfBAі̎擾
 */
static int FASTCALL fdc_chkd77(int drive)
{
	int i;
	int handle;
	int count;
	int max_track;
	int trk;
	DWORD offset;
	DWORD len;
	DWORD trkoffset;
	DWORD min_offset;
	BYTE buf[0x2b0];

	/*  */
	for (i=0; i<FDC_MEDIAS; i++) {
		fdc_foffset[drive][i] = 0;
		fdc_name[drive][i][0] = '\0';
	}
	count = 0;
	offset = 0;

	/* t@CI[v */
	handle = file_open(fdc_fname[drive], OPEN_R);
	if (handle == -1) {
		return count;
	}

	/* fBA[v */
	while (count < FDC_MEDIAS) {
		/* V[N */
		if (!file_seek(handle, offset)) {
			file_close(handle);
			return count;
		}

		/* ǂݍ */
		if (!file_read(handle, buf, 0x02b0)) {
			file_close(handle);
			return count;
		}

		/* ^Cv`FbNB2D,2DD̂ݑΉ */
		if ((buf[0x1b] != 0x00) && (buf[0x1b] != 0x10)) {
			file_close(handle);
			return count;
		}
		if (buf[0x1b] == 0x00) {
			max_track = 84;
		}
		else {
			max_track = 164;
		}

		/* ANZXĂŏIgbN̎Zo */
		min_offset = (0x0020 + max_track * 4);
		for (trk = 0; trk < max_track; trk++) {
			if ((DWORD)(0x0020 + trk * 4) >= min_offset) {
				break;
			}

			trkoffset = buf[0x20 + trk * 4 + 3];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 2];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 1];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 0];
			if ((trkoffset < min_offset) && (trkoffset >= 0x20)) {
				min_offset = trkoffset;
			}
		}
		max_track = (min_offset - 0x20) / 4;

		/* ok,t@CAItZbgi[ */
		buf[0x10] = '\0';
		memcpy(fdc_name[drive][count], buf, 17);
		fdc_foffset[drive][count] = offset;

		/* next */
		len = 0;
		len |= buf[0x1f];
		len *= 256;
		len |= buf[0x1e];
		len *= 256;
		len |= buf[0x1d];
		len *= 256;
		len |= buf[0x1c];
		if (len == 0){
			file_close(handle);
			return count;
		}
		for (trk = 0; trk < max_track; trk++) {
			trkoffset = buf[0x20 + trk * 4 + 3];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 2];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 1];
			trkoffset *= 256;
			trkoffset |= buf[0x20 + trk * 4 + 0];
			if (trkoffset >= len) {
				file_close(handle);
				return count;
			}
		}
		offset += len;
		count++;
	}

	/* ő僁fBAɒB */
	file_close(handle);
	return count;
}

/*
 *	VFDt@C`FbN
 */
static int FASTCALL fdc_chkvfd(int drive)
{
	int	i;
	int handle;
	BYTE buf[0x20];

	/*  */
	for (i=0; i<FDC_MEDIAS; i++) {
		fdc_foffset[drive][i] = 0;
		fdc_name[drive][i][0] = '\0';
	}

	/* t@CI[v */
	handle = file_open(fdc_fname[drive], OPEN_R);
	if (handle == -1) {
		return 0;
	}

	/* wb_̃`FbN */
	if (!file_seek(handle, 0)) {
		file_close(handle);
		return 0;
	}
	if (!file_read(handle, buf, 0x0020)) {
		file_close(handle);
		return 0;
	}
	file_close(handle);

	if (	(buf[0x00] != 0xe0) || (buf[0x01] != 0x01) ||
			(buf[0x02] != 0x00) || (buf[0x03] != 0x00)) {
		/* gbN0̃ItZbg0x01e0ȊOȂVFDł͂Ȃ(蔲c) */
		return 0;
	}

	return 1;
}

/*
 *	fBXNt@Cݒ
 */
int FASTCALL fdc_setdisk(int drive, char *fname)
{
	BOOL writep;
	int handle;
	DWORD fsize;
	int count;
	BOOL flag;
	int i;

	ASSERT((drive >= 0) && (drive < FDC_DRIVES));

	/* mbgfBɂꍇ */
	if (fname == NULL) {
		fdc_ready[drive] = FDC_TYPE_NOTREADY;
		fdc_fname[drive][0] = '\0';
		return 1;
	}

	/* t@CI[vAt@CTCY𒲂ׂ */
	if (strlen(fname) >= sizeof(fdc_fname[drive])) {
		fdc_ready[drive] = FDC_TYPE_NOTREADY;
		fdc_fname[drive][0] = '\0';
		return 1;
	}
	strcpy(fdc_fname[drive], fname);
	writep = FALSE;
	handle = file_open(fdc_fname[drive], OPEN_RW);
	if (handle == -1) {
		handle = file_open(fdc_fname[drive], OPEN_R);
		if (handle == -1) {
			fdc_ready[drive] = FDC_TYPE_NOTREADY;
			return 0;
		}
		writep = TRUE;
	}
	fsize = file_getsize(handle);
	file_close(handle);

	/*
	 * 2Dt@C
	 */
	if (fsize == 327680) {
		/* ^CvAݑݒ */
		fdc_ready[drive] = FDC_TYPE_2D;
		fdc_fwritep[drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(drive, 0)) {
			fdc_ready[drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[drive] = FALSE;
		fdc_medias[drive] = 1;
		return 1;
	}

	/*
	 * 2DDt@C
	 */
	if (fsize == 655360) {
		/* ^CvAݑݒ */
		fdc_ready[drive] = FDC_TYPE_2DD;
		fdc_fwritep[drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(drive, 0)) {
			fdc_ready[drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[drive] = FALSE;
		fdc_medias[drive] = 1;
		return 1;
	}

	/*
	 * VFDt@C
	 */
	/* t@C */
	if (fdc_chkvfd(drive) != 0) {
		fdc_ready[drive] = FDC_TYPE_VFD;
		fdc_fwritep[drive] = writep;

		/* fBAݒ */
		if (!fdc_setmedia(drive, 0)) {
			fdc_ready[drive] = FDC_TYPE_NOTREADY;
			return 0;
		}

		/* BꎞCWFNg */
		fdc_teject[drive] = FALSE;
		fdc_medias[drive] = 1;
		return 1;
	}

	/*
	 * D77t@C
	 */
	fdc_ready[drive] = FDC_TYPE_D77;
	fdc_fwritep[drive] = writep;

	/* t@C */
	count = fdc_chkd77(drive);
	if (count == 0){
		fdc_ready[drive] = FDC_TYPE_NOTREADY;
		return 0;
	}

	/* fBAݒ */
	flag = FALSE;
	for (i = 0; i < 16; i++) {
		if (fdc_setmedia(drive, i)) {
			flag = TRUE;
			break;
		}
	}
	if (!flag) {
		fdc_ready[drive] = FDC_TYPE_NOTREADY;
		return 0;
	}

	/* BꎞCWFNg */
	fdc_teject[drive] = FALSE;
	fdc_medias[drive] = (BYTE)count;
	return count;
}

/*-[ FDD Cxg ]-----------------------------------------------------*/

/*
 *	FDD [^[Cxg (WAIT)
 */
static BOOL FASTCALL fdd_motor_event(void)
{
	return TRUE;
}

/*-[ FDCR}h EFCg[h ]---------------------------------------*/

/*
 *	FDC DataRequestCxg (WAIT)
 */
static BOOL FASTCALL fdc_drq_event(void)
{
	/* Xgf[^`FbN */
	if (fdc_drqirq & 0x80) {
		return fdc_lost_event();
	}

	/* f[^Ă̂DRQݒ肵Čp */
	if (fdc_dataptr) {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq |= 0x80;
		return TRUE;
	}

	/* f[^ĂȂ̂RECORD NOT FOUNDŏI */
	fdc_status |= FDC_ST_RECNFND;
	fdc_status &= ~FDC_ST_BUSY;
	fdc_access[fdc_drvreg] = FDC_ACCESS_READY;

	/* 荞ݔ */
	mainetc_fdc();
	fdc_drqirq = 0x40;

	return FALSE;
}

/*
 *	FDC V[NCxg (WAIT)
 */
static BOOL FASTCALL fdd_seek_event(void)
{
	/* V[NI */
	if (fdc_seek_track < 0) {
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;

		/* BUSY/HEAD ENGAGEDtO𗎂Ƃ */
		fdc_status &= ~(FDC_ST_BUSY | FDC_ST_HEADENG);
		fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
		schedule_delevent(EVENT_FDD_SEEK);
	}
	else if (fdc_seek_track-- == 0) {
		/* V[NI̓EFCg} */
		schedule_setevent(EVENT_FDD_SEEK, 20000, fdd_seek_event);
	}
	else {
		/* TEh */
		if (fdc_sound) {
			wav_notify(SOUND_FDDSEEK);
		}
	}

	return TRUE;
}

/*
 *	FDC V[NCxgݒ (SOUND ON)
 */
static void FASTCALL fdc_setseekevent(BYTE track)
{
	/* V[NgbNۑ */
	fdc_seek_track = track;

	/* Cxgݒ */
	if (track > 0) {
		schedule_setevent(EVENT_FDD_SEEK, fdc_steprate[fdc_command & 3], fdd_seek_event);
	}
	else {
		/* wbhړȂĂ΂炭̊ԂBUSYԂɂ */
		schedule_setevent(EVENT_FDD_SEEK, 300, fdd_seek_event);
	}
}

/*-[ FDCR}h ]----------------------------------------------------------*/

/*
 *	[^[]ʒu擾
 */
static int FASTCALL fdd_get_currentpos(void)
{
	event_t *ev;
	DWORD pos;
	DWORD exec;

	/* ݈ʒu */
	ev = schedule_getevent(EVENT_FDD_MOTOR);
	pos = ev->reload - ev->current;

	/* sȂ玞Ԃ𑫂 */
	if (run_flag) {
		exec = (DWORD)maincpu.total;
		exec *= 10000;
		exec /= schedule_get_cycle();
		pos += exec;
		if (pos >= ev->reload) {
			pos -= ev->reload;
		}
	}

	return (int)pos;
}

/*
 *	ZN^ʒuvZ
 */
static void FASTCALL fdc_calc_pos(void)
{
	int i;
	int max_track;
	int track;
	int gap3;
	int len;
	BYTE *p;

	ASSERT(fdc_ready[fdc_drvreg] != FDC_TYPE_NOTREADY);

	/* ς݂ȂXLbv */
	if (fdi_init) {
		return;
	}

	/*  */
	fdi_init = TRUE;
	fdi_secs = 0;

	/* fBXN^Cv`FbN */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) ||
		(fdc_ready[fdc_drvreg] == FDC_TYPE_2DD)) {
		track = fdc_track[fdc_drvreg];
#if XM7_VER >= 3
		if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2D) && fdc_2ddmode) {
			/* fBA^Cv(2D) && hCu[h(2DD) */
			track >>= 1;
		}
		else if ((fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) && !fdc_2ddmode) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#else
		if (fdc_ready[fdc_drvreg] == FDC_TYPE_2DD) {
			/* fBA^Cv(2DD) && hCu[h(2D) */
			track <<= 1;
		}
#endif

#if XM7_VER >= 3
		if (fdc_2ddmode) {
			max_track = 80;
		}
		else {
			max_track = 40;
		}
#else
		max_track = 40;
#endif
		if (track >= max_track) {
			return;
		}

		/* ZN^=16CTCY=256oCgAGAP3=51oCg */
		fdi_secs = 16;
		len = 256;
		gap3 = 51;

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[0] = (80 + 12 + 4 + 50) * 32;
		fdi_dtpos[0] = fdi_idpos[0] + (12 + 4 + 6 + 22) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs; i++) {
			fdi_idpos[i] = fdi_idpos[i - 1] + (12 + 10 + 22 + 12 + 4 + len + 2 + gap3) * 32;
			fdi_dtpos[i] = fdi_idpos[i] + (12 + 4 + 6 + 22) * 32;
		}
		return;
	}

	if (fdc_ready[fdc_drvreg] == FDC_TYPE_VFD) {
		track = fdc_track[fdc_drvreg];
#if XM7_VER >= 3
		if (fdc_2ddmode) {
			track >>= 1;
		}
#endif
		/* ZN^ */
		fdi_secs = fdc_header[fdc_drvreg][track * 6 + 5];
		if (fdi_secs > 96) {
			fdi_secs = 96;
		}

		if (fdi_secs == 0) {
			return;
		}

		/* TCY */
		len = fdc_header[fdc_drvreg][track * 6 + 4];
		if (len == 0) {
			len = 128;
		}
		else {
			len *= 256;
		}

		/* GAP3 */
		if (fdi_secs <= 5) {
			gap3 = 0x74;
		}
		else {
			if (fdi_secs <= 10) {
				gap3 = 0x54;
			}
			else {
				if (fdi_secs <= 16) {
					gap3 = 0x33;
				}
				else {
					gap3 = 0x10;
				}
			}
		}

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[0] = (80 + 12 + 4 + 50) * 32;
		fdi_dtpos[0] = fdi_idpos[0] + (12 + 4 + 6 + 22) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs; i++) {
			fdi_idpos[i] = fdi_idpos[i - 1] + (12 + 10 + 22 + 12 + 4 + len + 2 + gap3) * 32;
			fdi_dtpos[i] = fdi_idpos[i] + (12 + 4 + 6 + 22) * 32;
		}
		return;
	}

	if (fdc_ready[fdc_drvreg] == FDC_TYPE_D77) {
		/* ZN^ */
		p = fdc_buffer;
		fdi_secs = p[0x0005];
		fdi_secs *= 256;
		fdi_secs |= p[0x0004];
		if (fdi_secs > 96) {
			fdi_secs = 96;
		}

		if (fdi_secs == 0) {
			return;
		}

		/* GAP3 */
		if (fdi_secs <= 5) {
			gap3 = 0x74;
		}
		else {
			if (fdi_secs <= 10) {
				gap3 = 0x54;
			}
			else {
				if (fdi_secs <= 16) {
					gap3 = 0x33;
				}
				else {
					gap3 = 0x10;
				}
			}
		}

		/* ŏ̃ZN^ʒu(ms) */
		fdi_idpos[0] = (80 + 12 + 4 + 50) * 32;
		fdi_dtpos[0] = fdi_idpos[0] + (12 + 4 + 6 + 22) * 32;

		/* ZN^ʒuɌvZ */
		for (i = 1; i < fdi_secs; i++) {
			/* TCY */
			len = p[0x000f];
			len *= 256;
			len |= p[0x000e];
			fdi_idpos[i] = fdi_idpos[i - 1] + (12 + 10 + 22 + 12 + 4 + len + 2 + gap3) * 32;
			fdi_dtpos[i] = fdi_idpos[i] + (12 + 4 + 6 + 22) * 32;
			p += len + 0x10;
		}
		return;
	}
}

/*
 *	FDC Xe[^X쐬
 */
static void FASTCALL fdc_make_stat(void)
{
	int pos;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		fdc_status |= FDC_ST_NOTREADY;
		return;
	}

	/* LȃhCu */
	if (fdc_ready[fdc_drvreg] == FDC_TYPE_NOTREADY) {
		fdc_status |= FDC_ST_NOTREADY;
	}
	else {
		fdc_status &= (BYTE)(~FDC_ST_NOTREADY);
	}

	/* ꎞCWFNg */
	if (fdc_teject[fdc_drvreg]) {
		fdc_status |= FDC_ST_NOTREADY;
	}

	/* CgveNg(ReadnR}h0) */
	if ((fdc_cmdtype == 2) || (fdc_cmdtype == 4) || (fdc_cmdtype == 6)) {
		fdc_status &= ~FDC_ST_WRITEP;
	}
	else {
		if (fdc_writep[fdc_drvreg] && !(fdc_status & FDC_ST_NOTREADY)) {
			fdc_status |= FDC_ST_WRITEP;
		}
		else {
			fdc_status &= ~FDC_ST_WRITEP;
		}
	}

	/* TYPE I ̂ */
	if (fdc_cmdtype != 1) {
		return;
	}

	/* TRACK00 */
	if (fdc_track[fdc_drvreg] == 0) {
		fdc_status |= FDC_ST_TRACK00;
	}
	else {
		fdc_status &= ~FDC_ST_TRACK00;
	}

	/* index */
	if (!(fdc_status & FDC_ST_NOTREADY)) {
		if (fdc_wait) {
			/* UNA */
			fdc_status &= ~FDC_ST_INDEX;

			/* ݂̉]ʒu擾 */
			pos = fdd_get_currentpos();

			/* CfbNXM̎(5000us) */
			if (pos < 5000) {
				fdc_status |= FDC_ST_INDEX;
			}
		} else {
			if (fdc_indexcnt == 0) {
				fdc_status |= FDC_ST_INDEX;
			}
			else {
				fdc_status &= ~FDC_ST_INDEX;
			}
			fdc_next_index();
		}
	}
}

/*
 *	TYPE I
 *	RESTORE
 */
static void FASTCALL fdc_restore(void)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype = 1;
	fdc_status = 0;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_make_stat();
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[fdc_drvreg];

	/* gbNrAXgAAǂݍ */
	if (fdc_track[fdc_drvreg] != 0) {
		fdc_track[fdc_drvreg] = 0;
		fdc_readbuf(fdc_drvreg);
	}
	fdc_seekvct = TRUE;

	/* Abvf[g(f[^WX^0ɂ) */
	fdc_trkreg = 0;
	fdc_datareg = 0;

	/* FDC荞 */
	fdc_drqirq = 0x00;
	if (fdc_wait) {
		fdc_setseekevent(prevtrk);
	}
	else {
		if (!mfd_irq_mask) {
			mainetc_fdc();
			fdc_drqirq |= (BYTE)0x40;
		}
	}

	/* Xe[^X */
	fdc_status = FDC_ST_BUSY;
	if (fdc_command & 0x08) {
		fdc_status |= FDC_ST_HEADENG;
	}
	fdc_make_stat();

	/* ANZX(SEEK) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	SEEK
 */
static void FASTCALL fdc_seek(void)
{
	BYTE target;
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype = 1;
	fdc_status = 0;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_make_stat();
		return;
	}

	/* Ƀxt@C */
	if (fdc_command & 0x04) {
		if (fdc_trkreg != fdc_track[fdc_drvreg]) {
			fdc_status |= FDC_ST_SEEKERR;
			/* C(ՂĂ΍) */
			fdc_trkreg = fdc_track[fdc_drvreg];
		}
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[fdc_drvreg];

	/* ΃V[N */
	target = (BYTE)(fdc_track[fdc_drvreg] + fdc_datareg - fdc_trkreg);
	if (fdc_datareg > fdc_trkreg) {
		fdc_seekvct = FALSE;
#if XM7_VER >= 2
		if (fdc_2ddmode) {
			if (target > 81) {
				target = 81;
			}
		}
		else {
			if (target > 41) {
				target = 41;
			}
		}
#else
		if (target > 41) {
			target = 41;
		}
#endif
	}
	else {
		fdc_seekvct = TRUE;
		/* gbNԍ̃I[o[t[h~ (OS-9΍) */
		if (fdc_track[fdc_drvreg] < (fdc_trkreg - fdc_datareg)) {
			target = 0;
		}
	}

	/* gbNrAV[NAǂݍ */
	if (fdc_track[fdc_drvreg] != target) {
		fdc_track[fdc_drvreg] = target;
		fdc_readbuf(fdc_drvreg);
	}

	/* Abvf[g */
	fdc_trkreg = fdc_datareg;	/* f[^WX^̒lf͗l */

	/* FDC荞 */
	fdc_drqirq = 0x00;
	if (fdc_wait) {
		fdc_setseekevent((BYTE)abs(target - prevtrk));
	}
	else {
		if (!mfd_irq_mask) {
			mainetc_fdc();
			fdc_drqirq |= (BYTE)0x40;
		}
	}

	/* Xe[^X */
	fdc_status |= FDC_ST_BUSY;
	if (fdc_command & 0x08) {
		fdc_status |= FDC_ST_HEADENG;
	}
	fdc_make_stat();

	/* ANZX(SEEK) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP IN
 */
static void FASTCALL fdc_step_in(void)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype = 1;
	fdc_status = 0;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_make_stat();
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[fdc_drvreg];

	/* Ƀxt@C */
	if (fdc_command & 0x04) {
		if (fdc_trkreg != fdc_track[fdc_drvreg]) {
			fdc_status |= FDC_ST_SEEKERR;
		}
	}

	/* XebvAǂݍ */
#if XM7_VER >= 2
	if (fdc_2ddmode) {
		if (fdc_track[fdc_drvreg] < 81) {
			fdc_track[fdc_drvreg]++;
			fdc_readbuf(fdc_drvreg);
		}
	}
	else {
		if (fdc_track[fdc_drvreg] < 41) {
			fdc_track[fdc_drvreg]++;
			fdc_readbuf(fdc_drvreg);
		}
	}
#else
	if (fdc_track[fdc_drvreg] < 41) {
		fdc_track[fdc_drvreg]++;
		fdc_readbuf(fdc_drvreg);
	}
#endif
	fdc_seekvct = FALSE;

	/* Abvf[g */
	if (fdc_command & 0x10) {
		if (fdc_trkreg < 255) {
			fdc_trkreg++;	/* gbNWX^̂̍XV݂̂Ǝv */
		}
	}

	/* FDC荞 */
	fdc_drqirq = 0x00;
	if (fdc_wait && (prevtrk != fdc_track[fdc_drvreg])) {
		fdc_setseekevent(1);
	}
	else {
		if (!mfd_irq_mask) {
			mainetc_fdc();
			fdc_drqirq |= (BYTE)0x40;
		}
	}

	/* Xe[^X */
	fdc_status |= FDC_ST_BUSY;
	if (fdc_command & 0x08) {
		fdc_status |= FDC_ST_HEADENG;
	}
	fdc_make_stat();

	/* ANZX(SEEK) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP OUT
 */
static void FASTCALL fdc_step_out(void)
{
	BYTE prevtrk;

	/* TYPE I */
	fdc_cmdtype = 1;
	fdc_status = 0;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_make_stat();
		return;
	}

	/* ݂̃gbNԍۑ */
	prevtrk = fdc_track[fdc_drvreg];

	/* Ƀxt@C */
	if (fdc_command & 0x04) {
		if (fdc_trkreg != fdc_track[fdc_drvreg]) {
			fdc_status |= FDC_ST_SEEKERR;
		}
	}

	/* XebvAǂݍ */
	if (fdc_track[fdc_drvreg] != 0) {
		fdc_track[fdc_drvreg]--;
		fdc_readbuf(fdc_drvreg);
	}
	fdc_seekvct = TRUE;

	/* Abvf[g */
	if (fdc_command & 0x10) {
		if (fdc_trkreg > 0) {
			fdc_trkreg--;	/* gbNWX^̂̍XV݂̂Ǝv */
		}
	}

	/* FDC荞 */
	fdc_drqirq = 0x00;
	if (fdc_wait && (prevtrk != fdc_track[fdc_drvreg])) {
		fdc_setseekevent(1);
	}
	else {
		if (!mfd_irq_mask) {
			mainetc_fdc();
			fdc_drqirq |= (BYTE)0x40;
		}
	}

	/* Xe[^X */
	fdc_status |= FDC_ST_BUSY;
	if (fdc_command & 0x08) {
		fdc_status |= FDC_ST_HEADENG;
	}
	fdc_make_stat();

	/* ANZX(SEEK) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_SEEK;
}

/*
 *	TYPE I
 *	STEP
 */
static void FASTCALL fdc_step(void)
{
	if (fdc_seekvct) {
		fdc_step_out();
	}
	else {
		fdc_step_in();
	}
}

/*
 *	TYPE II, III
 *	READ/WRITE Tu
 */
static BOOL FASTCALL fdc_rw_sub(void)
{
	fdc_status = 0;

	/* hCu`FbN */
	if (fdc_drvreg >= FDC_DRIVES) {
		LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
		fdc_make_stat();
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		return FALSE;
	}

	/* NOT READY`FbN */
	if ((fdc_ready[fdc_drvreg] == FDC_TYPE_NOTREADY) || fdc_teject[fdc_drvreg]) {
		fdc_make_stat();
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		return FALSE;
	}

	return TRUE;
}

/*
 *	TYPE II
 *	READ DATA
 */
static void FASTCALL fdc_read_data(void)
{
	BYTE stat;
	int secidx;
	int time;

	/* TYPE II, Read */
	fdc_cmdtype = 2;

	/* {`FbN */
	if (!fdc_rw_sub()) {
		return;
	}

	if (fdc_wait) {
		/* |WVݒ */
		fdc_calc_pos();
	}

	/* ZN^ */
	if (fdc_command & 0x02) {
		stat = fdc_readsec(
				fdc_trkreg, fdc_secreg, (BYTE)((fdc_command & 0x08) >> 3),
				TRUE, &secidx, fdd_get_currentpos());
	}
	else {
		stat = fdc_readsec(
				fdc_trkreg, fdc_secreg, fdc_sidereg,
				FALSE, &secidx, fdd_get_currentpos());
	}

	/* ɃXe[^Xݒ肷 */
	fdc_status = stat;

	/* RECORD NOT FOUND ? */
	if (fdc_status & FDC_ST_RECNFND) {
		/* ]f[^ */
		fdc_dataptr = NULL;
		fdc_datareg = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status |= FDC_ST_BUSY;

			/* tONAăCxgōĐݒ肷 */
			fdc_status &= ~FDC_ST_RECNFND;
			schedule_setevent(EVENT_FDC_L, FDD_ROTATION_TIME * 6, fdc_drq_event);
			fdc_access[fdc_drvreg] = FDC_ACCESS_READ;
			return;
		}
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
		return;
	}

	/* ŏ̃f[^ݒ */
	fdc_status |= FDC_ST_BUSY;
	fdc_datareg = fdc_dataptr[0];
	if (fdc_wait) {
		/* YZN^܂ł̎ */
		time = 0;
		if (secidx < fdi_secs) {
			time = fdi_dtpos[secidx] - fdd_get_currentpos();
		}
		if (time <= 0) {
			time += FDD_ROTATION_TIME;
		}

		/* ZN^܂ł̗\莞Ԃݒ */
		schedule_setevent(EVENT_FDC_L, (DWORD)time, fdc_drq_event);
		fdc_drqirq = 0x00;
	}
	else {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq = 0x80;
	}

	/* ANZX(READ) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
#if XM7_VER >= 2
	dmac_start();
#endif
	if (!fdc_wait) {
		schedule_setevent(EVENT_FDC_L, 30 * 1000, fdc_lost_event);
	}
}

/*
 *	TYPE II
 *	WRITE DATA
 */
static void FASTCALL fdc_write_data(void)
{
	BYTE stat;
	int secidx;
	int time;

	/* TYPE II, Write */
	fdc_cmdtype = 3;

	/* {`FbN */
	if (!fdc_rw_sub()) {
		return;
	}

	/* WRITE PROTECT`FbN */
	if (fdc_writep[fdc_drvreg] != 0) {
		fdc_make_stat();
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		return;
	}

	if (fdc_wait) {
		/* |WVݒ */
		fdc_calc_pos();
	}

	/* ZN^ */
	if (fdc_command & 0x02) {
		stat = fdc_readsec(
			fdc_trkreg, fdc_secreg, (BYTE)((fdc_command & 0x08) >> 3),
			TRUE, &secidx, fdd_get_currentpos());
	}
	else {
		stat = fdc_readsec(
			fdc_trkreg, fdc_secreg, fdc_sidereg,
			FALSE, &secidx, fdd_get_currentpos());
	}

	/* ɃXe[^Xݒ肷 */
	fdc_status = stat;

	/* RECORD NOT FOUND ? */
	if (fdc_status & FDC_ST_RECNFND) {
		/* ]f[^ */
		fdc_dataptr = NULL;
		fdc_datareg = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status |= FDC_ST_BUSY;

			/* tONAăCxgŐݒ肷 */
			fdc_status &= ~FDC_ST_RECNFND;
			schedule_setevent(EVENT_FDC_L, FDD_ROTATION_TIME * 6, fdc_drq_event);
			fdc_access[fdc_drvreg] = FDC_ACCESS_WRITE;
			return;
		}
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
		return;
	}

	/* DRQ */
	fdc_status |= FDC_ST_BUSY;
	if (fdc_wait) {
		/* YZN^܂ł̎ */
		time = 0;
		if (secidx < fdi_secs) {
			time = fdi_dtpos[secidx] - fdd_get_currentpos();
		}
		if (time <= 0) {
			time += FDD_ROTATION_TIME;
		}

		/* ZN^܂ł̗\莞Ԃݒ */
		schedule_setevent(EVENT_FDC_L, (DWORD)time, fdc_drq_event);
		fdc_drqirq = 0x00;
	}
	else {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq = 0x80;
	}

	/* ANZX(WRITE) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_WRITE;

	/* DMA]JnAXgf[^Cxgݒ */
#if XM7_VER >= 2
	dmac_start();
#endif
	if (!fdc_wait) {
		schedule_setevent(EVENT_FDC_L, 30 * 1000, fdc_lost_event);
	}
}

/*
 *	TYPE III
 *	READ ADDRESS
 */
static void FASTCALL fdc_read_addr(void)
{
	int i;
	int cur;
	int idx;
	int time;

	/* TYPE III, Read Address */
	fdc_cmdtype = 4;

	/* {`FbN */
	if (!fdc_rw_sub()) {
		return;
	}

	if (fdc_wait) {
		/* C[W̃ZN^ʒuvZ */
		fdc_calc_pos();

		/* ݂̉]ʒu */
		cur = fdd_get_currentpos();

		/* ߂̃ZN^T */
		idx = -1;
		time = 0;
		if (fdi_secs > 0) {
			for (i = 0; i < fdi_secs; i++) {
				if (fdi_idpos[i] < FDD_ROTATION_TIME && fdi_idpos[i] > cur) {
					break;
				}
			}

			/* ZN^m */
			idx = i;
			if (idx == fdi_secs) {
				idx = 0;
			}

			/* ݈ʒu̎ */
			time = fdi_idpos[idx] - cur;
			if (time <= 0) {
				time += FDD_ROTATION_TIME;
			}
		}
	} else {
		/* gbN擪̑΃ZN^ԍ擾 */
		idx = fdc_next_index();
	}

	/* AtH[}bg`FbN */
	if (idx == -1) {
		/* ]f[^ */
		fdc_dataptr = NULL;
		fdc_datareg = 0x00;
		if (fdc_wait) {
			/* Cxg܂łBUSY */
			fdc_status |= FDC_ST_BUSY;

			/* tONAăCxgŐݒ肷 */
			fdc_status &= ~FDC_ST_RECNFND;
			schedule_setevent(EVENT_FDC_L, FDD_ROTATION_TIME * 6, fdc_drq_event);
			fdc_access[fdc_drvreg] = FDC_ACCESS_READ;
			return;
		}
		/* RECORD NOT FOUND */
		fdc_status |= FDC_ST_RECNFND;

		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
		return;
	}

	/* f[^쐬 */
	fdc_makeaddr(idx);

	/* ŏ̃f[^ݒ */
	fdc_status |= FDC_ST_BUSY;
	fdc_datareg = fdc_dataptr[0];
	if (fdc_wait) {
		schedule_setevent(EVENT_FDC_L, (DWORD)time, fdc_drq_event);
		fdc_drqirq = 0x00;
	}
	else {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq = 0x80;
	}

	/* ANZX(READ) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
#if XM7_VER >= 2
	dmac_start();
#endif
	if (!fdc_wait) {
		schedule_setevent(EVENT_FDC_L, 10 * 1000, fdc_lost_event);
	}
}

/*
 *	TYPE III
 *	READ TRACK
 */
static void FASTCALL fdc_read_track(void)
{
	int time;

	/* TYPE III, Read Track */
	fdc_cmdtype = 6;

	/* {`FbN */
	if (!fdc_rw_sub()) {
		return;
	}

	/* f[^쐬 */
	fdc_make_track();

	/* ŏ̃f[^ݒ */
	fdc_status |= FDC_ST_BUSY;
	fdc_datareg = fdc_dataptr[0];
	if (fdc_wait) {
		time = FDD_ROTATION_TIME - fdd_get_currentpos();
		if (time <= 0) {
			time = FDD_ROTATION_TIME;
		}
		schedule_setevent(EVENT_FDC_L, (DWORD)time, fdc_drq_event);
		fdc_drqirq = 0x00;
	}
	else {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq = 0x80;
	}

	/* ANZX(READ) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_READ;

	/* DMA]JnAXgf[^Cxgݒ */
#if XM7_VER >= 2
	dmac_start();
#endif
	if (!fdc_wait) {
		schedule_setevent(EVENT_FDC_L, 150 * 1000, fdc_lost_event);
	}
}

/*
 *	TYPE III
 *	WRITE TRACK
 */
static void FASTCALL fdc_write_track(void)
{
	int time;

	/* TYPE III, Write */
	fdc_cmdtype = 5;

	/* {`FbN */
	if (!fdc_rw_sub()) {
		return;
	}

	/* WRITE PROTECT`FbN */
	if (fdc_writep[fdc_drvreg] != 0) {
		fdc_make_stat();
		/* FDC荞 */
		mainetc_fdc();
		fdc_drqirq = 0x40;
		return;
	}

	/* DRQ */
	fdc_status |= FDC_ST_BUSY;
	if (fdc_wait) {
		time = FDD_ROTATION_TIME - fdd_get_currentpos();
		if (time <= 0) {
			time = FDD_ROTATION_TIME;
		}
		schedule_setevent(EVENT_FDC_L, (DWORD)time, fdc_drq_event);
		fdc_drqirq = 0x00;
	}
	else {
		fdc_status |= FDC_ST_DRQ;
		fdc_drqirq = 0x80;
	}

	fdc_dataptr = fdc_buffer;
	fdc_totalcnt = 0x1800;
	fdc_onetime = 32;
	fdc_nowcnt = 0;

	/* ANZX(WRITE) */
	fdc_access[fdc_drvreg] = FDC_ACCESS_WRITE;

	/* DMA]JnAXgf[^Cxgݒ */
#if XM7_VER >= 2
	dmac_start();
#endif
	if (!fdc_wait) {
		schedule_setevent(EVENT_FDC_L, 150 * 1000, fdc_lost_event);
	}
}

/*
 *	TYPE IV
 *	FORCE INTERRUPT
 */
static void FASTCALL fdc_force_intr(void)
{
	/* WRITE TRACKI */
	if (fdc_cmdtype == 5) {
		if (!fdc_writetrk()) {
			fdc_status |= FDC_ST_WRITEFAULT;
		}
	}

	/* ANZX~ */
	if (!fdc_wait) {
		schedule_delevent(EVENT_FDC_L);
	}

	/* obt@C */
	fdc_readbuf(fdc_drvreg);
	fdc_dataptr = NULL;

	/* Xe[^X */
	switch (fdc_access[fdc_drvreg]) {
		case FDC_ACCESS_READ:
		case FDC_ACCESS_WRITE:
			/* s̃R}hƓXe[^X */
			fdc_status &= ~FDC_ST_BUSY;
			break;

		case FDC_ACCESS_SEEK:
		case FDC_ACCESS_READY:
			fdc_status = 0;
			fdc_indexcnt = 0;
			fdc_make_stat();
			break;
	}

	/* ŃANZXREADY */
	fdc_access[fdc_drvreg] = FDC_ACCESS_READY;

	/* ꂩĂ΁AIRQ(ƂĂ͕s\) */
	if (fdc_command & 0x0f) {
		mainetc_fdc();
		fdc_drqirq = 0x40;
	}
	else {
		fdc_drqirq = 0x00;
	}
}

/*
 *	}`ZN^
 *	Cxg
 */
static BOOL FASTCALL fdc_multi_event(void)
{
	if (fdc_drqirq & 0x10) {
		fdc_drqirq = 0x08;
		/* ZN^WX^UpAZN^ */
		fdc_secreg++;
		schedule_setevent(EVENT_FDC_M, 30, fdc_multi_event);
		return TRUE;
	}

	if (fdc_drqirq & 0x08) {
		fdc_drqirq = 0x00;
		/* ZN^ */
		if (fdc_cmdtype == 2) {
			fdc_read_data();
			fdc_drqirq |= 0x20;
		}
		if (fdc_cmdtype == 3) {
			fdc_write_data();
			fdc_drqirq |= 0x20;
		}
	}

	/* Cxg폜ďI */
	schedule_delevent(EVENT_FDC_M);
	return TRUE;
}

/*
 *	Xgf[^
 *	Cxg
 */
static BOOL FASTCALL fdc_lost_event(void)
{
	if (fdc_dataptr && (fdc_drqirq & 0x80)) {
		/* R}hł؂AXgf[^ */
		fdc_dataptr = NULL;
		fdc_status |= FDC_ST_LOSTDATA;
		fdc_status &= ~FDC_ST_DRQ;
		fdc_status &= ~FDC_ST_BUSY;
		fdc_access[fdc_drvreg] = FDC_ACCESS_READY;

		/* 荞ݔ */
		mainetc_fdc();
		fdc_drqirq = 0x40;
	}

	/* Cxg폜ďI */
	schedule_delevent(EVENT_FDC_L);
	return TRUE;
}


/*
 *	R}h
 */
static void FASTCALL fdc_process_cmd(void)
{
	BYTE high;

	high = (BYTE)(fdc_command >> 4);

	/* EFCgtOݒ */
	fdc_wait = fdc_waitmode;

	/* f[^]sĂ΁A~߂ */
	fdc_dataptr = NULL;

	/*  */
	switch (high) {
		/* restore */
		case 0x00:
			fdc_restore();
			break;
		/* seek */
		case 0x01:
			fdc_seek();
			break;
		/* step */
		case 0x02:
		case 0x03:
			fdc_step();
			break;
		/* step in */
		case 0x04:
		case 0x05:
			fdc_step_in();
			break;
		/* step out */
		case 0x06:
		case 0x07:
			fdc_step_out();
			break;
		/* read data */
		case 0x08:
		case 0x09:
			fdc_read_data();
			break;
		/* write data */
		case 0x0a:
		case 0x0b:
			fdc_write_data();
			break;
		/* read address */
		case 0x0c:
			fdc_read_addr();
			break;
		/* force interrupt */
		case 0x0d:
			fdc_force_intr();
			break;
		/* read track */
		case 0x0e:
			fdc_read_track();
			break;
		/* write track */
		case 0x0f:
			fdc_write_track();
			break;
		/* ȊO */
		default:
			ASSERT(FALSE);
			break;
	}
}

/*
 *	FDC
 *	PoCgǂݏo
 */
BOOL FASTCALL fdc_readb(WORD addr, BYTE *dat)
{
#if XM7_VER >= 2
	BYTE tmp;
#endif
	event_t *ev;

	/* FM-7[hFDC͂ȂɂȂ */
	if (!fdc_enable && (model_type == MODEL_FM7)) {
		return FALSE;
	}

	switch (addr) {
		/* Xe[^XWX^ */
		case 0xfd18:
			fdc_make_stat();
			*dat = fdc_status;
			/* Xe[^X`FbN() */
			fdc_drqirq |= (BYTE)0x20;
			/* BUSY */
			if ((fdc_status & FDC_ST_BUSY) && (fdc_dataptr == NULL)) {
				if (!fdc_wait && (fdc_cmdtype == 1)) {
					/* IRQ On */
					fdc_drqirq |= (BYTE)0x40;

					/* HEAD ENGAGEDtO𗎂Ƃ */
					fdc_status &= ~FDC_ST_HEADENG;
				}
				if (!fdc_wait || (fdc_cmdtype != 1) || (fdc_drqirq & 0x40)) {
					/* BUSYtO𗎂Ƃ */
					fdc_status &= ~FDC_ST_BUSY;
					fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
				}
				/* xBUSY */
				return TRUE;
			}
			/* FDC荞݂AŎ~߂ */
			mfd_irq_flag = FALSE;
			fdc_drqirq &= ~0x40;
			maincpu_irq();
			return TRUE;

		/* gbNWX^ */
		case 0xfd19:
			*dat = fdc_trkreg;
			return TRUE;

		/* ZN^WX^ */
		case 0xfd1a:
			*dat = fdc_secreg;
			return TRUE;

		/* f[^WX^ */
		case 0xfd1b:
			*dat = fdc_datareg;
			/* JE^ */
			if (fdc_dataptr && (fdc_drqirq & 0x20)) {
				fdc_nowcnt++;
				/* DRQ,IRQ */
				if (fdc_nowcnt == fdc_totalcnt) {
					if (fdc_wait) {
						schedule_delevent(EVENT_FDC_L);
					}
					fdc_status &= ~FDC_ST_BUSY;
					fdc_status &= ~FDC_ST_DRQ;
					fdc_drqirq &= (BYTE)~0x80;

					if ((fdc_cmdtype == 2) && (fdc_command & 0x10)) {
						/* }`ZN^ */
						fdc_status |= FDC_ST_BUSY;
						fdc_dataptr = NULL;
						fdc_drqirq = 0x10;
						schedule_setevent(EVENT_FDC_M, 30, fdc_multi_event);
						return TRUE;
					}
					/* VOZN^ */
					fdc_dataptr = NULL;
					mainetc_fdc();
					fdc_drqirq = 0x40;
					fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
					/* Read Track͕KLOST DATAAobt@C */
					if (fdc_cmdtype == 6) {
						fdc_status |= FDC_ST_LOSTDATA;
						fdc_readbuf(fdc_drvreg);
					}
				}
				else {
					fdc_datareg = fdc_dataptr[fdc_nowcnt];
					if (fdc_wait) {
						/* 1oCg]ԂقȂƂĐݒ */
						ev = schedule_getevent(EVENT_FDC_L);
						if (!ev->flag || ev->reload != (DWORD)fdc_onetime) {
							schedule_setevent(
								EVENT_FDC_L, (DWORD)fdc_onetime, fdc_drq_event);
						}

						fdc_status &= (BYTE)~FDC_ST_DRQ;
						fdc_drqirq &= (BYTE)~0x80;
					}
					else {
						fdc_drqirq |= (BYTE)0x80;
					}
				}
			}
			return TRUE;

		/* wbhWX^ */
		case 0xfd1c:
			*dat = (BYTE)(fdc_sidereg | 0xfe);
			return TRUE;

		/* hCuWX^ */
		case 0xfd1d:
#if XM7_VER >= 2
			if (fdc_motor) {
				*dat = (BYTE)(0xbc | fdc_drvregP);
			}
			else {
				*dat = (BYTE)(0x3c | fdc_drvregP);
			}
#else
			if (fdc_motor) {
				*dat = (BYTE)(0xbc | fdc_drvreg);
			}
			else {
				*dat = (BYTE)(0x3c | fdc_drvreg);
			}
#endif
			return TRUE;

		/* [hWX^ */
		case 0xfd1e:
#if XM7_VER >= 2
			if (model_type <= MODEL_FM77AV) {
				/* AV20/AV40/AV20EX/AV40EX̂ݗL */
				*dat = 0xff;
				return TRUE;
			}

			/* _hCuEhCuΉǂݏo */
			tmp = (BYTE)(fdc_logidrv << 2) | (fdc_physdrv[fdc_logidrv]);
			/* 2DD[hǂݏo */
			if (fdc_2ddmode) {
				*dat = (BYTE)(tmp | 0xb0);
			}
			else {
				*dat = (BYTE)(tmp | 0xf0);
			}
#else
			*dat = 0xFF;
#endif
			return TRUE;

		/* DRQ,IRQ */
		case 0xfd1f:
			*dat = (BYTE)(fdc_drqirq | 0x3f);
			/* DRQ܂IRQ`FbN() */
			fdc_drqirq |= (BYTE)0x20;
			/* f[^]ŖΏI(z̐_a) */
			if (fdc_dataptr == NULL) {
				if (!fdc_wait && (fdc_cmdtype == 1) && (fdc_status & FDC_ST_BUSY)) {
					/* IRQ On */
					fdc_drqirq |= (BYTE)0x40;

					/* HEAD ENGAGEDtO𗎂Ƃ */
					fdc_status &= ~FDC_ST_HEADENG;
				}
				if (!fdc_wait || (fdc_cmdtype != 1) || (fdc_drqirq & 0x40)) {
					/* BUSYtO𗎂Ƃ */
					fdc_status &= ~FDC_ST_BUSY;
					fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
				}
			}
			return TRUE;
	}

	return FALSE;
}

/*
 *	FDC
 *	PoCg
 */
BOOL FASTCALL fdc_writeb(WORD addr, BYTE dat)
{
#if XM7_VER >= 2
	BOOL tmp;
#endif
	event_t *ev;

	/* FM-7[hFDC͂ȂɂȂ */
	if (!fdc_enable && (model_type == MODEL_FM7)) {
		return FALSE;
	}

	/* AhX`FbN */
	if ((addr & 0xfff8) == 0xfd18) {
		fdc_use = TRUE;
	}

	switch (addr) {
		/* R}hWX^ */
		case 0xfd18:
#if XM7_VER >= 2
			fdc_dsenable = 0x00;	/* R}hsDS ENABLEԂƂ */
#endif
			fdc_command = dat;
			fdc_process_cmd();
			return TRUE;

		/* gbNWX^ */
		case 0xfd19:
			fdc_trkreg = dat;
			/* R}hs̃gbNWX^΍ */
			/* (F-BASIC V3.3L3x/V3.4L2x VOLCOPY/SUBSET(VCOPYEB)) */
			if (((fdc_status & FDC_ST_BUSY) && (fdc_nowcnt == 0)) &&
				((fdc_cmdtype == 2) || (fdc_cmdtype == 3))) {
				fdc_process_cmd();
			}
			return TRUE;

		/* ZN^WX^ */
		case 0xfd1a:
			fdc_secreg = dat;
			/* R}hs̃ZN^WX^΍ */
			/* (F-BASIC V3.3L3x/V3.4L2x VOLCOPY/SUBSET(VCOPYEB)) */
			if (((fdc_status & FDC_ST_BUSY) && (fdc_nowcnt == 0)) &&
				((fdc_cmdtype == 2) || (fdc_cmdtype == 3))) {
				fdc_process_cmd();
			}
			return TRUE;

		/* f[^WX^ */
		case 0xfd1b:
			fdc_datareg = dat;
			/* JE^ */
			if (fdc_dataptr && (fdc_drqirq & 0x20)) {
				fdc_dataptr[fdc_nowcnt] = fdc_datareg;
				fdc_nowcnt++;
				/* DRQ,IRQ */
				if (fdc_nowcnt == fdc_totalcnt) {
					if (fdc_wait) {
						schedule_delevent(EVENT_FDC_L);
					}
					fdc_status &= ~FDC_ST_DRQ;
					fdc_drqirq &= (BYTE)(~0x80);
					fdc_status &= ~FDC_ST_BUSY;

					if (fdc_cmdtype == 3) {
						if (!fdc_writesec()) {
							fdc_status |= FDC_ST_WRITEFAULT;
						}
					}
					if (fdc_cmdtype == 5) {
						if (!fdc_writetrk()) {
							fdc_status |= FDC_ST_WRITEFAULT;
						}
						/* tH[}bĝߎgpobt@C */
						fdc_readbuf(fdc_drvreg);
					}
					if ((fdc_cmdtype == 3) && (fdc_command & 0x10)) {
						/* }`ZN^ */
						fdc_status |= FDC_ST_BUSY;
						fdc_dataptr = NULL;
						fdc_drqirq = 0x10;
						schedule_setevent(EVENT_FDC_M, 30, fdc_multi_event);
						return TRUE;
					}
					fdc_dataptr = NULL;
					mainetc_fdc();
					fdc_drqirq = 0x40;
					fdc_access[fdc_drvreg] = FDC_ACCESS_READY;
				}
				else {
					if (fdc_wait) {
						/* 1oCg]ԂقȂƂĐݒ */
						ev = schedule_getevent(EVENT_FDC_L);
						if (!ev->flag || ev->reload != (DWORD)fdc_onetime) {
							schedule_setevent(
								EVENT_FDC_L, (DWORD)fdc_onetime, fdc_drq_event);
						}

						fdc_status &= (BYTE)~FDC_ST_DRQ;
						fdc_drqirq &= (BYTE)~0x80;
					}
					else {
						fdc_drqirq |= (BYTE)0x80;
					}
				}
			}
			return TRUE;

		/* wbhWX^ */
		case 0xfd1c:
			if ((dat & 0x01) != fdc_sidereg) {
				fdc_sidereg = (BYTE)(dat & 0x01);
				fdc_readbuf(fdc_drvreg);
			}
			return TRUE;

		/* hCuWX^ */
		case 0xfd1d:
			/* hCuύXȂAfdc_readbuf */
#if XM7_VER >= 2
			if (fdc_drvregP != (dat & 0x03)) {
				fdc_drvregP = (BYTE)(dat & 0x03);
				fdc_drvreg = (BYTE)fdc_physdrv[fdc_drvregP];
				fdc_readbuf(fdc_drvreg);
			}
#else
			if (fdc_drvreg != (dat & 0x03)) {
				fdc_drvreg = (BYTE)(dat & 0x03);
				fdc_readbuf(fdc_drvreg);
			}
#endif
			fdc_motor = (BYTE)(dat & 0x80);
			/* hCuȂA[^~߂ */
			if (fdc_drvreg >= FDC_DRIVES) {
				LOG1(Warning, "drvreg:%d\n", fdc_drvreg);
				fdc_motor = 0;
			}

			/* [^[Cxg */
			ev = schedule_getevent(EVENT_FDD_MOTOR);
			if (fdc_motor) {
				if (ev->flag == EVENT_NOTUSE) {
					schedule_setevent(EVENT_FDD_MOTOR, FDD_ROTATION_TIME, fdd_motor_event);
				}
			} else {
				if (ev->flag == EVENT_ENABLED) {
					schedule_delevent(EVENT_FDD_MOTOR);
				}
			}

#if XM7_VER >= 2
			if (model_type >= MODEL_FM77AV) {	/* FM77AV܂ނ̂ȁH */
				/* DS ENABLE͋L̂(ۂɎgpɂ͋@ˑ̑ΉKv) */
				fdc_dsenable = (BYTE)(dat & 0x40);
			}
#endif
			return TRUE;

		/* [hWX^ */
		case 0xfd1e:
#if XM7_VER >= 2
			if (model_type <= MODEL_FM77AV) {
				/* AV20/AV40/AV20EX/AV40EX̂ݗL */
				return TRUE;
			}

			/* 2DD؂芷 */
			tmp = fdc_2ddmode;

			if (dat & 0x40) {
				fdc_2ddmode = FALSE;
			}
			else {
				fdc_2ddmode = TRUE;
			}

			/* _hCuƕhCȗΉύX */
			fdc_logidrv = (BYTE)((dat & 0x0c) >> 2);
			if (dat & 0x10) {
				/*  */
				fdc_physdrv[fdc_logidrv] = (BYTE)(dat & 0x03);

				/* JghCu̐ݒ肪ύXꂽꍇ̏ */
				if (fdc_logidrv == fdc_drvregP) {
					fdc_drvreg = (BYTE)fdc_physdrv[fdc_logidrv];
					fdc_readbuf(fdc_drvreg);
					return TRUE;
				}
			}

			/* 2DD؂芷sꂽꍇ͍ēǂݍ */
			if (fdc_2ddmode != tmp) {
				fdc_readbuf(fdc_drvreg);
			}
#endif
			return TRUE;
	}

	return FALSE;
}

/*
 *	FDC
 *	Z[u
 */
BOOL FASTCALL fdc_save(int fileh)
{
	int i;

	LOG0(Normal, "Z[u");

	/* t@C֌WɎĂ */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_ready[i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_write(fileh, (BYTE*)fdc_fname[i], 256 + 1)) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_media[i])) {
			return FALSE;
		}
	}

	/* Ver4g */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_bool_write(fileh, fdc_teject[i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_track[i])) {
			return FALSE;
		}
	}
	if (!file_write(fileh, fdc_buffer, 0x2000)) {
		return FALSE;
	}

	/* fdc_dataptr͊Ɉˑf[^|C^ */
	if (!fdc_dataptr) {
		if (!file_word_write(fileh, 0x2000)) {
			return FALSE;
		}
	}
	else {
		if (!file_word_write(fileh, (WORD)(fdc_dataptr - &fdc_buffer[0]))) {
			return FALSE;
		}
	}

	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_write(fileh, fdc_seekofs[i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_write(fileh, fdc_secofs[i])) {
			return FALSE;
		}
	}

	/* I/O */
	if (!file_byte_write(fileh, fdc_command)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_status)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_trkreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_secreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_datareg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_sidereg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_drvreg)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_motor)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_drqirq)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_cmdtype)) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_write(fileh, fdc_totalcnt)) {
		return FALSE;
	}
	if (!file_word_write(fileh, fdc_nowcnt)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, fdc_seekvct)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_indexcnt)) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_access[i])) {
			return FALSE;
		}
	}
	if (!file_bool_write(fileh, fdc_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, fdc_use)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, TRUE)) {
		return FALSE;
	}

	/* Ver9.05/7.05g */
	if (!file_bool_write(fileh, fdc_wait)) {
		return FALSE;
	}

#if XM7_VER >= 2
	/* Ver8g */
	if (!file_byte_write(fileh, fdc_2ddmode)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_logidrv)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fdc_drvregP)) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_write(fileh, fdc_physdrv[i])) {
			return FALSE;
		}
	}
#endif

	if (!file_bool_write(fileh, TRUE)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_byte_write(fileh, fdc_dsenable)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	FDC
 *	[h
 */
BOOL FASTCALL fdc_load(int fileh, int ver)
{
	int i;
	BYTE ready[FDC_DRIVES];
	char fname[FDC_DRIVES][256 + 1];
	BYTE media[FDC_DRIVES];
	WORD offset;
	BOOL tmp;
	int pathlen;

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	/* t@C̍ő啶 */
#if XM7_VER >= 3
	if (((ver >= 715) && (ver <= 799)) || (ver >= 915)) {
#elif XM7_VER >= 2
	if (ver >= 715) {
#else
	if (((ver >= 305) && (ver <= 399)) || (ver >= 404)) {
#endif
		pathlen = 256;
	}
	else {
		pathlen = 128;
	}

	/* t@C֌WɎĂ */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &ready[i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_read(fileh, (BYTE*)fname[i], pathlen + 1)) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &media[i])) {
			return FALSE;
		}
	}

	/* ă}Eg݂ */
	for (i=0; i<FDC_DRIVES; i++) {
		fdc_setdisk(i, NULL);
		if (ready[i] != FDC_TYPE_NOTREADY) {
			fdc_setdisk(i, fname[i]);
			if (fdc_ready[i] != FDC_TYPE_NOTREADY) {
				if (fdc_medias[i] >= (media[i] + 1)) {
					fdc_setmedia(i, media[i]);
				}
			}
		}
	}

	/* Ver4g că`FbNӖȂł(` */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_bool_read(fileh, &fdc_teject[i])) {
			return FALSE;
		}
	}

	/* t@CXe[^X */
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &fdc_track[i])) {
			return FALSE;
		}
	}
	if (!file_read(fileh, fdc_buffer, 0x2000)) {
		return FALSE;
	}

	/* fdc_dataptr͊Ɉˑf[^|C^ */
	if (!file_word_read(fileh, &offset)) {
		return FALSE;
	}
	if (offset >= 0x2000) {
		fdc_dataptr = NULL;
	}
	else {
		fdc_dataptr = &fdc_buffer[offset];
	}

	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_read(fileh, &fdc_seekofs[i])) {
			return FALSE;
		}
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_dword_read(fileh, &fdc_secofs[i])) {
			return FALSE;
		}
	}

	/* I/O */
	if (!file_byte_read(fileh, &fdc_command)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_status)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_trkreg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_secreg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_datareg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_sidereg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_drvreg)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_motor)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_drqirq)) {
		return FALSE;
	}
	if (ver < 600) {
		fdc_drqirq |= (BYTE)0x20;
	}
	if (!file_byte_read(fileh, &fdc_cmdtype)) {
		return FALSE;
	}

	/* ̑ */
	if (!file_word_read(fileh, &fdc_totalcnt)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &fdc_nowcnt)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &fdc_seekvct)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &fdc_indexcnt)) {
		return FALSE;
	}
	for (i=0; i<FDC_DRIVES; i++) {
		if (!file_byte_read(fileh, &fdc_access[i])) {
			return FALSE;
		}
	}
#if XM7_VER >= 3
	if (((ver >= 1203) && (ver <= 1399)) || (ver >= 1403)) {
#elif XM7_VER >= 2
	if (ver >= 1203) {
#else
	if (ver >= 1003) {
#endif
		if (!file_bool_read(fileh, &fdc_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &fdc_use)) {
			return FALSE;
		}
	}
	else {
		fdc_enable = TRUE;
		fdc_use = TRUE;
	}
	/* o[WƂ̌݊p(u[gtO) */
	if (!file_bool_read(fileh, &tmp)) {
		return FALSE;
	}

	/* Cxg */
	schedule_handle(EVENT_FDD_MOTOR, fdd_motor_event);
	schedule_handle(EVENT_FDC_M, fdc_multi_event);
#if XM7_VER >= 3
	if ((ver >= 905) || ((ver >= 705) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 705) {
#else
	if (ver >= 300) {
#endif
		/* Ver9.05/7.05g */
		if (!file_bool_read(fileh, &fdc_wait)) {
			return FALSE;
		}
		schedule_handle(EVENT_FDD_SEEK, fdd_seek_event);
	}
	else {
		/* EFCg[hFALSE */
		fdc_wait = FALSE;

		/* V[NCxg(zbgZbgCxg)폜 */
		schedule_delevent(EVENT_FDD_SEEK);
	}
	/* FDANZXEFCgLȏꍇ̓CxgnhύX */
	if (fdc_wait) {
		schedule_handle(EVENT_FDC_L, fdc_drq_event);
	}
	else {
		schedule_handle(EVENT_FDC_L, fdc_lost_event);
	}

#if XM7_VER >= 2
	/* Ver8g */
#if XM7_VER >= 3
	if (ver < 800) {
#else
	if (ver < 1200) {
#endif
		fdc_2ddmode = FALSE;
		fdc_logidrv = 0;
		fdc_drvregP = fdc_drvreg;
		/* _hCuhCuɐݒ */
		for (i=0; i<FDC_DRIVES; i++) {
			fdc_physdrv[i] = (BYTE)i;
		}
	}
	else {
		if (!file_byte_read(fileh, &fdc_2ddmode)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &fdc_logidrv)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &fdc_drvregP)) {
			return FALSE;
		}
		for (i=0; i<FDC_DRIVES; i++) {
			if (!file_byte_read(fileh, &fdc_physdrv[i])) {
				return FALSE;
			}
		}
	}
#endif

#if XM7_VER >= 3
	if (((ver >= 720) && (ver <= 799)) ||
		((ver >= 920) && (ver <= 999)) ||
		((ver >= 1209) && (ver <= 1399)) ||
		((ver >= 1409) && (ver <= 1599))) {
#elif XM7_VER >= 2
	if (((ver >= 720) && (ver <= 799)) ||
		((ver >= 1209) && (ver <= 1399))) {
#else
	if ((ver >= 310) && (ver <= 399)) {
#endif
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
	}

#if XM7_VER >= 2
#if XM7_VER >= 3
	if (((ver >= 722) && (ver <= 799)) ||
		((ver >= 922) && (ver <= 999)) ||
		((ver >= 1209) && (ver <= 1399)) ||
		((ver >= 1409) && (ver <= 1599))) {
#else
	if (((ver >= 722) && (ver <= 799)) ||
		((ver >= 1209) && (ver <= 1399))) {
#endif
		if (!file_byte_read(fileh, &fdc_dsenable)) {
			return FALSE;
		}
	}
	else {
		fdc_dsenable = 0x00;
	}
#endif

	/* g */
	fdi_init = FALSE;
	fdc_onetime = 32;

	return TRUE;
}
#endif	/* XM7_VER == 1 && defined(SFDC) */
