/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *	Copyright (C) 2021-2023 GIMONS
 *	line printer support 2010-2023 by Ben.JP
 *
 *	[ Win32API j[R}h ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <shellapi.h>
#include <stdlib.h>
#include <assert.h>
#include <tchar.h>
#define DIRECTINPUT_VERSION		0x0300		/* DirectX3w */
#include <dinput.h>
#include <direct.h>
#include "xm7.h"
#include "device.h"
#include "fdc.h"
#include "keyboard.h"
#include "tapelp.h"
#include "tools.h"
#include "mouse.h"
#include "rtc.h"
#include "display.h"
#include "subctrl.h"
#include "jsubsys.h"
#include "bubble.h"
#if XM7_VER == 1
#include "mainetc.h"
#endif	/* XM7_VER == 1 */
#include "w32.h"
#include "w32_bar.h"
#include "w32_draw.h"
#include "w32_snd.h"
#include "w32_sch.h"
#include "w32_sub.h"
#include "w32_cfg.h"
#include "w32_res.h"
#include "w32_kbd.h"
#include "w32_comm.h"
#include "w32_lpr.h"

/*
 *	O[o [N
 */
char InitialDir[INITIAL_DIRS][_MAX_PATH];
char StatePath[_MAX_PATH];
char QuickPath[_MAX_PATH];
BOOL bResumeState;
BOOL bBackupState;
BOOL bStateSave;
BOOL bOFNCentering;
#ifdef KBDPASTE
HWND hKeyStrokeDialog;					/* L[͎x_CAOnh */
BOOL bKeyStrokeModeless;				/* L[͎x_CAO[hXtO */
#endif

/*
 *	X^eBbN [N
 */
static char DiskTitle[16 + 1];
static BOOL DiskMedia;
static BOOL DiskFormat;
static BOOL DiskInterleave;
#if XM7_VER == 1 && defined(BUBBLE)
static char BubbleTitle[16 + 1];
static BOOL BubbleMedia;
static BOOL BubbleFormat;
#endif
static double dRatio;
static WORD wMenuID = 0;
#ifdef KBDPASTE
static char KeyStrokeString[256];
static BOOL fOpenIME;					/* IMEJĂ邩̃tO */
static DWORD fdwConversion;				/* L[̓[h̃ftHgl */
static DWORD fdwSentence;				/* L[ϊ[h̃ftHgl */
#endif

/*
 *	^錾
 */
typedef enum __MIDL___MIDL_itf_inputscope_0000_0001 {
	IS_DEFAULT	= 0,
	IS_HIRAGANA	= 44,
	IS_KATAKANA_HALFWIDTH	= 45,
	IS_KATAKANA_FULLWIDTH	= 46,
} InputScope;
typedef HRESULT (WINAPI *SETINPUTSCOPE)(HWND, InputScope);

/*
 *	u[g\tO}N
 */
#define availableSFDBOOT	(((model_type == MODEL_FM8) && available_sfdboot8) || \
							 ((model_type == MODEL_FM77) && available_sfdboot))
#define availableBBLBOOT	(((model_type == MODEL_FM8) && available_bblboot8) || \
							 ((model_type == MODEL_FM7) && available_bblboot) || \
							 ((model_type == MODEL_FM77) && available_bblboot))
#define availableBBLBOOT128	(((model_type == MODEL_FM8) && available_bblboot8) || \
							 ((model_type == MODEL_FM77) && available_bblboot))

/*-[ ^Cg̓_CAO ]------------------------------------------------*/

/*
 *	^Cg̓_CAO
 *	_CAO
 */
static BOOL FASTCALL TitleDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_TITLEEDIT);
	ASSERT(hWnd);
	SetWindowText(hWnd, DiskTitle);

#if XM7_VER >= 3
	/* fBA^Cv */
	CheckDlgButton(hDlg, IDC_TITLE2D, BST_CHECKED);
#elif XM7_VER == 1 && defined(SFDC)
	/* fBA^Cv */
	CheckDlgButton(hDlg, IDC_TITLE2D, BST_CHECKED);
#endif

	/* [U[fBXNݒ擾 */
	CheckDlgButton(hDlg, IDC_TITLEUSRDISK, BST_UNCHECKED);

	/* C^[[uݒ擾 */
	CheckDlgButton(hDlg, IDC_TITLEINTERLEAVE, BST_CHECKED);
	hWnd = GetDlgItem(hDlg, IDC_TITLEINTERLEAVE);
	EnableWindow(hWnd, FALSE);

	return TRUE;
}

/*
 *	^Cg̓_CAO
 *	_CAOOK
 */
static void FASTCALL TitleDlgOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_TITLEEDIT);
	ASSERT(hWnd);

	/* 擾ARs[ */
	GetWindowText(hWnd, string, sizeof(string) - 1);
	memset(DiskTitle, 0, sizeof(DiskTitle));
	string[16] = '\0';
	strncpy(DiskTitle, string, sizeof(DiskTitle));

	/* fBA^Cv擾 */
#if XM7_VER >= 3
	if (IsDlgButtonChecked(hDlg, IDC_TITLE2DD)) {
		DiskMedia = TRUE;
	}
	else {
		DiskMedia = FALSE;
	}
#elif XM7_VER == 1 && defined(SFDC)
	if (IsDlgButtonChecked(hDlg, IDC_TITLE2HD)) {
		DiskMedia = TRUE;
	}
	else {
		DiskMedia = FALSE;
	}
#else
	/* V2ł2D̂ */
	DiskMedia = FALSE;
#endif

	/* [U[fBXNݒ擾 */
	if (IsDlgButtonChecked(hDlg, IDC_TITLEUSRDISK)) {
		DiskFormat = TRUE;
	}
	else {
		DiskFormat = FALSE;
	}

	/* C^[[uݒ擾 */
	if (IsDlgButtonChecked(hDlg, IDC_TITLEINTERLEAVE)) {
		DiskInterleave = TRUE;
	}
	else {
		DiskInterleave = FALSE;
	}
}

/*
 *	^Cg̓_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK TitleDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	HWND hWnd;

	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return TitleDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						TitleDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					SetMenuExitTimer();
					return TRUE;
			}

			switch (LOWORD(wParam)) {
				/* [U[fBXNtO */
				case IDC_TITLEUSRDISK:
					hWnd = GetDlgItem(hDlg, IDC_TITLEINTERLEAVE);
					if (IsDlgButtonChecked(hDlg, IDC_TITLEUSRDISK) == BST_CHECKED) {
						EnableWindow(hWnd, TRUE);
					}
					else {
						EnableWindow(hWnd, FALSE);
					}
					break;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*-[ ^Cg̓_CAO(2D/VFDϊp) ]----------------------------------*/

/*
 *	^Cg̓_CAO(2D/VFDϊp)
 *	_CAO
 */
static BOOL FASTCALL TitleDlg2DInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_TITLEEDIT2D);
	ASSERT(hWnd);
	SetWindowText(hWnd, DiskTitle);

	return TRUE;
}

/*
 *	^Cg̓_CAO(2D/VFDϊp)
 *	_CAOOK
 */
static void FASTCALL TitleDlg2DOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_TITLEEDIT2D);
	ASSERT(hWnd);

	/* 擾ARs[ */
	GetWindowText(hWnd, string, sizeof(string) - 1);
	memset(DiskTitle, 0, sizeof(DiskTitle));
	string[16] = '\0';
	strncpy(DiskTitle, string, sizeof(DiskTitle));
}

/*
 *	^Cg̓_CAO(2D/VFDϊp)
 *	_CAOvV[W
 */
static BOOL CALLBACK TitleDlg2DProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return TitleDlg2DInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						TitleDlg2DOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					SetMenuExitTimer();
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*-[ oufBA^CvI_CAO ]------------------------------------*/

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	oufBA^CvI_CAO
 *	_CAO
 */
static BOOL FASTCALL BubbleMediaTypeDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_TITLEEDIT);
	ASSERT(hWnd);
	SetWindowText(hWnd, BubbleTitle);

	/* fBA^Cv */
	CheckDlgButton(hDlg, IDC_MEDIATYPE32KB, BST_CHECKED);

	/* tH[}bg^Cv */
	CheckDlgButton(hDlg, IDC_MEDIAFORMATB77, BST_CHECKED);

	return TRUE;
}

/*
 *	oufBA^CvI_CAO
 *	_CAOOK
 */
static void FASTCALL BubbleMediaTypeDlgOK(HWND hDlg)
{
	ASSERT(hDlg);

	if (IsDlgButtonChecked(hDlg, IDC_MEDIAFORMATBBL)) {
		BubbleFormat = FALSE;
	}
	else {
		BubbleFormat = TRUE;
	}

	/* fBA^Cv擾 */
	if (IsDlgButtonChecked(hDlg, IDC_MEDIATYPE128KB)) {
		BubbleMedia = TRUE;
	}
	else {
		BubbleMedia = FALSE;
	}
}

/*
 *	oufBA^CvI_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK BubbleMediaTypeDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return BubbleMediaTypeDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						BubbleMediaTypeDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					SetMenuExitTimer();
					return TRUE;
				case IDC_MEDIAFORMATB77:
					EnableWindow(GetDlgItem(hDlg, IDC_TITLEEDIT), TRUE);
					return TRUE;
				case IDC_MEDIAFORMATBBL:
					EnableWindow(GetDlgItem(hDlg, IDC_TITLEEDIT), FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

/*-[ t@CIR_CAO ]-----------------------------------------*/

/*
 *	t@CIR_CAO
 *	
 */
static void FASTCALL FileDialogInit(HWND hDlg)
{
	RECT drect;
	RECT prect;
	HWND hWnd;

	/* _CAOCEChE̒Ɋ񂹂 */
	hWnd = GetParent(hDlg);
	GetWindowRect(hMainWnd, &prect);
	GetWindowRect(hWnd, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hWnd, drect.left, drect.top, drect.right, drect.bottom, FALSE);
}

/*
 *	t@CIR_CAO
 *	HOOK֐
 */
static UINT CALLBACK FileSelectHook(
						HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	UNUSED(wParam);
	UNUSED(lParam);

	switch (uMsg) {
		case WM_INITDIALOG:
			FileDialogInit(hDlg);
			return 0;
	}

	return 0;
}

/*
 *	t@CIR_CAO
 */
BOOL FASTCALL FileSelectSub(BOOL bOpen, UINT uFilterID, char *path, char *defext, BYTE IniDirNo)
{
	OPENFILENAME ofn;
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME + _MAX_EXT];
	char filter[4096];
	int i, j;

	ASSERT((bOpen == TRUE) || (bOpen == FALSE));
	ASSERT(uFilterID > 0);
	ASSERT(path);

	keyboard_inputmask(TRUE);

	/* f[^쐬 */
	memset(&ofn, 0, sizeof(ofn));
	memset(path, 0, _MAX_PATH);
	memset(fname, 0, sizeof(fname));
	ofn.lStructSize = 76;	/* sizeof(ofn)V5g܂ */
	ofn.hwndOwner = hMainWnd;

	LoadString(hResInstance, uFilterID, filter, sizeof(filter));
	j = strlen(filter);
	for (i=0; i<j; i++) {
		if (filter[i] == '|') {
			filter[i] = '\0';
		}
	}

	ofn.lpstrFilter = filter;
	ofn.lpstrFile = path;
	ofn.nMaxFile = _MAX_PATH;
	ofn.lpstrFileTitle = fname;
	ofn.nMaxFileTitle = sizeof(fname);
	ofn.lpstrDefExt = defext;
	ofn.lpstrInitialDir = InitialDir[IniDirNo];
	ofn.lpfnHook = FileSelectHook;

	/* R_CAOs */
	if (bOpen) {
		ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY |
					OFN_FILEMUSTEXIST;
		if (bOFNCentering) {
			ofn.Flags |= OFN_ENABLEHOOK;
		}
		if (!GetOpenFileName(&ofn)) {
			WAIT_KEY_POP();
			keyboard_inputmask(FALSE);
			SetMenuExitTimer();
			return FALSE;
		}
	}
	else {
		ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY |
					OFN_OVERWRITEPROMPT;
		if (bOFNCentering) {
			ofn.Flags |= OFN_ENABLEHOOK;
		}
		if (!GetSaveFileName(&ofn)) {
			WAIT_KEY_POP();
			keyboard_inputmask(FALSE);
			SetMenuExitTimer();
			return FALSE;
		}
	}

	/* fBNgۑ */
	_splitpath(path, InitialDir[IniDirNo], dir, NULL, NULL);
	if (dir[strlen(dir)-1] == '\\') {
		/* Ō̃pX؂L͋Iɍ */
		dir[strlen(dir)-1] = '\0';
	}
	strncat(InitialDir[IniDirNo], dir,
			sizeof(InitialDir[IniDirNo]) - strlen(InitialDir[IniDirNo]) - 1);

	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();
	return TRUE;
}

/*-[ AXyNgݒ_CAO ]-------------------------------------------*/

/*
 *	AXyNgݒ_CAO
 *	XN[
 */
static BOOL FASTCALL AspectRatioDlgVScroll(HWND hDlg, WORD wPos, HWND hWnd)
{
	char string[128];
	HWND hSubWnd;

	UNUSED(wPos);

	/* EBhEnh`FbN */
	if (hWnd == GetDlgItem(hDlg, IDC_ASPECTSPIN)) {
		/* AXyNg */
		/* |WVAofBEChEɒlݒ */
		hSubWnd = GetDlgItem(hDlg, IDC_ASPECTEDIT);
		ASSERT(hSubWnd);
		_snprintf(string, sizeof(string), "%.3f", dRatio);
		SetWindowText(hSubWnd, string);
		return TRUE;
	}
	return FALSE;
}

/*
 *	AXyNgݒ_CAO
 *	_CAO
 */
static BOOL FASTCALL AspectRatioDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	char string[128];

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* GfBbgeLXg */
	memset(string, 0, sizeof(string));
	hWnd = GetDlgItem(hDlg, IDC_ASPECTEDIT);
	ASSERT(hWnd);
	_snprintf(string, sizeof(string), "%.3f", dRatio);
	SetWindowText(hWnd, string);

	return TRUE;
}

/*
 *	AXyNgݒ_CAO
 *	_CAOOK
 */
static void FASTCALL AspectRatioDlgOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_ASPECTEDIT);
	ASSERT(hWnd);

	/* 擾ARs[ */
	memset(string, 0, sizeof(string));
	GetWindowText(hWnd, string, sizeof(string) - 1);
	dRatio = atof(string);
}

/*
 *	AXyNgݒ_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK AspectRatioDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	char string[128];
	LPNMUPDOWN lpnmud;
	HWND hWnd;

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return AspectRatioDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						AspectRatioDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					SetMenuExitTimer();
					return TRUE;
				case IDC_ASPECT_MB273XX:
				case IDC_ASPECT_FMTVXXX:
					if (LOWORD(wParam) == IDC_ASPECT_MB273XX) {
						dRatio = ASPECT_RATIO_MB27;
					}
					else {
						dRatio = ASPECT_RATIO_FMTV;
					}
					/* GfBbgeLXg */
					memset(string, 0, sizeof(string));
					hWnd = GetDlgItem(hDlg, IDC_ASPECTEDIT);
					ASSERT(hWnd);
					_snprintf(string, sizeof(string), "%.3f", dRatio);
					SetWindowText(hWnd, string);
					return TRUE;
			}
			break;

		/* ʒm */
		case WM_NOTIFY:
			if (wParam == (WPARAM)IDC_ASPECTSPIN) {
				lpnmud = (LPNMUPDOWN)lParam;
				if (lpnmud->hdr.code == UDN_DELTAPOS) {
					/* GfBbgeLXg */
					hWnd = GetDlgItem(hDlg, IDC_ASPECTEDIT);
					ASSERT(hWnd);
					GetWindowText(hWnd, string, (int)sizeof(string));
					dRatio = (double)atof(string) - (double)lpnmud->iDelta / (double)1000.0;
					if (dRatio < 0.5) {
						dRatio = 0.5;
					}
					else if (dRatio > 2.0) {
						dRatio = 2.0;
					}
				}
				return TRUE;
			}
			break;

		/* XN[ */
		case WM_VSCROLL:
			return AspectRatioDlgVScroll(hDlg, HIWORD(wParam), (HWND)lParam);
	}

	/* ȊÓAFALSE */
	return FALSE;
}

#ifdef KBDPASTE
/*-[ L[͎x_CAO ]-----------------------------------------------*/

/*
 *	L[͎x_CAO
 *	InputScopeύX
 */
static void FASTCALL SetInputScope(HWND hEdit, InputScope dwInputScope)
{
	char path[_MAX_PATH];
	HMODULE hMSCTF;
	SETINPUTSCOPE pfnSetInputScope;

	GetSystemDirectory(path, sizeof(path));
	strncat(path, "\\msctf.dll", sizeof(path) - strlen(path) - 1);
	hMSCTF = LoadLibrary(path);
	pfnSetInputScope = (SETINPUTSCOPE)GetProcAddress(hMSCTF, "SetInputScope");
	if (pfnSetInputScope) {
		(*pfnSetInputScope)(hEdit, dwInputScope);
	}

	FreeLibrary(hMSCTF);
}

/*
 *	L[͎x_CAO
 *	_CAO
 */
static BOOL FASTCALL KeyStrokeDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	HIMC hImm;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/*  */
	memset(KeyStrokeString, 0, sizeof(KeyStrokeString));

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_KEYSTROKEEDIT);
	ASSERT(hWnd);

	/* IMMݒ𔼊pJiɐݒ */
	SetFocus(hWnd);					/* WIn8/Win10 ATOKp */
	if (bWin8flag) {
		SetInputScope(hWnd, IS_KATAKANA_HALFWIDTH);
	}
	hImm = ImmGetContext(hWnd);
	fOpenIME = ImmGetOpenStatus(hImm);
	ImmGetConversionStatus(hImm, &fdwConversion, &fdwSentence);
	ImmSetOpenStatus(hImm, TRUE);
	ImmSetConversionStatus(hImm,
			IME_CMODE_NATIVE | IME_CMODE_KATAKANA |
			IME_CMODE_ROMAN, IME_SMODE_NONE);
	ImmReleaseContext(hWnd, hImm);		/* gp͉ */

	return TRUE;
}

/*
 *	L[͎x_CAO
 *	_CAOOK
 */
static void FASTCALL KeyStrokeDlgOK(HWND hDlg)
{
	HWND hWnd;

	ASSERT(hDlg);

	/* GfBbgeLXg */
	hWnd = GetDlgItem(hDlg, IDC_KEYSTROKEEDIT);
	ASSERT(hWnd);

	/* 擾ARs[ */
	GetWindowText(hWnd, (LPTSTR)KeyStrokeString, sizeof(KeyStrokeString) - 1);
	SetWindowText(hWnd, (LPTSTR)"");
}

/*
 *	L[͎x_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK KeyStrokeDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	HWND hWnd;
	HIMC hImm;

	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return KeyStrokeDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						KeyStrokeDlgOK(hDlg);
						if (strlen(KeyStrokeString) == 0) {
							break;
						}
						if (hKeyStrokeDialog) {
							/* 쐬 */
							LockVM();
							PasteKbd((char *)KeyStrokeString);
							UnlockVM();
							SetForegroundWindow(hMainWnd);
							break;
						}
					}

					/* GfBbgeLXg */
					hWnd = GetDlgItem(hDlg, IDC_KEYSTROKEEDIT);
					ASSERT(hWnd);

					/* IMMݒ̕A */
					SetFocus(hWnd);					/* Win8/Win10 ATOKp */
					if (bWin8flag) {
						SetInputScope(hWnd, IS_DEFAULT);
					}
					hImm = ImmGetContext(hWnd);
					ImmSetOpenStatus(hImm, fOpenIME);
					ImmSetConversionStatus(hImm, fdwConversion, fdwSentence);
					ImmReleaseContext(hWnd, hImm);		/* gp͉ */

					if (hKeyStrokeDialog) {
						DestroyWindow(hKeyStrokeDialog);
						hKeyStrokeDialog = NULL;
					}
					else {
						EndDialog(hDlg, LOWORD(wParam));
					}
					InvalidateRect(hDrawWnd, NULL, FALSE);
					SetMenuExitTimer();
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}
#endif

/*-[ ėpTu ]-------------------------------------------------------------*/

/*
 *	j[Enable
 */
void FASTCALL EnableMenuSub(HMENU hMenu, UINT uID, BOOL bEnable)
{
	ASSERT(hMenu);
	ASSERT(uID > 0);

	if (bEnable) {
		EnableMenuItem(hMenu, uID, MF_BYCOMMAND | MF_ENABLED);
	}
	else {
		EnableMenuItem(hMenu, uID, MF_BYCOMMAND | MF_GRAYED);
	}
}

/*
 *	j[Enable (ʒuw)
 */
void FASTCALL EnableMenuPos(HMENU hMenu, UINT uPos, BOOL bEnable)
{
	ASSERT(hMenu);
	ASSERT(uPos > 0);

	if (bEnable) {
		EnableMenuItem(hMenu, uPos, MF_BYPOSITION | MF_ENABLED);
	}
	else {
		EnableMenuItem(hMenu, uPos, MF_BYPOSITION | MF_GRAYED);
	}
}

/*
 *	j[Check
 */
void FASTCALL CheckMenuSub(HMENU hMenu, UINT uID, BOOL bCheck)
{
	ASSERT(hMenu);
	ASSERT(uID > 0);

	if (bCheck) {
		CheckMenuItem(hMenu, uID, MF_BYCOMMAND | MF_CHECKED);
	}
	else {
		CheckMenuItem(hMenu, uID, MF_BYCOMMAND | MF_UNCHECKED);
	}
}

/*-[ t@Cj[ ]-----------------------------------------------------*/

/*
 *	Xe[g[h
 */
static void FASTCALL StateLoad(char *path)
{
	char string[256];
	int state;

	state = system_load(path);
	if (state == STATELOAD_ERROR) {
		/* {̓ǂݍݒ̃G[̂݃Zbg */
		system_reset();
		StatePath[0] = '\0';
		SetAllMRUState();
		OnRefresh(hMainWnd);
	}
	if (state == STATELOAD_SUCCESS) {
		strncpy(StatePath, path, sizeof(StatePath));
		SetAllMRUState();
		OnRefresh(hMainWnd);
	}
	else if (state == STATELOAD_OPENERR) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_STATENOEXIST, string, sizeof(string));
		MessageBox(hMainWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
	}
	else {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_STATEERROR, string, sizeof(string));
		MessageBox(hMainWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
	}

	/* Esǂ̃P[Xłݒf[^ޔ */
	StoreCfg();
#if XM7_VER == 1 && defined(BUBBLE)
	StoreCfgBMC();
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
}

/*
 *	t@CtbZ[W{bNX
 */
static int FASTCALL FileMessageBox(HWND hWnd, UINT uID, char *path, UINT uType)
{
	char string[256];
	char buffer[256];
	char tmp[256];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	int nret;

	memset(&tmp, 0, sizeof(tmp));
	if (path != NULL) {
		if (path[0] != '\0') {
			_splitpath(path, drive, dir, fname, ext);
			if (strlen(dir) > 1) {
				strncpy(dir, "\\...\\", sizeof(dir));
			}
			strncpy(tmp, drive, sizeof(tmp));
			strncat(tmp, dir, sizeof(tmp)-strlen(tmp)-1);
			strncat(tmp, fname, sizeof(tmp)-strlen(tmp)-1);
			strncat(tmp, ext, sizeof(tmp)-strlen(tmp)-1);
		}
	}

	keyboard_inputmask(TRUE);
	LoadString(hResInstance, uID, buffer, sizeof(buffer));
	_snprintf(string, sizeof(string), buffer, tmp);
	nret = MessageBox(hWnd, string, "XM7", uType);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();
	return nret;
}

/*
 *	j[Xgǉ
 */
static void FASTCALL AddMenuList(HMENU hMenu, UINT uItem, WORD wIDS,
									WORD wID, int type, BOOL bflag)
{
	HMENU hSubMenu;
	MENUITEMINFO mii;
	int	i, j, k;
	char buffer[256+1];
	char string[256+1];
	UINT uitem2;

	ASSERT(hMenu);

	/* Tuj[nh擾 */
	hSubMenu = GetSubMenu(hMenu, uItem);

	/* Tuj[ׂč폜 */
	while (GetMenuItemCount(hSubMenu) > 0) {
		DeleteMenu(hSubMenu, 0, MF_BYPOSITION);
	}

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING | MFT_RADIOCHECK;

	/* Xgǉ */
	for (i=0; i<=XDL_FILES; i++) {
		char fname[_MAX_FNAME];
		mii.wID = wID + i;
		mii.fState = MFS_ENABLED;

		/* 擾Ă݂ */
		if (i == 0) {
			LoadString(hResInstance, wIDS, buffer, sizeof(buffer));
		}
		else {
			GetXDLFile(type, i - 1, buffer, sizeof(buffer));
		}
		if (buffer[0] == '\0') {
			break;
		}

		/* ΃j[ɒǉ */
		_splitpath(buffer, NULL, NULL, fname, NULL);
		strncpy(buffer, fname, sizeof(buffer));

		/* vtBbNX΍ */
		_snprintf(string, sizeof(string), "&%c ", '0' + i);
		k = strlen(string);
		for (j = 0; j < (int)strlen(buffer); j++) {
			if (buffer[j] == '&') {
				string[k++] = '&';
			}
			string[k++] = buffer[j];
		}
		string[k] = '\0';
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hSubMenu, i, TRUE, &mii);
	}

	/* XgȂ */
	if (i <= 1) {
		EnableMenuItem(hMenu, uItem, MF_BYPOSITION | MF_GRAYED);
	}
	else {
		if (bflag) {
			uitem2 = GetXDLSel(type) + 1;
		}
		else {
			uitem2 = 0;
		}
		CheckMenuRadioItem(hSubMenu, wID, wID + i, wID + uitem2, MF_BYCOMMAND);
	}
}

/*
 *	t@CXV
 */
static UINT FASTCALL AppendMenuHistory(HMENU hMenu, UINT uItem, UINT uID, int type, BOOL enable)
{
	MENUITEMINFO mii;
	char string[256+1];
	char buffer[256+1];
	int	i, j, k;

	ASSERT(hMenu);

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	if (enable) {
		mii.fState = MFS_ENABLED;
	}
	else {
		mii.fState = MFS_GRAYED;
	}

	/*  */
	if (GetMRUNum(type) != 0) {
		/* Zp[^} */
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(hMenu, uItem++, TRUE, &mii);
		mii.fType = MFT_STRING;

		/*  */
		for (i=0; i<MRU_FILES; i++) {
			char drive[_MAX_DRIVE];
			char dir[_MAX_DIR];
			char fname[_MAX_FNAME];
			char ext[_MAX_EXT];

			/* 擾Ă݂ */
			GetMRUFile(type, i, buffer, sizeof(buffer));
			if (buffer[0] == '\0') {
				break;
			}

			/* ΃j[ɒǉ */
			_splitpath(buffer, drive, dir, fname, ext);
			if (strlen(dir) > 1) {
				strncpy(dir, "\\...\\", sizeof(dir));
			}
			strncpy(buffer, drive, sizeof(buffer));
			strncat(buffer, dir, sizeof(buffer)-strlen(buffer)-1);
			strncat(buffer, fname, sizeof(buffer)-strlen(buffer)-1);
			strncat(buffer, ext, sizeof(buffer)-strlen(buffer)-1);

			mii.wID = uID + i;
			/* vtBbNX΍ */
			_snprintf(string, sizeof(string), "&%d ", i + 1);
			k = strlen(string);
			for (j = 0; j < (int)strlen(buffer); j++) {
				if (buffer[j] == '&') {
					string[k++] = '&';
				}
				string[k++] = buffer[j];
			}
			string[k] = '\0';
			mii.dwTypeData = string;
			mii.cch = strlen(string);
			InsertMenuItem(hMenu, uItem++, TRUE, &mii);
		}
	}

	return uItem;
}

/*
 *	ReLXgj[o͏
 */
static void FASTCALL OnContextualSubMenu(HWND hWnd, HMENU hMenu, WORD wID, int x, int y)
{
	POINT point;
	HMENU hSubMenu;
	WORD wRet;

	/* Tuj[\\[XIDۑ */
	wMenuID = wID;

	/* ReLXgj[s */
	point.x = x;
	point.y = y;
	hSubMenu = GetSubMenu(hMenu, 0);
	wRet = (WORD)TrackPopupMenu(hSubMenu, TPM_RECURSE|TPM_RETURNCMD, point.x, point.y, 0, hWnd, NULL);
	if (wRet != 0) {
		keyboard_inputmask(TRUE);
		EnterMenu(hWnd, FALSE);
		OnCommand(hWnd, wRet);
		ExitMenu(FALSE);
		OnExitMenuLoop();
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
		bMenuLoop = FALSE;
	}
}

/*
 *	qXg폜
 */
static void FASTCALL DeleteHistory(HWND hWnd, WORD wID)
{
	char string[256 + 1];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	int type;
	int index;
	int nret;

	/* XgZbg */
	if ((wID >= IDM_STATE_FILE0) && (wID <= IDM_STATE_FILE9)) {
		type = MRU_STATE;
		index = wID - IDM_STATE_FILE0;
	}
#if XM7_VER == 1 && defined(SFDC)
	if (((wID >= IDM_D7_FILE0) && (wID <= IDM_D7_FILE9))) {
		type = MRU_SFD3;
		index = wID - IDM_D7_FILE0;
	}
	if (((wID >= IDM_D6_FILE0) && (wID <= IDM_D6_FILE9))) {
		type = MRU_SFD2;
		index = wID - IDM_D6_FILE0;
	}
	if (((wID >= IDM_D5_FILE0) && (wID <= IDM_D5_FILE9))) {
		type = MRU_SFD1;
		index = wID - IDM_D5_FILE0;
	}
	if (((wID >= IDM_D4_FILE0) && (wID <= IDM_D4_FILE9))) {
		type = MRU_SFD0;
		index = wID - IDM_D4_FILE0;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if ((wID >= IDM_D3_FILE0) && (wID <= IDM_D3_FILE9)) {
		type = MRU_MFD3;
		index = wID - IDM_D3_FILE0;
	}
	if ((wID >= IDM_D2_FILE0) && (wID <= IDM_D2_FILE9)) {
		type = MRU_MFD2;
		index = wID - IDM_D2_FILE0;
	}
	if ((wID >= IDM_D1_FILE0) && (wID <= IDM_D1_FILE9)) {
		type = MRU_MFD1;
		index = wID - IDM_D1_FILE0;
	}
	if ((wID >= IDM_D0_FILE0) && (wID <= IDM_D0_FILE9)) {
		type = MRU_MFD0;
		index = wID - IDM_D0_FILE0;
	}
	if ((wID >= IDM_T_FILE0) && (wID <= IDM_T_FILE9)) {
		type = MRU_TAPE;
		index = wID - IDM_T_FILE0;
	}
#if XM7_VER == 1 && defined(BUBBLE)
	if ((wID >= IDM_B3_FILE0) && (wID <= IDM_B3_FILE9)) {
		type = MRU_BBL128_1;
		index = wID - IDM_B3_FILE0;
	}
	if ((wID >= IDM_B2_FILE0) && (wID <= IDM_B2_FILE9)) {
		type = MRU_BBL128_0;
		index = wID - IDM_B2_FILE0;
	}
	if ((wID >= IDM_B1_FILE0) && (wID <= IDM_B1_FILE9)) {
		type = MRU_BBL1;
		index = wID - IDM_B1_FILE0;
	}
	if ((wID >= IDM_B0_FILE0) && (wID <= IDM_B0_FILE9)) {
		type = MRU_BBL0;
		index = wID - IDM_B0_FILE0;
	}
#endif

	/* 擾Ă݂ */
	GetMRUFile(type, index, string, sizeof(string));
	nret = FileMessageBox(hWnd, IDS_HISTORY_DELETE, string, MB_ICONQUESTION | MB_YESNO);
	/* mFʂɂ */
	switch (nret) {
		/* YES */
		case IDYES:
			/* Xe[gZ[u */
			LockVM();
			StopSnd();
			DelMRUFile(type, index);
			PlaySnd();
			UnlockVM();
			break;

		/* NO */
		case IDNO:
			break;
	}
}

#ifdef FILTERDLL
/*
 *	AXyNgݒ菈
 */
static void FASTCALL AspectSetting(HWND hWnd, WORD wID)
{
	int nret;

	ASSERT(hWnd);
	UNUSED(wID);

	/* AXyNg */
	dRatio = dAspectRatio;
	nret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_ASPECTRATIODLG),
						hWnd, AspectRatioDlgProc);
	if (nret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	/* AXyNgύX */
	if (dRatio < 0.5) {
		dRatio = 0.5;
	}
	else if (dRatio > 2.0) {
		dRatio = 2.0;
	}
	dAspectRatio = dRatio;

	/* AXyNg䒲tOύX */
	bAspectRatio = TRUE;
	StoreCfg();

	/* \eXV */
	OnSize(hWnd, 640, 400);
	InvalidateRect(hDrawWnd, NULL, FALSE);

	PlaySnd();
	ResetSch();
	UnlockVM();

	SetMenuExitTimer();
}
#endif

/*-[ t@Cj[ ]-----------------------------------------------------*/

/*
 *	J(O)
 */
static void FASTCALL OnOpen(HWND hWnd)
{
	char path[_MAX_PATH];

	ASSERT(hWnd);

	/* t@CITu */
	if (!FileSelectSub(TRUE, IDS_STATEFILTER, path, NULL, STATE_FILE_DIR)) {
		return;
	}

	/* Xe[g[h */
	LockVM();
	StopSnd();
	StateLoad(path);
	PlaySnd();
	ResetSch();
	UnlockVM();

	/* ʍĕ` */
	OnRefresh(hWnd);

	/* \eXV */
	OnSize(hWnd, 640, 400);
}

/*
 *	Otĕۑ(A)
 */
static void FASTCALL OnSaveAs(HWND hWnd)
{
	char path[_MAX_PATH];

	ASSERT(hWnd);

	/* t@CITu */
	if (!FileSelectSub(FALSE, IDS_STATEFILTER, path, "XM7", STATE_FILE_DIR)) {
		return;
	}

	/* Xe[gZ[u */
	LockVM();
	StopSnd();
	if (bBackupState) {
		/* Xe[g̑Ҕ */
		file_rename_state(path);
	}
	if (!system_save(path)) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_STATESAVEERROR, path, sizeof(path));
		MessageBox(hWnd, path, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	else {
		strncpy(StatePath, path, sizeof(StatePath));
		SetMRUFile(MRU_STATE, StatePath);
	}
	SetMenuExitTimer();
	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	㏑ۑ(S)
 */
static void FASTCALL OnSave(HWND hWnd)
{
	char string[128];

	/* ܂ۑĂȂ΁AO */
	if ((StatePath[0] == '\0') ||
		(!strncmp(StatePath, QuickPath, _MAX_PATH))) {
		OnSaveAs(hWnd);
		return;
	}

	/* Xe[gZ[u */
	LockVM();
	StopSnd();
	if (bBackupState) {
		/* Xe[g̑Ҕ */
		file_rename_state(StatePath);
	}
	if (!system_save(StatePath)) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_STATESAVEERROR, string, sizeof(string));
		MessageBox(hWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
	}
	else {
		SetMRUFile(MRU_STATE, StatePath);
	}
	PlaySnd();
	ResetSch();
	UnlockVM();
}

#ifdef LPRINT
/*
 *	(P)
 */
static void FASTCALL OnPrint(void)
{
	if (lp_mode == LP_ENABLE) {
		lp_printfile();
	}
}
#endif

/*
 *	Zbg(R)
 */
static void FASTCALL OnReset(HWND hWnd)
{
	LockVM();
	system_reset();
	StatePath[0] = '\0';
	SetAllMRUState();
	ResetSch();
	UnlockVM();

	/* ĕ` */
	OnRefresh(hWnd);
}

/*
 *	zbgZbg(H)
 */
static void FASTCALL OnHotReset(HWND hWnd)
{
	LockVM();
	system_hotreset();
	ResetSch();
	UnlockVM();

	/* ĕ` */
	OnRefresh(hWnd);
}

/*
 *	TAB+Zbg(F)
 */
#if XM7_VER >= 2
static void FASTCALL OnTabReset(HWND hWnd)
{
#if XM7_VER >= 3
	if (!init_is_exsx ||
		((model_type != MODEL_FM77AV40EX) &&
		 (model_type != MODEL_FM77AV20EX))) {
#else
	if (!init_is_exsx ||
		(model_type != MODEL_FM77AV20EX)) {
#endif
		return;
	}

	LockVM();
	system_tabreset();
	ResetSch();
	UnlockVM();

	/* ĕ` */
	OnRefresh(hWnd);
}
#endif

/*
 *	BASIC[h(B)
 */
static void FASTCALL OnBasic(void)
{
	LockVM();
	boot_mode = BOOT_BASIC;
	StoreCfg();
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		mainmem_transfer_boot();
	}
#else
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = TRUE;
	}
#endif
	UnlockVM();
}

/*
 *	DOS[h(D)
 */
static void FASTCALL OnDos(void)
{
	LockVM();
	boot_mode = BOOT_DOS;
	StoreCfg();
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		mainmem_transfer_boot();
	}
#else
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = FALSE;
	}
#endif
	UnlockVM();
}

#if XM7_VER == 1
/*
 *	1MB(SFD) DOS[h(1)
 */
static void FASTCALL OnDosSfd(void)
{
	LockVM();
	boot_mode = BOOT_DOS_SFD;
	StoreCfg();
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = FALSE;
	}
	UnlockVM();
}

/*
 *	BASIC(DEBUG)[h(R)
 */
static void FASTCALL OnBasicRam(void)
{
	LockVM();
	boot_mode = BOOT_BASIC_DBG;
	StoreCfg();
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = FALSE;
	}
	UnlockVM();
}

/*
 *	ou[h(U)
 */
static void FASTCALL OnBubble32(void)
{
	LockVM();
	boot_mode = BOOT_BUBBLE;
	StoreCfg();
	mainmem_bubblemode_boot();
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = FALSE;
	}
	UnlockVM();
}

/*
 *	128KBou[h(L)
 */
static void FASTCALL OnBubble128(void)
{
	LockVM();
	boot_mode = BOOT_BUBBLE128;
	StoreCfg();
	mainmem_bubblemode_boot();
	if ((model_type == MODEL_FM8) &&
		(banksel_en != BANKSEL_ENABLE)) {
		basicrom_en = FALSE;
	}
	UnlockVM();
}
#endif	/* XM7_VER == 1 */

/*
 *	NCbN[h(K)
 */
static void FASTCALL OnQuickLoad(HWND hWnd)
{
	ASSERT(hWnd);

	/* Xe[g[h */
	LockVM();
	StopSnd();
	StateLoad(QuickPath);
	PlaySnd();
	ResetSch();
	UnlockVM();

	/* ʍĕ` */
	OnRefresh(hWnd);

	/* \eXV */
	OnSize(hWnd, 640, 400);
}

/*
 *	NCbNZ[u(Q)
 */
static void FASTCALL OnQuickSave(HWND hWnd)
{
	char string[128];

	/* Xe[gZ[u */
	LockVM();
	StopSnd();
	if (bBackupState) {
		/* Xe[g̑Ҕ */
		file_rename_state(QuickPath);
	}
	if (!system_save(QuickPath)) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_STATESAVEERROR, string, sizeof(string));
		MessageBox(hWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
	}
	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	I(X)
 */
static void FASTCALL OnExit(HWND hWnd)
{
	/* EChEN[Y */
	PostMessage(hWnd, WM_CLOSE, 0, 0);
}

/*
 *	t@C(F)j[
 */
static BOOL OnFile(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	switch (wID) {
		/* I[v */
		case IDM_OPEN:
			OnOpen(hWnd);
			return TRUE;

		/* ㏑ۑ */
		case IDM_SAVE:
			OnSave(hWnd);
			return TRUE;

		/* Oĕۑ */
		case IDM_SAVEAS:
			OnSaveAs(hWnd);
			return TRUE;

#ifdef LPRINT
		/*  */
		case IDM_PRINT:
			OnPrint();
			return TRUE;
#endif

		/* Zbg */
		case IDM_RESET:
			OnReset(hWnd);
			return TRUE;

		/* zbgZbg */
		case IDM_HOTRESET:
			OnHotReset(hWnd);
			return TRUE;

#if XM7_VER >= 2
		/* TAB+Zbg */
		case IDM_TABRESET:
			OnTabReset(hWnd);
			return TRUE;
#endif

		/* BASIC[h */
		case IDM_BASIC:
			OnBasic();
			return TRUE;

		/* DOS[h */
		case IDM_DOS:
			OnDos();
			return TRUE;

#if XM7_VER == 1
		/* 1MB/8" DOS[h */
		case IDM_DOS_SFD:
			OnDosSfd();
			return TRUE;

		/* BASIC(DEBUG)[h */
		case IDM_BASIC_DBG:
			OnBasicRam();
			return TRUE;

		/* BUBBLE[h */
		case IDM_BUBBLE:
			OnBubble32();
			return TRUE;

		/* 128KB BUBBLE[h */
		case IDM_BUBBLE128:
			OnBubble128();
			return TRUE;
#endif	/* XM7_VER == 1 */

		/* NCbN[h */
		case IDM_QLOAD:
			OnQuickLoad(hWnd);
			return TRUE;

		/* NCbNZ[u */
		case IDM_QSAVE:
			OnQuickSave(hWnd);
			return TRUE;

		/* I */
		case IDM_EXIT:
			OnExit(hWnd);
			return TRUE;
	}

	/* XgZbg */
	if ((wID >= IDM_STATE_FILE0) && (wID <= IDM_STATE_FILE9)) {
		char string[256 + 1];
		GetMRUFile(MRU_STATE, wID - IDM_STATE_FILE0, string, sizeof(string));

		LockVM();
		StopSnd();
		StateLoad(string);
		PlaySnd();
		ResetSch();
		UnlockVM();

		/* ʍĕ` */
		OnRefresh(hWnd);

		/* \eXV */
		OnSize(hWnd, 640, 400);

		return TRUE;
	}

	return FALSE;
}

/*
 *	t@C(F)j[XV
 */
static void FASTCALL OnFilePopup(HMENU hMenu)
{
	MENUITEMINFO mii;
	char string[128];
	UINT uitem;
	UINT id;

	ASSERT(hMenu);

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;
	uitem = 12;
#if XM7_VER >= 2
	uitem += 1;		/* TAB+Zbg */
#endif
#if XM7_VER == 1
 #ifdef SFDC
	uitem += 1;		/* 1MB DOS[h/8" DOS[h */
 #endif
 #ifdef BUBBLE
	uitem += 2;		/* ou[h/128KBou[h */
 #endif
	uitem += 1;		/* BASIC(DEBUG)[h */
#endif
#ifdef LPRINT
	uitem += 2;		/* /Zp[^ */
#endif

	/* Zp[^āAȍ~̃j[ׂ͂č폜 */
	while (GetMenuItemCount(hMenu) > (int)uitem) {
		DeleteMenu(hMenu, uitem, MF_BYPOSITION);
	}

	/* } */
	uitem = AppendMenuHistory(hMenu, uitem, IDM_STATE_FILE0, MRU_STATE, TRUE);

	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* Ij[ǉ */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.wID = IDM_EXIT;
	LoadString(hResInstance, IDS_EXIT, string, sizeof(string));
	mii.dwTypeData = string;
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

	/* u[g[hݒ */
	switch (boot_mode) {
		case BOOT_BASIC:
			id = IDM_BASIC;
			break;
		case BOOT_DOS:
			id = IDM_DOS;
			break;
#if XM7_VER == 1 && defined(SFDC)
		case BOOT_DOS_SFD:
			id = IDM_DOS_SFD;
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1
		case BOOT_BASIC_DBG:
			id = IDM_BASIC_DBG;
			break;
#endif	/* XM7_VER == 1 */
#if XM7_VER == 1 && defined(BUBBLE)
		case BOOT_BUBBLE:
			id = IDM_BUBBLE;
			break;
		case BOOT_BUBBLE128:
			id = IDM_BUBBLE128;
			break;
#endif
		default:
			ASSERT(FALSE);
			break;
	}
#if XM7_VER == 1
	CheckMenuRadioItem(hMenu, IDM_BASIC, IDM_BASIC_DBG, id, MF_BYCOMMAND);
#else
	CheckMenuRadioItem(hMenu, IDM_BASIC, IDM_DOS, id, MF_BYCOMMAND);
#endif

	/* Xe[gۑ@\ */
	EnableMenuSub(hMenu, IDM_SAVE, bStateSave);
	EnableMenuSub(hMenu, IDM_SAVEAS, bStateSave);
	EnableMenuSub(hMenu, IDM_QLOAD, bStateSave);
	EnableMenuSub(hMenu, IDM_QSAVE, bStateSave);

#if XM7_VER == 1
#ifdef SFDC
	EnableMenuSub(hMenu, IDM_DOS_SFD, availableSFDBOOT);
#endif
	if ((model_type == MODEL_FM8)) {
		EnableMenuSub(hMenu, IDM_BASIC_DBG, available_dbgboot8);
	}
	else {
		EnableMenuSub(hMenu, IDM_BASIC_DBG, (model_type == MODEL_FM77));
	}
#ifdef BUBBLE
	EnableMenuSub(hMenu, IDM_BUBBLE, availableBBLBOOT);
	EnableMenuSub(hMenu, IDM_BUBBLE128, availableBBLBOOT128);
#endif
#endif	/* XM7_VER == 1 */

#if XM7_VER >= 3
	EnableMenuSub(hMenu, IDM_TABRESET,	((model_type == MODEL_FM77AV40EX) ||
										 (model_type == MODEL_FM77AV20EX)) &&
										init_is_exsx);
#elif XM7_VER >= 2
	EnableMenuSub(hMenu, IDM_TABRESET,	(model_type == MODEL_FM77AV20EX) &&
										init_is_exsx);
#endif

#ifdef LPRINT
	EnableMenuSub(hMenu, IDM_PRINT,		(lp_mode == LP_ENABLE));
#endif
}

/*
 *	t@C(F)ReLXgj[
 */
static BOOL FASTCALL OnFileContextualMenu(HWND hWnd, WORD wID, int x, int y)
{
	ASSERT(hWnd);

	/* qXg */
	if ((wID >= IDM_STATE_FILE0) && (wID <= IDM_STATE_FILE9)) {
		OnContextualSubMenu(hWnd, hHistoryMenu, wID, x, y);
		return TRUE;
	}

	return FALSE;
}

/*-[ fBXNj[ ]-----------------------------------------------------*/

/*
 *	hCuJ
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskOpen(int Type, int Drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskOpen(int Drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char path[_MAX_PATH];

#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((Drive >= 0) || (Drive <= 3));

	/* t@CI */
#if XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_M) {
		if (!FileSelectSub(TRUE, IDS_DISKFILTER, path, NULL, DISK_IMAGE_DIR)) {
			return;
		}
	}
	else {
		if (!FileSelectSub(TRUE, IDS_DISKFILTER_SFD, path, NULL, DISK_IMAGE_DIR)) {
			return;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (!FileSelectSub(TRUE, IDS_DISKFILTER, path, NULL, DISK_IMAGE_DIR)) {
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* Zbg */
	LockVM();
#if XM7_VER == 1 && defined(SFDC)
	fdc_setdisk(Type, Drive, path);
#else	/* XM7_VER == 1 && defined(SFDC) */
	fdc_setdisk(Drive, path);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetAllMRUState();
	ResetSch();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	hCuJ
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskBoth(int Type, int Drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskBoth(int Drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char path[_MAX_PATH];

	/* t@CI */
#if XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_M) {
		if (!FileSelectSub(TRUE, IDS_DISKFILTER, path, NULL, DISK_IMAGE_DIR)) {
			return;
		}
	}
	else {
		if (!FileSelectSub(TRUE, IDS_DISKFILTER_SFD, path, NULL, DISK_IMAGE_DIR)) {
			return;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (!FileSelectSub(TRUE, IDS_DISKFILTER, path, NULL, DISK_IMAGE_DIR)) {
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* Zbg */
	LockVM();
#if XM7_VER == 1 && defined(SFDC)
	fdc_setdisk(Type, Drive + 0, path);
	fdc_setdisk(Type, Drive + 1, NULL);
	if ((fdc_ready[Type][Drive + 0] != FDC_TYPE_NOTREADY) && (fdc_medias[Type][Drive + 0] >= 2)) {
		fdc_setdisk(Type, Drive + 1, path);
		fdc_setmedia(Type, Drive + 1, 1);
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	fdc_setdisk(Drive + 0, path);
	fdc_setdisk(Drive + 1, NULL);
	if ((fdc_ready[Drive + 0] != FDC_TYPE_NOTREADY) && (fdc_medias[Drive + 0] >= 2)) {
		fdc_setdisk(Drive + 1, path);
		fdc_setmedia(Drive + 1, 1);
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetAllMRUState();
	ResetSch();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	hCu
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskReplace(int Type, int Drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskReplace(int Drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char cfname[256+1];	/* t@C */
	BYTE nmedia;		/* fBAZNg */

#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((Drive >= 0) || (Drive <= 3));

	/* Zbg */
	LockVM();
#if XM7_VER == 1 && defined(SFDC)
	if (fdc_ready[Type][Drive + 0] == FDC_TYPE_NOTREADY) {
		memset(cfname, 0, sizeof(cfname));
	}
	else {
		strncpy(cfname, fdc_fname[Type][Drive + 0], sizeof(cfname));
	}
	nmedia = fdc_media[Type][Drive + 0];
	if (fdc_ready[Type][Drive + 1] == FDC_TYPE_NOTREADY) {
		fdc_setdisk(Type, Drive + 0, NULL);
	}
	else {
		fdc_setdisk(Type, Drive + 0, fdc_fname[Type][Drive + 1]);
		fdc_setmedia(Type, Drive + 0, fdc_media[Type][Drive + 1]);
	}
	if (strlen(cfname) == 0) {
		fdc_setdisk(Type, Drive + 1, NULL);
	}
	else {
		fdc_setdisk(Type, Drive + 1, cfname);
		fdc_setmedia(Type, Drive + 1, nmedia);
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (fdc_ready[Drive + 0] == FDC_TYPE_NOTREADY) {
		memset(cfname, 0, sizeof(cfname));
	}
	else {
		strncpy(cfname, fdc_fname[Drive + 0], sizeof(cfname));
	}
	nmedia = fdc_media[Drive + 0];
	if (fdc_ready[Drive + 1] == FDC_TYPE_NOTREADY) {
		fdc_setdisk(Drive + 0, NULL);
	}
	else {
		fdc_setdisk(Drive + 0, fdc_fname[Drive + 1]);
		fdc_setmedia(Drive + 0, fdc_media[Drive + 1]);
	}
	if (strlen(cfname) == 0) {
		fdc_setdisk(Drive + 1, NULL);
	}
	else {
		fdc_setdisk(Drive + 1, cfname);
		fdc_setmedia(Drive + 1, nmedia);
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetAllMRUState();
	ResetSch();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	fBXNCWFNg
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskEject(int Type, int Drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskEject(int Drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((Drive >= 0) || (Drive <= 3));

	/* CWFNg */
	LockVM();
#if XM7_VER == 1 && defined(SFDC)
	fdc_setdisk(Type, Drive, NULL);
#else	/* XM7_VER == 1 && defined(SFDC) */
	fdc_setdisk(Drive, NULL);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	SetAllMRUState();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	fBXN݋֎~
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskProtect(int Type, int Drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskProtect(int Drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((Drive >= 0) || (Drive <= 3));

	/* ݋֎~؂ւ */
	LockVM();
#if XM7_VER == 1 && defined(SFDC)
	if (fdc_writep[Type][Drive]) {
		fdc_setwritep(Type, Drive, FALSE);
	}
	else {
		fdc_setwritep(Type, Drive, TRUE);
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (fdc_writep[Drive]) {
		fdc_setwritep(Drive, FALSE);
	}
	else {
		fdc_setwritep(Drive, TRUE);
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ResetSch();
	UnlockVM();
}

/*
 *	SfBXNCWFNg
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskDriveEject(int Type)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskDriveEject(void)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	int	i;

#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* CWFNg */
	LockVM();
	for (i=0; i<4; i++) {
#if XM7_VER == 1 && defined(SFDC)
		fdc_setdisk(Type, i, NULL);
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setdisk(i, NULL);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}
	SetAllMRUState();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	fBXN(1)(0)j[
 */
static BOOL FASTCALL OnDisk(WORD wID)
{
	switch (wID) {
		/* J */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_D0OPEN:
			OnDiskOpen(FDC_M, 0);
			break;
		case IDM_D1OPEN:
			OnDiskOpen(FDC_M, 1);
			break;
		case IDM_D2OPEN:
			OnDiskOpen(FDC_M, 2);
			break;
		case IDM_D3OPEN:
			OnDiskOpen(FDC_M, 3);
			break;
		case IDM_D4OPEN:
			OnDiskOpen(FDC_S, 0);
			break;
		case IDM_D5OPEN:
			OnDiskOpen(FDC_S, 1);
			break;
		case IDM_D6OPEN:
			OnDiskOpen(FDC_S, 2);
			break;
		case IDM_D7OPEN:
			OnDiskOpen(FDC_S, 3);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		case IDM_D0OPEN:
			OnDiskOpen(0);
			break;
		case IDM_D1OPEN:
			OnDiskOpen(1);
			break;
		case IDM_D2OPEN:
			OnDiskOpen(2);
			break;
		case IDM_D3OPEN:
			OnDiskOpen(3);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */

		/* hCuŊJ */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_DBOPEN:
			OnDiskBoth(FDC_M, 0);
			break;
		case IDM_DB23OPEN:
			OnDiskBoth(FDC_M, 2);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		case IDM_DBOPEN:
			OnDiskBoth(0);
			break;
		case IDM_DB23OPEN:
			OnDiskBoth(2);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_DB45OPEN:
			OnDiskBoth(FDC_S, 0);
			break;
		case IDM_DB67OPEN:
			OnDiskBoth(FDC_S, 2);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */

		/* hCu */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_DBREPLACE:
			OnDiskReplace(FDC_M, 0);
			break;
		case IDM_DB23REPLACE:
			OnDiskReplace(FDC_M, 2);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		case IDM_DBREPLACE:
			OnDiskReplace(0);
			break;
		case IDM_DB23REPLACE:
			OnDiskReplace(2);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_DB45REPLACE:
			OnDiskReplace(FDC_S, 0);
			break;
		case IDM_DB67REPLACE:
			OnDiskReplace(FDC_S, 2);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */

		/* O */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_D0EJECT:
			OnDiskEject(FDC_M, 0);
			break;
		case IDM_D1EJECT:
			OnDiskEject(FDC_M, 1);
			break;
		case IDM_D2EJECT:
			OnDiskEject(FDC_M, 2);
			break;
		case IDM_D3EJECT:
			OnDiskEject(FDC_M, 3);
			break;
		case IDM_MFD_EJECT:
			OnDiskDriveEject(FDC_M);
			break;
		case IDM_D4EJECT:
		case IDM_D4FORCE:
			OnDiskEject(FDC_S, 0);
			break;
		case IDM_D5EJECT:
		case IDM_D5FORCE:
			OnDiskEject(FDC_S, 1);
			break;
		case IDM_D6EJECT:
		case IDM_D6FORCE:
			OnDiskEject(FDC_S, 2);
			break;
		case IDM_D7EJECT:
		case IDM_D7FORCE:
			OnDiskEject(FDC_S, 3);
			break;
		case IDM_SFD_EJECT:
			OnDiskDriveEject(FDC_S);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		case IDM_D0EJECT:
			OnDiskEject(0);
			break;
		case IDM_D1EJECT:
			OnDiskEject(1);
			break;
		case IDM_D2EJECT:
			OnDiskEject(2);
			break;
		case IDM_D3EJECT:
			OnDiskEject(3);
			break;
		case IDM_MFD_EJECT:
			OnDiskDriveEject();
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */

		/* ݋֎~ */
#if XM7_VER == 1 && defined(SFDC)
		case IDM_D0WRITE:
			OnDiskProtect(FDC_M, 0);
			break;
		case IDM_D1WRITE:
			OnDiskProtect(FDC_M, 1);
			break;
		case IDM_D2WRITE:
			OnDiskProtect(FDC_M, 2);
			break;
		case IDM_D3WRITE:
			OnDiskProtect(FDC_M, 3);
			break;
		case IDM_D4WRITE:
			OnDiskProtect(FDC_S, 0);
			break;
		case IDM_D5WRITE:
			OnDiskProtect(FDC_S, 1);
			break;
		case IDM_D6WRITE:
			OnDiskProtect(FDC_S, 2);
			break;
		case IDM_D7WRITE:
			OnDiskProtect(FDC_S, 3);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		case IDM_D0WRITE:
			OnDiskProtect(0);
			break;
		case IDM_D1WRITE:
			OnDiskProtect(1);
			break;
		case IDM_D2WRITE:
			OnDiskProtect(2);
			break;
		case IDM_D3WRITE:
			OnDiskProtect(3);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}

	/* fBA */
	if ((wID >= IDM_D0MEDIA00) && (wID <= IDM_D0MEDIA15)) {
		LockVM();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setmedia(FDC_M, 0, wID - IDM_D0MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setmedia(0, wID - IDM_D0MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D1MEDIA00) && (wID <= IDM_D1MEDIA15)) {
		LockVM();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setmedia(FDC_M, 1, wID - IDM_D1MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setmedia(1, wID - IDM_D1MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D2MEDIA00) && (wID <= IDM_D2MEDIA15)) {
		LockVM();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setmedia(FDC_M, 2, wID - IDM_D2MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setmedia(2, wID - IDM_D2MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D3MEDIA00) && (wID <= IDM_D3MEDIA15)) {
		LockVM();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setmedia(FDC_M, 3, wID - IDM_D3MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setmedia(3, wID - IDM_D3MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((wID >= IDM_D4MEDIA00) && (wID <= IDM_D4MEDIA15)) {
		LockVM();
		fdc_setmedia(FDC_S, 0, wID - IDM_D4MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D5MEDIA00) && (wID <= IDM_D5MEDIA15)) {
		LockVM();
		fdc_setmedia(FDC_S, 1, wID - IDM_D5MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D6MEDIA00) && (wID <= IDM_D6MEDIA15)) {
		LockVM();
		fdc_setmedia(FDC_S, 2, wID - IDM_D6MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_D7MEDIA00) && (wID <= IDM_D7MEDIA15)) {
		LockVM();
		fdc_setmedia(FDC_S, 3, wID - IDM_D7MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* XgZbg */
#if XM7_VER == 1 && defined(SFDC)
	if (((wID >= IDM_D3_FILE0) && (wID <= IDM_D0_FILE9)) ||
		((wID >= IDM_D7_FILE0) && (wID <= IDM_D4_FILE9))) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	if ((wID >= IDM_D3_FILE0) && (wID <= IDM_D0_FILE9)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		char path[256 + 1];
#if XM7_VER == 1 && defined(SFDC)
		int type;
#endif	/* XM7_VER == 1 && defined(SFDC) */
		int drive;
		int mrutype;
		int index;
		int media;

		if ((wID >= IDM_D0_FILE0) && (wID <= IDM_D0_FILE9)) {
#if XM7_VER == 1 && defined(SFDC)
			type	= FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			drive	= 0;
			mrutype	= MRU_MFD0;
			index	= wID - IDM_D0_FILE0;
		}
		if ((wID >= IDM_D1_FILE0) && (wID <= IDM_D1_FILE9)) {
#if XM7_VER == 1 && defined(SFDC)
			type	= FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			drive	= 1;
			mrutype	= MRU_MFD1;
			index	= wID - IDM_D1_FILE0;
		}
		if ((wID >= IDM_D2_FILE0) && (wID <= IDM_D2_FILE9)) {
#if XM7_VER == 1 && defined(SFDC)
			type	= FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			drive	= 2;
			mrutype	= MRU_MFD2;
			index	= wID - IDM_D2_FILE0;
		}
		if ((wID >= IDM_D3_FILE0) && (wID <= IDM_D3_FILE9)) {
#if XM7_VER == 1 && defined(SFDC)
			type	= FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			drive	= 3;
			mrutype	= MRU_MFD3;
			index	= wID - IDM_D3_FILE0;
		}
#if XM7_VER == 1 && defined(SFDC)
		if ((wID >= IDM_D4_FILE0) && (wID <= IDM_D4_FILE9)) {
			type	= FDC_S;
			drive	= 0;
			mrutype	= MRU_SFD0;
			index	= wID - IDM_D4_FILE0;
		}
		if ((wID >= IDM_D5_FILE0) && (wID <= IDM_D5_FILE9)) {
			type	= FDC_S;
			drive	= 1;
			mrutype	= MRU_SFD1;
			index	= wID - IDM_D5_FILE0;
		}
		if ((wID >= IDM_D6_FILE0) && (wID <= IDM_D6_FILE9)) {
			type	= FDC_S;
			drive	= 2;
			mrutype	= MRU_SFD2;
			index	= wID - IDM_D6_FILE0;
		}
		if ((wID >= IDM_D7_FILE0) && (wID <= IDM_D7_FILE9)) {
			type	= FDC_S;
			drive	= 3;
			mrutype	= MRU_SFD3;
			index	= wID - IDM_D7_FILE0;
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */

		GetMRUFile(mrutype, index, path, sizeof(path));
		media = GetMRUMedia(mrutype, index);

		LockVM();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setdisk(type, drive, path);
		if (fdc_ready[type][drive] != FDC_TYPE_NOTREADY) {
			if (fdc_ready[type][drive] == FDC_TYPE_D77) {
				if (media >= 0) {
					fdc_setmedia(type, drive, media);
				}
			}
		}
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setdisk(drive, path);
		if (fdc_ready[drive] != FDC_TYPE_NOTREADY) {
			if (fdc_ready[drive] == FDC_TYPE_D77) {
				if (media >= 0) {
					fdc_setmedia(drive, media);
				}
			}
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
		else {
			FileMessageBox(hMainWnd, IDS_FILE_NOT_FOUND, path, MB_ICONSTOP | MB_OK);
		}
		SetAllMRUState();
		ResetSch();
		UnlockVM();

		/* \eXV */
		if (hMainWnd) {
			OnSize(hMainWnd, 640, 400);
		}
	}

	return FALSE;
}

/*
 *	fBXN(1)(0)j[XV
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnDiskPopup(HMENU hMenu, int Type)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnDiskPopup(HMENU hMenu)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	HMENU hSubMenu;
	MENUITEMINFO mii;
	char string[256];
	char buffer[256];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	UINT uitem;
	int drive, offset, i, j, k;
	BOOL enable;

	ASSERT(hMenu);
#if XM7_VER == 1 && defined(SFDC)
	ASSERT((Type == FDC_M) || (Type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */

	enable = TRUE;

#if 0
	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}
#endif

	/* ItZbgm */
#if 0//XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_S) {
		enable = fdc_enable[FDC_S];
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fType = MFT_STRING;
	if (enable) {
		mii.fState = MFS_ENABLED;
	}
	else {
		mii.fState = MFS_GRAYED;
	}

	for(drive = 0; drive < 4; drive ++) {

#if 0
		/* ItZbgm */
		offset = (IDM_D1OPEN - IDM_D0OPEN) * drive;
#if XM7_VER == 1 && defined(SFDC)
		if (Type != FDC_M) {
			offset += IDM_D4OPEN - IDM_D0OPEN;
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
#endif

		/* hCun */
#if 0
		if (!enable ||
#if XM7_VER == 1 && defined(SFDC)
			fdc_ready[Type][drive] == FDC_TYPE_NOTREADY) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			fdc_ready[drive] == FDC_TYPE_NOTREADY) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
			mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
			mii.wID = IDM_D0OPEN + offset;
			LoadString(hResInstance, IDS_DISK0OPEN + drive, string, sizeof(string));
		}
		else
#endif
		{
#if 0
			if (!enable) {
				mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
			}
			else
#endif
			{
				mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
			}
			mii.wID = 0;
			//mii.hSubMenu = CreatePopupMenu();
			mii.hSubMenu = GetSubMenu(hMenu, drive);
			LoadString(hResInstance, IDS_DISK0 + drive, buffer, sizeof(buffer));
			if ((enable) &&
 #if XM7_VER == 1 && defined(SFDC)
				(fdc_ready[Type][drive] != FDC_TYPE_NOTREADY)) {
				_splitpath(fdc_fname[Type][drive], NULL, NULL, fname, ext);
 #else	/* XM7_VER == 1 && defined(SFDC) */
				(fdc_ready[drive] != FDC_TYPE_NOTREADY)) {
				_splitpath(fdc_fname[drive], NULL, NULL, fname, ext);
 #endif	/* XM7_VER == 1 && defined(SFDC) */
				_snprintf(string, sizeof(string), "%s - %s%s", buffer, fname,
						  ext);
			}
			else {
				strncpy(string, buffer, sizeof(string));
			}
		}
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		RemoveMenu(hMenu, drive, MF_BYPOSITION);
		InsertMenuItem(hMenu, drive, TRUE, &mii);
	}

#if 0
	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, 4, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* I[v */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
#if XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_M) {
		mii.wID = IDM_DBOPEN;
	}
	else {
		mii.wID = IDM_DB45OPEN;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	mii.wID = IDM_DBOPEN;
#endif	/* XM7_VER == 1 && defined(SFDC) */
	LoadString(hResInstance, IDS_DISKBOTH, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, 5, TRUE, &mii);

	/* 2,3I[v */
#if XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_M) {
		mii.wID = IDM_DB23OPEN;
	}
	else {
		mii.wID = IDM_DB67OPEN;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	mii.wID = IDM_DB23OPEN;
#endif	/* XM7_VER == 1 && defined(SFDC) */
	LoadString(hResInstance, IDS_DISKBOTH23, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, 6, TRUE, &mii);

	/* CWFNg */
#if XM7_VER == 1 && defined(SFDC)
	if (Type == FDC_M) {
		mii.wID = IDM_MFD_EJECT;
		LoadString(hResInstance, IDS_MFD_EJECT, string, sizeof(string));
	}
	else {
		mii.wID = IDM_SFD_EJECT;
		LoadString(hResInstance, IDS_SFD_EJECT, string, sizeof(string));
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	mii.wID = IDM_MFD_EJECT;
	LoadString(hResInstance, IDS_MFD_EJECT, string, sizeof(string));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, 7, TRUE, &mii);
#endif

	/* Tuj[ */
	for(drive = 0; drive < 4; drive ++) {

#if 0
#if 0
		/* fBXN}ĂȂ΁A܂ */
#if XM7_VER == 1 && defined(SFDC)
		if (fdc_ready[Type][drive] == FDC_TYPE_NOTREADY || !enable) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (fdc_ready[drive] == FDC_TYPE_NOTREADY || !enable) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
#else
		if (!enable) {
#endif
			continue;
		}
#endif

		hSubMenu = GetSubMenu(hMenu, drive);

		/* j[ׂč폜 */
		while (GetMenuItemCount(hSubMenu) > 0) {
			DeleteMenu(hSubMenu, 0, MF_BYPOSITION);
		}

		/* ItZbgm */
		offset = (IDM_D1OPEN - IDM_D0OPEN) * drive;
#if XM7_VER == 1 && defined(SFDC)
		if (Type != FDC_M) {
			offset += IDM_D4OPEN - IDM_D0OPEN;
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
		uitem = 0;

		/* j[\̏ */
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.fType = MFT_STRING;

		/* I[v */
		mii.wID = IDM_D0OPEN + offset;
		LoadString(hResInstance, IDS_DISKOPEN, string, sizeof(string));
		mii.dwTypeData = string;
#if XM7_VER == 1 && defined(SFDC)
		if ((Type == FDC_S) && (fdc_doorlock[drive])) {
			mii.fState = MFS_GRAYED;
		}
		else {
			mii.fState = MFS_ENABLED;
		}
#else	/* XM7_VER == 1 && defined(SFDC) */
		mii.fState = MFS_ENABLED;
#endif	/* XM7_VER == 1 && defined(SFDC) */
		mii.cch = strlen(string);
		InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

		/* CWFNg */
		mii.wID = IDM_D0EJECT + offset;
#if XM7_VER == 1 && defined(SFDC)
		if ((fdc_ready[Type][drive] == FDC_TYPE_NOTREADY) ||
			((Type == FDC_S) && (fdc_doorlock[drive]))) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
			mii.fState = MFS_GRAYED;
		}
		else {
			mii.fState = MFS_ENABLED;
		}
		LoadString(hResInstance, IDS_DISKEJECT, string, sizeof(string));
		mii.cch = strlen(string);
		InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

#if 0
		/* Zp[^} */
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

		/* ꎞo */
		mii.wID = IDM_D0TEMP + offset;
		LoadString(hResInstance, IDS_DISKTEMP, string, sizeof(string));
#if XM7_VER == 1 && defined(SFDC)
		if ((fdc_ready[Type][drive] == FDC_TYPE_NOTREADY) ||
			((Type == FDC_S) && (fdc_doorlock[drive]))) {
			mii.fState = MFS_GRAYED;
		}
		else if (fdc_teject[Type][drive]) {
			mii.fState = MFS_CHECKED | MFS_ENABLED;
		}
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
			mii.fState = MFS_GRAYED;
		}
		else if (fdc_teject[drive]) {
			mii.fState = MFS_CHECKED | MFS_ENABLED;
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
		else {
			mii.fState = MFS_ENABLED;
		}
		mii.fType = MFT_STRING;
		mii.cch = strlen(string);
		InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
#endif

		/* CgveNg */
		mii.wID = IDM_D0WRITE + offset;
		LoadString(hResInstance, IDS_DISKPROTECT, string, sizeof(string));
#if XM7_VER == 1 && defined(SFDC)
		if (fdc_ready[Type][drive] == FDC_TYPE_NOTREADY) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
			mii.fState = MFS_GRAYED;
		}
		else {
#if XM7_VER == 1 && defined(SFDC)
			if (((Type == FDC_S) && (fdc_doorlock[drive])) ||
				(fdc_fwritep[Type][drive])) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			if (fdc_fwritep[drive]) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
				mii.fState = MFS_GRAYED;
			}
			else {
				mii.fState = MFS_ENABLED;
			}
#if XM7_VER == 1 && defined(SFDC)
			if ((fdc_fwritep[Type][drive]) || (fdc_writep[Type][drive])) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			if ((fdc_fwritep[drive]) || (fdc_writep[drive])) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
				mii.fState |= MFS_CHECKED;
			}
		}
		mii.fType = MFT_STRING;
		mii.cch = strlen(string);
		InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

#if XM7_VER == 1 && defined(SFDC)
		if (Type == FDC_S) {
			/* Zp[^} */
			mii.fType = MFT_SEPARATOR;
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

			/* CWFNg */
			mii.wID = IDM_D0FORCE + offset;
			LoadString(hResInstance, IDS_DISKFORCE, string, sizeof(string));
			if ((fdc_ready[Type][drive] == FDC_TYPE_NOTREADY) ||
				(!fdc_doorlock[drive])) {
				mii.fState = MFS_GRAYED;
			}
			else {
				mii.fState = MFS_ENABLED;
			}
			mii.fType = MFT_STRING;
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */

#if 0	/* 2D/2DD/2HD/VFDt@C͏wɕ\̂ō폜 */
		/* 2D/2DD/2HD/VFDȂꏈ */
#if XM7_VER == 1 && defined(SFDC)
		if ((fdc_ready[Type][drive] == FDC_TYPE_2D) ||
			(fdc_ready[Type][drive] == FDC_TYPE_2HD) ||
			(fdc_ready[Type][drive] == FDC_TYPE_VFD)) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if ((fdc_ready[drive] == FDC_TYPE_2D) ||
#if XM7_VER >= 3
			(fdc_ready[drive] == FDC_TYPE_2DD) ||
#endif
			(fdc_ready[drive] == FDC_TYPE_VFD)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */

			/* Zp[^} */
			mii.fType = MFT_SEPARATOR;
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

			mii.wID = IDM_D0MEDIA00 + offset;
			mii.fState = MFS_CHECKED | MFS_ENABLED;
			mii.fType = MFT_STRING | MFT_RADIOCHECK;
#if XM7_VER == 1 && defined(SFDC)
			if (fdc_ready[Type][drive] == FDC_TYPE_2D) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			if (fdc_ready[drive] == FDC_TYPE_2D) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
				LoadString(hResInstance, IDS_DISK2D, string, sizeof(string));
			}
#if XM7_VER == 1 && defined(SFDC)
			else if (fdc_ready[Type][drive] == FDC_TYPE_2HD) {
				LoadString(hResInstance, IDS_DISK2HD, string, sizeof(string));
			}
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER >= 3
			else if (fdc_ready[drive] == FDC_TYPE_2DD) {
				LoadString(hResInstance, IDS_DISK2DD, string, sizeof(string));
			}
#endif
			else {
				LoadString(hResInstance, IDS_DISKVFD, string, sizeof(string));
			}
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
		}
#endif

		/* D77Ȃ烁fBA */
#if XM7_VER == 1 && defined(SFDC)
		if (fdc_ready[Type][drive] == FDC_TYPE_D77) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (fdc_ready[drive] == FDC_TYPE_D77) {
#endif	/* XM7_VER == 1 && defined(SFDC) */

			/* Zp[^} */
			mii.fType = MFT_SEPARATOR;
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

#if XM7_VER == 1 && defined(SFDC)
			for (i=0; i<fdc_medias[Type][drive]; i++) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			for (i=0; i<fdc_medias[drive]; i++) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
				mii.wID = IDM_D0MEDIA00 + offset + i;
#if XM7_VER == 1 && defined(SFDC)
				if ((Type == FDC_S) && fdc_doorlock[drive]) {
					mii.fState = MFS_GRAYED;
				}
				else {
					mii.fState = MFS_ENABLED;
				}
				if (fdc_media[Type][drive] == i) {
					mii.fState |= MFS_CHECKED;
				}
#else	/* XM7_VER == 1 && defined(SFDC) */
				mii.fState = MFS_ENABLED;
				if (fdc_media[drive] == i) {
					mii.fState |= MFS_CHECKED;
				}
#endif	/* XM7_VER == 1 && defined(SFDC) */
				mii.fType = MFT_STRING | MFT_RADIOCHECK;
#if XM7_VER == 1 && defined(SFDC)
				if (strlen(fdc_name[Type][drive][i]) == 0) {
#else	/* XM7_VER == 1 && defined(SFDC) */
				if (strlen(fdc_name[drive][i]) == 0) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
					LoadString(hResInstance, IDS_MEDIA_NAME, buffer, sizeof(buffer));
					_snprintf(string, sizeof(string), buffer, i + 1);
				}
				else {
					/* vtBbNX΍ */
					k = 0;
#if XM7_VER == 1 && defined(SFDC)
					for (j=0; j<(int)strlen(fdc_name[Type][drive][i]); j++) {
						if (fdc_name[Type][drive][i][j] == '&') {
							string[k++] = '&';
						}
						string[k++] = fdc_name[Type][drive][i][j];
					}
#else	/* XM7_VER == 1 && defined(SFDC) */
					for (j=0; j<(int)strlen(fdc_name[drive][i]); j++) {
						if (fdc_name[drive][i][j] == '&') {
							string[k++] = '&';
						}
						string[k++] = fdc_name[drive][i][j];
					}
#endif	/* XM7_VER == 1 && defined(SFDC) */
					string[k] = '\0';
				}
				mii.cch = strlen(string);
				InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
			}
		}

		/* } */
		{
			UINT uID = IDM_D0_FILE0 - drive * 10;
			int type = MRU_MFD0 + drive;
			BOOL enable2 = enable;
#if XM7_VER == 1 && defined(SFDC)
			if (Type == FDC_S) {
				uID = IDM_D4_FILE0 - drive * 10;
				type = MRU_SFD0 + drive;
				if (fdc_doorlock[drive]) {
					enable2 = FALSE;
				}
			}
#endif	/* XM7_VER == 1 && defined(SFDC) */
			AppendMenuHistory(hSubMenu, uitem, uID, type, enable2);
		}
	}
}

/*
 *	fBXN(1)(0)ReLXgj[
 */
static BOOL FASTCALL OnDiskContextualMenu(HWND hWnd, WORD wID, int x, int y)
{
	ASSERT(hWnd);

	/* fBXNENbNj[ */
	if ((wID >= IDM_D0MEDIA00) && (wID <= IDM_D0MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D1MEDIA00) && (wID <= IDM_D1MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D2MEDIA00) && (wID <= IDM_D2MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D3MEDIA00) && (wID <= IDM_D3MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((wID >= IDM_D4MEDIA00) && (wID <= IDM_D4MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D5MEDIA00) && (wID <= IDM_D5MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D6MEDIA00) && (wID <= IDM_D6MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_D7MEDIA00) && (wID <= IDM_D7MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* qXg */
	if ((wID >= IDM_D3_FILE0) && (wID <= IDM_D0_FILE9)) {
		OnContextualSubMenu(hWnd, hHistoryMenu, wID, x, y);
		return TRUE;
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((wID >= IDM_D7_FILE0) && (wID <= IDM_D4_FILE9)) {
		OnContextualSubMenu(hWnd, hHistoryMenu, wID, x, y);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	return FALSE;
}

/*-[ e[vj[ ]-------------------------------------------------------*/

/*
 *  e[vI[v
 */
static void FASTCALL OnTapeOpen(void)
{
	char path[_MAX_PATH];

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_TAPEFILTER, path, NULL, TAPE_IMAGE_DIR)) {
		return;
	}

	/* Zbg */
	LockVM();
	tape_setfile(path);
	SetAllMRUState();
	ResetSch();
	UnlockVM();
}

/*
 *	e[vCWFNg
 */
static void FASTCALL OnTapeEject(void)
{
	/* CWFNg */
	LockVM();
	tape_setfile(NULL);
	SetAllMRUState();
	UnlockVM();
}

/*
 *	߂
 */
static void FASTCALL OnRew(void)
{
	HCURSOR hCursor;

	/* ߂ */
	hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	LockVM();
	StopSnd();

	tape_rew();

	PlaySnd();
	ResetSch();
	UnlockVM();
	SetCursor(hCursor);
}

/*
 *	ŏ܂Ŋ߂
 */
static void FASTCALL OnRewTop(void)
{
	LockVM();
	StopSnd();

	tape_rewtop();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	
 */
static void FASTCALL OnFF(void)
{
	HCURSOR hCursor;

	/* ߂ */
	hCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	LockVM();
	StopSnd();

	tape_ff();

	PlaySnd();
	ResetSch();
	UnlockVM();
	SetCursor(hCursor);
}

/*
 *	^
 */
static void FASTCALL OnRec(void)
{
	/* ^ */
	LockVM();
	if (tape_rec) {
		tape_setrec(FALSE);
	}
	else {
		tape_setrec(TRUE);
	}
	UnlockVM();
}

#if XM7_VER == 1 && defined(BUBBLE)
/*-[ ouj[ ]-------------------------------------------------------*/

/*
 *  ouI[v
 */
static void FASTCALL OnBubbleOpen(int Type, int Unit)
{
	char path[_MAX_PATH];

	ASSERT((Type == BMC_32) || (Type == BMC_128));
	ASSERT((Unit == 0) || (Unit == 1));

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_BUBBLEFILTER, path, NULL, BUBBLE_IMAGE_DIR)) {
		return;
	}

	/* Zbg */
	LockVM();
	bmc_setfile(Type, Unit, path);
	SetAllMRUState();
	ResetSch();
	UnlockVM();
}

/*
 *  jbgŊJ
 */
static void FASTCALL OnBubbleBoth(int Type)
{
	char path[_MAX_PATH];

	ASSERT((Type == BMC_32) || (Type == BMC_128));

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_BUBBLEFILTER, path, NULL, BUBBLE_IMAGE_DIR)) {
		return;
	}

	/* Zbg */
	LockVM();
	bmc_setfile(Type, 0, path);
	bmc_setfile(Type, 1, NULL);
	if ((bmc_ready[Type][0] != BMC_TYPE_NOTREADY) && (bmc_medias[Type][0] >= 2)) {
		bmc_setfile(Type, 1, path);
		bmc_setmedia(Type, 1, 1);
	}
	SetAllMRUState();
	ResetSch();
	UnlockVM();
}

/*
 *	jbg
 */
static void FASTCALL OnBubbleReplace(int Type, int Unit)
{
	char cfname[256+1];	/* t@C */
	BYTE nmedia;		/* fBAZNg */

	ASSERT((Type == BMC_32) || (Type == BMC_128));
	ASSERT((Unit == 0) || (Unit == 1));

	/* Zbg */
	LockVM();
	if (bmc_ready[Type][Unit + 0] == BMC_TYPE_NOTREADY) {
		memset(cfname, 0, sizeof(cfname));
	}
	else {
		strncpy(cfname, bmc_fname[Type][Unit + 0], sizeof(cfname));
	}
	nmedia = bmc_media[Type][Unit + 0];
	if (bmc_ready[Type][Unit + 1] == BMC_TYPE_NOTREADY) {
		bmc_setfile(Type, Unit + 0, NULL);
	}
	else {
		bmc_setfile(Type, Unit + 0, bmc_fname[Type][Unit + 1]);
		bmc_setmedia(Type, Unit + 0, bmc_media[Type][Unit + 1]);
	}
	if (strlen(cfname) == 0) {
		bmc_setfile(Type, Unit + 1, NULL);
	}
	else {
		bmc_setfile(Type, Unit + 1, cfname);
		bmc_setmedia(Type, Unit + 1, nmedia);
	}
	SetAllMRUState();
	ResetSch();
	UnlockVM();

	/* \eXV */
	if (hMainWnd) {
		OnSize(hMainWnd, 640, 400);
	}
}

/*
 *	ouCWFNg
 */
static void FASTCALL OnBubbleEject(int Type, int Unit)
{
	ASSERT((Type == BMC_32) || (Type == BMC_128));
	ASSERT((Unit == 0) || (Unit == 1));

	/* CWFNg */
	LockVM();
	bmc_setfile(Type, Unit, NULL);
	SetAllMRUState();
	UnlockVM();
}

#if 0
/*
 *	ouꎞo
 */
static void FASTCALL OnBubbleTemp(int Type, int Unit)
{
	ASSERT((Type == BMC_32) || (Type == BMC_128));
	ASSERT((Unit == 0) || (Unit == 1));

	/* ݋֎~؂ւ */
	LockVM();
	if (bmc_teject[Type][Unit]) {
		bmc_teject[Type][Unit] = FALSE;
	}
	else {
		bmc_teject[Type][Unit] = TRUE;
	}
	UnlockVM();
}
#endif

/*
 *	ou݋֎~
 */
static void FASTCALL OnBubbleProtect(int Type, int Unit)
{
	ASSERT((Type == BMC_32) || (Type == BMC_128));
	ASSERT((Unit == 0) || (Unit == 1));

	/* ݋֎~؂ւ */
	LockVM();
	if (bmc_writep[Type][Unit]) {
		bmc_setwritep(Type, Unit, FALSE);
	}
	else {
		bmc_setwritep(Type, Unit, TRUE);
	}
	ResetSch();
	UnlockVM();
}
#endif

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	SCWFNg
 */
static void FASTCALL OnCasetteEject(void)
{
	int	i;

	/* CWFNg */
	LockVM();
	tape_setfile(NULL);
	for (i=0; i<2; i++) {
		bmc_setfile(BMC_32, i, NULL);
		bmc_setfile(BMC_128, i, NULL);
	}
	SetAllMRUState();
	UnlockVM();
}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

/*
 *	e[v(A)j[
 */
static BOOL FASTCALL OnTape(WORD wID)
{
	switch (wID) {
		/* J */
		case IDM_TOPEN:
			OnTapeOpen();
			return TRUE;

		/* O */
		case IDM_TEJECT:
			OnTapeEject();
			return TRUE;

		/* ߂ */
		case IDM_REW:
			OnRew();
			return TRUE;

#if 0
		/* ŏ܂Ŋ߂ */
		case IDM_REWTOP:
			OnRewTop();
			return TRUE;
#endif

		/*  */
		case IDM_FF:
			OnFF();
			return TRUE;

		/* ^ */
		case IDM_REC:
			OnRec();
			return TRUE;

#if XM7_VER == 1 && defined(BUBBLE)
		/* J */
		case IDM_B0OPEN:
			OnBubbleOpen(BMC_32, 0);
			return TRUE;
		case IDM_B1OPEN:
			OnBubbleOpen(BMC_32, 1);
			return TRUE;
		case IDM_B2OPEN:
			OnBubbleOpen(BMC_128, 0);
			return TRUE;
		case IDM_B3OPEN:
			OnBubbleOpen(BMC_128, 1);
			return TRUE;

		/* jbgŊJ */
		case IDM_BBOPEN:
			OnBubbleBoth(BMC_32);
			break;
		case IDM_BB23OPEN:
			OnBubbleBoth(BMC_128);
			break;

		/* jbg */
		case IDM_BBREPLACE:
			OnBubbleReplace(BMC_32, 0);
			break;
		case IDM_BB23REPLACE:
			OnBubbleReplace(BMC_128, 0);
			break;

		/* O */
		case IDM_B0EJECT:
			OnBubbleEject(BMC_32, 0);
			return TRUE;
		case IDM_B1EJECT:
			OnBubbleEject(BMC_32, 1);
			return TRUE;
		case IDM_B2EJECT:
			OnBubbleEject(BMC_128, 0);
			return TRUE;
		case IDM_B3EJECT:
			OnBubbleEject(BMC_128, 1);
			return TRUE;

		/* ꎞCWFNg */
#if 0
		case IDM_B0TEMP:
			OnBubbleTemp(BMC_32, 0);
			return TRUE;
		case IDM_B1TEMP:
			OnBubbleTemp(BMC_32, 1);
			return TRUE;
		case IDM_B2TEMP:
			OnBubbleTemp(BMC_128, 0);
			return TRUE;
		case IDM_B3TEMP:
			OnBubbleTemp(BMC_128, 1);
			return TRUE;
#endif

		/* ݋֎~ */
		case IDM_B0WRITE:
			OnBubbleProtect(BMC_32, 0);
			return TRUE;
		case IDM_B1WRITE:
			OnBubbleProtect(BMC_32, 1);
			return TRUE;
		case IDM_B2WRITE:
			OnBubbleProtect(BMC_128, 0);
			return TRUE;
		case IDM_B3WRITE:
			OnBubbleProtect(BMC_128, 1);
			return TRUE;

		/* SCWFNg */
		case IDM_CAS_EJECT:
			OnCasetteEject();
			break;
#endif
	}

	/* XgZbgie[vj */
	if ((wID >= IDM_T_FILE0) && (wID <= IDM_T_FILE9)) {
		char path[256 + 1];

		GetMRUFile(MRU_TAPE, wID - IDM_T_FILE0, path, sizeof(path));

		LockVM();
		tape_setfile(path);
		if (tape_fileh == -1) {
			FileMessageBox(hMainWnd, IDS_FILE_NOT_FOUND, path, MB_ICONSTOP | MB_OK);
		}
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}

#if XM7_VER == 1 && defined(BUBBLE)
	/* fBAiouJZbgj */
	if ((wID >= IDM_B0MEDIA00) && (wID <= IDM_B0MEDIA15)) {
		LockVM();
		bmc_setmedia(BMC_32, 0, wID - IDM_B0MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
		return TRUE;
	}
	if ((wID >= IDM_B1MEDIA00) && (wID <= IDM_B1MEDIA15)) {
		LockVM();
		bmc_setmedia(BMC_32, 1, wID - IDM_B1MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
		return TRUE;
	}
	if ((wID >= IDM_B2MEDIA00) && (wID <= IDM_B2MEDIA15)) {
		LockVM();
		bmc_setmedia(BMC_128, 0, wID - IDM_B2MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
	if ((wID >= IDM_B3MEDIA00) && (wID <= IDM_B3MEDIA15)) {
		LockVM();
		bmc_setmedia(BMC_128, 1, wID - IDM_B3MEDIA00);
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}

	/* XgZbgiouJZbgj */
	if ((wID >= IDM_B3_FILE0) && (wID <= IDM_B0_FILE9)) {
		char path[256 + 1];
		int type;
		int unit;
		int mrutype;
		int index;
		int media;

		if ((wID >= IDM_B0_FILE0) && (wID <= IDM_B0_FILE9)) {
			type	= BMC_32;
			unit	= 0;
			mrutype	= MRU_BBL0;
			index	= wID - IDM_B0_FILE0;
		}
		if ((wID >= IDM_B1_FILE0) && (wID <= IDM_B1_FILE9)) {
			type	= BMC_32;
			unit	= 1;
			mrutype	= MRU_BBL1;
			index	= wID - IDM_B1_FILE0;
		}
		if ((wID >= IDM_B2_FILE0) && (wID <= IDM_B2_FILE9)) {
			type	= BMC_128;
			unit	= 0;
			mrutype	= MRU_BBL128_0;
			index	= wID - IDM_B2_FILE0;
		}
		if ((wID >= IDM_B3_FILE0) && (wID <= IDM_B3_FILE9)) {
			type	= BMC_128;
			unit	= 1;
			mrutype	= MRU_BBL128_1;
			index	= wID - IDM_B3_FILE0;
		}

		GetMRUFile(mrutype, index, path, sizeof(path));
		media = GetMRUMedia(mrutype, index);

		LockVM();
		bmc_setfile(type, unit, path);
		if (bmc_ready[type][unit] != BMC_TYPE_NOTREADY) {
			if (bmc_ready[type][unit] == BMC_TYPE_B77) {
				if (media >= 0) {
					bmc_setmedia(type, unit, media);
				}
			}
		}
		else {
			FileMessageBox(hMainWnd, IDS_FILE_NOT_FOUND, path, MB_ICONSTOP | MB_OK);
		}
		SetAllMRUState();
		ResetSch();
		UnlockVM();
	}
#endif

	return FALSE;
}

/*
 *	e[v(A)j[XV
 */
static void FASTCALL OnTapePopup(HMENU hMenu)
{
	MENUITEMINFO mii;
	char string[128];
	UINT uitem;

	ASSERT(hMenu);

	uitem = 0;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* I[v */
	mii.wID = IDM_TOPEN;
	LoadString(hResInstance, IDS_TAPEOPEN, string, sizeof(string));
	mii.dwTypeData = string;
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

	if (tape_fileh == -1) {
		mii.fState = MFS_GRAYED;
	}

	/* CWFNg */
	mii.wID = IDM_TEJECT;
	LoadString(hResInstance, IDS_TAPEEJECT, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* ߂ */
	mii.wID = IDM_REW;
	LoadString(hResInstance, IDS_TAPEREW, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

#if 0
	/* ŏ܂Ŋ߂ */
	mii.wID = IDM_REWTOP;
	LoadString(hResInstance, IDS_TAPEREWTOP, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
#endif

	/*  */
	mii.wID = IDM_FF;
	LoadString(hResInstance, IDS_TAPEFF, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* ^ */
	mii.wID = IDM_REC;
	LoadString(hResInstance, IDS_TAPEREC, string, sizeof(string));
	mii.cch = strlen(string);
	if (tape_writep || tape_fileh == -1) {
		mii.fState = MFS_GRAYED;
	}
	else {
		if (tape_rec) {
			mii.fState = MFS_CHECKED | MFS_ENABLED;
		}
		else {
			mii.fState = MFS_ENABLED;
		}
	}
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);

#ifndef BUBBLE
	if ((!bMountFileName) && (tape_fileh != -1)) {
		char buffer[256];
		char tmp[256];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
		int i,j;

		/* Zp[^} */
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);
		mii.fType = MFT_STRING;

		/* fBAt@C */
		mii.wID = IDM_TMEDIA;
		mii.fState = MFS_CHECKED | MFS_ENABLED;
		mii.fType = MFT_STRING | MFT_RADIOCHECK;

		/* t@Cl[{gq̂ݎo */
		_splitpath(tape_fname, NULL, NULL, fname, ext);
		_snprintf(tmp, sizeof(tmp), "%s%s", fname, ext);

		/* vtBbNX΍ */
		j = 0;
		for (i=0; i<(int)strlen(tmp); i++) {
			if (tmp[i] == '&') {
				buffer[j++] = '&';
			}
			buffer[j++] = tmp[i];
		}
		buffer[j] = '\0';
		mii.dwTypeData = buffer;
		mii.cch = strlen(buffer);
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	}
#endif

	/* } */
	AppendMenuHistory(hMenu, uitem, IDM_T_FILE0, MRU_TAPE, TRUE);
}

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	JZbg(C)j[XV
 */
static void FASTCALL OnCassettePopup(HMENU hMenu)
{
	HMENU hSubMenu;
	MENUITEMINFO mii;
	char string[256];
	char buffer[256];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	UINT uitem, utape, ububble[2][2], uid;
	int type;
	int unit;
	int offset;
	int shift;
	int i,j,k;

	ASSERT(hMenu);

#if 0
	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}
#endif

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;
	uitem = 0;

	/* e[v */
	utape = uitem++;
#if 0
	if (tape_fileh == -1) {
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.wID = IDM_TOPEN;
		LoadString(hResInstance, IDS_TAPE0OPEN, string, sizeof(string));
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, utape, TRUE, &mii);
	}
	else
#endif
	{
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
		mii.wID = IDM_TOPEN;
		//mii.hSubMenu = CreatePopupMenu();
		mii.hSubMenu = GetSubMenu(hMenu, utape);
		LoadString(hResInstance, IDS_TAPE, buffer, sizeof(buffer));
		if (tape_fileh != -1) {
			_splitpath(tape_fname, NULL, NULL, fname, ext);
			_snprintf(string, sizeof(string), "%s - %s%s", buffer, fname,
					  ext);
		}
		else {
			strncpy(string, buffer, sizeof(string));
		}
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		RemoveMenu(hMenu, utape, MF_BYPOSITION);
		InsertMenuItem(hMenu, utape, TRUE, &mii);
	}

	for(type = 0; type < 2; type ++) {
#if 0
		/* Zp[^} */
		mii.fType = MFT_SEPARATOR;
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);

		/* "ouz_jbg" */
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.fType = MFT_STRING;
		mii.fState = MFS_GRAYED;
		mii.wID = 0;
		LoadString(hResInstance, IDS_BUBBLE + type, string, sizeof(string));
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);
#else
		uitem++;

		/* "ouz_jbg" */
		EnableMenuItem(hMenu, uitem, MF_BYPOSITION | MF_GRAYED);
		uitem++;
#endif

		mii.fType = MFT_STRING;
#if 0
		if (bmc_enable[type]) {
			mii.fState = MFS_ENABLED;
		}
		else {
			mii.fState = MFS_GRAYED;
		}
#else
		mii.fState = MFS_ENABLED;
#endif

		for(unit = 0; unit < 2; unit ++) {

			/* ItZbgm */
			shift = unit + type * 2;
#if 0
			if (!type && !unit) {
				offset = 0;
			}
			else if (!type && unit) {
				offset = IDM_B1OPEN - IDM_B0OPEN;
			}
			else if (type && !unit) {
				offset = IDM_B2OPEN - IDM_B0OPEN;
			}
			else {
				offset = IDM_B3OPEN - IDM_B0OPEN;
			}
#endif

			/* ouz_jbgn */
			ububble[type][unit] = uitem;
#if 0
			if ((!bmc_enable[type]) ||
				(bmc_ready[type][unit] == BMC_TYPE_NOTREADY)) {
				mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
				mii.wID = IDM_B0OPEN + offset;
				LoadString(hResInstance, IDS_BUBBLE0OPEN + shift, string, sizeof(string));
			}
			else
#endif
			{
#if 0
				if (!bmc_enable[type]) {
					mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
				}
				else
#endif
				{
					mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
				}
				mii.wID = 0;
				//mii.hSubMenu = CreatePopupMenu();
				mii.hSubMenu = GetSubMenu(hMenu, uitem);
				LoadString(hResInstance, IDS_BUBBLE0 + shift, buffer, sizeof(buffer));
#if 0
				if ((bmc_enable[type]) &&
					(bmc_ready[type][unit] != BMC_TYPE_NOTREADY)) {
#else
				if (bmc_ready[type][unit] != BMC_TYPE_NOTREADY) {
#endif
					_splitpath(bmc_fname[type][unit], NULL, NULL, fname, ext);
					_snprintf(string, sizeof(string), "%s - %s%s", buffer,
							  fname, ext);
				}
				else {
					strncpy(string, buffer, sizeof(string));
				}
			}
			mii.dwTypeData = string;
			mii.cch = strlen(string);
			RemoveMenu(hMenu, uitem, MF_BYPOSITION);
			InsertMenuItem(hMenu, uitem++, TRUE, &mii);
		}

#if 0
		/* I[v */
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		if (!type) {
			mii.wID = IDM_BBOPEN;
			uid = IDS_BUBBLEBOTH;
		}
		else {
			mii.wID = IDM_BB23OPEN;
			uid = IDS_BUBBLEBOTHX;
		}
		LoadString(hResInstance, uid, string, sizeof(string));
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);
#else
		EnableMenuItem(hMenu, uitem++, MF_BYPOSITION | mii.fState);
		EnableMenuItem(hMenu, uitem++, MF_BYPOSITION | mii.fState);
#endif
	}

#if 0
	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* CWFNg */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.wID = IDM_CAS_EJECT;
	mii.fState = MFS_ENABLED;
	LoadString(hResInstance, IDS_CAS_EJECT, string, sizeof(string));
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, uitem++, TRUE, &mii);
#endif

	/* e[vETuj[ */
#if 0
	if (tape_fileh != -1) {
#else
	{
#endif
		hSubMenu = GetSubMenu(hMenu, utape);
		OnTapePopup(hSubMenu);
	}

	/* ouJZbgETuj[ */
	for(type = 0; type < 2; type ++) {
		for(unit = 0; unit < 2; unit ++) {
#if 0
#if 0
			/* ouJZbgZbgĂȂ΁A܂ */
			if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY || !bmc_enable[type]) {
#else
			if (!bmc_enable[type]) {
#endif
				continue;
			}
#endif

			hSubMenu = GetSubMenu(hMenu, ububble[type][unit]);

			/* j[ׂč폜 */
			while (GetMenuItemCount(hSubMenu) > 0) {
				DeleteMenu(hSubMenu, 0, MF_BYPOSITION);
			}

			/* ItZbgm */
			shift = unit + type * 2;
			if (!type && !unit) {
				offset = 0;
			}
			else if (!type && unit) {
				offset = IDM_B1OPEN - IDM_B0OPEN;
			}
			else if (type && !unit) {
				offset = IDM_B2OPEN - IDM_B0OPEN;
			}
			else {
				offset = IDM_B3OPEN - IDM_B0OPEN;
			}
			uitem = 0;

			/* j[\̏ */
			memset(&mii, 0, sizeof(mii));
			mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
			mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
			mii.fType = MFT_STRING;

			/* I[v */
			mii.wID = IDM_B0OPEN + offset;
			LoadString(hResInstance, IDS_BUBBLEOPEN, string, sizeof(string));
			mii.dwTypeData = string;
			mii.fState = MFS_ENABLED;
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

			/* CWFNg */
			mii.wID = IDM_B0EJECT + offset;
			if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY) {
				mii.fState = MFS_GRAYED;
			}
			else {
				mii.fState = MFS_ENABLED;
			}
			LoadString(hResInstance, IDS_BUBBLEEJECT, string, sizeof(string));
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

#if 0
			/* Zp[^} */
			mii.fType = MFT_SEPARATOR;
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

			/* ꎞo */
			mii.wID = IDM_B0TEMP + offset;
			LoadString(hResInstance, IDS_BUBBLETEMP, string, sizeof(string));
			if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY) {
				mii.fState = MFS_GRAYED;
			}
			else if (bmc_teject[type][unit]) {
				mii.fState = MFS_CHECKED | MFS_ENABLED;
			}
			else {
				mii.fState = MFS_ENABLED;
			}
			mii.fType = MFT_STRING;
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
#endif

			/* CgveNg */
			mii.wID = IDM_B0WRITE + offset;
			LoadString(hResInstance, IDS_BUBBLEPROTECT, string, sizeof(string));
			if (bmc_ready[type][unit] == BMC_TYPE_NOTREADY) {
				mii.fState = MFS_GRAYED;
			}
			else {
				if (bmc_fwritep[type][unit]) {
					mii.fState = MFS_GRAYED;
				}
				else {
					mii.fState = MFS_ENABLED;
				}
				if ((bmc_fwritep[type][unit]) || (bmc_writep[type][unit])) {
					mii.fState |= MFS_CHECKED;
				}
			}
			mii.fType = MFT_STRING;
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

#if 0	/* t@C͏wɕ\̂ō폜 */
			/* Zp[^} */
			mii.fType = MFT_SEPARATOR;
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

			/* fBAt@C */
			mii.wID = IDM_B0MEDIA00 + offset;
			mii.fState = MFS_CHECKED | MFS_ENABLED;
			mii.fType = MFT_STRING | MFT_RADIOCHECK;

			/* t@Cl[{gq̂ݎo */
			_splitpath(bmc_fname[type][unit], NULL, NULL, fname, ext);
			_snprintf(buffer, sizeof(buffer), "%s%s", fname, ext);

			/* vtBbNX΍ */
			j = 0;
			for (i=0; i<(int)strlen(buffer); i++) {
				if (buffer[i] == '&') {
					string[j++] = '&';
				}
				string[j++] = buffer[i];
			}
			string[j] = '\0';
			mii.cch = strlen(string);
			InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
#endif

			/* B77Ȃ烁fBA */
			if (bmc_ready[type][unit] == BMC_TYPE_B77) {

				/* Zp[^} */
				mii.fType = MFT_SEPARATOR;
				InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);

				for (i=0; i<bmc_medias[type][unit]; i++) {
					mii.wID = IDM_B0MEDIA00 + offset + i;
					mii.fState = MFS_ENABLED;
					if (bmc_media[type][unit] == i) {
						mii.fState |= MFS_CHECKED;
					}
					mii.fType = MFT_STRING | MFT_RADIOCHECK;
					if (strlen(bmc_name[type][unit][i]) == 0) {
						LoadString(hResInstance, IDS_MEDIA_NAME, buffer, sizeof(buffer));
						_snprintf(string, sizeof(string), buffer, i + 1);
					}
					else {
						/* vtBbNX΍ */
						k = 0;
						for (j=0; j<(int)strlen(bmc_name[type][unit][i]); j++) {
							if (bmc_name[type][unit][i][j] == '&') {
								string[k++] = '&';
							}
							string[k++] = bmc_name[type][unit][i][j];
						}
						string[k] = '\0';
					}
					mii.cch = strlen(string);
					InsertMenuItem(hSubMenu, uitem++, TRUE, &mii);
				}
			}

			/* } */
#if 0
			AppendMenuHistory(hSubMenu, uitem++, IDM_B0_FILE0 - shift * 10, MRU_BBL0 + shift, bmc_enable[type]);
#else
			AppendMenuHistory(hSubMenu, uitem++, IDM_B0_FILE0 - shift * 10, MRU_BBL0 + shift, TRUE);
#endif
		}
	}
}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

/*
 *	e[v(A)ReLXgj[
 */
static BOOL FASTCALL OnTapeContextualMenu(HWND hWnd, WORD wID, int x, int y)
{
	ASSERT(hWnd);

#if XM7_VER == 1 && defined(BUBBLE)
	/* ouJZbgENbNj[ */
	if ((wID >= IDM_B0MEDIA00) && (wID <= IDM_B0MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_B1MEDIA00) && (wID <= IDM_B1MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_B2MEDIA00) && (wID <= IDM_B2MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
	if ((wID >= IDM_B3MEDIA00) && (wID <= IDM_B3MEDIA15)) {
		OnContextualSubMenu(hWnd, hMediaMenu, wID, x, y);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

	/* qXg */
	if ((wID >= IDM_T_FILE0) && (wID <= IDM_T_FILE9)) {
		OnContextualSubMenu(hWnd, hHistoryMenu, wID, x, y);
		return TRUE;
	}
#if XM7_VER == 1 && defined(BUBBLE)
	if ((wID >= IDM_B3_FILE0) && (wID <= IDM_B0_FILE9)) {
		OnContextualSubMenu(hWnd, hHistoryMenu, wID, x, y);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

	return FALSE;
}

/*-[ \j[ ]---------------------------------------------------------*/

/*
 *	tbs[fBXNRg[(F)
 */
static void FASTCALL OnFDC(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_FDC]) {
		PostMessage(hSubWnd[SWND_FDC], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_FDC] = CreateFDC(hDrawWnd, SWND_FDC);
}

/*
 *	ouRg[(B)
 */
#if XM7_VER == 1 && defined(BUBBLE)
static void FASTCALL OnBMC(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_BMC]) {
		PostMessage(hSubWnd[SWND_BMC], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_BMC] = CreateBMC(hDrawWnd, SWND_BMC);
}
#endif

/*
 *	FMWX^(O)
 */
static void FASTCALL OnOPNReg(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_OPNREG]) {
		PostMessage(hSubWnd[SWND_OPNREG], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_OPNREG] = CreateOPNReg(hDrawWnd, SWND_OPNREG);
}

/*
 *	FMfBXvC(D)
 */
static void FASTCALL OnOPNDisp(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_OPNDISP]) {
		PostMessage(hSubWnd[SWND_OPNDISP], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_OPNDISP] = CreateOPNDisp(hDrawWnd, SWND_OPNDISP);
}

/*
 *	TuCPURg[(C)
 */
static void FASTCALL OnSubCtrl(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_SUBCTRL]) {
		PostMessage(hSubWnd[SWND_SUBCTRL], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_SUBCTRL] = CreateSubCtrl(hDrawWnd, SWND_SUBCTRL);
}

/*
 *	pbgWX^(P)
 */
static void FASTCALL OnPaletteReg(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_PALETTE]) {
		PostMessage(hSubWnd[SWND_PALETTE], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_PALETTE] = CreatePaletteReg(hDrawWnd, SWND_PALETTE);
}

	/*
 *	L[{[h(K)
 */
static void FASTCALL OnKeyboard(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_KEYBOARD]) {
		PostMessage(hSubWnd[SWND_KEYBOARD], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_KEYBOARD] = CreateKeyboard(hDrawWnd, SWND_KEYBOARD);
}

/*
 *	_Z/(L)
 */
#if XM7_VER >= 2
static void FASTCALL OnALULine(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_ALULINE]) {
		PostMessage(hSubWnd[SWND_ALULINE], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_ALULINE] = CreateALULine(hDrawWnd, SWND_ALULINE);
}
#endif

/*
 *	Ǘ(M)
 */
static void FASTCALL OnMMR(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_MMR]) {
		PostMessage(hSubWnd[SWND_MMR], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_MMR] = CreateMMR(hDrawWnd, SWND_MMR);
}

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
/*
 *	DMARg[(A)
 */
static void FASTCALL OnDMAC(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_DMAC]) {
		PostMessage(hSubWnd[SWND_DMAC], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_DMAC] = CreateDMAC(hDrawWnd, SWND_DMAC);
}
#endif

/*
 *	ROM(N)
 */
static void FASTCALL OnKanji(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_KANJI]) {
		PostMessage(hSubWnd[SWND_KANJI], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_KANJI] = CreateKanji(hDrawWnd, SWND_KANJI);
}

/*
 *	JZbge[vv^(T)
 */
static void FASTCALL OnTapeLPWnd(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_TAPELP]) {
		PostMessage(hSubWnd[SWND_TAPELP], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_TAPELP] = CreateTapeLP(hDrawWnd, SWND_TAPELP);
}

/*
 *	Xe[^Xo[(S)
 */
static void FASTCALL OnStatus(void)
{
	/* Xe[^Xo[LłȂ΁AȂ */
	if (!hStatusBar) {
		return;
	}

	if (IsWindowVisible(hStatusBar)) {
		/*  */
		ShowWindow(hStatusBar, SW_HIDE);
	}
	else {
		/* \ */
		ShowWindow(hStatusBar, SW_SHOW);
	}

	/* Xe[^Xo[\ */
	bStatusBar = IsWindowVisible(hStatusBar);
	LockVM();
	StoreCfg();
	UnlockVM();

	/* t[EChẼTCY␳ */
	OnSize(hMainWnd, 640, 400);
}

/*
 *	ŐV̏ɍXV(R)
 */
void FASTCALL OnRefresh(HWND hWnd)
{
	int i;

	ASSERT(hWnd);
	ASSERT(hDrawWnd);

	/* tAZuEChE@AhXXV */
	if ((bSyncDisasm[0]) && (nAddrModeDisAsm == 0)) {
		AddrDisAsm(MAINCPU, maincpu.pc);
	}
	if (bSyncDisasm[1]) {
		AddrDisAsm(SUBCPU, subcpu.pc);
	}
#ifdef JSUB
	if (bSyncDisasm[2]) {
		AddrDisAsm(JSUBCPU, jsubcpu.pc);
	}
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	if ((bSyncDisasm[3]) && (nAddrModeDisAsmZ80 == 0)) {
		AddrDisAsm(MAINZ80, (WORD)mainz80.pc);
	}
#endif

	/* FM/PSGWX^/fBXvCEChEXV */
	ReSizeOPNReg();
	ReSizeOPNDisp();

	/* CEChE */
	InvalidateRect(hWnd, NULL, FALSE);
	InvalidateRect(hDrawWnd, NULL, FALSE);

	/* TuEChE */
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i]) {
			InvalidateRect(hSubWnd[i], NULL, FALSE);
		}
	}
}

/*
 *	sɓ(Y)
 */
static void FASTCALL OnSync(void)
{
	bSync = (!bSync);
}

/*
 *	tXN[(U)
 */
static void FASTCALL OnFullScreen(HWND hWnd)
{
	BOOL bRun;
	int i;

	ASSERT(hWnd);

#ifdef LPRINT
	/* LP_CAO\̓tXN[s */
	if (lp_getdialog()) {
		char string[128];
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_LPRDIALOGWRN, string, sizeof(string));
		MessageBox(hWnd, string, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
		SetMenuExitTimer();
		return;
	}
#endif

#ifdef KBDPASTE
	if (hKeyStrokeDialog) {
		/* EChEN[Y */
		PostMessage(hKeyStrokeDialog, WM_CLOSE, 0, 0);
		Sleep(1);
	}
#endif

	/* VMbNAXgbv */
	LockVM();
	bRun = run_flag;
	run_flag = FALSE;
	StopSnd();

	/* [h؂ւ */
	if (bFullScreen) {
		ModeDraw(hWnd, FALSE);

		if (!bFullScreen) {
			/* ׂẴTuEBhEtXN[ȌԂɖ߂ */
			for (i=0; i<SWND_MAXNUM; i++) {
				if (hSubWnd[i]) {
					if (bShowSubWindow[i]) {
						ShowWindow(hSubWnd[i], SW_RESTORE);
					}
				}
			}
			if (bPopupSwnd) {
				SetForegroundWindow(hWnd);
			}
		}
	}
	else {
		ModeDraw(hWnd, TRUE);

		if (bFullScreen) {
			/* ׂẴTuEBhEB */
			for (i=0; i<SWND_MAXNUM; i++) {
				bShowSubWindow[i] = FALSE;
				if (hSubWnd[i]) {
					if (ShowWindow(hSubWnd[i], SW_HIDE)) {
						bShowSubWindow[i] = TRUE;
					}
				}
			}
		}
	}

	/* VMAbN */
	StoreCfg();
	run_flag = bRun;
	ResetSch();
	UnlockVM();
	PlaySnd();
}

/*
 *	fBXvCTCY(Z)
 */
static void FASTCALL OnScalingSize(HWND hWnd, int nSize)
{
	ASSERT((nSize >= 0) || (nSize <= 7));

	/* fBXvCTCYύX */
	LockVM();
	nScalingSize = nSize;
	StoreCfg();
	UnlockVM();

	/* \eXV */
	OnSize(hWnd, 640, 400);
	InvalidateRect(hDrawWnd, NULL, FALSE);
}

#ifdef FILTERDLL
/*
 *	AXyNg䒲(&E)
 */
static void FASTCALL OnAspectRatio(HWND hWnd)
{
	/* AXyNg䒲tOύX */
	LockVM();
	bAspectRatio = !bAspectRatio;
	StoreCfg();
	UnlockVM();

	/* \eXV */
	OnSize(hWnd, 640, 400);
	InvalidateRect(hDrawWnd, NULL, FALSE);
}
#endif

/*
 *	\(V)j[
 */
static BOOL FASTCALL OnView(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	switch (wID) {
		/* FDC */
		case IDM_FDC:
			OnFDC();
			return TRUE;

		/* ouRg[ */
#if XM7_VER == 1 && defined(BUBBLE)
		case IDM_BMC:
			OnBMC();
			return TRUE;
#endif

		/* OPNWX^ */
		case IDM_OPNREG:
			OnOPNReg();
			return TRUE;

		/* OPNfBXvC */
		case IDM_OPNDISP:
			OnOPNDisp();
			return TRUE;

		/* TuCPURg[ */
		case IDM_SUBCTRL:
			OnSubCtrl();
			return TRUE;

		/* pbgWX^ */
		case IDM_PALETTE:
			OnPaletteReg();
			return TRUE;

		/* L[{[h */
		case IDM_KEYBOARD:
			OnKeyboard();
			return TRUE;

#if XM7_VER >= 2
		/* _Z/ */
		case IDM_ALULINE:
			OnALULine();
			return TRUE;
#endif

		/* Ǘ */
		case IDM_MMR:
			OnMMR();
			return TRUE;

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
		/* DMARg[ */
		case IDM_DMAC:
			OnDMAC();
			return TRUE;
#endif

		/* ROM */
		case IDM_KANJI:
			OnKanji();
			return TRUE;

		/* JZbge[vv^ */
		case IDM_TAPELP:
			OnTapeLPWnd();
			return TRUE;

		/* Xe[^Xo[ */
		case IDM_STATUS:
			OnStatus();
			return TRUE;

		/* ŐV̏ɍXV */
		case IDM_REFRESH:
			OnRefresh(hWnd);
			return TRUE;

		/* sɓ */
		case IDM_SYNC:
			OnSync();
			return TRUE;

#ifdef FILTERDLL
		/* x0.5 */
		case IDM_DISPLAYSIZEx050:
			OnScalingSize(hWnd, DISPLAYSIZE_050);
			return TRUE;
#endif

		/* x1.0 */
		case IDM_DISPLAYSIZEx100:
			OnScalingSize(hWnd, DISPLAYSIZE_100);
			return TRUE;

#ifdef FILTERDLL
		/* x1.2 */
		case IDM_DISPLAYSIZEx120:
			OnScalingSize(hWnd, DISPLAYSIZE_120);
			break;

		/* x1.5 */
		case IDM_DISPLAYSIZEx150:
			OnScalingSize(hWnd, DISPLAYSIZE_150);
			return TRUE;

		/* x1.8 */
		case IDM_DISPLAYSIZEx180:
			OnScalingSize(hWnd, DISPLAYSIZE_180);
			return TRUE;
#endif

		/* x2.0 */
		case IDM_DISPLAYSIZEx200:
			OnScalingSize(hWnd, DISPLAYSIZE_200);
			return TRUE;

		/* x3.0 */
		case IDM_DISPLAYSIZEx300:
			OnScalingSize(hWnd, DISPLAYSIZE_300);
			return TRUE;

		/* x4.0 */
		case IDM_DISPLAYSIZEx400:
			OnScalingSize(hWnd, DISPLAYSIZE_400);
			return TRUE;

#ifdef FILTERDLL
		/* AXyNg䒲 */
		case IDM_ASPECTRATIO:
			OnAspectRatio(hWnd);
			return TRUE;
#endif

		/* tXN[ */
		case IDM_FULLSCREEN:
			OnFullScreen(hWnd);
			return TRUE;
	}

	return FALSE;
}

/*
 *	\(V)j[XV
 */
static void FASTCALL OnViewPopup(HMENU hMenu)
{
	UINT uitem, id;

	/* TuEChEQ */
	CheckMenuSub(hMenu, IDM_FDC, (BOOL)hSubWnd[SWND_FDC]);
#if XM7_VER == 1 && defined(BUBBLE)
	CheckMenuSub(hMenu, IDM_BMC, (BOOL)hSubWnd[SWND_BMC]);
#endif
	CheckMenuSub(hMenu, IDM_OPNREG, (BOOL)hSubWnd[SWND_OPNREG]);
	CheckMenuSub(hMenu, IDM_OPNDISP, (BOOL)hSubWnd[SWND_OPNDISP]);
	CheckMenuSub(hMenu, IDM_PALETTE, (BOOL)hSubWnd[SWND_PALETTE]);
	CheckMenuSub(hMenu, IDM_SUBCTRL, (BOOL)hSubWnd[SWND_SUBCTRL]);
#if XM7_VER >= 2
	CheckMenuSub(hMenu, IDM_ALULINE, (BOOL)hSubWnd[SWND_ALULINE]);
#endif
	CheckMenuSub(hMenu, IDM_KEYBOARD, (BOOL)hSubWnd[SWND_KEYBOARD]);
	CheckMenuSub(hMenu, IDM_MMR, (BOOL)hSubWnd[SWND_MMR]);
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	CheckMenuSub(hMenu, IDM_DMAC, (BOOL)hSubWnd[SWND_DMAC]);
#endif
	CheckMenuSub(hMenu, IDM_KANJI, (BOOL)hSubWnd[SWND_KANJI]);
	CheckMenuSub(hMenu, IDM_TAPELP, (BOOL)hSubWnd[SWND_TAPELP]);

	/* ̑ */
	if (hStatusBar) {
		CheckMenuSub(hMenu, IDM_STATUS, IsWindowVisible(hStatusBar));
	}
	else {
		CheckMenuSub(hMenu, IDM_STATUS, FALSE);
	}
	CheckMenuSub(hMenu, IDM_SYNC, bSync);
	CheckMenuSub(hMenu, IDM_FULLSCREEN, bFullScreen);

	/* tXN[̃TuEBhEj[ */
	EnableMenuSub(hMenu, IDM_FDC, !bFullScreen);
#if XM7_VER == 1 && defined(BUBBLE)
	EnableMenuSub(hMenu, IDM_BMC, !bFullScreen);
#endif
	EnableMenuSub(hMenu, IDM_OPNREG, !bFullScreen);
	EnableMenuSub(hMenu, IDM_OPNDISP, !bFullScreen);
	EnableMenuSub(hMenu, IDM_PALETTE, !bFullScreen);
	EnableMenuSub(hMenu, IDM_SUBCTRL, !bFullScreen);
#if XM7_VER >= 2
	EnableMenuSub(hMenu, IDM_ALULINE, !bFullScreen);
#endif
	EnableMenuSub(hMenu, IDM_KEYBOARD, !bFullScreen);
	EnableMenuSub(hMenu, IDM_MMR, !bFullScreen);
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	EnableMenuSub(hMenu, IDM_DMAC, !bFullScreen);
#endif
	EnableMenuSub(hMenu, IDM_KANJI, !bFullScreen);
	EnableMenuSub(hMenu, IDM_TAPELP, !bFullScreen);

	uitem = 12;
#if XM7_VER == 1 && defined(BUBBLE)
	uitem += 1;		/* ouRg[ */
#endif
#if XM7_VER >= 2
	uitem += 1;		/* _Z/ */
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	uitem += 1;		/* DMARg[ */
#endif
	EnableMenuPos(hMenu, uitem += 1, !bFullScreen);
	switch (nScalingSize) {
		case DISPLAYSIZE_100:		/* x1.0(640x400) */
		default:
			id = IDM_DISPLAYSIZEx100;
			break;

#ifdef FILTERDLL
		case DISPLAYSIZE_150:		/* x1.5(960x600) */
			id = IDM_DISPLAYSIZEx150;
			break;
#endif

		case DISPLAYSIZE_200:		/* x2.0(1280x800) */
			id = IDM_DISPLAYSIZEx200;
			break;

#ifdef FILTERDLL
		case DISPLAYSIZE_050:		/* x0.5(320x200) */
			id = IDM_DISPLAYSIZEx050;
			break;

		case DISPLAYSIZE_180:		/* x1.8(1152x720) */
			id = IDM_DISPLAYSIZEx180;
			break;

		case DISPLAYSIZE_120:		/* x1.2(768x480) */
			id = IDM_DISPLAYSIZEx120;
			break;
#endif

		case DISPLAYSIZE_300:		/* x3.0(1920x1200) */
			id = IDM_DISPLAYSIZEx300;
			break;

		case DISPLAYSIZE_400:		/* x4.0(2560x1600) */
			id = IDM_DISPLAYSIZEx400;
			break;
	}
	CheckMenuRadioItem(hMenu, IDM_DISPLAYSIZEx100, IDM_DISPLAYSIZEx400, id, MF_BYCOMMAND);
#ifdef FILTERDLL
	EnableMenuSub(hMenu, IDM_DISPLAYSIZEx150, (BOOL)(bFilterDll && bUseFilterDll));
	EnableMenuSub(hMenu, IDM_DISPLAYSIZEx050, (BOOL)(bFilterDll && bUseFilterDll));
	EnableMenuSub(hMenu, IDM_DISPLAYSIZEx180, (BOOL)(bFilterDll && bUseFilterDll));
	EnableMenuSub(hMenu, IDM_DISPLAYSIZEx120, (BOOL)(bFilterDll && bUseFilterDll));
	CheckMenuSub(hMenu, IDM_ASPECTRATIO, bAspectRatio);
	EnableMenuSub(hMenu, IDM_ASPECTRATIO, (BOOL)(!bFullScreen && bFilterDll && bUseFilterDll));
#endif
}

#ifdef FILTERDLL
/*
 *	\(V)ReLXgj[
 */
static BOOL FASTCALL OnViewContextualMenu(HWND hWnd, WORD wID, int x, int y)
{
	ASSERT(hWnd);

	/* AXyNg䒲 */
	if (wID == IDM_ASPECTRATIO) {
		/* O摜tB^DLLgp̂ݐݒ\ */
		if (bFilterDll && bUseFilterDll) {
			OnContextualSubMenu(hWnd, hAspectMenu, wID, x, y);
			return TRUE;
		}
	}

	return FALSE;
}
#endif

/*-[ fobOj[ ]-----------------------------------------------------*/

/*
 *	s(X)
 */
static void FASTCALL OnExec(void)
{
	/* ɎsȂAȂ */
	if (run_flag) {
		return;
	}

	/* X^[g */
	LockVM();
	stopreq_flag = FALSE;
	run_flag = TRUE;
	UnlockVM();
}

/*
 *	~(B)
 */
static void FASTCALL OnBreak(void)
{
	/* ɒ~ԂȂAȂ */
	if (!run_flag) {
		return;
	}

	/* ~ */
	LockVM();
	stopreq_flag = TRUE;
	UnlockVM();
}

/*
 *	g[X(T)
 */
static void FASTCALL OnTrace(HWND hWnd)
{
	ASSERT(hWnd);

	/* ~ԂłȂ΁A^[ */
	if (run_flag) {
		return;
	}

	/* s */
	schedule_trace();

	/* \XV */
	OnRefresh(hWnd);
}

/*
 *	u[N|Cg(B)
 */
static void FASTCALL OnBreakPoint(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_BREAKPOINT]) {
		PostMessage(hSubWnd[SWND_BREAKPOINT], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_BREAKPOINT] = CreateBreakPoint(hDrawWnd, SWND_BREAKPOINT);
}

/*
 *	XPW[(S)
 */
static void FASTCALL OnScheduler(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_SCHEDULER]) {
		PostMessage(hSubWnd[SWND_SCHEDULER], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_SCHEDULER] = CreateScheduler(hDrawWnd, SWND_SCHEDULER);
}

/*
 *	CPUWX^(C)
 */
static void FASTCALL OnCPURegister(int nCPU)
{
	int index;

#ifndef JSUB
	ASSERT((nCPU >= MAINCPU) && (nCPU <= SUBCPU));
#elif XM7_VER == 1 && defined(Z80CARD)
	ASSERT((nCPU >= MAINCPU) && (nCPU <= MAINZ80));
#else
	ASSERT((nCPU >= MAINCPU) && (nCPU <= JSUBCPU));
#endif
	ASSERT(hDrawWnd);

	/* CfbNX */
	index = SWND_CPUREG_MAIN + nCPU;

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[index]) {
		PostMessage(hSubWnd[index], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[index] = CreateCPURegister(hDrawWnd, (BYTE)nCPU, index);
}

/*
 *	tAZu(D)
 */
static void FASTCALL OnDisAsm(int nCPU)
{
	int index;

#ifndef JSUB
	ASSERT((nCPU >= MAINCPU) && (nCPU <= SUBCPU));
#elif XM7_VER == 1 && defined(Z80CARD)
	ASSERT((nCPU >= MAINCPU) && (nCPU <= MAINZ80));
#else
	ASSERT((nCPU >= MAINCPU) && (nCPU <= JSUBCPU));
#endif
	ASSERT(hDrawWnd);

	/* CfbNX */
	index = SWND_DISASM_MAIN + nCPU;

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[index]) {
		PostMessage(hSubWnd[index], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[index] = CreateDisAsm(hDrawWnd, (BYTE)nCPU, index);
}

/*
 *	_v(M)
 */
static void FASTCALL OnMemory(int nCPU)
{
	int index;

#ifndef JSUB
	ASSERT((nCPU >= MAINCPU) && (nCPU <= SUBCPU));
#else
	/* ̃P[X̂Z80[hpEChE݂͑Ȃ */
	ASSERT((nCPU >= MAINCPU) && (nCPU <= JSUBCPU));
#endif
	ASSERT(hDrawWnd);

	/* CfbNX */
	index = SWND_MEMORY_MAIN + nCPU;

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[index]) {
		PostMessage(hSubWnd[index], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[index] = CreateMemory(hDrawWnd, (BYTE)nCPU, index);
}

/*
 *	荞(I)
 */
static void FASTCALL OnIRQ(int nCPU)
{
	int index;

#ifndef JSUB
	ASSERT((nCPU >= MAINCPU) && (nCPU <= SUBCPU));
#else
	ASSERT((nCPU >= MAINCPU) && (nCPU <= JSUBCPU));
#endif
	ASSERT(hDrawWnd);

	/* CfbNX */
	index = SWND_IRQ_MAIN + nCPU;

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[index]) {
		PostMessage(hSubWnd[index], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[index] = CreateIRQ(hDrawWnd, (BYTE)nCPU, index);
}

/*
 *	O(L)
 */
static void FASTCALL OnLog(void)
{
	ASSERT(hDrawWnd);

	/* EChE݂΁AN[Ywo */
	if (hSubWnd[SWND_LOG]) {
		PostMessage(hSubWnd[SWND_LOG], WM_CLOSE, 0, 0);
		return;
	}

	/* EChE쐬 */
	hSubWnd[SWND_LOG] = CreateLog(hDrawWnd, SWND_LOG);
}

/*
 *	fobO(D)j[
 */
static BOOL FASTCALL OnDebug(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	switch (wID) {
		/* s */
		case IDM_EXEC:
			OnExec();
			return TRUE;

		/* u[N */
		case IDM_BREAK:
			OnBreak();
			return TRUE;

		/* g[X */
		case IDM_TRACE:
			OnTrace(hWnd);
			return TRUE;

		/* u[N|Cg */
		case IDM_BREAKPOINT:
			OnBreakPoint();
			return TRUE;

		/* XPW[ */
		case IDM_SCHEDULER:
			OnScheduler();
			return TRUE;

		/* CPUWX^(C) */
		case IDM_CPU_MAIN:
			OnCPURegister(MAINCPU);
			return TRUE;

		/* CPUWX^(Tu) */
		case IDM_CPU_SUB:
			OnCPURegister(SUBCPU);
			return TRUE;

		/* tAZu(C) */
		case IDM_DISASM_MAIN:
			OnDisAsm(MAINCPU);
			return TRUE;

		/* tAZu(Tu) */
		case IDM_DISASM_SUB:
			OnDisAsm(SUBCPU);
			return TRUE;

		/* _v(C) */
		case IDM_MEMORY_MAIN:
			OnMemory(MAINCPU);
			return TRUE;

		/* _v(Tu) */
		case IDM_MEMORY_SUB:
			OnMemory(SUBCPU);
			return TRUE;

#ifdef JSUB
		/* CPUWX^({Tu) */
		case IDM_CPU_JSUB:
			OnCPURegister(JSUBCPU);
			return TRUE;

		/* tAZu({Tu) */
		case IDM_DISASM_JSUB:
			OnDisAsm(JSUBCPU);
			return TRUE;

		/* _v({Tu) */
		case IDM_MEMORY_JSUB:
			OnMemory(JSUBCPU);
			return TRUE;
#endif

#if XM7_VER == 1 && defined(Z80CARD)
		/* CPUWX^(CZ80) */
		case IDM_CPU_MAINZ80:
			OnCPURegister(MAINZ80);
			return TRUE;

		/* tAZu(CZ80) */
		case IDM_DISASM_MAINZ80:
			OnDisAsm(MAINZ80);
			return TRUE;
#endif

		/* 荞(C) */
		case IDM_IRQ_MAIN:
			OnIRQ(MAINCPU);
			return TRUE;

		/* 荞(Tu) */
		case IDM_IRQ_SUB:
			OnIRQ(SUBCPU);
			return TRUE;

		/* O */
		case IDM_LOG:
			OnLog();
			return TRUE;
	}

	return FALSE;
}

/*
 *	fobO(D)j[XV
 */
static void FASTCALL OnDebugPopup(HMENU hMenu)
{
	ASSERT(hMenu);

	/* s */
	EnableMenuSub(hMenu, IDM_EXEC, !run_flag);
	EnableMenuSub(hMenu, IDM_BREAK, run_flag);
	EnableMenuSub(hMenu, IDM_TRACE, !run_flag);

	/* TuEChEQ */
	CheckMenuSub(hMenu, IDM_BREAKPOINT, (BOOL)hSubWnd[SWND_BREAKPOINT]);
	CheckMenuSub(hMenu, IDM_SCHEDULER, (BOOL)hSubWnd[SWND_SCHEDULER]);
	CheckMenuSub(hMenu, IDM_CPU_MAIN, (BOOL)hSubWnd[SWND_CPUREG_MAIN]);
	CheckMenuSub(hMenu, IDM_CPU_SUB, (BOOL)hSubWnd[SWND_CPUREG_SUB]);
	CheckMenuSub(hMenu, IDM_DISASM_MAIN, (BOOL)hSubWnd[SWND_DISASM_MAIN]);
	CheckMenuSub(hMenu, IDM_DISASM_SUB, (BOOL)hSubWnd[SWND_DISASM_SUB]);
	CheckMenuSub(hMenu, IDM_MEMORY_MAIN, (BOOL)hSubWnd[SWND_MEMORY_MAIN]);
	CheckMenuSub(hMenu, IDM_MEMORY_SUB, (BOOL)hSubWnd[SWND_MEMORY_SUB]);
#ifdef JSUB
	CheckMenuSub(hMenu, IDM_CPU_JSUB, (BOOL)hSubWnd[SWND_CPUREG_JSUB]);
	CheckMenuSub(hMenu, IDM_DISASM_JSUB, (BOOL)hSubWnd[SWND_DISASM_JSUB]);
	CheckMenuSub(hMenu, IDM_MEMORY_JSUB, (BOOL)hSubWnd[SWND_MEMORY_JSUB]);
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	CheckMenuSub(hMenu, IDM_CPU_MAINZ80, (BOOL)hSubWnd[SWND_CPUREG_Z80]);
	CheckMenuSub(hMenu, IDM_DISASM_MAINZ80, (BOOL)hSubWnd[SWND_DISASM_Z80]);
#endif
	CheckMenuSub(hMenu, IDM_IRQ_MAIN, (BOOL)hSubWnd[SWND_IRQ_MAIN]);
	CheckMenuSub(hMenu, IDM_IRQ_SUB, (BOOL)hSubWnd[SWND_IRQ_SUB]);
	CheckMenuSub(hMenu, IDM_LOG, (BOOL)hSubWnd[SWND_LOG]);

	/* tXN[̃TuEBhEj[ */
	EnableMenuSub(hMenu, IDM_BREAKPOINT, !bFullScreen);
	EnableMenuSub(hMenu, IDM_SCHEDULER, !bFullScreen);
	EnableMenuSub(hMenu, IDM_CPU_MAIN, !bFullScreen);
	EnableMenuSub(hMenu, IDM_CPU_SUB, !bFullScreen);
	EnableMenuSub(hMenu, IDM_DISASM_MAIN, !bFullScreen);
	EnableMenuSub(hMenu, IDM_DISASM_SUB, !bFullScreen);
	EnableMenuSub(hMenu, IDM_MEMORY_MAIN, !bFullScreen);
	EnableMenuSub(hMenu, IDM_MEMORY_SUB, !bFullScreen);
#ifdef JSUB
	EnableMenuSub(hMenu, IDM_CPU_JSUB, !bFullScreen);
	EnableMenuSub(hMenu, IDM_DISASM_JSUB, !bFullScreen);
	EnableMenuSub(hMenu, IDM_MEMORY_JSUB, !bFullScreen);
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	EnableMenuSub(hMenu, IDM_CPU_MAINZ80, !bFullScreen);
	EnableMenuSub(hMenu, IDM_DISASM_MAINZ80, !bFullScreen);
#endif
	EnableMenuSub(hMenu, IDM_IRQ_MAIN, !bFullScreen);
	EnableMenuSub(hMenu, IDM_IRQ_SUB, !bFullScreen);
	EnableMenuSub(hMenu, IDM_LOG, !bFullScreen);
}

/*-[ c[j[ ]-------------------------------------------------------*/

/*
 *	}EX[h؂芷
 */
#ifdef MOUSE
void FASTCALL MouseModeChange(BOOL flag)
{
	/* [hωȂ΋A */
	if (mos_capture == flag) {
		return;
	}

	/* }EXLv`tOݒ */
	mos_capture = flag;

	/* Xe[^Xo[ɏԕ\ */
	if (flag) {
		SetStatusMessage(IDS_MOUSE_ENABLE);
	}
	else {
		SetStatusMessage(IDS_MOUSE_DISABLE);
	}
}
#endif

/*
 *	AWXg
 */
#if XM7_VER >= 2
static void FASTCALL OnTimeAdjust(void)
{
	/* Đݒ肷 */
	rtc_time_adjust();

	/* Ô߃XPW[ */
	rtc_reset();
}
#endif

/*
 *	}EX[h؂芷(M)
 */
#ifdef MOUSE
static void FASTCALL OnMouseMode(void)
{
	/* }EXLv`tO𔽓]ă[h؂ւ */
	MouseModeChange(!mos_capture);
}
#endif

/*
 *	ʃLv`(C)
 */
static void FASTCALL OnGrpCapture(void)
{
	char path[_MAX_PATH];

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_GRPCAPFILTER, path, "BMP", BMP_FILE_DIR)) {
		return;
	}

	/* Lv` */
	LockVM();
	StopSnd();
	capture_to_bmp(path, bFullScan, nMonitorType, bPseudo400Line);
	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	ʃLv`2
 */
static void FASTCALL OnGrpCapture2(void)
{
	char path[_MAX_PATH];

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_GRPCAPFILTER, path, "BMP", BMP_FILE_DIR)) {
		return;
	}

	/* Lv` */
	LockVM();
	StopSnd();
	capture_to_bmp2(path, nMonitorType, bPseudo400Line);
	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	WAVLv`(W)
 */
static void FASTCALL OnWavCapture(HWND hWnd)
{
	char path[_MAX_PATH];

	ASSERT(hWnd);

	/* ɃLv`ȂAN[Y */
	if (hWavCapture >= 0) {
		LockVM();
		CloseCaptureSnd();
		UnlockVM();
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_WAVCAPFILTER, path, "WAV", WAV_FILE_DIR)) {
		return;
	}

	/* Lv` */
	LockVM();
	OpenCaptureSnd(path);
	UnlockVM();

	/*  */
	if (hWavCapture < 0) {
		LockVM();
		StopSnd();

		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_WAVCAPERROR, path, sizeof(path));
		MessageBox(hWnd, path, "XM7", MB_ICONSTOP | MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);

		PlaySnd();
		ResetSch();
		UnlockVM();
	}
	SetMenuExitTimer();
}

/*
 *	VKfBXN쐬(D)
 */
static void FASTCALL OnNewDisk(HWND hWnd)
{
	char path[_MAX_PATH];
	int ret;
	BOOL err;

	ASSERT(hWnd);

	/* ^Cg */
	strncpy(DiskTitle, "Default", sizeof(DiskTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG),
						hWnd, TitleDlgProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_NEWDISKFILTER, path, "D77", DISK_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	if (DiskFormat) {
		err = make_new_userdisk(path, DiskTitle, DiskMedia, DiskInterleave);
	}
	else {
		err = make_new_d77(path, DiskTitle, DiskMedia);
	}
	if (err) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_NEWDISKOK, path, sizeof(path));
		MessageBox(hWnd, path, "XM7", MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	VKe[v쐬(T)
 */
static void FASTCALL OnNewTape(HWND hWnd)
{
	char path[_MAX_PATH];

	ASSERT(hWnd);

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_TAPEFILTER, path, "T77", TAPE_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	if (make_new_t77(path)) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_NEWTAPEOK, path, sizeof(path));
		MessageBox(hWnd, path, "XM7", MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	VKouJZbg쐬(B)
 */
static void FASTCALL OnNewBubble(HWND hWnd)
{
	char path[_MAX_PATH];
	int ret;
	BOOL err;

	ASSERT(hWnd);

	/* ^Cg */
	strncpy(BubbleTitle, "Default", sizeof(BubbleTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_MEDIATYPEDLG),
						hWnd, BubbleMediaTypeDlgProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* t@CI */
	if (BubbleFormat) {
		if (!FileSelectSub(FALSE, IDS_B77FILTER, path, "B77",
			BUBBLE_IMAGE_DIR)) {
			return;
		}
	}
	else {
		if (!FileSelectSub(FALSE, IDS_BBLFILTER, path, "BBL",
			BUBBLE_IMAGE_DIR)) {
			return;
		}
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	if (BubbleFormat) {
		err = make_new_bubble(path, BubbleTitle, BubbleMedia);
	}
	else {
		err = make_new_bubble(path, NULL, BubbleMedia);
	}
	if (err) {
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_NEWBUBBLEOK, path, sizeof(path));
		MessageBox(hWnd, path, "XM7", MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}
#endif

/*
 *	VFDD77ϊ(V)
 */
static void FASTCALL OnVFD2D77(HWND hWnd)
{
	char src[_MAX_PATH];
	char dst[_MAX_PATH];
	int ret;

	ASSERT(hWnd);

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_VFDFILTER, src, "VFD", DISK_IMAGE_DIR)) {
		return;
	}

	/* ^Cg */
	strncpy(DiskTitle, "Default", sizeof(DiskTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG_2D),
						hWnd, TitleDlg2DProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_NEWDISKFILTER, dst, "D77", DISK_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	keyboard_inputmask(TRUE);
	if (conv_vfd_to_d77(src, dst, DiskTitle)) {
		LoadString(hResInstance, IDS_CONVERTOK, src, sizeof(src));
	}
	else {
		LoadString(hResInstance, IDS_CONVERTFAIL, src, sizeof(src));
	}
	MessageBox(hWnd, src, "XM7", MB_OK);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	2DD77ϊ(2)
 */
static void FASTCALL On2D2D77(HWND hWnd)
{
	char src[_MAX_PATH];
	char dst[_MAX_PATH];
	int ret;

	ASSERT(hWnd);

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_2DFILTER, src, "2D", DISK_IMAGE_DIR)) {
		return;
	}

	/* ^Cg */
	strncpy(DiskTitle, "Default", sizeof(DiskTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG_2D),
						hWnd, TitleDlg2DProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_NEWDISKFILTER, dst, "D77", DISK_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	keyboard_inputmask(TRUE);
	if (conv_2d_to_d77(src, dst, DiskTitle)) {
		LoadString(hResInstance, IDS_CONVERTOK, src, sizeof(src));
	}
	else {
		LoadString(hResInstance, IDS_CONVERTFAIL, src, sizeof(src));
	}
	MessageBox(hWnd, src, "XM7", MB_OK);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	VTPT77ϊ(P)
 */
static void FASTCALL OnVTP2T77(HWND hWnd)
{
	char src[_MAX_PATH];
	char dst[_MAX_PATH];

	ASSERT(hWnd);

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_VTPFILTER, src, "VTP", TAPE_IMAGE_DIR)) {
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_TAPEFILTER, dst, "T77", TAPE_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	keyboard_inputmask(TRUE);
	if (conv_vtp_to_t77(src, dst)) {
		LoadString(hResInstance, IDS_CONVERTOK, src, sizeof(src));
	}
	else {
		LoadString(hResInstance, IDS_CONVERTFAIL, src, sizeof(src));
	}
	MessageBox(hWnd, src, "XM7", MB_OK);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	BBLB77ϊ(L)
 */
static void FASTCALL OnBBL2B77(HWND hWnd)
{
	char src[_MAX_PATH];
	char dst[_MAX_PATH];
	int ret;

	ASSERT(hWnd);

	/* t@CI */
	if (!FileSelectSub(TRUE, IDS_BBLFILTER, src, "BBL", BUBBLE_IMAGE_DIR)) {
		return;
	}

	/* ^Cg */
	strncpy(DiskTitle, "Default", sizeof(DiskTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG_2D),
						hWnd, TitleDlg2DProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* t@CI */
	if (!FileSelectSub(FALSE, IDS_B77FILTER, dst, "B77", BUBBLE_IMAGE_DIR)) {
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	keyboard_inputmask(TRUE);
	if (conv_bbl_to_b77(src, dst, DiskTitle)) {
		LoadString(hResInstance, IDS_CONVERTOK, src, sizeof(src));
	}
	else {
		LoadString(hResInstance, IDS_CONVERTFAIL, src, sizeof(src));
	}
	MessageBox(hWnd, src, "XM7", MB_OK);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}
#endif

#ifdef KBDPASTE
/*
 *	Nbv{[h\t(E)
 */
static void FASTCALL OnPaste(HWND hWnd)
{
	ASSERT(hWnd);

	/* y[Xg҂Ԑݒ肪Ȃꍇ */
	if ((uPasteWait == 0) && (uPasteWaitCntl == 0)) {
		return;
	}

	/* s */
	LockVM();
	StopSnd();

	PasteClipboardKbd(hWnd);

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	L[͎x(K)
 */
static void FASTCALL OnKeyStroke(HWND hWnd)
{
	int ret;

	ASSERT(hWnd);

	/* y[Xg҂Ԑݒ肪Ȃꍇ */
	if ((uPasteWait == 0) && (uPasteWaitCntl == 0)) {
		return;
	}

	/* s */
	if (bFullScreen || !bKeyStrokeModeless) {
		/* dɊJȂ */
		if (hKeyStrokeDialog) {
			/* EChEN[Y */
			PostMessage(hKeyStrokeDialog, WM_CLOSE, 0, 0);
			Sleep(1);
		}

		/* L[͎x_CAO */
		ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_KEYSTROKEDLG),
						hWnd, KeyStrokeDlgProc);
		if (ret != IDOK) {
			SetMenuExitTimer();
			return;
		}

		/* 쐬 */
		LockVM();
		PasteKbd((char *)KeyStrokeString);
		InvalidateRect(hDrawWnd, NULL, FALSE);
		UnlockVM();

		SetMenuExitTimer();
	}
	else {
		/* dɊJȂ */
		if (hKeyStrokeDialog) {
			SetForegroundWindow(hKeyStrokeDialog);
			return;
		}

		/* L[͎x_CAO */
		hKeyStrokeDialog = CreateDialog(
			(HINSTANCE)GetWindowLong(hWnd, GWL_HINSTANCE),
			MAKEINTRESOURCE(IDD_KEYSTROKEDLG),
			hWnd,
			(DLGPROC)KeyStrokeDlgProc);
		if (hKeyStrokeDialog == NULL) {
			return;
		}
		ShowWindow(hKeyStrokeDialog, SW_SHOW);
		UpdateWindow(hKeyStrokeDialog);
	}
}
#endif

/*
 *	TEho̓[h؂ւ
 */
static void FASTCALL OnChgSound(void)
{
	LockVM();

	/* TEh[hύX */
	uStereoOut = (uStereoOut + 1) % 5;
	SetStatusMessage(IDS_SND_MONAURAL + uStereoOut);

	/* Kp */
	ApplySnd();
	UnlockVM();
}

/*
 *	TCNX`[[h؂ւ
 */
static void FASTCALL OnChgCycleSteal(void)
{
	LockVM();

	/* TCNX`[[hύX */
	cycle_steal = !cycle_steal;

	StoreCfg();

	/* łsubclock_modeKv͂Ȃ */
	if (cycle_steal) {
		SetStatusMessage(IDS_ENABLE_CYCLESTEAL);
	}
	else {
		SetStatusMessage(IDS_DISABLE_CYCLESTEAL);
	}

	UnlockVM();
}

/*
 *	fBXN^CgύX
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL OnSetDiskTitle(HWND hWnd, int Type, int Drive, int Index)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL OnSetDiskTitle(HWND hWnd, int Drive, int Index)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char string[128];
	int ret, i;
	BOOL err;
	BYTE media;
	BOOL teject;

	ASSERT(hWnd);

#if XM7_VER == 1 && defined(SFDC)
	/* D77`FbN */
	if (fdc_ready[Type][Drive] != FDC_TYPE_D77) {
		return;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	/* D77`FbN */
	if (fdc_ready[Drive] != FDC_TYPE_D77) {
		return;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* ^Cg */
#if XM7_VER == 1 && defined(SFDC)
	strncpy(DiskTitle, fdc_name[Type][Drive][Index], sizeof(DiskTitle));
#else	/* XM7_VER == 1 && defined(SFDC) */
	strncpy(DiskTitle, fdc_name[Drive][Index], sizeof(DiskTitle));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG_2D),
						hWnd, TitleDlg2DProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

#if XM7_VER == 1 && defined(SFDC)
	err = set_title_media(fdc_fname[Type][Drive], Index, DiskTitle);
#else	/* XM7_VER == 1 && defined(SFDC) */
	err = set_title_media(fdc_fname[Drive], Index, DiskTitle);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (err) {
		for (i = 0; i < FDC_DRIVES; i ++) {
#if XM7_VER == 1 && defined(SFDC)
			if (fdc_ready[Type][i] == FDC_TYPE_NOTREADY) {
				continue;
			}
			media = fdc_media[Type][i];
			teject = fdc_teject[Type][i];
			fdc_setdisk(Type, i, fdc_fname[Type][i]);
			fdc_setmedia(Type, i, media);
			fdc_teject[Type][i] = teject;
#else	/* XM7_VER == 1 && defined(SFDC) */
			if (fdc_ready[i] == FDC_TYPE_NOTREADY) {
				continue;
			}
			media = fdc_media[i];
			teject = fdc_teject[i];
			fdc_setdisk(i, fdc_fname[i]);
			fdc_setmedia(i, media);
			fdc_teject[i] = teject;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			SetAllMRUState();
		}
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_SETTITLEOK, string, sizeof(string));
		MessageBox(hWnd, string, "XM7", MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}

/*
 *	ouJZbg^CgύX
 */
#if XM7_VER == 1 && defined(BUBBLE)
static void FASTCALL OnSetBubbleTitle(HWND hWnd, int Type, int Unit, int Index)
{
	char string[128];
	int ret, i;
	BOOL err;
	BYTE media;
	BOOL teject;

	ASSERT(hWnd);

	/* `FbN */
	if (bmc_ready[Type][Unit] != BMC_TYPE_B77) {
		return;
	}

	/* ^Cg */
	strncpy(DiskTitle, bmc_name[Type][Unit][Index], sizeof(DiskTitle));
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_TITLEDLG_2D),
						hWnd, TitleDlg2DProc);
	if (ret != IDOK) {
		SetMenuExitTimer();
		return;
	}

	/* 쐬 */
	LockVM();
	StopSnd();

	err = set_title_media(bmc_fname[Type][Unit], Index, DiskTitle);
	if (err) {
		for (i = 0; i < BMC_UNITS_32; i ++) {
			if (bmc_ready[Type][i] == BMC_TYPE_NOTREADY) {
				continue;
			}
			media = bmc_media[Type][i];
			teject = bmc_teject[Type][i];
			bmc_setfile(Type, i, bmc_fname[Type][i]);
			bmc_setmedia(Type, i, media);
			bmc_teject[Type][i] = teject;
			SetAllMRUState();
		}
		keyboard_inputmask(TRUE);
		LoadString(hResInstance, IDS_SETTITLEOK, string, sizeof(string));
		MessageBox(hWnd, string, "XM7", MB_OK);
		WAIT_KEY_POP();
		keyboard_inputmask(FALSE);
	}
	SetMenuExitTimer();

	PlaySnd();
	ResetSch();
	UnlockVM();
}
#endif	/* XM7_VER == 1 && defined(SFDC) */

/*
 *	c[(T)j[
 */
static BOOL FASTCALL OnTool(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	switch (wID) {
		/* ݒ */
		case IDM_CONFIG:
			OnConfig(hWnd);
			return TRUE;

#ifdef MOUSE
		/* }EX[h؂芷 */
		case IDM_MOUSEMODE:
			if (bDetectMouse) {
				OnMouseMode();
			}
			return TRUE;

		/* }EX[h؂芷(F11,j[\) */
		case IDM_MOUSEMODE2:
			if (!kbd_table[DIK_F11] && bDetectMouse) {
				OnMouseMode();
			}
			return TRUE;

		/* }EX[hL */
		case IDM_MOUSEON:
			if (bDetectMouse) {
				MouseModeChange(TRUE);
			}
			return TRUE;

		/* }EX[h */
		case IDM_MOUSEOFF:
			if (bDetectMouse) {
				MouseModeChange(FALSE);
			}
			return TRUE;
#endif

#if XM7_VER >= 2
		/* AWXg */
		case IDM_TIMEADJUST:
			OnTimeAdjust();
			return TRUE;
#endif

		/* ʃLv` */
		case IDM_GRPCAP:
			OnGrpCapture();
			return TRUE;

		/* ʃLv`2 */
		case IDM_GRPCAP2:
			OnGrpCapture2();
			return TRUE;

		/* WAVLv` */
		case IDM_WAVCAP:
			OnWavCapture(hWnd);
			return TRUE;

		/* VKfBXN */
		case IDM_NEWDISK:
			OnNewDisk(hWnd);
			return TRUE;

		/* VKe[v */
		case IDM_NEWTAPE:
			OnNewTape(hWnd);
			return TRUE;

#if XM7_VER == 1 && defined(BUBBLE)
		/* VKouJZbg */
		case IDM_NEWBUBBLE:
			OnNewBubble(hWnd);
			return TRUE;
#endif

		/* VFDD77 */
		case IDM_VFD2D77:
			OnVFD2D77(hWnd);
			return TRUE;

		/* 2DD77 */
		case IDM_2D2D77:
			On2D2D77(hWnd);
			return TRUE;

		/* VTPT77 */
		case IDM_VTP2T77:
			OnVTP2T77(hWnd);
			return TRUE;

#if XM7_VER == 1 && defined(BUBBLE)
		/* BBLB77 */
		case IDM_BBL2B77:
			OnBBL2B77(hWnd);
			return TRUE;
#endif

#ifdef KBDPASTE
		/* Nbv{[h\t */
		case IDM_PASTE:
			OnPaste(hWnd);
			return TRUE;

		/* L[͎x */
		case IDM_KEYSTROKE:
			OnKeyStroke(hWnd);
			return TRUE;
#endif

		/* TEho͐؂ւ(j[\) */
		case IDM_CHG_SOUNDMODE:
			OnChgSound();
			return TRUE;

		/* TCNX`[[h؂ւ(j[\) */
		case IDM_CYCLESTEAL:
			OnChgCycleSteal();
			return TRUE;
	}

#ifdef FMDLL
	/* XgZbg */
	if ((wID >= IDM_EXTSND0) && (wID <= IDM_EXTSND9)) {
		int index = wID - IDM_EXTSND0;
		if (bUseFmDll) {
			if (index == GetXDLSel(XDL_SND) + 1) {
				return TRUE;
			}
		}
		else {
			if (index == 0) {
				return TRUE;
			}
		}
		if (index == 0) {
			bUseFmDll = FALSE;
		}
		else {
			char string[256 + 1];
			bUseFmDll = TRUE;
			SetXDLSel(XDL_SND, index - 1);
			GetXDLFile(XDL_SND, index - 1, string, sizeof(string));
			strncpy(szFmDll, string, sizeof(szFmDll));
		}
		LockVM();
		StoreCfg();
		StopSnd();
		CleanSnd();
		ApplySnd();
		PlaySnd();
		ResetSch();
		UnlockVM();
		return TRUE;
	}
#endif

#ifdef FILTERDLL
	/* XgZbg */
	if ((wID >= IDM_EXTFLT0) && (wID <= IDM_EXTFLT9)) {
		int index = wID - IDM_EXTFLT0;
		if (bUseFilterDll) {
			if (index == GetXDLSel(XDL_FLT) + 1) {
				return TRUE;
			}
		}
		else {
			if (index == 0) {
				return TRUE;
			}
		}
		if (index == 0) {
			bUseFilterDll = FALSE;
		}
		else {
			char string[256 + 1];
			bUseFilterDll = TRUE;
			SetXDLSel(XDL_FLT, index - 1);
			GetXDLFile(XDL_FLT, index - 1, string, sizeof(string));
			strncpy(szFilterDll, string, sizeof(szFilterDll));
		}
		if (!(bFilterDll && bUseFilterDll)) {
			if ((nScalingSize != DISPLAYSIZE_200) &&
				(nScalingSize != DISPLAYSIZE_300) &&
				(nScalingSize != DISPLAYSIZE_400)) {
				nScalingSize = DISPLAYSIZE_100;	/* x1.0(640x400) */
			}
		}
		LockVM();
		StoreCfg();
		UnlockVM();
		if (!bFullScreen) {
			LockVM();
			StopSnd();
			ReDraw(hWnd);
			PlaySnd();
			ResetSch();
			UnlockVM();

			/* ʍĕ` */
			OnRefresh(hWnd);

			/* \eXV */
			OnSize(hWnd, 640, 400);

			InvalidateRect(hDrawWnd, NULL, FALSE);
		}
		return TRUE;
	}
#endif

	return FALSE;
}

/*
 *	c[(T)j[XV
 */
static void FASTCALL OnToolPopup(HMENU hMenu)
{
	UINT uitem, uitem2;
	HMENU hSubMenu;
	MENUITEMINFO mii;
	char string[256+1];
	char buffer[256+1];
	int	i, j, k;

	ASSERT(hMenu);

#ifdef MOUSE
	/* j[ڂ炢폜 */
	DeleteMenu(hMenu, IDM_MOUSEMODE, MF_BYCOMMAND);

	/* ߂čڂ˂ */
	if (kbd_table[DIK_F11]) {
		LoadString(hResInstance, IDS_MOUSEMODE_DISF11, string, sizeof(string));
	}
	else {
		LoadString(hResInstance, IDS_MOUSEMODE_ENBF11, string, sizeof(string));
	}

	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;
	mii.wID = IDM_MOUSEMODE;
	mii.dwTypeData = string;
	mii.cch = strlen(string);
#if XM7_VER >= 2
	InsertMenuItem(hMenu, IDM_TIMEADJUST, MF_BYCOMMAND, &mii);
#else
	InsertMenuItem(hMenu, 2, MF_BYPOSITION, &mii);
#endif
	EnableMenuItem(hMenu, IDM_MOUSEMODE, !bDetectMouse);

	/* }EX[h */
	CheckMenuSub(hMenu, IDM_MOUSEMODE, mos_capture);
#endif

	/* WAVLv` nhŃI[v */
	CheckMenuSub(hMenu, IDM_WAVCAP, (hWavCapture >= 0));

	uitem = 12;
#ifdef MOUSE
	uitem += 1;		/* }EX[h */
#endif
#if XM7_VER >= 2
	uitem += 1;		/* AWXg */
#endif
#if defined(MOUSE) || (XM7_VER >= 2)
	uitem += 1;		/* Zp[^ */
#endif
#ifdef S98LOG
	uitem += 1;		/* S98Lv` */
#endif
#if XM7_VER == 1 && defined(BUBBLE)
	uitem += 2;		/* ouJZbg쐬/BBLB77ϊ */
#endif

#ifdef FMDLL
	/* OXgǉ */
	AddMenuList(hMenu, uitem, IDS_FMGEN_BUILDIN, IDM_EXTSND0, XDL_SND, bUseFmDll);
	if (bRomeo && bUseRomeo) {
		/* ROMEOgp͑Is */
		EnableMenuItem(hMenu, uitem ++, MF_BYPOSITION | MF_GRAYED);
	}
	if (bSCCI_available && bUseSCCI) {
		/* SCCIgp͑Is */
		EnableMenuItem(hMenu, uitem ++, MF_BYPOSITION | MF_GRAYED);
	}
	uitem ++;
#endif

#ifdef FILTERDLL
	/* O摜tB^Xgǉ */
	AddMenuList(hMenu, uitem ++, IDS_NO_FILTER, IDM_EXTFLT0, XDL_FLT, bUseFilterDll);
#endif

#ifdef KBDPASTE
	/* Nbv{[h\t/L[͎x֘Aj[ڂU폜 */
	while (GetMenuItemCount(hMenu) > (int)uitem) {
		DeleteMenu(hMenu, uitem, MF_BYPOSITION);
	}

	/* y[Xg҂Ԑݒ肪΍ēxǉ */
	if ((uPasteWait != 0) || (uPasteWaitCntl != 0)) {
		/* j[\̏ */
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */

		/* Zp[^} */
		mii.fType = MFT_SEPARATOR;
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.fState = MFS_ENABLED;
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);

		LoadString(hResInstance, IDS_PASTE, string, sizeof(string));
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.fType = MFT_STRING;
		mii.fState = MFS_ENABLED;
		mii.wID = IDM_PASTE;
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);

		LoadString(hResInstance, IDS_KEYSTROKE, string, sizeof(string));
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
		mii.fType = MFT_STRING;
		mii.fState = MFS_ENABLED;
		mii.wID = IDM_KEYSTROKE;
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, uitem++, TRUE, &mii);
	}
#endif
}

/*-[ EBhEj[ ]---------------------------------------------------*/

/*
 *	SEChE̕\ԂꊇĕύX
 */
static void FASTCALL ShowAllWindowSub(int CmdShow)
{
	int i;

	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i]) {
			ShowWindow(hSubWnd[i], CmdShow);
		}
	}
}

/*
 *	d˂ĕ\(C)
 */
static void FASTCALL OnCascade(void)
{
	/* d˂ĕ\ */
	ASSERT(hDrawWnd);
	CascadeWindows(hDrawWnd, 0, NULL, 0, NULL);
}

/*
 *	ׂĕ\(T)
 */
static void FASTCALL OnTile(void)
{
	/* ׂĕ\ */
	ASSERT(hDrawWnd);
	TileWindows(hDrawWnd, MDITILE_VERTICAL, NULL, 0, NULL);
}

/*
 *	SăACR(I)
 */
static void FASTCALL OnIconic(void)
{
	ShowAllWindowSub(SW_MINIMIZE);
}

/*
 *	ACR̐(A)
 */
static void FASTCALL OnArrangeIcon(void)
{
	/* ACR̐ */
	ASSERT(hDrawWnd);
	ArrangeIconicWindows(hDrawWnd);
}

/*
 *	SĉB(H)
 */
static void FASTCALL OnHide(void)
{
	ShowAllWindowSub(SW_HIDE);
}

/*
 *	Sĕ(R)
 */
static void FASTCALL OnRestore(void)
{
	ShowAllWindowSub(SW_RESTORE);

	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}
}

/*
 *	Sĕ(O)
 */
static void FASTCALL OnClose(void)
{
	int i;

	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i]) {
			DestroyWindow(hSubWnd[i]);
			hSubWnd[i] = NULL;
		}
	}
}

/*
 *	EBhE(W)j[
 */
static BOOL FASTCALL OnWindow(WORD wID)
{
	int i;

	switch (wID) {
		/* d˂ĕ\ */
		case IDM_CASCADE:
			OnCascade();
			return TRUE;

		/* ׂĕ\ */
		case IDM_TILE:
			OnTile();
			return TRUE;

		/* SăACR */
		case IDM_ICONIC:
			OnIconic();
			return TRUE;

		/* ACR̐ */
		case IDM_ARRANGEICON:
			OnArrangeIcon();
			return TRUE;

		/* SĉB */
		case IDM_ALLHIDE:
			OnHide();
			return TRUE;

		/* Sĕ */
		case IDM_ALLRESTORE:
			OnRestore();
			return TRUE;

		/* Sĕ */
		case IDM_ALLCLOSE:
			OnClose();
			return TRUE;
	}

	/* EBhEI */
	if ((wID >= IDM_SWND00) && (wID < IDM_SWND00 + SWND_MAXNUM)) {
		for (i=0; i<SWND_MAXNUM; i++) {
			if (hSubWnd[i] == NULL) {
				continue;
			}

			/* JEg_EAOK */
			if (wID == IDM_SWND00) {
				ShowWindow(hSubWnd[i], SW_RESTORE);
				SetWindowPos(hSubWnd[i], HWND_TOP, 0, 0, 0, 0,
					SWP_NOMOVE | SWP_NOSIZE);

				if (bPopupSwnd) {
					SetForegroundWindow(hMainWnd);
				}
				break;
			}
			else {
				wID--;
			}
		}
	}

	return FALSE;
}

/*
 *	EBhE(W)j[XV
 */
static void FASTCALL OnWindowPopup(HMENU hMenu)
{
	int i;
	BOOL flag;
	MENUITEMINFO mii;
	UINT nID;
	int count;
	char string[128];

	ASSERT(hMenu);

	/* tXN[̐擪7̖yу|bvAbv̐擪2+4Ԗڂ̖ */
	EnableMenuSub(hMenu, IDM_CASCADE, !bFullScreen && !bPopupSwnd);
	EnableMenuSub(hMenu, IDM_TILE, !bFullScreen && !bPopupSwnd);
	EnableMenuSub(hMenu, IDM_ICONIC, !bFullScreen);
	EnableMenuSub(hMenu, IDM_ARRANGEICON, !bFullScreen && !bPopupSwnd);
	EnableMenuSub(hMenu, IDM_ALLHIDE, !bFullScreen);
	EnableMenuSub(hMenu, IDM_ALLRESTORE, !bFullScreen);
	EnableMenuSub(hMenu, IDM_ALLCLOSE, !bFullScreen);

	/* 擪̂Vcč폜 */
	while (GetMenuItemCount(hMenu) > 7) {
		DeleteMenu(hMenu, 7, MF_BYPOSITION);
	}

	/* LȃTuEChEȂ΁Â܂܃^[ */
	flag = FALSE;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] != NULL) {
			flag = TRUE;
		}
	}
	if (!flag) {
		return;
	}

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* Zp[^} */
	mii.fType = MFT_SEPARATOR;
	InsertMenuItem(hMenu, 7, TRUE, &mii);
	mii.fType = MFT_STRING;

	/* EChE^CgZbg */
	count = 0;
	nID = IDM_SWND00;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] != NULL) {
			/* j[} */
			mii.wID = nID;
			memset(string, 0, sizeof(string));
			GetWindowText(hSubWnd[i], string, sizeof(string) - 1);
			mii.dwTypeData = string;
			mii.cch = strlen(string);
			InsertMenuItem(hMenu, count + 8, TRUE, &mii);
			EnableMenuSub(hMenu, nID, !bFullScreen);

			/*  */
			nID++;
			count++;
		}
	}
}

/*-[ wvj[ ]-------------------------------------------------------*/

#ifdef RES_DLL
/*
 *	ύX(L)
 */
static void FASTCALL OnLangage(void)
{
	HMENU hMenu;

	if (bExistResourceDll) {
		/* TuEBhESĕ */
		OnClose();

		bLanguageMode = !bLanguageMode;
		if (bLanguageMode) {
			hResInstance = hInstRes;
		}
		else {
			hResInstance = hAppInstance;
		}

		LockVM();

		/* Cj[[h */
		hMenu = GetMenu(hMainWnd);
		SetMenu(hMainWnd, NULL);
		DestroyMenu(hMenu);
		hMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MAINMENU));
		SetMenu(hMainWnd, hMenu);

		/* fBAj[[h */
		DestroyMenu(hMediaMenu);
		hMediaMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MEDIAMENU));

		/* qXgj[[h */
		DestroyMenu(hHistoryMenu);
		hHistoryMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_HISTORYMENU));

		/* AXyNgݒ胁j[[h */
		DestroyMenu(hAspectMenu);
		hAspectMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_ASPECTMENU));

		/* ʃbZ[W[h */
		LoadMessageScreen();

		UnlockVM();
	}
}
#endif

/*
 *	wv(H)j[
 */
static BOOL FASTCALL OnHelp(HWND hWnd, WORD wID)
{
	switch (wID) {
		/* o[W */
		case IDM_ABOUT:
			OnAbout(hWnd);
			return TRUE;

#ifdef RES_DLL
		/* ύX */
		case IDM_LANGUAGE:
			OnLangage();
			return TRUE;
#endif
	}

	return FALSE;
}

#ifdef RES_DLL
/*
 *	wv(H)j[XV
 */
static void FASTCALL OnHelpPopup(HMENU hMenu)
{
	int i;
	BOOL flag;
	MENUITEMINFO mii;
	UINT nID;
	int count;
	char string[128];

	ASSERT(hMenu);

	/* \[XDLLȂΉȂ */
	if ((!bUseLanguageMode) ||
		(!bExistResourceDll)) {
		return;
	}

	/* 擪1폜 */
	if (GetMenuItemCount(hMenu) > 1) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;	/* sizeof(mii)WINVER>=0x0500 */
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/*  */
	mii.wID = IDM_LANGUAGE;
	LoadString(hResInstance, IDS_LANGUAGE, string, sizeof(string));
#if 1	/* V1.2L12/V2.9L52/V3.4L52̉ݒ */
	if (!strlen(string)) {
		strncpy(string, "&Japanese Mode", sizeof(string));
	}
#endif
	mii.dwTypeData = string;
	mii.cch = strlen(string);
	InsertMenuItem(hMenu, 0, TRUE, &mii);
}
#endif

/*-[ ReLXgj[ ]-------------------------------------------------*/

/*
 *	fBA^CgύX(&T)
 */
static BOOL FASTCALL OnMediaChangeTitle(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	if ((wID >= IDM_D0MEDIA00) && (wID <= IDM_D0MEDIA15)) {
#if XM7_VER == 1 && defined(SFDC)
		OnSetDiskTitle(hWnd, FDC_M, 0, wID - IDM_D0MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		OnSetDiskTitle(hWnd, 0, wID - IDM_D0MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		return TRUE;
	}
	if ((wID >= IDM_D1MEDIA00) && (wID <= IDM_D1MEDIA15)) {
#if XM7_VER == 1 && defined(SFDC)
		OnSetDiskTitle(hWnd, FDC_M, 1, wID - IDM_D1MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		OnSetDiskTitle(hWnd, 1, wID - IDM_D1MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		return TRUE;
	}
	if ((wID >= IDM_D2MEDIA00) && (wID <= IDM_D2MEDIA15)) {
#if XM7_VER == 1 && defined(SFDC)
		OnSetDiskTitle(hWnd, FDC_M, 2, wID - IDM_D2MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		OnSetDiskTitle(hWnd, 2, wID - IDM_D2MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		return TRUE;
	}
	if ((wID >= IDM_D3MEDIA00) && (wID <= IDM_D3MEDIA15)) {
#if XM7_VER == 1 && defined(SFDC)
		OnSetDiskTitle(hWnd, FDC_M, 3, wID - IDM_D3MEDIA00);
#else	/* XM7_VER == 1 && defined(SFDC) */
		OnSetDiskTitle(hWnd, 3, wID - IDM_D3MEDIA00);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		return TRUE;
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((wID >= IDM_D4MEDIA00) && (wID <= IDM_D4MEDIA15)) {
		OnSetDiskTitle(hWnd, FDC_S, 0, wID - IDM_D4MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_D5MEDIA00) && (wID <= IDM_D5MEDIA15)) {
		OnSetDiskTitle(hWnd, FDC_S, 1, wID - IDM_D5MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_D6MEDIA00) && (wID <= IDM_D6MEDIA15)) {
		OnSetDiskTitle(hWnd, FDC_S, 2, wID - IDM_D6MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_D7MEDIA00) && (wID <= IDM_D7MEDIA15)) {
		OnSetDiskTitle(hWnd, FDC_S, 3, wID - IDM_D7MEDIA00);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(BUBBLE)
	if ((wID >= IDM_B0MEDIA00) && (wID <= IDM_B0MEDIA15)) {
		OnSetBubbleTitle(hWnd, BMC_32, 0, wID - IDM_B0MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_B1MEDIA00) && (wID <= IDM_B1MEDIA15)) {
		OnSetBubbleTitle(hWnd, BMC_32, 1, wID - IDM_B1MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_B2MEDIA00) && (wID <= IDM_B2MEDIA15)) {
		OnSetBubbleTitle(hWnd, BMC_128, 0, wID - IDM_B2MEDIA00);
		return TRUE;
	}
	if ((wID >= IDM_B3MEDIA00) && (wID <= IDM_B3MEDIA15)) {
		OnSetBubbleTitle(hWnd, BMC_128, 1, wID - IDM_B3MEDIA00);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

	return FALSE;
}

/*
 *	qXg폜(&D)
 */
static BOOL FASTCALL OnHistoryDelete(HWND hWnd, WORD wID)
{
	/* qXgENbNj[ */
	if ((wID >= IDM_STATE_FILE0) && (wID <= IDM_STATE_FILE9)) {
		DeleteHistory(hWnd, wID);
		return TRUE;
	}
	if ((wID >= IDM_D3_FILE0) && (wID <= IDM_D0_FILE9)) {
		DeleteHistory(hWnd, wID);
		return TRUE;
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((wID >= IDM_D7_FILE0) && (wID <= IDM_D4_FILE9)) {
		DeleteHistory(hWnd, wID);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if ((wID >= IDM_T_FILE0) && (wID <= IDM_T_FILE9)) {
		DeleteHistory(hWnd, wID);
		return TRUE;
	}
#if XM7_VER == 1 && defined(BUBBLE)
	if ((wID >= IDM_B3_FILE0) && (wID <= IDM_B0_FILE9)) {
		DeleteHistory(hWnd, wID);
		return TRUE;
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

	return FALSE;
}

#ifdef FILTERDLL
/*
 *	AXyNgݒ(&A)
 */
static BOOL FASTCALL OnAspectSetting(HWND hWnd, WORD wID)
{
	/* qXgENbNj[ */
	if (wID == IDM_ASPECTRATIO) {
		AspectSetting(hWnd, wID);
		return TRUE;
	}

	return FALSE;
}
#endif

/*
 *	ReLXgj[̌Ăяo
 */
static BOOL FASTCALL OnSubMenu(HWND hWnd, WORD wID)
{
	switch (wID) {
		/* fBA^CgύX(&T) */
		case IDM_MEDIATITLE:
			OnMediaChangeTitle(hWnd, wMenuID);
			return TRUE;

		/* qXg폜(&D) */
		case IDM_HISTORY_DELETE:
			OnHistoryDelete(hWnd, wMenuID);
			return TRUE;

#ifdef FILTERDLL
		/* AXyNgݒ(&A) */
		case IDM_ASPECTSETTING:
			OnAspectSetting(hWnd, wMenuID);
			return TRUE;
#endif
	}

	return FALSE;
}

/*-[ j[R}h ]-------------------------------------------------*/

/*
 *	j[R}h
 */
void FASTCALL OnCommand(HWND hWnd, WORD wID)
{
	ASSERT(hWnd);

	if (OnFile(hWnd, wID)) {
		return;
	}
	if (OnDisk(wID)) {
		return;
	}
	if (OnTape(wID)) {
		return;
	}
	if (OnView(hWnd, wID)) {
		return;
	}
	if (OnDebug(hWnd, wID)) {
		return;
	}
	if (OnTool(hWnd, wID)) {
		return;
	}
	if (OnWindow(wID)) {
		return;
	}
	if (OnHelp(hWnd, wID)) {
		return;
	}
	if (OnSubMenu(hWnd, wID)) {
		return;
	}
}

/*
 *	j[R}hXV
 */
void FASTCALL OnMenuPopup(HWND hWnd, HMENU hSubMenu, UINT uPos)
{
	HMENU hMenu;

	ASSERT(hWnd);
	ASSERT(hSubMenu);

	/* Cj[̍XV`FbN */
	hMenu = GetMenu(hWnd);
	if (GetSubMenu(hMenu, uPos) != hSubMenu) {
		return;
	}

	/* bNKv */
	LockVM();

	switch (uPos) {
		/* t@C */
		case FILE_MENU:
			OnFilePopup(hSubMenu);
			break;

#if XM7_VER == 1 && defined(SFDC)
		/* 1MB FDD */
		case SFD_MENU:
			OnDiskPopup(hSubMenu, FDC_S);
			break;

		/* 320KB FDD */
		case MFD_MENU:
			OnDiskPopup(hSubMenu, FDC_M);
			break;
#else	/* XM7_VER == 1 && defined(SFDC) */
		/* hCu */
		case MFD_MENU:
			OnDiskPopup(hSubMenu);
			break;
#endif	/* XM7_VER == 1 && defined(SFDC) */

#if XM7_VER == 1 && defined(BUBBLE)
		/* JZbg */
		case CASETTE_MENU:
			OnCassettePopup(hSubMenu);
			break;
#else	/* XM7_VER == 1 && defined(BUBBLE) */
		/* e[v */
		case TAPE_MENU:
			OnTapePopup(hSubMenu);
			break;
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

		/* \ */
		case VIEW_MENU:
			OnViewPopup(hSubMenu);
			break;

		/* fobO */
		case DEBUG_MENU:
			OnDebugPopup(hSubMenu);
			break;

		/* c[ */
		case TOOL_MENU:
			OnToolPopup(hSubMenu);
			break;

		/* EBhE */
		case WINDOW_MENU:
			OnWindowPopup(hSubMenu);
			break;

#ifdef RES_DLL
		/* wv */
		case HELP_MENU:
			OnHelpPopup(hSubMenu);
			break;
#endif
	}

	/* AbN */
	UnlockVM();
}

/*
 *	ReLXgj[
 */
void FASTCALL OnContextualMenu(HWND hWnd, WORD wID, int x, int y)
{
	ASSERT(hWnd);

	if (OnFileContextualMenu(hWnd, wID, x, y)) {
		return;
	}
	if (OnDiskContextualMenu(hWnd, wID, x, y)) {
		return;
	}
	if (OnTapeContextualMenu(hWnd, wID, x, y)) {
		return;
	}
#ifdef FILTERDLL
	if (OnViewContextualMenu(hWnd, wID, x, y)) {
		return;
	}
#endif
}

/*-[ hbOhbvER}hC ]-------------------------------*/

/*
 *	t@ChbvTu
 */
void FASTCALL OnDropSub(char *path)
{
	char dir[_MAX_DIR];
	char ext[_MAX_EXT];
	char InitDir[_MAX_PATH];

	ASSERT(path);

	/* gq */
	_splitpath(path, InitDir, dir, NULL, ext);
	strncat(InitDir, dir, sizeof(InitDir) - strlen(InitDir) - 1);

	/* D77 */
	if (stricmp(ext, ".D77") == 0) {
		strncpy(InitialDir[DISK_IMAGE_DIR], InitDir,
			sizeof(InitialDir[DISK_IMAGE_DIR]));
		LockVM();
		StopSnd();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setdisk(FDC_M, 0, path);
		fdc_setdisk(FDC_M, 1, NULL);
		fdc_setdisk(FDC_M, 2, NULL);
		fdc_setdisk(FDC_M, 3, NULL);
		if (fdc_ready[FDC_M][0] != FDC_TYPE_NOTREADY) {
			if (fdc_medias[FDC_M][0] >= 2) {
				fdc_setdisk(FDC_M, 1, path);
				fdc_setmedia(FDC_M, 1, 1);
			}
			if ((boot_mode != BOOT_BASIC) &&
				(boot_mode != BOOT_DOS) &&
				(boot_mode != BOOT_BASIC_DBG)) {
				boot_mode = BOOT_BASIC;
			}
			system_reset();
			StatePath[0] = '\0';
		}
		if (fdc_enable[FDC_S]) {
			fdc_setdisk(FDC_S, 0, path);
			fdc_setdisk(FDC_S, 1, NULL);
			fdc_setdisk(FDC_S, 2, NULL);
			fdc_setdisk(FDC_S, 3, NULL);
			if (fdc_ready[FDC_S][0] != FDC_TYPE_NOTREADY) {
				if (fdc_medias[FDC_S][0] >= 2) {
					fdc_setdisk(FDC_S, 1, path);
					fdc_setmedia(FDC_S, 1, 1);
				}
				if (availableSFDBOOT) {
					if (boot_mode != BOOT_DOS_SFD) {
						boot_mode = BOOT_DOS_SFD;
					}
					system_reset();
					StatePath[0] = '\0';
				}
			}
		}
		SetAllMRUState();
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setdisk(0, path);
		fdc_setdisk(1, NULL);
		fdc_setdisk(2, NULL);
		fdc_setdisk(3, NULL);
		if ((fdc_ready[0] != FDC_TYPE_NOTREADY) && (fdc_medias[0] >= 2)) {
			fdc_setdisk(1, path);
			fdc_setmedia(1, 1);
		}
#if XM7_VER == 1
		if ((fdc_ready[0] != FDC_TYPE_NOTREADY) &&
			(boot_mode != BOOT_BASIC) &&
			(boot_mode != BOOT_DOS) &&
			(boot_mode != BOOT_BASIC_DBG)) {
			boot_mode = BOOT_BASIC;
		}
#endif	/* XM7_VER == 1 */
		system_reset();
		StatePath[0] = '\0';
		SetAllMRUState();
		OnRefresh(hMainWnd);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		PlaySnd();
		ResetSch();
		UnlockVM();
	}

	/* 2D/2DD/VFD */
#if XM7_VER >= 3
	if ((stricmp(ext, ".2D") == 0) || (stricmp(ext, ".2DD") == 0) ||
		(stricmp(ext, ".VFD") == 0)) {
#else
	if ((stricmp(ext, ".2D") == 0) || (stricmp(ext, ".VFD") == 0)) {
#endif
		strncpy(InitialDir[DISK_IMAGE_DIR], InitDir,
			sizeof(InitialDir[DISK_IMAGE_DIR]));
		LockVM();
		StopSnd();
#if XM7_VER == 1 && defined(SFDC)
		fdc_setdisk(FDC_M, 0, path);
		fdc_setdisk(FDC_M, 1, NULL);
		fdc_setdisk(FDC_M, 2, NULL);
		fdc_setdisk(FDC_M, 3, NULL);
		if ((fdc_ready[FDC_M][0] != FDC_TYPE_NOTREADY) &&
			(boot_mode != BOOT_BASIC) &&
			(boot_mode != BOOT_DOS) &&
			(boot_mode != BOOT_BASIC_DBG)) {
			boot_mode = BOOT_BASIC;
		}
#else	/* XM7_VER == 1 && defined(SFDC) */
		fdc_setdisk(0, path);
		fdc_setdisk(1, NULL);
		fdc_setdisk(2, NULL);
		fdc_setdisk(3, NULL);
#if XM7_VER == 1
		if ((fdc_ready[0] != FDC_TYPE_NOTREADY) &&
			(boot_mode != BOOT_BASIC) &&
			(boot_mode != BOOT_DOS) &&
			(boot_mode != BOOT_BASIC_DBG)) {
			boot_mode = BOOT_BASIC;
		}
#endif
#endif	/* XM7_VER == 1 && defined(SFDC) */
		system_reset();
		StatePath[0] = '\0';
		SetAllMRUState();
		OnRefresh(hMainWnd);
		PlaySnd();
		ResetSch();
		UnlockVM();
	}

#if XM7_VER == 1 && defined(SFDC)
	if (fdc_enable[FDC_S]) {
		/* 2HD */
		if (stricmp(ext, ".2HD") == 0) {
			strncpy(InitialDir[DISK_IMAGE_DIR], InitDir, sizeof(InitialDir[DISK_IMAGE_DIR]));
			LockVM();
			StopSnd();
			fdc_setdisk(FDC_S, 0, path);
			fdc_setdisk(FDC_S, 1, NULL);
			fdc_setdisk(FDC_S, 2, NULL);
			fdc_setdisk(FDC_S, 3, NULL);
			if ((fdc_ready[FDC_S][0] != FDC_TYPE_NOTREADY) &&
				(availableSFDBOOT)) {
				if (boot_mode != BOOT_DOS_SFD) {
					boot_mode = BOOT_DOS_SFD;
				}
				system_reset();
				StatePath[0] = '\0';
			}
			SetAllMRUState();
			PlaySnd();
			ResetSch();
			UnlockVM();
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* T77 */
	if (stricmp(ext, ".T77") == 0) {
		strncpy(InitialDir[TAPE_IMAGE_DIR], InitDir,
			sizeof(InitialDir[TAPE_IMAGE_DIR]));
		LockVM();
		tape_setfile(path);
		SetAllMRUState();
		UnlockVM();
	}

#if XM7_VER == 1 && defined(BUBBLE)
	if (bmc_enable[BMC_32] || bmc_enable[BMC_128]) {
		/* B77 */
		if (stricmp(ext, ".B77") == 0) {
			strncpy(InitialDir[BUBBLE_IMAGE_DIR], InitDir,
				sizeof(InitialDir[BUBBLE_IMAGE_DIR]));
			LockVM();
			StopSnd();
			if (bmc_enable[BMC_32]) {
				bmc_setfile(BMC_32, 0, path);
				bmc_setfile(BMC_32, 1, NULL);
				if (bmc_ready[BMC_32][0] != BMC_TYPE_NOTREADY) {
					if (bmc_medias[BMC_32][0] >= 2) {
						bmc_setfile(BMC_32, 1, path);
						bmc_setmedia(BMC_32, 1, 1);
					}
					if (availableBBLBOOT) {
						if (boot_mode != BOOT_BUBBLE) {
							boot_mode = BOOT_BUBBLE;
							mainmem_bubblemode_boot();
						}
						system_reset();
						StatePath[0] = '\0';
					}
				}
			}
			if (bmc_enable[BMC_128]) {
				bmc_setfile(BMC_128, 0, path);
				bmc_setfile(BMC_128, 1, NULL);
				if (bmc_ready[BMC_128][0] != BMC_TYPE_NOTREADY) {
					if (bmc_medias[BMC_128][0] >= 2) {
						bmc_setfile(BMC_128, 1, path);
						bmc_setmedia(BMC_128, 1, 1);
					}
					if (availableBBLBOOT128) {
						if (boot_mode != BOOT_BUBBLE128) {
							boot_mode = BOOT_BUBBLE128;
							mainmem_bubblemode_boot();
						}
						system_reset();
						StatePath[0] = '\0';
					}
				}
			}
			SetAllMRUState();
			PlaySnd();
			ResetSch();
			UnlockVM();
		}

		/* BBL */
		if (stricmp(ext, ".BBL") == 0) {
			strncpy(InitialDir[BUBBLE_IMAGE_DIR], InitDir,
				sizeof(InitialDir[BUBBLE_IMAGE_DIR]));
			LockVM();
			StopSnd();
			if (bmc_enable[BMC_32]) {
				bmc_setfile(BMC_32, 0, path);
				bmc_setfile(BMC_32, 1, NULL);
				if ((bmc_ready[BMC_32][0] != BMC_TYPE_NOTREADY) &&
					(availableBBLBOOT)) {
					if (boot_mode != BOOT_BUBBLE) {
						boot_mode = BOOT_BUBBLE;
						mainmem_bubblemode_boot();
					}
					system_reset();
					StatePath[0] = '\0';
				}
			}
			if (bmc_enable[BMC_128]) {
				bmc_setfile(BMC_128, 0, path);
				bmc_setfile(BMC_128, 1, NULL);
				if ((bmc_ready[BMC_128][0] != BMC_TYPE_NOTREADY) &&
					(availableBBLBOOT128)) {
					if (boot_mode != BOOT_BUBBLE128) {
						boot_mode = BOOT_BUBBLE128;
						mainmem_bubblemode_boot();
					}
					system_reset();
					StatePath[0] = '\0';
				}
			}
			SetAllMRUState();
			PlaySnd();
			ResetSch();
			UnlockVM();
		}
	}
#endif

	/* XM7 */
	if (stricmp(ext, ".XM7") == 0) {
		strncpy(InitialDir[STATE_FILE_DIR], InitDir,
			sizeof(InitialDir[STATE_FILE_DIR]));
		LockVM();
		StopSnd();
		StateLoad(path);
		StoreCfg();
		PlaySnd();
		ResetSch();
		UnlockVM();
	}

	/* \eXV */
	if (hDrawWnd) {
		InvalidateRect(hDrawWnd, NULL, FALSE);
	}
}

/*
 *	Xe[^Xo[
 *	t@ChbvTu
 */
void FASTCALL OnBarDropSub(char *path, POINT point)
{
	char dir[_MAX_DIR];
	char ext[_MAX_EXT];
	char InitDir[_MAX_PATH];
	int drive;
#if XM7_VER == 1 && defined(SFDC)
	int type;
#endif	/* XM7_VER == 1 && defined(SFDC) */

	ASSERT(path);

	/* gq */
	_splitpath(path, InitDir, dir, NULL, ext);
	strncat(InitDir, dir, sizeof(InitDir) - strlen(InitDir) - 1);

	/* fBXNC[W */
#if XM7_VER == 1 && defined(SFDC)
	if ((point.x >= uPaneX[0]) && (point.x < uPaneX[8])) {
		/* hCu */
		if (point.x >= uPaneX[7]) {
			type = FDC_M;
			drive = 0;
		}
		else if (point.x >= uPaneX[6]) {
			type = FDC_M;
			drive = 1;
		}
		else if (point.x >= uPaneX[5]) {
			type = FDC_M;
			drive = 2;
		}
		else if (point.x >= uPaneX[4]) {
			type = FDC_M;
			drive = 3;
		}
		else if (point.x >= uPaneX[3]) {
			type = FDC_S;
			drive = 0;
		}
		else if (point.x >= uPaneX[2]) {
			type = FDC_S;
			drive = 1;
		}
		else if (point.x >= uPaneX[1]) {
			type = FDC_S;
			drive = 2;
		}
		else {
			type = FDC_S;
			drive = 3;
		}
		if ((type == FDC_S) && !fdc_enable[FDC_S]) {
			OnDropSub(path);
			return;
		}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if ((point.x >= uPaneX[0]) && (point.x < uPaneX[4])) {
		/* hCu */
		if (point.x >= uPaneX[3]) {
			drive = 0;
		}
		else if (point.x >= uPaneX[2]) {
			drive = 1;
		}
		else if (point.x >= uPaneX[1]) {
			drive = 2;
		}
		else {
			drive = 3;
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */

		/* D77 */
		if (stricmp(ext, ".D77") == 0) {
			strncpy(InitialDir[DISK_IMAGE_DIR], InitDir,
				sizeof(InitialDir[DISK_IMAGE_DIR]));
			LockVM();
			StopSnd();
#if XM7_VER == 1 && defined(SFDC)
			fdc_setdisk(type, drive, path);
#else	/* XM7_VER == 1 && defined(SFDC) */
			fdc_setdisk(drive, path);
#endif	/* XM7_VER == 1 && defined(SFDC) */
			SetAllMRUState();
			PlaySnd();
			ResetSch();
			UnlockVM();
			return;
		}

		/* 2D/2DD/2HD/VFD */
#if XM7_VER >= 3
		if ((stricmp(ext, ".2D") == 0) || (stricmp(ext, ".2DD") == 0) ||
			(stricmp(ext, ".VFD") == 0)) {
#elif XM7_VER == 1 && defined(SFDC)
		if ((stricmp(ext, ".2D") == 0) || (stricmp(ext, ".2HD") == 0) ||
			(stricmp(ext, ".VFD") == 0)) {
#else
		if ((stricmp(ext, ".2D") == 0) || (stricmp(ext, ".VFD") == 0)) {
#endif
			strncpy(InitialDir[DISK_IMAGE_DIR], InitDir,
				sizeof(InitialDir[DISK_IMAGE_DIR]));
			LockVM();
			StopSnd();
#if XM7_VER == 1 && defined(SFDC)
			fdc_setdisk(type, drive, path);
#else	/* XM7_VER == 1 && defined(SFDC) */
			fdc_setdisk(drive, path);
#endif	/* XM7_VER == 1 && defined(SFDC) */
			SetAllMRUState();
			PlaySnd();
			ResetSch();
			UnlockVM();
			return;
		}
	}

	/* fBXNC[WȊOE͈͊O */
	OnDropSub(path);
}

/*
 *	t@Chbv
 */
void FASTCALL OnDropFiles(HANDLE hDrop)
{
	char path[_MAX_PATH];
	POINT point;
	POINT spoint;
	HWND hwnd;

	ASSERT(hDrop);

	/* t@C󂯎 */
	DragQueryPoint(hDrop, &point);
	DragQueryFile(hDrop, 0, path, sizeof(path));
	DragFinish(hDrop);

	/*  */
	spoint = point;
	ClientToScreen(hMainWnd, &spoint);
	hwnd = WindowFromPoint(spoint);
	if (hwnd == hStatusBar) {
		/* Xe[^Xo[ւ̃hbv */
		OnBarDropSub(path, point);
	}
	else {
		/* h[EBhEւ̃hbv͏] */
		OnDropSub(path);
	}

	LockVM();
	SetAllMRUState();
	UnlockVM();
}

/*
 *	R}hC
 */
void FASTCALL OnCmdLine(LPSTR lpCmdLine)
{
	char dir[_MAX_DIR];
	char ext[_MAX_EXT];
	char path[_MAX_PATH];
	char fullpath[_MAX_PATH];
	char d77_path[_MAX_PATH];
	char InitDir[_MAX_PATH];
	LPSTR p;
	LPSTR q;
	BOOL flag;
	int drive;
#if XM7_VER == 1
 #ifdef SFDC
	int sdrive;
 #endif
 #ifdef BUBBLE
	char bbl_path[_MAX_PATH];
	int unit1, unit2;
 #endif
	BOOL boot;
#elif XM7_VER == 1 && defined(BUBBLE)
	int unit;
	char bbl_path[_MAX_PATH];
#endif
	BOOL tape_set;

	ASSERT(lpCmdLine);

	/* [N */
	drive = 0;
#if XM7_VER == 1
 #ifdef SFDC
	sdrive = 0;
 #endif
 #ifdef BUBBLE
	unit1 = unit2 = 0;
 #endif
	boot = FALSE;
#elif XM7_VER == 1 && defined(BUBBLE)
	unit = 0;
#endif
	tape_set = FALSE;
	d77_path[0] = '\0';
#if XM7_VER == 1 && defined(BUBBLE)
	bbl_path[0] = '\0';
#endif

	/* t@CXLbv */
	p = lpCmdLine;
	flag = FALSE;
	for (;;) {
		/* I`FbN */
		if (*p == '\0') {
			return;
		}

		/* NH[g`FbN */
		if (*p == '"') {
			flag = !flag;
		}

		/* Xy[X`FbN */
		if ((*p == ' ') && !flag) {
			break;
		}

		/*  */
		p++;
	}

	/* VMbN */
	LockVM();
	StopSnd();

	/* R}hCIvVI܂Ń[v */
	while (*p) {
		path[0] = '\0';

		/* Xy[Xǂݔ΂ */
		for (;;) {
			if (*p != ' ') {
				break;
			}

			/*  */
			p++;
		}

		if (*p == '"') {
			/* NH[g`FbN */
			p++;
			q = path;

			/* NH[go܂ő */
			for (;;) {
				*q = *p++;

				/* NH[gI`FbN */
				if (*q == '\0') {
					path[0] = '\0';
					break;
				}

				/* NH[g`FbN */
				if (*q == '"') {
					*q = '\0';
					break;
				}

				/*  */
				q++;
			}
		}
		else if (*p) {
			/* NH[gȂ */
			q = path;

			/* Xy[Xo邩񂪏I܂ő */
			for (;;) {
				*q = *p++;

				/* ؂蕶`FbN */
				if ((*q == '\0') || (*q == ' ')) {
					*q = '\0';
					break;
				}

				/*  */
				q++;
			}
		}

		/* I`FbN */
		if (!path[0]) {
			break;
		}

		/* tpX𐶐 */
		_fullpath(fullpath, path, _MAX_PATH);

		/* gq */
		_splitpath(fullpath, InitDir, dir, NULL, ext);
		strncat(InitDir, dir, sizeof(InitDir) - strlen(InitDir) - 1);

		/* D77/2D/2DD/VFD */
#if XM7_VER >= 3
		if ((stricmp(ext, ".D77") == 0) || (stricmp(ext, ".VFD") == 0) ||
			(stricmp(ext, ".2D") == 0) || (stricmp(ext, ".2DD") == 0)) {
#elif XM7_VER == 1 && defined(SFDC)
		if ((stricmp(ext, ".D77") == 0) || (stricmp(ext, ".VFD") == 0) ||
			(stricmp(ext, ".2D") == 0) || (stricmp(ext, ".2HD") == 0)) {
#else
		if ((stricmp(ext, ".D77") == 0) || (stricmp(ext, ".VFD") == 0) ||
			(stricmp(ext, ".2D") == 0)) {
#endif

#if XM7_VER == 1 && defined(SFDC)
			if ((drive == 0) && (sdrive == 0)) {
				/* 1ڂ̏ꍇApXۑ */
				strncpy(InitialDir[DISK_IMAGE_DIR], InitDir, sizeof(InitialDir[DISK_IMAGE_DIR]));

				/* D771w肵ꍇ̂߂Ƀt@Cۑ */
				if (stricmp(ext, ".D77") == 0) {
					strncpy(d77_path, fullpath, sizeof(d77_path));
				}

				/* 320KB/1MBɃhCu0`3̃}Eg */
				fdc_setdisk(FDC_M, 0, NULL);
				fdc_setdisk(FDC_M, 1, NULL);
				fdc_setdisk(FDC_M, 2, NULL);
				fdc_setdisk(FDC_M, 3, NULL);
				fdc_setdisk(FDC_S, 0, NULL);
				fdc_setdisk(FDC_S, 1, NULL);
				fdc_setdisk(FDC_S, 2, NULL);
				fdc_setdisk(FDC_S, 3, NULL);
			}
#endif	/* XM7_VER == 1 && defined(SFDC) */

			/* 4܂Ń}Eg\ */
#if XM7_VER == 1 && defined(SFDC)
			if (drive < 4 && stricmp(ext, ".2HD") != 0) {
#else	/* XM7_VER == 1 && defined(SFDC) */
			if (drive < 4) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if !(XM7_VER == 1 && defined(SFDC))
				if (drive == 0) {
					/* 1ڂ̏ꍇApXۑ */
					strncpy(InitialDir[DISK_IMAGE_DIR], InitDir,
						sizeof(InitialDir[DISK_IMAGE_DIR]));

					/* D771w肵ꍇ̂߂Ƀt@Cۑ */
					if (stricmp(ext, ".D77") == 0) {
						strncpy(d77_path, fullpath, sizeof(d77_path));
					}
					/* hCu0`3̃}Eg */
					fdc_setdisk(0, NULL);
					fdc_setdisk(1, NULL);
					fdc_setdisk(2, NULL);
					fdc_setdisk(3, NULL);
				}
#endif	/* !(XM7_VER == 1 && defined(SFDC)) */

#if XM7_VER == 1 && defined(SFDC)
				/* fBXN}Eg */
				fdc_setdisk(FDC_M, drive, fullpath);

				/* }EgȂhCuԍ(=}Eg)+1 */
				if (fdc_ready[FDC_M][drive] != FDC_TYPE_NOTREADY) {
					drive ++;
					if ((drive == 1) && !boot &&
						(boot_mode != BOOT_BASIC) &&
						(boot_mode != BOOT_DOS) &&
						(boot_mode != BOOT_BASIC_DBG)) {
						boot_mode = BOOT_BASIC;
					}
					boot = TRUE;
				}
#else	/* XM7_VER == 1 && defined(SFDC) */
				/* fBXN}Eg */
				fdc_setdisk(drive, fullpath);

				/* }EgȂhCuԍ(=}Eg)+1 */
				if (fdc_ready[drive] != FDC_TYPE_NOTREADY) {
					drive ++;
#if XM7_VER == 1
					if ((drive == 1) && !boot &&
						(boot_mode != BOOT_BASIC) &&
						(boot_mode != BOOT_DOS) &&
						(boot_mode != BOOT_BASIC_DBG)) {
						boot_mode = BOOT_BASIC;
					}
					boot = TRUE;
#endif	/* XM7_VER == 1 */
				}
#endif	/* XM7_VER == 1 && defined(SFDC) */
			}

#if XM7_VER == 1 && defined(SFDC)
			/* 4܂Ń}Eg\ */
			if (sdrive < 4 && fdc_enable[FDC_S] &&
				stricmp(ext, ".VFD") != 0 && stricmp(ext, ".2D") != 0) {
				/* fBXN}Eg */
				fdc_setdisk(FDC_S, sdrive, fullpath);

				/* }EgȂhCuԍ(=}Eg)+1 */
				if (fdc_ready[FDC_S][sdrive] != FDC_TYPE_NOTREADY) {
					sdrive ++;
					if ((sdrive == 1) && !boot && (availableSFDBOOT)) {
						boot_mode = BOOT_DOS_SFD;
						boot = TRUE;
					}
				}
			}
#endif	/* XM7_VER == 1 && defined(SFDC) */
		}

		/* T77 */
		if (stricmp(ext, ".T77") == 0) {
			/* LȂ̂1̂ */
			if (!tape_set) {
				/* pXۑ */
				strncpy(InitialDir[TAPE_IMAGE_DIR], InitDir,
					sizeof(InitialDir[TAPE_IMAGE_DIR]));

				/* e[v}Eg */
				tape_setfile(fullpath);

				/* }EgȂe[v}Eg֎~tO𗧂Ă */
				if (tape_fileh != -1) {
					tape_set = TRUE;
				}
			}
		}

#if XM7_VER == 1 && defined(BUBBLE)
		/* B77/BBL */
		if (((stricmp(ext, ".B77") == 0) || (stricmp(ext, ".BBL") == 0)) &&
			(bmc_enable[BMC_32] || bmc_enable[BMC_128])) {

			if ((unit1 == 0) && (unit2 == 0)) {
				/* 1ڂ̏ꍇApXۑ */
				strncpy(InitialDir[BUBBLE_IMAGE_DIR], InitDir,
					sizeof(InitialDir[BUBBLE_IMAGE_DIR]));

				/* 1w肵ꍇ̂߂Ƀt@Cۑ */
				if (stricmp(ext, ".BBL") == 0) {
					strncpy(bbl_path, fullpath, sizeof(bbl_path));
				}

				/* 32KB/128KBɃjbg1̃}Eg */
				bmc_setfile(BMC_32, 0, NULL);
				bmc_setfile(BMC_32, 1, NULL);
				bmc_setfile(BMC_128, 0, NULL);
				bmc_setfile(BMC_128, 1, NULL);
			}

			/* 2܂Ń}Eg\ */
			if ((unit1 < 2) && bmc_enable[BMC_32]) {
				/* ouJZbg}Eg */
				bmc_setfile(BMC_32, unit1, fullpath);

				/* }EgȂ烆jbgԍ(=}Eg)+1 */
				if (bmc_ready[BMC_32][unit1] != BMC_TYPE_NOTREADY) {
					unit1 ++;
					if ((unit1 == 1) && !boot && (availableBBLBOOT)) {
						if (boot_mode != BOOT_BUBBLE) {
							boot_mode = BOOT_BUBBLE;
							mainmem_bubblemode_boot();
						}
						boot = TRUE;
					}
				}
			}

			/* 2܂Ń}Eg\ */
			if ((unit2 < 2) && bmc_enable[BMC_128]) {
				/* ouJZbg}Eg */
				bmc_setfile(BMC_128, unit2, fullpath);

				if (bmc_ready[BMC_128][unit2] != BMC_TYPE_NOTREADY) {
					unit2 ++;
					if ((unit2 == 1) && !boot && (availableBBLBOOT128)) {
						if (boot_mode != BOOT_BUBBLE128) {
							boot_mode = BOOT_BUBBLE128;
							mainmem_bubblemode_boot();
						}
						boot = TRUE;
					}
				}
			}
		}
#endif

		/* XM7 */
		if (stricmp(ext, ".XM7") == 0) {
			strncpy(InitialDir[STATE_FILE_DIR], InitDir,
				sizeof(InitialDir[STATE_FILE_DIR]));
			StateLoad(fullpath);
			StoreCfg();

			/* Xe[g[hɃ}Egŝŏł؂ */
			drive = 0;
#if XM7_VER == 1
 #ifdef SFDC
			sdrive = 0;
 #endif
			boot = FALSE;
#elif XM7_VER == 1 && defined(BUBBLE)
			unit = 0;
#endif
			break;
		}
	}

	/* fBXNC[Wt@Cw萔1̏ꍇ̓ꏈ */
	if (drive == 1) {
#if XM7_VER == 1 && defined(SFDC)
		/* D77̏ꍇA2ڂ΃hCu1Ƀ}Eg */
		if (d77_path[0]) {
			if ((fdc_ready[FDC_M][0] != FDC_TYPE_NOTREADY) && (fdc_medias[FDC_M][0] >= 2)) {
				fdc_setdisk(FDC_M, 1, d77_path);
				fdc_setmedia(FDC_M, 1, 1);
			}
		}
#else	/* XM7_VER == 1 && defined(SFDC) */

		/* D77̏ꍇA2ڂ΃hCu1Ƀ}Eg */
		if (d77_path[0]) {
			if ((fdc_ready[0] != FDC_TYPE_NOTREADY) && (fdc_medias[0] >= 2)) {
				fdc_setdisk(1, d77_path);
				fdc_setmedia(1, 1);
			}
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}
#if XM7_VER == 1 && defined(SFDC)
	if (sdrive == 1) {
		/* D77̏ꍇA2ڂ΃hCu1Ƀ}Eg */
		if (d77_path[0]) {
			if ((fdc_ready[FDC_S][0] != FDC_TYPE_NOTREADY) && (fdc_medias[FDC_S][0] >= 2)) {
				fdc_setdisk(FDC_S, 1, d77_path);
				fdc_setmedia(FDC_S, 1, 1);
			}
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

#if XM7_VER == 1 && defined(BUBBLE)
	/* ouC[Wt@Cw萔1̏ꍇ̓ꏈ */
	if (unit1 == 1) {
		/* B77̏ꍇA2{ڂ΃hCu1Ƀ}Eg */
		if (bbl_path[0]) {
			if ((bmc_ready[BMC_32][0] != BMC_TYPE_NOTREADY) && (bmc_medias[BMC_32][0] >= 2)) {
				bmc_setfile(BMC_32, 1, bbl_path);
				bmc_setmedia(BMC_32, 1, 1);
			}
		}
	}
	if (unit2 == 1) {
		/* B77̏ꍇA2{ڂ΃hCu1Ƀ}Eg */
		if (bbl_path[0]) {
			if ((bmc_ready[BMC_128][0] != BMC_TYPE_NOTREADY) && (bmc_medias[BMC_128][0] >= 2)) {
				bmc_setfile(BMC_128, 1, bbl_path);
				bmc_setmedia(BMC_128, 1, 1);
			}
		}
	}
#endif

#if XM7_VER == 1
	/* u[g\ȃC[W}EgꂽꍇAVMZbg */
	if (boot) {
#else	/* XM7_VER == 1 */
	/* fBXNC[W}EgꂽꍇAVMZbg */
	if (drive >= 1) {
#endif	/* XM7_VER == 1 */
		system_reset();
		StatePath[0] = '\0';
		OnRefresh(hMainWnd);
	}

	SetAllMRUState();

	/* VMAbN */
	PlaySnd();
	ResetSch();
	UnlockVM();

	/* \eXV */
	if (hDrawWnd) {
		InvalidateRect(hDrawWnd, NULL, FALSE);
	}
}
#endif	/* _WIN32 */
