/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2023 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2023 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2023 Toma
 *
 *	[ Win32API TuEBhEP ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <mmsystem.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include "xm7.h"
#include "event.h"
#include "jsubsys.h"
#include "keyboard.h"
#include "mmr.h"
#include "mouse.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_sch.h"
#include "w32_sub.h"
#include "w32_kbd.h"

/*
 *	O[o [N
 */
HWND hSubWnd[SWND_MAXNUM];				/* TuEChE */
BOOL bShowSubWindow[SWND_MAXNUM];		/* TuEBhE\ */
BOOL bPopupSwnd;						/* TuEBhE|bvAbv */
BOOL bSwndVisible;						/* TuEBhE\tO */
BOOL bResumeSwnd;						/* TuEBhEԂۑ */
BOOL bResumeBootSwnd;					/* NɃTuEBhEԂ𕜋A */
int nSwndFlag[SWND_MAXNUM];				/* TuEBhE */
RECT stSwndRect[SWND_MAXNUM];			/* TuEBhEʒu */
WORD wDisAsm[MAXCPU];					/* tAZu AhX */
int nHeightDisAsm[MAXCPU];				/* tAZu EBhEcTCY */
int nAddrSegDisAsm;						/* tAZuZOg */
int nAddrModeDisAsm;					/* tAZuAhX[h */
#if XM7_VER == 1 && defined(Z80CARD)
int nAddrBnkDisAsmZ80;					/* tAZuZ80oN */
int nAddrModeDisAsmZ80;					/* tAZuZ80AhX[h */
#endif
WORD wMemory[MAXCPU];					/* _v AhX */
int nHeightMemory[MAXCPU];				/* _v EBhEcTCY */
BOOL bAsciiDump[MAXCPU];				/* _v ASCII\ */
BOOL bKanjiDump[MAXCPU];				/* _v \ */
int nAddrSegDump;						/* _vZOg */
int nAddrModeDump;						/* _vAhX[h */
LONG lLogHeight;						/* O EBhEcTCY */
LONG lLogWidth;							/* O EBhETCY */

/*
 *	X^eBbN [N
 */
static WORD AddrHistory[16];			/* AhXqXg */
static WORD AddrBuf;					/* AhXobt@ */
static int AddrNum;						/* AhXqXg */
static BYTE nBreakCPU;					/* u[N|Cg ݒΏCPU */
static BYTE *pBreakPoint;				/* u[N|Cg Drawobt@ */
static HMENU hBreakPoint;				/* u[N|Cg j[nh */
static POINT PosBreakPoint;				/* u[N|Cg }EXʒu */
static BYTE *pScheduler;				/* XPW[ Drawobt@ */
static DWORD dwSchedulerTime[2];		/* XPW[  */
static DWORD dwSchedulerFrame[2];		/* XPW[ t[ */
static BYTE *pCPURegister[MAXCPU];		/* CPUWX^ Drawobt@ */
static HMENU hCPURegister[MAXCPU];		/* CPUWX^ j[nh */
static BYTE *pDisAsm[MAXCPU];			/* tAZu Drawobt@ */
static HMENU hDisAsm[MAXCPU];			/* tAZu j[nh */
static POINT PosDisAsmPoint;			/* tAZu }EXʒu */
static BYTE *pMemory[MAXCPU];			/* _v Drawobt@ */
static HMENU hMemory[MAXCPU];			/* _v j[nh */
static BYTE nMemDlgCPU;					/* ύX ύXΏCPU */
static WORD wMemDlgAddr;				/* ύX AhX */
static WORD nMemDlgByte;				/* ύX AhXe */
static BYTE nRegDlgCPU;					/* WX^ύX ύXΏCPU */
static BYTE nRegDlgNo;					/* WX^ύX ύXΏۃWX^ */
static BYTE nMemSrchCPU;				/*  ΏCPU */
static WORD wMemSrchAddr;				/*  JnAhX */
static WORD wMemSrchAddrSave[MAXCPU];	/*  JnAhX main */
static char szMemSrchString[128];		/*  Ώۃf[^ */
static WORD wStackAddr[16];				/* VXeX^bNWvAhX */
static WORD wBreakAddr[16];				/* u[N|CgWvAhX */
static WORD wVectorAddr[16];			/* xN^[WvAhX */
static char szDisAsmWndName[128];		/* tAZu LvV */
static char szMemoryWndName[128];		/* _v LvV */
#if XM7_VER == 1 && defined(Z80CARD)
static char szDisAsmWndNameZ80[128];	/* tAZuZ80 LvV */
#endif

/*
 *	Oe[u
 */
static const struct HEADERTABLE {
	UINT id;
	int fmt;
	int cx;
} headertable[] = {
	{IDS_SWND_LOG_NUMBER,	LVCFMT_RIGHT,	7},		/* J0(ʂԍ) */
	{IDS_SWND_LOG_RTIME,	LVCFMT_RIGHT,	17},	/* J1() */
	{IDS_SWND_LOG_VTIME,	LVCFMT_RIGHT,	10},	/* J2(z) */
	{IDS_SWND_LOG_PC,		LVCFMT_CENTER,	13},	/* J3(PC) */
	{IDS_SWND_LOG_LEVEL,	LVCFMT_CENTER,	6},		/* J4(x) */
	{IDS_SWND_LOG_DEVICE,	LVCFMT_CENTER,	7},		/* J5(foCX) */
	{IDS_SWND_LOG_MSG,		LVCFMT_LEFT,	42},	/* J6(bZ[W) */
};

/*
 *	OǗ\
 */
typedef struct {
	DWORD number;						/* ʂԍ(ZbgŃNA) */
	DWORD time;							/* z */
	DWORD id;							/* foCXID */
	DWORD pc1;							/* vOJE^(MAIN) */
	DWORD pc2;							/* vOJE^(DSUB) */
#ifdef JSUB
	DWORD pc3;							/* vOJE^(JSUB) */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	DWORD pc4;							/* vOJE^(Z80) */
#endif
	loglevel level;						/* x */
	char *string;						/*  */
	time_t timer;						/*  */
} logdata_t;

enum {
	LogMax = 0x4000						/* OőL^(2̔{ł邱) */
};

/*
 *	OǗX^eBbN [N
 */
static HWND hLogWnd;
static int nLogTop = 0;					/* 擪|C^(łÂ) */
static int nLogNum = 0;					/* OL^ */
static int nLogCnt = 1;					/* OJEg */
static int nLogDsp = 0;					/* O\Jnԍ */
static logdata_t *stLogData[LogMax]		/* O|C^ */
								= {0};

/*
 *	vg^Cv錾
 */
static void FASTCALL PaintMemory(HWND hWnd);

/*
 *	VtgJIS}N+ӂ
 */
#define isKanji(p)	((BYTE)((p ^ 0x20) - 0xa1) <= 0x3b)
#define isKanji2(p)	(((BYTE)(p + 3) >= 0x43) && (p != 0x7f))
#ifdef RES_DLL
#define isJapanese	(!bLanguageMode)
#else
#define isJapanese	(PRIMARYLANGID(GetUserDefaultLangID()) == LANG_JAPANESE)
#endif

#if XM7_VER >= 3
#define	IDM_DIS_MAX	(IDM_DIS_SEG00 + 0x07)
#define	IDM_DMP_MAX	(IDM_DMP_SEG00 + 0x07)
#elif XM7_VER >= 2
#define	IDM_DIS_MAX	(IDM_DIS_SEG00 + 0x03)
#define	IDM_DMP_MAX	(IDM_DMP_SEG00 + 0x03)
#else
#define	IDM_DIS_MAX	(IDM_DIS_SEG00 + 0x3f)
#define	IDM_DMP_MAX	(IDM_DMP_SEG00 + 0x3f)
#endif

/*
 *	16Fpbge[u
 */
const RGBQUAD rgbQuadTable16[] = {
	/*  B     G     R   Reserve */
	{ 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0xff, 0x00 },
	{ 0xff, 0x00, 0xff, 0x00 },
	{ 0x00, 0xff, 0x00, 0x00 },
	{ 0xff, 0xff, 0x00, 0x00 },
	{ 0x00, 0xff, 0xff, 0x00 },
	{ 0xff, 0xff, 0xff, 0x00 },

	{ 0x3f, 0x3f, 0x3f, 0x00 },	/* ÊD */
	{ 0xff, 0xbf, 0x00, 0x00 },	/* F */
	{ 0x00, 0xdf, 0xff, 0x00 },	/* F */
	{ 0x00, 0xaf, 0x7f, 0x00 },	/*  */
	{ 0xbf, 0x9f, 0xff, 0x00 },	/* Î */
	{ 0x9f, 0xff, 0x3f, 0x00 },	/* GhO[() */
	{ 0x7f, 0xbf, 0x00, 0x00 }, /* GhO[() */
	{ 0xcf, 0xcf, 0xcf, 0x00 },	/* D */
};

/*
 *	TuEChE쐬֐e[u
 */
static const CREATESWNDFUNC	CreateSwndFunc[SWND_MAXNUM] = {
	CreateBreakPoint,					/* u[N|Cg */
	CreateScheduler,					/* XPW[ */
	CreateCPURegisterMain,				/* CPUWX^ C */
	CreateCPURegisterSub,				/* CPUWX^ Tu */
#ifdef JSUB
	CreateCPURegisterJSub,				/* CPUWX^ {Tu */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	CreateCPURegisterZ80,				/* CPUWX^ C(Z80) */
#endif
	CreateDisAsmMain,					/* tAZu C */
	CreateDisAsmSub,					/* tAZu Tu */
#ifdef JSUB
	CreateDisAsmJSub,					/* tAZu {Tu */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	CreateDisAsmZ80,					/* tAZu C(Z80) */
#endif
	CreateMemoryMain,					/* _v C */
	CreateMemorySub,					/* _v Tu */
#ifdef JSUB
	CreateMemoryJSub,					/* _v {Tu */
#endif
	CreateFDC,							/* FDC */
	CreateOPNReg,						/* OPNWX^ */
	CreateOPNDisp,						/* OPNfBXvC */
	CreateSubCtrl,						/* TuCPURg[ */
#if XM7_VER >= 2
	CreateALULine,						/* _Z/ */
#endif
	CreateKeyboard,						/* L[{[h */
	CreateMMR,							/* MMR */
	CreatePaletteReg,					/* pbgWX^ */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
	CreateDMAC,							/* DMAC */
#endif
	CreateIRQMain,						/* IRQ C */
	CreateIRQSub,						/* IRQ Tu */
#if XM7_VER == 1 && defined(BUBBLE)
	CreateBMC,							/* ouRg[ */
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
	CreateKanji,						/* ROM */
	CreateTapeLP,						/* JZbge[vv^ */
	CreateLog,							/* O */
};

/*-[ ėpEBhE ]----------------------------------------------------*/

/*
 *	TuEBhE
 *	TCYύX
 */
static void FASTCALL WindowSizing(HWND hWnd, LPRECT rect, BYTE **buf)
{
	RECT wrect;
	POINT tpoint;
	POINT bpoint;
	int tframe;
	int bframe;
	int x;
	int y;

	/* t[E^Cgo[̃TCY擾 */
	bframe = GetSystemMetrics(SM_CYSIZEFRAME);
	tframe = GetSystemMetrics(SM_CYCAPTION) + bframe;

	/* VNCAg̈vZ */
	tpoint.x = 0;
	tpoint.y = rect->top + tframe;
	bpoint.x = 0;
	bpoint.y = rect->bottom - bframe;
	ScreenToClient(hWnd, &tpoint);
	ScreenToClient(hWnd, &bpoint);

	/* NCAg̈𕶎̔{ɍ킹 */
	tpoint.y = (tpoint.y / lCharHeight) * lCharHeight;
	bpoint.y = (bpoint.y / lCharHeight) * lCharHeight;

	/* XN[W֕ϊ */
	ClientToScreen(hWnd, &tpoint);
	ClientToScreen(hWnd, &bpoint);

	/* EBhETCY𒲐 */
	GetWindowRect(hWnd, &wrect);
	rect->left = wrect.left;
	rect->top = tpoint.y;
	rect->right = wrect.right;
	rect->bottom = bpoint.y;
	x = rect->right / lCharWidth;
	y = rect->bottom / lCharHeight;

	/* ĉݒ(AdjustWindowRectɌ̒l) */
	if (bPopupSwnd) {
		AdjustWindowRect(rect,	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU |
								WS_CAPTION | WS_VISIBLE |
								WS_MINIMIZEBOX | WS_SIZEBOX, FALSE);
	}
	else {
		AdjustWindowRect(rect,	WS_CHILD | WS_OVERLAPPED | WS_SYSMENU |
								WS_CAPTION | WS_VISIBLE |
								WS_MINIMIZEBOX | WS_CLIPSIBLINGS |
								WS_SIZEBOX, FALSE);
	}
	rect->left = wrect.left;
	rect->right = wrect.right;

	/* obt@Ď擾A0xFFŖ߂ */
	free(*buf);
	*buf = malloc(2 * x * y);
	ASSERT(*buf);
	memset(*buf, 0xff, 2 * x * y);
}

/*
 *	TuEBhE
 *	eLXg`
 */
void FASTCALL DrawWindowText(HDC hDC, BYTE *ptr, int x, int y)
{
	BYTE *p;
	BYTE *q;
	int xx, yy;
	HFONT hBackup;

	/* tHgZNg */
	hBackup = SelectObject(hDC, hFont);

	/* r` */
	p = ptr;
	q = &p[x * y];
	for (yy=0; yy<y; yy++) {
		for (xx=0; xx<x; xx++) {
			if (ptr) {
				if (*p != *q) {
					TextOut(hDC, xx * lCharWidth, yy * lCharHeight,
						(LPCTSTR)p, 1);
					*q = *p;
				}
				p++;
				q++;
			}
		}
	}

	/* I */
	SelectObject(hDC, hBackup);
}

/*
 *	TuEBhE
 *	eLXg` (\Ή)
 */
void FASTCALL DrawWindowTextKanji(HDC hDC, BYTE *ptr, int x, int y)
{
	BYTE *p;
	BYTE *q;
	int xx, yy;
	HFONT hBackup;

	/* tHgZNg */
	hBackup = SelectObject(hDC, hFont);

	/* r` */
	for (yy=0; yy<y; yy++) {
		for (xx=0; xx<x; xx++) {
			p = &ptr[x * yy + xx];
			q = &p[x * y];
			if (ptr) {
				if (isKanji(*p)) {
					if (*(WORD *)p != *(WORD *)q) {
						TextOut(hDC, xx * lCharWidth, yy * lCharHeight,
							(LPCTSTR)p, 2);
						*(WORD *)q = *(WORD *)p;
					}
					xx++;
				}
				else {
					if (*p != *q) {
						TextOut(hDC, xx * lCharWidth, yy * lCharHeight,
							(LPCTSTR)p, 1);
						*q = *p;
					}
				}
			}
		}
	}

	/* I */
	SelectObject(hDC, hBackup);
}

/*
 *	TuEBhE
 *	eLXg`(]t)
 */
void FASTCALL DrawWindowText2(HDC hDC, BYTE *ptr, int x, int y)
{
	BYTE *p;
	BYTE *q;
	int xx, yy;
	HFONT hBackup;
	COLORREF tcolor, bcolor;
	char dat;

	/* tHgZNg */
	hBackup = SelectObject(hDC, hFont);

	/* r` */
	p = ptr;
	q = &p[x * y];
	for (yy=0; yy<y; yy++) {
		for (xx=0; xx<x; xx++) {
			if (ptr) {
				if (*p != *q) {
					if (*p & 0x80) {
						/* ]\ */
						dat = (char)(*p & 0x7f);
						bcolor = GetBkColor(hDC);
						tcolor = GetTextColor(hDC);
						SetTextColor(hDC, bcolor);
						SetBkColor(hDC, tcolor);
						TextOut(hDC, xx * lCharWidth, yy * lCharHeight, &dat, 1);
						SetTextColor(hDC, tcolor);
						SetBkColor(hDC, bcolor);
					}
					else {
						/* ʏ\ */
						TextOut(hDC, xx * lCharWidth, yy * lCharHeight,
							(LPCTSTR)p, 1);
					}
					*q = *p;
				}
				p++;
				q++;
			}
		}
	}

	/* I */
	SelectObject(hDC, hBackup);
}

/*
 *	TuEChE
 *	WM_DESTROY
 */
void FASTCALL DestroySubWindow(HWND hWnd, BYTE **pBuf, HMENU hmenu)
{
	int i;
	WINDOWPLACEMENT wndpl;
	RECT rect;

	GetWindowRect(hDrawWnd, &rect);

	/* CEChE֎ʒm */
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			hSubWnd[i] = NULL;

			wndpl.length = sizeof(WINDOWPLACEMENT);
			GetWindowPlacement(hWnd, &wndpl);
			if ((wndpl.showCmd != SW_SHOWMINIMIZED) && bResumeSwnd) {
				/* |bvAbvTuEBhEҔ */
				GetWindowRect(hWnd, &stSwndRect[i]);
				if (!bPopupSwnd) {
					stSwndRect[i].left -= rect.left;
					stSwndRect[i].top -= rect.top;
				}
			}
			nSwndFlag[i] = 0;

			/* j[폜 */
			if (hmenu) {
				DestroyMenu(hmenu);
			}

			/* Drawobt@폜 */
			if (pBuf && *pBuf) {
				free(*pBuf);
				*pBuf = NULL;
			}
		}
	}

	/* h[EChEĕ` */
	InvalidateRect(hDrawWnd, NULL, FALSE);
}

/*
 *	TuEBhE
 *	ʒuZo
 */
void FASTCALL PositioningSubWindow(HWND hParent, LPRECT rect, int index)
{
	RECT wrect;

	/* TuEBhEʒuZo */
	if (bResumeSwnd && (nSwndFlag[index] >= 0)) {
		*rect = stSwndRect[index];
		while (rect->left > GetSystemMetrics(SM_CXSCREEN) - 80) {
			rect->left -= 40;
		}
		while (rect->top > GetSystemMetrics(SM_CYSCREEN) - 20) {
			rect->top -= 40;
		}
	}
	else {
		rect->left = lCharWidth * index;
		rect->top = lCharHeight * index;
		rect->right = 0;
		rect->bottom = 0;
		if (rect->top >= 380) {
			rect->top -= 380;
		}
		if (bPopupSwnd) {
			GetWindowRect(hParent, (LPRECT)&wrect);
			rect->left += wrect.left;
			rect->top += wrect.top;
		}
	}
	if (bResumeSwnd) {
		nSwndFlag[index] = 1;
	}
}

/*
 *	TuEBhE
 *	
 */
void FASTCALL RestoreSubWindow(void)
{
	int i;
	WORD wAddr;

	if (!bResumeSwnd) {
		return;
	}

	/* Misc */
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i]) {
			continue;
		}
		if (nSwndFlag[i] > 0) {
			/* EChE쐬 */
			switch (i) {
			case SWND_DISASM_MAIN:		/* tAZu C */
			case SWND_DISASM_SUB:		/* tAZu Tu */
#ifdef JSUB
			case SWND_DISASM_JSUB:		/* tAZu {Tu */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
			case SWND_DISASM_Z80:		/* tAZu Z80 */
#endif
				wAddr = wDisAsm[i - SWND_DISASM_MAIN];
				break;
			case SWND_MEMORY_MAIN:		/* _v C */
			case SWND_MEMORY_SUB:		/* _v Tu */
#ifdef JSUB
			case SWND_MEMORY_JSUB:		/* _v {Tu */
#endif
				wAddr = wMemory[i - SWND_MEMORY_MAIN];
				break;
			}
			hSubWnd[i] = (CreateSwndFunc[i])(hDrawWnd, i);
			switch (i) {
			case SWND_DISASM_MAIN:		/* tAZu C */
			case SWND_DISASM_SUB:		/* tAZu Tu */
#ifdef JSUB
			case SWND_DISASM_JSUB:		/* tAZu {Tu */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
			case SWND_DISASM_Z80:		/* tAZu Z80 */
#endif
				AddrDisAsm(i - SWND_DISASM_MAIN, wAddr);
				break;
			case SWND_MEMORY_MAIN:		/* _v C */
			case SWND_MEMORY_SUB:		/* _v Tu */
#ifdef JSUB
			case SWND_MEMORY_JSUB:		/* _v {Tu */
#endif
				AddrMemory(i - SWND_MEMORY_MAIN, wAddr);
				break;
			}
		}
	}
	if (hMainWnd) {
		SetForegroundWindow(hMainWnd);
	}
}

/*
 *	TuEChE
 *	WM_SIZE
 */
void FASTCALL MoveSubWindow(HWND hWnd)
{
	int i;
	RECT rect;

	/* |bvAbvEۑ */
	GetWindowRect(hDrawWnd, &rect);
	if (bResumeSwnd &&
		(!(GetWindowLong(hWnd, GWL_STYLE) & (WS_MAXIMIZE | WS_MINIMIZE)))) {

		/* CEChE֎ʒm */
		for (i=0; i<SWND_MAXNUM; i++) {
			if (hSubWnd[i] == hWnd) {
				/* |bvAbvTuEBhEҔ */
				GetWindowRect(hWnd, &stSwndRect[i]);
				if (!bPopupSwnd) {
					stSwndRect[i].left -= rect.left;
					stSwndRect[i].top -= rect.top;
				}
			}
		}
	}
}

/*
 *	TuEChE
 *	̋󔒂폜
 */
static void FASTCALL rtrim(char *pstring)
{
	int n;

	/* NULL`FbN */
	if (pstring == NULL) {
		return;
	}
	n = strlen(pstring);
	while ((--n >= 0) && (pstring[n] == ' '));
	pstring[n + 1] = '\0';
}

/*
 *	TuEChE
 *	Nbv{[ho^
 */
void FASTCALL SetClipboardSubWindow(HWND hWnd, BYTE *pDraw, int nwidth, int nheight, BOOL mode)
{
	HGLOBAL hg;
	PTSTR strMem;
	int i, j;
	BYTE *p;

	if (!OpenClipboard(hWnd)) {
		return;
	}
	EmptyClipboard();
	hg = GlobalAlloc(GHND, (size_t)(2 * (nwidth + 2) * nheight));
	strMem = (PTSTR)GlobalLock(hg);

	p = malloc(nwidth + 3);
	if (p) {
		for (i = 0; i < nheight; i ++) {
			memset(p, 0, nwidth + 3);
			strncpy((char *)p, (char *)(pDraw + nwidth * i), nwidth);
			if (mode) {
				for (j = 0; j < (int)strlen((char *)p); j++) {
					p[j] &= (BYTE)~0x80;
				}
			}
			rtrim((char *)p);
			strncat((char *)p, "\n", nwidth + 2);
			lstrcat(strMem, (LPCTSTR)p);
		}
		free(p);
	}

	GlobalUnlock(hg);
	SetClipboardData(CF_TEXT, hg);
	CloseClipboard();
}

/*-[ EBhE[N ]------------------------------------------------*/

/*
 *	TuEBhE
 *	[N
 */
void FASTCALL InitSubWndWork(BOOL mode)
{
	int i;

	for (i=0; i<MAXCPU; i++) {
		nHeightDisAsm[i] = 8;			/* tAZu EBhEcTCY */
#if XM7_VER == 1 && defined(Z80CARD)
		if (i == MAINZ80) {
			continue;
		}
#endif
		nHeightMemory[i] = 8;			/* _v EBhEcTCY */
		bAsciiDump[i] = FALSE;			/* _v ASCII\ */
		bKanjiDump[i] = FALSE;			/* _v \ */
		wMemSrchAddrSave[i] = 0;		/*  JnAhX */
	}

	nAddrSegDisAsm = 0;					/* tAZuZOg */
	nAddrModeDisAsm = 0;				/* tAZuAhX[h */
#if XM7_VER == 1 && defined(Z80CARD)
	nAddrBnkDisAsmZ80 = 0;				/* tAZuZ80oN */
	nAddrModeDisAsmZ80 = 0;				/* tAZuZ80AhX[h */
#endif
	nAddrSegDump = 0;					/* _vZOg */
	nAddrModeDump = 0;					/* _vAhX[h */
#if XM7_VER == 1 && defined(BUBBLE)
	bBMC = FALSE;						/* BMC \[h */
#endif

	if (mode) {
		/* N݂̂̏ */
		bPopupSwnd = TRUE;
	}

	memset(szDisAsmWndName, 0, sizeof(szDisAsmWndName));	/* tAZu LvV */
	memset(szMemoryWndName, 0, sizeof(szMemoryWndName));	/* _v LvV */
#if XM7_VER == 1 && defined(Z80CARD)
	memset(szDisAsmWndNameZ80, 0, sizeof(szDisAsmWndNameZ80));	/* tAZuZ80 LvV */
#endif

	/* TuEBhE */
	for (i=0; i<SWND_MAXNUM; i++) {
		nSwndFlag[i] = -1;
		memset(&stSwndRect[i], 0, sizeof(RECT));
	}
}

/*-[ AhXqXg ]------------------------------------------------------*/

/*
 *	AhXqXg
 *	R{{bNXւ̑}
 */
static void FASTCALL InsertHistory(HWND hWnd)
{
	int i;
	char string[128];

	/* qXg} */
	(void)ComboBox_ResetContent(hWnd);
	for (i=AddrNum; i>0; i--) {
		_snprintf(string, sizeof(string), "%04X", AddrHistory[i - 1]);
		(void)ComboBox_AddString(hWnd, string);
	}
}

/*
 *	AhXqXg
 *	ǉ
 */
static void FASTCALL AddAddrHistory(WORD addr)
{
	int i;

	/* do^`FbN */
	for (i=0; i<AddrNum; i++) {
		if (AddrHistory[i] == addr) {
			return;
		}
	}

	/* qXgVtgA}AJEgAbv */
	for (i=14; i>=0; i--) {
		AddrHistory[i + 1] = AddrHistory[i];
	}
	AddrHistory[0] = addr;
	if (AddrNum < 16) {
		AddrNum++;
	}
}

/*-[ AhX̓_CAO ]------------------------------------------------*/

/*
 *	AhX̓_CAO
 *	_CAO
 */
static BOOL FASTCALL AddrDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	char string[128];

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_ADDRCOMBO);
	ASSERT(hWnd);
	InsertHistory(hWnd);

	/* AhXݒ */
	_snprintf(string, sizeof(string), "%04X", AddrBuf);
	ComboBox_SetText(hWnd, string);

	return TRUE;
}

/*
 *	AhX̓_CAO
 *	_CAOOK
 */
static void FASTCALL AddrDlgOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_ADDRCOMBO);
	ASSERT(hWnd);

	/* ݂̒l擾 */
	ComboBox_GetText(hWnd, string, sizeof(string) - 1);
	AddrBuf = (WORD)strtol(string, NULL, 16);

	/* qXgǉ */
	AddAddrHistory(AddrBuf);
}

/*
 *	AhX̓_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK AddrDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return AddrDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						AddrDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*
 *	AhX
 */
static BOOL FASTCALL AddrDlg(HWND hWnd, WORD *pAddr)
{
	int ret;

	ASSERT(hWnd);
	ASSERT(pAddr);

	/* AhXZbg */
	AddrBuf = *pAddr;

	/* [__CAOs */
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_ADDRDLG), hWnd, AddrDlgProc);
	if (ret != IDOK) {
		return FALSE;
	}

	/* AhXZbgAA */
	*pAddr = AddrBuf;
	return TRUE;
}

/*-[ u[N|Cgݒ_CAO ]----------------------------------------*/

/*
 *	u[N|Cgݒ_CAO
 *	_CAO
 */
static BOOL FASTCALL BreakPointDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	char string[128];

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_BREAKP_ADDRCOMBO);
	ASSERT(hWnd);
	InsertHistory(hWnd);

	/* AhXݒ */
	_snprintf(string, sizeof(string), "%04X", AddrBuf);
	ComboBox_SetText(hWnd, string);

	/* ΏCPUݒ */
	switch (nBreakCPU) {
		case MAINCPU :	CheckDlgButton(hDlg, IDC_BREAKP_CPU_MAIN, BST_CHECKED);
						break;
		case SUBCPU :	CheckDlgButton(hDlg, IDC_BREAKP_CPU_SUB, BST_CHECKED);
						break;
#ifdef JSUB
		case JSUBCPU :	CheckDlgButton(hDlg, IDC_BREAKP_CPU_JSUB, BST_CHECKED);
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	CheckDlgButton(hDlg, IDC_BREAKP_CPU_MAINZ80, BST_CHECKED);
						break;
#endif
#endif
	}

	return TRUE;
}

/*
 *	u[N|Cgݒ_CAO
 *	_CAOOK
 */
static void FASTCALL BreakPointDlgOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_BREAKP_ADDRCOMBO);
	ASSERT(hWnd);

	/* ΏCPU擾 */
	if (IsDlgButtonChecked(hDlg, IDC_BREAKP_CPU_MAIN) == BST_CHECKED) {
		nBreakCPU = MAINCPU;
	}
#ifdef JSUB
	else if (IsDlgButtonChecked(hDlg, IDC_BREAKP_CPU_JSUB) == BST_CHECKED) {
		nBreakCPU = JSUBCPU;
	}
#endif
#if XM7_VER == 1
#ifdef Z80CARD
	else if (IsDlgButtonChecked(hDlg, IDC_BREAKP_CPU_MAINZ80) == BST_CHECKED) {
		nBreakCPU = MAINZ80;
	}
#endif
#endif
	else {
		nBreakCPU = SUBCPU;
	}

	/* ݂̒l擾 */
	ComboBox_GetText(hWnd, string, sizeof(string) - 1);
	AddrBuf = (WORD)strtol(string, NULL, 16);

	/* qXgǉ */
	AddAddrHistory(AddrBuf);
}

/*
 *	u[N|Cgݒ_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK BreakPointDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return BreakPointDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						BreakPointDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*
 *	u[N|Cgݒ
 */
static BOOL FASTCALL BreakPointDlg(HWND hWnd, int num)
{
	int ret;

	ASSERT(hWnd);

	/* AhXZbg */
	if (breakp[num].flag != BREAKP_NOTUSE) {
		AddrBuf = breakp[num].addr;
		nBreakCPU = (BYTE)breakp[num].cpu;
	}
	else {
		AddrBuf = 0;
		nBreakCPU = MAINCPU;
	}

	/* [__CAOs */
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_BREAKPDLG), hWnd, BreakPointDlgProc);
	if (ret != IDOK) {
		return FALSE;
	}

	/* AhXZbgAA */
	schedule_setbreak2(num, nBreakCPU, AddrBuf);
	return TRUE;
}

/*-[ u[N|Cg ]------------------------------------------------------*/

/*
 *	u[N|CgEChE
 *	ZbgAbv
 */
static void FASTCALL SetupBreakPoint(BYTE *p, int x, int y)
{
	int i;
	char string[128];
	char tmp[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* VMbN */
	LockVM();

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* u[N|Cg[v */
	for (i=0; i<BREAKP_MAXNUM; i++) {
		/* 쐬 */
		string[0] = '\0';
		if (breakp[i].flag == BREAKP_NOTUSE) {
			_snprintf(string, sizeof(string), "%2d ------------------", i + 1);
		}
		else {
			switch (breakp[i].cpu) {
				case MAINCPU :	strncpy(tmp, "Main", sizeof(tmp));
								break;
				case SUBCPU :	strncpy(tmp, "Sub ", sizeof(tmp));
								break;
#ifdef JSUB
				case JSUBCPU :	strncpy(tmp, "Jsub", sizeof(tmp));
								break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
				case MAINZ80 : strncpy(tmp, "Z80 ", sizeof(tmp));
								break;
#endif
#endif
			}
			_snprintf(string, sizeof(string), "%2d %s %04X ",
				i + 1, tmp, breakp[i].addr);
			switch (breakp[i].flag) {
				case BREAKP_ENABLED:
					strncat(string, " Enabled", sizeof(string) - strlen(string) - 1);
					break;
				case BREAKP_DISABLED:
					strncat(string, "Disabled", sizeof(string) - strlen(string) - 1);
					break;
			}
		}

		/* Rs[ */
		memcpy(&p[x * (i % (BREAKP_MAXNUM / 2)) + (i / (BREAKP_MAXNUM / 2)) * 23], string, strlen(string));
	}

	/* VMAbN */
	UnlockVM();
}

/*
 *	u[N|CgEChE
 *	`
 */
static void FASTCALL DrawBreakPoint(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pBreakPoint) {
		return;
	}
	SetupBreakPoint(pBreakPoint, x, y);

	/* ` */
	DrawWindowText(hDC, pBreakPoint, x, y);
}

/*
 *	u[N|CgEChE
 *	tbV
 */
void FASTCALL RefreshBreakPoint(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_BREAKPOINT] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_BREAKPOINT];
	hDC = GetDC(hWnd);
	DrawBreakPoint(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	u[N|CgEChE
 *	ĕ`
 */
static void FASTCALL PaintBreakPoint(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pBreakPoint;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawBreakPoint(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	u[N|Cg
 *	R}h
 */
static void FASTCALL CmdBreakPoint(HWND hWnd, WORD wID)
{
	int y, x, num;
	POINT point;

	ASSERT(hWnd);

	/* CfbNXԍ擾 */
	point = PosBreakPoint;
	y =		point.y / lCharHeight;
	x =		point.x / lCharWidth;
	num =	y + (x / 23) * (BREAKP_MAXNUM / 2);
	if ((num < 0) || (num >= BREAKP_MAXNUM) || ((x >= 21) && (x <= 22))) {
		return;
	}

	/* R}h */
	switch (wID) {
		/* Wv */
		case IDM_BREAKP_JUMP:
			if (breakp[num].flag != BREAKP_NOTUSE) {
				AddrDisAsm((BYTE)breakp[num].cpu, breakp[num].addr);
			}
			break;

		/* ݒ */
		case IDM_BREAKP_SET:
			BreakPointDlg(hWnd, num);
			break;

		/* Cl[u */
		case IDM_BREAKP_ENABLE:
			if (breakp[num].flag == BREAKP_DISABLED) {
				breakp[num].flag = BREAKP_ENABLED;
				InvalidateRect(hWnd, NULL, FALSE);
				RefreshDisAsm();
			}
			break;

		/* fBZ[u */
		case IDM_BREAKP_DISABLE:
			if (breakp[num].flag == BREAKP_ENABLED) {
				breakp[num].flag = BREAKP_DISABLED;
				InvalidateRect(hWnd, NULL, FALSE);
				RefreshDisAsm();
			}
			break;

		/* NA */
		case IDM_BREAKP_CLEAR:
			if (breakp[num].flag != BREAKP_NOTUSE) {
				schedule_setbreak(breakp[num].cpu, breakp[num].addr);
			}
			InvalidateRect(hWnd, NULL, FALSE);
			RefreshDisAsm();
			break;

		/* SăNA */
		case IDM_BREAKP_ALL:
			for (num=0; num<BREAKP_MAXNUM; num++) {
				if (breakp[num].flag != BREAKP_NOTUSE) {
					schedule_setbreak(breakp[num].cpu, breakp[num].addr);
				}
			}
			InvalidateRect(hWnd, NULL, FALSE);
			RefreshDisAsm();
			break;

		/* SăCl[u */
		case IDM_BREAKP_ENABLE_ALL:
			for (num=0; num<BREAKP_MAXNUM; num++) {
				if (breakp[num].flag == BREAKP_DISABLED) {
					breakp[num].flag = BREAKP_ENABLED;
				}
			}
			InvalidateRect(hWnd, NULL, FALSE);
			RefreshDisAsm();
			break;

		/* SăfBZ[u */
		case IDM_BREAKP_DISABLE_ALL:
			for (num=0; num<BREAKP_MAXNUM; num++) {
				if (breakp[num].flag == BREAKP_ENABLED) {
					breakp[num].flag = BREAKP_DISABLED;
				}
			}
			InvalidateRect(hWnd, NULL, FALSE);
			RefreshDisAsm();
			break;
	}
}

/*
 *	u[N|CgEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK BreakPointProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	POINT point;
	int y, x, i;
	HMENU hMenu;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintBreakPoint(hWnd);
			UnlockVM();
			return 0;

		/* NbN */
		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return 0;
			}
#endif

			/* J[\ʒuA */
			y = HIWORD(lParam) / lCharHeight;
			x = LOWORD(lParam) / lCharWidth;
			i =	y + (x / 23) * (BREAKP_MAXNUM / 2);
			if ((x >= 21) && (x <= 22)) {
				return 0;
			}
			if ((i >= 0) && (i < BREAKP_MAXNUM)) {
				if (breakp[i].flag != BREAKP_NOTUSE) {
					AddrDisAsm((BYTE)breakp[i].cpu, breakp[i].addr);
				}
			}
			return 0;

		/* ReLXgj[ */
		case WM_RBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return 0;
			}
#endif

			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			PosBreakPoint = point;
			hMenu = GetSubMenu(hBreakPoint, 0);
			ClientToScreen(hWnd, &point);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* j[폜 */
			DestroyMenu(hBreakPoint);

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pBreakPoint, NULL);

			UnlockVM();
			break;

		/* R}h */
		case WM_COMMAND:
			CmdBreakPoint(hWnd, LOWORD(wParam));
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	u[N|CgEChE
 *	EChE쐬
 */
HWND FASTCALL CreateBreakPoint(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_BreakPoint";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 44;
	rect.bottom = lCharHeight * BREAKP_MAXNUM / 2;

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_BREAKPOINT,
				szWndName, sizeof(szWndName));
	pBreakPoint = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* j[[h */
	hBreakPoint = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_BREAKPOINTMENU));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = BreakPointProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ XPW[EChE ]------------------------------------------------*/

static char *pszSchedulerTitle[] = {
	"Main Timer",
	"Sub  Timer",
	"OPN Timer A",
	"OPN Timer B",
	"Key Repeat",
	"BEEP",
	"V-SYNC",
	"V/H-BLANK",
#if XM7_VER >= 2
	"Line LSI",
	"RTC Clock",	/* V3.2L02` */
#else
	"Text Blink",
#ifdef SFDC
	"SFDD(Seek)",
#else
	"(Reserved)",
#endif
#endif
	"WHG Timer A",
	"WHG Timer B",
	"THG Timer A",
	"THG Timer B",
#if XM7_VER == 1 && defined(SFDC)
	"MFDC(M.Sec)",
	"MFDC(Lost)",
	"MFDD(Seek)",
#else
	"FDC (M.Sec)",
	"FDC (Lost)",
	"FDD (Seek)",
#endif
	"Tape Sound",
#ifdef MOUSE
	"Mouse Lost",
#else
	"(Reserved)",
#endif
	"DMA Start",
#if XM7_VER >= 2
	"KeyEnc ACK",
#elif defined(SFDC)
	"SFDC NMI",
#else
	"(Reserved)",
#endif
#ifdef RSC
	"RS TxTiming",
	"RS RxTiming",
#else
	"(Reserved)",
	"(Reserved)",
#endif
#if XM7_VER >= 2
	"KeyEnc BEEP",
	"KeyEnc MSG",
#elif defined(SFDC)
	"SFDC(M.Sec)",
	"SFDC(Lost)",
#else
	"(Reserved)",
	"(Reserved)",
#endif
#ifdef MOUSE
	"PTM Timer",
#else
	"(Reserved)",
#endif
#ifdef BUBBLE
	"BMC(32-L)",
	"BMC(32-M)",
#else	/* XM7_VER == 1 && defined(BUBBLE) */
	"(Reserved)",
	"(Reserved)",
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
#if defined(SFDC)
	"MFDD_MOTOR",
	"SFDD_MOTOR",
#else
	"FDD_MOTOR",
	"(Reserved)",
#endif
#ifdef BUBBLE
	"BMC(128-L)",
	"BMC(128-M)",
#else	/* XM7_VER == 1 && defined(BUBBLE) */
	"(Reserved)",
	"(Reserved)",
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
};

/*
 *	XPW[EChE
 *	ZbgAbv
 */
static void FASTCALL SetupScheduler(BYTE *p, int x, int y)
{
	int i;
	int j;
	char string[128];
	DWORD dwFrame;
	DWORD dwTime;

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* VMbN */
	LockVM();

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* s */
	_snprintf(string, sizeof(string), "Execute Time %12d ms",
		dwExecTotal / 1000);
	memcpy(p, string, strlen(string));

	/* t[g[^ */
	dwFrame = dwDrawTotal - dwSchedulerFrame[0];
	dwTime = timeGetTime() - dwSchedulerTime[0];
	if (dwTime > 0) {
		dwFrame = dwFrame * 100000 / dwTime;
	}
	else {
		dwSchedulerTime[0] = timeGetTime();
		dwFrame = 0;
	}
	_snprintf(string, sizeof(string), "Frame Rate (Total) %3d.%02d fps",
		dwFrame / 100, dwFrame % 100);
	memcpy(&p[x * 1], string, strlen(string));

	/* t[2sec */
	dwTime = timeGetTime() - dwSchedulerTime[1];
	if (dwTime > 0) {
		dwFrame = dwDrawTotal - dwSchedulerFrame[1];
		dwFrame = dwFrame * 100000 / dwTime;
	}
	else {
		dwFrame = 0;
	}
	if ((dwTime < 0) || (dwTime > 2000)) {
		dwSchedulerTime[1] = timeGetTime();
		dwSchedulerFrame[1] = dwDrawTotal;
	}
	_snprintf(string, sizeof(string), "Frame Rate (Latest)%3d.%02d fps",
		dwFrame / 100, dwFrame % 100);
	memcpy(&p[x * 2], string, strlen(string));

	/* [v */
	j = 4;
	for (i=0; i<EVENT_MAXNUM; i++) {
		/* ^Cg */
		memcpy(&p[x * j], pszSchedulerTitle[i], strlen(pszSchedulerTitle[i]));

		/* JgA[h */
		if (event[i].flag != EVENT_NOTUSE) {
			_snprintf(string, sizeof(string), "%4d.%03dms",
				event[i].current / 1000, event[i].current % 1000);
			memcpy(&p[x * j + 12], string, strlen(string));

			_snprintf(string, sizeof(string), "(%4d.%03dms)",
				event[i].reload / 1000, event[i].reload % 1000);
			memcpy(&p[x * j + 23], string, strlen(string));

			/* Xe[^X */
			switch (event[i].flag) {
				case EVENT_ENABLED:
					strncpy(string, " Enabled", sizeof(string));
					break;

				case EVENT_DISABLED:
					strncpy(string, "Disabled", sizeof(string));
					break;

				default:
					ASSERT(FALSE);
					break;
			}
		}
		else {
			strncpy(string, "", sizeof(string));
		}

		memcpy(&p[x * j + 36], string, strlen(string));
		j++;
	}

	/* I[o[TCN */
	strncpy(string,  "|  Over Cycle", sizeof(string));
	memcpy(&p[x * 0 + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "| Main %7d", main_overcycles);
	memcpy(&p[x * 1 + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "| Sub  %7d", sub_overcycles);
	memcpy(&p[x * 2 + 30], string, strlen(string));

#ifdef DEBUG
	if (dwFrame == 0) {
		dwTime = 100000;
	}
	else {
		dwTime = 100000 / dwFrame;
	}

	/* (ł΂΁`p) */
	_snprintf(string, sizeof(string), "Exec Time (CPU)  %4d.%03d ms (%3d%%)",
		dwCpuExecTime / 1000, dwCpuExecTime % 1000, dwCpuExecTime / dwTime);
	memcpy(&p[x * (j + 1)], string, strlen(string));

	_snprintf(string, sizeof(string), "Exec Time (DRAW) %4d.%03d ms (%3d%%)",
		dwDrawExecTime / 1000, dwDrawExecTime % 1000, dwDrawExecTime / dwTime);
	memcpy(&p[x * (j + 2)], string, strlen(string));

	_snprintf(string, sizeof(string), "Exec Time (POLL) %4d.%03d ms (%3d%%)",
		dwPollExecTime / 1000, dwPollExecTime % 1000, dwPollExecTime / dwTime);
	memcpy(&p[x * (j + 3)], string, strlen(string));

	i = (dwTime * 100) - (dwCpuExecTime + dwDrawExecTime + dwPollExecTime);
	if (i < 0) {
		i = 0;
	}
	_snprintf(string, sizeof(string), "Idle Time        %4d.%03d ms (%3d%%)",
		i / 1000, i % 1000, i / dwTime);
	memcpy(&p[x * (j + 4)], string, strlen(string));
#endif

	/* VMAbN */
	UnlockVM();
}

/*
 *	XPW[EChE
 *	`
 */
static void FASTCALL DrawScheduler(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pScheduler) {
		return;
	}
	SetupScheduler(pScheduler, x, y);

	/* ` */
	DrawWindowText(hDC, pScheduler, x, y);
}

/*
 *	XPW[EChE
 *	tbV
 */
void FASTCALL RefreshScheduler(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_SCHEDULER] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_SCHEDULER];
	hDC = GetDC(hWnd);
	DrawScheduler(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	XPW[EChE
 *	ĕ`
 */
static void FASTCALL PaintScheduler(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pScheduler;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawScheduler(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	XPW[EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK SchedulerProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintScheduler(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pScheduler, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	XPW[EChE
 *	EChE쐬
 */
HWND FASTCALL CreateScheduler(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Scheduler";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 44;
#ifdef DEBUG
	rect.bottom = lCharHeight * (EVENT_MAXNUM + 9);
#else
	rect.bottom = lCharHeight * (EVENT_MAXNUM + 4);
#endif

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_SCHEDULER,
				szWndName, sizeof(szWndName));
	pScheduler = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* ԁAt[ */
	dwSchedulerTime[0] = timeGetTime();
	dwSchedulerTime[1] = dwSchedulerTime[0];
	dwSchedulerFrame[0] = dwDrawTotal;
	dwSchedulerFrame[1] = dwSchedulerFrame[0];

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = SchedulerProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ WX^eύX_CAO ]--------------------------------------------*/

/*
 *	WX^eύX_CAO
 *	_CAO
 */
static BOOL FASTCALL CPURegDlgInit(HWND hDlg)
{
	static const char *RegName[] = {
		"CC", "A", "B", "DP", "D", "X", "Y", "U", "S", "PC"
	};
#if XM7_VER == 1 && defined(Z80CARD)
	static const char *RegNameZ80[] = {
		"AF",	"BC",	"DE",	"HL",	"IX",	"IY",	"SP",	"PC",
	};
#endif

	HWND hWnd;
	RECT prect;
	RECT drect;
	cpu6809_t *pReg;
	char tmp[16];
	char tmp2[128];
	char string[128];
	WORD dat;
	UINT id;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* ݒ */
	switch (nRegDlgCPU) {
		case MAINCPU :	id = IDS_SWND_CPU_MAIN;
						pReg = &maincpu;
						break;
		case SUBCPU :	id = IDS_SWND_CPU_SUB;
						pReg = &subcpu;
						break;
#ifdef JSUB
		case JSUBCPU :	id = IDS_SWND_CPU_JSUB;
						pReg = &jsubcpu;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	id = IDS_SWND_CPU_MAINZ80;
						pReg = NULL;
						break;
#endif
#endif
	}
	LoadString(hResInstance, id, tmp, sizeof(tmp));

	/* AhXݒ */
	hWnd = GetDlgItem(hDlg, IDC_REGLABEL);
	ASSERT(hWnd);
#if XM7_VER == 1 && defined(Z80CARD)
	if (nRegDlgCPU == MAINZ80) {
		if (LoadString(hResInstance, IDS_SWND_REG_MESSAGE,
						tmp2, sizeof(tmp2)) == 0) {
			_snprintf(string, sizeof(string), "%sWX^ (%s)",
				RegNameZ80[nRegDlgNo], "Z80");
		}
		else {
			_snprintf(string, sizeof(string), tmp2,
				RegNameZ80[nRegDlgNo], "Z80");
		}
	}
	else {
		if (LoadString(hResInstance, IDS_SWND_REG_MESSAGE,
						tmp2, sizeof(tmp2)) == 0) {
			_snprintf(string, sizeof(string), "%sWX^ (%s)",
				RegName[nRegDlgNo], tmp);
		}
		else {
			_snprintf(string, sizeof(string), tmp2, RegName[nRegDlgNo], tmp);
		}
	}
#else
	if (LoadString(hResInstance, IDS_SWND_REG_MESSAGE,
					tmp2, sizeof(tmp2)) == 0) {
		_snprintf(string, sizeof(string), "%sWX^ (%s)",
			RegName[nRegDlgNo], tmp);
	}
	else {
		_snprintf(string, sizeof(string), tmp2, RegName[nRegDlgNo], tmp);
	}
#endif
	SetWindowText(hWnd, string);

	/* f[^ݒ */
	hWnd = GetDlgItem(hDlg, IDC_REGEDIT);
	ASSERT(hWnd);

#if XM7_VER == 1 && defined(Z80CARD)
	if (nRegDlgCPU == MAINZ80) {
		switch (nRegDlgNo) {
			case 0:	/* AF */
					dat = (WORD)((mainz80.regs8[REGID_A] << 8) |
						mainz80.regs8[REGID_F]);
					break;
			case 1:	/* BC */
					dat = (WORD)((mainz80.regs8[REGID_B] << 8) |
						mainz80.regs8[REGID_C]);
					break;
			case 2:	/* DE */
					dat = (WORD)((mainz80.regs8[REGID_D] << 8) |
						mainz80.regs8[REGID_E]);
					break;
			case 3:	/* HL */
					dat = (WORD)((mainz80.regs8[REGID_H] << 8) |
						mainz80.regs8[REGID_L]);
					break;
			case 4:	/* IX */
					dat = (WORD)((mainz80.regs8[REGID_IXH] << 8) |
						mainz80.regs8[REGID_IXL]);
					break;
			case 5:	/* IY */
					dat = (WORD)((mainz80.regs8[REGID_IYH] << 8) |
						mainz80.regs8[REGID_IYL]);
					break;
			case 6:	/* SP */
					dat = mainz80.sp;
					break;
			case 7:	/* PC */
					dat = mainz80.pc;
					break;
			default:
					ASSERT(FALSE);
		}
		_snprintf(string, sizeof(string), "%04X", dat);
		SetWindowText(hWnd, string);

		return TRUE;
	}
#endif

	switch (nRegDlgNo) {
		case 0:	/* CC */
				dat = pReg -> cc;
				break;
		case 1:	/* A */
				dat = pReg -> acc.h.a;
				break;
		case 2:	/* B */
				dat = pReg -> acc.h.b;
				break;
		case 3:	/* DP */
				dat = pReg -> dp;
				break;
		case 4:	/* D */
				dat = pReg -> acc.d;
				break;
		case 5:	/* X */
				dat = pReg -> x;
				break;
		case 6:	/* Y */
				dat = pReg -> y;
				break;
		case 7:	/* U */
				dat = pReg -> u;
				break;
		case 8:	/* S */
				dat = pReg -> s;
				break;
		case 9:	/* PC */
				dat = pReg -> pc;
				break;
		default:
				ASSERT(FALSE);
	}
	if (nRegDlgNo < 4) {
		/* 8rbgWX^ */
		_snprintf(string, sizeof(string), "%02X", dat);
	}
	else {
		/* 16rbgWX^ */
		_snprintf(string, sizeof(string), "%04X", dat);
	}
	SetWindowText(hWnd, string);

	return TRUE;
}

/*
 *	WX^eύX_CAO
 *	_CAOOK
 */
static void FASTCALL CPURegDlgOK(HWND hDlg)
{
	HWND hWnd;
	cpu6809_t *pReg;
	char string[128];
	WORD dat;

	ASSERT(hDlg);
	LockVM();

	/* ݒ */
	switch (nRegDlgCPU) {
		case MAINCPU :	pReg = &maincpu;
						break;
		case SUBCPU :	pReg = &subcpu;
						break;
#ifdef JSUB
		case JSUBCPU :	pReg = &jsubcpu;
						break;
#endif
	}

	/* ݂̒l擾 */
	hWnd = GetDlgItem(hDlg, IDC_REGEDIT);
	ASSERT(hWnd);
	GetWindowText(hWnd, string, sizeof(string) - 1);
	dat = (WORD)strtol(string, NULL, 16);

#if XM7_VER == 1 && defined(Z80CARD)
	if (nRegDlgCPU == MAINZ80) {
		/* f[^ݒ */
		switch (nRegDlgNo) {
			case 0:	/* AF */
					mainz80.regs8[REGID_A] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_F] = (BYTE)(dat & 0xff);
					break;
			case 1:	/* BC */
					mainz80.regs8[REGID_B] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_C] = (BYTE)(dat & 0xff);
					break;
			case 2:	/* DE */
					mainz80.regs8[REGID_D] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_E] = (BYTE)(dat & 0xff);
					break;
			case 3:	/* HL */
					mainz80.regs8[REGID_H] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_L] = (BYTE)(dat & 0xff);
					break;
			case 4:	/* IX */
					mainz80.regs8[REGID_IXH] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_IXL] = (BYTE)(dat & 0xff);
					break;
			case 5:	/* IY */
					mainz80.regs8[REGID_IYH] = (BYTE)((dat >> 8) & 0xff);
					mainz80.regs8[REGID_IYL] = (BYTE)(dat & 0xff);
					break;
			case 6:	/* SP */
					mainz80.sp = dat;
					break;
			case 7:	/* PC */
					mainz80.pc = dat;
					break;
			default:
					ASSERT(FALSE);
		}
	}
	else {
		/* f[^ݒ */
		switch (nRegDlgNo) {
			case 0:	/* CC */
					pReg -> cc = (BYTE)dat;
					break;
			case 1:	/* A */
					pReg -> acc.h.a = (BYTE)dat;
					break;
			case 2:	/* B */
					pReg -> acc.h.b = (BYTE)dat;
					break;
			case 3:	/* DP */
					pReg -> dp = (BYTE)dat;
					break;
			case 4:	/* D */
					pReg -> acc.d = dat;
					break;
			case 5:	/* X */
					pReg -> x = dat;
					break;
			case 6:	/* Y */
					pReg -> y = dat;
					break;
			case 7:	/* U */
					pReg -> u = dat;
					break;
			case 8:	/* S */
					pReg -> s = dat;
					break;
			case 9:	/* PC */
					pReg -> pc = dat;
					break;
			default:
					ASSERT(FALSE);
		}
	}
#else
	/* f[^ݒ */
	switch (nRegDlgNo) {
		case 0:	/* CC */
				pReg -> cc = (BYTE)dat;
				break;
		case 1:	/* A */
				pReg -> acc.h.a = (BYTE)dat;
				break;
		case 2:	/* B */
				pReg -> acc.h.b = (BYTE)dat;
				break;
		case 3:	/* DP */
				pReg -> dp = (BYTE)dat;
				break;
		case 4:	/* D */
				pReg -> acc.d = dat;
				break;
		case 5:	/* X */
				pReg -> x = dat;
				break;
		case 6:	/* Y */
				pReg -> y = dat;
				break;
		case 7:	/* U */
				pReg -> u = dat;
				break;
		case 8:	/* S */
				pReg -> s = dat;
				break;
		case 9:	/* PC */
				pReg -> pc = dat;
				break;
		default:
				ASSERT(FALSE);
	}
#endif

	UnlockVM();
}

/*
 *	WX^eύX_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK CPURegDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return CPURegDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						CPURegDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*
 *	WX^eύX
 */
static BOOL FASTCALL CPURegDlg(HWND hWnd, BYTE nCPU, int x, int y)
{
	int ret;

	ASSERT(hWnd);

	/* p[^ۑEݒ */
	nRegDlgCPU = nCPU;

	if (x == -1) {
		/* WX^ԍZbg */
		nRegDlgNo = (BYTE)y;
	}
#if XM7_VER == 1 && defined(Z80CARD)
	else if (nCPU == MAINZ80) {
		/* WX^ԍZbg */
		nRegDlgNo = (BYTE)(((x / 20) * 4) + y);

		/* W`FbN */
		if ((x % 20) >= 8) {
			return FALSE;
		}
	}
#endif
	else {
		/* WX^ԍZbg */
		nRegDlgNo = (BYTE)(((x / 10) * 4) + y);
		if (nRegDlgNo >= 4) {
			nRegDlgNo ++;
		}

		/* IRWX^EW`FbN */
		if (nRegDlgNo >= 10) {
			return FALSE;
		}
	}

	/* [__CAOs */
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_REGDLG), hWnd, CPURegDlgProc);
	if (ret != IDOK) {
		return FALSE;
	}
	return TRUE;
}

/*-[ CPUWX^EChE ]-------------------------------------------------*/

/*
 *	CPUWX^EChE
 *	ZbgAbv
 */
static void FASTCALL SetupCPURegister(BYTE nCPU, BYTE *p, int x, int y)
{
	char buf[128];
	cpu6809_t *pReg;

	ASSERT(nCPU < MAXCPU);
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* WX^obt@𓾂 */
	switch (nCPU) {
		case MAINCPU :	pReg = &maincpu;
						break;
		case SUBCPU :	pReg = &subcpu;
						break;
#ifdef JSUB
		case JSUBCPU :	pReg = &jsubcpu;
						break;
#endif
	}

	/* Zbg */
	_snprintf(buf, sizeof(buf),
		"CC   %02X   X  %04X   PC %04X", pReg->cc, pReg->x, pReg->pc);
	memcpy(&p[0 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"A    %02X   Y  %04X   IR %04X", pReg->acc.h.a, pReg->y, pReg->intr);
	memcpy(&p[1 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"B    %02X   U  %04X", pReg->acc.h.b, pReg->u);
	memcpy(&p[2 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"DP   %02X   S  %04X", pReg->dp, pReg->s);
	memcpy(&p[3 * x + 0], buf, strlen(buf));
}

/*
 *	CPUWX^EChE(Z80)
 *	ZbgAbv
 */
#if XM7_VER == 1 && defined(Z80CARD)
static void FASTCALL SetupCPURegisterZ80(BYTE nCPU, BYTE *p, int x, int y)
{
	char buf[128];

//	ASSERT(nCPU != MAINZ80);
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);
	UNUSED(nCPU);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* Zbg */
	_snprintf(buf, sizeof(buf),
		"AF  %02X%02X  AF' %04X  IX  %02X%02X",
		mainz80.regs8[REGID_A], mainz80.regs8[REGID_F], mainz80.saf,
		mainz80.regs8[REGID_IXH], mainz80.regs8[REGID_IXL]);
	memcpy(&p[0 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"BC  %02X%02X  BC' %04X  IY  %02X%02X",
		mainz80.regs8[REGID_B], mainz80.regs8[REGID_C], mainz80.sbc,
		mainz80.regs8[REGID_IYH], mainz80.regs8[REGID_IYL]);
	memcpy(&p[1 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"DE  %02X%02X  DE' %04X  SP  %04X",
		mainz80.regs8[REGID_D], mainz80.regs8[REGID_E], mainz80.sde,
		mainz80.sp);
	memcpy(&p[2 * x + 0], buf, strlen(buf));
	_snprintf(buf, sizeof(buf),
		"HL  %02X%02X  HL' %04X  PC  %04X",
		mainz80.regs8[REGID_H], mainz80.regs8[REGID_L], mainz80.shl,
		mainz80.pc);
	memcpy(&p[3 * x + 0], buf, strlen(buf));
}
#endif

/*
 *	CPUWX^EChE
 *	`
 */
static void FASTCALL DrawCPURegister(HWND hWnd, HDC hDC)
{
	int i;
	BYTE nCPU;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			nCPU = (BYTE)((i - SWND_CPUREG_MAIN) % MAXCPU);
			p = pCPURegister[nCPU];
			break;
		}
	}
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
#if XM7_VER == 1 && defined(Z80CARD)
	if (nCPU == MAINZ80) {
		SetupCPURegisterZ80(nCPU, p, x, y);
	}
	else {
		SetupCPURegister(nCPU, p, x, y);
	}
#else
	SetupCPURegister(nCPU, p, x, y);
#endif

	/* ` */
	DrawWindowText(hDC, p, x, y);
}

/*
 *	CPUWX^EChE
 *	tbV
 */
void FASTCALL RefreshCPURegister(void)
{
	HWND hWnd;
	HDC hDC;

	/* CCPU */
	if (hSubWnd[SWND_CPUREG_MAIN]) {
		hWnd = hSubWnd[SWND_CPUREG_MAIN];
		hDC = GetDC(hWnd);
		DrawCPURegister(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

	/* TuCPU */
	if (hSubWnd[SWND_CPUREG_SUB]) {
		hWnd = hSubWnd[SWND_CPUREG_SUB];
		hDC = GetDC(hWnd);
		DrawCPURegister(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

#ifdef JSUB
	/* {TuCPU */
	if (hSubWnd[SWND_CPUREG_JSUB]) {
		hWnd = hSubWnd[SWND_CPUREG_JSUB];
		hDC = GetDC(hWnd);
		DrawCPURegister(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
#endif

#if XM7_VER == 1
#ifdef Z80CARD
	/* CCPU(Z80) */
	if (hSubWnd[SWND_CPUREG_Z80]) {
		hWnd = hSubWnd[SWND_CPUREG_Z80];
		hDC = GetDC(hWnd);
		DrawCPURegister(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
#endif
#endif
}

/*
 *	CPUWX^EChE
 *	ĕ`
 */
static void FASTCALL PaintCPURegister(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	int i;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			p = pCPURegister[((i - SWND_CPUREG_MAIN) % MAXCPU)];
			break;
		}
	}

	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawCPURegister(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	CPUWX^EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK CPURegisterProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam, BYTE nCPU)
{
	HMENU hMenu;
	POINT point;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintCPURegister(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pCPURegister[nCPU], hCPURegister[nCPU]);

			UnlockVM();
			break;

		/* WX^eύX */
		case WM_LBUTTONDBLCLK:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* J[\ʒuA */
			point.x = LOWORD(lParam) / lCharWidth;
			point.y = HIWORD(lParam) / lCharHeight;
			CPURegDlg(hWnd, nCPU, point.x, point.y);
			return 0;

		/* ReLXgj[ */
		case WM_CONTEXTMENU:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* ReLXgj[[h */
			hMenu = GetSubMenu(hCPURegister[nCPU], 0);

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* R}h */
		case WM_COMMAND:
#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				CPURegDlg(hWnd, nCPU, -1, LOWORD(wParam) - IDM_REG_AF);
			}
			else {
				CPURegDlg(hWnd, nCPU, -1, LOWORD(wParam) - IDM_REG_CC);
			}
#else
			CPURegDlg(hWnd, nCPU, -1, LOWORD(wParam) - IDM_REG_CC);
#endif
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	CPUWX^EChE(C)
 *	EChEvV[W
 */
static LRESULT CALLBACK CPURegisterProcMain(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return CPURegisterProc(hWnd, message, wParam, lParam, MAINCPU);
}

/*
 *	CPUWX^EChE(Tu)
 *	EChEvV[W
 */
static LRESULT CALLBACK CPURegisterProcSub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return CPURegisterProc(hWnd, message, wParam, lParam, SUBCPU);
}

/*
 *	CPUWX^EChE({Tu)
 *	EChEvV[W
 */
#ifdef JSUB
static LRESULT CALLBACK CPURegisterProcJsub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return CPURegisterProc(hWnd, message, wParam, lParam, JSUBCPU);
}
#endif

/*
 *	CPUWX^EChE(CZ80)
 *	EChEvV[W
 */
#if XM7_VER == 1
#ifdef Z80CARD
static LRESULT CALLBACK CPURegisterProcZ80(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return CPURegisterProc(hWnd, message, wParam, lParam, MAINZ80);
}
#endif
#endif

/*
 *	CPUWX^EChE
 *	EChE쐬
 */
HWND FASTCALL CreateCPURegister(HWND hParent, BYTE nCPU, int index)
{
	WNDCLASSEX wcex;
	char szClassNameMain[] = "XM7_CPURegisterMain";
	char szClassNameSub[] = "XM7_CPURegisterSub";
#ifdef JSUB
	char szClassNameJsub[] = "XM7_CPURegisterJsub";
#endif
#if XM7_VER == 1
#ifdef Z80CARD
	char szClassNameZ80[] = "XM7_CPURegisterZ80";
#endif
#endif
	char *szClassName;
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	UINT id;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 27;
	rect.bottom = lCharHeight * 4;

	/* EChE^CgAobt@m */
	switch (nCPU) {
		case MAINCPU :	id = IDS_SWND_CPUREG_MAIN;
						szClassName = szClassNameMain;
						break;
		case SUBCPU :	id = IDS_SWND_CPUREG_SUB;
						szClassName = szClassNameSub;
						break;
#ifdef JSUB
		case JSUBCPU :	id = IDS_SWND_CPUREG_JSUB;
						szClassName = szClassNameJsub;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	id = IDS_SWND_CPUREG_Z80;
						szClassName = szClassNameZ80;
						rect.right = lCharWidth * 28;
						rect.bottom = lCharHeight * 4;
						break;
#endif
#endif
	}
	LoadString(hResInstance, id, szWndName, sizeof(szWndName));
	pCPURegister[nCPU] = malloc(2 * (rect.right / lCharWidth) *
							(rect.bottom / lCharHeight));
#if XM7_VER == 1 && defined(Z80CARD)
	if (nCPU == MAINZ80) {
		hCPURegister[nCPU] = LoadMenu(hResInstance,
								MAKEINTRESOURCE(IDR_CPUREGZ80MENU));
	}
	else {
		hCPURegister[nCPU] = LoadMenu(hResInstance,
								MAKEINTRESOURCE(IDR_CPUREGMENU));
	}
#else
	hCPURegister[nCPU] = LoadMenu(hResInstance,
							MAKEINTRESOURCE(IDR_CPUREGMENU));
#endif

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS;
	switch (nCPU) {
		case MAINCPU :	wcex.lpfnWndProc = CPURegisterProcMain;
						break;
		case SUBCPU :	wcex.lpfnWndProc = CPURegisterProcSub;
						break;
#ifdef JSUB
		case JSUBCPU :	wcex.lpfnWndProc = CPURegisterProcJsub;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	wcex.lpfnWndProc = CPURegisterProcZ80;
						break;
#endif
#endif
	}
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

HWND FASTCALL CreateCPURegisterMain(HWND hParent, int index)
{
	return	CreateCPURegister(hParent, MAINCPU, index);
}

HWND FASTCALL CreateCPURegisterSub(HWND hParent, int index)
{
	return	CreateCPURegister(hParent, SUBCPU, index);
}

#ifdef JSUB
HWND FASTCALL CreateCPURegisterJSub(HWND hParent, int index)
{
	return	CreateCPURegister(hParent, JSUBCPU, index);
}
#endif

#if XM7_VER == 1 && defined(Z80CARD)
HWND FASTCALL CreateCPURegisterZ80(HWND hParent, int index)
{
	return	CreateCPURegister(hParent, MAINZ80, index);
}
#endif

/*-[ X^bNEu[N|CgWv ]------------------------------------*/

/*
 *	X^bNWvj[}
 */
static void FASTCALL InsertStackJumpMenu(HMENU hMenu, BYTE nCPU, UINT id)
{
	BYTE (FASTCALL *readbnio)(WORD addr);
	cpu6809_t *pReg;
	MENUITEMINFO mii;
	char string[128];
	WORD addr;
	int i;

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* ڂǉ */
	for (i=15; i>=0; i--) {
		mii.wID = id + i;
		switch (nCPU) {
#if XM7_VER == 1 && defined(Z80CARD)
			case MAINZ80 :
#endif
			case MAINCPU :	readbnio = mainmem_readbnio;
							pReg = &maincpu;
							break;
			case SUBCPU :	readbnio = submem_readbnio;
							pReg = &subcpu;
							break;
#ifdef JSUB
			case JSUBCPU :	readbnio = jsubmem_readbnio;
							pReg = &jsubcpu;
							break;
#endif
		}
#if XM7_VER == 1 && defined(Z80CARD)
		if (nCPU  == MAINZ80) {
			addr = (WORD)(readbnio((WORD)(mainz80.sp + i + 1)) << 8);
			addr |= (WORD)readbnio((WORD)(mainz80.sp + i));
		}
		else {
			addr = (WORD)(readbnio((WORD)(pReg->s + i)) << 8);
			addr |= (WORD)readbnio((WORD)(pReg->s + i + 1));
		}
#else
		addr = (WORD)(readbnio((WORD)(pReg->s + i)) << 8);
		addr |= (WORD)readbnio((WORD)(pReg->s + i + 1));
#endif

		wStackAddr[i] = addr;
#if XM7_VER == 1 && defined(Z80CARD)
		if (nCPU  == MAINZ80) {
			_snprintf(string, sizeof(string), "SP+%d ($%04X)", i, addr);
		}
		else {
			_snprintf(string, sizeof(string), "%d,S ($%04X)", i, addr);
		}
#else
		_snprintf(string, sizeof(string), "%d,S ($%04X)", i, addr);
#endif
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, 0, TRUE, &mii);
	}
}

/*
 *	u[N|CgWvj[}
 */
static void FASTCALL InsertBreakPointMenu(HMENU hMenu, BYTE nCPU, UINT id)
{
	MENUITEMINFO mii;
	char string[128];
	int no;
	int i;

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* ڂǉ */
	no = 0;
	for (i=15; i>=0; i--) {
		if ((breakp[i].flag != BREAKP_NOTUSE) && (breakp[i].cpu == nCPU)) {
			mii.wID = id + no;
			wBreakAddr[no++] = breakp[i].addr;
			_snprintf(string, sizeof(string), "%d : $%04X",
				i+1, breakp[i].addr);
			mii.dwTypeData = string;
			mii.cch = strlen(string);
			InsertMenuItem(hMenu, 0, TRUE, &mii);
		}
	}
	if (no == 0) {
		/* 1u[N|CgȂꍇ */
		mii.wID = id;
		mii.fState = MFS_GRAYED;
		LoadString(hResInstance, IDS_SWND_BREAKPOINT_NONE,
				string, sizeof(string));
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, 0, TRUE, &mii);
	}
}

/*
 *	xN^[j[}
 */
static void FASTCALL InsertVectorMenu(HMENU hMenu, BYTE nCPU, UINT id)
{
	static const struct VECTOR {
		const WORD addr;
		const char * const title;
	} vector[MAXCPU][16] = {
		{	{	0xfbfe,	"BASIC RESET"		},
			{	0xfbfc,	"BASIC HOT RESET"	},
			{	0xfbfa,	"BIOS(RCB)"			},
			{	0xfba7,	"BIOS(REGISER)"		},
			{	0xfb9b,	"OPN BIOS"			},
			{	0,		""					},
		},
		{	{	0,		""},				},
#ifdef JSUB
		{	{	0,		""},				},
#endif
#if XM7_VER == 1 && defined(Z80CARD)
		{	{	0,		""},				},
#endif
	};
	BYTE (FASTCALL *readbnio)(WORD addr);
	MENUITEMINFO mii;
	char c, string[128];
	WORD addr;
	int i, num;

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	switch (nCPU) {
#if XM7_VER == 1 && defined(Z80CARD)
		case MAINZ80 :
#endif
		case MAINCPU :	readbnio = mainmem_readbnio;
						break;
		case SUBCPU :	readbnio = submem_readbnio;
						break;
#ifdef JSUB
		case JSUBCPU :	readbnio = jsubmem_readbnio;
						break;
#endif
	}

	switch (nCPU) {
		case MAINCPU :
			memset(string, 0, sizeof(string));
			num = 0;
			if (string[0] == 0) {
				/* F-BASIC V1.0 ($8900-$8909) */
				for (i = 0; i <= 0x8909 - 0x8900; i ++) {
					c = (char)readbnio((WORD)(0x8900 + i));
					if (c < '0' || c > '9') {
						string[0] = 0;
						break;
					}
					string[i] = c;
				}
			}
			if (string[0] == 0) {
				/* F-BASIC V2.0-V3.0 ($FBA5-$FBAE) */
				for (i = 0; i <= 0xfbae - 0xfba5; i ++) {
					c = (char)readbnio((WORD)(0xfba5 + i));
					if (c < '0' || c > '9') {
						string[0] = 0;
						break;
					}
					string[i] = c;
				}
			}
			if (string[0] == 0) {
				/* F-BASIC V3.3-V5.0 ($FB9D-$FBA6) */
				for (i = 0; i <= 0xfba6 - 0xfb9d; i ++) {
					c = (char)readbnio((WORD)(0xfb9d + i));
					if ((i != 3) && (c < '0' || c > '9')) {
						string[0] = 0;
						break;
					}
					string[i] = c;
				}
			}
			if (string[0] != '0')	break;
			if (string[1] > '5')	break;
			else if (string[1] < '3') {
				/* F-BASIC V1.0-V2.2 */
				num = 3;
				break;
			}
			else if (string[1] > '3') {
				/* F-BASIC V4.0-V5.0(v̂Hw) */
				num = 4;
				break;
			}
			else if (string[2] < '3') {
				/* F-BASIC V3.0-V3.1 */
				num = 3;
				break;
			}
			else if (string[2] < '5') {
				/* F-BASIC V3.3-V3.4 */
				num = 5;
				break;
			}
			else {
				/* F-BASIC V3.5 */
				num = 4;
				break;
			}
		case SUBCPU :
			num = 0;
			break;
#ifdef JSUB
		case JSUBCPU :
			num = 0;
			break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :
			num = 0;
			break;
#endif
#endif
		default:
			ASSERT(FALSE);
	}

	/* ڂǉ */
	for (i = 0; i < num/*vector[nCPU][i].addr != 0*/; i ++) {
		mii.wID = id + i;
		addr = (WORD)(readbnio((WORD)(vector[nCPU][i].addr)) << 8);
		addr |= (WORD)readbnio((WORD)(vector[nCPU][i].addr + 1));
		wVectorAddr[i] = addr;
		_snprintf(string, sizeof(string), "$%04X : %s", addr, vector[nCPU][i].title);
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, i, TRUE, &mii);
	}

	if (num == 0) {
		/* 1xN^[Ȃꍇ */
		mii.wID = id;
		mii.fState = MFS_GRAYED;
		LoadString(hResInstance, IDS_SWND_BREAKPOINT_NONE,
				string, sizeof(string));
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, 0, TRUE, &mii);
	}
}

/*
 *	AhXj[ݒ
 */
static void FASTCALL InsertPysicAddrMenu(HMENU hMenu, BYTE nCPU, UINT id)
{
	MENUITEMINFO mii;
	char string[128];
	int no;
	int i;

	UNUSED(nCPU);

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* ڂǉ */
#if XM7_VER >= 3 || XM7_VER == 1
#if XM7_VER >= 3
	if (model_type >= MODEL_FM77AV40) {
#else
	if (mmr_extram == EXRAM_960KB) {
#endif
		no = 0x0f;
	}
	else {
		no = 0x03;
	}
#else
	no = 0x03;
#endif
	for (i=no; i>=0; i--) {
		mii.wID = id + i;
		_snprintf(string, sizeof(string), "$%X0000-$%XFFFF", i, i);
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, 0, TRUE, &mii);
	}
}

/*
 *	MSRj[ݒ
 */
static void FASTCALL InsertSegmentMenu(HMENU hMenu, BYTE nCPU, UINT id)
{
	MENUITEMINFO mii;
	char string[128];
	int no;
	int i;

	UNUSED(nCPU);

	/* j[\̏ */
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = 44;
	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
	mii.fType = MFT_STRING;
	mii.fState = MFS_ENABLED;

	/* j[ׂč폜 */
	while (GetMenuItemCount(hMenu) > 0) {
		DeleteMenu(hMenu, 0, MF_BYPOSITION);
	}

	/* ڂǉ */
#if XM7_VER >= 2
	if (model_type > MODEL_FM77AV20) {
		no = 0x07;
	}
#else
	if (mmr_extram == EXRAM_960KB) {
		no = 0x3f;
	}
#endif
	else {
		no = 0x03;
	}
	for (i=no; i>=0; i--) {
		mii.wID = id + i;
		_snprintf(string, sizeof(string), "%d", i);
		mii.dwTypeData = string;
		mii.cch = strlen(string);
		InsertMenuItem(hMenu, 0, TRUE, &mii);
	}
}

/*-[ tAZuEChE ]------------------------------------------------*/

/*
 *	tAZuEChE
 *	AhXݒ
 */
void FASTCALL AddrDisAsm(BYTE nCPU, WORD wAddr)
{
	SCROLLINFO sif;
	HWND hWnd;

	/* EChEWIg𓾂 */
	hWnd = NULL;
	switch (nCPU) {
		case MAINCPU :	hWnd = hSubWnd[SWND_DISASM_MAIN];
						break;
		case SUBCPU :	hWnd = hSubWnd[SWND_DISASM_SUB];
						break;
#ifdef JSUB
		case JSUBCPU :	hWnd = hSubWnd[SWND_DISASM_JSUB];
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	hWnd = hSubWnd[SWND_DISASM_Z80];
						break;
#endif
#endif
	}
	if (!hWnd) {
		/* EChEȂ΂̂܂܃^[ */
		return;
	}

	/* SCROLLINFO\̂̏ */
	memset(&sif, 0, sizeof(sif));
	sif.cbSize = sizeof(sif);
	sif.fMask = SIF_POS;
	sif.nPos = (int)wAddr;

	/* ݃`FbNݒ */
	wDisAsm[nCPU] = wAddr;
	SetScrollInfo(hWnd, SB_VERT, &sif, TRUE);

	/* tbV */
	RefreshDisAsm();
}

/*
 *	tAZuEBhE
 *	TCY
 */
static void FASTCALL ResizeDisAsm(HWND hWnd, WORD cx, WORD cy, BYTE nCPU)
{
	RECT crect;
	RECT wrect;
	int width;

	ASSERT(hWnd);

	/* ŏ̏ꍇ́AȂ */
	if ((cx == 0) && (cy == 0)) {
		return;
	}

	/* TCYύX */
	width = 47;
	nHeightDisAsm[nCPU] = cy / lCharHeight;
	if (nHeightDisAsm[nCPU] < 1) {
		nHeightDisAsm[nCPU] = 1;
	}

	/* Ċm */
	free(pDisAsm[nCPU]);
	pDisAsm[nCPU] = malloc(2 * width * nHeightDisAsm[nCPU]);
	memset(pDisAsm[nCPU], 0xff, 2 * width * nHeightDisAsm[nCPU]);
	ASSERT(pDisAsm[nCPU]);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += (width * lCharWidth);
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += (nHeightDisAsm[nCPU] * lCharHeight);
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
}

/*
 *	tAZuEChE
 *	ZbgAbv
 */
static void FASTCALL SetupDisAsm(BYTE nCPU, BYTE *p, int x, int y)
{
	char string[128];
	int addr;
	int i;
	int j;
	int k;
	int ret;
	WORD pc;

	ASSERT(nCPU < MAXCPU);
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* ݒ */
	switch (nCPU) {
		case MAINCPU :	pc = maincpu.pc;
						break;
		case SUBCPU :	pc = subcpu.pc;
						break;
#ifdef JSUB
		case JSUBCPU :	pc = jsubcpu.pc;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	pc = (WORD)mainz80.pc;
						break;
#endif
#endif
	}
	addr = (int)wDisAsm[nCPU];

	/* LvVXV */
	if (nCPU == MAINCPU) {
		switch (nAddrModeDisAsm) {
		case 1:
			_snprintf(string, sizeof(string), " [$%05X - $%05X]",
						(nAddrSegDisAsm << 16), (nAddrSegDisAsm << 16) + 0xffff);
			break;
		case 2:
			_snprintf(string, sizeof(string), " [Segment:$%02X]", nAddrSegDisAsm);
			break;
		case 0:
			memset(string, 0, sizeof(string));
			break;
		default:
			ASSERT((nAddrModeDisAsm >= 0) && (nAddrModeDisAsm <= 2));
		}
		if (strncmp(szDisAsmWndName, string, sizeof(szDisAsmWndName))) {
			strncpy(szDisAsmWndName, string, sizeof(szDisAsmWndName));
			LoadString(hResInstance, IDS_SWND_DISASM_MAIN, string, sizeof(string));
			strncat(string, szDisAsmWndName, sizeof(string));
			SetWindowText(hSubWnd[SWND_DISASM_MAIN], string);
		}
	}
#if XM7_VER == 1 && defined(Z80CARD)
	if (nCPU == MAINZ80) {
		char string[128];
		switch (nAddrModeDisAsmZ80) {
		case 1:
			_snprintf(string, sizeof(string), " [$%05X - $%05X]",
						(nAddrBnkDisAsmZ80 << 16), (nAddrBnkDisAsmZ80 << 16) + 0xffff);
			break;
		case 0:
			memset(string, 0, sizeof(string));
			break;
		default:
			ASSERT((nAddrModeDisAsmZ80 >= 0) && (nAddrModeDisAsmZ80 <= 2));
		}
		if (strncmp(szDisAsmWndNameZ80, string, sizeof(szDisAsmWndNameZ80))) {
			strncpy(szDisAsmWndNameZ80, string, sizeof(szDisAsmWndNameZ80));
			LoadString(hResInstance, IDS_SWND_DISASM_Z80, string, sizeof(string));
			strncat(string, szDisAsmWndNameZ80, sizeof(string));
			SetWindowText(hSubWnd[SWND_DISASM_Z80], string);
		}
	}
#endif

	for (i=0; i<y; i++) {
		 /* tAZu */
#if XM7_VER == 1 && defined(Z80CARD)
		if (nCPU == MAINZ80) {
			ret = disline80(nCPU, (WORD)addr, string, nAddrBnkDisAsmZ80, nAddrModeDisAsmZ80);
		}
		else {
			ret = disline(nCPU, (WORD)addr, string, nAddrSegDisAsm, nAddrModeDisAsm);
		}
#else
		ret = disline(nCPU, (WORD)addr, string, nAddrSegDisAsm, nAddrModeDisAsm);
#endif

		/* Zbg */
		memcpy(&p[x * i + 3], string, strlen(string));

#if XM7_VER == 1 && defined(Z80CARD)
		if (((nCPU == MAINCPU) && (nAddrModeDisAsm == 0)) ||
			(nCPU == SUBCPU) ||
			(nCPU == JSUBCPU) ||
			((nCPU == MAINZ80) && (nAddrModeDisAsmZ80 == 0))) {
#else
		if ((nCPU != MAINCPU) || (nAddrModeDisAsm == 0)) {
#endif

			/* }[N */
			if (pc == addr) {
				p[x * i + 2] = '>';
			}
			for (j=0; j<BREAKP_MAXNUM; j++) {
				if (breakp[j].addr != addr) {
					continue;
				}
				if (breakp[j].cpu != nCPU) {
					continue;
				}
				if (breakp[j].flag == BREAKP_NOTUSE) {
					continue;
				}
				if (breakp[j].flag == BREAKP_DISABLED) {
					continue;
				}

				/* u[N|Cg */
				p[x * i + 0] = (BYTE)((j < 9 ? ' ' : '1'));
				p[x * i + 1] = (BYTE)((j < 9 ? '1' + j : 0x27 + j));

				/* ] */
				for (k=0; k<x; k++) {
					 p[x * i + k] = (BYTE)(p[x * i + k] | 0x80);
				}
			}
		}

		/* ZAI[o[`FbN */
		addr += ret;
		if (addr >= 0x10000) {
			break;
		}
	}
}

/*
 *	tAZuEChE
 *	`
 */
static void FASTCALL DrawDisAsm(HWND hWnd, HDC hDC)
{
	int i;
	BYTE nCPU;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			nCPU = (BYTE)((i - SWND_DISASM_MAIN) % MAXCPU);
			p = pDisAsm[nCPU];
			break;
		}
	}
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	SetupDisAsm(nCPU, p, x, y);

	/* ` */
	DrawWindowText2(hDC, p, x, y);
}

/*
 *	tAZuEChE
 *	tbV
 */
void FASTCALL RefreshDisAsm(void)
{
	HWND hWnd;
	HDC hDC;

	/* CCPU */
	if (hSubWnd[SWND_DISASM_MAIN]) {
		hWnd = hSubWnd[SWND_DISASM_MAIN];
		hDC = GetDC(hWnd);
		DrawDisAsm(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

	/* TuCPU */
	if (hSubWnd[SWND_DISASM_SUB]) {
		hWnd = hSubWnd[SWND_DISASM_SUB];
		hDC = GetDC(hWnd);
		DrawDisAsm(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

#ifdef JSUB
	/* {TuCPU */
	if (hSubWnd[SWND_DISASM_JSUB]) {
		hWnd = hSubWnd[SWND_DISASM_JSUB];
		hDC = GetDC(hWnd);
		DrawDisAsm(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
#endif

#if XM7_VER == 1
#ifdef Z80CARD
	/* CCPU(Z80) */
	if (hSubWnd[SWND_DISASM_Z80]) {
		hWnd = hSubWnd[SWND_DISASM_Z80];
		hDC = GetDC(hWnd);
		DrawDisAsm(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
#endif
#endif
}

/*
 *	tAZuEChE
 *	ĕ`
 */
static void FASTCALL PaintDisAsm(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	int i;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			p = pDisAsm[((i - SWND_DISASM_MAIN) % MAXCPU)];
			break;
		}
	}

	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawDisAsm(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	tAZuEChE
 *	NbN
 */
static void FASTCALL LButtonDisAsm(HWND hWnd, POINT point, BOOL mode)
{
	char string[128];
	int addr;
	BOOL flag;
	BYTE nCPU;
	int i;
	int y;
	int ret;

	ASSERT(hWnd);

	/* sJEg𓾂 */
	y = point.y / lCharHeight;

	/* ۂɋtAZuĂ݂ */
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			nCPU = (BYTE)((i - SWND_CPUREG_MAIN) % MAXCPU);
			addr = (int)wDisAsm[nCPU];
			break;
		}
	}

	/* tAZu [v */
	for (i=0; i<y; i++) {
#if XM7_VER == 1 && defined(Z80CARD)
		if (nCPU == MAINZ80) {
			ret = disline80(nCPU, (WORD)addr, string, nAddrBnkDisAsmZ80, nAddrModeDisAsmZ80);
		}
		else {
			ret = disline(nCPU, (WORD)addr, string, nAddrSegDisAsm, nAddrModeDisAsm);
		}
#else
		ret = disline(nCPU, (WORD)addr, string, nAddrSegDisAsm, nAddrModeDisAsm);
#endif
		addr += ret;
		if (addr >= 0x10000) {
			return;
		}
	}

	if (mode) {
		/* Wv */
		switch (nCPU) {
			case MAINCPU :	maincpu.pc = (WORD)addr;
							break;
			case SUBCPU :	subcpu.pc = (WORD)addr;
							break;
#ifdef JSUB
			case JSUBCPU :	jsubcpu.pc = (WORD)addr;
							break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
			case MAINZ80 :	mainz80.pc = (Uint32)addr;
							break;
#endif
#endif
		}
	}
	else {
		/* u[N|Cg on/off */
		flag = FALSE;
		for (i=0; i<BREAKP_MAXNUM; i++) {
			if ((breakp[i].cpu == nCPU) && (breakp[i].addr == addr)) {
				if (breakp[i].flag != BREAKP_NOTUSE) {
					breakp[i].flag = BREAKP_NOTUSE;
					flag = TRUE;
					break;
				}
			}
		}
		if (!flag) {
			schedule_setbreak(nCPU, (WORD)addr);
		}

		/* u[N|CgEChEAĕ`悷 */
		if (hSubWnd[SWND_BREAKPOINT]) {
			InvalidateRect(hSubWnd[SWND_BREAKPOINT], NULL, FALSE);
		}
	}

	/* ĕ` */
	InvalidateRect(hWnd, NULL, FALSE);
}

/*
 *	tAZuEChE
 *	R}h
 */
static void FASTCALL CmdDisAsm(HWND hWnd, WORD wID, BYTE nCPU)
{
	WORD target;
	cpu6809_t *cpu;

	/* CPU\̌ */
	switch (nCPU) {
#if XM7_VER == 1 && defined(Z80CARD)
		case MAINZ80 :
#endif
		case MAINCPU :	cpu = &maincpu;
						break;
		case SUBCPU :	cpu = &subcpu;
						break;
#ifdef JSUB
		case JSUBCPU :	cpu = &jsubcpu;
						break;
#endif
	}

	/* ^[QbgAhX */
	switch (wID) {
		case IDM_DIS_ADDR:
			target = wDisAsm[nCPU];
			if (!AddrDlg(hWnd, &target)) {
				return;
			}
			break;

		case IDM_DIS_PC:
			target = cpu->pc;
			break;
		case IDM_DIS_X:
			target = cpu->x;
			break;
		case IDM_DIS_Y:
			target = cpu->y;
			break;
		case IDM_DIS_U:
			target = cpu->u;
			break;
		case IDM_DIS_S:
			target = cpu->s;
			break;
#if XM7_VER == 1 && defined(Z80CARD)
		case IDM_DIS_PC_Z80:
			target = mainz80.pc;
			break;
		case IDM_DIS_BC:
			target = (WORD)((mainz80.regs8[REGID_B]) << 8 |
				mainz80.regs8[REGID_C]);
			break;
		case IDM_DIS_DE:
			target = (WORD)((mainz80.regs8[REGID_D]) << 8 |
				mainz80.regs8[REGID_E]);
			break;
		case IDM_DIS_HL:
			target = (WORD)((mainz80.regs8[REGID_H]) << 8 |
				mainz80.regs8[REGID_L]);
			break;
		case IDM_DIS_IX:
			target = (WORD)((mainz80.regs8[REGID_IXH]) << 8 |
				mainz80.regs8[REGID_IXL]);
			break;
		case IDM_DIS_IY:
			target = (WORD)((mainz80.regs8[REGID_IYH]) << 8 |
				mainz80.regs8[REGID_IYL]);
			break;
		case IDM_DIS_SP:
			target = mainz80.sp;
			break;
#endif

		case IDM_DIS_RESET:
			target = (WORD)((cpu->readmem(0xfffe) << 8) | cpu->readmem(0xffff));
			break;
		case IDM_DIS_NMI:
			target = (WORD)((cpu->readmem(0xfffc) << 8) | cpu->readmem(0xfffd));
			break;
		case IDM_DIS_SWI:
			target = (WORD)((cpu->readmem(0xfffa) << 8) | cpu->readmem(0xfffb));
			break;
		case IDM_DIS_IRQ:
			target = (WORD)((cpu->readmem(0xfff8) << 8) | cpu->readmem(0xfff9));
			break;
		case IDM_DIS_FIRQ:
			target = (WORD)((cpu->readmem(0xfff6) << 8) | cpu->readmem(0xfff7));
			break;
		case IDM_DIS_SWI2:
			target = (WORD)((cpu->readmem(0xfff4) << 8) | cpu->readmem(0xfff5));
			break;
		case IDM_DIS_SWI3:
			target = (WORD)((cpu->readmem(0xfff2) << 8) | cpu->readmem(0xfff3));
			break;
		case IDM_DIS_JUMP:
			LButtonDisAsm(hWnd, PosDisAsmPoint, TRUE);
#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				target = (WORD)mainz80.pc;
			}
			else {
				target = cpu->pc;
			}
#else
			target = cpu->pc;
#endif
			break;
		case IDM_DIS_SYNCPC:
			bSyncDisasm[nCPU] = (!bSyncDisasm[nCPU]);
			return;

		case IDM_DIS_LOGIC:
#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				nAddrBnkDisAsmZ80 = 0;						/* tAZuoN */
				nAddrModeDisAsmZ80 = 0;						/* tAZuAhX[h */
			}
			else {
				nAddrSegDisAsm = 0;							/* tAZuZOg */
				nAddrModeDisAsm = 0;						/* tAZuAhX[h */
			}
#else
			nAddrSegDisAsm = 0;								/* tAZuZOg */
			nAddrModeDisAsm = 0;							/* tAZuAhX[h */
#endif
			return;
		case IDM_DIS_COPY:
			SetClipboardSubWindow(hWnd, pDisAsm[nCPU], 47, nHeightDisAsm[nCPU], TRUE);
			return;
		default:
			if ((wID >= IDM_DIS_STACK) && (wID <= IDM_DIS_STACK + 15)) {
				target = (WORD)wStackAddr[wID - IDM_DIS_STACK];
			}
			else if ((wID >= IDM_DIS_BREAK) && (wID <= IDM_DIS_BREAK + 15)) {
				target = (WORD)wBreakAddr[wID - IDM_DIS_BREAK];
			}
			else if ((wID >= IDM_DIS_VECTOR) && (wID <= IDM_DIS_VECTOR + 15)) {
				target = (WORD)wVectorAddr[wID - IDM_DIS_VECTOR];
			}
			else {
				target = wDisAsm[nCPU];
				if ((wID >= IDM_DIS_PADR0) && (wID <= IDM_DIS_PADR0 + 0x0f)) {
#if XM7_VER == 1 && defined(Z80CARD)
					if (nCPU == MAINZ80) {
						nAddrBnkDisAsmZ80 = (wID - IDM_DIS_PADR0);	/* tAZuoN */
						nAddrModeDisAsmZ80 = 1;				/* tAZuAhX[h */
					}
					else {
						nAddrSegDisAsm = (wID - IDM_DIS_PADR0);	/* tAZuZOg */
						nAddrModeDisAsm = 1;				/* tAZuAhX[h */
					}
#else
					nAddrSegDisAsm = (wID - IDM_DIS_PADR0);	/* tAZuZOg */
					nAddrModeDisAsm = 1;					/* tAZuAhX[h */
#endif
				}
				else if ((wID >= IDM_DIS_SEG00) && (wID <= IDM_DIS_MAX)) {
					nAddrSegDisAsm = (wID - IDM_DIS_SEG00);	/* tAZuZOg */
					nAddrModeDisAsm = 2;					/* tAZuAhX[h */
				}
				else if ((wID >= IDM_DIS_LADR0) && (wID <= IDM_DIS_LADR0 + 0x0f)) {
					target = (wID - IDM_DIS_LADR0) << 12;
				}
			}
			break;
	}

	/* ݒ聕XV */
	AddrDisAsm(nCPU, (WORD)target);
}

/*
 *	tAZuEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK DisAsmProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam, BYTE nCPU)
{
	static BOOL bResize = FALSE;	/* 蓮TCYtO */
	WORD wAddr;
	char string[128];
	int ret;
	int i;
	POINT point;
	HMENU hMenu;
	RECT rect;
	BOOL bpysic;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintDisAsm(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pDisAsm[nCPU], hDisAsm[nCPU]);

			UnlockVM();
			break;

		/* EChETCYύX */
		case WM_SIZE:
			if (bResize) {
				bResize = FALSE;
				ResizeDisAsm(hWnd, LOWORD(lParam), HIWORD(lParam), nCPU);
			}
			break;

		/* EChETCYύX */
		case WM_SIZING:
			WindowSizing(hWnd, (LPRECT)lParam, &pDisAsm[nCPU]);
			bResize = TRUE;
			break;

		/* EBhETCYύXbZ[W */
		case WM_WINDOWPOSCHANGING:
			bResize = TRUE;
			break;

		/* ŏTCY */
		case WM_GETMINMAXINFO:
			rect.left = 0;
			rect.right = 47 * lCharWidth + GetSystemMetrics(SM_CXVSCROLL);
			rect.top = 0;
			rect.bottom = lCharHeight;
			if (bPopupSwnd) {
				AdjustWindowRect(&rect, WS_POPUP | WS_OVERLAPPED | WS_SYSMENU |
										WS_CAPTION | WS_VISIBLE |
										WS_MINIMIZEBOX | WS_SIZEBOX, FALSE);
			}
			else {
				AdjustWindowRect(&rect, WS_CHILD | WS_OVERLAPPED | WS_SYSMENU |
										WS_CAPTION | WS_VISIBLE |
										WS_MINIMIZEBOX | WS_CLIPSIBLINGS |
										WS_SIZEBOX, FALSE);
			}
			((LPMINMAXINFO)lParam)->ptMinTrackSize.x = rect.right - rect.left;
			((LPMINMAXINFO)lParam)->ptMaxTrackSize.x = rect.right - rect.left;
			((LPMINMAXINFO)lParam)->ptMinTrackSize.y = rect.bottom - rect.top;
			return 0;

		/* ReLXgj[ */
		case WM_RBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* ReLXgj[[h */
			hMenu = GetSubMenu(hDisAsm[nCPU], 0);

#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				HMENU hIrqVectorMenu;
				MENUITEMINFO mii;

				hIrqVectorMenu = GetSubMenu(hMenu, 4);

				/* j[ׂč폜 */
				while (GetMenuItemCount(hIrqVectorMenu) > 0) {
					DeleteMenu(hIrqVectorMenu, 0, MF_BYPOSITION);
				}

				/* j[\̏ */
				memset(&mii, 0, sizeof(mii));
				mii.cbSize = 44;
				mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID;
				mii.fType = MFT_STRING;

				/* 荞݃xN^[Wvj[ݒ */
				mii.wID = IDM_DIS_VECTOR;
				mii.fState = MFS_GRAYED;
				LoadString(hResInstance, IDS_SWND_BREAKPOINT_NONE,
						string, sizeof(string));
				mii.dwTypeData = string;
				mii.cch = strlen(string);
				InsertMenuItem(hIrqVectorMenu, 0, TRUE, &mii);
			}
#endif

			/* VXeX^bNWvj[ݒ */
			InsertStackJumpMenu(GetSubMenu(hMenu, 5), nCPU, IDM_DIS_STACK);

			/* u[N|Cgj[ݒ */
			InsertBreakPointMenu(GetSubMenu(hMenu, 6), nCPU, IDM_DIS_BREAK);

			/* ̑xN^[Wvj[ݒ */
			InsertVectorMenu(GetSubMenu(hMenu, 7), nCPU, IDM_DIS_VECTOR);

			/* AhX[h`FbN */
#if XM7_VER >= 2
			bpysic = ((nCPU == MAINCPU) && (model_type >= MODEL_FM77AV));
#else
			bpysic = ((nCPU == MAINCPU) && (model_type == MODEL_FM77));
#endif
#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				for (i = 1; i <= 6; i ++) {
					EnableMenuPos(hMenu, i, (nAddrModeDisAsmZ80 == 0));
				}
				EnableMenuPos(hMenu, 4, (nAddrModeDisAsmZ80 != 1));
				EnableMenuPos(hMenu, 7, (nAddrModeDisAsmZ80 != 1));
#if 0	/* RZ192Ή܂ŕ */
				EnableMenuSub(hMenu, IDM_DIS_LOGIC, TRUE);
				EnableMenuPos(hMenu, 11, TRUE);
#else
				EnableMenuSub(hMenu, IDM_DIS_LOGIC, FALSE);
				EnableMenuPos(hMenu, 11, FALSE);
#endif

				EnableMenuPos(hMenu, 12, FALSE);
				CheckMenuSub(hMenu, IDM_DIS_SYNCPC, bSyncDisasm[nCPU] && (nAddrModeDisAsmZ80 == 0));
				EnableMenuSub(hMenu, IDM_DIS_SYNCPC, bSync && (nAddrModeDisAsmZ80 == 0));

				/* AhXj[ݒ */
				InsertPysicAddrMenu(GetSubMenu(hMenu, 11), nCPU, IDM_DIS_PADR0);

#if 0	/* RZ192Ή܂ŕ */
				if (nAddrModeDisAsmZ80 == 1) {
					CheckMenuRadioItem(GetSubMenu(hMenu, 11), 0x00, 0x0f, nAddrBnkDisAsmZ80, MF_BYPOSITION);
				}
				CheckMenuRadioItem(hMenu, 10, 11, 10 + nAddrModeDisAsmZ80, MF_BYPOSITION);
#endif
			}
			else {
				for (i = 1; i <= 6; i ++) {
					EnableMenuPos(hMenu, i, !bpysic || (nAddrModeDisAsm == 0));
				}
				EnableMenuPos(hMenu, 4, !bpysic || (nAddrModeDisAsm != 1));
				EnableMenuPos(hMenu, 7, !bpysic || (nAddrModeDisAsm != 1));
				EnableMenuSub(hMenu, IDM_DIS_LOGIC, bpysic);
				EnableMenuPos(hMenu, 11, bpysic);

				EnableMenuPos(hMenu, 12, bpysic);
				CheckMenuSub(hMenu, IDM_DIS_SYNCPC, bSyncDisasm[nCPU] && (!bpysic || (nAddrModeDisAsm == 0)));
				EnableMenuSub(hMenu, IDM_DIS_SYNCPC, bSync && (!bpysic || (nAddrModeDisAsm == 0)));
				if (bpysic) {
					/* AhXj[ݒ */
					InsertPysicAddrMenu(GetSubMenu(hMenu, 11), nCPU, IDM_DIS_PADR0);

					/* MSRj[ݒ */
					InsertSegmentMenu(GetSubMenu(hMenu, 12), nCPU, IDM_DIS_SEG00);

					if (nAddrModeDisAsm == 1) {
						CheckMenuRadioItem(GetSubMenu(hMenu, 11), 0x00, 0x0f, nAddrSegDisAsm, MF_BYPOSITION);
					}
					else if (nAddrModeDisAsm == 2) {
						CheckMenuRadioItem(GetSubMenu(hMenu, 12), 0x00, 0x3f, nAddrSegDisAsm, MF_BYPOSITION);
					}
					CheckMenuRadioItem(hMenu, 10, 12, 10 + nAddrModeDisAsm, MF_BYPOSITION);
				}
			}
#else
			for (i = 1; i <= 6; i ++) {
				EnableMenuPos(hMenu, i, !bpysic || (nAddrModeDisAsm == 0));
			}
			EnableMenuPos(hMenu, 4, !bpysic || (nAddrModeDisAsm != 1));
			EnableMenuPos(hMenu, 7, !bpysic || (nAddrModeDisAsm != 1));
			EnableMenuSub(hMenu, IDM_DIS_LOGIC, bpysic);
			EnableMenuPos(hMenu, 11, bpysic);

			EnableMenuPos(hMenu, 12, bpysic);
			CheckMenuSub(hMenu, IDM_DIS_SYNCPC, bSyncDisasm[nCPU] && (!bpysic || (nAddrModeDisAsm == 0)));
			EnableMenuSub(hMenu, IDM_DIS_SYNCPC, bSync && (!bpysic || (nAddrModeDisAsm == 0)));
			if (bpysic) {
				/* AhXj[ݒ */
				InsertPysicAddrMenu(GetSubMenu(hMenu, 11), nCPU, IDM_DIS_PADR0);

				/* MSRj[ݒ */
				InsertSegmentMenu(GetSubMenu(hMenu, 12), nCPU, IDM_DIS_SEG00);

				if (nAddrModeDisAsm == 1) {
					CheckMenuRadioItem(GetSubMenu(hMenu, 11), 0x00, 0x0f, nAddrSegDisAsm, MF_BYPOSITION);
				}
				else if (nAddrModeDisAsm == 2) {
					CheckMenuRadioItem(GetSubMenu(hMenu, 12), 0x00, 0x3f, nAddrSegDisAsm, MF_BYPOSITION);
				}
				CheckMenuRadioItem(hMenu, 10, 12, 10 + nAddrModeDisAsm, MF_BYPOSITION);
			}
#endif

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			PosDisAsmPoint = point;
			ClientToScreen(hWnd, &point);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* NbNE_uNbN */
		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* _AhX[hȊO͖ */
#if XM7_VER == 1 && defined(Z80CARD)
			if (nCPU == MAINZ80) {
				if (nAddrModeDisAsmZ80 != 0) {
					return FALSE;
				}
			}
			else {
				if (nAddrModeDisAsm != 0) {
					return FALSE;
				}
			}
#else
			if (nAddrModeDisAsm != 0) {
				return FALSE;
			}
#endif

			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			LButtonDisAsm(hWnd, point, FALSE);
			break;

		/* R}h */
		case WM_COMMAND:
			CmdDisAsm(hWnd, LOWORD(wParam), nCPU);
			break;

		/* XN[o[ */
		case WM_VSCROLL:
			/* ^Cv */
			wAddr = wDisAsm[nCPU];

			/* ANV */
			switch (LOWORD(wParam)) {
				/* gbv */
				case SB_TOP:
					wAddr = 0;
					break;
				/* I[ */
				case SB_BOTTOM:
					wAddr = 0xffff;
					break;
				/* Ps */
				case SB_LINEUP:
					if (wAddr > 0) {
						wAddr--;
					}
					break;
				/* Ps(͍Hv) */
				case SB_LINEDOWN:
#if XM7_VER == 1 && defined(Z80CARD)
					if (nCPU == MAINZ80) {
						ret = disline80(nCPU, wAddr, string, nAddrBnkDisAsmZ80, nAddrModeDisAsmZ80);
					}
					else {
						ret = disline(nCPU, wAddr, string, nAddrSegDisAsm, nAddrModeDisAsm);
					}
#else
					ret = disline(nCPU, wAddr, string, nAddrSegDisAsm, nAddrModeDisAsm);
#endif
					i = (int)wAddr;
					i += ret;
					if (i < 0x10000) {
						wAddr += (WORD)ret;
					}
					break;
				/* y[WAbv */
				case SB_PAGEUP:
					if (wAddr < 0x80) {
						wAddr = 0;
					}
					else {
						wAddr -= (WORD)0x80;
					}
					break;
				/* y[W_E */
				case SB_PAGEDOWN:
					if (wAddr >= 0xff80) {
						wAddr = 0xffff;
					}
					else {
						wAddr += (WORD)0x80;
					}
					break;
				/* ڎw */
				case SB_THUMBTRACK:
					wAddr = HIWORD(wParam);
					if (wAddr >= 0xffff) {
						wAddr = 0xffff;
					}
					break;
			}
			AddrDisAsm(nCPU, wAddr);
			RefreshDisAsm();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	tAZuEChE(C)
 *	EChEvV[W
 */
static LRESULT CALLBACK DisAsmProcMain(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return DisAsmProc(hWnd, message, wParam, lParam, MAINCPU);
}

/*
 *	tAZuEChE(Tu)
 *	EChEvV[W
 */
static LRESULT CALLBACK DisAsmProcSub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return DisAsmProc(hWnd, message, wParam, lParam, SUBCPU);
}

#ifdef JSUB
/*
 *	tAZuEChE({Tu)
 *	EChEvV[W
 */
static LRESULT CALLBACK DisAsmProcJsub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return DisAsmProc(hWnd, message, wParam, lParam, JSUBCPU);
}
#endif

#if XM7_VER == 1
#ifdef Z80CARD
/*
 *	tAZuEChE(CZ80)
 *	EChEvV[W
 */
static LRESULT CALLBACK DisAsmProcZ80(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return DisAsmProc(hWnd, message, wParam, lParam, MAINZ80);
}
#endif
#endif

/*
 *	tAZuEChE
 *	EChE쐬
 */
HWND FASTCALL CreateDisAsm(HWND hParent, BYTE nCPU, int index)
{
	WNDCLASSEX wcex;
	char szClassNameMain[] = "XM7_DisAsmMain";
	char szClassNameSub[] = "XM7_DisAsmSub";
#ifdef JSUB
	char szClassNameJsub[] = "XM7_DisAsmJsub";
#endif
#if XM7_VER == 1
#ifdef Z80CARD
	char szClassNameZ80[] = "XM7_DisAsmZ80";
#endif
#endif
	char *szClassName;
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	SCROLLINFO si;
	UINT id;
	WORD pc;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 47;

	/* EChE^CgAobt@mہAj[h */
	switch (nCPU) {
		case MAINCPU :	id = IDS_SWND_DISASM_MAIN;
						pc = maincpu.pc;
						szClassName = szClassNameMain;
						break;
		case SUBCPU :	id = IDS_SWND_DISASM_SUB;
						pc = subcpu.pc;
						szClassName = szClassNameSub;
						break;
#ifdef JSUB
		case JSUBCPU :	id = IDS_SWND_DISASM_JSUB;
						pc = jsubcpu.pc;
						szClassName = szClassNameJsub;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	id = IDS_SWND_DISASM_Z80;
						pc = (WORD)mainz80.pc;
						szClassName = szClassNameZ80;
						break;
#endif
#endif
	}
	rect.bottom = lCharHeight * nHeightDisAsm[nCPU];
	LoadString(hResInstance, id, szWndName, sizeof(szWndName));
	pDisAsm[nCPU] = malloc(2 * (rect.right / lCharWidth) * nHeightDisAsm[nCPU]);
	wDisAsm[nCPU] = pc;
#if XM7_VER == 1 && defined(Z80CARD)
	if (nCPU == MAINZ80) {
		hDisAsm[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_DISASMZ80MENU));
	}
	else {
		hDisAsm[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_DISASMMENU));
	}
#else
	hDisAsm[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_DISASMMENU));
#endif

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	switch (nCPU) {
		case MAINCPU :	wcex.lpfnWndProc = DisAsmProcMain;
						break;
		case SUBCPU :	wcex.lpfnWndProc = DisAsmProcSub;
						break;
#ifdef JSUB
		case JSUBCPU :	wcex.lpfnWndProc = DisAsmProcJsub;
						break;
#endif
#if XM7_VER == 1
#ifdef Z80CARD
		case MAINZ80 :	wcex.lpfnWndProc = DisAsmProcZ80;
						break;
#endif
#endif
	}
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_VSCROLL | WS_SIZEBOX;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS
					| WS_VSCROLL | WS_SIZEBOX;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);

		/* XN[o[̐ݒ肪Kv */
		memset(&si, 0, sizeof(si));
		si.cbSize = sizeof(si);
		si.fMask = SIF_ALL;
		si.nMin = 0;
		si.nMax = 0x100fe;
		si.nPage = 0x100;
		si.nPos = pc;
		SetScrollInfo(hWnd, SB_VERT, &si, TRUE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

HWND FASTCALL CreateDisAsmMain(HWND hParent, int index)
{
	return	CreateDisAsm(hParent, MAINCPU, index);
}

HWND FASTCALL CreateDisAsmSub(HWND hParent, int index)
{
	return	CreateDisAsm(hParent, SUBCPU, index);
}

#ifdef JSUB
HWND FASTCALL CreateDisAsmJSub(HWND hParent, int index)
{
	return	CreateDisAsm(hParent, JSUBCPU, index);
}
#endif

#if XM7_VER == 1 && defined(Z80CARD)
HWND FASTCALL CreateDisAsmZ80(HWND hParent, int index)
{
	return	CreateDisAsm(hParent, MAINZ80, index);
}
#endif

/*-[ eύX_CAO ]----------------------------------------------*/

/*
 *	eύX_CAO
 *	_CAO
 */
static BOOL FASTCALL MemoryChangeDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	char tmp[16];
	char string[128];
	UINT id;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* AhXݒ */
	hWnd = GetDlgItem(hDlg, IDC_MEMLABEL);
	ASSERT(hWnd);
	switch (nMemDlgCPU) {
		case MAINCPU :	id = IDS_SWND_CPU_MAIN;
						break;
		case SUBCPU :	id = IDS_SWND_CPU_SUB;
						break;
#ifdef JSUB
		case JSUBCPU :	id = IDS_SWND_CPU_JSUB;
						break;
#endif
	}
	LoadString(hResInstance, id, tmp, sizeof(tmp));
	_snprintf(string, sizeof(string), "$%04X (%s)", wMemDlgAddr, tmp);
	SetWindowText(hWnd, string);

	/* f[^ݒ */
	hWnd = GetDlgItem(hDlg, IDC_MEMEDIT);
	ASSERT(hWnd);
	_snprintf(string, sizeof(string), "%02X", nMemDlgByte);
	SetWindowText(hWnd, string);

	return TRUE;
}

/*
 *	eύX_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK MemoryChangeDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	HWND hWnd;
	char string[128];

	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return MemoryChangeDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						/* ݂̒l擾 */
						hWnd = GetDlgItem(hDlg, IDC_MEMEDIT);
						ASSERT(hWnd);
						GetWindowText(hWnd, string, sizeof(string) - 1);
						if (string[0] == 0x27) {
							nMemDlgByte = (BYTE)string[1];
						}
						else {
							nMemDlgByte = (BYTE)strtol(string, NULL, 16);
						}
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*
 *	eύX
 */
static BOOL FASTCALL MemoryChange(HWND hWnd, BYTE nCPU, int x, int y)
{
	BYTE (FASTCALL *readbnio)(WORD addr);
	void (FASTCALL *writeb)(WORD addr, BYTE dat);
	int height;
	int ret;

	ASSERT(hWnd);

	/* p[^ۑEݒ */
	switch (nCPU) {
		case MAINCPU :	readbnio = mainmem_readbnio;
						writeb = mainmem_writeb;
						break;
		case SUBCPU :	readbnio = submem_readbnio;
						writeb = submem_writeb;
						break;
#ifdef JSUB
		case JSUBCPU :	readbnio = jsubmem_readbnio;
						writeb = jsubmem_writeb;
						break;
#endif
	}
	ASSERT(readbnio);
	ASSERT(writeb);

	nMemDlgCPU = nCPU;
	height = nHeightMemory[nCPU];
	if (bAsciiDump[nCPU]) {
		height -= 3;
	}

	/* AhXZbg */
	if ((x < 6) || (x > 53) || ((x % 3) == 2) ||
		(y < 0) || (y >= height)) {
		/* W͈͊O */
		return FALSE;
	}
	wMemDlgAddr = (WORD)(wMemory[nCPU] & 0xfff0);
	wMemDlgAddr += (WORD)((x - 6) / 3);
	wMemDlgAddr += (WORD)(y * 16);

	/* f[^Zbg */
	if ((nCPU == MAINCPU) && (nAddrModeDump != 0)) {
		nMemDlgByte = mainmem_readbsegnio(wMemDlgAddr, nAddrSegDump, nAddrModeDump);
	}
	else {
		nMemDlgByte = readbnio(wMemDlgAddr);
	}

	/* [__CAOs */
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_MEMDLG), hWnd, MemoryChangeDlgProc);
	if (ret != IDOK) {
		return FALSE;
	}

	/* f[^Zbg */
	LockVM();
	if ((nCPU == MAINCPU) && (nAddrModeDump != 0)) {
		mainmem_writebseg(wMemDlgAddr, (BYTE)nMemDlgByte, nAddrSegDump, nAddrModeDump);
	}
	else {
		writeb(wMemDlgAddr, (BYTE)nMemDlgByte);
	}
	InvalidateRect(hWnd, NULL, FALSE);
	UnlockVM();

	return TRUE;
}

/*-[ _CAO ]--------------------------------------------------*/

/*
 *	_CAO
 *	_CAO
 */
static BOOL FASTCALL MemorySearchDlgInit(HWND hDlg)
{
	HWND hWnd;
	RECT prect;
	RECT drect;
	char string[128];
	WORD addr;

	ASSERT(hDlg);

	/* eEChE̒ɐݒ */
	hWnd = GetParent(hDlg);
	GetWindowRect(hWnd, &prect);
	GetWindowRect(hDlg, &drect);
	drect.right -= drect.left;
	drect.bottom -= drect.top;
	drect.left = (prect.right - prect.left) / 2 + prect.left;
	drect.left -= (drect.right / 2);
	drect.top = (prect.bottom - prect.top) / 2 + prect.top;
	drect.top -= (drect.bottom / 2);
	MoveWindow(hDlg, drect.left, drect.top, drect.right, drect.bottom, FALSE);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_SEARCHADDR);
	ASSERT(hWnd);
	InsertHistory(hWnd);

	/* AhXݒ */
	addr = wMemSrchAddrSave[nMemSrchCPU];
	_snprintf(string, sizeof(string), "%04X", addr);
	ComboBox_SetText(hWnd, string);

	/* ݒ */
	hWnd = GetDlgItem(hDlg, IDC_SEARCHDATA);
	ASSERT(hWnd);
	SetWindowText(hWnd, (LPSTR)szMemSrchString);

	return TRUE;
}

/*
 *	_CAO
 *	_CAOOK
 */
static void FASTCALL MemorySearchDlgOK(HWND hDlg)
{
	HWND hWnd;
	char string[128];

	ASSERT(hDlg);

	/* ݂̒l擾 */
	hWnd = GetDlgItem(hDlg, IDC_SEARCHDATA);
	ASSERT(hWnd);
	GetWindowText(hWnd, szMemSrchString, sizeof(szMemSrchString) - 1);

	/* R{{bNX */
	hWnd = GetDlgItem(hDlg, IDC_SEARCHADDR);
	ASSERT(hWnd);

	/* ݂̒l擾 */
	ComboBox_GetText(hWnd, string, sizeof(string) - 1);
	wMemSrchAddr = (WORD)strtol(string, NULL, 16);

	/* qXgǉ */
	AddAddrHistory(wMemSrchAddr);
}

/*
 *	_CAO
 *	_CAOvV[W
 */
static BOOL CALLBACK MemorySearchDlgProc(HWND hDlg, UINT iMsg,
									WPARAM wParam, LPARAM lParam)
{
	UNUSED(lParam);

	switch (iMsg) {
		/* _CAO */
		case WM_INITDIALOG:
			return MemorySearchDlgInit(hDlg);

		/* R}h */
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				/* OKELZ */
				case IDOK:
				case IDCANCEL:
					WAIT_KEY_POP();
					if (LOWORD(wParam) == IDOK) {
						MemorySearchDlgOK(hDlg);
					}
					EndDialog(hDlg, LOWORD(wParam));
					InvalidateRect(hDrawWnd, NULL, FALSE);
					return TRUE;
			}
			break;
	}

	/* ȊÓAFALSE */
	return FALSE;
}

/*
 *	
 */
static BOOL FASTCALL MemorySearch(HWND hWnd, BYTE nCPU)
{
	BYTE (FASTCALL *readbnio)(WORD addr);
	BYTE searchdata[128];
	char string[128];
	char tmp[128];
	char *p;
	char *q;
	BOOL flag;
	BYTE dat;
	int datasize;
	int i;
	int j;
	int ret;

	ASSERT(hWnd);

	/* [__CAOs */
	nMemSrchCPU = nCPU;
	ret = DialogBox(hResInstance, MAKEINTRESOURCE(IDD_SEARCHDLG), hWnd, MemorySearchDlgProc);
	if (ret != IDOK) {
		return FALSE;
	}

	/* [Nݒ */
	switch (nCPU) {
		case MAINCPU :	readbnio = mainmem_readbnio;
						break;
		case SUBCPU :	readbnio = submem_readbnio;
						break;
#ifdef JSUB
		case JSUBCPU :	readbnio = jsubmem_readbnio;
						break;
#endif
	}
	wMemSrchAddrSave[nCPU] = wMemSrchAddr;
	memset(searchdata, 0, sizeof(searchdata));
	memset(string, 0, sizeof(string));
	memset(tmp, 0, sizeof(tmp));

	/* ϊ */
	if (szMemSrchString[0] == '#') {
		/* 16iƂċ镶̂ݔo */
		p = &szMemSrchString[1];
		q = string;
		while (*p != '\0') {
			if ((	((*p >= '0') && (*p <= '9')) ||
					((*p >= 'A') && (*p <= 'F')) ||
					((*p >= 'a') && (*p <= 'f')))) {
				*q++ = *p;
			}
			p++;
		}

		/* f[^֕ϊ */
		datasize = 0;
		p = string;

		while (*p != '\0') {
			memcpy(tmp, p, 2);
			searchdata[datasize ++] = (BYTE)strtol(tmp, NULL, 16);
			p += 2;
		}
	}
	else {
		/* w */
		datasize = strlen(szMemSrchString);
		memcpy(searchdata, szMemSrchString, datasize);
	}

	/* f[^ȂꍇȂ */
	if (datasize == 0) {
		return FALSE;
	}

	/* f[^ */
	flag = FALSE;
	for (i=wMemSrchAddr; i<65536 - datasize; i++) {
		if ((nCPU == MAINCPU) && (nAddrModeDump != 0)) {
			dat = mainmem_readbsegnio((WORD)i, nAddrSegDump, nAddrModeDump);
		}
		else {
			dat = readbnio((WORD)i);
		}

		/* 擪f[^vꍇɎc̃f[^`FbN*/
		if (searchdata[0] == dat) {
			for (j=1; j<datasize; j++) {
				if ((nCPU == MAINCPU) && (nAddrModeDump != 0)) {
					dat = mainmem_readbsegnio((WORD)(i + j), nAddrSegDump, nAddrModeDump);
				}
				else {
					dat = readbnio((WORD)(i + j));
				}
				if (searchdata[j] != dat) {
					break;
				}
			}
			if (j == datasize) {
				/*  */
				LoadString(hResInstance, IDS_SWND_SRCH_FOUND, tmp, sizeof(tmp));
				_snprintf(string, sizeof(string), tmp, i);
				flag = TRUE;
				break;
			}
		}
	}

	if (!flag) {
		/* ł */
		LoadString(hResInstance, IDS_SWND_SRCH_NOTFOUND, string, sizeof(string));
	}

	/* ʂ\ */
	keyboard_inputmask(TRUE);
	MessageBox(hMainWnd, string, "XM7", MB_OK | MB_ICONINFORMATION);
	WAIT_KEY_POP();
	keyboard_inputmask(FALSE);
	SetMenuExitTimer();

	if (flag) {
		/* łꍇ͂̃AhXɔ΂ */
		AddrMemory(nCPU, (WORD)i);

		/* JnAhXXV */
		wMemSrchAddrSave[nCPU] = (WORD)(i + datasize);
	}

	return TRUE;
}

/*-[ _vEChE ]------------------------------------------------*/

/*
 *	_vEChE
 *	AhXݒ
 */
void FASTCALL AddrMemory(BYTE nCPU, WORD wAddr)
{
	SCROLLINFO sif;

	wAddr &= (WORD)(0xfff0);

	memset(&sif, 0, sizeof(sif));
	sif.cbSize = sizeof(sif);
	sif.fMask = SIF_POS;
	sif.nPos = (int)(wAddr >> 4);

	/* ݃`FbNݒ */
	wMemory[nCPU] = wAddr;
	switch (nCPU) {
		case MAINCPU :	if (hSubWnd[SWND_MEMORY_MAIN]) {
							SetScrollInfo(hSubWnd[SWND_MEMORY_MAIN], SB_VERT,
								&sif, TRUE);
						}
						break;
		case SUBCPU :	if (hSubWnd[SWND_MEMORY_SUB]) {
							SetScrollInfo(hSubWnd[SWND_MEMORY_SUB], SB_VERT,
								&sif, TRUE);
						}
						break;
#ifdef JSUB
		case JSUBCPU :	if (hSubWnd[SWND_MEMORY_JSUB]) {
							SetScrollInfo(hSubWnd[SWND_MEMORY_JSUB], SB_VERT,
								&sif, TRUE);
						}
						break;
#endif
	}

	/* tbV */
	RefreshMemory();
}

/*
 *	_vEBhE
 *	TCY
 */
static void FASTCALL ResizeMemory(HWND hWnd, WORD cx, WORD cy, BYTE nCPU)
{
	RECT crect;
	RECT wrect;
	SCROLLINFO si;
	int width;
	int height;

	ASSERT(hWnd);

	/* ŏ̏ꍇ́AȂ */
	if ((cx == 0) && (cy == 0)) {
		return;
	}

	/* cTCYύX */
	nHeightMemory[nCPU] = (WORD)(cy / lCharHeight);
	if (nHeightMemory[nCPU] < 1) {
		nHeightMemory[nCPU] = 1;
	}

	/* TCYύX */
	if (bAsciiDump[nCPU]) {
		width = 75;
	}
	else {
		width = 53;
	}

	/* Ċm */
	free(pMemory[nCPU]);
	pMemory[nCPU] = malloc(2 * width * nHeightMemory[nCPU]);
	ASSERT(pMemory[nCPU]);
	memset(pMemory[nCPU], 0xff, 2 * width * nHeightMemory[nCPU]);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += (width * lCharWidth);
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += (nHeightMemory[nCPU] * lCharHeight);
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);

	/* XN[o[̐ݒ */
	memset(&si, 0, sizeof(si));
	height = nHeightMemory[nCPU];
	if (bAsciiDump[nCPU]) {
		height -= 3;
	}
	si.cbSize = sizeof(si);
	si.fMask = SIF_ALL;
	si.nMin = 0;
	si.nMax = 0x100f - height;
	si.nPage = 0x10;
	si.nPos = (wMemory[nCPU] >> 4);
	if (si.nPos > si.nMax) {
		si.nPos = si.nMax;
	}
	SetScrollInfo(hWnd, SB_VERT, &si, TRUE);
}

/*
 *	_vEBhE
 *	TCY(sw)
 */
static void FASTCALL ResizeMemoryLines(HWND hWnd, int lines, BYTE nCPU)
{
	RECT rect;

	GetWindowRect(hWnd, &rect);
	ResizeMemory(hWnd, 	(WORD)(rect.right - rect.left + 1),
						(WORD)(lCharHeight * lines), nCPU);
}

/*
 *	_vEChE
 *	ZbgAbv
 */
static void FASTCALL SetupMemory(BYTE nCPU, BYTE *p, int x, int y)
{
	BYTE (FASTCALL *readbnio)(WORD addr);
	char string[128];
	char temp[32];
	WORD addr;
	int i;
	int j;
	int height;
	BYTE mem[16];
	BYTE tsum[16];
	BYTE sum;
	BYTE knjsave;
	int offset;
	BOOL isj;

	ASSERT(nCPU < MAXCPU);
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);
	memset(tsum, 0, 16);

	/* [N */
	knjsave = 0;
	isj = isJapanese;

	/* AhX擾 */
	addr = (WORD)(wMemory[nCPU] & 0xfff0);
	height = nHeightMemory[nCPU];
	switch (nCPU) {
		case MAINCPU :	readbnio = mainmem_readbnio;
						break;
		case SUBCPU :	readbnio = submem_readbnio;
						break;
#ifdef JSUB
		case JSUBCPU :	readbnio = jsubmem_readbnio;
						break;
#endif
	}

	/* LvVXV */
	if (nCPU == MAINCPU) {
		switch (nAddrModeDump) {
		case 1:
			_snprintf(string, sizeof(string), " [$%05X - $%05X]",
						(nAddrSegDump << 16), (nAddrSegDump << 16) + 0xffff);
			break;
		case 2:
			_snprintf(string, sizeof(string), " [Segment:$%02X]", nAddrSegDump);
			break;
		default:
		case 0:
			memset(string, 0, sizeof(string));
			break;
		}
		if (strncmp(szMemoryWndName, string, sizeof(szMemoryWndName))) {
			strncpy(szMemoryWndName, string, sizeof(szMemoryWndName));
			LoadString(hResInstance, IDS_SWND_MEMORY_MAIN, string, sizeof(string));
			strncat(string, szMemoryWndName, sizeof(string));
			SetWindowText(hSubWnd[SWND_MEMORY_MAIN], string);
		}
	}

	/* ASCII_v[h 㕔KCh */
	if (bAsciiDump[nCPU]) {
		strncpy(string, "Addr ", sizeof(string));
		for (j=0; j<16; j++) {
			_snprintf(temp, sizeof(temp), " +%01X", j);
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}
		strncat(string, " Sum", sizeof(string) - strlen(string) - 1);

		/* Rs[ */
		memcpy(p, string, strlen(string));

		offset = 1;
		height -= 3;
	}
	else {
		offset = 0;
	}

	/* [v */
	for (i=0; i<height; i++) {
		/* 쐬 */
		_snprintf(string, sizeof(string), "%04X:", addr);
		sum = 0;
		for (j=0; j<16; j++) {
			if ((nCPU == MAINCPU) && (nAddrModeDump != 0)) {
				mem[j] = mainmem_readbsegnio((WORD)(addr + j), nAddrSegDump, nAddrModeDump);
			}
			else {
				mem[j] = readbnio((WORD)(addr + j));
			}
			_snprintf(temp, sizeof(temp), " %02X", mem[j]);
			tsum[j] += mem[j];
			sum += mem[j];
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		if (bAsciiDump[nCPU]) {
			/* `FbNTEKCh */
			_snprintf(temp, sizeof(temp), " :%02X  ................", sum);
			strncat(string, temp, sizeof(string) - strlen(string) - 1);

			/* R[h2sɂ܂ꍇ̏ */
			if (isj && isKanji(knjsave)) {
				string[58] = knjsave;
				knjsave = 1;
			}

			/* ASCII_v */
			for (j=0; j<16; j++) {
				if ((mem[j] < 0x20) || (mem[j] == 0x7f) || (mem[j] >= 0xfd)) {
					/* Rg[R[h(\) */
					knjsave = 0;
				}
				else if (!isj && (mem[j] & 0x80)) {
					/* {ꃂ[h 0x80`0xFF(\) */
					knjsave = 0;
				}
				else if (bKanjiDump[nCPU] && isKanji(mem[j]) && (knjsave == 0)) {
					/*  */

					if (j == 15) {
						/* sɂ܂ꍇ͔\ */
						string[j + 59] = ' ';
						knjsave = mem[j];
					}
					else {
						if (isKanji2(mem[j + 1])) {
							/* 2oCgƂȂꍇ */
							string[j + 59] = mem[j];
							knjsave = 1;
						}
						else {
							/* R[hُ(\) */
							knjsave = 0;
						}
					}
				}
				else {
					if ((mem[j] == 0xa0) && (knjsave == 0)) {
						/* f[^$A0ŃVtgJIS2oCgłȂꍇ */
						string[j + 59] = ' ';
					}
					else if ((!isKanji(mem[j]) && (mem[j] != 0x80)) || (knjsave == 1)) {
						/* ANK ܂ VtgJIS2oCg */
						string[j + 59] = mem[j];
					}
					knjsave = 0;
				}
			}
		}

		/* Rs[ */
		memcpy(&p[x * (i + offset)], string, strlen(string));

		/*  */
		addr += (WORD)0x0010;
		if (addr == 0) {
			break;
		}
	}

	/* ASCII_v[h rEcTEubNT */
	if (bAsciiDump[nCPU]) {
		/* r쐬 */
		memset(string, '-', 75);
		memset(&string[57], ' ', 2);
		string[76] = '\0';

		/* Rs[ */
		memcpy(&p[x * (height + 1)], string, strlen(string));

		/* cT쐬 */
		_snprintf(string, sizeof(string), "Sum :", addr);
		sum = 0;
		for (j=0; j<16; j++) {
			_snprintf(temp, sizeof(temp), " %02X", tsum[j]);
			sum += tsum[j];
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		/* `FbNT */
		_snprintf(temp, sizeof(temp), " :%02X", sum);
		strncat(string, temp, sizeof(string) - strlen(string) - 1);

		/* Rs[ */
		memcpy(&p[x * (height + 2)], string, strlen(string));
	}
}

/*
 *	_vEChE
 *	`
 */
static void FASTCALL DrawMemory(HWND hWnd, HDC hDC)
{
	int i;
	BYTE nCPU;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			nCPU = (BYTE)((i - SWND_MEMORY_MAIN) % MAXCPU);
			p = pMemory[nCPU];
			break;
		}
	}
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	SetupMemory(nCPU, p, x, y);

	/* ` */
	if (bKanjiDump[nCPU] && isJapanese) {
		DrawWindowTextKanji(hDC, p, x, y);
	}
	else {
		DrawWindowText(hDC, p, x, y);
	}
}

/*
 *	_vEChE
 *	tbV
 */
void FASTCALL RefreshMemory(void)
{
	HWND hWnd;
	HDC hDC;

	/* CCPU */
	if (hSubWnd[SWND_MEMORY_MAIN]) {
		hWnd = hSubWnd[SWND_MEMORY_MAIN];
		hDC = GetDC(hWnd);
		DrawMemory(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

	/* TuCPU */
	if (hSubWnd[SWND_MEMORY_SUB]) {
		hWnd = hSubWnd[SWND_MEMORY_SUB];
		hDC = GetDC(hWnd);
		DrawMemory(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

#ifdef JSUB
	/* {TuCPU */
	if (hSubWnd[SWND_MEMORY_JSUB]) {
		hWnd = hSubWnd[SWND_MEMORY_JSUB];
		hDC = GetDC(hWnd);
		DrawMemory(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
#endif
}

/*
 *	_vEChE
 *	ĕ`
 */
static void FASTCALL PaintMemory(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	int i;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
#if XM7_VER == 1 && defined(Z80CARD)
			if (((i - SWND_CPUREG_MAIN) % MAXCPU) == MAINZ80) {
				continue;
			}
#endif
			p = pMemory[((i - SWND_MEMORY_MAIN) % MAXCPU)];
			break;
		}
	}

	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawMemory(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	_vEChE
 *	R}h
 */
static void FASTCALL CmdMemory(HWND hWnd, WORD wID, BYTE nCPU)
{
	WORD target;
	cpu6809_t *cpu;
	BYTE height;
#if XM7_VER == 1 && defined(Z80CARD)
	WORD tmp;
#endif

	/* CPU\̌ */
	switch (nCPU) {
		case MAINCPU :	cpu = &maincpu;
						break;
		case SUBCPU :	cpu = &subcpu;
						break;
#ifdef JSUB
		case JSUBCPU :	cpu = &jsubcpu;
						break;
#endif
	}

	/* ^[QbgAhX */
	switch (wID) {
		case IDM_DMP_ADDR:
			target = wMemory[nCPU];
			if (!AddrDlg(hWnd, &target)) {
				return;
			}
			break;

		case IDM_DMP_BLK128:
		case IDM_DMP_BLK256:
			if (wID == IDM_DMP_BLK128) {
				height = 8;
			}
			else if (wID == IDM_DMP_BLK256) {
				height = 16;
			}
			if (bAsciiDump[nCPU]) {
				height += (BYTE)3;
			}
			ResizeMemoryLines(hWnd, height, nCPU);
			return;

		case IDM_DMP_ASCII:
			bAsciiDump[nCPU] = !bAsciiDump[nCPU];

			/* cTCYύX */
			height = nHeightMemory[nCPU];
			if (bAsciiDump[nCPU]) {
				height += (BYTE)3;
			}
			else {
				height -= (BYTE)3;
			}
			ResizeMemoryLines(hWnd, height, nCPU);
			PaintMemory(hWnd);
			return;

		case IDM_DMP_KANJI:
			bKanjiDump[nCPU] = !bKanjiDump[nCPU];
			InvalidateRect(hWnd, NULL, FALSE);
			return;

		case IDM_DMP_SEARCH:
			MemorySearch(hWnd, nCPU);
			return;

		case IDM_DMP_PC:
			target = cpu->pc;
			break;
		case IDM_DMP_X:
			target = cpu->x;
			break;
		case IDM_DMP_Y:
			target = cpu->y;
			break;
		case IDM_DMP_U:
			target = cpu->u;
			break;
		case IDM_DMP_S:
			target = cpu->s;
			break;
		case IDM_DMP_DP:
			target = (WORD)(cpu->dp << 8);
			break;

#if XM7_VER == 1 && defined(Z80CARD)
		case IDM_DMP_PC_Z80:
			target = (WORD)mainz80.pc;
			break;
		case IDM_DMP_BC:
			target = (WORD)((mainz80.regs8[REGID_B]) << 8 |
					mainz80.regs8[REGID_C]);
			break;
		case IDM_DMP_DE:
			target = (WORD)((mainz80.regs8[REGID_D]) << 8 |
				mainz80.regs8[REGID_E]);
			break;
		case IDM_DMP_HL:
			target = (WORD)((mainz80.regs8[REGID_H]) << 8 |
				mainz80.regs8[REGID_L]);
			break;
		case IDM_DMP_IX:
			target = (WORD)((mainz80.regs8[REGID_IXH]) << 8 |
				mainz80.regs8[REGID_IXL]);
			break;
		case IDM_DMP_IY:
			target = (WORD)((mainz80.regs8[REGID_IYH]) << 8 |
				mainz80.regs8[REGID_IYL]);
			break;
		case IDM_DMP_SP:
			target = mainz80.sp;
			break;
#endif

		case IDM_DMP_RESET:
			target = (WORD)((cpu->readmem(0xfffe) << 8) | cpu->readmem(0xffff));
			break;
		case IDM_DMP_NMI:
			target = (WORD)((cpu->readmem(0xfffc) << 8) | cpu->readmem(0xfffd));
			break;
		case IDM_DMP_SWI:
			target = (WORD)((cpu->readmem(0xfffa) << 8) | cpu->readmem(0xfffb));
			break;
		case IDM_DMP_IRQ:
			target = (WORD)((cpu->readmem(0xfff8) << 8) | cpu->readmem(0xfff9));
			break;
		case IDM_DMP_FIRQ:
			target = (WORD)((cpu->readmem(0xfff6) << 8) | cpu->readmem(0xfff7));
			break;
		case IDM_DMP_SWI2:
			target = (WORD)((cpu->readmem(0xfff4) << 8) | cpu->readmem(0xfff5));
			break;
		case IDM_DMP_SWI3:
			target = (WORD)((cpu->readmem(0xfff2) << 8) | cpu->readmem(0xfff3));
			break;
		case IDM_DMP_IO:
			switch (nCPU) {
				case MAINCPU :	target = 0xfd00;
								break;
				case SUBCPU :	target = 0xd400;
								break;
#ifdef JSUB
				case JSUBCPU :	target = 0x9f00;
								break;
#endif
				default :		target = 0x0000;
			}
			break;

		case IDM_DMP_LOGIC:
			nAddrSegDump = 0;							/* _vZOg */
			nAddrModeDump = 0;							/* _vAhX[h */
			return;
		case IDM_DMP_COPY:
			if (bAsciiDump[nCPU]) {
				SetClipboardSubWindow(hWnd, pMemory[nCPU], 75, nHeightMemory[nCPU], FALSE);
			}
			else {
				SetClipboardSubWindow(hWnd, pMemory[nCPU], 53, nHeightMemory[nCPU], FALSE);
			}
			return;
		default:
			if ((wID >= IDM_DMP_STACK) && (wID <= IDM_DMP_STACK + 15)) {
				target = (WORD)wStackAddr[wID - IDM_DMP_STACK];
			}
			else if ((wID >= IDM_DMP_VECTOR) && (wID <= IDM_DMP_VECTOR + 15)) {
				target = (WORD)wVectorAddr[wID - IDM_DIS_VECTOR];
			}
			else {
				target = wMemory[nCPU];
				if ((wID >= IDM_DMP_PADR0) && (wID <= IDM_DMP_PADR0 + 0x0f)) {
					nAddrSegDump = (wID - IDM_DMP_PADR0);	/* _vZOg */
					nAddrModeDump = 1;					/* _vAhX[h */
				}
				else if ((wID >= IDM_DMP_SEG00) && (wID <= IDM_DMP_MAX)) {
					nAddrSegDump = (wID - IDM_DMP_SEG00);	/* _vZOg */
					nAddrModeDump = 2;					/* _vAhX[h */
				}
				else if ((wID >= IDM_DMP_LADR0) && (wID <= IDM_DMP_LADR0 + 0x0f)) {
					target = (wID - IDM_DMP_LADR0) << 12;
				}
			}
			break;
	}

	/* ݒ聕XV */
	AddrMemory(nCPU, target);
}

/*
 *	_vEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK MemoryProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam, BYTE nCPU)
{
	static BOOL bResize = FALSE;		/* 蓮TCYtO */
	HMENU hMenu;
	POINT point;
	WORD wAddr;
	int height;
	UINT idcheck;
	RECT rect;
	BOOL bpysic;
	int i;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintMemory(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pMemory[nCPU], hMemory[nCPU]);

			UnlockVM();
			break;

		/* EChETCYύX */
		case WM_SIZE:
			if (bResize) {
				bResize = FALSE;
				ResizeMemory(hWnd, LOWORD(lParam), HIWORD(lParam), nCPU);
			}
			break;

		/* EChETCYύX */
		case WM_SIZING:
			WindowSizing(hWnd, (LPRECT)lParam, &pMemory[nCPU]);
			bResize = TRUE;
			break;

		/* EBhETCYύXbZ[W */
		case WM_WINDOWPOSCHANGING:
			bResize = TRUE;
			break;

		/* ŏTCY */
		case WM_GETMINMAXINFO:
			rect.left = 0;
			rect.right = 53 * lCharWidth + GetSystemMetrics(SM_CXVSCROLL);
			rect.top = 0;
			rect.bottom = lCharHeight;

			/* ASCII_v[h̏cTXy[Xl */
			if (bAsciiDump[nCPU]) {
				rect.right += (22 * lCharWidth);
				rect.bottom += (lCharHeight * 3);
			}

			if (bPopupSwnd) {
				AdjustWindowRect(&rect, WS_POPUP | WS_OVERLAPPED | WS_SYSMENU |
										WS_CAPTION | WS_VISIBLE |
										WS_MINIMIZEBOX | WS_SIZEBOX, FALSE);
			}
			else {
				AdjustWindowRect(&rect, WS_CHILD | WS_OVERLAPPED | WS_SYSMENU |
										WS_CAPTION | WS_VISIBLE |
										WS_MINIMIZEBOX | WS_CLIPSIBLINGS |
										WS_SIZEBOX, FALSE);
			}
			((LPMINMAXINFO)lParam)->ptMinTrackSize.x = rect.right - rect.left;
			((LPMINMAXINFO)lParam)->ptMaxTrackSize.x = rect.right - rect.left;
			((LPMINMAXINFO)lParam)->ptMinTrackSize.y = rect.bottom - rect.top;
			return 0;

		/* eύX */
		case WM_LBUTTONDBLCLK:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* J[\ʒuA */
			point.x = LOWORD(lParam) / lCharWidth;
			point.y = HIWORD(lParam) / lCharHeight;
			if (bAsciiDump[nCPU]) {
				point.y --;
			}

			MemoryChange(hWnd, nCPU, point.x, point.y);
			return 0;

		/* ReLXgj[ */
		case WM_CONTEXTMENU:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (mos_capture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* Tuj[o */
			hMenu = GetSubMenu(hMemory[nCPU], 0);
			height = nHeightMemory[nCPU];
			if (bAsciiDump[nCPU]) {
				height -= 3;
			}

			/* \TCY`FbN */
			CheckMenuSub(hMenu, IDM_DMP_BLK128, FALSE);
			CheckMenuSub(hMenu, IDM_DMP_BLK256, FALSE);
			if (height == 8) {
				idcheck = IDM_DMP_BLK128;
			}
			else if (height == 16) {
				idcheck = IDM_DMP_BLK256;
			}
			else {
				idcheck = 0;
			}
			if (idcheck > 0) {
				CheckMenuRadioItem(hMenu, IDM_DMP_BLK128, IDM_DMP_BLK256, idcheck, MF_BYCOMMAND);
			}

			/* ASCII_v`FbN */
			CheckMenuSub(hMenu, IDM_DMP_ASCII, bAsciiDump[nCPU]);

			/* \`FbN */
			if (isJapanese) {
				EnableMenuSub(hMenu, IDM_DMP_KANJI, bAsciiDump[nCPU]);
				CheckMenuSub(hMenu, IDM_DMP_KANJI, bKanjiDump[nCPU]);
			}
			else {
				EnableMenuSub(hMenu, IDM_DMP_KANJI, FALSE);
				CheckMenuSub(hMenu, IDM_DMP_KANJI, FALSE);
			}

			/* VXeX^bNWvj[ݒ */
			InsertStackJumpMenu(GetSubMenu(hMenu, 10), nCPU, IDM_DMP_STACK);

			/* xN^[Wvj[ݒ */
			InsertVectorMenu(GetSubMenu(hMenu, 12), nCPU, IDM_DMP_VECTOR);

			/* AhX[h`FbN */
#if XM7_VER >= 2
			bpysic = ((nCPU == MAINCPU) && (model_type >= MODEL_FM77AV));
#else
			bpysic = ((nCPU == MAINCPU) && (model_type == MODEL_FM77));
#endif
			EnableMenuPos(hMenu, 8, !bpysic || nAddrModeDump == 0);
			EnableMenuPos(hMenu, 9, !bpysic || nAddrModeDump != 1);
			EnableMenuPos(hMenu, 10, !bpysic || nAddrModeDump == 0);
//			EnableMenuPos(hMenu, 11, !bpysic || nAddrModeDump != 1);
			EnableMenuSub(hMenu, IDM_DMP_IO, !bpysic || nAddrModeDump != 1);
			EnableMenuPos(hMenu, 12, !bpysic || nAddrModeDump != 1);
			EnableMenuPos(hMenu, 15, bpysic);
			EnableMenuPos(hMenu, 16, bpysic);
			EnableMenuPos(hMenu, 17, bpysic);
			if (bpysic) {
				/* AhXj[ݒ */
				InsertPysicAddrMenu(GetSubMenu(hMenu, 16), nCPU, IDM_DMP_PADR0);

				/* MSRj[ݒ */
				InsertSegmentMenu(GetSubMenu(hMenu, 17), nCPU, IDM_DMP_SEG00);

				if (nAddrModeDump == 1) {
					CheckMenuRadioItem(GetSubMenu(hMenu, 16), 0x00, 0x0f, nAddrSegDump, MF_BYPOSITION);
				}
				else if (nAddrModeDump == 2) {
					CheckMenuRadioItem(GetSubMenu(hMenu, 17), 0x00, 0x3f, nAddrSegDump, MF_BYPOSITION);
				}
				CheckMenuRadioItem(hMenu, 15, 17, 15 + nAddrModeDump, MF_BYPOSITION);
			}

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* R}h */
		case WM_COMMAND:
			CmdMemory(hWnd, LOWORD(wParam), nCPU);
			break;

		/* XN[o[ */
		case WM_VSCROLL:
			/* ^Cv */
			wAddr = wMemory[nCPU];
			height = nHeightMemory[nCPU];
			if (bAsciiDump[nCPU]) {
				height -= 3;
			}

			/* ANV */
			switch (LOWORD(wParam)) {
				/* gbv */
				case SB_TOP:
					wAddr = 0;
					break;
				/* I[ */
				case SB_BOTTOM:
					wAddr = (WORD)(0x10000 - height * 16);
					break;
				/* Ps */
				case SB_LINEUP:
					if (wAddr >= 0x0010) {
						wAddr -= (WORD)0x0010;
					}
					break;
				/* Ps */
				case SB_LINEDOWN:
//					if (wAddr < 0x10000 - (WORD)(height * 16)) {	/* ĈȑÕp^[Ƃ */
					if (wAddr < 0xfff0) {
						wAddr += (WORD)0x0010;
					}
					break;
				/* y[WAbv */
				case SB_PAGEUP:
					if (wAddr <= (height * 16)) {
						wAddr = 0;
					}
					else {
						wAddr -= (WORD)(height * 0x10);
					}
					break;
				/* y[W_E */
				case SB_PAGEDOWN:
//					if (wAddr >= (0x10000 - height * 16)) {
					if (wAddr >= (0xfff0 - height * 16)) {
//						wAddr = (0x10000 - height * 16);
						wAddr = 0xfff0;
					}
					else {
						wAddr += (WORD)(height * 16);
					}
					break;
				/* ڎw */
				case SB_THUMBTRACK:
					wAddr = (WORD)(HIWORD(wParam) * 16);
//					if (wAddr >= (0x1000f - (DWORD)(height * 16))) {
//						wAddr = 0x10000 - (DWORD)(height * 16);
//					}
					break;
			}
			wAddr &= (WORD)0xfff0;
			AddrMemory(nCPU, wAddr);
			RefreshMemory();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	_vEChE(C)
 *	EChEvV[W
 */
static LRESULT CALLBACK MemoryProcMain(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return MemoryProc(hWnd, message, wParam, lParam, MAINCPU);
}

/*
 *	_vEChE(Tu)
 *	EChEvV[W
 */
static LRESULT CALLBACK MemoryProcSub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return MemoryProc(hWnd, message, wParam, lParam, SUBCPU);
}

/*
 *	_vEChE({Tu)
 *	EChEvV[W
 */
#ifdef JSUB
static LRESULT CALLBACK MemoryProcJsub(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	return MemoryProc(hWnd, message, wParam, lParam, JSUBCPU);
}
#endif

/*
 *	_vEChE
 *	EChE쐬
 */
HWND FASTCALL CreateMemory(HWND hParent, BYTE nCPU, int index)
{
	WNDCLASSEX wcex;
	char szClassNameMain[] = "XM7_MemoryMain";
	char szClassNameSub[] = "XM7_MemorySub";
#ifdef JSUB
	char szClassNameJsub[] = "XM7_MemoryJsub";
#endif
	char *szClassName;
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	int height;
	SCROLLINFO si;
	UINT id;
	WORD pc;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);

	/* EChE^CgAobt@mہAj[[h */
	switch (nCPU) {
		case MAINCPU :	id = IDS_SWND_MEMORY_MAIN;
						pc = maincpu.pc;
						szClassName = szClassNameMain;
						break;
		case SUBCPU :	id = IDS_SWND_MEMORY_SUB;
						pc = subcpu.pc;
						szClassName = szClassNameSub;
						break;
#ifdef JSUB
		case JSUBCPU :	id = IDS_SWND_MEMORY_JSUB;
						pc = jsubcpu.pc;
						szClassName = szClassNameJsub;
						break;
#endif
	}
	rect.bottom = lCharHeight * nHeightMemory[nCPU];
	LoadString(hResInstance, id, szWndName, sizeof(szWndName));
	pMemory[nCPU] = malloc(2 * 75 * nHeightMemory[nCPU]);
	wMemory[nCPU] = pc;
#if XM7_VER == 1 && defined(Z80CARD)
	if (nCPU == MAINCPU) {
		hMemory[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MEMDMPMAINMENU));
	}
	else {
		hMemory[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MEMDMPMENU));
	}
#else
	hMemory[nCPU] = LoadMenu(hResInstance, MAKEINTRESOURCE(IDR_MEMDMPMENU));
#endif

	/* EBhE */
	if (bAsciiDump[nCPU]) {
		rect.right = lCharWidth * 75;
	}
	else {
		rect.right = lCharWidth * 53;
	}

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS;
	switch (nCPU) {
		case MAINCPU :	wcex.lpfnWndProc = MemoryProcMain;
						break;
		case SUBCPU :	wcex.lpfnWndProc = MemoryProcSub;
						break;
#ifdef JSUB
		case JSUBCPU :	wcex.lpfnWndProc = MemoryProcJsub;
						break;
#endif
	}
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER
					| WS_VSCROLL | WS_SIZEBOX;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS
					| WS_VSCROLL | WS_SIZEBOX;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);

		/* XN[o[̐ݒ肪Kv */
		memset(&si, 0, sizeof(si));
		si.cbSize = sizeof(si);
		si.fMask = SIF_ALL;
		si.nMin = 0;
		si.nPage = 0x10;
		height = nHeightMemory[nCPU];
		si.nPos = (pc >> 4);
		if (bAsciiDump) {
			height -= 3;
		}
		si.nMax = 0x100f - height;
		SetScrollInfo(hWnd, SB_VERT, &si, TRUE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

HWND FASTCALL CreateMemoryMain(HWND hParent, int index)
{
	return	CreateMemory(hParent, MAINCPU, index);
}

HWND FASTCALL CreateMemorySub(HWND hParent, int index)
{
	return	CreateMemory(hParent, SUBCPU, index);
}

#ifdef JSUB
HWND FASTCALL CreateMemoryJSub(HWND hParent, int index)
{
	return	CreateMemory(hParent, JSUBCPU, index);
}
#endif

/*-[ OEChE ]--------------------------------------------------------*/

/*
 *	OEChE
 *	ftHgTCY擾
 */
void FASTCALL GetDefaultSizeLog(RECT *rect, int y)
{
	HDC hDC;
	TEXTMETRIC tm;
	int i;

	/* eLXggbN擾 */
	hDC = GetDC(hMainWnd);
	GetTextMetrics(hDC, &tm);
	ReleaseDC(hMainWnd, hDC);

	memset(rect, 0, sizeof(RECT));
	for (i=0; i<countof(headertable); i++) {
		/* Ђ畽ϕTCY瓾ꂽTCY𑫂 */
		rect->right += headertable[i].cx * tm.tmAveCharWidth;
	}
	rect->bottom = (tm.tmHeight + tm.tmExternalLeading) * y;
}

/*
 *	OEChE
 *	Oݒ
 */
static void FASTCALL LogAddString(DWORD id, loglevel level, const char *string)
{
	int index;
	time_t timer;

	ASSERT(string);

	/* ݎ擾 */
	time(&timer);

	/* bN */
	LockVM();

	/* ǉʒu߂ */
	if (nLogNum < LogMax) {
		ASSERT(nLogTop == 0);
		index = nLogNum;
		ASSERT(stLogData[index] == NULL);
		nLogNum++;
	}
	else {
		ASSERT(nLogNum == LogMax);
		index = nLogTop;
		nLogTop++;
		nLogTop &= (LogMax - 1);
	}

	/* ɑ݂Ή */
	if (stLogData[index]) {
		if (stLogData[index]->string) {
			/* |C^ */
			free(stLogData[index]->string);
		}
		/* f[^{̂ */
		free(stLogData[index]);

		/* NULL */
		stLogData[index] = NULL;
	}

	/* \̂zu */
	stLogData[index] = (void *)malloc(sizeof(logdata_t));
	stLogData[index]->number = nLogCnt++;
	stLogData[index]->time = dwExecTotal;
	stLogData[index]->id = id;
	stLogData[index]->pc1 = maincpu.pc;
	stLogData[index]->pc2 = subcpu.pc;
#ifdef JSUB
	stLogData[index]->pc3 = jsubcpu.pc;
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	stLogData[index]->pc4 = (WORD)mainz80.pc;
#endif
	stLogData[index]->level = level;
	stLogData[index]->timer = timer;

	/* 񏈗 */
	stLogData[index]->string = (char *)malloc(strlen(string) + 1);
	memcpy(stLogData[index]->string, string, strlen(string) + 1);

	/* OAIɈxNA(16ڈƂ) */
	if (nLogCnt >= 0x60000000) {
		int i;
		int index;

		index = nLogTop;
		for (i=0; i<nLogNum; i++) {
			/* f[^ */
			if (stLogData[index]) {
				if (stLogData[index]->string) {
					/* |C^ */
					free(stLogData[index]->string);
				}
				/* f[^{̂ */
				free(stLogData[index]);

				/* NULL */
				stLogData[index] = NULL;
			}

			/* CfbNXړ */
			index++;
			index &= (LogMax - 1);
		}

		/* Ǘ[NNA */
		nLogTop = 0;
		nLogNum = 0;
		nLogCnt = 1;
	}

	nLogDsp = index - nLogTop - lLogHeight / lCharWidth;
	if (nLogDsp < 0) {
		nLogDsp = 0;
	}

	/* AbN */
	UnlockVM();
}

/*
 *	OEChE
 *	TCY
 */
static void FASTCALL ResizeLog(HWND hWnd, WORD cx, WORD cy)
{
	ASSERT(hWnd);

	/* ŏ̏ꍇ́AȂ */
	if ((cx == 0) && (cy == 0)) {
		return;
	}

	lLogWidth = cx;							/* O EBhETCY */
	lLogHeight = cy;						/* O EBhEcTCY */
	MoveWindow(hWnd, 0, 0, cx, cy, /*TRUE*/FALSE);
}

/*
 *	OEChE
 *	tbV
 */
void FASTCALL RefreshLog(void)
{
	static int nNum = -1, nTop = -1;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_LOG] == NULL) {
		return;
	}

	/* Ȏ݃`FbN */
	if (nLogNum == 0) {
		return;
	}

	/* O̒ǉ`FbN */
	if ((nNum == nLogNum) && (nTop == nLogTop)) {
		return;
	}
	nNum = nLogNum;
	nTop = nLogTop;

	/* XV */
	ListView_EnsureVisible(hLogWnd, nLogNum - 1, TRUE);
	InvalidateRect(hLogWnd, NULL, FALSE);
}

/*
 *	OEChE
 *	XgRg[쐬
 */
static HWND FASTCALL CreateLogList(HWND hWnd, HINSTANCE hInst)
{
	HDC hDC;
	TEXTMETRIC tm;
	int i;
	char string[128];
	LVCOLUMN lvc;

	hLogWnd = CreateWindowEx(	0,
								WC_LISTVIEW,
								"",
								WS_CHILD | WS_VISIBLE | LVS_REPORT |
								LVS_OWNERDATA,
								0,
								0,
								0,
								0,
								hWnd,
								(HMENU)IDC_LOGLIST,
								hInst,
								NULL);
	ListView_SetItemCountEx(hLogWnd, LogMax, LVSICF_NOINVALIDATEALL);

	/* eLXggbN擾 */
	hDC = GetDC(hLogWnd);
	GetTextMetrics(hDC, &tm);
	ReleaseDC(hLogWnd, hDC);

	/* }[v */
	for (i=0; i<countof(headertable); i++) {
		/* eLXg[h */
		LoadString(hResInstance, headertable[i].id, string, sizeof(string));

		/* J\̂쐬 */
		lvc.mask = LVCF_FMT | LVCF_SUBITEM | LVCF_WIDTH | LVCF_TEXT;
		lvc.fmt = headertable[i].fmt;
		lvc.cx = headertable[i].cx * tm.tmAveCharWidth;
		lvc.pszText = string;
		lvc.cchTextMax = strlen(string);
		lvc.iSubItem = i;

		/* J} */
		ListView_InsertColumn(hLogWnd, i, &lvc);
	}

	return	hLogWnd;
}

/*
 *	OEChE
 *	\
 */
static BOOL FASTCALL LogListGetDispInfo(LVITEM *pitem)
{
	int index;
	logdata_t *pstLogData;
	char string[256+1];
	char buffer[256+1];
	struct tm *t_st;
	DWORD dwTime;

	if (!(pitem->mask & LVIF_TEXT)) {
		return	FALSE;
	}

	if (nLogNum == 0) {
		return	FALSE;
	}
	index = pitem->iItem;
	if (nLogNum == LogMax) {
		index += nLogTop;
		index &= (LogMax - 1);
	}
	pstLogData = stLogData[index];
	if (pstLogData == NULL) {
		return	FALSE;
	}

	switch(pitem->iSubItem) {
	case 0:
		/* No. */
		memset(string, 0, sizeof(string));
		_snprintf(string, sizeof(string), "%d", pstLogData->number);
		memset(pitem->pszText, 0, pitem->cchTextMax);
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 1:
		/*  */
		t_st = localtime(&(pstLogData->timer));
		memset(string, 0, sizeof(string));
		_snprintf(string, sizeof(string), "%04d/", t_st->tm_year + 1900);
		_snprintf(buffer, sizeof(buffer), "%02d/", t_st->tm_mon + 1);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		_snprintf(buffer, sizeof(buffer), "%02d ", t_st->tm_mday);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		//t_st->tm_wday
		_snprintf(buffer, sizeof(buffer), "%2d:", t_st->tm_hour);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		_snprintf(buffer, sizeof(buffer), "%02d:", t_st->tm_min);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		_snprintf(buffer, sizeof(buffer), "%02d", t_st->tm_sec);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		memset(pitem->pszText, 0, pitem->cchTextMax);
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 2:
		/* z (s.ms.us) */
		memset(string, 0, sizeof(string));
		dwTime = (DWORD)pstLogData->time;
		_snprintf(string, sizeof(string), "%d.", dwTime / (1000 * 1000));
		dwTime %= (1000 * 1000);
		_snprintf(buffer, sizeof(buffer), "%03d.",dwTime / 1000);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		dwTime %= 1000;
		_snprintf(buffer, sizeof(buffer), "%03d",dwTime);
		strncat(string, buffer, sizeof(string)-strlen(string)-1);
		memset(pitem->pszText, 0, pitem->cchTextMax);
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 3:
		/* PC */
		memset(pitem->pszText, 0, pitem->cchTextMax);
#if XM7_VER == 1 && defined(Z80CARD) && defined(JSUB)
		_snprintf(string, sizeof(string), "$%04X/$%04X/$%04X/$%04X",
		pstLogData->pc1, pstLogData->pc2, pstLogData->pc3, pstLogData->pc4);
#elif defined(JSUB)
		_snprintf(string, sizeof(string), "$%04X/$%04X/$%04X",
		pstLogData->pc1, pstLogData->pc2, pstLogData->pc3);
#else
		_snprintf(string, sizeof(string), "$%04X/$%04X",
		pstLogData->pc1, pstLogData->pc2);
#endif
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 4:
		/* x */
		memset(pitem->pszText, 0, pitem->cchTextMax);
		switch (pstLogData->level) {
		/* ڍ */
		case Detail:
			LoadString(hResInstance, IDS_SWND_LOG_DETAIL, string, sizeof(string));
			break;
		/* ʏ */
		case Normal:
			LoadString(hResInstance, IDS_SWND_LOG_NORMAL, string, sizeof(string));
			break;
		/* x */
		case Warning:
			LoadString(hResInstance, IDS_SWND_LOG_WARNING, string, sizeof(string));
			break;
		/* fobO */
		case Debug:
			LoadString(hResInstance, IDS_SWND_LOG_DEBUG, string, sizeof(string));
			break;
		/* ̑(肦Ȃ) */
		default:
			ASSERT(FALSE);
			memset(string, 0, sizeof(string));
			break;
		}
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 5:
		/* foCX */
		memset(pitem->pszText, 0, pitem->cchTextMax);
		_snprintf(string, sizeof(string), "%c%c%c%c",
			(BYTE)(pstLogData->id >> 24),
			(BYTE)(pstLogData->id >> 16),
			(BYTE)(pstLogData->id >> 8),
			(BYTE)(pstLogData->id));
		strncpy(pitem->pszText, string, pitem->cchTextMax - 1);
		break;
	case 6:
		/* bZ[W */
		memset(pitem->pszText, 0, pitem->cchTextMax);
		strncpy(pitem->pszText, pstLogData->string, pitem->cchTextMax - 1);
		break;
	}
	return	TRUE;
}

/*
 *	OEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK LogProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	static BOOL bResize = FALSE;	/* 蓮TCYtO */
	LPNMHDR lpNmhdr;
	NMLVDISPINFO *lpDispInfo;

	/* bZ[W */
	switch (message) {
		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, NULL, NULL);

			UnlockVM();
			break;

		/* EChE쐬 */
		case WM_CREATE:
			CreateLogList(hWnd, ((LPCREATESTRUCT)lParam)->hInstance);
			return	0;

		/* EChE */
		case WM_CLOSE:
			/* OEBhE폜 */
			DestroyWindow(hLogWnd);
			hLogWnd = NULL;
			break;

		/* EChETCYύX */
		case WM_SIZE:
			if (bResize) {
				bResize = FALSE;
				ResizeLog(hLogWnd, LOWORD(lParam), HIWORD(lParam));
			}
			break;

		/* EChETCYύX */
		case WM_SIZING:
			bResize = TRUE;
			break;

		/* EBhETCYύXbZ[W */
		case WM_WINDOWPOSCHANGING:
			bResize = TRUE;
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;

		/* ʒm */
		case WM_NOTIFY:
			if (wParam == (WPARAM)IDC_LOGLIST) {
				lpNmhdr = (LPNMHDR)lParam;
				if (lpNmhdr->code == LVN_GETDISPINFO) {
					lpDispInfo = (NMLVDISPINFO*)lParam;
					if (LogListGetDispInfo(&lpDispInfo->item)) {
						return 0;
					}
				}
			}
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	OEChE
 *	EChE쐬
 */
HWND FASTCALL CreateLog(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Log";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lLogWidth;
	rect.bottom = lLogHeight;

	/* EChE^CgAobt@m */
	LoadString(hResInstance, IDS_SWND_LOG, szWndName, sizeof(szWndName));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = LogProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER
					| WS_SIZEBOX;
	}
	else {
		dwStyle = 	WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS
					| WS_SIZEBOX;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*
 *	OEChE
 *	tH[}bg
 */
void AddLogPrintf(DWORD id, loglevel level, const char *format, ...)
{
	char buffer[0x200];
	va_list args;
	va_start(args, format);

	/* tH[}bg */
	vsprintf(buffer, format, args);

	/* ϒI */
	va_end(args);

	/* bZ[Wǉ */
	LogAddString(id, level, buffer);
}

#endif	/* _WIN32 */
